/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
 * iptree.c
 *
 * This is a data structure for providing marked sets of ip addresses
 * in a tree format.
 *
 * This revision of the iptree module now moves it off into a cleaner
 * separate space and has replaced the read and write statements with
 * atomic operations.
 */

#include "silk.h"

RCSIDENT("$SiLK: /src/ds/iptree/iptree.c,v 1.14 2005/10/10 17:08:01 thomasm Exp $");

#include "iptree.h"


/* LOCAL FUNCTION */

static int _skTreeLoad(
    FILE           *dataFile,
    int             swapFlag,
    uint64_t        toRead,
    skIPTree_t    **dataTree);

static int _skTreeSaveWrite(
    FILE               *dataFile,
    const skIPTree_t   *tgtTree,
    const char         *fileName);

static void skIPTreePrintCIDR(
    FILE               *outFile,
    const skIPTree_t   *tree,
    int                 ipFormat);

static void handleCidrRange(
    FILE       *outFile,
    uint32_t    low,
    uint32_t    high,
    int         ipFormat);


/*
 *  Allocation and creation function; initializes a new tree at the
 *  space specified by **newTree and sets the contents to empty.  See
 *  header for details.
 */
int skIPTreeCreate(skIPTree_t **newTree)
{
    assert(newTree);
    *newTree = calloc(1, sizeof(skIPTree_t));
    if (*newTree == NULL) {
        return SKIP_ERR_TREEMEM;
    }
    return SKIP_OK;
}


/*
 *  Frees space associated with *oldTree.
 */
void skIPTreeDelete(skIPTree_t **oldTree)
{
    int i;

    if (*oldTree != NULL) {
        for (i = 0; i < SKIP_BBLOCK_COUNT; i++) {
            if ((*oldTree)->nodes[i] != NULL) {
                free((*oldTree)->nodes[i]);
                (*oldTree)->nodes[i] = NULL;
            }
        }
        free(*oldTree);
        *oldTree = NULL;
    };
}


/*
 *  Write the IP Tree to disk.  See the header for details.
 */
int skIPTreeSave(const char *fileName, const skIPTree_t *tree)
{
    FILE *dataFile;
    int errCode;

    if (fileExists(fileName)) {
        skAppPrintErr(("Output IP Tree file '%s' exists.\n"
                       "\t Delete or move it before trying to write"),
                      fileName);
        return SKIP_ERR_FILEFIND;
    }

    dataFile = fopen(fileName, "wb");
    if (!dataFile) {
        skAppPrintErr("Error opening IP Tree output file '%s': %s",
                      fileName, strerror(errno));
        return SKIP_ERR_FILEIO;
    }

    errCode = _skTreeSaveWrite(dataFile, tree, fileName);
    if (errCode) {
        fclose(dataFile);
        return errCode;
    }

    if (fclose(dataFile) == EOF) {
        goto skIPTreeSaveWriteError;
    }

    return SKIP_OK;

    /*
     * error handling is down here.
     * wmtWriteError: Any error that's going to reqiure a file close.
     */
  skIPTreeSaveWriteError:
    skAppPrintErr("Error writing set to file '%s'", fileName);
    fclose(dataFile);
    return SKIP_ERR_FILEIO;
}


/*
 *  Write IP Tree to 'dataStream'.  See header for details.
 */
int skIPTreeWrite(FILE *dataStream, const skIPTree_t *tree)
{
    return _skTreeSaveWrite(dataStream, tree, NULL);
}


/*
 *  Support function for skIPTreeSave() and skIPTreeWrite().  Writes
 *  the tree in 'dataTree' to 'dataStream'.  'fileName' is the source
 *  of the dataStream and is used only for error messages.
 */
static int _skTreeSaveWrite(
    FILE               *dataStream,
    const skIPTree_t   *tree,
    const char         *fileName)
{
    uint32_t i;
    uint32_t j;
    uint32_t jOffset;
    uint32_t addrSet;
    skIPTreeHeader_t tgtHeader;
    skIPNode_t *workNode;

    /*
     * Prep and write the IPTree's header information.
     */
    PREPHEADER(&(tgtHeader.gHdr));
    tgtHeader.gHdr.type = FT_IPTREE;
    tgtHeader.gHdr.version = 1;
    if (1 != fwrite(&tgtHeader, sizeof(skIPTreeHeader_t), 1, dataStream)) {
        goto skIPTreeSaveWriteError;
    }

    workNode = NULL;
    /*
     * Blocks are written in the form
     * <a><b><c>0<block>
     * where a, b, and c are the first 3 octets.
     */
    for (i = 0 ; i < SKIP_BBLOCK_COUNT; i++) {
        workNode = tree->nodes[i];
        if (workNode != NULL) {
            /*
             * non null node we now walk through the node to see what
             * is here, and write any non-null children to disk.
             */
            j = 0;
            do {
                if (workNode->addressBlock[j]) {
                    /*
                     * I've neglected #define abstractions here
                     * because the values are so eccentric I think
                     * it's better to just keep them visible.
                     *
                     * Step 1: isolate out the address position as a byte.
                     */
                    addrSet = ((i << 16) | ((j >> 3) << 8)) & 0xFFFFFF00;
                    if (1 != fwrite(&addrSet, sizeof(uint32_t), 1, dataStream)){
                        goto skIPTreeSaveWriteError;
                    }
                    jOffset = (j & 0x7F8);
                    if (8 != fwrite((void *) (workNode->addressBlock + jOffset),
                                    sizeof(uint32_t), 8, dataStream))
                    {
                        goto skIPTreeSaveWriteError;
                    }
                    j = ((j & 0x7F8) + 8);
                } else {
                    j = j + 1;
                }
            } while (j < SKIP_BBLOCK_SIZE);
        }
    }

    if (fflush(dataStream) == EOF) {
        goto skIPTreeSaveWriteError;
    }

    return 0;

    /*
     * error handling is down here.
     * wmtWriteError: Any error that's going to reqiure a file close.
     */
  skIPTreeSaveWriteError:
    if (fileName) {
        skAppPrintErr("Error writing IP Tree to file '%s'", fileName);
    } else {
        skAppPrintErr("Error writing IP Tree to stream");
    }
    return SKIP_ERR_FILEIO;
}



/*
 *  Reads IP Tree from fileName.  Basically a wrapper around
 *  skIPTreeRead().  See header for details.
 */
int skIPTreeLoad(const char *fileName, skIPTree_t **newTree)
{
    FILE *dataStream = NULL;
    int errCode;

    assert(newTree);

    dataStream = fopen(fileName, "rb");
    if (dataStream == NULL) {
        /* cache error; check if file exists */
        int my_error = errno;
        if (!fileExists(fileName)) {
            skAppPrintErr("Cannot load IP Tree from non-existent file '%s'",
                          fileName);
            return SKIP_ERR_FILEFIND;
        }
        skAppPrintErr("Cannot open IP Tree file '%s': %s",
                      fileName, strerror(my_error));
        return SKIP_ERR_FILEIO;
    }

    errCode = skIPTreeRead(dataStream, newTree);
    if (dataStream != NULL) {
        fclose(dataStream);
    }

    return errCode;
}


/*
 *  Read 'newTree' from the data stream 'dataStream'.  See header for details.
 */
int skIPTreeRead(FILE *dataStream, skIPTree_t **newTree)
{
    skIPTreeHeader_t th;
    int  swapFlag, errCode;

    assert(newTree);

    if ((*newTree) != NULL) {
        skAppPrintErr("newTree is already initialized, can't use.");
        return SKIP_ERR_UNKNOWN;
    }

    if (dataStream == NULL) {
        skAppPrintErr("Stream is NULL, not a valid input stream.");
        return SKIP_ERR_UNKNOWN;
    }

    /* Read header; determine if we need to set swapFlag */
    if (!fread((void *) &th, sizeof(skIPTreeHeader_t), 1, dataStream)) {
        skAppPrintErr("cannot read RW header from file");
        errCode = SKIP_ERR_FILEIO;
        goto skIPTreeLoadError;
    }
    if (CHECKMAGIC(&(th.gHdr))) {
        skAppPrintErr("invalid header; file is not an RW file");
        errCode = SKIP_ERR_FILEIO;
        goto skIPTreeLoadError;
    }
    if (FT_IPTREE != th.gHdr.type) {
        skAppPrintErr("file is not an IP Tree file");
        errCode = SKIP_ERR_FILEIO;
        goto skIPTreeLoadError;
    }
    /* there really should be a version check here as well, but we've
     * been extremely sloppy about setting version numbers on the set
     * files; to add it now would break a lot of existing sets.
     */
    /*
     *  if (th.gHdr.version > 1) {
     *      skAppPrintErr("Cannot process version $d IP Tree files.",
     *                    th.gHdr.version);
     *      errCode = SKIP_ERR_FILEIO;
     *      goto skIPTreeLoadError;
     *  }
     */

#if IS_LITTLE_ENDIAN
    swapFlag = th.gHdr.isBigEndian;
#else
    swapFlag = !th.gHdr.isBigEndian;
#endif

    errCode = skIPTreeCreate(newTree);
    if (errCode != SKIP_OK) {
        goto skIPTreeLoadError;
    }

    errCode = _skTreeLoad(dataStream, swapFlag, 0, newTree);
    if (errCode != SKIP_OK) {
        goto skIPTreeLoadError;
    }

    return SKIP_OK;

  skIPTreeLoadError:
    /* Do cleanup (Delete tree) and return */
    if ((*newTree) != NULL) {
        skIPTreeDelete(newTree);
    }

    return errCode;
}



/*
 *  Print a textual prepresenation of the IP Tree.  See header for
 *  details.
 */
void skIPTreePrint(FILE *outFile, const skIPTree_t *tree, int ipFormat)
{
    int i, j;

    if (ipFormat & SKIP_IPF_CIDR) {
        skIPTreePrintCIDR(outFile, tree, (ipFormat & ~SKIP_IPF_CIDR));
        return;
    }

    for (i = 0; i < SKIP_BBLOCK_COUNT; i++) {
        if (tree->nodes[i] != NULL) {
            for (j = 0; j < SKIP_BBLOCK_COUNT; j++) {
                if (skIPTreeHasMark(j, tree->nodes[i]->addressBlock)) {
                    skIPTreePrintAddr(outFile, ((i << 16) | j), ipFormat);
                };
            }
        }
    }
}


/*
 *  Like skIPTreePrint(), but prints IP ranges in CIDR notation.
 *  Loops over all IPs in the tree to find ranges of consecutive IPs.
 *  When it finds a gap between two IP addresses, it calls
 *  handleCidrRange() to print the existing CIDR range, then begins a
 *  new range.
 */
static void skIPTreePrintCIDR(
    FILE               *outFile,
    const skIPTree_t   *tree,
    int                 ipFormat)
{
    int i, j;
    int initial_ip = 1;
    uint32_t addr;
    uint32_t low = 0;
    uint32_t high = 0;

    for (i = 0; i < SKIP_BBLOCK_COUNT; i++) {
        if (tree->nodes[i] != NULL) {
            for (j = 0; j < SKIP_BBLOCK_COUNT; j++) {
                if (skIPTreeHasMark(j, tree->nodes[i]->addressBlock)) {
                    addr = ((i << 16) | j);
                    if (initial_ip) {
                        /* first IP begins a range */
                        low = high = addr;
                        initial_ip = 0;
                    } else if (addr == (high + 1)) {
                        /* add this addr to the current range */
                        high = addr;
                    } else {
                        /* print current range of IP addresses */
                        handleCidrRange(outFile, low, high, ipFormat);
                        /* start a new range */
                        low = high = addr;
                    }
                }
            }
        }
    }

    if (!initial_ip) {
        /* handle final range if we saw any IP addresses */
        handleCidrRange(outFile, low, high, ipFormat);
    }
}


/*
 *  Print the IPs from 'low' to 'high' inclusive in as few CIDR ranges
 *  as possible.
 */
static void handleCidrRange(
    FILE       *outFile,
    uint32_t    low,
    uint32_t    high,
    int         ipFormat)
{
    /* macro to print an IP */
#define PRINT_IP(outf, fmt, ip)                                 \
    switch (fmt) {                                              \
      case SKIP_IPF_HEX:                                        \
        fprintf((outf), "0x%08lx", (unsigned long)(ip));        \
        break;                                                  \
      case SKIP_IPF_DEC:                                        \
        fprintf((outf), "%lu", (unsigned long)(ip));            \
        break;                                                  \
      case SKIP_IPF_ZERO:                                       \
        fprintf((outf), "%s", num2dot0(ip));                    \
        break;                                                  \
      case SKIP_IPF_DOT:                                        \
      default:                                                  \
        fprintf((outf), "%s", num2dot(ip));                     \
        break;                                                  \
    }

    int mask;
    uint32_t bitmask;
    uint32_t range_low, range_high;

    /* A CIDR block cannot start on odd value; if low is odd, print it
     * and increment to next value */
    if (low & 0x01) {
        PRINT_IP(outFile, ipFormat, low);
        fprintf(outFile, "\n");

        /* only one IP? */
        if (low == high) {
            return;
        }
        ++low;
    }

    while (low != high) {
        mask = 32;
        do {
            --mask;
            bitmask = (~((uint32_t)0) << (33 - mask));
            range_low = low & bitmask;
            range_high = low | ~bitmask;
        } while ((range_high <= high) && (range_low == low));
        PRINT_IP(outFile, ipFormat, low);
        fprintf(outFile, "/%d\n", mask);

        /* compute high value of this range */
        range_high = low | ((~((uint32_t)0)) >> mask);
        if (high == range_high) {
            return;
        }
        low = range_high + 1;
    }

    /* Print file (single) IP address */
    PRINT_IP(outFile, ipFormat, low);
    fprintf(outFile, "\n");
}


/*
 *  Return a count of the number of IPs in tree.  See header for details.
 */
uint64_t skIPTreeCountIPs(const skIPTree_t *tree)
{
    int i, j, k;
    uint64_t count = 0;

    for (i = 0; i < SKIP_BBLOCK_COUNT; i++) {
        if (tree->nodes[i] != NULL) {
            /* Break the abstraction to speed up the counting */
            for (j = 0; j < SKIP_BBLOCK_COUNT / 32; ++j) {
                if (tree->nodes[i]->addressBlock[j]) {
                    for (k = 0; k < 32; ++k) {
                        if (tree->nodes[i]->addressBlock[j] & (1 << k)) {
                            ++count;
                        }
                    }
                }
            }
        }
    }

    return count;
}


/*
 * NAME
 * int _skTreeLoad(FILE *dataStream, int swapFlag, skIPTree_t **dataTree)
 *
 * ABSTRACT
 *
 * This is a low-level loading routine for snarfing an iptree off
 * of disk from *dataStream.  This is effectively the disk->data structure
 * translator, since we use the same forma for ip and tag trees
 * (and will until there's a substantial performance bonus for doing the
 * tag tree indexing from square one).
 *
 * PARAMETERS
 *
 * FILE *dataStream: Pointer to the FILE structure containing the
 *                 tree.
 *                 PR: Pointer is located at the beginning of the IP
 *                     tree data.
 *                 PO: Pointer is located at the end of the IP tree data.
 *
 * int swapFlag: Determined by parent loading routine, governs big or
 *               little endian loading.
 *
 * skIPTree_t **dataTree: Pointer to the IPtree structure.
 *                 PR: Empty & Null;
 *                 PC: Filled on success; empty & Null on error.
 *
 * RETURNS
 *
 * Status value: {SKIP_OK, SKIP_ERR_FILEIO, SKIP_ERR_TREEMEM, SKIP_ERR_UNKNOWN}
 *
 * SIDE EFFECTS
 *
 * Memory consumed for the construction of the ip tree.
 *
 * FAILS
 *
 * memory allocation failures will result in SKIP_ERR_TREEMEM, errors
 * reading will result in SKIP_ERR_FILEIO
 *
 * NOTES
 *
 * This is a purely in-module routine intended to provide a clean load
 * for both IPTree and TagTree loading.
 *
 */
static int _skTreeLoad(
    FILE *dataStream,
    int   swapFlag,
    uint64_t UNUSED(toRead),
    skIPTree_t **dataTree)
{
    uint32_t tBuffer[9];
    uint32_t blockStart, rootAddr;
    uint32_t total;
    int errCode, i;
    total = 0 ;

    while (fread(tBuffer, sizeof(uint32_t), 9, dataStream)) {
        if (swapFlag) {
            for (i = 0; i < 9; ++i) {
                tBuffer[i] = BSWAP32(tBuffer[i]);
            }
        }
        rootAddr = (tBuffer[0] & 0xFFFF0000) >> 16;
        blockStart = (tBuffer[0] >> 8) & 0xFF;
        if ((*dataTree)->nodes[rootAddr] == ((skIPNode_t *) NULL)) {
            (*dataTree)->nodes[rootAddr] = calloc(1, sizeof(skIPNode_t));
            if ((*dataTree)->nodes[rootAddr] == NULL) {
                errCode = SKIP_ERR_TREEMEM;
                goto _skTreeLoadError;
            }
        }
        total = 0;
        for (i = 0 ; i < 9; i++) {
            total += tBuffer[i];
        }
        if (total == 0) break;
        memcpy(((*dataTree)->nodes[rootAddr]->addressBlock +
                (blockStart * 8)), tBuffer + 1, 32);
    }

    return SKIP_OK;

    /*
     * Error handling and cleanup is managed down here
     */
  _skTreeLoadError:
    if (*dataTree != NULL) {
        skIPTreeDelete(dataTree);
    }
    return errCode;
}


/* Merge tree into result_tree */
int skIPTreeUnion(
    skIPTree_t         *result_tree,
    const skIPTree_t   *tree)
{
    int i, j;

    for (i = 0; i < SKIP_BBLOCK_COUNT; ++i) {
        if (NULL != tree->nodes[i]) {
            if (NULL != result_tree->nodes[i]) {
                /* need to merge */
                for (j = 0; j < SKIP_BBLOCK_SIZE; ++j) {
                    result_tree->nodes[i]->addressBlock[j] |=
                        tree->nodes[i]->addressBlock[j];
                }
            } else {
                /* copy block from tree to result_tree */
                result_tree->nodes[i] = malloc(sizeof(skIPNode_t));
                if (result_tree->nodes[i] == NULL) {
                    return 1;
                }
                memcpy(result_tree->nodes[i], tree->nodes[i],
                       sizeof(skIPNode_t));
            }
        }
    }

    return 0;
}


/* Turn off bits of 'result_tree' that are off in 'tree'. */
void skIPTreeIntersect(
    skIPTree_t         *result_tree,
    const skIPTree_t   *tree)
{
    int i, j;

    for (i = 0; i < SKIP_BBLOCK_COUNT; ++i) {
        if (result_tree->nodes[i] != NULL) {
            if (tree->nodes[i] == NULL) {
                /* This /16 is off in 'tree', turn off in 'result_tree' */
                free(result_tree->nodes[i]);
                result_tree->nodes[i] = NULL;
            } else  {
                /* Need to intersect the bits in the /16 */
                for (j = 0; j < SKIP_BBLOCK_SIZE; ++j) {
                    result_tree->nodes[i]->addressBlock[j] &=
                        tree->nodes[i]->addressBlock[j];
                }
            }
        }
        /* ELSE this /16 is off in the 'result_tree'.  Leave it alone. */
    }
}


void skIPTreeSubstract(
    skIPTree_t         *result_tree,
    const skIPTree_t   *tree)
{
    int i, j;

    for (i = 0; i < SKIP_BBLOCK_COUNT; ++i) {
        if ((result_tree->nodes[i] != NULL) && (tree->nodes[i] != NULL)) {
            /* Need to intersect with the complement in the /16 */
            for (j = 0; j < SKIP_BBLOCK_SIZE; ++j) {
                result_tree->nodes[i]->addressBlock[j] &=
                    ~(tree->nodes[i]->addressBlock[j]);
            }
        }
        /* ELSE This /16 is off in at least one tree.  Leave it alone. */
    }
}


/* ITERATOR CODE */


/* Create iterator */
int skIPTreeIteratorCreate(
    skIPTreeIterator_t **out_iter,
    const skIPTree_t    *tree)
{
    assert(out_iter);
    if (tree == NULL) {
        return -1;
    }

    *out_iter = malloc(sizeof(skIPTreeIterator_t));
    if (*out_iter == NULL) {
        return -1;
    }

    (*out_iter)->tree = tree;
    skIPTreeIteratorReset(*out_iter);

    return 0;
}


/* Destroy iterator */
void skIPTreeIteratorDestroy(skIPTreeIterator_t **out_iter)
{
    if (*out_iter) {
        (*out_iter)->tree = NULL;
        free(*out_iter);
        *out_iter = NULL;
    }
}


#define FIND_NEXT_ITER_TOP(iter)                                        \
    for ( ; (iter)->top_16 < SKIP_BBLOCK_COUNT; ++(iter)->top_16) {     \
        if ((iter)->tree->nodes[(iter)->top_16] != NULL) {              \
            /* found a /16 with data */                                 \
            break;                                                      \
        }                                                               \
    }

/* Get next entry in tree */
skIPTreeIteratorStatus_t skIPTreeIteratorNext(
    uint32_t           *out_addr,
    skIPTreeIterator_t *iter)
{
    assert(out_addr);

    while (iter->top_16 < SKIP_BBLOCK_COUNT) {
        for ( ; iter->bot_16 < SKIP_BBLOCK_COUNT; ++iter->bot_16) {
            if (skIPTreeHasMark(iter->bot_16,
                                iter->tree->nodes[iter->top_16]->addressBlock))
            {
                /* Generate the IP address */
                *out_addr = ((iter->top_16) << 16) | iter->bot_16;
                /* Prepare to search from next IP address */
                ++iter->bot_16;
                return SKIP_ITERATOR_OK;
            }
        }

        /* reset bottom counter; move top counter to next valid block */
        iter->bot_16 = 0;
        ++iter->top_16;
        FIND_NEXT_ITER_TOP(iter);
    }

    return SKIP_ITERATOR_NO_MORE_ENTRIES;
}


/* Reset iterator */
void skIPTreeIteratorReset(skIPTreeIterator_t *iter)
{
    /* Set bottom to 0; set top to first /16 with data */
    iter->top_16 = iter->bot_16 = 0;
    
    FIND_NEXT_ITER_TOP(iter);
}
