/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/
#ifndef _BAGTREE_H
#define _BAGTREE_H

#include "silk.h"

RCSIDENTVAR(rcsID_BAGTREE_H, "$SiLK: /src/include/bagtree.h,v 1.10 2005/05/20 17:54:10 thomasm Exp $");


/*
** bagtree.h
**
** Christopher Lee
** 2004-11-04
**
** Bagtree is a static (but variable) depth tree, where a key is
** encoded into the tree structure, and blocks of counters are stored
** in the leaf nodes.
**
** The structure currently uses a single 32b counter.  That can
** change, but the library will need a little revision.  I did a bit
** of forward thinking on making it work, but didn't have time to
** implement and test it all the way through.
**
** Similarly, it would be nice to have more-variable size keys, but
** that has not been worked out either.
*/

#include "utils.h"

/*
**
**  Exported API declarations
**
*/

typedef enum skBag_err_e {
    SKBAG_OK = 0,
    SKBAG_ERR_MEMORY = 1,
    SKBAG_ERR_KEY_NOT_FOUND = 2,
    SKBAG_ERR_INPUT = 3,
    SKBAG_ERR_OP_BOUNDS = 4
} skBag_err_t;

typedef uint8_t skBag_level_t; /* the number of a level */
typedef uint8_t skBag_levelsize_t; /* the number of bits encoded on a level */
typedef uint32_t skBag_blocksize_t; /* the number of entries in a node/leaf */
typedef uint32_t skBag_key_t; /* a key encoded in the tree */
typedef uint64_t skBag_counter_t; /* a counter */

typedef union skBag_node_u {
    union skBag_node_u *p_child;
    skBag_counter_t *p_leaf;
} skBag_node_t;

typedef struct _skBag_stats {
    uint32_t nodes; /* count of internal nodes allocated */
    uint64_t nodes_size; /* number of bytes allocated to nodes */
    uint32_t leaves; /* count of leaf blocks allocated */
    uint64_t leaves_size; /* number of bytes allocated to leaves */
    uint64_t keys_inserted; /* count of entries inserted in the tree */
    uint64_t unique_keys; /* count of entries inserted in the tree */
    skBag_counter_t max_counter; /* maximum counter value */
    skBag_key_t min_key; /* minimum key inserted */
    skBag_key_t max_key; /* maximum key inserted */
} skBag_stats_t;

typedef struct _skBag_header {
    skBag_node_t root; /* the root of the bagtree */
    skBag_level_t levels; /* number of levels in the tree */
    skBag_levelsize_t *rg_level_size; /* array of length "levels",
                                         each entry is the number of
                                         bits encoded in a level. */
    skBag_stats_t *p_stats;
} skBag_header_t;

typedef struct _skBag_iterator {
    skBag_header_t *p_hdr; /* pointer to the header to which this
                              iterator was created */
    skBag_blocksize_t *rg_offset_path; /* path of offsets through the
                                          tree, p_hdr->levels in
                                          size */
    int f_skip_counter; /* if true, iterator skips the first counter
                           in the leaf block, so that we don't re-find
                           the same node with subsequent calls to
                           getNext(). */
} skBag_iterator_t;


/*
**
**  Exported Bagtree API functions
**
*/


/*
** skBag_alloc
**
** allocate memory for a new bagtree, and set "p_p_hdr" to point to
** it.  the new bagtree should be "levels" levels deep (counting leaf
** nodes), and each level should encode "rg_level_size[level]" bits of
** data.
**
** for example, to create a 9/9/9/5 IP space tree:
**  BagTreeHeader *p_hdr;
**  uint8_t level_sizes[4] = { 9, 9, 9, 5 };
**  skBag_alloc(&p_hdr, 4, level_sizes);
**
** Input:
**
**   p_p_hdr - a pointer to the skBag_header_t pointer the caller wants
**   to fill
**
**   levels - the total number of levels of the bagtree, including
**   root node and leaf level
**
**   rg_level_size - an array of length levels, where each entry is
**   the number of key bits encoded at that level
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_alloc(
    skBag_header_t    **p_p_hdr,
    skBag_level_t       levels,
    skBag_levelsize_t  *rg_level_size);


/*
** skBag_free
**
**   free all memory associated with a bagtree
**
** Input:
**
**   p_hdr - a pointer to the skBag_header_t pointer the caller wants
**   to free
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_free(skBag_header_t *p_hdr);


/*
** skBag_addToCounter
**
**   given a pointer to a key, increment the counter associated with
**   that key
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_key - a pointer to the key to whose counter you want to add
**
**   p_counter - a pointer to the counter whose values you are adding
**
** Output:
**
**   a skBag_err_t error code
**
**   SKBAG_ERR_OP_BOUNDS if the addition would exceed the counter size
**   and cause it to loop.
*/
skBag_err_t skBag_addToCounter(
    skBag_header_t     *p_hdr,
    skBag_key_t        *p_key,
    skBag_counter_t    *p_counter);


/*
** skBag_getCounter
**
**   lookup tree at p_hdr, get counter associated with key p_key, and
**   put its value in p_counter.
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_key - the key whose counter is to be retrieved
**
**   p_counter - a pointer to the counter the consumer wants to fill
**
** Output:
**
**   a skBag_err_r error code
**
**   changes counter pointed to by p_counter
*/
skBag_err_t skBag_getCounter(
    skBag_header_t     *p_hdr,
    skBag_key_t        *p_key,
    skBag_counter_t    *p_p_counter);


/*
** skBag_incrCounter
**
**   given a pointer to a key, increment the counter associated with
**   that key
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_key - a pointer to the key whose counter is to be retrieved
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_incrCounter(
    skBag_header_t     *p_hdr,
    skBag_key_t        *p_key);


/*
** skBag_setCounter
**
**   given a pointer to a key, set the counter associated with that
**   key to a given value
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_key - a pointer to the key whose counter is to be retrieved
**
**   p_value - a pointer to the value to which to set the counter
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_setCounter(
    skBag_header_t     *p_hdr,
    skBag_key_t        *p_key,
    skBag_counter_t    *p_value);


/*
** skBag_subtractFromCounter
**
**   given a pointer to a key, increment the counter associated with
**   that key
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_key - a pointer to the key to from whose counter you want to subtract
**
**   p_counter - a pointer to the counter whose value you are subtracting
**
** Output:
**
**   a skBag_err_t error code
**
**   SKBAG_ERR_OP_BOUNDS if the subtraction would cause the counter to
**   become negative
*/
skBag_err_t skBag_subtractFromCounter(
    skBag_header_t     *p_hdr,
    skBag_key_t        *p_key,
    skBag_counter_t    *p_counter);





/*
**
** Exported Bagtree Iterator API functions
**
*/


/*
** skBag_allocIterator
**
**   create a new bagtree iterator, and initialize it so the first
**   call to skBag_getNext will return the counter associated with the
**   first key
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_iter - a pointer to the iterator pointer the consumer wants to
**   fill
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_allocIterator(
    skBag_header_t     *p_hdr,
    skBag_iterator_t  **p_iter);


/*
** skBag_freeIterator
**
**   deallocate all memory associated with a bagtree iterator
**
** Input:
**
**   p_iter - a pointer to the iterator to free
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_freeIterator(skBag_iterator_t *p_iter);


/*
** skBag_resetIterator
**
**   reset the iterator so the first call to skBag_getNext will return
**   the counter associated with the first key
**
** Input:
**
**   p_iter - a pointer to the iterator to reset
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_resetIterator(skBag_iterator_t *p_iter);


/*
** skBag_getNext
**
**   get the next key/counter pair associated with a given iterator
**
** Input:
**
**   p_iter - a pointer to the iterator
**
**   p_key - a pointer to the key to fill with the next key
**
**   p_counter - a pointer to the counter to fill with the next
**   counter
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_getNext(
    skBag_iterator_t   *p_iter,
    skBag_key_t        *p_key,
    skBag_counter_t    *p_counter);





/*
**
** Exported Bagtree I/O API functions
**
*/


/*
** skBag_printTreeStats
**
**   print out metadata on how the bagtree is performing
**
** Input:
**
**   p_hdr - a pointer to the bagtree
**
**   fp_out - an open output stream
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_printTreeStats(
    skBag_header_t     *p_hdr,
    FILE               *fp_out);


/*
** skBag_writeBinary
**
**   serialize bag to output stream
**
** Input:
**
**   p_hdr - a pointer to the bagtree
**
**   fp_out - an open output stream
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_writeBinary(
    skBag_header_t     *p_hdr,
    FILE               *fp_out);

/*
** skBag_readBinary
**
**   read serialized bag from input stream
**
** Input:
**
**   p_p_hdr - a pointer to the bagtree
**
**   fp_in - an open input stream
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_readBinary(
    skBag_header_t    **p_hdr,
    FILE               *fp_in);


/*
**
**  Internal declarations
**
*/

#define MIN_LEVELS 1
#define MAX_LEVELS 32
#define MIN_KEY_SIZE 1
#define MAX_KEY_SIZE 32

#endif /* _BAGTREE_H */
