/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/
#ifndef _DYNLIB_H
#define _DYNLIB_H

#include "silk.h"

RCSIDENTVAR(rcsID_DYNLIB_H, "$SiLK: /src/include/dynlib.h,v 1.17 2005/05/20 17:54:11 thomasm Exp $");

#include "utils.h"
#include "rwpack.h"


/*
 *  Each plug-in should define a dynlib_api_version() routine that
 *  returns this value.  This allows the application to skip an old
 *  plug-in without a SEGV.
 */
#define DYNLIB_API_VERSION  3


/*
 *  The plug-in's setup() method should return on of these values, as
 *  explained below.
 */
#define DYNLIB_FAILED	1
#define DYNLIB_WILLPROCESS 2
#define DYNLIB_WONTPROCESS 3


/*
 *  dynlib provides an interface to C-plug-ins---dynamic libraries
 *  loaded at run time.
 *
 *  TO USE A PLUG-IN:
 *
 *  The application that wants to use a plug-in does the following:
 *
 *  1. Calls dynlibCreate(appType) where 'appType' is one of the
 *  symbols listed in the 'dynlibSymbolId' enumeration and describes
 *  the type of the application; this function will return an
 *  interface pointer, 'dlISP'.
 *
 *  2. Calls dynlibLoad(dlISP, path) where 'path' is the path to the
 *  dynamic library.  If 'path' does not contain a slash, dynlibLoad()
 *  will first look for the plug-in
 *  $SILK_PATH/SILK_SUBDIR_PLUGINS/path.  If that fails (or if 'path'
 *  contains a slash), 'path' is handed to dlopen() unaltered.
 *  dlopen() may consult LD_LIBRARY_PATH (or similar).  If dlopen()
 *  fails, DYNLIB_FAILED is returned.
 *
 *  If dlopen() succeeds, dynlibLoad() will verify that the plug-in's
 *  version matches the DYNLIB_API_VERSION given above and that the
 *  plug-in has all the required subroutines.  If it does not,
 *  DYNLIB_FAILED is returned.  Finally, dynlibLoad() will call the
 *  plug-in's setup() method and the result of that method is returned
 *  to the caller.  The details of the setup() method are described
 *  below.
 *
 *  2a. Should the application desire to print any options associated
 *  with the plug-in, it may call dynlibOptionsUsage(dlISP) which will
 *  call the plug-in's optionsUage() method, if it exists.
 *
 *  3. Some plug-ins may require additional configuration prior to
 *  options handling.  The dynlibConfigure() method will call the
 *  plug-in's configure() method, if it exists.
 *
 *  4. Once the application has called the options parser, it can use
 *  dynlibCheckActive(dlISP) to see if the plug-in should be called as
 *  part of the application's processing--some applications only
 *  become active when the user specifies an option related to the
 *  plug-in.  Some applications may have to explicitly initialize and
 *  activate the plug-in by calling the dynlibInitialize() method.
 *
 *  5. If the plug-in is active, dynlibGetRWProcessor(dlISP) returns a
 *  pointer to the function that the application should call.  For a
 *  filtering application, dynlibGetRWProcessor() returns a pointer to
 *  the plug-in's filter() subroutine.
 *
 *  6. Once processing is completed, the application should call
 *  dynlibTeardown(dlISP) which will free the dlISP data structure and
 *  call teardown() on the plug-in.
 *
 *
 *
 *  TO BE A PLUG-IN:
 *
 *  The code that expects to by used by the dynlib interface must
 *  provide the following C subroutines:
 *
 *  - int setup(dynlibInfoStruct *dlISP, dynlibSymbolId appType);
 *
 *    This subroutine should set up the plug-in; if the plug-in has its
 *    own options, they can to be registered now.  setup() should
 *    return one of:
 *
 *      DYNLIB_FAILED - if processing fails
 *      DYNLIB_WONTPROCESS - if application should do normal output
 *      DYNLIB_WILLPROCESS - if this plug-in takes over output
 *
 *  - void teardown(dynlibSymbolId appType);
 *
 *    Any cleanup the plug-in requires should happen in this
 *    subroutine.
 *
 *  In addition, the following subroutines MAY exist in the plug-in:
 *
 *  - int dynlib_api_version(void) { return DYNLIB_API_VERSION; }
 *
 *    This subroutine, if it exists, is called first to verify that
 *    the plug-in and application are using a common API.  If it does
 *    not exist, the application will assume the APIs are correct.
 *
 *  - int configure(dynlibInfoStruct *dlISP, void *data);
 *
 *    This subroutine is optional.  It provides a way to pass
 *    arbitrary data between the plug-in and the application.  It is
 *    often used as an alternate way to configure the options.  The
 *    exact purpose of this routine will depend on the application and
 *    the plug-in.
 *
 *  - int initialize(dynlibInfoStruct *dlISP, dynlibSymbolId appType);
 *
 *    This subroutine should do any "expensive" initialization
 *    required by the plug-in in preparation for calling one of the
 *    interface subroutines listed below.
 *
 *  - void optionsUsage(dynlibSymbolId appType, FILE *fh);
 *
 *    If the plug-in has registered options, this subroutine should
 *    print a them and a short usage message to the given file handle.
 *
 *  To be used by an application type, the plug-in must provide the
 *  necessary interface subroutine.  The interface routine depends on
 *  the 'appType' that is specified in the dynlibCreate() call that
 *  the application makes:
 *
 *  DYNLIB_EXCL_FILTER: used by rwfilter
 *    int check(rwRec *rwRec);
 *
 *    This subroutine takes over all determination as to whether the
 *    rwRec should pass or fail the filter.  The subroutine should
 *    return 0 if the rwRec passes the filter; 1 otherwise.
 *
 *  DYNLIB_SHAR_FILTER: used by rwfilter
 *    int filter(rwRec *rwRec);
 *
 *    This subroutine works in tadem with the standard filtering rules
 *    of rwfilter (and other shared filter plug-ins) to determine
 *    whether the rwRec should pass the filter.  This subroutine may
 *    not see all rwRec records since they may be filtered by one of
 *    the standard rules before reaching this plug-in.  The subroutine
 *    should return 0 if the rwRec passes the filter; 1 otherwise.
 *
 *    The dynlib plug-in code will call this filter() routine ONCE per
 *    rw-record.  If the user has specified multiple filter-rules
 *    switches which a single plug-in handles, the plug-in must process
 *    all the switches before returning.
 *
 *  DYNLIB_CUT: used by rwcut
 *    int cut(unsigned int field, char *out, size_t len_out, rwRec *rwrec)
 *
 *  Arguments:
 *      which field to output (indexed from 1)
 *      a string buffer to hold the output
 *      the length of the buffer
 *      a RW record
 *  Returns:
 *      if field is zero, the number of supported fields
 *      if out is NULL and rwrec is NULL, the size of buffer needed
 *         for a title for this field
 *      if out is NULL and rwrec is non-NULL, the size of buffer
 *         needed for the value to be output for this field
 *      if out is non-NULL and rwrec is NULL, outputs the title,
 *         returns length of buffer necessary to hold full string
 *      if out is non-NULL and rwrec is non-NULL, outputs the value,
 *         returns length of buffer necessary to hold full string
 *
 *
 *  DYNLIB_COUNT: used by rwcount
 *    int count(rwRec *rwRec)
 *
 *    TO BE DECIDED
 *
 *  DYNLIB_SET: used by rwset
 *    int set(rwRec *rwRec)
 *
 *    TO BE DECIDED
 *
 *  DYNLIB_TOTAL: used by rwtotal
 *    int total(rwRec *rwRec)
 *
 *    TO BE DECIDED
 *
 */

/*
 *  A list of application types.  One of these should be passed to
 *  dynlibCreate().
 */
typedef enum _dynlibSymbolId {
    DYNLIB_SETUP,		/* internal use -- must be first*/
    DYNLIB_TEARDOWN,		/* internal use */
    DYNLIB_VERSION,		/* internal use */
    DYNLIB_CONFIGURE,		/* internal use */
    DYNLIB_INITIALIZE,		/* internal use */
    DYNLIB_OPT_USAGE,		/* internal use */
    DYNLIB_EXCL_FILTER,		/* rwfilter: override other filter rules  */
    DYNLIB_SHAR_FILTER,		/* rwfilter: augemnt other filter rules */
    DYNLIB_CUT,			/* rwcut: output of data */
    DYNLIB_COUNT,		/* rwcount: to be decided */
    DYNLIB_SET,			/* rwset: to be decided */
    DYNLIB_TOTAL,		/* rwtotal: to be decided */
    DYNLIB_VOID			/* internal use -- must be last */
} dynlibSymbolId;


/*
 *  Pointer to function returning int---the actual argument list
 *  varies by function.  We store an array of these in the
 *  dynlibInfoStruct, and it is the return type of
 *  dynlibGetRWProcessor().
 */
typedef int (*dynlib_fn_t)();


/*
 *  The structure that holds information about the plug-in.
 */
struct _dynlibInfoStruct;
typedef struct _dynlibInfoStruct dynlibInfoStruct;
typedef dynlibInfoStruct * dynlibInfoStructPtr;


/*
 *  dynlibCreate:
 *	get space and initialize struct
 *  Inputs:
 * 	The type of application
 *  Outputs:
 *	dynlibInfoStruct *dlISP;
 */
dynlibInfoStruct *dynlibCreate(dynlibSymbolId appType);


/*
 *  dynlibLoad:
 *	check all stuff about potential dynamic lib: load the library
 *	and check that it has the necessary functions.  To find Call
 *	the dynamic library's setup() subroutine which may register
 *	options
 *  Input:
 *	dynlibInfoStruct *dlISP
 *	path to dyn lib
 *  Output:
 *	0 if ok; 1 else
 *  Side Effects:
 *	struct dlISP filled.
 */
int dynlibLoad(dynlibInfoStruct *dlISP, const char *dlPath);


/*
 *  dynlibConfigure
 *
 *      Do any additional setup/configuration by calling the plug-in's
 *      configure() method.  This function returns 0 if the plug-in
 *      has a configure() method, and 1 otherwise.  The return status
 *      of the confiugre() method is returned in *out_status.
 */
int dynlibConfigure(dynlibInfoStruct *dlISP, void *data);


/*
 *  dynlibInitialize:
 *      Call the dynamic library's initalize() subroutine.  The idea
 *      is to put expensive initialization code in the initialize()
 *      routine, and only call it once the application knows the
 *      dynamic library will be used.
 *
 *      Once the initialize() routine returns 0, this wrapper will not
 *      call it again; if you must call initialize() multiple times,
 *      use the dynlibClearInitialized() macro to clear the
 *      'initialized()-was-called' flag.
 *  Input:
 *	dynlibInfoStruct *dlISP
 *  Output:
 *      Returns 0 if
 *        -- initialized() does not exist, or
 *        -- initialized() returns 0, or
 *        -- initialized() previously returned 0
 *      Otherwise returns status of call to initialize()
 *  Side Effects:
 *      Depends on shared library
 */
int dynlibInitialize(dynlibInfoStruct *dlISP);


/*
 *  dynlibOptionsUsage:
 *	Calls the dynamic library's optionsUsage() subroutine
 *  Input:
 *	dynlibInfoStruct *dlISP
 *      FILE *fh
 *  Output:
 *	none
 *  Side Effects:
 *	Depends on shared library
 */
void dynlibOptionsUsage(dynlibInfoStruct *dlISP, FILE *fh);


/*
 *  dynlibGetRWProcessor:
 *	Returns a pointer to the function applicable to this app-type;
 *	e.g., for DYNLIB_CUT returns a pointer to cut().  The
 *	existence of the function was checked in dynlibLoad
 *  Input:
 *	dynlibInfoStruct *dlISP
 *  Output:
 *	function ptr
 *  Side Effects:
 *	None.
 */
dynlib_fn_t dynlibGetRWProcessor(dynlibInfoStruct *dlISP);


/*
 *  dynlibTeardown:
 *	close and free mem
 *  Inputs:
 * 	dynlibInfoStruct *dlISP
 *  Outputs:
 *	None
 */
void dynlibTeardown(dynlibInfoStruct *dlISP);


/*
 *  dynlibGetPath
 *	Return the path to the dynamic library
 *  Inputs:
 * 	dynlibInfoStruct *dlISP
 *  Outputs:
 *	A \0 terminated string.  String should be considered
 *	read-only.  Returns the path passed to dlopen to open
 *      the dynamic library.
 *  Side Effects:
 *	None.
 */
const char *dynlibGetPath(const dynlibInfoStruct *dlIPS);


/*
 *  dynlibCheckLoaded
 *	Returns status of loading and checking the plug-in
 *  Inputs:
 * 	dynlibInfoStruct *dlISP
 *  Outputs:
 *	Return 1 if the plug-in was successfully loaded and checked, 0
 *	otherwise.
 *  Side Effects:
 *	None.
 *  Notes:
 *	dlISP may be NULL
 */
int dynlibCheckLoaded(const dynlibInfoStruct *dlISP);


/*
 *  dynlibGetAppType
 *	Returns the application type as set in dynlibCreate()
 *  Inputs:
 * 	dynlibInfoStruct *dlISP
 *  Outputs:
 *	The dynlibSymbolId for this application
 *  Side Effects:
 *	None.
 */
dynlibSymbolId dynlibGetAppType(const dynlibInfoStruct *dlISP);


/*
 *  dynlibGetStatus
 *	Return the status of the call to dynlibLoad()
 *  Inputs:
 * 	dynlibInfoStruct *dlISP
 *  Outputs:
 *	An integer value of:
 *
 *      DYNLIB_FAILED - if processing fails
 *      DYNLIB_WONTPROCESS - if application should do normal output
 *      DYNLIB_WILLPROCESS - if this plug-in takes over output
 *  Side Effects:
 *	None.
 */
int dynlibGetStatus(const dynlibInfoStruct *dlISP);


/*
 *  dynlibGetAppContext
 *	Return the application context: a pointer to a struct
 *	containing the application's name and its options
 *
 *      Plug-ins should call skAppContextSet() with the value returned
 *      by this macro in their setup() routines.  This will ensure
 *      that the plug-in's options are in the application's main
 *      options structure.

 *  Inputs:
 * 	dynlibInfoStruct *dlISP
 *  Outputs:
 *	The global application context object, skAppContext_t*
 *  Side Effects:
 *	None.
 */
skAppContext_t *dynlibGetAppContext(const dynlibInfoStruct *dlISP);


/*
 *  dynlibCheckActive
 *	Return non-zero if this plug-in will be processing the
 *	rwRecPtr's and thus should be called.
 *  Inputs:
 * 	dynlibInfoStruct *dlISP
 *  Outputs:
 *	The state of the active flag
 *  Side Effects:
 *	None.
 *  Notes:
 *	dlISP may be NULL
 */
int dynlibCheckActive(const dynlibInfoStruct *dlISP);


/*
 *  dynlibMakeActive
 *	Mark the plug-in as being active: i.e., the plug-in will be
 *	processing the rwRecPtr's
 *  Inputs:
 * 	dynlibInfoStruct *dlISP
 *  Outputs:
 *	None.
 *  Side Effects:
 *	None.
 */
void dynlibMakeActive(dynlibInfoStruct *dlISP);


/*
 *  dynlibClearInitialized
 *	Clear the flag that indicates that the initialize() routine
 *	has been run.  See dynlibInitialize() for details.
 *  Inputs:
 * 	dynlibInfoStruct *dlISP
 *  Outputs:
 *	None.
 *  Side Effects:
 *	None.
 */
void dynlibClearInitialized(dynlibInfoStruct *dlISP);


/*
 *  dynlibSetDoNotClose
 *	Set the do-not-close flag to TRUE.  This will prevent
 *	dlclose() from being called on the plug-in.  We do this do
 *	work around an Oracle problem with unloading libraries.
 *  Inputs:
 * 	dynlibInfoStruct *dlISP
 *  Outputs:
 *	None.
 *  Side Effects:
 *	None.
 */
void dynlibSetDoNotClose(dynlibInfoStruct *dlISP);


/*
 *  dynlibLastError
 *	Return the most recent value from dlerror()
 *  Inputs:
 * 	dynlibInfoStruct *dlISP
 *  Outputs:
 *	A \0 terminated string.  String should be considered
 *	read-only.  The string only contains a valid error message
 *	immediately after an error occurred.  At other times, the
 *	string may be empty or contain a previous error.
 *  Side Effects:
 *	None.
 */
const char *dynlibLastError(const dynlibInfoStruct *dlISP);


/* These are the functions on the plug-in that this code may call */
int dynlib_api_version(/*void*/);
int setup(/*dynlibInfoStruct *dlISP, dynlibSymbolId appType*/);
int configure(/*dynlibInfoStruct *dlISP, void *data*/);
void teardown(/*dynlibSymbolId appType*/);
void optionsUsage(/*dynlibSymbolId appType, FILE *fh*/);
int initialize(/*dynlibInfoStruct *dlISP, dynlibSymbolId appType*/);
int filter(/*rwRec *rwrec*/);
int cut(/*unsigned int type, char *out, size_t len_out, rwRec *rwrec*/);


#endif /*  _DYNLIB_H */
