/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/
#ifndef _IPTREE_H
#define _IPTREE_H

#include "silk.h"

RCSIDENTVAR(rcsID_IPTREE_H, "$SiLK: /src/include/iptree.h,v 1.22 2005/10/26 14:09:45 thomasm Exp $");

/*
 * iptree.h
 *
 * Michael Collins
 * May 6th, 2003
 *
 * This is a tree structure for ip addresses.  It provides both
 * macro and microtree structures.
 *
 * This structure now provides support for three classes of trees:
 *
 * iptrees: trees which just contain a bitmap of ip addresses.
 * indexed trees: an indexed tree will provide a function which
 * given ip address can show the ip address' order in the set.
 *
 * tagged tree: a tagged tree contains an ancillary pointer to some
 * data array (which is going to be effectively rle encoded) and given an
 * ip address, can determine the associated tag for the ip.
 */

#include "silk.h"
#include "utils.h"


/*
 * Now we define our basic tree types.
 */
#define SKIP_BBLOCK_COUNT 65536
#define SKIP_BBLOCK_SIZE 2048

/*
 *  Error return values
 */
enum skIPTreeErrors {
    SKIP_OK = 0,
    SKIP_ERR_FILEFIND,
    SKIP_ERR_FILEIO,
    SKIP_ERR_TREEMEM,
    SKIP_ERR_UNKNOWN
};

enum skIPTreeTypes {
    SKIP_IPTREE_MACROTREE,
    SKIP_IPTREE_TAGTREE,
    SKIP_IPTREE_INDEXTREE
};


/*
 *  Formats for textual output.  See skIPTreePrintAddr() below.
 */
enum skIPTreeFormats {
    SKIP_IPF_DOT = 0,   /* Dotted quad:  10.0.0.1 */
    SKIP_IPF_DEC,       /* Base-10 integer:  167772161 */
    SKIP_IPF_HEX,       /* Base-16 integer:  0a000000 */
    SKIP_IPF_ZERO,      /* Zero-padded dotted-quad:  010.000.000.001 */

    /* Next value can be OR'ed with other values */
    SKIP_IPF_CIDR = 128
};


/*
 * IP-Tree
 *
 * An IP-Tree consists of a 64k-node root, followed by pointers to
 * IP-Nodes, each of which contains 64k spots for marking addresses.
 */
typedef struct _skIPNode_t skIPNode_t;

/* THE TREE */
typedef struct skIPTree_t {
    uint32_t    totalAddrs;
    skIPNode_t *nodes[SKIP_BBLOCK_COUNT];
} skIPTree_t;

/* A NODE */
struct _skIPNode_t {
    uint16_t    parentAddr;
    uint32_t    addressBlock[SKIP_BBLOCK_SIZE];
};

#define skIPTreeHasMark(space, dataBlock) \
    ((dataBlock)[(space) >> 5] & (1 << ((space) & 0x1F)))
/*
 *    INTERNAL.  Check for bit 'space' in node 'dataBlock'.  Space
 *    should be the lower 16 bits of an address.
 */

#define skIPTreeMarkSpot(addr, dataBlock) \
    ((dataBlock)[(((addr) & 0xFFFF) >> 5)] |= (1 << ((addr) & 0x1F)))
/*
 *    INTERNAL.  Set bit 'addr' in node 'dataBlock'
 */

#define skIPTreeCheckAddress(ipAddr, m)                                      \
    ((m)->nodes[((ipAddr) >> 16)]                                            \
     && skIPTreeHasMark(((ipAddr) & 0xFFFF),                                 \
                        (m)->nodes[((ipAddr) >> 16)]->addressBlock))
/*
 *    Return TRUE if ipAddr is set on tree 'm'; FALSE otherwise
 */


#define skIPTreeAddAddress(addr, dt)                                    \
    do {                                                                \
        if ( ! (dt)->nodes[(addr)>>16] ) {                              \
            (dt)->nodes[(addr)>>16] = calloc(1, sizeof(skIPNode_t));    \
            (dt)->nodes[(addr)>>16]->parentAddr = (addr) >> 16;         \
        }                                                               \
        skIPTreeMarkSpot((addr),(dt)->nodes[(addr)>>16]->addressBlock); \
    } while(0)
/*
 *    Add the IP Address 'addr' to the binary IPSet 'dt'.
 */



/*
 *    On disk format: the header is simply the usual SiLK generic
 *    header.
 */
typedef struct skIPTreeHeader_t{
    genericHeader gHdr;
} skIPTreeHeader_t;


/*
 *    An interator structure.
 */
typedef struct {
    const skIPTree_t   *tree;
    uint32_t            top_16;
    uint32_t            bot_16;
} skIPTreeIterator_t;


/*
 *    Status of an interator.
 */
typedef enum {
    /* More entries */
    SKIP_ITERATOR_OK=0,
    /* No more entries */
    SKIP_ITERATOR_NO_MORE_ENTRIES
} skIPTreeIteratorStatus_t;


/*
 * IP Address printing Macro
 */
#define skIPTreePrintAddrInt(outF, addr)        \
    (fprintf((outF), "%10u\n", (addr)))

#define skIPTreePrintAddrHex(outF, addr)        \
    (fprintf((outF), "%08x\n", (addr)))

#define skIPTreePrintAddrDot(outF, addr)        \
    (fprintf((outF), "%s\n", num2dot(addr)))

#define skIPTreePrintAddrZero(outF, addr)       \
    (fprintf((outF), "%s\n", num2dot0(addr)))

#define skIPTreePrintAddr(outF, addr, mode)     \
    switch (mode) {                             \
      case SKIP_IPF_DEC:                        \
        skIPTreePrintAddrInt((outF), (addr));   \
        break;                                  \
      case SKIP_IPF_HEX:                        \
        skIPTreePrintAddrHex((outF), (addr));   \
        break;                                  \
      case SKIP_IPF_ZERO:                       \
        skIPTreePrintAddrZero((outF), (addr));  \
        break;                                  \
      case SKIP_IPF_DOT:                        \
      default:                                  \
        skIPTreePrintAddrDot((outF), (addr));   \
        break;                                  \
    }


/* FUNCTION PROTOTYPES */

int skIPTreeCreate(skIPTree_t **newTree);
/*
 *    Allocation and creation function; initializes a new tree at the space
 *    specified by **newTree and sets the contents to empty.
 *
 *    IPTree **newTree - pointer to the tree to initialize
 *
 *    Returns SKIP_OK if everything went well, else SKIP_ERR_TREEMEM
 *    on a malloc failure.
 *
 *    skIPTreeDelete() is the corresponding free function.
 */


void skIPTreeDelete(skIPTree_t **oldTree);
/*
 *    Frees the space associated with *oldTree.
 *    On completion **oldTree == NULL
 */


int skIPTreeRead(FILE *dataStream, skIPTree_t **newTree);
/*
 *    Creates(allocates) a new IP Tree at the location pointed at by
 *    'newTree' and fills it with the IP Tree data that the function
 *    reads from the stream 'dataStream'.
 *
 *    Returns one of the following:
 *        SKIP_OK on success
 *        SKIP_ERR_FILEIO - error during a read call
 *        SKIP_ERR_TREEMEM - allocating memory.
 *        SKIP_ERR_UNKNOWN - newTree is nonNull.
 *
 *    Fails idempotently - if there's been a failure, newTree is set
 *    back to null and deleted.
 */


int skIPTreeLoad(const char *fileName, skIPTree_t **newTree);
/*
 *    Creates(allocates) a new IP Tree at the location pointed at by
 *    'newTree' and fills it with the IP Tree data that the function
 *    reads from the file 'fileName'.
 *
 *    This function is similar to skIPTreeRead(), except that this
 *    function reads from a file.
 *
 *    Returns one of the values listed for skIPTreeRead(), or
 *    one of these additional values:
 *
 *        SKIP_ERR_FILEFIND - error in opening a file that doesn't exist
 *
 *    Fails idempotently - if there's been a failure, newTree is set
 *    back to null and deleted, file handles are closed, etc.
 */


int skIPTreeWrite(FILE *dataStream, const skIPTree_t *tree);
/*
 *   Write the IP tree at 'tree' the output stream 'dataStream'.  The
 *   disk format is specified in iptree.api.
 *
 *   Returns one of the following values:
 *
 *        SKIP_OK on success.
 *        SKIP_ERR_FILEFIND if the file already exists.
 *        SKIP_ERR_FILEIO if there's an error writing the data to disk.
 *        SKIP_ERR_TREEMEM on a memory allocatin problem for the tree.
 *        SKIP_ERR_UNKNOWN for any other reason.
 */


int skIPTreeSave(const char *filename, const skIPTree_t *tree);
/*
 *    Writes the IP Tree at 'tree' to the disk file 'filename'; the
 *    disk format is specified in iptree.api.
 *
 *    This function is similar to skIPTreeWrite(), except this
 *    function writes to a disk file.
 *
 *    Returns one of the values listed in skIPTreeWrite(), or the
 *    following additional values:
 *
 *        SKIP_ERR_FILEFIND if the file already exists.
 */


void skIPTreePrint(FILE *outFile, const skIPTree_t *tree, int ipFormat);
/*
 *    Prints, to the stream 'outFile', a textual representation of the
 *    IP Tree given by 'tree'.  The parameter 'ipFormat' decribes how
 *    to print the tree, it can be one of the values described in the
 *    "enum skIPTreeFormats".  If 'ipFormat' has been OR'ed with the
 *    SKIP_IPF_CIDR value, the output will be in CIDR notation.
 */


uint64_t skIPTreeCountIPs(const skIPTree_t *tree);
/*
 *    Returns a count of the number of ip addresses marked in the tree;
 */


int skIPTreeUnion(
    skIPTree_t         *result_tree,
    const skIPTree_t   *tree);
/*
 *    Perform the union of 'result_tree' and 'tree', with the result
 *    in 'result_tree'; i.e., merge 'tree' into 'result_tree'.
 *    Returns 0 on success, or 1 on memory allocation error.
 */


void skIPTreeIntersect(
    skIPTree_t         *result_tree,
    const skIPTree_t   *tree);
/*
 *    Perform an intersection of 'result_tree' and 'tree', with the
 *    result in the 'result_tree'; i.e., turn off all addresses in
 *    'result_tree' that are off in 'tree'.
 */


void skIPTreeSubstract(
    skIPTree_t         *result_tree,
    const skIPTree_t   *tree);
/*
 *    Subtract 'tree' from 'result_tree' with the result in the
 *    'result_tree'; i.e., if an address is off in 'tree', do not
 *    modify the value in 'result_tree', otherwise, turn off that
 *    address in 'result_tree'.
 */


int skIPTreeIteratorCreate(
    skIPTreeIterator_t **out_iter,
    const skIPTree_t    *tree);
/*
 *    Creates a new iterator at the address pointed to by 'out_iter'
 *    that will iterate over all the entries in the IPSet 'tree'.
 *    Returns 0 on success, non-zero otherwise.
 */


void skIPTreeIteratorDestroy(skIPTreeIterator_t **out_iter);
/*
 *    Destroys the iterator pointed to by 'out_iter'.
 */


skIPTreeIteratorStatus_t skIPTreeIteratorNext(
    uint32_t           *out_addr,
    skIPTreeIterator_t *iter);
/*
 *    If there are more entries in the IPSet, this function puts the
 *    next IP Address that exists in the IPTree into the location
 *    referenced by 'out_addr' and returns SKIP_ITERATOR_OK.
 *    Otherwise, 'out_addr' is not touched and
 *    SKIP_ITERATOR_NO_MORE_ENTRIES is returned.
 */


void skIPTreeIteratorReset(skIPTreeIterator_t *iter);
/*
 *    Resets the iterator 'iter' to begin looping through the entries
 *    in the IPSet again.
 */


#endif /* _IPTREE_H */
