/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/
#ifndef _UTILS_H
#define _UTILS_H

#include "silk.h"

RCSIDENTVAR(rcsID_UTILS_H, "$SiLK: /src/include/utils.h,v 1.58 2005/11/15 17:50:28 thomasm Exp $");

/*
 *  a collection of utility routines.
 *
 *  Suresh L Konda
 */


#ifdef SK_NEED_GETOPT
#include "gnu_getopt.h"
#else
#include <getopt.h>
#endif

#define isRunningOnNode()  0


/*
**
**    utils.c
**
*/

typedef int (*sk_msg_fn_t)(const char *, ...);
/*
 *    The type of message functions.  These should use the same
 *    semantics as printf.
 */

typedef int (*sk_msg_vargs_fn_t)(const char *, va_list);
/*
 *    The type of message functions with the arguments expanded to a
 *    variable argument list.
 */

extern int skMsgNone(const char *, ...);
/*
 *    The null message function.  Does nothing and returns zero.
 */

extern int skMsgNoneV(const char *, va_list);
/*
 *    The null message function.  Does nothing and returns zero.
 */

typedef struct skAppContext skAppContext_t;
/*
 *    Struct that contains global information about the application:
 *    its name, full path, options.
 */

extern void skAppRegister(const char *name);
/*
 *    Register the application.  Other functions below assume you've
 *    passed argv[0] as the value of 'name'.
 */

extern const char *skAppRegisteredName(void);
/*
 *    Return the name that was used to register the application.  This
 *    is the value that was passed to skAppRegister().  The return
 *    value should be considered read-only.
 */

extern const char *skAppName(void);
/*
 *    Return a short name for the application.  This is the basename
 *    of the registered name.  The return value should be considered
 *    read-only.
 */

extern const char *skAppFullPathname(void);
/*
 *    Return the full path to application.  This will consult the PATH
 *    envar and cgetcwd() to find the complete path to the
 *    application.  The return value should be considered read-only.
 */

extern char *skAppDirParentDir(char *buf, size_t buf_len);
/*
 *    Return the application's directory's parent directory in 'buf',
 *    a character array of buf_len bytes. e.g., if the rwfilter
 *    application lives in "/usr/local/bin/rwfilter", this function
 *    puts "/usr/local" into buf.  Return value is a pointer to 'buf',
 *    or NULL on error.
 */

extern void skAppContextSet(skAppContext_t *a_context);
/*
 *    Use '*a_context' as the global application context.
 */

extern skAppContext_t *skAppContextGet(void);
/*
 *    Return the address of the application's global context.
 */

extern void skAppContextSetErrStream(FILE *f);
/*
 *    Sets the error stream to 'f' for any messages printed with
 *    skAppPrintErrV().  The function has no effect when 'f' is NULL;
 *    to disable all output from skAppPrintErr(), set the error
 *    function to &skMsgNoneV.
 */

extern void skAppContextSetErrFunction(sk_msg_vargs_fn_t fn);
/*
 *    skAppPrintErr() will call this funtion to print its arguments.
 *    By default, skAppPrintErr() uses skAppPrintErrV().
 */

extern int skAppPrintErrV(const char *fmt, va_list args);
/*
 *    The default function used by skAppPrintErr() to print error
 *    messages.
 */

#ifdef TEST_PRINTF_FORMATS
#define skAppPrintErr printf
#else
extern int skAppPrintErr(const char *fmt, ...);
#endif
/*
 *    Prints the application name, a colon, the result of formatting
 *    the arguments using v*printf(), and a newline to stderr.
 */


/*
**
**    options.c
**
*/

/* long option arg types */
#define NO_ARG          0
#define REQUIRED_ARG    1
#define OPTIONAL_ARG    2

/* macro to print whether switch requires an argument */
#define SK_OPTION_HAS_ARG(switch)             \
    (((switch).has_arg == REQUIRED_ARG)       \
     ? "Req Arg"                              \
     : (((switch).has_arg == OPTIONAL_ARG)    \
        ? "Opt Arg"                           \
        : (((switch).has_arg == NO_ARG)       \
           ? "No Arg"                         \
           : "BAD 'has_arg' VALUE")))

/* name of option for app to load a dynamlic library */
#define OPT_DYNAMIC_LIBRARY "dynamic-library"


/* generic types for dealing with opaque handles */
typedef void *clientData;
typedef int (* optHandler)(clientData cData, int optIndex, char *optArg);
typedef void (* usageFn)(void);


int optionsSetup(usageFn fn);
/* returns 0 (ok), 1 (not ok) */

void optionsTeardown(void);

int optionsRegister(void *inOptions, optHandler handler, clientData cData);

int optionsHandleConfFile(char *filename);
/*
 * optionsHandleConfFile:
 *
 *     Loads a configuration file.  The configuration file consists of
 *     a series of newline-terminated lines.  A line consisting of
 *     only whitespace, or whose first non-whitespace character is a
 *     `#' character is ignored.  All other lines should consist of a
 *     single option name followed by the option's value (if any),
 *     seperated by whitespace.  Whitespace at the beginning and end
 *     of the line is ignored.
 *
 * BUGS:
 *     If you intersperse switches (options) and arguments, arguments
 *     before the configuration file is parsed will not be seen.
 *
 *  Return:
 *      0 if ok. -1 else
 */

int optionsParse(int argc, char **argv);
/*
 *  optionsParse:
 *      Adjust the global options array to allow for the help
 *      option. If help is selected by the user, call the stashed
 *      usageFunction.  Parse input options given a set of
 *      pre-registered options and their handlers.  For each
 *      legitimate option, call the handler.
 *  SideEffects:
 *      The individual handlers update whatever datastruture they wish
 *      to via the clientData argument to the handler.
 *  Return:
 *      optind which points at the first non-option argument passed if
 *      all is OK.  If not OK, the return -1 for error.
 */




/*
**
**    times.c
**
*/

int32_t julianDate(int yr, int mo, int day);
/* jdate with 1999/12/1 == 0 */


void gregorianDate(int32_t jd, int32_t *yr, int32_t *mo, int32_t *day);
/* with jdate with 1999/12/1 == 0 */


#if SK_HAVE_TIMEGM
#  define SK_TIME_GM(tm) timegm(tm)
#else
#  define SK_TIME_GM(tm) sk_timegm(tm)
#endif
time_t sk_timegm(struct tm *tm);
/*
 *    Does the reverse of gmtime().  Takes a time structure and
 *    returns seconds since the epoch in the UTC timezone
 */


char *timestamp(uint32_t t);
/*
 *    Treats t as seconds since the UNIX epoch and returns a pointer
 *    to a static string buffer containing the UTC time in the form:
 *
 *      "mm/dd/yyyy HH:MM:SS"
 */


char *timestamp_r(uint32_t t, char *outbuf, size_t outbuf_size);
/*
 *    Thread-safe version of timestamp().
 */


int getJDFromDate(char *line, int *hour);
/*
 * getJDFromDate:
 *      Given a string of YYYY/MM/DD:HH:MM:SS
 *      return a julian date calculated from 0 for 12/1/1999.
 *      If the hour (etc) are given return the hour in the
 *      int pointer vaiable
 * Input:
 *      Date string
 * Outputs:
 *      JD as returned value.
 * Side Effect:
 *      the int pointer argument is filled with the hour if given
 *      or -1 otherwise.
 *      Exits on error.
 */


int maxDayInMonth(int yr, int mo);
/*
 *  maxDayInMonth:
 *      Use fixed array and julianDate and gregorianDate to compute
 *      the maximum day in a given month/year
 *      exit(1) on failure.
 *
 *  Months are in the 1..12 range and NOT 0..11
 *
 */


/*
**
**    silkfilesys.c
**
*/

char *baseName(const char *fp);
/*
 *    Strip directory prefix from the file path fp.  Returns a pointer
 *    to a static string buffer.
 */


char *baseName_r(char *dest, const char *src, size_t dest_size);
/*
 *    Thread safe version of baseName()
 */


char *dirName(const char *fp);
/*
 *    Strip file name suffix from the file path fp.  Returns a pointer
 *    to a static string buffer.
 */


char *dirName_r(char *dest, const char *src, size_t dest_size);
/*
 *    Thread safe version of dirName()
 */


#define FILEIsATty(fd)          isatty(fileno(fd))
/*
 *    Returns 1 if the FILE* fd is a tty, 0 otherwise
 */


int isFIFO(const char *name);
/*
 *    Returns 1 if name exists and is a FIFO; returns 0 otherwise.
 */


int dirExists(const char *dName);
/*
 *    Returns 1 if dName exists and is a directory; returns 0
 *    otherwise.
 */


int fileExists(const char *fName);
/*
 *    Returns 1 if fName exists and is a regular file; returns 0
 *    otherwise.
 */


off_t fileSize(const char *fName);
/*
 *    Returns the size of the file fName.  Returns 0 if file is empty
 *    or if it does not exist; use fileExists() to check for existence
 */


#define getRLockFD(fd)         silkFileLocks((fd), F_RDLCK, F_SETLK)
#define getRLockFILE(fp)       getRLockFD(fileno(fp))
/*
 *    int ret = getRLockFD(int fd);           // file descriptor
 *    int ret = getRLockFILE(FILE *fp);       // file pointer
 *
 *    Get an exclusive read lock on the opened file.  It does not
 *    wait.
 *
 *    Returns: 0 OK; 1 failed;
 *
 */


#define getRLockFD_Wait(fd)    silkFileLocks((fd), F_RDLCK, F_SETLKW)
#define getRLockFILE_Wait(fp)  getRLockFD_Wait(fileno(fp))
/*
 *    int ret = getRLockFD_Wait(int fd);      // file descriptor
 *    int ret = getRLockFILE_Wait(FILE *fp);  // file pointer
 *
 *    Get an exclusive read lock on the opened file.  It waits till it
 *    gets the lock indefinitely.
 *
 *    Returns: 0 OK; 1 failed;
 *
 */


#define getWLockFD(fd)         silkFileLocks((fd), F_WRLCK, F_SETLK)
#define getWLockFILE(fp)       getWLockFD(fileno(fp))
/*
 *    int ret = getWLockFD(int fd);           // file descriptor
 *    int ret = getWLockFILE(FILE *fp);       // file pointer
 *
 *    Get an exclusive write lock on the opened file.  It does not
 *    wait.
 *
 *    Returns: 0 OK; 1 failed;
 *
 */


#define getWLockFD_Wait(fd)    silkFileLocks((fd), F_WRLCK, F_SETLKW)
#define getWLockFILE_Wait(fp)  getWLockFD_Wait(fileno(fp))
/*
 *    int ret = getWLockFD_Wait(int fd);      // file descriptor
 *    int ret = getWLockFILE_Wait(FILE *fp);  // file pointer
 *
 *    Get an exclusive write lock on the opened file.  It waits till
 *    it gets the lock indefinitely.
 *
 *    Input: the file descriptor
 *
 *    Returns: 0 OK; 1 failed;
 *
 */


#define releaseLockFD(fd)      silkFileLocks((fd), F_UNLCK, F_SETLK)
#define releaseLockFILE(fp)    releaseLockFD(fileno(fp))
/*
 *    int ret = releaseLockFD(int fd);        // file descriptor
 *    int ret = releaseLockFILE(FILE *fp);    // file pointer
 *
 *    Release a previously locked file.
 *
 *    Returns: 0 OK; 1 failed;
 *
 */


int silkFileLocks(int fd, short type, int cmd);
/*
 *    Perform a locking operation on the specified file.  For use by
 *    the convenience macros above, which see.
 */


char *findFile(const char *fName);
/*
 *  findFile(fName);
 *
 *  Find the given file in one of several places:
 *
 *  -- See if the environment variable named by the cpp macro
 *     ENV_SILK_PATH (normally SILK_PATH) is defined.  If so, look in
 *     $SILK_PATH/share/fName
 *  -- For historical reasons, look in $SILK_PATH/fName
 *  -- Take the full path to the application (/yadda/yadda/bin/app),
 *     lop off the app's parent directory (/yadda/yadda), and check
 *     the "/share" subdir (/yadda/yadda/share/fName
 *
 *  Inputs:
 *      filename to find
 *
 *  Outputs:
 *      Return char * if found NULL else.  When non-null is returned,
 *      the caller should free() the char* when finished.
 */


char *skutilsFindPluginPath(const char *dlPath, char *path, size_t path_len);
/*
 *  skutilsFindPluginPath(dlPath, path, path_len);
 *
 *  Find the given plugin path, dlPath, in one of several places:
 *
 *  -- If dlPath contains a slash, return NULL.
 *  -- See if the environment variable named by the cpp macro
 *     ENV_SILK_PATH (normally SILK_PATH) is defined.  If so, look in
 *     $SILK_PATH/share/fName
 *  -- For historical reasons, look in $SILK_PATH/fName
 *  -- Take the full path to the application (/yadda/yadda/bin/app),
 *     lop off the app's parent directory (/yadda/yadda), and check
 *     the "/share" subdir (/yadda/yadda/share/fName
 *
 *  Inputs:
 *      - name or path of plug-in to find
 *      - buffer to hold the path
 *      - length of the buffer
 *
 *  Outputs:
 *      Return NULL if dlPath was not found; otherwise return a char*
 *      which is the buffer passed into the subroutine.
 */



int openFile(const char *FName, int mode, FILE **fp, int *isPipe);
/*
 *  status = openFile(file, mode, &fp, &isPipe);
 *
 *    Open 'file' as a pipe or as a regular file depending on whether
 *    it is a gzipped file or not.  A file is considered gzipped if
 *    its name contains the string ".gz\0" or ".gz.".
 *
 *    The name of the file is given in the C-string 'file'; 'mode'
 *    determines whether to open 'file' for reading (mode==0) or
 *    writing (mode==1).
 *
 *    The file pointer to the newly opened file is put into 'fp'.  The
 *    value pointed to by 'isPipe' is set to 0 if fopen() was used to
 *    open the file, or 1 if popen() was used.  The caller is
 *    responsible for calling fclose() or pclose() as appropriate.
 *
 *    The function returns 0 on success, or 1 on failure.
 */


int mkDirPath(const char *directory);
/*
 *  mkDirPath:
 *      make full directory path including parent if required
 *  Input: char * path
 *  Output: 0 on success, 1 on failure
 */


int copyFile(const char *srcPath, const char *destPath);
/*
 * copyFile:
 *      Copies the file "source" to "dest".  "Dest" may be a file or a
 *      directory.
 * Input: char * source
 *        char * dest
 * Output: 0 on success, errno on failure.
 */

int moveFile(const char *srcPath, const char *destPath);
/*
 * moveFile:
 *      Moves the file "source" to "dest".  "Dest" may be a file or a
 *      directory.
 * Input: char * source
 *        char * dest
 * Output: 0 on success, errno on failure.
 */

int skOpenPagerWhenStdoutTty(
    FILE  **output_stream,
    char  **pager);
/*
 *    Attempts to redirect the '*output_stream' to the paging program
 *    '*pager.'
 *
 *    If output changed so that it goes to a pager, 1 is returned; if
 *    the output is unchanged, 0 is returned.  If an error occurred in
 *    invoking the pager, -1 is returned.
 *
 *    If the '*output_stream' is NULL, it is assumed that the output
 *    was being sent to stdout.  If the '*output_stream' is not stdout
 *    or not a terminal, no paging is preformed and 0 is returned.
 *
 *    If '*pager' is NULL, the environment variable SILK_PAGER is
 *    checked for the paging program; if that is NULL, the PAGER
 *    environment variable is checked.  If that is also NULL, no
 *    paging is performed and 0 is returned.
 *
 *    If there was a problem invoking the pager, an error is printed
 *    and -1 is returned.
 *
 *    If the pager was started, the *output_stream value is set to the
 *    pager stream, and *pager is set to the name of the pager (if the
 *    user's environment was consulted), and 1 is returned.  To close
 *    the pager, use the skClosePager() function, or call pclose() on
 *    the *output_stream.
 *
 *    Due to the race condition of checking the status of the child
 *    process, it is possible for 1 to be returned even though the
 *    pager has exited.  In this case the tool's output will be lost.
 */


void skClosePager(FILE *output_stream, const char *pager);
/*
 *    If skOpenPagerWhenStdoutTty() returns a positive value, use this
 *    function to close the pager stream.  Prints an error if the
 *    close fails.
 */


/*
**
**    silkstring.c
**
*/


#define MALLOCCOPY(a,b) {if (!((a)=strdup(b))) {fprintf(stderr,"MALLOCCOPY: Out of memory at %s:%d\n", __FILE__, __LINE__);exit(1);}}
/*
 *    Duplicate the C string 'b' into 'a'.  Exits if not enough memory.
 */


char *num2dot(uint32_t ip);
/*
 *    Converts the integer form of an IPv4 IP address to the dotted-quad
 *    version.  ip is taken to be in native byte order; returns a
 *    pointer to a static string buffer.
 *
 *    Returns NULL on error.
 */


char *num2dot_r(uint32_t ip, char *outbuf, size_t outbuf_size);
/*
 *    Thread safe version of num2dot()
 */


char *num2dot0(uint32_t ip);
/*
 *    Like num2dot(), but will zero-pad the octects: num2dot0(0) will
 *    return "000.000.000.000"
 */


char *num2dot0_r(uint32_t ip, char *outbuf, size_t outbuf_size);
/*
 *    Thread safe version of num2dot0().
 */


uint32_t dot2num(const char *ip);
/*
 *    Converts the string buffer ip containing an IPv4 dotted-quad IP
 *    address into an integer in the native byte order.
 *
 *    Returns 0xFFFFFFFF on error.
 */


char *tcpflags_string(const uint8_t flags);
/*
 *    Return an 8 character string denoting which TCP flags are set.  If
 *    all flags are on, FSRPAUEC is returned.  For any flag that is off,
 *    a space appears in place of the character.  Returns a pointer to a
 *    static buffer.
 */


int strip(char *line);
/*
 *    Strips all whitespace from the start and end of line.  Modifies
 *    line in place.
 */


void lower(char *cp);
/*
 *    Converts uppercase letters in cp to lowercase.  Modifies the
 *    string in place.
 */


void upper(char *cp);
/*
 *    Converts lowercase letters in cp to uppercase.  Modifies the
 *    string in place.
 */


int skStringParseNumberList(
    uint32_t                  **number_list,
    uint32_t                   *number_count,
    const char                 *input,
    uint32_t                    min_val,
    uint32_t                    max_val,
    uint32_t                    max_number_count);
/*
 *    Given a C-string containing a list---i.e., comma or hyphen
 *    delimited set---of non-negative integers, e.g., "4,3,2-6",
 *    allocate and return, via the 'number_list' parameter, an array
 *    whose values are the numbers the list contains, breaking ranges
 *    into a list of numbers.  If duplicates appear in the input, they
 *    will appear in the return value.  Order is maintained.  Thus
 *    given the C-string 'input' of "4,3,2-6", the function will set
 *    *number_list to a newly allocated array containing
 *    {4,3,2,3,4,5,6}.  The number of entries in the array is returned
 *    in '*number_count'.  The list of number is limited by the
 *    'min_val' and 'max_val' parameters; a 'max_val' of 0 means no
 *    maximum.  The maximum size of the array to be returned is given
 *    by 'max_number_count'; when this value is 0 and max_value is not
 *    zero, it is set to the number of possible values (1+max-min);
 *    when max_number_count is 0 and max_value is 0, it is set to a
 *    large (2^24) number of entries.  In all cases, the function
 *    tries to keep the returned array as small as possible.  On
 *    success, 0 is returned.
 *
 *    INPUT:
 *      number_list -- the address in which to return the array
 *      number_count -- the address in which to store the
 *          number valid elements in the returned array.
 *      input -- the string buffer to be parsed
 *      min_value -- the minimum allowed value in user input
 *      max_value -- the maximum allowed value in user input.  When
 *          max_value is 0, there is no maximum.
 *      max_number_count -- the maximum number of entries the array
 *          returned in 'number_list' is allowed to have.
 *
 *    The caller should free() the returned array when finished
 *    processing.
 *
 *    On error, an error message is printed (via skAppPrintErr()), and
 *    -1 is returned.
 */


int skStringParseNumberListToBitmap(
    uint32_t                   *out_bitmap,
    const char                 *input,
    uint32_t                    bitmap_size);
/*
 *    Similar to skStringParseNumberList(), except that instead of
 *    returning an array, bits are set in the bitmap 'out_bitmap'.
 *    'out_bitmap' should have been declared with
 *
 *        BITMAP_DECLARE('out_bitmap', 'bitmap_size').
 *
 *    The input may have values from 0 to (bitmap_size-1).
 */


uint8_t *skParseNumberList(
    const char *input,
    uint8_t    minValue,
    uint8_t    maxValue,
    uint8_t   *returnValueCount);
/*
 * skParseNumberList
 *
 *    Given a list (i.e., comma or hyphen delimited set of
 *    non-negative integers), return an array whose values are the
 *    numbers the list contains, breaking ranges into a list of
 *    numbers.  If duplicates appear in the input, they will appear in
 *    the return value.  Order is maintained.  Thus the input
 *    "4,3,2-6" will return an array containing {4,3,2,3,4,5,6}.
 *
 *    INPUT:
 *      input -- the string buffer to be parsed
 *      minValue -- the minimum allowed value in user input
 *      maxValue -- the maximum allowed value in user input
 *      returnValueCount -- the address in which to store the
 *          number valid elements in the returned array.
 *
 *    The caller should free() the returned array when finished
 *    processing.
 *
 *    On error, an error message is printed (via skAppPrintErr()),
 *    *returnValueCount is set to 0, and NULL is returned.
 */


int skIPStringToBitmap(uint32_t bitmap[4][8], const char *ip_string);
/*
 *    Takes a C-string of an IPv4 address or range in the dotted-quad
 *    or CIDR notation and returns an array of four bitmaps of 256
 *    bits each, where each bitmap represents the one octet.  Returns
 *    0 and puts result into 'bitmap' if parsing was successful;
 *    returns 1 if 'ip_string' was empty; returns 2 if it contained
 *    only whitespace; returns -1 if there was text that could not be
 *    parsed.
 *
 *    The 'ip_string' can be a CIDR notation "1.2.3.0/24", an ordinary
 *    dotted quad "1.2.3.4", a dotted-quad with an 'x' respresented an
 *    entire octet "1.2.3.x" (equivalent to "1.2.3.0/24"), or a dotted
 *    quad with lists or ranges in any or all octets "1.2,3.4,5.6,7"
 *    or "1.2.3.0-255".
 */


int skStringParseDatetime(
    uint32_t   *time_val,
    const char *time_string,
    int        *resulting_precision);
/*
 *    Attempts to parse the 'time_string' as a date in the form
 *    YYYY/MM/DD[:HH[:MM[:SS]]].  Sets *time_val to that time in UNIX
 *    epoch seconds.  Assumes the time in UTC.
 *
 *    If 'resulting_precision' is non-null, it is filled with the
 *    index of the last "section" of date parsed.  If it is NULL, the
 *    precision is not recorded (but parsing still occurs).
 *
 *    The "sections" are:
 *        Years parsed = 1
 *        Months parsed = 2
 *        Days parsed = 3
 *        Hours parsed = 4
 *        Minutes parsed = 5
 *        Seconds parsed = 6
 *
 *    (Note that skStringParseDatetime() will return -1 if either
 *    date does not contain at least Day precision.)
 *
 *    Return 0 on success.  Returns -1 and prints an message on an
 *    error; error contions include an empty string, malformed date,
 *    or extra text after the date.
 */


int skStringParseDatetimeRange(
    uint32_t   *start,
    uint32_t   *end,
    const char *s_datetime,
    int        *start_precision,
    int        *end_precision);
/*
 *    Attempts to parse 's_datetime' as a datetime or a range of
 *    datetimes.  If one date is found, it is stored in 'start' (and
 *    the end date is set to UINT32_MAX).
 *
 *    If two dash-separated dates are found, the first is stored in
 *    'start' and the second is stored in 'end'.
 *
 *    If 'start_precision' is not NULL, it is filled with the index of
 *    the last "section" of date parsed.  If it is NULL, the precision
 *    is not recorded (but parsing still occurs).
 *
 *    Similarly, if 'end_precision' is not NULL and a second date was
 *    parsed, it is filled with the index of the last "section" of
 *    date parsed.  If it is NULL, the precision is not recorded (but
 *    parsing still occurs).  (see skStringParseDate() for a
 *    description of "sections" of a date).
 *
 *    (Note that skStringParseDatetimeRange() will return -1 if either
 *    date does not contain at least Day precision.)
 *
 *    Returns -1 if the date found is unparsable, or if a dash is
 *    found, the date on either side is unparsable.
 *
 *    Returns 1 if two dates are found and parsed, but the end-date is
 *    earlier than the start_date.
 *
 *    Returns 0 if dates are parsed correctly.
 */


int skDatetimeCeiling(
    uint32_t   *result,
    uint32_t    datetime,
    int         precision);
/*
 *    Takes 'datetime' (which is a UTC timestamp) and a 'precision' to
 *    which it was parsed (see skStringParseDate() for a description
 *    of "sections" of a date).
 *
 *    Timestamps are usually parsed as the earliest-possible timestamp
 *    that meets the constraints parsed.  For example, "1990/12/25:05"
 *    is assumed to mean "1990/12/25:05:00:00".
 *
 *    skDatetimeCeiling() calculates a new timestamp of the
 *    latest-possible timestamp that meets the constraints parsed.
 *    For example, an input of "1990/12/25:05" would result in the
 *    timestamp for "1990/12/25:05:59:59".
 *
 *    skDatetimeCeiling() is commonly used to calculate the endpoint
 *    for a range of dates.  For example, "2004/12/25-2004/12/26:3"
 *    should represent everything between "2004/12/25:00:00:00" and
 *    "2004/12/26:03:59:59"
 */


int skStringParseUint32(
    uint32_t   *result_val,
    const char *int_string,
    uint32_t    min_val,
    uint32_t    max_val);
/*
 *    Attempts to parse the C string 'int_string' as an unsigned
 *    32-bit integer.  In addition, verifies that the value is between
 *    'min_val' and 'max_val' inclusive.  When 'max_val' is 0, then
 *    there is no maximum.  Puts the result into location pointed to
 *    by 'result_val'.  Ignores any whitespace around the number.
 *
 *    Returns a value >= 0 the string was parsable and the value was
 *    within the limits.  Returns 0 when nothing other than whitespace
 *    was present.  A positive return value is the index into
 *    'int_string' of the first non-whitespace character following the
 *    successfully parsed number; e.g; "1x" would set *result_val to 1
 *    and return 2.
 *
 *    A negative return value signals an error: -1 if 'int_string; is
 *    NULL; -2 if it is empty or contained only whitespace; -3 if
 *    string is unparseable; -4 if value overflows the parser; -11 if
 *    value is less than 'min_val'; -12 is value is greater than
 *    'max_val'.
 */


int skStringParseUint64(
    uint64_t   *result_val,
    const char *int_string,
    uint64_t    min_val,
    uint64_t    max_val);
/*
 *    As skStringParseUint32(), except that it attempts to parse the
 *    C-string 'int_string' as an unsigned 64-bit integer.
 */



/* options flags to pass to skStringParseHumanUint64() */
typedef enum {
    SK_HUMAN_NORMAL = 0,
    SK_HUMAN_LOWER_TRADITIONAL = 0,
    SK_HUMAN_UPPER_TRADITIONAL = 0,
    SK_HUMAN_LOWER_SI = 1,
    SK_HUMAN_UPPER_SI = 2,
    SK_HUMAN_MID_WS = 4,
    SK_HUMAN_MID_NO_WS = 0,
    SK_HUMAN_END_WS = 0,
    SK_HUMAN_END_NO_WS = 8
} skHumanFlags_t;

int skStringParseHumanUint64(
    uint64_t       *result_val,
    const char     *int_string,
    skHumanFlags_t  parse_flags);
/*
 *    Attempts to parse the C string 'int_string' as an unsigned
 *    64-bit integer, ignoring leading whitespace.  In addition,
 *    handles human suffixes such as k, m, g, and t.  Puts the result
 *    into location pointed to by 'result_val'.
 *
 *    Returns a value >= 0 the string was parsable.  Returns 0 when
 *    nothing other than parsed whitespace was present.  A positive
 *    return value is the index into 'int_string' of the first
 *    non-parsed character following the successfully parsed number;
 *    e.g; "1x" would set *result_val to 1 and return 2.
 *
 *    A negative return value signals an error: -1 if 'int_string; is
 *    NULL; -2 if it is empty or contained only whitespace; -3 if
 *    string is unparseable; -4 if value overflows the parser; -5 if
 *    value underflows the parser; -6 if value is NAN.
 *
 *    parse_flags is made by or-ing flags from skHumanFlags_t together.
 *    These flags affect the parsing as follows:
 *    SK_HUMAN_NORMAL             use defaults
 *    SK_HUMAN_LOWER_TRADITIONAL  use 1024 instead of 1000 for k, etc.
 *    SK_HUMAN_UPPER_TRADITIONAL  use 1024 instead of 1000 for K, etc.
 *    SK_HUMAN_LOWER_SI           use 1000 instead of 1024 for k, etc.
 *    SK_HUMAN_UPPER_SI           use 1000 instead of 1024 for K, etc.
 *    SK_HUMAN_MID_WS             ignore white space between number & suffix
 *    SK_HUMAN_MID_NO_WS          don't parse white space betwn number & suffix
 *    SK_HUMAN_END_WS             parse & ignore white space after suffixn
 *    SK_HUMAN_END_NO_WS          do not parse white space after suffix
 */



/* Private macro used by GET_MASKED_BITS and SET_MASKED_BITS.  This
   creates a 32-bit integer with the first `s' least-significat bits
   turned on. */
#define _BITMASK(s) (((s) >= 32) ? (~(uint32_t)0) : (~((~(uint32_t)0) << (s))))

/*
 * GET_MASKED_BITS(x, o, s)
 *
 * Given an integer value (<=32 bits) x, return an integer
 * representing the set of bits within x starting at offset o (from
 * the least-significant bit) of size s in bits.  Works on 32, 16, and
 * 8 bit integers.
 *
 * 76543210  GET_MASKED_BITS(x, 2, 4) would return the value represent by 
 * ..xxxx..  the middle 4 bits of a single byte, as shown at the left.
 */
#define GET_MASKED_BITS(x, o, s) (((x) >> (o)) & _BITMASK((s)))

/*
 * SET_MASKED_BITS(x, v, o, s)
 *
 * Given an integer variable (<=32 bits) x, sets x to the value
 * created by seting the set of bits within x starting at offset o
 * (from the least-significant bit) of size s in bits to v.  This is
 * the setting equivalent to GET_MASKED_BITS.  This works on variables
 * which are 32, 16, or 8 bit integers. 
 */
#define SET_MASKED_BITS(x, v, o, s)                                     \
    ((x) = ((x) & (~(_BITMASK((s)) << (o)))) | (((v) & _BITMASK((s))) << (o)))


/*  skbitmap.c */


typedef struct {
    uint32_t   *map;
    uint32_t    num_bits;
    uint32_t    count;
} sk_bitmap_t;


/* Internal macros.  Not for public use. */
#define _BMAP_INDEX(p)      ((p) >> 5)
#define _BMAP_OFFSET(p)     (1 << ((p) & 0x1F))
#define _BMAP_IS_SET(b, p)  ((b)->map[_BMAP_INDEX(p)] & _BMAP_OFFSET(p))


int skBitmapCreate(sk_bitmap_t **bitmap_out, uint32_t num_bits);
/*
 *    Create a new bitmap capable of holding 'num_bits' bits.  Return
 *    the bitmap in the location given by 'bitmap_out'.  Returns 0 if
 *    successful; -1 for a memory allocation error or if num_bits is
 *    zero.
 */


void skBitmapDestroy(sk_bitmap_t **bitmap);
/*
 *    Destroy the bitmap at location given by 'bitmap'.
 */


void skBitmapClearAllBits(sk_bitmap_t *bitmap);
/*
 *    Turn OFF all the bits in 'bitmap' and set the high-bit count to
 *    zero.
 */


uint32_t skBitmapGetSizeF(const sk_bitmap_t *bitmap);
#ifdef SKBITMAP_DEBUG
#  define skBitmapGetSize(bitmap) skBitmapGetSizeF(bitmap)
#else
#  define skBitmapGetSize(bitmap) ((bitmap)->num_bits)
#endif
/*
 *    Return the number of bits that 'bitmap' can hold.
 */


uint32_t skBitmapGetHighCountF(const sk_bitmap_t *bitmap);
#ifdef SKBITMAP_DEBUG
#  define skBitmapGetHighCount(bitmap) skBitmapGetHighCountF(bitmap)
#else
#  define skBitmapGetHighCount(bitmap) ((bitmap)->count)
#endif
/*
 *    Return the number of bits in 'bitmap' that are ON.
 */


int skBitmapGetBitF(const sk_bitmap_t *bitmap, uint32_t pos);
#ifdef SKBITMAP_DEBUG
#  define skBitmapGetBit(bitmap, pos) skBitmapGetBitF(bitmap, pos)
#else
#  define skBitmapGetBit(bitmap, pos)           \
    (((pos) >= (bitmap)->num_bits)              \
     ? -1                                       \
     : (_BMAP_IS_SET((bitmap), (pos)) ? 1 : 0))
#endif
/*
 *    Return 1 if the bit at position 'pos' in 'bitmap' is ON; return
 *    0 if it is OFF.
 */


void skBitmapSetBitF(sk_bitmap_t *bitmap, uint32_t pos);
#ifdef SKBITMAP_DEBUG
#  define skBitmapSetBit(bitmap, pos) skBitmapSetBitF(bitmap, pos)
#else
#  define skBitmapSetBit(bitmap, pos)                                   \
    if (((pos)>=(bitmap)->num_bits) || _BMAP_IS_SET((bitmap), (pos))) { \
        /* no-op */                                                     \
    } else {                                                            \
        (bitmap)->map[_BMAP_INDEX(pos)] |= _BMAP_OFFSET(pos);           \
        ++(bitmap)->count;                                              \
    }
#endif
/*
 *    Turn ON the bit at position 'pos' in 'bitmap'.  Adjust the
 *    bitmap's high-bit counter.
 */


void skBitmapClearBitF(sk_bitmap_t *bitmap, uint32_t pos);
#ifdef SKBITMAP_DEBUG
#  define skBitmapClearBit(bitmap, pos) skBitmapClearBitF(bitmap, pos)
#else
#  define skBitmapClearBit(bitmap, pos)                                 \
    if (((pos)>=(bitmap)->num_bits) || !_BMAP_IS_SET((bitmap),(pos))) { \
        /* no-op */                                                     \
    } else {                                                            \
        (bitmap)->map[_BMAP_INDEX(pos)] &= ~(_BMAP_OFFSET(pos));        \
        --(bitmap)->count;                                              \
    }
#endif
/*
 *    Turn OFF the bit at position 'pos' in 'bitmap'.  Adjust the
 *    bitmap's high-bit counter.
 */



/*
 *  BITMAP_DECLARE(var_name, size);
 *
 *    Declares the bitmap 'var_name' that will hold size bits numbered
 *    from 0 to size-1.
 */
#define BITMAP_DECLARE(name, size) \
    uint32_t name[((size) >> 5) + ((((size) & 0x1F) == 0) ? 0 : 1)]

/*
 *  BITMAP_INIT(name);
 *
 *    Clears all the bits in the bitmap named 'name'.  This macro must
 *    appear in the same scope as the BITMAP_DECLARE() macro.
 */
#define BITMAP_INIT(name) memset(&name, 0, sizeof(name))

/*
 *  BITMAP_SETBIT(name, pos);
 *  is_set = BITMAP_GETBIT(name, pos);
 *
 *    Set or get the bit as position 'pos' in bitmap 'name'.
 */
#define BITMAP_SETBIT(name, pos) ((name)[_BMAP_INDEX(pos)] |= _BMAP_OFFSET(pos))
#define BITMAP_GETBIT(name, pos) (((name)[_BMAP_INDEX(pos)] & _BMAP_OFFSET(pos)) ? 1 : 0)



/* TCP FLAGS HANDLING */

/* Sets any high-flags in 'flags' to high in 'var'.  Other high-flags
 * in 'var' are not affected. */
#define TCP_FLAG_SET_FLAG( var, flags ) ((var) |= (flags))

/* Returns 1 if the high-flags in 'flags' are also high in 'var';
 * returns 0 otherwise. */
#define TCP_FLAG_TEST( var, flags ) ( (((flag) & (var)) != 0) ? 1 : 0 )

/* Returns 1 if, for all high-flags in 'mask', the only high-bits in
 * 'var' are thost that are set in 'high'; return 0 otherwise.  See
 * skStringParseTCPFlagsHighMask() for details. */
#define TCP_FLAG_TEST_HIGH_MASK( var, high, mask ) \
    ( (((var) & (mask)) == ((high) & (mask))) ? 1 : 0 )



int skStringParseTCPFlags(
    uint8_t    *result,
    const char *flag_string);
/*
 *    Parses 'flag_string' as a block of TCP flags.  Flag strings can
 *    only contain the following characters:
 *        F (FIN)
 *        S (SYN)
 *        R (RESET)
 *        P (PUSH)
 *        A (ACK)
 *        U (URGENT)
 *        E (ECE)
 *        C (CWR)
 *
 *    Returns 0 when 'flag_string' contains only TCP flag characters and
 *    whitespace.
 *
 *    When an unrecognized character is seen, parsing stops and the
 *    function returns the position (1-based) of the that character in
 *    'flag_string'---ie, returns the number of characters successfully
 *    parsed.
 *
 *    Assumes flag_string is a NULL-terminated string.  Overwrites the
 *    value of 'result'.
 */


int skStringParseTCPFlagsHighMask(
    uint8_t    *high,
    uint8_t    *mask,
    const char *flag_string);
/*
 *    Parses 'flag_string' as a high/mask pair of blocks of TCP flags.
 *    Both the high and mask portions of the string are flag strings,
 *    and are interpreted by skStringParseTCPFlags().  Any TCP flags
 *    not listed in 'mask' can have any value; flags listed 'high'
 *    must be set; flags listed in 'mask' but not in 'high' must be
 *    low.  It is an error if a flag is listed in 'high' but not in
 *    'mask'.
 *
 *    For example: "AS/ASFR" means ACK,SYN must be high, FIN,RST must
 *    be low, and the other flags--PSH,URG,ECE,CWR--may have any
 *    value.
 *
 *    If the string is unparsable as high/mask pairs, returns -1.
 *
 *    If the high portion of the string contains a flag not found in
 *    the mask portion of the string, returns -1.
 *
 *    If successful, returns 0.
 *
 *    The inputs 'high' and 'mask' are overwritten with the bitmap of
 *    the flags set.
 */


/* Whether to support 8 bit TCP flags. */
#define TCP_8_BIT_FLAGS 1

/*
 * flag definitions.
 */
#define CWR_FLAG (1 << 7)         /* 128 */
#define ECE_FLAG (1 << 6)         /*  64 */
#define URG_FLAG (1 << 5)         /*  32 */
#define ACK_FLAG (1 << 4)         /*  16 */
#define PSH_FLAG (1 << 3)         /*   8 */
#define RST_FLAG (1 << 2)         /*   4 */
#define SYN_FLAG (1 << 1)         /*   2 */
#define FIN_FLAG (1)              /*   1 */

#endif /* _UTILS_H */
