/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  Interface for converting from NetFlow v5 PDUs (aka cflowd records)
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/flowsource/cflowd2fc.c,v 1.13 2005/10/26 16:17:23 dugganm Exp $");

#include "cflowd2fc.h"

/* Encapsulates information from a netflowV5 header.  The resulting
 * cflowdTimeInfo_t is used to do time calculations when processing
 * individual records. */
void cflowdTimeInfoSetup(v5Header *hdr, cflowdTimeInfo_t *ti)
{
    /* GMT when router was last re-booted based on this PDU */
    uint32_t sysUptime;
    ldiv_t hdr_sys;
    ldiv_t hdr_unix;

    sysUptime = ntohl(hdr->SysUptime);

    /* sysUpTime is in milliseconds, convert to seconds/milliseconds */
    hdr_sys = ldiv(sysUptime, 1000);

    /* unix seconds/nanoseconds to seconds/milliseconds */
    hdr_unix.quot = ntohl(hdr->unix_secs);
    hdr_unix.rem = ntohl(hdr->unix_nsecs) / 1000000L; /* 1e6 */

    /* subtract sysUpTime from unix seconds */
    hdr_unix.quot -= hdr_sys.quot;

    /* borrow a sec? */
    if (hdr_sys.rem > hdr_unix.rem) {
        --hdr_unix.quot;
        hdr_unix.rem += 1000;
    }
    hdr_unix.rem -= hdr_sys.rem;

    ti->rtr_unix_secs = hdr_unix.quot;
    ti->rtr_unix_msecs = hdr_unix.rem;
}




/* Fills out a rwRec in a specific byte order from a v5pdu record.
   The variable ti can be NULL, in which case the start time will not
   necessarily reflect reality. */
void cflowdToGeneric(
    v5Record           *v5RPtr,
    rwRec              *rec,
    cflowdTimeInfo_t   *ti,
    silk_endian_t       endianness)
{
    uint32_t s, e;
    ldiv_t sTime, eTime;
    ldiv_t duration;
    uint16_t sTime_frac, duration_frac;

    /* Setup start and duration */
    e = ntohl(v5RPtr->Last);
    s = ntohl(v5RPtr->First);
    eTime = ldiv(e, 1000);
    sTime = ldiv(s, 1000);
    duration = ldiv(e - s, 1000);
    if (ti) {
        CONVERTTIME(sTime, *ti);
        CONVERTTIME(eTime, *ti);
    }
    sTime.quot = htonl(sTime.quot);
    sTime_frac = htons(sTime.rem);
    duration.quot = htonl(duration.quot);
    duration_frac = htons(duration.rem);

    switch (endianness)
    {
      case SILK_ENDIAN_LITTLE:
#if IS_LITTLE_ENDIAN
      case SILK_ENDIAN_NATIVE:
#endif
        rec->sIP.ipnum = BSWAP32(v5RPtr->srcaddr.ipnum);
        rec->dIP.ipnum = BSWAP32(v5RPtr->dstaddr.ipnum);
        rec->sPort = BSWAP16(v5RPtr->srcport);
        rec->dPort = BSWAP16(v5RPtr->dstport);
        rec->proto = v5RPtr->prot;
        rec->flags = v5RPtr->tcp_flags;
        rec->input = ntohs(v5RPtr->input);
        rec->output = ntohs(v5RPtr->output);
        rec->nhIP.ipnum = BSWAP32(v5RPtr->nexthop.ipnum);
        rec->sTime = BSWAP32(sTime.quot);
        rec->sTime_msec = BSWAP16(sTime_frac);
        rec->pkts = BSWAP32(v5RPtr->dPkts);
        rec->bytes = BSWAP32(v5RPtr->dOctets);
        rec->elapsed = BSWAP32(duration.quot);
        rec->elapsed_msec = BSWAP16(duration_frac);
        break;
      case SILK_ENDIAN_BIG:
      case SILK_ENDIAN_ANY:
#if IS_BIG_ENDIAN
      case SILK_ENDIAN_NATIVE:
#endif
        rec->sIP.ipnum = v5RPtr->srcaddr.ipnum;
        rec->dIP.ipnum = v5RPtr->dstaddr.ipnum;
        rec->sPort = v5RPtr->srcport;
        rec->dPort = v5RPtr->dstport;
        rec->proto = v5RPtr->prot;
        rec->flags = v5RPtr->tcp_flags;
        rec->input = ntohs(v5RPtr->input);
        rec->output = ntohs(v5RPtr->output);
        rec->nhIP.ipnum = v5RPtr->nexthop.ipnum;
        rec->sTime = sTime.quot;
        rec->sTime_msec = sTime_frac;
        rec->pkts = v5RPtr->dPkts;
        rec->bytes = v5RPtr->dOctets;
        rec->elapsed = duration.quot;
        rec->elapsed_msec = duration_frac;
        break;

      default:
        skAppPrintErr("Option %d not handled in switch at %s:%d",
                      endianness, __FILE__, __LINE__);
        abort();
    }

    rec->rest_flags = 0;
    rec->tcp_state = SK_TCPSTATE_NO_INFO;
    rec->application = 0;
}
