/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  Interface to pull a single flow from a NetFlow v5 PDU
**
*/


#include "silk.h"

RCSIDENT("$SiLK: /src/rw/flowsource/pdusource.c,v 1.8 2005/09/08 13:02:06 dugganm Exp $");

#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include "utils.h"
#include "bigsockbuf.h"
#include "v5pdu.h"
#include "cflowd2fc.h"
#include "pdusource.h"
#include "flowcap-stats.h"
#include "udpsource.h"

#define log if (source->logfn) source->logfn

typedef struct pdu_source_t
{
    flowcap_stats_t statistics;

    udpSource_t source;
    int source_index;

    cflowdTimeInfo_t ti;

    sk_msg_fn_t logfn;

    uint8_t count;                /* Number of recs left in pdu */
    v5PDU *pdu;                   /* Current pdu */

    uint32_t flowSeqNumbers[65536];
    BITMAP_DECLARE(engineMasks, 65536);

    int sockbufsize;
    in_addr_t listen_addr;
    int port;

    unsigned badpdu : 1;          /* Encountered a bad rec in this pdu */
    unsigned invertable : 1;       /* This source produces header data */
    unsigned send_head : 1;        /* Send header data */
    unsigned file : 1;             /* File-based source */

} pdu_source_t;

typedef struct pdu_source_pool_item_t 
{
    const probe_def_t *probe;
    udpSource_t source;
    int index;
} pdu_source_pool_item_t;

static udpSource_t pduUdpSourceCreate(
    int port, 
    in_addr_t from_address,
    in_addr_t listen_address,
    uint32_t max_pdus, 
    int sockbufsize,
    sk_msg_fn_t logfn);
static udpSource_t pduUdpFileSourceCreate(const char *path, sk_msg_fn_t logfn);

pduSource_t pduSourceCreate(
    int port, 
    in_addr_t from_address,
    in_addr_t listen_address,
    uint32_t max_pdus, 
    int sockbufsize,
    sk_msg_fn_t logfn)
{
    udpSource_t udpsource;
    pduSource_t source;

    udpsource = pduUdpSourceCreate(port, from_address, listen_address, 
                                   max_pdus, sockbufsize, logfn);
    if (udpsource == NULL) {
        return NULL;
    }

    source = (pduSource_t)calloc(1, sizeof(pdu_source_t));
    if (source == NULL) {
        udpSourceDestroy(udpsource);
        return NULL;
    }

    source->source = udpsource;
    source->source_index = 0;
    source->logfn = logfn;
    source->listen_addr = listen_address;
    source->sockbufsize = sockbufsize;
    source->port = port;

    return source;
}


pduSource_t pduFileSourceCreate(
    const char *path, 
    sk_msg_fn_t logfn)
{
    udpSource_t udpsource;
    pduSource_t source;

    udpsource = pduUdpFileSourceCreate(path, logfn);
    if (udpsource == NULL) {
        return NULL;
    }

    source = (pduSource_t)calloc(1, sizeof(pdu_source_t));
    if (source == NULL) {
        udpSourceDestroy(udpsource);
        return NULL;
    }

    source->file = 1;
    source->source = udpsource;
    source->logfn = logfn;

    return source;
}


void pduSourceDestroy(pduSource_t source)
{
    udpSourceDestroy(source->source);
    free(source);
}


static void handleBadPacket(pduSource_t source, v5PDU *UNUSED(pdu))
{
    source->statistics.badPkts++;
}


static v5PDU *next_pdu(pduSource_t source)
{
    assert (source != NULL);

    /* Infloop; exit by return only */
    while (1) {
        v5PDU *pdu;
        uint16_t count;
        uint16_t engine;
        uint32_t flow_sequence;

        pdu = (v5PDU *)udpNextByIndex(source->source, source->source_index);
        if (pdu == NULL) {
            return NULL;
        }

        source->statistics.procPkts++;

        count = ntohs(pdu->hdr.count);
        if (count > V5PDU_MAX_RECS || count == 0) {
            /* reject packet */
            handleBadPacket(source, pdu);
            continue;
        }

        /* handle seq numbers here */
        flow_sequence = ntohl(pdu->hdr.flow_sequence);
        engine = (pdu->hdr.engine_type << 8) | pdu->hdr.engine_id;
        if (BITMAP_GETBIT(source->engineMasks, engine)) {
            /* check hdr seqnum against expected seq num */
            if (flow_sequence < source->flowSeqNumbers[engine]) {
                /*
                ** Out of sequence packet.  Reduce missing flow count.
                ** However, do not change the expected seq num.
                */
                source->statistics.missingRecs -= count;
            } else {
                /* Check if this is a later packet and handling missing recs */
                if (flow_sequence > source->flowSeqNumbers[engine]) {
                    /* Increase missing flow count */
                    source->statistics.missingRecs += 
                        flow_sequence - source->flowSeqNumbers[engine];
                    log("Missing netflow records: %lld/%llu == %7.4g%%",
                        source->statistics.missingRecs,
                        source->statistics.pktCount,
                        source->statistics.pktCount ?
                        ((float)source->statistics.missingRecs /
                         (float)(source->statistics.pktCount +
                                 source->statistics.missingRecs)) * 100.0 :
                        100.0);
                }
                /* Update the next expected seq */
                source->flowSeqNumbers[engine] = flow_sequence + count;

            }
        } else {
            /* A new engine. Mark and record */
            BITMAP_SETBIT(source->engineMasks, engine);
            source->flowSeqNumbers[engine] = flow_sequence + count;
        }

        cflowdTimeInfoSetup(&pdu->hdr, &source->ti);

        return pdu;
    }
}


static int pduSourceGetNextRec(pduSource_t source)
{
    assert (source != NULL);

    /* Infloop; exit by return only */
    while (1) {
        v5Record *v5RPtr;

        /* If we need a pdu, get a new one, otherwise we are not finished
           with the last. */
        if (source->count == 0) {
            source->pdu = next_pdu(source);
            if (source->pdu == NULL) {
                return -1;
            }
            source->count = ntohs(source->pdu->hdr.count);
            source->badpdu = 0;
            if (source->invertable) {
                source->send_head = 1;
                return 0;
            }
        }

        /* Get next record, and decrement counter*/
        v5RPtr = &source->pdu->data[ntohs(source->pdu->hdr.count)
                                    - source->count--];

        /* Check for zero packets or bytes.  No need for byteswapping
           when checking zero. */
        if (v5RPtr->dPkts == 0 || v5RPtr->dOctets == 0) {
            log("Netflow record has zero packets or bytes.");
            if (!source->badpdu) {
                handleBadPacket(source, source->pdu);
                source->badpdu = 1;
            }
            if (!source->invertable) {
                continue;
            }
        }

        /* Check to see if more packets them bytes. */
        if (ntohl(v5RPtr->dPkts) > ntohl(v5RPtr->dOctets)) {
            log("Netflow record has more packets them bytes.");
            if (!source->badpdu) {
                handleBadPacket(source, source->pdu);
                source->badpdu = 1;
            }
            if (!source->invertable) {
                continue;
            }
        }

        /* Check for earlier end time than start time */
        if (ntohl(v5RPtr->Last) < ntohl(v5RPtr->First)) {
            log("Netflow record has earlier end time than start time.");
            if (!source->badpdu) {
                handleBadPacket(source, source->pdu);
                source->badpdu = 1;
            }
            if (!source->invertable) {
                continue;
            }
        }

        /* Check for bogosities in how the ICMP type/code are set.  It
           should be in dest port, but sometimes it is backwards in src
           port. */
        if (v5RPtr->prot == 1 &&  /* ICMP */
            v5RPtr->dstport == 0) /* No byteswapping for check against 0 */
        {
            uint32_t *ports = (uint32_t *)&v5RPtr->srcport;
            *ports = BSWAP32(*ports); /* This will swap src into dest,
                                         while byteswapping. */
        }

        source->statistics.pktCount++;

        return 0;
    }
}




int pduSourceGetGeneric(
    pduSource_t source, 
    rwRec *rec,
    silk_endian_t endianness)
{
    if (pduSourceGetNextRec(source)) {
        return -1;
    }

    /* If a header was set, discard it.  We do not use it. */
    if (source->send_head) {
        source->send_head = 0;
        if (pduSourceGetNextRec(source)) {
            return -1;
        }
    }        
    cflowdToGeneric(&source->pdu->data[ntohs(source->pdu->hdr.count) -
                                       (source->count + 1)],
                    rec, &source->ti, endianness);
    return 0;
}


static udpSource_t pduUdpSourceCreate(
    int port, 
    in_addr_t from_address,
    in_addr_t listen_address,
    uint32_t max_pdus, 
    int sockbufsize,
    sk_msg_fn_t logfn)
{
    struct sockaddr_in addr;
    int sock;

    /* Get a socket */
    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock == -1) {
        return NULL;
    }

    /* Make the socket receiver buffer size as close as we can to
       sockbufsize. */
    bigsockbuf(sock, SO_RCVBUF, sockbufsize);

    /* Set up the address structure */
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(listen_address);
    addr.sin_port = htons(port);

    /* Bind socket to port */
    if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
        close(sock);
        return NULL;
    }

    return udpSourceCreate(sock, from_address, V5PDU_LEN, max_pdus, logfn);
}


static udpSource_t pduUdpFileSourceCreate(const char *path,
                                          sk_msg_fn_t logfn)
{
    return udpFileSourceCreate(path, V5PDU_LEN, logfn);
}

pduSourcePool_t pduSourcePoolCreate()
{
    return skVectorNew(sizeof(pdu_source_pool_item_t));
}

void pduSourcePoolDestroy(pduSourcePool_t pool)
{
    skVectorDestroy(pool);
}

pduSource_t pduSourceCreateFromProbeDef(
    pduSourcePool_t pool,
    const probe_def_t *probe,
    uint32_t max_pdus, 
    int sockbufsize,
    sk_msg_fn_t logfn)
{
    uint8_t i;
    in_addr_t paddr;
    in_addr_t haddr;
    uint16_t pport;
    pduSource_t source;
    pdu_source_pool_item_t current;
    int rv;

    assert(pool);
    assert(probe);

    rv = skProbeGetListenAsHost(&paddr, &pport, probe);
    if (rv == -1) {
        return NULL;
    }
    skProbeGetAcceptFromHost(&haddr, probe);

    /* Loop through the udpSources to see if one matches the probe. */
    for (i = 0; i < skVectorGetCount(pool); i++)
    {
        in_addr_t addr;
        uint16_t port;
        
        skVectorGetValue(&current, pool, i);
        
        rv = skProbeGetListenAsHost(&addr, &port, current.probe);
        if (rv == -1) {
            continue;
        }

        if (port == pport && addr == paddr) {
            in_addr_t from_addr;
            skProbeGetAcceptFromHost(&from_addr, current.probe);

            if (from_addr == haddr) {
                /* Matches in all respects.  Create a duplicate. */
                source = (pduSource_t)calloc(1, sizeof(pdu_source_t));
                if (source == NULL) {
                    return NULL;
                }
                
                source->source = current.source;
                udpSourceIncRef(current.source);
                source->source_index = current.index;
                source->logfn = logfn;
                
                return source;
            }
            if (from_addr == INADDR_ANY ||
                haddr == INADDR_ANY) {
                /* Can't mix INADDR_ANY and specific ips */
                return NULL;
            }

            /* Matches in all but from_addr.  Add that addr. */
            source = (pduSource_t)calloc(1, sizeof(pdu_source_t));
            if (source == NULL) {
                return NULL;
            }
                
            source->source = current.source;
            udpSourceIncRef(current.source);
            source->source_index = udpSourceAddAddress(source->source, haddr);
            source->logfn = logfn;
                
            return source;
        }
    }
    
    /* No match.  Add new source. */
    source = pduSourceCreate(pport, haddr, paddr, max_pdus, 
                             sockbufsize, logfn);
    if (source == NULL) {
        return NULL;
    }
    current.source = source->source;
    current.index = source->source_index;
    current.probe = probe;
    skVectorAppendValue(pool, &current);

    return source;
}

int pduSourceReset(pduSource_t source)
{
    struct sockaddr_in addr;
    int sock = 0;

    udpSourceResetStop(source->source, 1);
    
    if (!source->file) {
        /* Get a socket */
        sock = socket(AF_INET, SOCK_DGRAM, 0);
        if (sock == -1) {
            return -1;
        }

        /* Make the socket receiver buffer size as close as we can to
        sockbufsize. */
        bigsockbuf(sock, SO_RCVBUF, source->sockbufsize);

        /* Set up the address structure */
        memset(&addr, 0, sizeof(addr));
        addr.sin_family = AF_INET;
        addr.sin_addr.s_addr = htonl(source->listen_addr);
        addr.sin_port = htons(source->port);

        /* Bind socket to port */
        if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
            close(sock);
            return -1;
        }
    }

    return udpSourceResetStart(source->source, &sock);
}
