/*-------------------------------------------------------------------------
 *  filebuf.c: Routines to read files using a buffer
 *
 *  Copyright (C) 2004 by Accelerated Servers,
 *  			  Todd K McAnally <todd@acceleratedservers.com>
 *  			  Charles Galpin  <charles@acceleratedservers.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id$
 *-----------------------------------------------------------------------*/


#include "silk.h"

#include <sys/times.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <netinet/in.h>
#include <unistd.h>

#include "filebuf.h"
#include "profile.h"

extern struct profile prof1;


static char tkmlbuf[9011200];

/*
 * fname - stolen from iod.c (pvfs) - takes an inode number and computes
 * the filename on a pvfs node.
 */


/*
 * open_buffered - open file and allocate buffer
 *   filename - name of file to open
 *   bufsize - number of bytes to allocate for buffer
 *   (returned) - pointer to allocated buffer structure
 */

bufStructPtr open_buffered(const char *filename, int UNUSED(bufsize), b_file_type b_type){

  bufStructPtr bstr;
  int fp = 0;

  if (!filename) {
    fprintf(stderr, "open_buffered: no filename given");
    return NULL;
  }

  switch (b_type) {
    case b_unix:
      fp = open(filename, O_RDONLY|O_NONBLOCK);
      break;
    case b_pvfs:
      break;
    case b_pvfs_local:
      break;
  }

  if (fp == -1) {
    fprintf(stderr, "open_buffered: open returned -1: %s\n",filename);
    return NULL;
  }


  bstr = (bufStructPtr)calloc(1, sizeof(struct bufStruct));
  if (!bstr) return NULL;

  /*
  bstr->buffer = calloc(bufsize, 1);
  if (!bstr->buffer) {
    free(bstr);
    return NULL;
  }

  bstr->buffer_size = bufsize;
  */

  bstr->buffer = tkmlbuf;
  bstr->buffer_size = sizeof(tkmlbuf);

  bstr->chars_in_buffer = 0;
  bstr->next_char_in_buffer = 0;
  bstr->chars_left_to_read = 0;
  bstr->fp = fp;
  bstr->b_type = b_type;
  bstr->node_base = 0;

  return bstr;
}


/* 
 * read_buffered - read data from file/buffer
 *   bstruct - pointer to structure returned by open_buffered
 *   buf - array to put data into
 *   readsize - number of bytes to load into buf
 *   (returned) - number of bytes actually loaded into buffer
 *     0 means EOF
 *     -1 means error
 */

int read_buffered(bufStructPtr bstr, uint8_t * buf, int readsize){
  int bytes_copied = 0;
  int bytes_to_copy = 0;
  int bytes_left_to_copy;

  if (!bstr) {
    fprintf(stderr, "read_buffered: bufStruct is null\n");
    return -1;
  }

  if (!buf) {
    fprintf(stderr, "read_buffered: buf is null\n");
    return -1;
  }

  if (readsize <= 0) {
    fprintf(stderr, "read_buffered: readsize is <= 0\n");
    return -1;
  }

  bytes_left_to_copy = readsize;

  while (!bstr->eof) {
    bytes_to_copy = MIN( bytes_left_to_copy, 
             bstr->chars_left_to_read);
    if (bytes_to_copy) {

      memcpy( &buf[bytes_copied], 
        &(bstr->buffer[bstr->next_char_in_buffer]), 
        bytes_to_copy);
  
      bstr->next_char_in_buffer += bytes_to_copy;
      bstr->chars_left_to_read = bstr->chars_in_buffer - bstr->next_char_in_buffer;
      bytes_left_to_copy -= bytes_to_copy;
      bytes_copied += bytes_to_copy;
    
      /* done, read all bytes */
      if (bytes_copied == readsize) return bytes_copied;

    }

    /*
    if (bytes_copied) {
        fprintf(stderr, "reloading buffer with bytes left to copy %d\n", bytes_copied);
    }
    */

    if (bstr->chars_left_to_read == 0) {
      /* do the read */
      switch (bstr->b_type) {
        case b_unix:
        case b_pvfs_local:
          bstr->chars_in_buffer = 
            read(bstr->fp, bstr->buffer, bstr->buffer_size);
          break;
        case b_pvfs:
          break;
      }

      /*
      if (bstr->chars_in_buffer != bstr->buffer_size) {
        fprintf(stderr, "asked for %d got %d\n", bstr->buffer_size, bstr->chars_in_buffer);
      }
      */
      if (bstr->chars_in_buffer == 0) {
        bstr->eof = 1;
        return bytes_copied;
      }
      bstr->next_char_in_buffer = 0;
      bstr->chars_left_to_read = bstr->chars_in_buffer - bstr->next_char_in_buffer;
    }
  } /* while */

  return 0;
}

/*
 * close_buffered - close file and free bufStruct
 *   (returned) 1 on success
 *             0 on failure
 */

int close_buffered(bufStructPtr bstr){
  if (bstr) {
    switch (bstr->b_type) {
      case b_unix:
      case b_pvfs_local:
        close(bstr->fp);
        break;
      case b_pvfs:
        break;
    }

    /*
           if (bstr->buffer) free(bstr->buffer);
    free(bstr);
    */
  }
  return 0;
}


off_t seek_buffered(bufStructPtr bstr, off_t offset, int where) {
  if (bstr) {
    switch (bstr->b_type) {
      case b_unix:
      case b_pvfs_local:
        return lseek(bstr->fp, offset, where);
        break;
      case b_pvfs:
        break;
    }

  }
  return 0;
}


#if 0 /* thomasm 2005.08.16 Moving to full header */
#endif /* 0 */
