/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwheaders.c
**
**  Routines for reading/writing rwfilter header files
**  (moved from util/filter/filterio.c)
**
**/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/librw/rwheaders.c,v 1.15 2005/11/08 18:05:13 thomasm Exp $");

#include "librw_priv.h"


/*
 * This ReadHeader and WriteHeader functions in this file work with
 * either file pointers (FILE*) or file descriptors (int).  The helper
 * functions use the following macros to call the appropriate C
 * function to read/write the data from/to the correct stream object.
 * These macros assume fread()/fwrite() are in use if the FILE* is not
 * NULL; otherwise the read()/write() system calls are used.  The
 * macros return TRUE if the call was successful, FALSE otherwise.
 */
#define WRITE(buffer, sz, filep, fd) \
    ((NULL != (filep)) \
     ? (1 == fwrite((buffer), (sz), 1, (filep)))        \
     : (((ssize_t)(sz)) == write((fd), (buffer), (sz))))

#define READ(buffer, sz, filep, fd) \
    ((NULL != (filep)) \
     ? (1 == fread((buffer), (sz), 1, (filep)))         \
     : (((ssize_t)(sz)) == read((fd), (buffer), (sz))))


int fdfilterWriteHeaderV1(
    rwIOStruct     *rwIOS,
    int             fildes)
{
    return _headersWriteV1(rwIOS, fildes);
}


int fdfilterReadHeaderV1(
    rwIOStruct     *rwIOS,
    int             fildes)
{
    return _headersReadV1(rwIOS, fildes);
}


int _headersCreateV1(
    rwIOStruct     *rwIOS)
{
    int rv = LIBRW_OK;
    size_t old_len;
    void *old_hdr;

    assert(rwIOS);
    assert(rwIOS->hdr);

    /* store old length and set new length */
    old_len = rwIOS->hdrLen;
    rwIOS->hdrLen = sizeof(filterHeaderV1);

    /* return if old length and new length are the same */
    if (old_len == rwIOS->hdrLen) {
        /* Assume this function was previously called.  Do not clear
         * anything in the header. */
        goto END;
    }

    /* the old header should be a genericHeader */
    assert(old_len == sizeof(genericHeader));

    /* store the current header, then grow it to hold the start time  */
    old_hdr = rwIOS->hdr;
    rwIOS->hdr = realloc(rwIOS->hdr, rwIOS->hdrLen);
    if (NULL == rwIOS->hdr) {
        rwIOS->hdr = old_hdr;
        rwIOS->hdrLen = old_len;
        rv = LIBRW_ERR_ALLOC;
        goto END;
    }

    /* clear newly allocated section */
    assert(rwIOS->hdrLen > old_len);
    memset((((uint8_t*)rwIOS->hdr) + old_len), 0, (rwIOS->hdrLen - old_len));

  END:
    return rv;
}


/*
 * filterAddFInfoToHeaderV1:
 *      add a new filter info header to the given header.
 * Inputs:
 *      filterHeaderV1Ptr, argc, argv
 * Outputs:
 *      0 if OK. 1 else
 */
int _headersAppendArgv(
    rwIOStruct     *rwIOS,
    int             argc,
    char * const   *argv)
{
    int rv = LIBRW_OK;
    int i;
    char *cp;
    uint32_t len;
    filterHeaderV1 *hdr;
    const char *app_name = skAppName();

    assert(rwIOS);

    /* create the header */
    rv = _headersCreateV1(rwIOS);
    if (rv) {
        goto END;
    }

    hdr = (filterHeaderV1*)rwGetHeader(rwIOS);
    assert(hdr);

    /* make certain there is something to do */
    if (argc == 0) {
        goto END;
    }

    /* grow the header if there are no entries or if the number of
     * entries is at capacity */
    if (hdr->filterCount == hdr->filterCapacity) {
        filterInfoV1 *new_array = NULL;
        uint32_t new_cap = 2 + hdr->filterCapacity;

        if (hdr->filterCapacity == 0) {
            new_array = malloc(new_cap * sizeof(filterInfoV1));
        } else {
            new_array = realloc(hdr->fiArray, new_cap*sizeof(filterInfoV1));
        }

        if (new_array == NULL) {
            rv = LIBRW_ERR_ALLOC;
            goto END;
        }
        hdr->fiArray = new_array;
        hdr->filterCapacity = new_cap;
    }

    /* compute the total length of the application name and the
     * command line arguments */
    len = strlen(app_name) + 1; /* allow for final \0 */
    for(i = 1; i < argc; i++) {
        len += strlen(argv[i]) + 1; /* allow for the \0 */
        /* truncate the command line if too long */
        if (len >= UINT16_MAX) {
            len -= strlen(argv[i]) + 1;
            argc = i;
            break;
        }
    }

    /* store length */
    hdr->fiArray[hdr->filterCount].byteCount = (uint16_t)len;

    /* malloc the required space for the command line */
    hdr->fiArray[hdr->filterCount].info = malloc(len * sizeof(char));
    if (NULL == hdr->fiArray[hdr->filterCount].info) {
        rv = LIBRW_ERR_ALLOC;
        goto END;
    }

    /* 'cp' will be the current position in the command line array */
    cp = hdr->fiArray[hdr->filterCount].info;

    /* store the application name */
    len = strlen(app_name) + 1;
    memcpy(cp, app_name, len);
    cp += len;

    /* store the command line */
    for(i = 1; i < argc; i++) {
        /* copy all including trailing NUL */
        len = strlen(argv[i]) + 1;
        memcpy(cp, argv[i], len);
        cp += len;
    }

    /* set total length and increment filter-count */
    hdr->totFilterLen += (sizeof(hdr->fiArray[hdr->filterCount].byteCount)
                          + hdr->fiArray[hdr->filterCount].byteCount);
    ++hdr->filterCount;

  END:
    return rv;
}
/* filteraddFInfoToHeaderV1 */


int _headersAppendFromFile(
    rwIOStruct         *rwIOS,
    const rwIOStruct   *source_rwIOS)
{
    int rv = LIBRW_OK;
    filterHeaderV1 *hdr;
    const filterHeaderV1 *src_hdr;
    uint32_t src, tgt;

    assert(rwIOS);
    assert(source_rwIOS);

    /* create the header */
    rv = _headersCreateV1(rwIOS);
    if (rv) {
        goto END;
    }

    hdr = (filterHeaderV1*)rwGetHeader(rwIOS);
    assert(hdr);

    src_hdr = (filterHeaderV1*)rwGetHeader(source_rwIOS);
    assert(src_hdr);

    /* make certain there is something to do */
    if (src_hdr->filterCount == 0) {
        goto END;
    }

    /* grow the header if there are no entries or if the number of
     * entries is at capacity */
    if ((hdr->filterCount + src_hdr->filterCount) >= hdr->filterCapacity) {
        filterInfoV1 *new_array = NULL;
        uint32_t new_cap = 2 + hdr->filterCapacity + src_hdr->filterCount;

        if (hdr->filterCapacity == 0) {
            new_array = malloc(new_cap * sizeof(filterInfoV1));
        } else {
            new_array = realloc(hdr->fiArray, new_cap*sizeof(filterInfoV1));
        }

        if (new_array == NULL) {
            rv = LIBRW_ERR_ALLOC;
            goto END;
        }
        hdr->fiArray = new_array;
        hdr->filterCapacity = new_cap;
    }

    /* loop over all command-line entries in the source header, and
     * copy them into the target header. */
    for (src=0, tgt=hdr->filterCount; src < src_hdr->filterCount; ++src,++tgt)
    {
        /* store length */
        hdr->fiArray[tgt].byteCount = src_hdr->fiArray[src].byteCount;

        /* malloc the required space for the command line */
        hdr->fiArray[tgt].info = malloc(hdr->fiArray[tgt].byteCount);
        if (NULL == hdr->fiArray[tgt].info) {
            rv = LIBRW_ERR_ALLOC;
            goto END;
        }

        /* copy command line */
        memcpy(hdr->fiArray[tgt].info, src_hdr->fiArray[src].info,
               hdr->fiArray[tgt].byteCount);

        /* update total length and increment filter-count */
        hdr->totFilterLen += (sizeof(hdr->fiArray[tgt].byteCount)
                              +  hdr->fiArray[tgt].byteCount);
        ++hdr->filterCount;
    }

  END:
    return rv;
}


int _headersReadV1(
    rwIOStruct     *rwIOS,
    int             fildes)
{
    int rv = LIBRW_OK;
    uint8_t padding[SK_MAX_RECORD_SIZE];
    filterHeaderV1 *hdr;
    uint32_t pad_len;
    uint16_t cmd_line_len;
    uint32_t i;

    /* make certain we are using the correct swapping below */
    assert(sizeof(cmd_line_len) == sizeof(hdr->fiArray[i].byteCount)
           && 2 == sizeof(cmd_line_len));

    /* create room for the command lines */
    rv = _headersCreateV1(rwIOS);
    if (rv) {
        goto END;
    }

    /* check record size */
    assert(0 < rwIOS->recLen);
    assert(rwIOS->recLen < SK_MAX_RECORD_SIZE);

    hdr = (filterHeaderV1*)rwGetHeader(rwIOS);

    /* read the number of filter information structs we have */
    if (!READ(&hdr->filterCount, sizeof(hdr->filterCount), rwIOS->FD, fildes)){
        rwIOS->errnum = errno;
        rv = LIBRW_ERR_READ;
        goto END;
    }
    if (rwIOS->swapFlag) {
        hdr->filterCount = BSWAP32(hdr->filterCount);
    }

    /* keep track of number of bytes of header on disk */
    hdr->totFilterLen = sizeof(genericHeader) + sizeof(hdr->filterCount);

    /* allocate the required space */
    if (hdr->filterCount > 0) {
        hdr->filterCapacity = 2 + hdr->filterCount;
        hdr->fiArray = malloc(hdr->filterCapacity * sizeof(filterInfoV1));
        if (hdr->fiArray == NULL) {
            rv = LIBRW_ERR_ALLOC;
            goto END;
        }
    }

    /* and load each one into newly malloc'd space */
    for (i = 0; i < hdr->filterCount; i++) {
        /* the info length */
        if (!READ(&cmd_line_len, sizeof(cmd_line_len), rwIOS->FD, fildes)) {
            rwIOS->errnum = errno;
            rv = LIBRW_ERR_READ;
            goto END;
        }
        if (rwIOS->swapFlag) {
            cmd_line_len = BSWAP16(cmd_line_len);
        }

        if (cmd_line_len) {
            /* get space for the struct in the array */
            hdr->fiArray[i].info = malloc(cmd_line_len * sizeof(char));
            if (NULL == hdr->fiArray[i].info) {
                hdr->filterCount = i;
                rv = LIBRW_ERR_ALLOC;
                goto END;
            }

            /* read in the real information */
            if (!READ(hdr->fiArray[i].info, cmd_line_len, rwIOS->FD, fildes)) {
                free(hdr->fiArray[i].info);
                rv = LIBRW_ERR_ALLOC;
                goto END;
            }
        }

        hdr->fiArray[i].byteCount = cmd_line_len;
        hdr->totFilterLen += cmd_line_len + sizeof(cmd_line_len);
    }

    /* Read the padding if it is present. */
    if (hdr->gHdr.version >= 2) {
        pad_len = rwIOS->recLen - (hdr->totFilterLen % rwIOS->recLen);
        /* if pad_len == recLen, the header is already an even
         * multiple of the record size, and there is no padding. */
        if (pad_len < rwIOS->recLen) {
            if ( !READ(&padding,  pad_len, rwIOS->FD, fildes)) {
                rwIOS->errnum = errno;
                rv = LIBRW_ERR_READ;
                goto END;
            }
        }
    }

  END:
    return rv;
}


int _headersWriteV1(
    rwIOStruct     *rwIOS,
    int             fildes)
{
    static uint8_t padding[SK_MAX_RECORD_SIZE];
    int rv = LIBRW_OK;
    filterHeaderV1 *hdr;
    size_t num_bytes;
    size_t bytes_written;
    uint32_t filter_count;
    uint16_t cmd_line_len;
    uint32_t i;

    /* make certain we are using the correct swapping below */
    assert(sizeof(filter_count) == sizeof(hdr->filterCount)
           && 4 == sizeof(filter_count));
    assert(sizeof(cmd_line_len) == sizeof(hdr->fiArray[i].byteCount)
           && 2 == sizeof(cmd_line_len));

    /* check input and record size */
    assert(rwIOS);
    assert(0 < rwIOS->recLen);
    assert(rwIOS->recLen < SK_MAX_RECORD_SIZE);

    hdr = (filterHeaderV1*)rwGetHeader(rwIOS);
    assert(hdr);

    /* We need to keep track of bytes written so we can pad out to a
     * multiple of the record length. */
    bytes_written = 0;

    /* generic header */
    num_bytes = sizeof(genericHeader);
    if ( !WRITE(hdr, num_bytes, rwIOS->FD, fildes)) {
        rwIOS->errnum = errno;
        rv = LIBRW_ERR_WRITE;
        goto END;
    }
    bytes_written += num_bytes;

    /* number of command lines */
    num_bytes = sizeof(hdr->filterCount);
    if ( !rwIOS->swapFlag) {
        filter_count = hdr->filterCount;
    } else {
        filter_count = BSWAP32(hdr->filterCount);
    }
    if ( !WRITE(&filter_count, num_bytes, rwIOS->FD, fildes)) {
        rwIOS->errnum = errno;
        rv = LIBRW_ERR_WRITE;
        goto END;
    }
    bytes_written += num_bytes;

    /* handle each command line */
    for (i = 0; i < hdr->filterCount; i++) {
        num_bytes = hdr->fiArray[i].byteCount;
        cmd_line_len = ((rwIOS->swapFlag) ? BSWAP16(num_bytes) : num_bytes);

        if ( !(WRITE(&cmd_line_len, sizeof(cmd_line_len), rwIOS->FD, fildes)
               && WRITE(hdr->fiArray[i].info, num_bytes, rwIOS->FD, fildes)))
        {
                rwIOS->errnum = errno;
                rv = LIBRW_ERR_WRITE;
                goto END;
        }
        bytes_written += num_bytes + sizeof(cmd_line_len);
    }

    /* If required, add padding so that header length is multiple of the
     * record length. */
    if (hdr->gHdr.version >= 2) {
        num_bytes = rwIOS->recLen - (bytes_written % rwIOS->recLen);
        /* if num_bytes == recLen, the header is already an even
         * multiple of the record size, and there is no padding. */
        if (num_bytes < rwIOS->recLen) {
            if ( !WRITE(padding, num_bytes, rwIOS->FD, fildes)) {
                rwIOS->errnum = errno;
                rv = LIBRW_ERR_WRITE;
                goto END;
            }
        }
    }

  END:
    return rv;
}
