/*
** Copyright (C) 2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  Functions to support binary output: opening binary SiLK files for
**  writing or appending.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/librw/rwio.c,v 1.6 2005/10/13 14:44:28 thomasm Exp $");

#include "librw_priv.h"


/* Buffer size for buffered I/O on the clusters */
#define BUFSIZE 90112


static int _ioIsModifiable(
    rwIOStruct     *rwIOS)
{
    if (rwIOS == NULL) {
        return LIBRW_ERR_NULL_AUGUMENT;
    }

    if (rwIOS->FD != NULL) {
        return LIBRW_ERR_ATTRIBUTE_FIXED;
    }

    if (rwIOS->io_mode == SK_RWIO_CLOSED) {
        return LIBRW_ERR_STREAM_CLOSED;
    }

    if (rwIOS->io_mode != SK_RWIO_WRITE) {
        return LIBRW_ERR_ATTRIBUTE_FIXED;
    }

    return LIBRW_OK;
}


static int _ioIsNotOpen(
    rwIOStruct     *rwIOS)
{
    if (rwIOS == NULL) {
        return LIBRW_ERR_NULL_AUGUMENT;
    }

    if (rwIOS->io_mode == SK_RWIO_CLOSED) {
        return LIBRW_ERR_STREAM_CLOSED;
    }

    if (rwIOS->FD != NULL) {
        return LIBRW_ERR_STREAM_OPEN;
    }

    return LIBRW_OK;
}


static int _ioHeaderWritten(
    rwIOStruct     *rwIOS)
{
    if (rwIOS == NULL) {
        return LIBRW_ERR_NULL_AUGUMENT;
    }

    if (rwIOS->io_mode == SK_RWIO_CLOSED) {
        return LIBRW_ERR_STREAM_CLOSED;
    }

    if (rwIOS->io_mode != SK_RWIO_WRITE) {
        return LIBRW_ERR_UNSUPPORT_STREAM;
    }

    /* If the write-function is set and the write-body-function is not
     * set, the header has been printed. */
    if ((rwIOS->rwWriteFn != NULL) && (rwIOS->writeBodyRecordsFn == NULL)) {
        return LIBRW_ERR_ATTRIBUTE_FIXED;
    }

    return LIBRW_OK;
}
    

static int _ioSetFileType(
    rwIOStruct     *rwIOS,
    uint8_t         fileType)
{
    assert(rwIOS);
    assert(rwGetHeader(rwIOS));
    ((genericHeader*)rwGetHeader(rwIOS))->type = fileType;
    return LIBRW_OK;
}


static int _ioSetFileVersion(
    rwIOStruct     *rwIOS,
    fileVersion_t   fileVersion)
{
    assert(rwIOS);
    assert(rwGetHeader(rwIOS));
    ((genericHeader*)rwGetHeader(rwIOS))->version = fileVersion;
    return LIBRW_OK;
}


static int _ioSetSwapFlag(
    rwIOStruct     *rwIOS)
{
    assert(rwIOS);
    assert(rwGetHeader(rwIOS));

    /* Do we swap the bytes on write? */
#if     IS_LITTLE_ENDIAN
    rwIOS->swapFlag = rwGetIsBigEndian(rwIOS);
#else
    rwIOS->swapFlag = !rwGetIsBigEndian(rwIOS);
#endif
    return LIBRW_OK;
}


static int _ioSetFileByteorder(
    rwIOStruct     *rwIOS,
    silk_endian_t   byte_order)
{
    genericHeader *gHdr;

    assert(rwIOS);
    assert(rwGetHeader(rwIOS));

    gHdr = (genericHeader*)rwGetHeader(rwIOS);

    switch (byte_order) {
      case SILK_ENDIAN_BIG:
#if !IS_LITTLE_ENDIAN
      case SILK_ENDIAN_NATIVE:
      case SILK_ENDIAN_ANY:
#endif
        gHdr->isBigEndian = 1;
        break;
      case SILK_ENDIAN_LITTLE:
#if IS_LITTLE_ENDIAN
      case SILK_ENDIAN_NATIVE:
      case SILK_ENDIAN_ANY:
#endif
        gHdr->isBigEndian = 0;
        break;
      default:
        assert(0);
        skAppPrintErr("Option %d not handled in switch() at %s:%d",
                      byte_order, __FILE__, __LINE__);
        abort();
    }

    return _ioSetSwapFlag(rwIOS);
}


static int _ioPathnameIsCompressed(
    const char     *pathname)
{
    const char *gz;

    gz = strstr(pathname, ".gz");
    return ((gz != NULL) && ((gz[3] == '\0') || (gz[3] == '.')));
}


static int _ioPathnameIsFiledesc(
    const char     *pathname)
{
    int fildes;
    if (1 == sscanf(pathname, "[%d]", &fildes)) {
        return fildes;
    }
    return -1;
}


static int _ioCheckPathName(
    const char     *pathname,
    rwio_mode_t     read_write_append)
{
    int rv = LIBRW_OK;

    /* check name */
    if (pathname == NULL || *pathname == '\0') {
        rv = LIBRW_ERR_NULL_AUGUMENT;
        goto END;
    }

    /* If pathname is "stdin" or "stdout", make certain stream is not
     * a terminal */
    if (0 == strcmp(pathname, "stdin")) {
        switch (read_write_append) {
          case SK_RWIO_CLOSED:
            rv = LIBRW_ERR_STREAM_CLOSED;
            break;
          case SK_RWIO_READ:
            if (FILEIsATty(stdin)) {
                rv = LIBRW_ERR_UNSUPPORT_STREAM;
            }
            break;
          case SK_RWIO_WRITE:
          case SK_RWIO_APPEND:
            /* cannot write or append to stdin */
            rv = LIBRW_ERR_UNSUPPORT_STREAM;
            break;
        }
        goto END;
    }

    if (0 == strcmp(pathname, "stdout")) {
        switch (read_write_append) {
          case SK_RWIO_CLOSED:
            rv = LIBRW_ERR_STREAM_CLOSED;
            break;
          case SK_RWIO_READ:
          case SK_RWIO_APPEND:
            /* cannot read from or append to stdout */
            rv = LIBRW_ERR_UNSUPPORT_STREAM;
            break;
          case SK_RWIO_WRITE:
            if (FILEIsATty(stdout)) {
                rv = LIBRW_ERR_UNSUPPORT_STREAM;
            }
            break;
        }
        goto END;
    }

    if (read_write_append == SK_RWIO_APPEND) {
        if (_ioPathnameIsCompressed(pathname)) {
            rv = LIBRW_ERR_UNSUPPORT_COMPRESS;
            goto END;
        }
        if (isFIFO(pathname)) {
            rv =  LIBRW_ERR_UNSUPPORT_STREAM;
            goto END;
        }
    }

  END:
    return rv;
}    


static int _ioOpenRead(
    rwIOStruct     *rwIOS)
{
    int rv = LIBRW_OK;
    int is_pipe;

    assert(rwIOS);
    assert(rwGetFileName(rwIOS));

    if (0 == strcmp(rwGetFileName(rwIOS), "stdin")) {
        rwIOS->FD = stdin;
        rwIOS->closeFn = NULL;
    } else if (rwIOS->pvfsFlag) {
    } else {
        if (openFile(rwGetFileName(rwIOS), 0, &(rwIOS->FD), &is_pipe)) {
            rwIOS->errnum = errno;
            rv = LIBRW_ERR_OPEN;
            goto END;
        }

        if (is_pipe) {
            rwIOS->closeFn = &pclose;
        } else {
            rwIOS->closeFn = &fclose;
        }
    }

  END:
    return rv;
}


static int _ioOpenAppend(
    rwIOStruct     *rwIOS)
{
    int rv = LIBRW_OK;

    assert(rwIOS);
    assert(rwGetFileName(rwIOS));

    /* Open file for read and write; position at start. */
    rwIOS->FD = fopen(rwGetFileName(rwIOS), "r+");
    if (rwIOS->FD == NULL) {
        rwIOS->errnum = errno;
        rv = LIBRW_ERR_OPEN;
        goto END;
    }
    rwIOS->closeFn = fclose;

  END:
    return rv;
}


static int _ioOpenWrite(
    rwIOStruct     *rwIOS)
{
    int rv = LIBRW_OK;
    int fildes;

    assert(rwIOS);
    assert(rwGetFileName(rwIOS));

    if (0 == strcmp(rwGetFileName(rwIOS), "stdout")) {
        rwIOS->FD = stdout;
        rwIOS->closeFn = NULL;
    } else if ((fildes = _ioPathnameIsFiledesc(rwGetFileName(rwIOS))) != -1) {
        /* Handle "[%d]": Convert the descriptor to a FILE pointer;
         * set close function */
        rwIOS->FD = fdopen(fildes, "w");
        if (rwIOS->FD == NULL) {
            rwIOS->errobj.num = fildes;
            rwIOS->errnum = errno;
            rv = LIBRW_ERR_FDOPEN;
            goto END;
        }
        rwIOS->closeFn = fclose;
    } else if ((0 == strcmp(rwGetFileName(rwIOS), "/dev/null"))
               || _ioPathnameIsCompressed(rwGetFileName(rwIOS))
               || isFIFO(rwGetFileName(rwIOS)))
    {
        /* Handle /dev/null, FIFOs, and compressed files */
        int is_pipe;

        if (openFile(rwGetFileName(rwIOS), 1, &(rwIOS->FD), &is_pipe)) {
            rwIOS->errnum = errno;
            rv = LIBRW_ERR_OPEN;
            goto END;
        }
        if (is_pipe) {
            rwIOS->closeFn = &pclose;
        } else {
            rwIOS->closeFn = &fclose;
        }
    } else {
        int mode, flags, file_desc;

        mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
        flags = O_WRONLY | O_CREAT | O_EXCL;
        file_desc = open(rwGetFileName(rwIOS), flags, mode);
        if (file_desc == -1) {
            rwIOS->errnum = errno;
            rv = LIBRW_ERR_OPEN;
            goto END;
        }
        /* Convert the descriptor to a pointer; set close function */
        rwIOS->FD = fdopen(file_desc, "w");
        if (rwIOS->FD == NULL) {
            rwIOS->errobj.num = file_desc;
            rwIOS->errnum = errno;
            rv = LIBRW_ERR_FDOPEN;
            goto END;
        }
        rwIOS->closeFn = fclose;
    }

  END:
    return rv;
}


static int _ioPrepareRead(
    rwIOStruct     *rwIOS)
{
    int rv = LIBRW_OK;
    genericHeader *gHdr;

    assert(rwIOS);
    assert(rwGetHeader(rwIOS));

    gHdr = (genericHeader*)rwGetHeader(rwIOS);

    if (rwIOS->pvfsFlag) {
        /* Note: we allocate an rwFileHeaderV0 header here. */
        rv = _packedfileheaderCreateV0(rwIOS);
        if (rv) {
            goto END;
        }

    } else {
        /* read header */
        if (!fread(gHdr, sizeof(genericHeader), 1, rwIOS->FD)) {
            rwIOS->errnum = errno;
            rv = LIBRW_ERR_READ;
            goto END;
        }
    }

    /* verify header is RW file */
    if (CHECKMAGIC(gHdr)) {
        rv = LIBRW_ERR_BAD_MAGIC;
        goto END;
    }

    /* swap bytes? */
    rv = _ioSetSwapFlag(rwIOS);

    /* handle specific type */
    switch(gHdr->type) {
      case FT_RWFILTER:
        rv = _filterioPrepareRead(rwIOS);
        break;

      case FT_RWGENERIC:
        rv = _genericioPrepareRead(rwIOS);
        break;

      case FT_RWNOTROUTED:
        rv = _notroutedioPrepareRead(rwIOS);
        break;

      case FT_RWROUTED:
        rv = _routedioPrepareRead(rwIOS);
        break;

      case FT_RWSPLIT:
        rv = _splitioPrepareRead(rwIOS);
        break;

      case FT_RWWWW:
        rv = _wwwioPrepareRead(rwIOS);
        break;

      default:
        rv = LIBRW_ERR_UNSUPPORT_TYPE;
        break;
    }

    if (rv) { goto END; }

    if (rwIOS->io_mode == SK_RWIO_APPEND) {
        /* Since we called the prepare-read function, the
         * writeBodyRecordsFn should be NULL.  However, if it is not
         * NULL, we want to use it instead of the write-initial-record
         * function.  Using the write-initial-record function would
         * put a header in the middle of the file we are appending
         * data onto. */
        if (rwIOS->writeBodyRecordsFn != NULL) {
            rwIOS->rwWriteFn = rwIOS->writeBodyRecordsFn;
            rwIOS->writeBodyRecordsFn = NULL;
        }

        /* go to the end of the file */
        if (-1 == fseek(rwIOS->FD, 0, SEEK_END)) {
            rwIOS->errnum = errno;
            rv = LIBRW_ERR_FSEEK;
            goto END;
        }
    }

  END:
    return rv;
}


static int _ioPrepareWrite(
    rwIOStruct     *rwIOS)
{
    int rv = LIBRW_OK;

    /* Verify type and version match */
    switch (rwGetFileType(rwIOS)) {
      case FT_RWFILTER:
        rv = _filterioPrepareWrite(rwIOS);
        break;

      case FT_RWGENERIC:
        rv = _genericioPrepareWrite(rwIOS);
        break;

      case FT_RWNOTROUTED:
        rv = _notroutedioPrepareWrite(rwIOS);
        break;

      case FT_RWROUTED:
        rv = _routedioPrepareWrite(rwIOS);
        break;

      case FT_RWSPLIT:
        rv = _splitioPrepareWrite(rwIOS);
        break;

      case FT_RWWWW:
        rv = _wwwioPrepareWrite(rwIOS);
        break;

      default:
        rv = LIBRW_ERR_UNSUPPORT_TYPE;
        break;
    }

    return rv;
}




int rwioCreate(
    rwIOStruct    **rwIOS,
    const char     *pathname,
    rwio_mode_t     read_write_append)
{
    int rv = LIBRW_OK;
    genericHeader *gHdr;

    if (rwIOS == NULL) {
        rv = LIBRW_ERR_NULL_AUGUMENT;
        goto END;
    }
    *rwIOS = NULL;

    if (read_write_append == SK_RWIO_CLOSED) {
        rv = LIBRW_ERR_NULL_AUGUMENT;
        goto END;
    }

    /* Allocate the rwIOStruct */
    *rwIOS = calloc(1, sizeof(rwIOStruct));
    if (NULL == *rwIOS) {
        rv = LIBRW_ERR_ALLOC;
        goto END;
    }

    /* store the parameters on the structure */
    (*rwIOS)->io_mode = read_write_append;

    /* file name */
    (*rwIOS)->fPath = strdup(pathname);
    if (NULL == (*rwIOS)->fPath) {
        rv = LIBRW_ERR_ALLOC;
        goto END;
    }

    /* Allocate a genericHeader; the header may be re-allocated
     * later once we know the file type. */
    (*rwIOS)->hdrLen = sizeof(genericHeader);
    gHdr = calloc(1, (*rwIOS)->hdrLen);
    if (NULL == gHdr) {
        rv = LIBRW_ERR_ALLOC;
        goto END;
    }
    (*rwIOS)->hdr = gHdr;

    /* Set to a generic file, default version, native byte order */
    PREPHEADER(gHdr);
#if IS_LITTLE_ENDIAN
    gHdr->isBigEndian = 0;
#else
    gHdr->isBigEndian = 1;
#endif
    gHdr->type = FT_RWGENERIC;
    gHdr->version = SK_FILE_VERSION_ANY;
    gHdr->cLevel = 0;

    /* Native format by default, so don't swap */
    (*rwIOS)->swapFlag = 0;

    /* No nodes. */
    (*rwIOS)->node = -1;

    /* set sensor ID and flow type from the file name, but don't
     * complain on error.  Maybe someday these will be stored in the
     * header of the file so we don't depend on the filename. */
    (*rwIOS)->sID = getSIDFromFName(pathname, 1);
    (*rwIOS)->flow_type = _readFilenameToFlowType(pathname);

    rv = _ioCheckPathName(pathname, read_write_append);

  END:
    return rv;
}


int rwioSetFileType(
    rwIOStruct     *rwIOS,
    uint8_t         file_type)
{
    int rv;

    /* verify librw understands the file_type */
    switch (file_type) {
      case FT_RWFILTER:
      case FT_RWGENERIC:
      case FT_RWNOTROUTED:
      case FT_RWROUTED:
      case FT_RWSPLIT:
      case FT_RWWWW:
        break;
      default:
        rv = LIBRW_ERR_UNSUPPORT_TYPE;
        goto END;
    }

    /* can we actually modify the file type? */
    rv = _ioIsModifiable(rwIOS);
    if (rv) { goto END; }

    /* Do not support changing the type once we've grown the header,
     * since it may mean re-allocating.... */
    if (rwIOS->hdrLen != sizeof(genericHeader)) {
        rv = LIBRW_ERR_ATTRIBUTE_FIXED;
        goto END;
    }

    rv = _ioSetFileType(rwIOS, file_type);

  END:
    return rv;
}


int rwioSetFileVersion(
    rwIOStruct     *rwIOS,
    fileVersion_t   file_version)
{
    int rv;

    /* No check of the version until we try to open the file */

    rv = _ioIsModifiable(rwIOS);
    if (rv) { goto END; }

    rv = _ioSetFileVersion(rwIOS, file_version);
    if (rv) { goto END; }

  END:
    return rv;
}


int rwioSetFileByteorder(
    rwIOStruct     *rwIOS,
    silk_endian_t   byte_order)
{
    int rv;

    rv = _ioIsModifiable(rwIOS);
    if (rv) { goto END; }

    rv = _ioSetFileByteorder(rwIOS, byte_order);
    if (rv) { goto END; }

  END:
    return rv;
}


int rwioSetFileSTime(
    rwIOStruct     *rwIOS,
    uint32_t        start_time,
    int             use_floor_hour)
{
    int rv = LIBRW_OK;

    rv = _ioHeaderWritten(rwIOS);
    if (rv) {
        goto END;
    }

    switch (rwGetFileType(rwIOS)) {
      case FT_RWNOTROUTED:
      case FT_RWROUTED:
      case FT_RWSPLIT:
      case FT_RWWWW:
        rv = _packedfileheaderSetSTime(rwIOS, start_time, use_floor_hour);
        break;

      case FT_RWFILTER:
      case FT_RWGENERIC:
      default:
        rv = LIBRW_ERR_UNSUPPORT_TYPE;
        break;
    }

  END:
    return rv;
}


int rwioAppendToHistoryFromFile(
    rwIOStruct         *rwIOS,
    const rwIOStruct   *source_rwIOS)
{
    int rv = LIBRW_OK;

    rv = _ioHeaderWritten(rwIOS);
    if (rv) { goto END; }

    if (rwGetFileType(rwIOS) != FT_RWFILTER) {
        rv = LIBRW_ERR_UNSUPPORT_TYPE;
        goto END;
    }

    if (rwGetFileType(source_rwIOS) != FT_RWFILTER) {
        rv = LIBRW_ERR_UNSUPPORT_TYPE;
        goto END;
    }

    rv = _headersAppendFromFile(rwIOS, source_rwIOS);

  END:
    return rv;
}


int rwioAppendToHistoryArgv(
    rwIOStruct     *rwIOS,
    int             argc,
    char * const   *argv)
{
    int rv = LIBRW_OK;

    rv = _ioHeaderWritten(rwIOS);
    if (rv) { goto END; }

    if (rwGetFileType(rwIOS) != FT_RWFILTER) {
        rv = LIBRW_ERR_UNSUPPORT_TYPE;
        goto END;
    }

    rv = _headersAppendArgv(rwIOS, argc, argv);

  END:
    return rv;
}


int rwioOpen(
    rwIOStruct     *rwIOS)
{
    int rv;

    rv = _ioIsNotOpen(rwIOS);
    if (rv) { goto END; }

    switch (rwIOS->io_mode) {
      case SK_RWIO_CLOSED:
        rv = LIBRW_ERR_STREAM_CLOSED;
        break;

      case SK_RWIO_WRITE:
        rv = _ioPrepareWrite(rwIOS);
        if (rv) { goto END; }
        rv = _ioOpenWrite(rwIOS);
        if (rv) { goto END; }
        break;

      case SK_RWIO_READ:
        rv = _ioOpenRead(rwIOS);
        if (rv) { goto END; }
        rv = _ioPrepareRead(rwIOS);
        if (rv) { goto END; }
        break;

      case SK_RWIO_APPEND:
        rv = _ioOpenAppend(rwIOS);
        if (rv) { goto END; }
        rv = _ioPrepareRead(rwIOS);
        if (rv) { goto END; }
        break;
    }

  END:
    return rv;
}


int rwioFDOpen(
    rwIOStruct     *rwIOS,
    int             file_desc)
{
    int rv;
    char *mode = NULL;

    rv = _ioIsNotOpen(rwIOS);
    if (rv) { goto END; }

    /* pre-open checks and set mode */
    switch (rwIOS->io_mode) {
      case SK_RWIO_CLOSED:
        rv = LIBRW_ERR_STREAM_CLOSED;
        break;
      case SK_RWIO_WRITE:
        rv = _ioPrepareWrite(rwIOS);
        mode = "w";
        break;
      case SK_RWIO_READ:
        mode = "r";
        break;
      case SK_RWIO_APPEND:
        mode = "r+";
        break;
    }
    if (rv) { goto END; }

    /* Convert the descriptor to a pointer; set close function */
    rwIOS->FD = fdopen(file_desc, mode);
    if (rwIOS->FD == NULL) {
        rwIOS->errobj.num = file_desc;
        rwIOS->errnum = errno;
        rv = LIBRW_ERR_FDOPEN;
        goto END;
    }
    rwIOS->closeFn = fclose;

    /* post-open checks */
    switch (rwIOS->io_mode) {
      case SK_RWIO_CLOSED:
      case SK_RWIO_WRITE:
        /* no checks */
        break;
      case SK_RWIO_READ:
      case SK_RWIO_APPEND:
        rv = _ioPrepareRead(rwIOS);
        break;
    }

  END:
    return rv;
}


int rwioMakeTemp(
    rwIOStruct     *rwIOS)
{
    int rv;
    int fd;

    rv = _ioIsNotOpen(rwIOS);
    if (rv) { goto END; }

    /* Temp files only make sense for writing */
    if (rwIOS->io_mode != SK_RWIO_WRITE) {
        rv = LIBRW_ERR_UNSUPPORT_STREAM;
        goto END;
    }

    rv = _ioPrepareWrite(rwIOS);
    if (rv) { goto END; }

    /* open file */
    fd = mkstemp(rwGetFileName(rwIOS));
    if (fd == -1) {
        rwIOS->errnum = errno;
        rv = LIBRW_ERR_MKSTEMP;
        goto END;
    }
    rwIOS->FD = fdopen(fd, "w");
    if (rwIOS->FD == NULL)
    {
        rwIOS->errobj.num = fd;
        rwIOS->errnum = errno;
        rv = LIBRW_ERR_FDOPEN;
        goto END;
    }

    rwIOS->closeFn = fclose;

  END:
    return rv;
}


int rwioLockFile(
    rwIOStruct    *rwIOS)
{
    int rv = LIBRW_OK;

    if (rwIOS == NULL) {
        rv = LIBRW_ERR_NULL_AUGUMENT;
        goto END;
    }

    if (NULL == rwIOS->FD) {
        rv = LIBRW_ERR_STREAM_NOT_OPEN;
        goto END;
    }

    if (rwIOS->io_mode == SK_RWIO_CLOSED) {
        rv = LIBRW_ERR_STREAM_CLOSED;
    }

    /* Don't try to lock anything that is not a real file */
    if (rwIOS->closeFn != &fclose) {
        goto END;
    }

    if (rwIOS->io_mode == SK_RWIO_READ) {
        rv = getRLockFILE_Wait(rwIOS->FD);
        if (rv != 0) {
            rwIOS->errnum = errno;
            rv = LIBRW_ERR_RLOCK;
        }
    } else {
        rv = getWLockFILE_Wait(rwIOS->FD);
        if (rv != 0) {
            rwIOS->errnum = errno;
            rv = LIBRW_ERR_WLOCK;
        }
    }

  END:
    return rv;
}


int rwioMakeDirectory(
    rwIOStruct     *rwIOS)
{
    int rv;
    char *cp;

    rv = _ioIsNotOpen(rwIOS);
    if (rv) { goto END; }

    if (rwIOS->io_mode != SK_RWIO_WRITE) {
        rv = LIBRW_ERR_UNSUPPORT_STREAM;
        goto END;
    }

    /* Create the directory to the file if the directory does not
     * exist.  Check for a '/' in the name, but make certain it
     * isn't the first '/' in the file path (e.g., "/foo") */
    cp = strrchr(rwIOS->fPath, '/');
    if ((cp != NULL) && (cp != rwIOS->fPath)) {
        *cp = '\0';
        if ( !dirExists(rwIOS->fPath)) {
            rv = mkDirPath(rwIOS->fPath);
            *cp = '/';
            if (rv) {
                rwIOS->errnum = errno;
                rv = LIBRW_ERR_MKDIRPATH;
                goto END;
            }
        }
        *cp = '/';
    }

  END:
    return rv;
}


int rwioWriteHeader(
    rwIOStruct     *rwIOS)
{
    int rv = LIBRW_OK;

    rv = _ioHeaderWritten(rwIOS);
    if (rv) { goto END; }

    if (NULL == rwIOS->FD) {
        rv = LIBRW_ERR_STREAM_NOT_OPEN;
        goto END;
    }

    /* To write the header, call the rwWrite function---which should
     * be a function that writes the header and an initial
     * record---with a NULL record. */
    rv = rwWrite(rwIOS, NULL);

    /* That function should have set this to NULL. */
    assert(rwIOS->writeBodyRecordsFn == NULL);

  END:
    return rv;
}


int rwioClose(
    rwIOStruct     *rwIOS)
{
    int rv = LIBRW_OK;

    if (rwIOS == NULL) {
        rv = LIBRW_ERR_NULL_AUGUMENT;
        goto END;
    }

    if (rwIOS->bufFlag) {
        /* TODO: handle close functions */
        if ( rwIOS->bufFD ) {
            close_buffered(rwIOS->bufFD);
        }
        rwIOS->bufFD = NULL;
    } else if (rwIOS->FD) {
        if (rwIOS->FD != stdin && rwIOS->closeFn) {
            rv = rwIOS->closeFn(rwIOS->FD);
        }
        rwIOS->FD = (FILE *)NULL;
    }

    /* Mark the rwIOS as closed and unable to be opened. */
    rwIOS->io_mode = SK_RWIO_CLOSED;

  END:
    return rv;
}


int rwioUnbindFile(
    FILE          **out_filep,
    rwIOStruct     *rwIOS)
{
    if (rwIOS == NULL) {
        return LIBRW_ERR_NULL_AUGUMENT;
    }

    *out_filep = rwIOS->FD;
    rwIOS->FD = NULL;
    rwIOS->closeFn = NULL;
    return LIBRW_OK;
}


int rwioDestroy(
    rwIOStruct    **rwIOS)
{
    int rv = LIBRW_OK;

    if (rwIOS == NULL || *rwIOS == NULL) {
        goto END;
    }

    if ((*rwIOS)->FD != NULL) {
        rv = rwioClose(*rwIOS);
        (*rwIOS)->FD = NULL;
    }
    if ((*rwIOS)->hdr) {
        free((*rwIOS)->hdr);
        (*rwIOS)->hdr = NULL;
    }
    if ((*rwIOS)->fPath) {
        free((*rwIOS)->fPath);
        (*rwIOS)->fPath = NULL;
    }
    free(*rwIOS);
    *rwIOS = NULL;

  END:
    return rv;
}


void rwioPrintLastErr(
    rwIOStruct         *rwIOS,
    librwErrorCodes_t   error_code,
    sk_msg_fn_t         errfn)
{
    const char *msg;
    char tstamp[256];

#ifdef TEST_PRINTF_FORMATS
#define errfn printf
#else
    if (errfn == NULL) {
        errfn = &skAppPrintErr;
    }
#endif

    switch (error_code) {
      case LIBRW_OK:
        errfn("Command completed successfully");
        break;

      case LIBRW_ERR_STIME_UNDRFLO:
        strcpy(tstamp, timestamp(rwIOS->errobj.rec->sTime));
        msg = "Record's start time less than that allowed in file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s < %s", msg,
                  rwGetFileName(rwIOS), tstamp,
                  timestamp(rwGetFileSTime(rwIOS)));
        }
        break;

      case LIBRW_ERR_STIME_OVRFLO:
        strcpy(tstamp, timestamp(rwIOS->errobj.rec->sTime));
        msg = "Record's start time greater than that allowed in file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s > %s", msg,
                  rwGetFileName(rwIOS), tstamp,
                  timestamp(MAX_START_TIME+rwGetFileSTime(rwIOS)));
        }
        break;

      case LIBRW_ERR_ELPSD_OVRFLO:
        msg = "Record's duration greater than that allowed in file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %d > %d", msg,
                  rwGetFileName(rwIOS), rwIOS->errobj.rec->elapsed,
                  MAX_ELAPSED_TIME);
        }
        break;

      case LIBRW_ERR_PKTS_OVRFLO:
        msg = "Record's packet count greater than that allowed in file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %d > %d", msg,
                  rwGetFileName(rwIOS), rwIOS->errobj.rec->pkts,
                  (MAX_PKTS * PKTS_DIVISOR));
        }
        break;

      case LIBRW_ERR_PKTS_ZERO:
        msg = "Record's packet count is zero while writting to file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s'", msg,
                  rwGetFileName(rwIOS));
        }
        break;

      case LIBRW_ERR_SNMP_OVRFLO:
        msg = "Record's SNMP index greater than that allowed in file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else if (rwIOS->errobj.rec->input > 255) {
            errfn("%s '%s': input %d > 255", msg,
                  rwGetFileName(rwIOS), rwIOS->errobj.rec->input);
        } else {
            errfn("%s '%s': output %d > 255", msg,
                  rwGetFileName(rwIOS), rwIOS->errobj.rec->output);
        }
        break;

      case LIBRW_ERR_SENSORID_OVRFLO:
        msg = "Record's Sensor ID greater than that allowed in file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %d > 255", msg,
                  rwGetFileName(rwIOS), rwIOS->errobj.rec->sID);
        }
        break;

      case LIBRW_ERR_PROTO_MISMATCH:
        msg = "Record's IP-protocol is not supported in file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %d > 255", msg,
                  rwGetFileName(rwIOS), rwIOS->errobj.rec->proto);
        }
        break;

      case LIBRW_ERR_WRITE:
        msg = "Cannot write to file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s", msg,
                  rwGetFileName(rwIOS), strerror(rwIOS->errnum));
        }
        break;

      case LIBRW_ERR_NULL_AUGUMENT:
        errfn("Unexpected NULL or empty value");
        break;

      case LIBRW_ERR_ALLOC:
        errfn("Memory allocation failed");
        break;

      case LIBRW_ERR_UNSUPPORT_TYPE:
        msg = "Unsupported operation on file's type";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s' type %s (0x%02x)", msg,
                  rwGetFileName(rwIOS),
                  rwGetFileOutputFormatName(rwGetFileType(rwIOS)),
                  rwGetFileType(rwIOS));
        }
        break;

      case LIBRW_ERR_UNSUPPORT_VERSION:
        msg = "This SiLK release does not support the type and version of";
        if (!rwIOS) {
            errfn("%s file", msg);
        } else {
            errfn("%s\n\t file '%s': %s(0x%02x) v%u", msg,
                  rwGetFileName(rwIOS),
                  rwGetFileOutputFormatName(rwGetFileType(rwIOS)),
                  rwGetFileType(rwIOS), rwGetFileVersion(rwIOS));
        }
        break;

      case LIBRW_ERR_UNSUPPORT_STREAM:
        msg = "Operation is not permitted on the stream/file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s'", msg,
                  rwGetFileName(rwIOS));
        }
        break;

      case LIBRW_ERR_ATTRIBUTE_FIXED:
        msg = "Cannot change attributes of this stream/file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s'", msg,
                  rwGetFileName(rwIOS));
        }
        break;

      case LIBRW_ERR_STREAM_OPEN:
        msg = "Operation not permitted on previously opened file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s'", msg,
                  rwGetFileName(rwIOS));
        }
        break;

      case LIBRW_ERR_STREAM_NOT_OPEN:
        msg = "Operation not permitted on unopened file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s'", msg,
                  rwGetFileName(rwIOS));
        }
        break;

      case LIBRW_ERR_STREAM_CLOSED:
        msg = "Operation not permitted on closed stream";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s'", msg,
                  rwGetFileName(rwIOS));
        }
        break;

      case LIBRW_ERR_OPEN:
        msg = "Cannot open file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s", msg,
              rwGetFileName(rwIOS), strerror(rwIOS->errnum));
        }
        break;

      case LIBRW_ERR_FDOPEN:
        msg = "Cannot open file descriptor for file as a stream";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s desc %d, file '%s': %s", msg,
                  rwIOS->errobj.num, rwGetFileName(rwIOS),
                  strerror(rwIOS->errnum));
        }
        break;

      case LIBRW_ERR_MKSTEMP:
        msg = "Cannot create temporary file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s", msg,
                  rwGetFileName(rwIOS), strerror(rwIOS->errnum));
        }
        break;

      case LIBRW_ERR_READ:
        msg = "Cannot read from file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s", msg,
                  rwGetFileName(rwIOS), strerror(rwIOS->errnum));
        }
        break;

      case LIBRW_ERR_WRITE_FILE_EXISTS:
        msg = "Cannot create new file over existing file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s'", msg,
                  rwGetFileName(rwIOS));
        }
        break;

      case LIBRW_ERR_MKDIRPATH:
        msg = "Cannot create directory to hold file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s", msg,
                  rwGetFileName(rwIOS), strerror(rwIOS->errnum));
        }
        break;

      case LIBRW_ERR_OPEN_NOEXIST:
        msg = "Cannot open non-existent file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s'", msg,
                  rwGetFileName(rwIOS));
        }
        break;

      case LIBRW_ERR_FSEEK:
        msg = "Cannot seek in file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s", msg,
                  rwGetFileName(rwIOS), strerror(rwIOS->errnum));
        }
        break;

      case LIBRW_ERR_UNSUPPORT_COMPRESS:
        msg = "Operation not permitted on compressed file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s'", msg,
                  rwGetFileName(rwIOS));
        }
        break;

      case LIBRW_ERR_BAD_MAGIC:
        msg = "File does not appear to be a SiLK data file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s: '%s'", msg,
                  rwGetFileName(rwIOS));
        }
        break;

      case LIBRW_ERR_MISMATCH_TYPE:
        msg = "Specified type does not match that of file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s (0x%02x)", msg,
                  rwGetFileName(rwIOS),
                  rwGetFileOutputFormatName(rwGetFileType(rwIOS)),
                  rwGetFileType(rwIOS));
        }
        break;

      case LIBRW_ERR_MISMATCH_VERSION:
        msg = "Specified version does not match that of file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %d", msg,
                  rwGetFileName(rwIOS), rwGetFileVersion(rwIOS));
        }
        break;

      case LIBRW_ERR_MISMATCH_ENDIAN:
        msg = "Specified byte order does not match that of file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s", msg,
                  rwGetFileName(rwIOS),
                  ((rwGetIsBigEndian(rwIOS) ? "Big" : "Little")));
        }
        break;

      case LIBRW_ERR_MISMATCH_STARTIME:
        msg = "Specified start time does not match that of file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s", msg,
                  rwGetFileName(rwIOS), timestamp(rwGetFileSTime(rwIOS)));
        }
        break;

      case LIBRW_ERR_WLOCK:
        msg = "Cannot get a write lock on file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s", msg,
                  rwGetFileName(rwIOS), strerror(rwIOS->errnum));
        }
        break;

      case LIBRW_ERR_RLOCK:
        msg = "Cannot get a read lock on file";
        if (!rwIOS) {
            errfn("%s", msg);
        } else {
            errfn("%s '%s': %s", msg,
                  rwGetFileName(rwIOS), strerror(rwIOS->errnum));
        }
        break;

    } /* switch */
}
