/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwread.c
**
**  library to open and read various raw file types.
**  also handle dumping to copyInputFD when required.
**
**  When dumping to copyInputFD, for all except rwfilter, write a
**  generic header and then write each record in rwRec format (i.e.,
**  already processed).  For rwfilter, write the filterheader (not
**  just the generic header) and write the rwFilterRec record (i.e.,
**  pre-processed).
**
**  Suresh Konda
*/

#define DECLARE_SITE_VARIABLES 1
#define DECLARE_FILEFORMAT_VARIABLES 1
#include "silk.h"

RCSIDENT("$SiLK: /src/rw/librw/rwread.c,v 1.44 2005/09/19 10:08:56 thomasm Exp $");

#include "librw_priv.h"
/* Get sensor id definitions and other file-related information */
#include "silk_site.h"


/* external variables */
struct profile prof1;


/* static variables */

/* set to 0 after the first file is opened in rwOpenFile() */
static uint32_t fileCount = 0;

/* number of output file formats */
static const uint32_t file_format_count = 
    ((sizeof(fileOutputFormats)/sizeof(char*)) - 1);



/*
 *  _readGetFilenameFlowType(filename);
 *
 *    Return the flowtype_t of the file based on the the prefix of
 *    'filename'.  Returns SILK_INVALID_FLOW_TYPE if the file does not
 *    have a prefix or if the prefix is not recognized.
 */
int _readFilenameToFlowType(const char *filename)
{
    const char *s;
    const char *n;
    unsigned int i;
    int len;

    /* Set 's' to the basename of the filename */
    s = strrchr(filename, '/');
    if (s) {
        s++;
    } else {
        s = filename;
    }

    /* find the prefix */
    n = strchr(s, '-');
    if (n == NULL) {
        return -1;
    }
    len = n - s;             /* Length of prefix */
    for (i = 0; i < numFileTypes; ++i) {
        if ((strncmp(outFInfo[i].filePrefix, s, len) == 0)
            && (outFInfo[i].filePrefix[len] == '\0'))
        {
            return i;
        }
    }

    /* not found */
    return SILK_INVALID_FLOW_TYPE;
}


uint8_t _readFilenameToFileFormat(const char *filename)
{
    int flow_type = _readFilenameToFlowType(filename);
    if (flow_type == SILK_INVALID_FLOW_TYPE) {
        /* any file format that librw doesn't understand will cause
         * librw to bow out. */
        return 0;
    }
    return (uint8_t)(outFInfo[flow_type].packType & 0xFF);
}


const char *rwGetFileOutputFormatName(uint32_t file_output_format)
{
    if (file_output_format > file_format_count) {
        return "UNKNOWN";
    }
     
    return fileOutputFormats[file_output_format];
}


uint32_t rwGetFileOutputFormatCount(void)
{
    return file_format_count;
}


/*
 *  Close file and return null.
 */
rwIOStruct *_errorReturn(rwIOStruct *rwIOS)
{
    rwCloseFile(rwIOS);
    if (fileCount > 0) {
        fileCount--;
    }
    return (rwIOStruct*) NULL;
}


/*
 *  Close the file and free all memory.
 */
int rwCloseFile(rwIOStruct *rwIOS)
{
    int rv = LIBRW_OK;

    if (rwIOS->bufFlag) {
        /* TODO: handle close functions */
        if ( rwIOS->bufFD ) {
            close_buffered(rwIOS->bufFD);
        }
        rwIOS->bufFD = NULL;
    } else if (rwIOS->FD) {
        if (rwIOS->FD != stdin && rwIOS->closeFn) {
            rv = rwIOS->closeFn(rwIOS->FD);
            if (rv) {
                skAppPrintErr("Error closing '%s': %s",
                              (rwIOS->fPath ? rwIOS->fPath : "<file>"),
                              strerror(errno));
            }
        }
        rwIOS->FD = (FILE *)NULL;
    }

    if (rwIOS->hdr) {
        free(rwIOS->hdr);
        rwIOS->hdr = NULL;
    }
    if (rwIOS->fPath) {
        free(rwIOS->fPath);
        rwIOS->fPath = NULL;
    }
    free(rwIOS);

    return rv;
}


/*
 *  Opens a file for reading.  See rwpack.h for details.
 */
rwIOStruct *rwOpenFile(const char *fPath, rwIOStruct *copyInput)
{
    rwIOStruct *rwIOS;
    int rv = LIBRW_OK;

    rv = rwioCreate(&rwIOS, fPath, SK_RWIO_READ);
    if (rv) { goto END; }

    rv = rwioOpen(rwIOS);
    if (rv) { goto END; }

    rwIOS->copyInputFD = copyInput;

  END:
    if (rv) {
        rwioPrintLastErr(rwIOS, rv, NULL);
        if (rwIOS) {
            rwioDestroy(&rwIOS);
        }
    }
    return rwIOS;
}
/* rwOpenFile */




/*
 *  Read a record into 'buf'.  Works on cluster or non-cluster.  See
 *  librw_priv.h for details.
 */
int _readReadRecord(void *buf, rwIOStruct *rwIOS)
{
    size_t count = 0;

    if (rwIOS->eofFlag) {
        return 1;
    }

    if (rwIOS->bufFlag) {
        count = read_buffered( rwIOS->bufFD, buf, rwIOS->recLen );
    } else {
        count = fread(buf, 1, rwIOS->recLen, rwIOS->FD);
    }

    if ( count != rwIOS->recLen) {
        /* EOF */
        rwIOS->eofFlag = 1;
        if ( rwIOS->bufFlag) {
            /* a read of 0 means a clean record boundry */
            if ( count != 0 ) {
                skAppPrintErr("Short read of %lu on %s",
                              (unsigned long)count, rwIOS->fPath);
            }
        } else {
            if (!feof(rwIOS->FD)) {
                /* some error */
                skAppPrintErr("Short read on %s", rwIOS->fPath);
            }
        }
        return 1;
    }

    return 0;
}
/* _readReadRecord */


/*
 * librwGetRecordSize(genericHeaderPtr gHdrPtr);
 *
 * Given a pointer to a genericHeader, return the size of the records
 * in the file.  Return 0 if the file does not contain fixed-size
 * records or if the information cannot be determined from the
 * genericHeader.
 */
size_t librwGetRecordSize(genericHeader *gHdr)
{
    switch (gHdr->type) {
      case FT_RWFILTER:
        return _filterioGetRecLen(gHdr->version);
      case FT_RWGENERIC:
        return _genericioGetRecLen(gHdr->version);
      case FT_RWNOTROUTED:
        return _notroutedioGetRecLen(gHdr->version);
      case FT_RWROUTED:
        return _routedioGetRecLen(gHdr->version);
      case FT_RWSPLIT:
        return _splitioGetRecLen(gHdr->version);
      case FT_RWWWW:
        return _wwwioGetRecLen(gHdr->version);
      case FT_RWACL:
      default:
        return 0u;
    }

    return 0u; /*NOTREACHED*/
}
/* librwGetRecordSize */


/*
 * librwComputeHeaderPadding(rwFileHeaderV0 *hdr);
 *
 * Compute the bytes of padding required to make the complete header a
 * multiple of the record size.  Return -1 if information cannot be
 * determined from the genericHeader.
 */
int librwComputeHeaderPadding(genericHeaderPtr gHdrPtr)
{
    int rec_size;
    int pad_length = 0;

    if (gHdrPtr->version < 2) {
        return 0;
    }

    switch (gHdrPtr->type) {
      case FT_RWACL:
      case FT_RWNOTROUTED:
      case FT_RWROUTED:
      case FT_RWSPLIT:
      case FT_RWWWW:
        /* These all have rwFileHeaderV0 headers */
        rec_size = librwGetRecordSize(gHdrPtr);
        pad_length = rec_size - (sizeof(rwFileHeaderV0) % rec_size);
        if (pad_length == rec_size) {
            pad_length = 0;
        }
        return pad_length;

      case FT_RWFILTER:
        /* Cannot be determined from just the genericHeader */
        return -1;

      case FT_RWGENERIC:
        /* This only has the genericHeader.  As of 2004.04.30, nothing is
         * writing version 2 RWGENERIC files */
        return 0;

      default:
        return -1;
    }

    return -1; /*NOTREACHED*/
}
/* librwComputeHeaderPadding */


/*  Fill in the bytes and packets fields in rwRP by expanding the
 *  values that were read from disk.  See librw_priv.h for details. */
void _readUnpackBytesPackets(
    rwGenericRec_V3    *rwRP,
    uint32_t            bpp,
    uint32_t            pkts,
    uint32_t            pflag)
{
    uint32_t bytes;
    int bPPkt, bPPFrac;
    div_t i;

    if (pflag) {
        pkts *= PKTS_DIVISOR;
    }

    /* Unpack the bpp value:  bPPkt:14; bPPFrac:6; */
    bPPkt = GET_MASKED_BITS(bpp, 6, 14);
    bPPFrac = GET_MASKED_BITS(bpp, 0, 6);

    /* convert fraction to whole number */
    i = div((bPPFrac * pkts), BPP_PRECN);

    bytes = ((bPPkt * pkts) + i.quot + ((i.rem >= BPP_PRECN_DIV_2) ? 1 : 0));

    rwRP->pkts = pkts;
    rwRP->bytes = bytes;
    rwRP->bpp = bpp;
}


void _readUnpackProtoFlags(
    rwGenericRec_V3    *rwRP,
    uint8_t             is_tcp,
    uint8_t             prot_flags,
    uint8_t             tcp_state,
    uint8_t             rest_flags)
{
    /* For some record types (e.g., RWWWW), proto is fixed at 6(TCP)
     * and there may be another value in the 'is_tcp' bit; ignore the
     * 'is_tcp' bit if the protocol is already set to TCP. */
    if ((rwRP->proto == 6) || (is_tcp == 1)) {
        /* Flow is TCP */
        rwRP->proto = 6;

        if (tcp_state == 0) {
            /* We don't have additional TCP info; 'prot_flags' holds
             * the flags. */
            rwRP->flags = prot_flags;
        } else {
            /* We have additional flow information; value in
             * prot_flags are the flags on the first packet. */
            rwRP->tcp_state  = tcp_state;
            rwRP->init_flags = prot_flags;
            rwRP->rest_flags = rest_flags;
            rwRP->flags      = (prot_flags | rest_flags);
        }
    } else {
        /* Flow is not TCP so there can be no additional TCP info.
         * 'prot_flags' holds the protocol.  Although 'flags' has no
         * real meaning here, the 'rest_flags' value has the value
         * that we got from the flow collector, so set 'rwRP->flags'
         * to that value. */
        rwRP->proto = prot_flags;
        rwRP->flags = rest_flags;
    }
}


/* Set values in rwRP by expanding the 'sbb' and 'pef' fields that
 * exist in the packed file formats.  See librw_priv.h for details. */
void _readUnpackSbbPef(
    rwGenericRec_V3    *rwRP,
    uint32_t            file_start_time,
    const uint32_t     *sbb,
    const uint32_t     *pef)
{
    uint32_t pkts, pktsFlag, bpp, start_time;

    /* pef: uint32_t pkts:20; uint32_t elapsed :11; uint32_t pktsFlag:1; */
    pkts = *pef >> 12;
    rwRP->elapsed = (*pef >> 1) & MASKARRAY_11;
    pktsFlag = *pef & MASKARRAY_01;

    /* sbb: uint32_t start_time:12; uint32_t bpp:20 */
    bpp = *sbb & MASKARRAY_20;
    start_time = *sbb >> 20;
    rwRP->sTime = file_start_time + start_time;

    _readUnpackBytesPackets(rwRP, bpp, pkts, pktsFlag);
}


void _readUnpackTimeBytesPktsFlags(
    rwGenericRec_V3    *rwRP,
    uint32_t            file_start_time,
    const uint32_t     *pkts_stime,
    const uint32_t     *bbe,
    const uint32_t     *msec_flags)
{
    uint32_t pkts, bpp, is_tcp, pflag;
    uint8_t prot_flags;

    /* pkts_stime: pkts:20; sTime: 12; */
    pkts = GET_MASKED_BITS(*pkts_stime, 12, 20);
    rwRP->sTime = file_start_time + GET_MASKED_BITS(*pkts_stime, 0, 12);

    /* bbe: bpp: 20; elapsed: 12 */
    bpp = GET_MASKED_BITS(*bbe, 12, 20);
    rwRP->elapsed = GET_MASKED_BITS(*bbe, 0, 12);

    /* msec_flags: sTime_msec:10; elaps_msec:10; pflag:1;
     *             is_tcp:1; pad:2, prot_flags:8;          */
    rwRP->sTime_msec = GET_MASKED_BITS(*msec_flags, 22, 10);
    rwRP->elapsed_msec = GET_MASKED_BITS(*msec_flags, 12, 10);
    pflag = GET_MASKED_BITS(*msec_flags, 11, 1);
    is_tcp = GET_MASKED_BITS(*msec_flags, 10, 1);
    prot_flags = (uint8_t)GET_MASKED_BITS(*msec_flags, 0, 8);

    if (rwRP->proto == 6) {
        /* caller has forced record to be TCP */
        rwRP->flags = prot_flags;
    } else if (is_tcp == 0) {
        /* flow is not TCP */
        rwRP->proto = prot_flags;
    } else {
        /* flow is TCP */
        rwRP->proto = 6;
        rwRP->flags = prot_flags;
    }

    /* unpack the bpp value into bytes and packets */
    _readUnpackBytesPackets(rwRP, bpp, pkts, pflag);
}


/* Move 'skip_count' records ahead.  See rwpack.h for details. */
int _readSkipRecords(
    rwIOStruct     *rwIOS,
    int             skip_count)
{
    uint8_t ar[SK_MAX_RECORD_SIZE];
    int i;

    for (i = 0; i < skip_count; i++) {
        if (_readReadRecord(ar, rwIOS)) {
            return 1; /* failed */
        }
    }

    return 0;
}
