/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
 * rwrtd2split
 *
 *  Read an input file in RWROUTED format and write the records to a
 *  new file in RWSPLIT format.
 */

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/librw/rwrtd2split.c,v 1.18 2005/09/16 21:54:42 thomasm Exp $");

#include "rwpack.h"
#include "utils.h"

/* LOCAL FUNCTIONS */

#define NUM_REQUIRED_ARGS 2

static void appUsage(void);             /* never returns */
static void appSetup(int argc, char **argv);

/* LOCAL VARIABLES */

static char *in_fpath;
static char *out_fpath;
static rwIOStruct *in_ios;
static rwIOStruct *out_ios;


/*
 * appUsage:
 *      print usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsage(void)
{
    fprintf(stderr, ("%s <inFile> <outFile> \n"
                     "  Write the <outFile> in RWSPLIT format given the\n"
                     "  <inFile> in RWROUTED format.\n"),
            skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    silk_endian_t endian;
    int rv;

    skAppRegister(argv[0]);

    if ((argc - 1) < NUM_REQUIRED_ARGS) {
        appUsage();         /* never returns */
    }

    in_fpath = argv[1];
    out_fpath = argv[2];

    /* Open input file; check its type */
    in_ios = rwOpenFile(in_fpath, NULL);
    if (in_ios  == NULL) {
        exit(EXIT_FAILURE);
    }
    if (rwGetFileType(in_ios) != FT_RWROUTED) {
        skAppPrintErr("Input file '%s' not in RWROUTED format",
                      in_fpath);
        exit(EXIT_FAILURE);
    }

    /* Determine byte order */
    if (rwGetIsBigEndian(in_ios)) {
        endian = SILK_ENDIAN_BIG;
    } else {
        endian = SILK_ENDIAN_LITTLE;
    }

    /* Output output file */
    if ((rv = rwioCreate(&out_ios, out_fpath, SK_RWIO_WRITE))
        || (rv = rwioSetFileType(out_ios, FT_RWSPLIT))
        || (rv = rwioSetFileVersion(out_ios, rwGetFileVersion(in_ios)))
        || (rv = rwioSetFileByteorder(out_ios, rwGetFileByteorder(in_ios)))
        || (rv = rwioSetFileSTime(out_ios, rwGetFileSTime(in_ios), 0)))
    {
        rwioPrintLastErr(out_ios, rv, &skAppPrintErr);
        skAppPrintErr("Unable to open output file '%s'.", out_fpath);
        exit(EXIT_FAILURE);
    }

    return;                       /* OK */
}


int main(int argc, char **argv)
{
    rwRec rwrec;
    int64_t rec_count = 0;
    int64_t file_size_real = 0;
    int64_t file_size_calc = 0;
    int64_t hdr_len;
    int64_t rec_len;

    /* Setup app: open input and output files; will exit(1) on error */
    appSetup(argc, argv);

    hdr_len = out_ios->hdrLen;
    rec_len = out_ios->recLen;

    /* Process body */
    while (rwRead(in_ios, &rwrec)) {
        rec_count++;
        if ( rwWrite(out_ios, &rwrec)) {
            skAppPrintErr("Error writing to '%s'.  Stopping copy.",
                          out_fpath);
            break;
        }
    }

    rwCloseFile(in_ios);
    in_ios = NULL;
    rwCloseFile(out_ios);
    out_ios = NULL;

    file_size_real = fileSize(out_fpath);
    file_size_calc = hdr_len + rec_len * rec_count;
    if (file_size_real != file_size_calc) {
        skAppPrintErr(("ERROR: output filesize mismatch."
                       " Calc. %lld vs real %lld"),
                      (long long)file_size_calc, (long long)file_size_real);
        exit(EXIT_FAILURE);
    }

    /* done */
    return 0;
}
