/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
** rwsid.c
** Suresh Konda
** 10/11/2002
** Module to deal with sensor id to sensor name mapping consistently.
** Basically exports 2 variables.  rwpack.h declares a macro to
** map an id to a name with checks.
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/librw/rwsid.c,v 1.15 2005/05/20 17:54:28 thomasm Exp $");

#define DECLARE_SENSORID_VARIABLES
#include "silk_site.h"

#include "librw_priv.h"



const char *sensorIdToName(sensorID_t id)
{
    if (id >= numSensors) {
        return "UNKN?";
    }
    return sensorInfo[id].sensorName;
}


sensorID_t sensorNameToId(const char *name)
{
    sensorID_t i;

    for (i = 0; i < numSensors; i++) {
        if (0 == strcmp(sensorInfo[i].sensorName, name)) {
            return i;
        }
    }

    return SENSOR_INVALID_SID;
}


int sensorNameIsLegal(const char *name)
{
    size_t len;
    const char *cp;

    /* check length; should be between 1 and SENSOR_MAX_SENSOR_NAME_LEN */
    len = strlen(name);
    if (len < 1) {
        /* skAppPrintErr("sensor name must contain at least 1 character"); */
        return -2;
    }
    if (len > SENSOR_MAX_SENSOR_NAME_LEN) {
        /*skAppPrintErr("sensor name can contain no more than %d characters",
         *              SENSOR_MAX_SENSOR_NAME_LEN);
         */
        return -3;
    }

    /* check for bad characters */
    for (cp = name; *cp; ++cp) {
        if (isspace((int)*cp)) {
            return (1 + cp - name);
            /*skAppPrintErr("sensor name cannot contain whitespace characters");
             *return 1;
             */
        }
        if (('/' == *cp) || ('_' == *cp)) {
            return (1 + cp - name);
            /*skAppPrintErr("sensor name contains illegal character '%c'",
             *              *cp);
             *return 1;
             */
        }
    }

    /* check that begins with a letter */
    if ( !isalpha((int)*name)) {
        /* skAppPrintErr("sensors must begin with a letter"); */
        return -1;
    }

    return 0;
}


int sensorCheckCounts(void)
{
    size_t sensor_count;

    sensor_count = sizeof(sensorInfo)/sizeof(sensorinfo_t) - 1;
    if (numSensors <= sensor_count) {
        return 0;
    }

    skAppPrintErr(("Declared number of sensors (%lu) > defined sensors (%lu)\n"
                   "Adjust the silk_site_*.h file and recompile"),
                  (unsigned long)numSensors, (unsigned long)sensor_count);
    return sensor_count;
}


sensorID_t getSIDFromFName(const char *fPath, int stdinFlag)
{
#if (SENSOR_COUNT == 0)
    return SENSOR_INVALID_SID;
#else
    /*
    ** as a stop gap measure, we have to stash the sensor id into the
    ** reader struct and get the value from the file name. When we move
    ** to the next version where the sid is embedded into the record itself
    ** we have to change this here and in each of the rwReadXXXX functions.
    ** The general filename format is
    **     <prefix>-<sensorid>_<YYYYMMDD.HH>[.gz]
    **
    ** For version 0, sensorid was X<nn>
    **
    ** For version 1, sensorid begins with an upper case letter and
    **     constains any character except underscore
    **
    ** These are stashed in sensorInfo
    */
    unsigned int i = 0;
    char *cp, *ep;
    unsigned int sid_len;
    char name[SENSOR_MAX_SENSOR_NAME_LEN + 1];

    cp = baseName(fPath);

    /* skip past <prefix>- */
    cp = strchr(cp, '-');
    if (!cp) {
        goto ERROR;
    }
    cp++;                 /* past - */

    if (!isupper((int)*cp)) {
        goto ERROR;
    }

    /* get end of sensor id */
    ep = strchr(cp, '_');
    if (!ep) {
        goto ERROR;
    }

    sid_len = (ep - cp);
    if (sid_len == 0 || sid_len > SENSOR_MAX_SENSOR_NAME_LEN) {
        goto ERROR;
    }

    /* copy sensor to name */
    strncpy(name, cp, sid_len);
    name[sid_len] = '\0';
    /* do the compare */
    for (i  = 0; i < numSensors; i++ ) {
        if (strcmp(sensorInfo[i].sensorName, name) == 0 ) {
            return i;
        }
    }

  ERROR:
    if (!stdinFlag) {
        if (0 == i) {
            skAppPrintErr("Warning: could not find Sensor ID. Setting to %u",
                          SENSOR_INVALID_SID);
        } else {
            skAppPrintErr(("Warning: could not match sensor name '%s'"
                           " to a sensor ID.  Setting to %u"),
                          name, SENSOR_INVALID_SID);
        }
    }
    return SENSOR_INVALID_SID;
#endif  /* #else clause of #if SENSOR_COUNT == 0 */
}


#if 0
/* parse comma separated list of sensor names and/or ids */
int sensorParseList(
    const char *sensor_input)
{
    char *number_list = NULL;
    unsigned int i;
    int found_sensor;
    sensorinfo_t *s_info;
    uint8_t list_len;
    uint8_t *parsed_list = NULL;
    char *err_name = NULL;
    const char *cp;
    const char *ep;
    int len;
    int rv = 0; /* return value */
    int number_list_len = 0;

    /* Look at the user's sensor_input as comma-delimited tokens.  For
     * each token, if it starts with a digit, copy the token to
     * number_list, otherwise, treat it as a sensor name and try to
     * find it in the list of sensors.  number_list is parsed after
     * all the named sensors have been removed.
     */

    cp = sensor_input;
    while (*cp) {
        /* find next comma-delimited token; set len to its length */
        ep = strchr(cp, ',');
        if (ep == cp) {
            /* double comma */
            ++cp;
            continue;
        }
        if (ep) {
            len = (size_t)(ep - cp);
            ++ep;
        } else {
            len = strlen(cp);
            ep = cp+len;
        }

        if (isdigit((int)*cp)) {
            /* copy number/range to number_list for later processing and go
             * to the next token.  This assumes sensor names never begin
             * with a digit. */
            if (!number_list) {
                number_list = calloc(sizeof(char), 1+strlen(sensor_input));
                if (!number_list) {
                    skAppPrintErr("Out of memory at %s:%d",
                                  __FILE__, __LINE__);
                    rv = 1;
                    goto END;
                }
            }
            strncpy(&(number_list[number_list_len]), cp, len+1);
            number_list_len += len+1;
        } else {
            /* look for the sensor as a name among all the sensors.
             * found_sensor is 0 if we don't find it; 1 if we do */
            found_sensor = 0;
            for (i = 0; i < numSensors; ++i) {
                s_info = &sensorInfo[i];
                if ((0 == strncmp(cp, s_info->sensorName, len)) &&
                    ('\0' == s_info->sensorName[len]))
                {
                    BITMAP_SETBIT(sensor_bits, i);
                    ++found_sensor;
                    break;
                }
            }

            if (found_sensor == 0) {
                if ( (err_name = malloc(len+1)) ) {
                    /* only show user the current name that is giving
                     * problems */
                    strncpy(err_name, cp, len);
                    err_name[len] = '\0';
                    cp = err_name;
                }
                skAppPrintErr("unknown sensor name '%s'", cp);
                rv = 1;
                goto END;
            }
        } /* if (isdigit()) .. else */

        cp = ep;
    } /* while (*cp) */

    if (0 == number_list_len) {
        /* all of input given as named sensors */
        rv = 0;
        goto END;
    }

    /* handle the numbers/ranges in user's input */
    number_list[number_list_len] = '\0';

    /* convert string of numbers/ranges to array of numbers */
    parsed_list = skParseNumberList(number_list, 0, numSensors, &list_len);
    if (!parsed_list) {
        skAppPrintErr("invalid sensor list '%s'",
                      number_list);
        rv = 1;
        goto END;
    }

    /* loop 'i' over all the sensors ids given on the command line */
    for (i = 0; i < list_len; ++i) {
        BITMAP_SETBIT(sensor_bits, parsed_list[i]);
    }

 END:
    if (parsed_list) {
        free(parsed_list);
    }
    if (number_list) {
        free(number_list);
    }
    if (err_name) {
        free(err_name);
    }
    return rv;
}
#endif /* 0 */
