/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  Functions to support binary output: opening binary SiLK files for
**  writing or appending.
**
*/


#include "silk.h"

RCSIDENT("$SiLK: /src/rw/librw/rwwrite.c,v 1.10 2005/09/16 21:54:42 thomasm Exp $");

#include "librw_priv.h"



/*  Convert bytes and packets fields in 'rwRP' to the values used in
 *  rwfilter output files and in the packed file formats.  See
 *  librw_priv.h. */
int _writePackBytesPackets(
    uint32_t               *bpp_out,
    uint32_t               *pkts_out,
    uint32_t               *pflag_out,
    const rwGenericRec_V3  *rwRP)
{
#define VERIFY_BPP 1  /* whether to verify bpp value in rwRP */
    uint32_t packets;
    uint32_t bytes;

    assert(bpp_out);
    assert(pkts_out);
    assert(pflag_out);
    assert(rwRP);

    /* Check for 0 value in 'pkts' field */
    packets = rwRP->pkts;
    if (packets == 0) {
        return LIBRW_ERR_PKTS_ZERO;
    }

    /* Set packets field; check for overflow */
    if (packets < MAX_PKTS) {
        *pkts_out = packets;
        *pflag_out = 0;
    } else {
        *pkts_out = packets / PKTS_DIVISOR;
        if (*pkts_out >= MAX_PKTS) {
            /* Double overflow in pkts */
            return LIBRW_ERR_PKTS_OVRFLO;
        }
        /* pktsFlag */
        *pflag_out = 1;
    }

    /* Use the bytes-per-packet field on rwGeneric if it is non-zero;
     * otherwise compute it. */
    /* bpp: uint32_t bPPkt:14;  uint32_t bPPFrac:6; */
    if ((rwRP->bpp != 0) && !VERIFY_BPP) {
        /* use value as is */
        *bpp_out = rwRP->bpp;
    } else {
        /* compute new value */
        bytes = rwRP->bytes;
        *bpp_out = (((bytes / packets) << 6) |
                    ((BPP_PRECN * (bytes % packets)) / packets));
#if VERIFY_BPP
        /* ignore fractional part in comparison */
        if ((rwRP->bpp >> 6) == (*bpp_out >> 6)) {
            /* two values are equal, use the value from the rwrec. */
            *bpp_out = rwRP->bpp;
        } else if (rwRP->bpp != 0) {
            /* don't print warning if bpp on rwrec was zero */
            skAppPrintErr("Bad value in bpp field");
        }
#endif /* VERIFY_BPP */
    }

    return LIBRW_OK;
}


/* Pack the protocol, flags, and TCP state fields.  See librw_priv.h */
void _writePackProtoFlags(
    uint8_t                *is_tcp_out,
    uint8_t                *prot_flags_out,
    uint8_t                *tcp_state_out,
    uint8_t                *rest_flags_out,
    const rwGenericRec_V3  *rwRP)
{
    if (rwRP->proto != 6) {
        /* Flow is not TCP, so there is no additional TCP info.  Set
         * '*rest_flags_out' to value of rwRP->flags. */
        *is_tcp_out = 0;
        *prot_flags_out = rwRP->proto;
        *tcp_state_out  = 0;
        *rest_flags_out = rwRP->flags;
    } else {
        /* Flow is TCP */
        *is_tcp_out = 1;

        if (rwRP->tcp_state == 0) {
            /* There is no additional TCP info. */
            *prot_flags_out = rwRP->flags;
            *tcp_state_out  = 0;
            *rest_flags_out = 0;
        } else {
            /* There is additional TCP info.  Put the initial TCP
             * flags into the '*prot_flags_out' value. */
            *prot_flags_out = rwRP->init_flags;
            *tcp_state_out  = rwRP->tcp_state;
            *rest_flags_out = rwRP->rest_flags;
        }
    }
}


/*  Compute the 'sbb' and 'pef' fields used in packed file formats.
 *  See librw_priv.h. */
int _writePackSbbPef(
    uint32_t                 *sbb_out,
    uint32_t                 *pef_out,
    const rwGenericRec_V3    *rwRP,
    uint32_t                  file_start_time)
{
    int rv = LIBRW_OK; /* return value */
    uint32_t start_time = RWREC_GET_STIME_SEC(rwRP);
    uint32_t elapsed = RWREC_GET_ELAPSED_SEC(rwRP);
    uint32_t pkts, bpp, pflag;

    if (elapsed >= MAX_ELAPSED_TIME) {
        rv = LIBRW_ERR_ELPSD_OVRFLO;
        goto END;
    }

    if (start_time < file_start_time) {
        rv = LIBRW_ERR_STIME_UNDRFLO;
        goto END;
    }
    start_time -= file_start_time;
    if (start_time >= MAX_START_TIME) {
        rv = LIBRW_ERR_STIME_OVRFLO;
        goto END;
    }

    rv = _writePackBytesPackets(&bpp, &pkts, &pflag, rwRP);
    if (rv) { goto END; }

    /* sbb: uint32_t sTime:12;  uint32_t bPPkt:14;  uint32_t bPPFrac:6; */
    *sbb_out = ((start_time << 20) | (bpp & MASKARRAY_20));

    /* pef: uint32_t pkts:20; uint32_t elapsed :11; uint32_t pktsFlag:1; */
    *pef_out = ((pkts << 12) | (elapsed << 1) | pflag);

  END:
    return rv;
}


/*
 *  _writePackTimeBytesPktsFlag(&pkts_stime_out, &bbe_out, &msec_flags_out,
 *                              rwRP, file_start_time);
 *
 *    Computes the 'sbb' and 'pef' fields used when packing split,
 *    web, routed, and notrouted V1 files, and rwfilter V2 files.
 *
 *    Uses the sTime, elapsed, pkts, and bytes fields in the rwGeneric
 *    record pointer rwRP to compute these values.  rwIOS is an
 *    rwIOStructPtr to the file containing the record--required for
 *    file's start time.
 *
 *    If the record has millisec values for sTime and/or elapsed, the
 *    sTime and elapsed fields are rounded if required.
 *
 *    sbb and pef are returned in native byte order.
 *
 *    Returns 0 on success or non-zero on these failures: rwRP's sTime
 *    is earlier than rwIOS's sTime or is too large; elapsed time
 *    is too large; packets field is too large.
 */
int _writePackTimeBytesPktsFlags(
    uint32_t               *pkts_stime_out,
    uint32_t               *bbe_out,
    uint32_t               *msec_flags_out,
    const rwGenericRec_V3  *rwRP,
    uint32_t                file_start_time)
{
    int rv = LIBRW_OK; /* return value */
    uint32_t start_time = rwRP->sTime;
    uint32_t pkts, bpp, pflag, is_tcp;
    uint8_t prot_flags;

    if (rwRP->elapsed >= MAX_ELAPSED_TIME) {
        rv = LIBRW_ERR_ELPSD_OVRFLO;
        goto END;
    }

    if (start_time < file_start_time) {
        rv = LIBRW_ERR_STIME_UNDRFLO;
        goto END;
    }
    start_time -= file_start_time;
    if (start_time >= MAX_START_TIME) {
        rv = LIBRW_ERR_STIME_OVRFLO;
        goto END;
    }

    rv = _writePackBytesPackets(&bpp, &pkts, &pflag, rwRP);
    if (rv) { goto END; }

    /* pkts_stime: pkts:20; sTime: 12; */
    *pkts_stime_out = ((pkts << 12) | start_time);

    /* bbe: bpp: 20; elapsed: 12 */
    *bbe_out = ((bpp << 12) | rwRP->elapsed);

    /* set is_tcp bit and prot_flags */
    if (rwRP->proto == 6) {
        is_tcp = 1;
        prot_flags = rwRP->flags;
    } else {
        is_tcp = 0;
        prot_flags = rwRP->proto;
    }

    /* msec_flags: sTime_msec:10; elaps_msec:10; pflag:1;
     *             is_tcp:1; pad:2; prot_flags:8;*/
    *msec_flags_out = ((rwRP->sTime_msec << 22)
                       | ((rwRP->elapsed_msec & MASKARRAY_10) << 12)
                       | (pflag ? (1 << 11) : 0)
                       | (is_tcp ? (1 << 10) : 0)
                       | prot_flags);

  END:
    return rv;
}


/* Write the record in 'data' to 'rwIOS'.  See librw_priv.h. */
int _writeRecordWrite(
    rwIOStruct         *rwIOS,
    const uint8_t      *data)
{
    if (fwrite(data, rwIOS->recLen, 1, rwIOS->FD)) {
        return LIBRW_OK;
    }
    rwIOS->errnum = errno;
    return LIBRW_ERR_WRITE;
}
