/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
 * addrcount.c
 *
 * this is the last major rwset tool I want to write, it takes in two streams
 * of data, an rwset and an rwfilter stream.  From this data, it then generates
 * a result - one of three outputs:
 *            totals (default) - outputs to screen a table containing the
 *                               ip address, bytes, packets, flows
 *            print-ips        - outputs to screen the ip addresses
 *            set-file         - outputs to screen the set data.
 *
 * So the reason for the second two is because I'm including three thresholds
 * here - bytes, packets & records.
 *
 * 12/2 notes.  Often, the best is the enemy of the good, I am implementing
 * a simple version of this application for now with the long-term plan being
 * that I am going to write a better, faster, set-friendly version later.
 */

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwaddrcount/rwaddrcount.c,v 1.25 2005/10/26 14:16:26 thomasm Exp $");

#include "iptree.h"
#include "rwpack.h"
#include "utils.h"
#include "iochecks.h"
#include "rwaddrcount.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* where to write output from --help */
#define USAGE_FH stdout

/* I'm #define the hash and comparison functions to avoid stack overhead */
#define HASHFUNC(value) \
    ((value ^ (value >> 7) ^ (value << 23)) % RWAC_ARRAYSIZE)

/* IPHASH:
 * rwrec: rwRecord to use.
 * useSrc: use Source IP address
 */
#define IPHASH(rwrec, useDst) \
    ((useDst) ? HASHFUNC((rwrec).dIP.ipnum) : HASHFUNC((rwrec).sIP.ipnum))
#define CMPFNC(rwrec, crPtr, useDst) \
    ((useDst) ? ((rwrec).dIP.ipnum == (crPtr)->value) : \
                ((rwrec).sIP.ipnum == (crPtr)->value))


/*
 * when generating ouput, this macro will evaluate to TRUE if the
 * record is within the limits given by the user and should be
 * printed/counted/used-to-genrate-output.  This macro uses the global
 * limit variables.
 */
#define IS_RECORD_WITHIN_LIMITS(count_rec)                     \
    ((count_rec)->bytes >= byteMin &&                          \
     (count_rec)->packets >= packetMin &&                      \
     (count_rec)->flows >= recMin &&                           \
     (!byteMax || (count_rec)->bytes <= byteMax) &&            \
     (!packetMax || (count_rec)->packets <= packetMax) &&      \
     (!recMax || (count_rec)->flows <= recMax))


/* LOCAL FUNCTIONS */

static void appUsage(void);             /* never returns */
static void appUsageLong(void);
static void appTeardown(void);
static void appSetup(int argc, char **argv);    /* never returns */
static void appOptionsUsage(FILE*);             /* application options usage */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);

/* EXPORTED VARIABLES */

rwac_print_mode_t printMode = 0;
uint8_t useDestFlag = 0;
uint8_t integer_ips_flag = 0;
uint8_t zero_pad_ips_flag = 0;
uint8_t sort_ips_flag = 0;
uint32_t byteMin, recMin, packetMin, byteMax, recMax, packetMax;
countRecord_t *countArray[RWAC_ARRAYSIZE];
uint32_t totalHashRecords;
char* outputSetFN;


/* LOCAL VARIABLES */

static iochecksInfoStruct *ioISP;
static rwRec currentRecord;     /* Record we're currently reading*/

/* whether to print column titles; default yes */
static uint8_t printTitles = 1;

/* name of program to run to page output */
static char *pager = NULL;


/* OPTIONS */

/* Names of options; keep the order in sync with appOptions[] */
typedef enum {
    RWACOPT_PRINT_RECS,
    RWACOPT_PRINT_STAT,
    RWACOPT_PRINT_IPS,
    RWACOPT_USE_DEST,
    RWACOPT_BYTE_MIN,
    RWACOPT_PKT_MIN,
    RWACOPT_REC_MIN,
    RWACOPT_BYTE_MAX,
    RWACOPT_PKT_MAX,
    RWACOPT_REC_MAX,
    RWACOPT_SET_FILE,
    RWACOPT_NO_TITLES,
    RWACOPT_INTEGER_IPS,
    RWACOPT_ZERO_PAD_IPS,
    RWACOPT_SORT_IPS,
    RWACOPT_PAGER
} appOptionsEnum;

struct option appOptions[] = {
    {"print-recs",   NO_ARG,       0, RWACOPT_PRINT_RECS},
    {"print-stat",   NO_ARG,       0, RWACOPT_PRINT_STAT},
    {"print-ips",    NO_ARG,       0, RWACOPT_PRINT_IPS},
    {"use-dest",     NO_ARG,       0, RWACOPT_USE_DEST},
    {"byte-min",     REQUIRED_ARG, 0, RWACOPT_BYTE_MIN},
    {"packet-min",   REQUIRED_ARG, 0, RWACOPT_PKT_MIN},
    {"rec-min",      REQUIRED_ARG, 0, RWACOPT_REC_MIN},
    {"byte-max",     REQUIRED_ARG, 0, RWACOPT_BYTE_MAX},
    {"packet-max",   REQUIRED_ARG, 0, RWACOPT_PKT_MAX},
    {"rec-max",      REQUIRED_ARG, 0, RWACOPT_REC_MAX},
    {"set-file",     REQUIRED_ARG, 0, RWACOPT_SET_FILE},
    {"no-titles",    NO_ARG,       0, RWACOPT_NO_TITLES},
    {"integer-ips",  NO_ARG,       0, RWACOPT_INTEGER_IPS},
    {"zero-pad-ips", NO_ARG,       0, RWACOPT_ZERO_PAD_IPS},
    {"sort-ips",     NO_ARG,       0, RWACOPT_SORT_IPS},
    {"pager",        REQUIRED_ARG, 0, RWACOPT_PAGER},
    {0,0,0,0}                     /* sentinel entry */
};

const char *appHelp[] = {
    "Print summary byte, packet, flow counts per IP record",
    "Print statistics (total bytes, packets, flows, unique IPs)",
    "Print IP addresses only to stdout",
    "Use destination IP address as key. Def. Source address",
    "Min byte count in an IP record for output. Def. 1",
    "Min packet count in an IP record for output. Def. 1",
    "Min record count in an IP record for output. Def. 1",
    "Max byte count in an IP record for output. Def. None",
    "Max packet count in an IP record for output. Def. None",
    "Max record count in an IP record for output. Def. None",
    "Write IPs to specified binary IPset file. Def. No",
    "Do not print column titles. Def. Print titles",
    "Print IP numbers as integers. Def. No",
    "Print IP numbers as zero-padded dotted-decimal. Def. No",
    "When printing results, sort by IP address. Def. No",
    "Program to invoke to page output. Def. $SILK_PAGER or $PAGER",
    (char *)NULL
};


/*
 * appUsage:
 *      print usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 *      print usage information to USAGE_FH and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
void appUsageLong(void)
{
#define USAGE_MSG                                                            \
    ("%s {--print-recs|--print-stat|--print-ips} [options] [files]\n"        \
     "\tSummarize SiLK Flow records by source or destination IP; with\n"     \
     "\tthe --print-recs option will produce textual ouput with counts of\n" \
     "\tbytes, packets, and flow records for each IP, and the time range\n"  \
     "\twhen the IP was active.  When no files are given on command line,\n" \
     "\tflows are read from STDIN.\n")
    
    fprintf(USAGE_FH, USAGE_MSG, skAppName());
    appOptionsUsage(USAGE_FH);
    exit(EXIT_FAILURE);
}


/*
 * appOptionsUsage:
 *      print options for this app to specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\noptions\n");
    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }
    return;
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    iochecksTeardown(ioISP);
    optionsTeardown();

    return;
}

/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    skAppRegister(argv[0]);

    /*
     * Variable initialization.
     */
    useDestFlag = 0;
    printMode = RWAC_PMODE_NONE;
    byteMin = 0;
    recMin = 0;
    packetMin = 0;
    byteMax = 0;
    recMax = 0;
    packetMax = 0;

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* create the input/output checker */
    ioISP = iochecksSetup(0, 0, argc, argv);

    /* initialize options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* register our options and handler */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* parse options */
    ioISP->firstFile = optionsParse(argc, argv);
    if (ioISP->firstFile < 0) {
        appUsage();/* never returns */
    }

    /* Use STDIN as an input stream if it is not a TTY; make certain
     * we have some input and we are either reading from STDIN or
     * using files listed the command line, but not both. */
    if (iochecksAcceptFromStdin(ioISP) || iochecksInputs(ioISP, 0)) {
        appUsage();
    }

    if (printMode == RWAC_PMODE_NONE) {
        skAppPrintErr("Specify --print-stat, --print-recs or --print-ips");
        appUsage();
    }

    /* Do they want the IPs in sorted order? */
    if (sort_ips_flag) {
        switch (printMode) {
          case RWAC_PMODE_IPS:
            printMode = RWAC_PMODE_SORTED_IPS;
            break;
          case RWAC_PMODE_RECS:
            printMode = RWAC_PMODE_SORTED_RECS;
            break;
          default:
            /* skAppPrintErr("--sort-ips switch ignored"); */
            break;
        }
    }

    /* Using both the integer-ips and zero-pad-ips options is
     * inconsistent.  Let integer-ips override zero-pad-ips and print
     * warning */
    if (integer_ips_flag && zero_pad_ips_flag) {
        skAppPrintErr("--integer-ips option overrides --zero-pad-ips\n"
                      "\t Will print IPs as integers.");
        zero_pad_ips_flag = 0;
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    (void) umask((mode_t) 0002);

    return;                       /* OK */
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int  appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    switch ((appOptionsEnum)opt_index) {
      case RWACOPT_USE_DEST:
        useDestFlag = 1;
        break;

      case RWACOPT_BYTE_MIN:
        if (skStringParseUint32(&byteMin, opt_arg, 0, 0) != 0) {
            skAppPrintErr("Bad argument to --%s switch: '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

      case RWACOPT_PKT_MIN:
        if (skStringParseUint32(&packetMin, opt_arg, 0, 0) != 0) {
            skAppPrintErr("Bad argument to --%s switch: '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

      case RWACOPT_REC_MIN:
        if (skStringParseUint32(&recMin, opt_arg, 0, 0) != 0) {
            skAppPrintErr("Bad argument to --%s switch: '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

      case RWACOPT_BYTE_MAX:
        if (skStringParseUint32(&byteMax, opt_arg, 0, 0) != 0) {
            skAppPrintErr("Bad argument to --%s switch: '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

      case RWACOPT_PKT_MAX:
        if (skStringParseUint32(&packetMax, opt_arg, 0, 0) != 0) {
            skAppPrintErr("Bad argument to --%s switch: '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

      case RWACOPT_REC_MAX:
        if (skStringParseUint32(&recMax, opt_arg, 0, 0) != 0) {
            skAppPrintErr("Bad argument to --%s switch: '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

      case RWACOPT_PRINT_STAT:
        printMode = RWAC_PMODE_STAT;
        break;

      case RWACOPT_PRINT_IPS:
        printMode = RWAC_PMODE_IPS;
        break;

      case RWACOPT_PRINT_RECS:
        printMode = RWAC_PMODE_RECS;
        break;

      case RWACOPT_SET_FILE:
        printMode = RWAC_PMODE_IPSETFILE;
        outputSetFN = opt_arg;
        break;

      case RWACOPT_NO_TITLES:
        printTitles = 0;
        break;

      case RWACOPT_INTEGER_IPS:
        integer_ips_flag = 1;
        break;

      case RWACOPT_ZERO_PAD_IPS:
        zero_pad_ips_flag = 1;
        break;

      case RWACOPT_SORT_IPS:
        sort_ips_flag = 1;
        break;

      case RWACOPT_PAGER:
        pager = opt_arg;
        break;
    };

    return 0;                     /* OK */
}


/*
 * int addRecord(rwRec *Location, rwRec *currentRecord)
 *
 * Adds the contents of a record to the values stored in currentRecord.
 *
 * This function is actually designed for new records or records that
 * already exist.  The "not existence" flag is the epoch start time
 * which, if set to zero is assumed to be a recently calloc'd record.
 */
static int addRecord(countRecord_t *tgt, rwRec *newRecord)
{
    tgt->bytes += newRecord->bytes;
    tgt->packets += newRecord->pkts;
    tgt->flows ++;
    if (!tgt->start) {
        tgt->value = useDestFlag ? newRecord->dIP.ipnum : newRecord->sIP.ipnum;
        tgt->start = newRecord->sTime;
        tgt->end = newRecord->sTime + newRecord->elapsed;
        totalHashRecords++;
    } else {
        tgt->start =
            (newRecord->sTime <= tgt->start) ? newRecord->sTime : tgt->start;
        tgt->end = (newRecord->sTime + newRecord->elapsed) > tgt->end ?
            (newRecord->sTime + newRecord->elapsed) : tgt->end;
    }
    return 0;
}



/*
 * SECTION: Dumping
 *
 * All the output routines are in this section of the text.
 */


/*
 *  countArrayToIPTree(&iptree, count_array);
 *
 *    Fills in the 'iptree' with all the IPs in the 'count_array'.
 */
static void countArrayToIPTree(
    skIPTree_t    **ip_tree,
    countRecord_t **count_array)
{
    countRecord_t *currentRec;
    uint32_t i;

    skIPTreeCreate(ip_tree);

    for (i = 0; i < RWAC_ARRAYSIZE; i++) {
        currentRec = count_array[i];
        while (currentRec != NULL) {
            if (IS_RECORD_WITHIN_LIMITS(currentRec)) {
                skIPTreeAddAddress((currentRec->value), (*ip_tree));
            };
            if (currentRec->nextRecord == count_array[i]) break;
            currentRec = currentRec->nextRecord;
        }
    }
}


/*
 *  int dumpRecords(outfp)
 *
 *    Dumps the addrcount contents as a record of bytes, packets,
 *    times &c to 'outfp'
 *
 *    This is the typical text output from addrcount.
 *
 */
static int dumpRecords(FILE *outfp)
{
    uint32_t i;
    countRecord_t *currentRec;

    if (printTitles) {
        fprintf(outfp, "%15s|%20s|%10s|%10s|%20s|%20s|\n",
                (useDestFlag ? "dIP" : "sIP"), "Bytes", "Packets", "Records",
                "Start_Time", "End_Time");
    }

    for (i = 0; i < RWAC_ARRAYSIZE; i++) {
        currentRec = countArray[i];
        while (currentRec != NULL) {
            if (IS_RECORD_WITHIN_LIMITS(currentRec)) {
                /* print IP in appropriate form */
                if (integer_ips_flag) {
                    fprintf(outfp, "%15lu|",
                            (unsigned long)currentRec->value);
                } else if (zero_pad_ips_flag) {
                    fprintf(outfp, "%15s|",
                            num2dot0(currentRec->value));
                } else {
                    fprintf(outfp, "%15s|",
                            num2dot(currentRec->value));
                }
                /* print "volumes" and time */
                fprintf(outfp, "%20llu|%10u|%10u|%20s|",
                        currentRec->bytes, currentRec->packets,
                        currentRec->flows, timestamp(currentRec->start));
                fprintf(outfp, "%20s|\n",
                        timestamp(currentRec->end));
            }
            if (currentRec->nextRecord == countArray[i]) break;
            currentRec = currentRec->nextRecord;
        }
    }
    return 0;
}


/*
 *  int dumpRecordsSorted(outfp)
 *
 *    Dumps the addrcount contents as a record of bytes, packets,
 *    times &c to 'outfp', sorted by the IP address.
 *
 */
static int dumpRecordsSorted(FILE *outfp)
{
    uint32_t ip, key;
    countRecord_t *currentRec;
    skIPTree_t *resultTree;
    skIPTreeIterator_t *iter;

    countArrayToIPTree(&resultTree, countArray);
    if (skIPTreeIteratorCreate(&iter, resultTree)) {
        skAppPrintErr("Unable to create IPTree iterator");
        return 1;
    }

    if (printTitles) {
        fprintf(outfp, "%15s|%20s|%10s|%10s|%20s|%20s|\n",
                (useDestFlag ? "dIP" : "sIP"), "Bytes", "Packets", "Records",
                "Start_Time", "End_Time");
    }

    while (skIPTreeIteratorNext(&ip, iter) == SKIP_ITERATOR_OK) {

        /* find the ip's entry in the hash table */
        key = HASHFUNC(ip);

        /* loop through the list of records at this hash table entry
         * until we find the one that has the IP we want */
        currentRec = countArray[key];
        while (currentRec != NULL) {
            if (ip != currentRec->value) {
                /* not the ip we wanted; goto next */
                currentRec = currentRec->nextRecord;
                /* if we've looped all the around, we missed the
                 * IP we wanted, and things are horked */
                assert(currentRec != countArray[key]);
                continue;
            }

            /* print IP in appropriate form */
            if (integer_ips_flag) {
                fprintf(outfp, "%15lu|",
                        (unsigned long)currentRec->value);
            } else if (zero_pad_ips_flag) {
                fprintf(outfp, "%15s|",
                        num2dot0(currentRec->value));
            } else {
                fprintf(outfp, "%15s|",
                        num2dot(currentRec->value));
            }
            /* print "volumes" and time */
            fprintf(outfp, "%20llu|%10u|%10u|%20s|",
                    currentRec->bytes, currentRec->packets,
                    currentRec->flows, timestamp(currentRec->start));
            fprintf(outfp, "%20s|\n",
                    timestamp(currentRec->end));
            break;
        }
    }

    skIPTreeIteratorDestroy(&iter);
    return 0;
}



/*
 *  int dumpIPs(outfp)
 *
 *    writes IP addresses to the stream 'outfp' in hash-table order
 *    (unsorted).
 *
 *    Returns 0 on success.
 */
static int dumpIPs(FILE *outfp)
{
    uint32_t i;
    countRecord_t *currentRec;

    if (printTitles) {
        fprintf(outfp, "%15s\n", (useDestFlag ? "dIP" : "sIP"));
    }

    for (i = 0; i < RWAC_ARRAYSIZE; i++) {
        currentRec = countArray[i];
        while (currentRec != NULL) {
            if (IS_RECORD_WITHIN_LIMITS(currentRec)) {
                if (integer_ips_flag) {
                    fprintf(outfp, "%15lu\n",
                            (unsigned long)currentRec->value);
                } else if (zero_pad_ips_flag) {
                    fprintf(outfp, "%15s\n",
                            num2dot0(currentRec->value));
                } else {
                    fprintf(outfp, "%15s\n",
                            num2dot(currentRec->value));
                }
            }
            if (currentRec->nextRecord == countArray[i]) break;
            currentRec = currentRec->nextRecord;
        }
    }
    return 0;
}



/*
 *  int dumpIPsSorted(outfp)
 *
 *    Writes the IPs to the stream 'outfp' in sorted order.
 *
 */
static int dumpIPsSorted(FILE *outfp)
{
    uint32_t ip;
    skIPTree_t *resultTree;
    skIPTreeIterator_t *iter;

    countArrayToIPTree(&resultTree, countArray);
    if (skIPTreeIteratorCreate(&iter, resultTree)) {
        skAppPrintErr("Unable to create IPTree iterator");
        return 1;
    }

    if (printTitles) {
        fprintf(outfp, "%15s\n", (useDestFlag ? "dIP" : "sIP"));
    }

    while (skIPTreeIteratorNext(&ip, iter) == SKIP_ITERATOR_OK) {
        if (integer_ips_flag) {
            fprintf(outfp, "%15lu\n",
                    (unsigned long)ip);
        } else if (zero_pad_ips_flag) {
            fprintf(outfp, "%15s\n",
                    num2dot0(ip));
        } else {
            fprintf(outfp, "%15s\n",
                    num2dot(ip));
        }
    }

    skIPTreeIteratorDestroy(&iter);
    return 0;
}



/*
 *  int dumpStats(outfp)
 *
 *    dumps a text string describing rwaddrcont results to the stream
 *    'outfp'.
 */
static int dumpStats(FILE *outfp)
{
    uint32_t i;
    uint32_t qual_ips;
    uint64_t qual_bytes, qual_packets, qual_flows;
    uint64_t tot_bytes,  tot_packets,  tot_flows;
    countRecord_t *currentRec;

    qual_ips = 0;
    qual_bytes = qual_packets = qual_flows = 0;
    tot_bytes  = tot_packets  = tot_flows  = 0;

    for (i = 0; i < RWAC_ARRAYSIZE; i++) {
        currentRec = countArray[i];
        while (currentRec != NULL) {
            tot_bytes   += currentRec->bytes;
            tot_packets += currentRec->packets;
            tot_flows   += currentRec->flows;

            if (IS_RECORD_WITHIN_LIMITS(currentRec)) {
                ++qual_ips;
                qual_bytes   += currentRec->bytes;
                qual_packets += currentRec->packets;
                qual_flows   += currentRec->flows;
            }
            if (currentRec->nextRecord == countArray[i]) break;
            currentRec = currentRec->nextRecord;
        }
    }

    fprintf(outfp, "%10s|%10s|%20s|%15s|%15s|\n",
            "", (useDestFlag ? "dIP_Uniq" : "sIP_Uniq"),
            "Bytes", "Packets", "Records");
    fprintf(outfp, "%10s|%10u|%20llu|%15llu|%15llu|\n",
            "Total", totalHashRecords, tot_bytes, tot_packets, tot_flows);
    if (byteMin || byteMax || recMin || recMax || packetMin || packetMax) {
        fprintf(outfp,
                "%10s|%10u|%20llu|%15llu|%15llu|\n",
                "Qualifying", qual_ips, qual_bytes, qual_packets, qual_flows);
    }

    return 0;
}



/*
 *  int dumpIPSet (char *setFileName)
 *
 *    Dumps the ip addresses counted during normal operation to disk
 *    in ip tree format.
 */
static int dumpIPSet(char *setFileName)
{
    skIPTree_t *resultTree;

    /* Create the ip tree */
    countArrayToIPTree(&resultTree, countArray);

    /*
     * Okay, now we write to disk.
     */
    if (skIPTreeSave(setFileName, resultTree)) {
        exit(EXIT_FAILURE);
    };
    return 0;
}



/*
 * main invocation
 */
int main(int argc, char **argv)
{
    int counter;
    uint32_t hashIndex, recCount;
    countRecord_t *hashRecord;
    countRecord_t *rootRecord;
    rwIOStruct *rwIOS;
    int using_pager = 0;
    FILE *stream_out = stdout;

    counter = hashIndex = recCount = 0;

    appSetup(argc, argv);                 /* never returns */

    /* Read in records */
    for (counter = ioISP->firstFile; counter < ioISP->fileCount ; counter++) {
        rwIOS = rwOpenFile(ioISP->fnArray[counter], NULL);
        if (NULL == rwIOS) {
            exit(EXIT_FAILURE);
        }
        while (rwRead(rwIOS, &currentRecord)) {
            recCount++;
            hashIndex = IPHASH(currentRecord, useDestFlag);
            /*
             * standard hash foo - check to see if we've got a value.
             * If not, create and stuff.  If so, check to see if they
             * match - if so, stuff.  If not, move down until you do -
             * if you find nothing, stuff.
             */
            if (countArray[hashIndex] == NULL) {
                countArray[hashIndex] = calloc(1, sizeof(countRecord_t));
                if (countArray[hashIndex] == NULL) {
                    skAppPrintErr("Error allocating memory for record");
                    exit(EXIT_FAILURE);
                }
                /*
                 * No error on callocing, continue
                 */
                if (addRecord(countArray[hashIndex], &currentRecord)) {
                    skAppPrintErr("Error creating record, terminating");
                    exit(EXIT_FAILURE);
                }
                /*
                 * Alright, I'm modifying the record set so that it's circular,
                 * the explicit rationale for this is that a loop (rather than
                 * an array) will allow me to effectively rotate the "contact
                 * point of the loop" so that the last used is always
                 * at the countArray[hashIndex] position.
                 */
                countArray[hashIndex]->nextRecord = countArray[hashIndex];
            } else {
                hashRecord = countArray[hashIndex];
                rootRecord = hashRecord;
                while (!CMPFNC(currentRecord, hashRecord, useDestFlag) &&
                      (hashRecord->nextRecord != rootRecord))
                {
                    hashRecord = hashRecord->nextRecord;
                }
                /*
                 * Alright, we've either hit the end of the linked
                 * list or we've found the value.  We check and handle
                 * the end of list case first.
                 */
                if (!CMPFNC(currentRecord, hashRecord, useDestFlag)) {
                    if ((hashRecord->nextRecord =
                        (countRecord_t *) calloc(1, sizeof(countRecord_t)))
                       == NULL)
                    {
                        skAppPrintErr("Error allocating memory for record");
                        exit(EXIT_FAILURE);
                    }
                    hashRecord = hashRecord->nextRecord;
                    hashRecord->nextRecord = rootRecord; /* Restore the loop */
                    /* We've now implicitly created an empty
                     * record and dropped it at currentRecord.
                     */
                }
                countArray[hashIndex] = hashRecord;
                if (addRecord(countArray[hashIndex], &currentRecord)) {
                    skAppPrintErr("Error creating record, terminating");
                    exit(EXIT_FAILURE);
                }
            }
        }
        rwCloseFile(rwIOS);
    }

    /* Invoke the pager when appropriate. */
    switch (printMode) {
      case RWAC_PMODE_STAT:
      case RWAC_PMODE_IPSETFILE:
        break;
      case RWAC_PMODE_RECS:
      case RWAC_PMODE_IPS:
      case RWAC_PMODE_SORTED_RECS:
      case RWAC_PMODE_SORTED_IPS:
        using_pager = skOpenPagerWhenStdoutTty(&stream_out, &pager);
        if (using_pager < 0) {
            exit(EXIT_FAILURE);
        }
        break;
      case RWAC_PMODE_NONE:
      case RWAC_PMODE_RWACFILE:
        abort();
    }

    /* Produce the output */
    switch (printMode) {
      case RWAC_PMODE_STAT:
        dumpStats(stream_out);
        break;
      case RWAC_PMODE_IPSETFILE:
        dumpIPSet(outputSetFN);
        break;
      case RWAC_PMODE_RECS:
        dumpRecords(stream_out);
        break;
      case RWAC_PMODE_IPS:
        dumpIPs(stream_out);
        break;
      case RWAC_PMODE_SORTED_RECS:
        dumpRecordsSorted(stream_out);
        break;
      case RWAC_PMODE_SORTED_IPS:
        dumpIPsSorted(stream_out);
        break;
      case RWAC_PMODE_NONE:
      case RWAC_PMODE_RWACFILE:
        abort();
    };

    /* close the pager */
    if (using_pager) {
        skClosePager(stream_out, pager);
    }

    /* output */
    appTeardown();

    return (0);
}
