/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**
**  rwappend.c
**
**  Suresh L Konda
**  8/10/2002
**      Append f2..fn to f1 as long as the headers match exactly.
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwappend/rwappend.c,v 1.15 2005/06/10 18:18:12 thomasm Exp $");

#include "rwpack.h"
#include "utils.h"


/* local defines and typedefs */
#define NUM_REQUIRED_ARGS       2
#define BUF_SIZE 8192

/* LOCAL FUNCTIONS */

static void appUsage(void);            /* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);    /* never returns */

/* LOCAL VARIABLES */

static  int inoutFD, inFD;
static  rwFileHeaderV0 inoutHdr, inHdr;
static  int exitVal = 0;
static  uint8_t buffer[BUF_SIZE];


/*
 * appUsage:
 *      print usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
void appUsage(void)
{
    fprintf(stderr, "%s <input-fPath1> <input-fPath2> [<input-fPath3>...]\n",
            skAppName());
    exit(1);
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    if (inoutFD > -1) {
        close(inoutFD);
    }
    if (inFD > -1) {
        close(inFD);
    }
    inoutFD = -1;
    inFD = -1;

    return;
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    skAppRegister(argv[0]);

    /* first do some sanity checks */
    if (argc <= NUM_REQUIRED_ARGS) {
        appUsage();         /* never returns */
    }
    inoutFD = inFD = -1;

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("registering appTeardown() failed");
        appTeardown();
        exit(1);
    }

    return;                       /* OK */
}


int main(int argc, char **argv)
{
    const char *inoutPath, *inPath;
    ssize_t rLen, wLen;           /* read/write lengths */
    uint64_t cumRead = 0;
    uint64_t cumWrite = 0;
    int curFileIndex;
    int thisExitVal;
    int pad;

    appSetup(argc, argv);                 /* never returns on error */

    /* Open the first file (the file we will append to).  Cannot be
       compressed. */
    inoutPath = argv[1];
    if (strstr(inoutPath, ".gz")) {
        skAppPrintErr("File 1 '%s' cannot be a compressed file", inoutPath);
        exit(1);
    }
    if (! fileExists(inoutPath) ) {
        skAppPrintErr("File 1 '%s' does not exist", inoutPath);
        exit(1);
    }
    inoutFD = open(inoutPath, O_RDWR);
    if (inoutFD < 0) {
        skAppPrintErr("unable to open file '%s': %s",
                      inoutPath, strerror(errno));
        exit(1);
    }

    /* read header */
    rLen = read(inoutFD, &inoutHdr, sizeof(inoutHdr));
    if (rLen < (ssize_t)sizeof(inoutHdr)) {
        skAppPrintErr("File %s too small", inoutPath);
        exit(1);
    }

    /* check file types */
    switch (inoutHdr.gHdr.type) {
      case FT_RWACL:
      case FT_RWNOTROUTED:
      case FT_RWROUTED:
      case FT_RWSPLIT:
      case FT_RWWWW:
        /* these are all o.k. nothing to do */
        break;
      default:
        /* anything else must fail */
        skAppPrintErr("input file '%s' has unsupported type 0x%x",
                      inoutPath, inoutHdr.gHdr.type);
        exit(1);
    }

    /* ok - skip to eof on inoutFD */
    if ( lseek(inoutFD, 0, SEEK_END) < 0) {
        skAppPrintErr("Unable to skip to EOF on %s", inoutPath);
        exit(1);
    }

    /* Loop over files 2..N and append them to the first */
    for (curFileIndex = 2; curFileIndex < argc; curFileIndex++ ) {
        thisExitVal = 0;

        /* open the file */
        inPath = argv[curFileIndex];
        if (strstr(inPath, ".gz")) {
            skAppPrintErr("File %d '%s' cannot be a compressed file",
                          curFileIndex, inPath);
            thisExitVal = 1;
            continue;
        }
        if (! fileExists(inPath) ) {
            skAppPrintErr("File %d '%s' does not exist", curFileIndex, inPath);
            thisExitVal = 1;
            continue;
        }
        if (strcmp(inoutPath, inPath) == 0) {
            skAppPrintErr("File 1 and File %d are identical", curFileIndex);
            thisExitVal = 1;
            continue;
        }
        inFD = open(inPath, O_RDONLY);
        if (inFD < 0) {
            skAppPrintErr("unable to open file '%s': %s",
                          inPath, strerror(errno));
            thisExitVal = 1;
            continue;
        }

        /* read header */
        rLen = read(inFD, &inHdr, sizeof(inHdr));
        if (rLen < (ssize_t)sizeof(inHdr)) {
            skAppPrintErr("File %s too small, no header", inPath);
            thisExitVal = 1;
            continue;
        }
        cumRead = rLen;

        /* compare headers */
        if (memcmp(&inoutHdr, &inHdr, sizeof(inHdr))) {
            skAppPrintErr("Headers do not match. Current file '%s'. Abort",
                          inPath);
            thisExitVal = 1;
            continue;
        }

        /* read header padding, if any */
        pad = librwComputeHeaderPadding((genericHeaderPtr)&inHdr);
        if (pad) {
            rLen = read(inFD, buffer, pad);
            if (rLen < pad) {
                skAppPrintErr("File %s too small, no header", inPath);
                thisExitVal = 1;
                continue;
            }
        }
        cumRead += pad;

        /* CONSTANTCONDITION */
        while (1) {
            rLen = read(inFD, buffer, BUF_SIZE);
            if (rLen < 0) {
                /* io error */
                skAppPrintErr("read error on %s after reading %llu bytes: %s",
                              inPath, cumRead, strerror(errno));
                break;
            } else if (rLen == 0) {
                break;                  /* EOF */
            }
            cumRead += rLen;
            wLen = write(inoutFD, buffer, rLen);
            if (wLen != rLen) {
                if (wLen == -1) {
                    skAppPrintErr("write error on '%s' after %llu bytes\n"
                                  "\t (currently reading file %s): %s",
                                  inoutPath, (unsigned long long)cumWrite,
                                  inPath, strerror(errno));
                } else {
                    skAppPrintErr(("write/read mismatch: %llu read from '%s'"
                                   " vs %llu written to '%s'\n"
                                   "\t after %llu total bytes written"),
                                  (unsigned long long)rLen, inPath,
                                  (unsigned long long)rLen, inoutPath,
                                  (unsigned long long)cumRead);
                }
                break;
            }
            cumWrite += wLen;
        } /* inner read loop */
        exitVal += thisExitVal;
        close(inFD);
        inFD = -1;
    } /* outer for loop */

    /* done with all files */
    fprintf(stdout, "Appended %llu bytes to %s\n",
            (unsigned long long)cumWrite, inoutPath);
    appTeardown();

    return exitVal; /* NOTREACHED */
}
