/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwcat.c
**
**  12/13/2002
**
**  Suresh L. Konda
**
**  Stream out a bunch of input files--given on the command
**  line--and/or stdin to stdout or given file Path
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwcat/rwcat.c,v 1.15 2005/09/16 21:58:36 thomasm Exp $");

#include "rwpack.h"
#include "utils.h"


/* TYPEDEFS AND DEFINES */

/* file handle for --help output */
#define USAGE_FH stdout


/* LOCAL FUNCTIONS */

static void appUsage(void);
static void appUsageLong(void);
static void appOptionsUsage(FILE *);
static int  appOptionsHandler(clientData, int, char *);
static void appSetup(int, char**);
static void appTeardown(void);
static int  catFile(const char *inFName);


/* LOCAL VARIABLES */

/* index into argv of first filename to process */
static int firstFile;

/* whether to read records from stdin */
static int read_from_stdin = 0;

/* xargs input stream; currently only reading from stdin is allowed */
static FILE *xargs_stream = NULL;

/* output stream */
static rwIOStructPtr outIOS;


/* OPTIONS */

enum appOptionsEnum {
    RWCAT_OUTPUT_PATH, RWCAT_PRINT_FILENAMES, RWCAT_XARGS
};

static struct option appOptions[] = {
    {"output-path",     REQUIRED_ARG, 0, RWCAT_OUTPUT_PATH},
    {"print-filenames", NO_ARG,       0, RWCAT_PRINT_FILENAMES},
    {"xargs",           NO_ARG,       0, RWCAT_XARGS},
    {0,0,0,0}                     /* sentinal entry */
};

static char *appHelp[] = {
    "full path name of the output file. Def: stdout",
    "print input filenames while processing",
    "treat stdin as a list of files to read, one name per line",
    (char *)NULL
};

static struct {
    const char *output_path;     /* name of output path */
    const char *xargs;           /* name of stream to read filenames from */
    uint8_t     print_filenames; /* whether to print file names */
} rwcat_opt = {
    NULL, NULL, 0
};



/*
 *  appUsage();
 *
 *    Print a short usage message to stderr and exit the program.
 *    This is used when the user has made an input error.  Use the
 *    function appUsageLong() to print full usage information.
 */
void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 *  appUsageLong();
 *
 *    Prints full usage information to USAGE_FH and exits the program.
 *    This function should be passed to optionsSetup() to print the
 *    program's usage when the "--help" option given.
 */
static void appUsageLong(void)
{
    fprintf(USAGE_FH,
            ("Usage: %s [app-options] [<inputFiles>] \n"
             "\t Reads binary RW records from stdin or from named files and\n"
             "\t prints them to the specified output file or to stdout.\n"),
            skAppName());
    appOptionsUsage(USAGE_FH);
    exit(0);
}


/*
 *  appOptionsUsage(fh);
 *
 *    Print, to the specified file pointer, each switch in the
 *    appOptions[] array, whether the switch takes an argument, and
 *    the switch's help text.
 */
static void appOptionsUsage(FILE *fh)
{
    int i;
    fprintf(fh, "\napp options\n");
    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }

    return;
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    switch (opt_index) {
      case RWCAT_OUTPUT_PATH:
        /* output file path: treat "stdout" as standard output (which is
         * the default); anything else as file */
        if (0 != strcmp(opt_arg, "stdout")) {
            /* check for multiple output arguments */
            if (NULL != rwcat_opt.output_path) {
                skAppPrintErr("multiple output files '%s', '%s'\n",
                              rwcat_opt.output_path, opt_arg);
                return 1;
            }
            rwcat_opt.output_path = strdup(opt_arg);
            if (!rwcat_opt.output_path) {
                skAppPrintErr("appOptionsHandler: strdup failed\n");
                exit(EXIT_FAILURE);
            }
        }
        break;

      case RWCAT_PRINT_FILENAMES:
        /* print input file names */
        rwcat_opt.print_filenames = 1;
        break;

      case RWCAT_XARGS:
        rwcat_opt.xargs = "stdin";
        xargs_stream = stdin;
        break;

      default:
        skAppPrintErr("Unhandled option %d in options handler at %s:%d",
                      opt_index, __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    } /* switch */

    return 0;                     /* OK */
}


/*
 *  appSetup(argc, argv)
 *    do all the setup for this application include setting up
 *    required modules etc.
 *  Arguments:
 *    argc, argv
 *  Returns: None.
 *  Side Effects:
 *    exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int i;
    int rv;

    skAppRegister(argv[0]);

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* initialize options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);/* never returns */
    }

    /* register the options handler */
    if (optionsRegister(appOptions, (optHandler) appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);
    }

    /* parse options */
    firstFile = optionsParse(argc, argv);
    assert(firstFile <= argc);
    if (firstFile < 0) {
        appUsage();         /* never returns */
    }

    /* Make certain we have some input.  If xargs was given, stdin is
     * a list of filenames for input.  Else if stdin is not a
     * terminal, read records from stdin.  Else if no files, no xargs,
     * and stdin is a terminal, complain about no input. */
    if (xargs_stream == NULL) {
        if ( !FILEIsATty(stdin)) {
            read_from_stdin = 1;
        } else if (firstFile == argc) {
            /* no file names were given. */
            skAppPrintErr("no file names given and stdin is a terminal");
            appUsage();
        }

        /* If file names given and one matches "stdin", disable
         * automatic reading from stdin, since user explicitly
         * requested stdin.  We do this to maintain consistency with
         * past behavior when we did not automatically read from
         * stdin. */
        for (i = firstFile; i < argc; ++i) {
            if (0 == strcmp("stdin", argv[i])) {
                if ( !read_from_stdin ) {
                    skAppPrintErr("stdin is connected to a terminal");
                    appUsage();
                }
                read_from_stdin = 0;
                break;
            }
        }
    }

    /* stdout is our default output path */
    if (NULL == rwcat_opt.output_path) {
        if (FILEIsATty(stdout)) {
            skAppPrintErr("stdout is connected to a terminal");
            appUsage();
        }
        rwcat_opt.output_path = "stdout";
    }

    /* open output */
    if ((rv = rwioCreate(&outIOS, rwcat_opt.output_path, SK_RWIO_WRITE))
         || (rv = rwioOpen(outIOS)))
    {
        rwioPrintLastErr(outIOS, rv, &skAppPrintErr);
        skAppPrintErr("Couldn't open output file.  Exiting.");
        exit(EXIT_FAILURE);
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    (void) umask((mode_t) 0002);
    return;/* OK */
}


/*
 * appTeardown:
 *      Teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 * All modules are torn down. Then application teardown takes place.
 * Static variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static uint8_t teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    if (outIOS) {
        rwCloseFile(outIOS);
    }

    optionsTeardown();

    return;
}


/*
 * catFile:
 *      Dump all records to outF
 */
static int catFile(const char *inFName)
{
    rwRec rwrec;
    rwIOStructPtr inIOS;
    uint32_t inC = 0;
    uint32_t outC = 0;

    inIOS = rwOpenFile(inFName, NULL);
    if (NULL == inIOS) {
        /* some error. the library would have dumped a msg */
        return 1;
    }

    if (rwcat_opt.print_filenames) {
        fprintf(stderr, "%s\n", inFName);
    }
#if     1                       /* set to 0  for testing */
    while (rwRead(inIOS, &rwrec)) {
        int rv;
        inC++;
        rv = rwWrite(outIOS, &rwrec);
        if (0 != rv) {
            rwioPrintLastErr(outIOS, rv, &skAppPrintErr);
            break;
        }
        outC++;
    }
#endif
    rwCloseFile(inIOS);
    inIOS = (rwIOStructPtr)NULL;
    if (rwcat_opt.print_filenames) {
        fprintf(stderr, "Read %u Wrote %u\n", inC, outC);
    }

    return 0;
}


/*
 *  status = catFilenamesFromStream(stream);
 *
 *    Treats 'stream' as containing a list of filenames, one per line.
 *    Calls catFile() for each filename (line of input) that is read.
 */
static int catFilenamesFromStream(FILE *stream)
{
    char line[1024];
    char *eol;

    while (!feof(stream)) {
        if (!fgets(line, sizeof(line), stream)) {
            continue;
        }
        eol = strchr(line, '\n');
        if (eol == line) {
            /* empty line; ignore */
            continue;
        }

        if (NULL != eol) {
            /* expected behavior: read an entire line */
            *eol = '\0';
        } else if (feof(stream)) {
            /* okay: last line did not end in newline */
        } else {
            /* bad: line was longer than sizeof(line).  read until
             * newline or eof, then through away line */
            skAppPrintErr("--xargs input line too long. ignored");
            while (fgets(line, sizeof(line), stream) && !strchr(line,'\n'))
                ; /* empty */
            continue;
        }

        /* Don't read binary data from the file we are currently using
         * as the input stream to xargs */
        if (0 == strcmp(line, rwcat_opt.xargs)) {
            skAppPrintErr("Ignoring filename '%s' in xargs input",
                          line);
            continue;
        }

        catFile(line);
    }

    return 0;
}


/*
 *  status = rwcatMain(argc, argv);
 *
 *    Call catFile() on each file, whatever its source.
 */
static int rwcatMain(int argc, char **argv)
{
    int counter;

    if (xargs_stream != NULL) {
        return catFilenamesFromStream(xargs_stream);
    }

    if (read_from_stdin) {
        catFile("stdin");
    }
    for(counter = firstFile; counter < argc ; counter++) {
        if (catFile(argv[counter])) {
            break;
        }
    }

    return 0;
}


int main(int argc, char ** argv)
{
    int rv;

    appSetup(argc, argv);
    rv = rwcatMain(argc, argv);
    appTeardown();

    return(rv);
}
