/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwcountutils.c
**
**  utility routines for rwcount.c
**
**  Michael P. Collins
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwcount/rwcountutils.c,v 1.19 2005/10/26 14:23:17 thomasm Exp $");

#include "rwpack.h"
#include "utils.h"
#include "iochecks.h"
#include "rwcount.h"


/* where to send --help output */
#define USAGE_FH stdout


/* required externs */
extern countFile countData;

/*
 * I/O stuff
 */

extern rwIOStructPtr rwIOSPtr;
extern iochecksInfoStructPtr ioISP;


static void appUsageLong(void);
void appTeardown(void);
void appOptionsUsage(FILE*);
int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);


/* OPTIONS SETUP */

typedef enum {
    RWCO_BIN_SIZE, RWCO_LOAD_SCHEME, RWCO_SKIP_ZEROES,
    RWCO_DELIMITER, RWCO_PRINT_FILENAMES, RWCO_NO_TITLES,
    RWCO_EPOCH_SLOTS, RWCO_BIN_SLOTS, RWCO_START_EPOCH,
    RWCO_PAGER
} appOptionsEnum;

static struct option appOptions[] = {
    /* Size of the bins in seconds */
    {"bin-size",        REQUIRED_ARG, 0, RWCO_BIN_SIZE},
    /* Bin splitting scheme */
    {"load-scheme",     REQUIRED_ARG, 0, RWCO_LOAD_SCHEME},
    {"skip-zeroes",     NO_ARG,       0, RWCO_SKIP_ZEROES},
    {"delimiter",       REQUIRED_ARG, 0, RWCO_DELIMITER},
    {"print-filenames", NO_ARG,       0, RWCO_PRINT_FILENAMES},
    {"no-titles",       NO_ARG,       0, RWCO_NO_TITLES},
    {"epoch-slots",     NO_ARG,       0, RWCO_EPOCH_SLOTS},
    {"bin-slots",       NO_ARG,       0, RWCO_BIN_SLOTS},
    {"start-epoch",     REQUIRED_ARG, 0, RWCO_START_EPOCH},
    {"pager",           REQUIRED_ARG, 0, RWCO_PAGER},
    {0,0,0,0}                     /* sentinal entry */
};

static const char *appHelp[] = {
    "Size of bins in seconds. Def. 30 seconds",
    ("Defines bin loading, options are:\n"
     "\t0: split values evenly across the bins the record covers\n"
     "\t1: fill first appropriate bin with values [Default]\n"
     "\t2: fill last appropriate bin\n"
     "\t3: fill centermost bin"),
    "Do not print bins that have no flows. Def. Print all",
    "Delimiter between columns. Def. '|'",
    "Print names of input files as they are opened. Def. No",
    "Do not print column titles. Def. Print titles",
    "Print bin labels using epoch time. Def. Human readable",
    "Print bin labels using the internal bin index. Def. No",
    "Print bins from this time forward. Def. First nonzero bin",
    "Program to invoke to page output. Def. $SILK_PAGER or $PAGER",
    (char *)NULL
};




/*
 * appUsage:
 * print usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 *      print usage information to USAGE_FH and exit with code 1.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                             \
    ("Usage: %s [options] [files]\n"                                          \
     "\tSummarize SiLK Flow records across time, producing textual output\n"  \
     "\twith counts of bytes, packets, and flow records for each time bin.\n" \
     "\tWhen no files given on command line, flows are read from STDIN.\n")

    fprintf(USAGE_FH, USAGE_MSG, skAppName());
    appOptionsUsage(USAGE_FH);
    exit(EXIT_FAILURE);
}


/*
 * appOptionsUsage:
 * print help for options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *  do NOT add non-option usage here (i.e., required and other optional args)
 */
void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\noptions\n");
    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }
    return;
}


/*
 *  appSetup
 *      do all the setup for this application include setting up required
 *       modules etc.
 *  Arguments:
 *      argc, argv
 *  Returns: None.
 *  Side Effects:
 *      exits with code 1 if anything does not work.
 */
void appSetup(int argc, char **argv)
{
    skAppRegister(argv[0]);

    /* make sure count of option's declarations and help-strings match */
    assert((sizeof(appOptions)/sizeof(struct option)) ==
           (sizeof(appHelp)/sizeof(char *)));

    /* Setup the options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);
    }

    /* Register our options and handler */
    if (optionsRegister(appOptions, (optHandler) appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);
    }

    ioISP = iochecksSetup(0, 0, argc, argv);

    /* parse options; print usage if error */
    ioISP->firstFile = optionsParse(argc, argv);
    if (ioISP->firstFile < 0) {
        appUsage();
    }

    /* Use STDIN as an input stream if it is not a TTY; make certain
     * we have some input and we are either reading from STDIN or
     * using files listed the command line, but not both. */
    if (iochecksAcceptFromStdin(ioISP) || iochecksInputs(ioISP, 0)) {
        appUsage();
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    (void) umask((mode_t) 0002);

    return;/* OK */
}


/*
 *  appOptionsHandler:
 *      called for each option the app has registered.
 *  Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 *  Returns:
 *      0 if OK. 1 else.
 *  Side Effects:
 *      Relevant options are set.
 */
int  appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    uint32_t opt_value;

    switch ((appOptionsEnum)opt_index) {
      case RWCO_LOAD_SCHEME:
        if (skStringParseUint32(&opt_value, opt_arg, 0, LOAD_MIDDLE)) {
            skAppPrintErr("unrecognized %s option '%s'.  Choose 0,1,2, or 3",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        switch ((bin_load_scheme_enum_t)opt_value) {
          case LOAD_MEAN:
          case LOAD_START:
          case LOAD_END:
          case LOAD_MIDDLE:
            countData.fileHeader->bin_load_scheme = opt_value;
            break;
        }
        break;

      case RWCO_BIN_SIZE:
        if (0 != skStringParseUint32(&opt_value, opt_arg, 1, 0))
        {
            skAppPrintErr("Invalid value for %s: '%s'.",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        countData.fileHeader->binSize = opt_value;
        break;

      case RWCO_SKIP_ZEROES:
        countData.fileHeader->skip_zeroes = 1;
        break;

      case RWCO_PRINT_FILENAMES:
        printFNameFlag = 1;
        break;

      case RWCO_DELIMITER:
        countData.fileHeader->delimiter = opt_arg[0];
        break;

      case RWCO_NO_TITLES:
        countData.fileHeader->no_titles = 1;
        break;

      case RWCO_EPOCH_SLOTS:
        countData.fileHeader->bin_label_flag = BIN_LABEL_EPOCH;
        break;

      case RWCO_BIN_SLOTS:
        countData.fileHeader->bin_label_flag = BIN_LABEL_INDEX;
        break;

      case RWCO_START_EPOCH:
        /* try to parse as epoch time; if that fails try as YYYY/MM/DD */
        if ((0 != skStringParseUint32(&opt_value, opt_arg, 1, 0))
            && (0 != skStringParseDatetime(&opt_value, opt_arg, NULL)))
        {
            skAppPrintErr("Invalid value for %s: '%s'.",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        countData.fileHeader->start_epoch = opt_value;
        break;

      case RWCO_PAGER:
        pager = opt_arg;
        break;
    }

    return 0;                     /* OK */
}


/*
 * appTeardown:
 *      Teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 * All modules are torn down. Then application teardown takes place.
 * Static variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
void appTeardown(void)
{
    static uint8_t teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    iochecksTeardown(ioISP);
    optionsTeardown();

    if (rwIOSPtr) {
        rwCloseFile(rwIOSPtr);
    }
    rwIOSPtr = (rwIOStructPtr )NULL;
    return;
}

