/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**
**  10/14/2002
** Suresh L. Konda
**
** fglob.c
**
** routines for globbing files in the special dir hierarchy we have for
** packed files.
**
** There are three externally visible routines:
**  fglobSetup(): setup stuff
**  fglobNext():  call repeatedly till no more files
**  fglobTeardown(): clean up
**  fglobValid(): 1 if user specified any params; 0 else.
**
**  There is also a set of GetXX routines to gain access to the internals
**  of the fglob data structure.
**
** Details are given as block comments before these routines.
**
*/

/* Include site-specific declarations */
#define DECLARE_FGLOB_VARIABLES
#define DECLARE_CLASSINFO_VARIABLES
#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwfilter/fglob.c,v 1.13 2005/10/18 16:03:47 thomasm Exp $");

#include "rwfilter.h"
#include "cfg_config.h"
#include "silk_site.h"


/* LOCAL TYPEDEFS AND DEFINES */

/* Whether to support the configuration file */
#ifndef SK_FGLOB_CONFIG_FILE
#define SK_FGLOB_CONFIG_FILE  0
#endif

/* name of configuration file */
#define CFG_CONFIGURATION_FILE "silk.cfg"

/* property names for defaults */
#define CFG_DEFAULT_DOMAIN     "DEFAULT-DOMAIN"
#define CFG_DEFAULT_DIRECTION  "DEFAULT-DIRECTION"
#define CFG_DEFAULT_ACTION     "DEFAULT-ACTION"

/* Maximum string length of direction:action */
#define CFG_FLOWPATH_MAX_STRLEN  128

/*
 * Whether the --class switch supports multiple classes.  Set this to
 * NUM_CLASSES (defined in silk_site_*.h) for multiple class support,
 * or 1 to support a single class.  THIS SHOULD NOT BE 0.
 */
#define FGLOB_NUM_CLASSES 1
#ifndef FGLOB_NUM_CLASSES
#define FGLOB_NUM_CLASSES NUM_CLASSES
#endif

/* Output handle for --print-missing-files output */
#define MISSING_FH stderr

/* FIXME: WHERE TO PUT THIS??? */
/* Maximum number of directions and actions. */
#define FGLOB_NUM_DIRECTIONS  NUM_FLOW_TYPES
#define FGLOB_NUM_ACTIONS     NUM_FLOW_TYPES

/*
 *  Structure for all pertinent information for a given find request.
 */
typedef struct fglobListStruct {
    /* flag: number of globbing options specified */
    int8_t          fg_user_option_count;
    /* flag: 1 if the globber was partly initialized; 2 if completely init'd */
    int8_t          fg_initialized;
    /* flag: 1 to print msg about existence of each file */
    int8_t          fg_missing;
    /* flag: 1 to verify sensor names in configuration file */
    int8_t          fg_verify_sensors;

    /* which adjustCounters*() function to call */
    int           (*fg_adjust_counters_fn)(void);

    /* configuration file info */
    cfg_config_t   *fg_config_file;
    /* the domain to use from the config file*/
    char           *fg_domain;
    /* flowpath information from config file for current direction/action */
    cfg_array_t    *fg_path_group;
    /* current flowpath; index into fg_path_group */
    int             fg_path_group_idx;

    /* specified directions; strings from user's input */
    char           *fg_direction_list[FGLOB_NUM_DIRECTIONS];
    /* number of directions designated: length of fg_direction_list */
    int             fg_direction_count;
    /* current direction when globbing; index into fg_direction_list */
    int             fg_direction_idx;
    /* copy of the user's input to --direction switch, parsed into words */
    char           *fg_direction;

    /* specified actions; strings from user's input */
    char           *fg_action_list[FGLOB_NUM_ACTIONS];
    /* number of actions designated: length of fg_action_list */
    int             fg_action_count;
    /* current action when globbing; index into fg_action_list */
    int             fg_action_idx;
    /* copy of the user's input to --action switch, parsed into words */
    char           *fg_action;

    /* specified class/types; values are indexes into the outFInfo[]
     * array that is defined in the silk_site_*.h headers. */
    int             fg_classtype_list[NUM_FLOW_TYPES];
    /* number of class/type pairs designated: length of fg_classtype_list */
    int             fg_classtype_count;
    /* current class/type when globbing; index into fg_classtype_list */
    int             fg_classtype_idx;

    /* list of sensor name's wanted for each class/type: pointers into
     * the sensorInfo[] array defined in the silk_site_*.h headers. */
    char           *fg_sensor_list[NUM_FLOW_TYPES][1+SENSOR_COUNT];
    /* for each class/type, number of sensors: length of fg_sensor_list[i] */
    int             fg_sensor_count[NUM_FLOW_TYPES];
    /* current sensor when globbing; index into
     * fg_sensor_list[fg_classtype_idx] */
    int             fg_sensor_idx;

    /* start, end, and current julian dates */
    int32_t         fg_julian_start;
    int32_t         fg_julian_end;
    int32_t         fg_julian_idx;

    /* current date */
    int32_t         fg_year;
    int32_t         fg_month;
    int32_t         fg_day;

    /* start, end, current hour and number of hours */
    int             fg_hour_start;
    int             fg_hour_end;
    int             fg_hour_idx;
    int             fg_hour_count;

    /* current directory path */
    char            fg_dir_path[PATH_MAX];

    /* name of current file */
    char            fg_file_name[PATH_MAX];

    /* user's input to various options.  This will be an array of
     * char*, index by the FGLOB_OPT_* enum.  The array will be
     * created once we know the number of options.*/
    char          **fg_option;

} fglobListStruct_t;



/* INTERNAL FUNCTIONS */

static int adjustCountersClassType(void);
#if SK_FGLOB_CONFIG_FILE
static int getPathGroup(void);
static int adjustCountersDirectionAction(void);
#endif

static int checkTypesForClass(void);
static int checkClass(int *num_classes, int class_list[FGLOB_NUM_CLASSES]);
static int checkSensors(void);

#if SK_FGLOB_CONFIG_FILE
static int parseConfigurationFile(void);
static int checkDomainDirectionAction(void);
#endif

static int fglobInit(void);
/* static void removeNullClasses(void); */
static int initTimes(void);
static int fglobHandler(clientData cData, int opt_index, char *opt_arg);
static void enableAllSensors(void);
static int parseSensorCmdLine(
    const char *,
    BITMAP_DECLARE(sensor_bits, (SENSOR_INVALID_SID+1)));


/* INTERNAL VARIABLES */

static fglobListStruct_t fglob_list;
static fglobListStruct_t * const fList = &fglob_list;


/* OPTIONS SETUP */

enum fglobOptionEnum {
    FGLOB_OPT_CLASS, FGLOB_OPT_TYPE, FGLOB_OPT_SENSORS,
#if SK_FGLOB_CONFIG_FILE
    FGLOB_OPT_DOMAIN, FGLOB_OPT_DIRECTION, FGLOB_OPT_ACTION,
#endif
    FGLOB_OPT_START_DATE, FGLOB_OPT_END_DATE,
    FGLOB_OPT_PRINT_MISSING_FILES,
#if SK_FGLOB_CONFIG_FILE
    FGLOB_OPT_VERIFY_SENSORS,
#endif
    FGLOB_OPT_ROOTDIR
};

static struct option fglobOptions[] = {
    {"class",                REQUIRED_ARG, 0, FGLOB_OPT_CLASS},
    {"type",                 REQUIRED_ARG, 0, FGLOB_OPT_TYPE},
    {"sensors",              REQUIRED_ARG, 0, FGLOB_OPT_SENSORS},
#if SK_FGLOB_CONFIG_FILE
    {"domain",               REQUIRED_ARG, 0, FGLOB_OPT_DOMAIN},
    {"direction",            REQUIRED_ARG, 0, FGLOB_OPT_DIRECTION},
    {"action",               REQUIRED_ARG, 0, FGLOB_OPT_ACTION},
#endif
    {"start-date",           REQUIRED_ARG, 0, FGLOB_OPT_START_DATE},
    {"end-date",             REQUIRED_ARG, 0, FGLOB_OPT_END_DATE},
    {"print-missing-files",  NO_ARG,       0, FGLOB_OPT_PRINT_MISSING_FILES},
#if SK_FGLOB_CONFIG_FILE
    {"verify-sensors",       NO_ARG,       0, FGLOB_OPT_VERIFY_SENSORS},
#endif
    {"data-rootdir",         REQUIRED_ARG, 0, FGLOB_OPT_ROOTDIR},
    {0, 0, 0, 0}
};

static char *fglobHelp[] = {
    NULL, /* class:     text is generated by fglobUsage() */
    NULL, /* type:      text is generated by fglobUsage() */
    NULL, /* sensor:    text is generated by fglobUsage() */
#if SK_FGLOB_CONFIG_FILE
    NULL, /* domain:    text is generated by fglobUsage() */
    NULL, /* direction: text is generated by fglobUsage() */
    NULL, /* action:    text is generated by fglobUsage() */
#endif
    "YYYY/MM/DD[:HH]. Default start of today",
    "YYYY/MM/DD[:HH]. Default: start-date",
    "print the names of missing files to stderr",
#if SK_FGLOB_CONFIG_FILE
    "verify that the sensor names in config file are valid during processing",
#endif
    "root of directory tree containing packed data",
    (char *)NULL
};



/*
 *  is_ok = fglobSetup();
 *
 *    Initialize the values in the global fList structure: set
 *    everything to zero; set times to -1.  Register fglob options and
 *    handler.  Check the environment for variable giving the location
 *    of the data.
 *
 *    Return 0 if OK. 1 else.
 */
int fglobSetup(void)
{
    char *rootdir_env;

    /* Verify programmer set options and help structs correctly */
    assert((sizeof(fglobHelp)/sizeof(char *))
           == (sizeof(fglobOptions)/sizeof(struct option)));

    /* zero out everything */
    memset(fList, 0, sizeof(fglobListStruct_t));

    /* Create the array to hold the option strings */
    fList->fg_option = calloc(1, sizeof(fglobHelp));
    if (NULL == fList->fg_option) {
        skAppPrintErr("fglobSetup: out of memory");
        return 1;
    }

    /* set start/end days/hours to be able to trap whether user gave
     * it explicitly
     */
    fList->fg_hour_start = fList->fg_hour_end = -1;
    fList->fg_julian_start = fList->fg_julian_end = -1;

    /* Check for data-rootdir set in the user's environment */
    rootdir_env = getenv(SK_FGLOB_ENVAR);
    if (rootdir_env) {
        /* ignore leading whitespace */
        while (isspace((int)*rootdir_env)) {
            ++rootdir_env;
        }
        if (*rootdir_env) {
            /* Should we check that this directory exists, or let the
             * users shoot themselves in the foot? */
            silk_fglob_rootdir = strdup(rootdir_env);
            if (!silk_fglob_rootdir) {
                skAppPrintErr("fglobSetup: strdup failed. out of memory");
                return 1;
            }
        }
    }

    /* Sanity check the sensors */
    if (sensorCheckCounts()) {
        return 1;
    }

    /* register options */
    if (optionsRegister((void *)fglobOptions, fglobHandler,
                        (clientData) fList)) {
        skAppPrintErr("fglobSetup: Unable to register options");
        return 1;
    }

    return 0;                     /* OK */
}


/*
 *  fglobUsageClass(fh)
 *
 *    Print the usage for the --class switch to the specified file
 *    handle.
 */
static void fglobUsageClass(FILE *fh)
{
    unsigned int j;

    /* CLASS: Read classInfo to get list of classes; use
     * DEFAULT_CLASS macro to print default class. */

    /* print first class */
    fprintf(fh, "class (%s", classInfo[0].className);

    /* now print remaining classes */
    for (j = 1; j < NUM_CLASSES; ++j) {
        fprintf(fh, ",%s", classInfo[j].className);
    }

    /* print default class*/
    fprintf(fh, "). Default %s\n",
            classInfo[DEFAULT_CLASS].className);
}


/*
 *  fglobUsageType(fh)
 *
 *    Print the usage for the --type switch to the specified file
 *    handle.
 */
static void fglobUsageType(FILE *fh)
{
    unsigned int j, k;

    /* TYPE: For each class, read the list of types and
     * default list of types from classInfo.  The string
     * names of the types are in the outFInfo
     * structure. */

    fprintf(fh, "type varies by class:\n");

    /* loop over all the classes */
    for (j = 0; j < NUM_CLASSES; ++j) {
        /* print class name and first type for this class */
        fprintf(fh, "\t%s (%s",
                classInfo[j].className,
                outFInfo[classInfo[j].typeList[0]].dirPrefix);
        /* print remaining types for the class */
        for (k = 1; k < classInfo[j].numTypes; ++k) {
            fprintf(fh, ",%s",
                    outFInfo[classInfo[j].typeList[k]].dirPrefix);
        }
        /* print first default type for this class */
        fprintf(fh, "). Default %s",
                outFInfo[classInfo[j].defaultList[0]].dirPrefix);
        /* print remaining default types for this class */
        for (k = 1; k < classInfo[j].numDefaults; ++k) {
            fprintf(fh, ",%s",
                    outFInfo[classInfo[j].defaultList[k]].dirPrefix);
        }
        fprintf(fh, "\n");
    }
}


/*
 *  fglobUsageSensors(fh)
 *
 *    Print the usage for the --sensor switch to the specified file
 *    handle.
 */
static void fglobUsageSensors(FILE *fh)
{
    fprintf(fh, "comma separated list of sensor names and/or IDs(0-%u).\n",
            numSensors-1);
}


#if SK_FGLOB_CONFIG_FILE
/*
 *  fglobUsageDomain(fh)
 *
 *    Print the usage for the --domain switch to the specified file
 *    handle.
 */
static void fglobUsageDomain(FILE *fh)
{
    cfg_map_t *domains;
    cfg_map_t *props;
    char *p;
    int i;

    fprintf(fh, "domain");

    if (NULL == fList->fg_config_file) {
        goto END;
    }

    /* Get the properties and the domains.  Return if domains are missing */
    props = cfg_get_config_properties(fList->fg_config_file);
    domains = cfg_get_config_domains(fList->fg_config_file);

    if ((domains == NULL) || (cfg_get_map_length(domains) < 1)) {
        goto END;
    }

    /* print first domain */
    fprintf(fh, " (%s", cfg_get_map_key(domains, 0));

    /* print remaining domains */
    for (i = 1; i < cfg_get_map_length(domains); ++i) {
        fprintf(fh, ",%s", cfg_get_map_key(domains, i));
    }
    fprintf(fh, ").");

    /* print default domain */
    if (props) {
        p = (char*)cfg_map_assoc(props, CFG_DEFAULT_DOMAIN);
        if (p) {
            fprintf(fh, " Default %s", p);
        }
    }

  END:
    fprintf(fh, "\n");
}


/*
 *  fglobUsageDirectionAction(fh, dir_act)
 *
 *    Print the usage for the --direction or --action switch to the
 *    specified file handle, 'fh'.  The 'dir_act' parameter determines
 *    whether the output for --direction (dir_act ==
 *    FGLOB_OPT_DIRECTION) or --action (FGLOB_OPT_ACTION) is printed.
 */
static void fglobUsageDirectionAction(FILE *fh, int dir_act)
{
    cfg_map_t *domain_list;
    cfg_map_t *dom;
    cfg_map_t *props;
    char       uniques[FGLOB_NUM_DIRECTIONS][CFG_FLOWPATH_MAX_STRLEN];
    int        uniq_count;
    int        is_dupe;
    char      *path;
    size_t     len;
    char *p;
    int i, j, k;

    fprintf(fh, "%s varies by domain.",
            ((FGLOB_OPT_DIRECTION == dir_act) ? "direction" : "action"));

    /* Can't do anything else with a configuration file */
    if (NULL == fList->fg_config_file) {
        goto END;
    }

    /* Get the properties and the domains.  Return if domains are missing */
    props = cfg_get_config_properties(fList->fg_config_file);
    domain_list = cfg_get_config_domains(fList->fg_config_file);

    if ((domain_list == NULL) || (cfg_get_map_length(domain_list) < 1)) {
        goto END;
    }

    /* Print the default, if we find it */
    if (props) {
        p = (char*)cfg_map_assoc(props, ((FGLOB_OPT_DIRECTION == dir_act)
                                         ? CFG_DEFAULT_DIRECTION
                                         : CFG_DEFAULT_ACTION));
        if (p) {
            fprintf(fh, " Default %s", p);
        }
    }

    /* Loop 'i' over all of the available domains. */
    for (i = 0; i < cfg_get_map_length(domain_list); ++i) {

        /* get a domain, making certain it has data. */
        dom = cfg_lookup_config_domain(fList->fg_config_file,
                                       cfg_get_map_key(domain_list, i));
        if ((NULL == dom) || (cfg_get_map_length(dom) < 1)) {
            continue;
        }

        /* To avoid printing the same direction or action multiple
         * times, we store the names in the uniques array and we keep
         * a count of the number of unique ones see in uniq_count. */
        uniq_count = 0;

        /* Loop 'j' over all the Path names for this domain 'dom' */
        for (j = 0; j < cfg_get_map_length(dom); ++j) {

            /* Get the current path (direction:action) */
            path = cfg_get_map_key(dom, j);

            /* Point 'p' at the direction or action part as
             * appropriate; set 'len' to the strlen of that segment,
             * and then copy the direction or action to the uniques
             * array.
             */
            p = strchr(path, ':');
            if (!p) {
                /* Warning: missing ':' in path name */
                continue;
            }
            if (FGLOB_OPT_DIRECTION == dir_act) {
                len = p - path;
                p = path;
            } else {
                ++p;
                len = strlen(p);
            }

            if (len >= CFG_FLOWPATH_MAX_STRLEN) {
                /* Warning: string too long */
                continue;
            }
            strncpy(uniques[uniq_count], p, len);
            uniques[uniq_count][len] = '\0';

            /* Compare this direction/action with all the ones we've
             * previously seen in this domain so we don't print
             * dupes. */
            is_dupe = 0;
            for (k = 0; k < uniq_count; ++k) {
                if (0 == strcmp(uniques[uniq_count], uniques[k])) {
                    /* we've already seen it */
                    is_dupe = 1;
                    break;
                }
            }

            if ( !is_dupe ) {
                /* we haven't seen it before, so print it */
                if (uniq_count == 0) {
                    /* This is the first we've seen, so print the domain
                     * name and opening paren. */
                    fprintf(fh, "\n\t%s (", cfg_get_map_key(domain_list, i));
                } else {
                    fprintf(fh, ",");
                }
                fprintf(fh, "%s", uniques[uniq_count]);
                ++uniq_count;
            }
        }
        fprintf(fh, ")");
    }

  END:
    fprintf(fh, "\n");
}
#endif /* SK_FGLOB_CONFIG_FILE */


/*
 *  fglobUsage(fh);
 *
 *    Print, to the given file handle, the usage for the command-line
 *    switches provided by the fglob library.
 */
void fglobUsage(FILE *fh)
{
    int opt;

    fprintf(fh, ("\nSELECTION SWITCHES choose which files to read"
                 " from the data store:\n\n"));

#if SK_FGLOB_CONFIG_FILE
    parseConfigurationFile();
#endif

    for (opt = 0; fglobOptions[opt].name; ++opt) {
        fprintf(fh, "--%s %s: ", fglobOptions[opt].name,
                SK_OPTION_HAS_ARG(fglobOptions[opt]));
        if (fglobHelp[opt]) {
            /* Simple help text is in fglobHelp[] array */
            fprintf(fh, "%s\n", fglobHelp[opt]);
        } else {
            /* Must generate the help: use the enum value to branch */
            switch (fglobOptions[opt].val) {

              case FGLOB_OPT_CLASS:
                fglobUsageClass(fh);
                break;

              case FGLOB_OPT_TYPE:
                fglobUsageType(fh);
                break;

              case FGLOB_OPT_SENSORS:
                fglobUsageSensors(fh);
                break;

#if SK_FGLOB_CONFIG_FILE
              case FGLOB_OPT_DOMAIN:
                fglobUsageDomain(fh);
                break;

              case FGLOB_OPT_DIRECTION:
                fglobUsageDirectionAction(fh, fglobOptions[opt].val);
                break;

              case FGLOB_OPT_ACTION:
                fglobUsageDirectionAction(fh, fglobOptions[opt].val);
                break;
#endif /* SK_FGLOB_CONFIG_FILE */

              default:
                fprintf(stderr, "Programmer error: no help for %s\n",
                        fglobOptions[opt].name);
                abort();
            }
        } /* else */
    } /* for */

    fprintf(fh, "Root of directory tree to find packed data: %s\n",
            silk_fglob_rootdir);

    return;
}


/*
 *  fglobTeardown()
 *
 *    Free the elements in fList which were malloc'ed.  Do not free
 *    the fList struct itself---it is a global static.  Multiple calls
 *    to this function are handled gracefully.
 */
void fglobTeardown(void)
{
    /* Idempotency check. */
    if (0 == fList->fg_user_option_count) {
        return;
    }
    fList->fg_user_option_count = 0;

    /* clean up our strdup()'s */
    if (fList->fg_direction) {
        free(fList->fg_direction);
        fList->fg_direction = NULL;
    }
    if (fList->fg_action) {
        free(fList->fg_action);
        fList->fg_action = NULL;
    }
    /* can't clean up silk_fglob_rootdir because sometimes it is static */

    /* teardown the config file */
    if (fList->fg_config_file) {
#if SK_FGLOB_CONFIG_FILE
        cfg_config_destroy(fList->fg_config_file);
#endif
        fList->fg_config_file = NULL;
    }

    /* free the options */
    if (fList->fg_option) {
        free(fList->fg_option);
        fList->fg_option = NULL;
    }

    return;
}


/*
 *  filename = fglobNext();
 *
 *    Return the name of next available file.  Returns NULL if all
 *    files have been processed.
 *
 *    Uses the adjust_counters_fn to call the appropriate function
 *    depending on whether the --class/type/sensor or
 *    --domain/direction/action switches were given.
 *
 *    Will complete the initialization of the library if required.
 */
char *fglobNext(void)
{
    int len;

    if (!fList->fg_initialized) {
        if (fglobInit()) {
            /* error */
            return NULL;
        }
    }

    /* keep adjusting the counters until we find a file that exists */
    while (fList->fg_adjust_counters_fn()) {

        /* set the prefix: this the full directory path---root dir,
         * dirPrefix, year, month, day--plus the class/type-dependent
         * file prefix. */
        if ('\0' == *(fList->fg_dir_path)) {
            fileTypeInfo *ft;
            ft = &outFInfo[fList->fg_classtype_list[fList->fg_classtype_idx]];

            snprintf(fList->fg_dir_path, sizeof(fList->fg_dir_path),
                     "%s/%s/%04d/%02d/%02d/%s",
                     silk_fglob_rootdir, ft->pathPrefix,
                     fList->fg_year, fList->fg_month, fList->fg_day,
                     ft->filePrefix);
        }

        /* Create the file name from the directory prefix and the
         * sensor name, year, month, day, and hour.  Store the length
         * so we can append a compression suffix (".gz") to it. */
        len = snprintf(fList->fg_file_name, sizeof(fList->fg_file_name),
                       "%s-%s_%04d%02d%02d.%02d",
                       fList->fg_dir_path,
                       (fList->fg_sensor_list[fList->fg_classtype_idx]
                        [fList->fg_sensor_idx]),
                       fList->fg_year, fList->fg_month,
                       fList->fg_day, fList->fg_hour_idx);

        if (!fileExists(fList->fg_file_name)) {
            /* check for compressed version of file */
            strcpy(&(fList->fg_file_name[len]), ".gz");
            if (!fileExists(fList->fg_file_name)) {
                if (fList->fg_missing) {
                    fList->fg_file_name[len] = '\0';
                    fprintf(MISSING_FH, "Missing %s\n", fList->fg_file_name);
                }
                continue;
            }
        }

        return fList->fg_file_name;
    }

    return NULL;
}


/*
 *  is_ok = fglobInit();
 *
 *    This is an internal function and should be called to initialize
 *    things based on the options given by the user.  Hence this must
 *    be called after the options are parsed.  To make life easier for
 *    the application programmer, this routine is called by the first
 *    call to fglobNext().
 *
 *    Calls the appropriate functions to initialize the library based
 *    on whether the --class/type/sensor or --domain/direction/action
 *    switches were given.  If neither were given, uses the default
 *    --class/type.
 *
 *    Return 0 for OK. 1 else;
 */
static int fglobInit(void)
{
    fList->fg_initialized = 1;

    /* set up the times */
    if (initTimes()) {
        /* error */
        return 1;
    }

#if SK_FGLOB_CONFIG_FILE
    if (fList->fg_option[FGLOB_OPT_DOMAIN]
        || fList->fg_option[FGLOB_OPT_DIRECTION]
        || fList->fg_option[FGLOB_OPT_ACTION])
    {

        fList->fg_adjust_counters_fn = &adjustCountersDirectionAction;

        if (parseConfigurationFile()) {
            /* error */
            return 1;
        }
        if (checkDomainDirectionAction()) {
            /* error */
            return 1;
        }
    } else
#endif /* SK_FGLOB_CONFIG_FILE */
    {
        fList->fg_adjust_counters_fn = &adjustCountersClassType;

        if (checkTypesForClass()) {
            /* error */
            return 1;
        }
        /* now check if the sensor list is valid given the class/type. */
        if (checkSensors()) {
            /* error */
            return 1;
        }
    }

    return 0;
}


/*
 *  is_ok = fglobHandler(cData, opt_index, opt_arg);
 *
 *    Called by options processor.  Return 1 on error, 0 on success.
 */
static int fglobHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    fList->fg_user_option_count++;

    switch (opt_index) {
      case FGLOB_OPT_CLASS:
      case FGLOB_OPT_TYPE:
      case FGLOB_OPT_SENSORS:
#if SK_FGLOB_CONFIG_FILE
      case FGLOB_OPT_DOMAIN:
      case FGLOB_OPT_DIRECTION:
      case FGLOB_OPT_ACTION:
#endif
      case FGLOB_OPT_START_DATE:
      case FGLOB_OPT_END_DATE:
        if (fList->fg_option[opt_index] != NULL) {
            skAppPrintErr("option --%s given multiple times",
                          fglobOptions[opt_index].name);
            return 1;
        }
        fList->fg_option[opt_index] = opt_arg;
        break;

      case FGLOB_OPT_PRINT_MISSING_FILES:
        fList->fg_missing = 1;
        break;

#if SK_FGLOB_CONFIG_FILE
      case FGLOB_OPT_VERIFY_SENSORS:
        fList->fg_verify_sensors = 1;
        break;
#endif

      case FGLOB_OPT_ROOTDIR:
        if (!dirExists(opt_arg)) {
            skAppPrintErr("Root data directory '%s' does not exist", opt_arg);
            return 1;
        }
        silk_fglob_rootdir = strdup(opt_arg);
        if (!silk_fglob_rootdir) {
            skAppPrintErr("fglobHandler: strdup failed: out of memory!");
            exit(1);
        }
        break;

      default:
        skAppPrintErr("fglobHandler: invalid index %d", opt_index);
        abort(); /* programmer error */

    } /* switch */

#if SK_FGLOB_CONFIG_FILE
    /* Make certain user is either using class/type/sensor switches or
     * direction/action switches but not a combination */
    if (((fList->fg_option[FGLOB_OPT_DOMAIN] != NULL)
         || (fList->fg_option[FGLOB_OPT_DIRECTION] != NULL)
         || (fList->fg_option[FGLOB_OPT_ACTION] != NULL))
        && ((fList->fg_option[FGLOB_OPT_CLASS] != NULL)
            || (fList->fg_option[FGLOB_OPT_TYPE] != NULL)
            || (fList->fg_option[FGLOB_OPT_SENSORS] != NULL)))
    {
        skAppPrintErr("The class/type/sensor switches are not compatiable\n"
                      "\t with the domain/direction/action switches.\n"
                      "\t Use one set or the other");
        return 1;
    }
#endif /* SK_FGLOB_CONFIG_FILE */

    return 0;                     /* OK */
}


/*
 *  have_more_data = adjustCountersClassType();
 *
 *    Go to the next (or first if fg_initialized is < 2) sensor, or
 *    type, or class, or hour, or day.  Return 0 if there is nothing
 *    else to iterate over, or 1 otherwise.
 *
 *    fglobInit() must be called before calling this function.
 *
 *    When fg_initialized < 2, the function assumes this is the first
 *    time it has been called and it will complete the initialization
 *    of the counters that was started by fglobInit(); otherwise it
 *    will increment the counters.
 */
static int adjustCountersClassType(void)
{
    if (fList->fg_initialized < 2) {
        /* was init called? */
        assert(fList->fg_initialized);

        /* This is the first call into this function.  All of the
         * indexes are at zero; there is nothing we need to lookup, so
         * simply set the "we've been here" flag and return. */
        fList->fg_initialized = 2;

        return 1;
    }

    /* This is not the first call into this function; need to
     * increment or reset the appropriate counters. */

    /* Make certain we haven't reached the end of the data. */
    if (fList->fg_hour_count < 0) {
        return 0;
    }

    /* First, see if we can increment the sensor */
    fList->fg_sensor_idx++;
    if(fList->fg_sensor_idx < fList->fg_sensor_count[fList->fg_classtype_idx]){
        /* we're good: looking at next sensor--same class, type & time */
        return 1;
    }
    /* On last sensor; reset and try class/type */
    fList->fg_sensor_idx = 0;

    /* Optimization: we are going to a different directory; set the
     * first character of fList->fg_dir_path to '\0' */
    *(fList->fg_dir_path) = '\0';

    /* Try to increment the class/type */
    if (fList->fg_classtype_count > 1) {
        fList->fg_classtype_idx++;
        if (fList->fg_classtype_idx < fList->fg_classtype_count) {
            /* we're good: looking at next class/type -- same time */
            return 1;
        }
        /* On last class/type; reset and try time */
        fList->fg_classtype_idx = 0;
    }

    /* Finally, increment the time: go to next hour */
    fList->fg_hour_count--;
    if (fList->fg_hour_count <= 0)
    {
        /* We're done. Make certain no more processing occurs */
        fList->fg_hour_count = -1;
        return 0;  /* done */
    }

    fList->fg_hour_idx++;
    if (fList->fg_hour_idx <= 23) {
        /* we're good: next hour, same day */
        return 1;
    }
    /* On last hour of the day; reset hour and move to next day */
    fList->fg_hour_idx = 0;
    fList->fg_julian_idx++;

    /* set the yr, mo, day from the fg_julian_idx */
    gregorianDate(fList->fg_julian_idx, &fList->fg_year,
                  &fList->fg_month, &fList->fg_day);

    return 1;
}


#if SK_FGLOB_CONFIG_FILE
/*
 *  found_it = getPathGroup();
 *
 *    This function finds the next VALID entry in the path-group list
 *    starting with the current entry.  If found, a positive value is
 *    returned and the fg_classtype_list[0] and fg_sensor_list[0][0]
 *    fields on fList are set to point to the correct values so a
 *    file-existence test can be performed.
 *
 *    If the current entry is not valid, getPathGroup() will search
 *    the remainder of the path-group list until if finds a valid
 *    entry or reaches the end of the path-group list.  If the end of
 *    the list is reached, the current path-group is destroyed,
 *    counters are reset, and 0 is returned.
 *
 *    An entry in the flow-group list is considered invalid when (1)it
 *    lists a flow-path that does not appear in the outFInfo[] array
 *    defined in silk_site_*.h, or when (2)the user has specified the
 *    --verify-sensors option and the flow-path lists a sensor that is
 *    not listed in the sensorInfo[] array also defined in
 *    silk_site_*.h.
 *
 *    If path-group is NULL, try to lookup a path-group from the
 *    configuration file using the current action and direction.  If
 *    the lookup fails, return 0 so that the next action or direction
 *    can be used.  If the lookup succeeds, find the first VALID entry
 *    in the path-group.
 */
static int getPathGroup(void)
{
    cfg_path_t     *path;
    int             out_f_info_index;
    unsigned int    j;

    /* If the path-group is NULL, attempt to lookup a new one from the
       config-file */
    if (fList->fg_path_group == NULL) {
        char flowpath[CFG_FLOWPATH_MAX_STRLEN];

        /* Create a flowpath; we can use sprintf() here since
         * we've already checked the lengths */
        sprintf(flowpath, "%s:%s",
                fList->fg_direction_list[fList->fg_direction_idx],
                fList->fg_action_list[fList->fg_action_idx]);
        fList->fg_path_group= cfg_config_lookup_date(fList->fg_config_file,
                                                     fList->fg_domain,
                                                     flowpath,
                                                     fList->fg_julian_idx,
                                                     fList->fg_hour_idx);
        if (fList->fg_path_group == NULL) {
            /* We either found a direction:action pair that is
             * invalid--which we warned the user about in
             * checkDomainDirectionAction()--or there is no data for
             * that particular direction:action pair for this hour.
             * In either case, return 0 here so we will increment the
             * counters. */
            return 0;
        }
    }

    while(fList->fg_path_group_idx<cfg_get_array_length(fList->fg_path_group)){

        /* get current path-group from path-group list */
        path = cfg_get_array_element(fList->fg_path_group,
                                     fList->fg_path_group_idx);

        /* Try to convert the path-group into a classtype index that
         * points into the outFInfo[] array. */
        out_f_info_index = -1;
        for (j = 0; j < NUM_FLOW_TYPES; ++j) {
            if(0==strcmp(outFInfo[j].filePrefix,cfg_get_path_flowlabel(path))){
                out_f_info_index = j;
                break;
            }
        }

        /* If path-group not found in outFInfo[], warn and try next
         * path-group */
        if (out_f_info_index == -1) {
            skAppPrintErr(("WARNING! config file error:"
                           "flowlabel '%s' not listed in silk_site_*.h"),
                          cfg_get_path_flowlabel(path));
            ++fList->fg_path_group_idx;
            continue;
        }

#if 0
        /* handle web data specially: if packType is web but the
         * user's query excludes web data (e.g, proto != 6), then skip
         * over the web files.  */
        if (outFInfo[out_f_info_index].packType == FT_RWWWW) {
            if () {
            }
        }
#endif

        /* If user wants to verify sensors, try to find it in the
         * sensorInfo[] list.  This lookup is not required--and thus a
         * user-option--since files use the sensor name which we
         * already have. */
        if (fList->fg_verify_sensors) {
            int sensor_idx = -1;
            /* find the sensor in sensorInfo */
            for (j = 0; j < numSensors; ++j) {
                if (0 == strcmp(sensorInfo[j].sensorName,
                                cfg_get_path_sensor(path)))
                {
                    sensor_idx = j;
                    break;
                }
            }
            if (sensor_idx == -1) {
                skAppPrintErr(("WARNING! config file error:"
                               "  sensor '%s' not listed in silk_site_*.h"),
                              cfg_get_path_sensor(path));
                /* Return here or just increment? */
                ++fList->fg_path_group_idx;
                continue;
            }
        }

        /* "Fake" fglob into believing that it is looking at the next
         * class/type/sensor combination. */
        fList->fg_classtype_list[0] = out_f_info_index;
        fList->fg_sensor_list[0][0] = cfg_get_path_sensor(path);
        fList->fg_path_group_idx++;
        return 1;
    }

    /* On last path-group; destroy the list, reset the counters, and
     * return. */
    cfg_array_path_destroy(fList->fg_path_group);
    fList->fg_path_group = NULL;
    fList->fg_path_group_idx = 0;

    return 0;
}


/*
 *  have_more_data = adjustCountersDirectionAction();
 *
 *    Go to the next (or first if fg_initialized < 2) flowpath, or
 *    action, or direction, or hour, or day.  Return 0 if there is
 *    nothing else to iterate over, or 1 otherwise.
 *
 *    fglobInit() must be called before calling this function.
 *
 *    When fg_initialized < 2, the function assumes this is the first
 *    time it has been called and it will complete the initialization
 *    of the counters that was started by fglobInit(); otherwise it
 *    will increment the counters.
 */
static int adjustCountersDirectionAction(void)
{
    int rv = 0;

    /* Since we have no way of knowing whether we are in the same
     * directory or not, always reset the dir_path to force it to be
     * recomputed. */
    *(fList->fg_dir_path) = '\0';

    if (fList->fg_initialized < 2) {
        /* was init called? */
        assert(fList->fg_initialized);

        /* This is the first call into this function.  All of the
         * indexes are at zero; set the "we've been here" flag, then
         * continue into the nested while() loops below to find the
         * first flow-path. */
        fList->fg_initialized = 2;
    }

    /*
     * continue to get data until we've seen all actions for all
     * directions for all hours.
     */
    while (fList->fg_hour_count > 0) {
        while (fList->fg_direction_idx < fList->fg_direction_count) {
            while (fList->fg_action_idx < fList->fg_action_count) {
                /* get next path-group, looking it up if needed */
                rv = getPathGroup();
                if (rv > 0) {
                    return rv;
                }

                /* try next action */
                fList->fg_action_idx++;
            }
            /* Reset action and try next direction */
            fList->fg_action_idx = 0;
            fList->fg_direction_idx++;
        }
        /* Reset direction and try next hour; adjust date if needed */
        fList->fg_direction_idx = 0;

        fList->fg_hour_count--;
        fList->fg_hour_idx++;
        if (fList->fg_hour_idx > 23) {
            /* On last hour of the day; reset hour and move to next day */
            fList->fg_hour_idx = 0;
            fList->fg_julian_idx++;

            /* set the yr, mo, day from the fg_julian_idx */
            gregorianDate(fList->fg_julian_idx, &fList->fg_year,
                          &fList->fg_month, &fList->fg_day);
        }
    }

    /* We're done. Make certain no more processing occurs */
    fList->fg_hour_count = -1;
    return 0;  /* done */
}
#endif /* SK_FGLOB_CONFIG_FILE */


/*
 *  is_ok = initTimes();
 *
 *    Verify the user's time inputs and initialize the julian dates
 *    and the hours based on the input.  Return 0 if all is well;
 *    otherwise print an error and return 1.
 */
static int initTimes(void)
{
    /* Parse the start and end dates from the user, if any */
    if (fList->fg_option[FGLOB_OPT_START_DATE]) {
        /* never returns on error */
        fList->fg_julian_start
            = getJDFromDate(fList->fg_option[FGLOB_OPT_START_DATE],
                            &fList->fg_hour_start);
        if (fList->fg_julian_start < 0) {
            return 1;  /* error */
        }
    }
    if (fList->fg_option[FGLOB_OPT_END_DATE]) {
        /* never returns on error */
        fList->fg_julian_end
            = getJDFromDate(fList->fg_option[FGLOB_OPT_END_DATE],
                            &fList->fg_hour_end);
        if (fList->fg_julian_end < 0) {
            return 1;  /* error */
        }
    }

    /* Use today when the user didn't enter a start date. */
    if (fList->fg_julian_start < 0) {
        time_t ts;
        struct tm *tms;
        if (time(&ts) == (time_t)-1) {
            skAppPrintErr("Unable to get current time of day: %s",
                          strerror(errno));
            return 1;
        }
        /* Should we use gmtime() here instead since that is how the
         * data is stored? */
        tms = localtime(&ts);
        fList->fg_julian_start =
            julianDate(tms->tm_year + 1900, tms->tm_mon + 1, tms->tm_mday);
    }

    if (fList->fg_julian_end < 0) {
        /* When the user does not give an explicit ending date, use
         * the start date */
        fList->fg_julian_end = fList->fg_julian_start;
    } else if (fList->fg_julian_end < fList->fg_julian_start) {
        skAppPrintErr("Ending date is earlier than start date");
        return 1;
    }

    if (fList->fg_hour_start < 0) {
        /* When no starting hour is set, use the entire day.  This
         * overrides any ending hour the user may have specified. */
        fList->fg_hour_start = 0;
        fList->fg_hour_end = 23;
    } else if (fList->fg_hour_end < 0) {
        /* When a starting hour is specified but no ending hour is
         * set, only look at one hour */
        fList->fg_hour_end = fList->fg_hour_start;
    } else if ((fList->fg_hour_end < fList->fg_hour_start)
               && (fList->fg_julian_end == fList->fg_julian_start)) {
        skAppPrintErr("Ending date is earlier than start date");
        return 1;
    }

    /* compute number of hours */
    fList->fg_hour_count = fList->fg_hour_end - fList->fg_hour_start + 1
        + (fList->fg_julian_end - fList->fg_julian_start) * 24;

    /* Set the current date and hour to the starting values. */
    fList->fg_julian_idx = fList->fg_julian_start;
    fList->fg_hour_idx = fList->fg_hour_start;

    /* set the yr, mo, day from the fg_julian_idx */
    gregorianDate(fList->fg_julian_idx,
                  &fList->fg_year, &fList->fg_month, &fList->fg_day);

    return 0;
}


/*
 *  is_ok = checkSensors()
 *
 *    Check that the sensor list is good given the class/type.
 *
 *    Parses the sensor option given by the user.  Supports sensor
 *    IDs or sensor Names.  The names are handled first; things that
 *    look like numbers are copied into another C-string and parsed
 *    later.  Returns 0 on success; 1 otherwise.
 */
static int checkSensors(void)
{
    BITMAP_DECLARE(sensor_bits, (SENSOR_INVALID_SID+1));
    unsigned int i;
    int j, k, m;
    int found_sensor;
    int c_number_from_classtype;
    int sensors_for_classtype[NUM_FLOW_TYPES];
    /* ptr to element of outFInfo from silk_site_*.h */
    fileTypeInfo *ft_info;
    /* ptr to element of sensorInfo from silk_site_*.h */
    sensorinfo_t *s_info;

    /* if user specified no sensors, use all sensors for the given
     * classtype */
    if (NULL == fList->fg_option[FGLOB_OPT_SENSORS]) {
        enableAllSensors();
        return 0;
    }

    /* We will count the number of sensors we get for each class/type
     * combination, so we can complain if the user failed to specify
     * sensors for a class/type. */
    memset(sensors_for_classtype, 0, sizeof(sensors_for_classtype));

    /* turn off all sensors */
    BITMAP_INIT(sensor_bits);

    /* convert the user's input to a bit string */
    if (parseSensorCmdLine(fList->fg_option[FGLOB_OPT_SENSORS], sensor_bits)) {
        /* error.  msg should already be printed */
        return 1;
    }

    /* For each sensor bit that is set, verify that the sensor is
     * valid for the specified class/type(s).
     */

    /* loop 'i' over all the known sensor ids from silk_site_*.h */
    for (i = 0; i < numSensors; ++i) {
        /* skip sensors the user didn't select */
        if ( !BITMAP_GETBIT(sensor_bits, i)) {
            continue;
        }
        s_info = &sensorInfo[i];
        found_sensor = 0;

        /* loop 'k' over all the fglob class/types */
        for (k = 0; k < fList->fg_classtype_count; ++k) {
            /* from the class/type get pointer to file type info. */
            ft_info = &outFInfo[fList->fg_classtype_list[k]];

            /* loop 'm' over all known classes to convert the class
             * name in the file type info struct to a class number. */
            c_number_from_classtype = -1;
            for (m = 0; m < NUM_CLASSES; ++m) {
                if (0 == strcmp(ft_info->className, classInfo[m].className)) {
                    c_number_from_classtype = m;
                    break;
                }
            }

            if (c_number_from_classtype == -1) {
                /* silk_site_*.h is internally inconsistent. abort()! */
                skAppPrintErr(("silk_site_*.h is inconsistent.\n"
                               "\t No class '%s' in classInfo[]"),
                              ft_info->className);
                abort();
            }

            /* loop 'j' over all classes that this sensor has data for */
            for (j = 0; j < s_info->numClasses; ++j) {

                /* compare the classNumber on the sensor with the
                 * class number from the class/type list */
                if (s_info->classNumbers[j] == c_number_from_classtype) {
                    /* add the sensor name to this classtype */
                    fList->fg_sensor_list[k][fList->fg_sensor_count[k]++] =
                        s_info->sensorName;
                    ++sensors_for_classtype[k];
                    ++found_sensor;
                }
            }
        }

        /* warn about unused sensor, but continue */
        if (0 == found_sensor) {
            skAppPrintErr(("ignoring sensor %u (%s)--"
                           "not used by specified class%s"),
                          i, s_info->sensorName,
                          ((fList->fg_classtype_count > 1) ? "es" : ""));
        }
    }

    /* For any class/type combinations for which we didn't find a
     * sensor, print and error and return. */
    for (j = 0; j < fList->fg_classtype_count; ++j) {
        if (sensors_for_classtype[j] == 0) {
            skAppPrintErr(("no corresponding sensors given for" /*CONCAT*/
                           " class/type pair '%s/%s'"),
                          outFInfo[fList->fg_classtype_list[j]].className,
                          outFInfo[fList->fg_classtype_list[j]].dirPrefix);
            return 1;
        }
    }

    return 0;
}


/*
 *  enableAllSensors();
 *
 *    For each of the configured classtypes, add every sensor that is
 *    valid for that classtype.
 */
static void enableAllSensors(void)
{
    unsigned int i;
    int j, k;
    const char *c_name_from_sensor;
    const char *c_name_from_classtype;
    /* ptr to element of sensorInfo from silk_site_*.h */
    sensorinfo_t *s_info;
    /* ptr to element of classInfo from silk_site_*.h */
    classInfoStruct *c_info;
    /* ptr to element of outFInfo from silk_site_*.h */
    fileTypeInfo *ft_info;

    /* loop 'i' over all the sensors defined in silk_site_*.h*/
    for (i = 0; i < numSensors; i++) {
        s_info = &sensorInfo[i];

        /* loop 'j' over all classes that this sensor has data for */
        for (j = 0; j < s_info->numClasses; ++j) {
            /* get the name of the sensor's class */
            c_info = &classInfo[s_info->classNumbers[j]];
            c_name_from_sensor = c_info->className;

            /* loop 'k' over all the fglob class/types we are using */
            for (k = 0; k < fList->fg_classtype_count; ++k) {
                /* from the class/type get the file type, from the
                 * file type get the class name */
                ft_info = &outFInfo[fList->fg_classtype_list[k]];
                c_name_from_classtype = ft_info->className;

                /* compare the two class names */
                if (0 == strcmp(c_name_from_sensor, c_name_from_classtype)) {
                    /* add this sensor to this classtype */
                    fList->fg_sensor_list[k][fList->fg_sensor_count[k]++] =
                        s_info->sensorName;
                }
            }
        }
    }
}


/*
 *  is_ok = parseSensorCmdLine(sensor_input, sensor_bitmap);
 *
 *    Parse the user's sensor input in sensor_input.  sensor_input can
 *    be a combination of named sensors and lone integers or integer
 *    ranges, separated by commas:  "S01,8,3-6,S02".
 *
 *    For each sensor, sets a bit in sensor_bitmap if that sensor is
 *    enabled.  The caller should bzero() sensor_bitmap before calling
 *    this function.
 *
 *    Function returns 0 on success or 1 on error: malloc error, bad
 *    sensor name, number out of range.
 */
static int parseSensorCmdLine(
    const char *sensor_input,
    BITMAP_DECLARE(sensor_bits, (1+SENSOR_INVALID_SID)))
{
    char *number_list = NULL;
    unsigned int i;
    int found_sensor;
    sensorinfo_t *s_info;
    uint32_t list_len;
    uint32_t *parsed_list = NULL;
    char *err_name = NULL;
    const char *cp;
    const char *ep;
    int len;
    int rv = 0; /* return value */
    int number_list_len = 0;

    /* Look at the user's sensor_input as comma-delimited tokens.  For
     * each token, if it starts with a digit, copy the token to
     * number_list, otherwise, treat it as a sensor name and try to
     * find it in the list of sensors.  number_list is parsed after
     * all the named sensors have been removed.
     */

    cp = sensor_input;
    while (*cp) {
        /* find next comma-delimited token; set len to its length */
        ep = strchr(cp, ',');
        if (ep == cp) {
            /* double comma */
            ++cp;
            continue;
        }
        if (ep) {
            len = (size_t)(ep - cp);
            ++ep;
        } else {
            len = strlen(cp);
            ep = cp+len;
        }

        if (isdigit((int)*cp)) {
            /* copy number/range to number_list for later processing and go
             * to the next token.  This assumes sensor names never begin
             * with a digit. */
            if (!number_list) {
                number_list = calloc(sizeof(char), 1+strlen(sensor_input));
                if (!number_list) {
                    skAppPrintErr("fglob.c checkSensors() out of memory");
                    rv = 1;
                    goto END;
                }
            }
            strncpy(&(number_list[number_list_len]), cp, len+1);
            number_list_len += len+1;
        } else {
            /* look for the sensor as a name among all the sensors.
             * found_sensor is 0 if we don't find it; 1 if we do */
            found_sensor = 0;
            for (i = 0; i < numSensors; ++i) {
                s_info = &sensorInfo[i];
                if ((0 == strncmp(cp, s_info->sensorName, len)) &&
                    ('\0' == s_info->sensorName[len]))
                {
                    BITMAP_SETBIT(sensor_bits, i);
                    ++found_sensor;
                    break;
                }
            }

            if (found_sensor == 0) {
                if ( (err_name = malloc(len+1)) ) {
                    /* only show user the current name that is giving
                     * problems */
                    strncpy(err_name, cp, len);
                    err_name[len] = '\0';
                    cp = err_name;
                }
                skAppPrintErr("unknown sensor name '%s'", cp);
                rv = 1;
                goto END;
            }
        } /* if (isdigit()) .. else */

        cp = ep;
    } /* while (*cp) */

    if (0 == number_list_len) {
        /* all of input given as named sensors */
        rv = 0;
        goto END;
    }

    /* handle the numbers/ranges in user's input */
    number_list[number_list_len] = '\0';

    /* convert string of numbers/ranges to array of numbers */
    if (skStringParseNumberList(&parsed_list, &list_len, number_list,
                                0, numSensors-1, 0))
    {
        skAppPrintErr("invalid sensor list '%s'",
                      number_list);
        rv = 1;
        goto END;
    }

    /* loop 'i' over all the sensors ids given on the command line */
    for (i = 0; i < list_len; ++i) {
        BITMAP_SETBIT(sensor_bits, parsed_list[i]);
    }

 END:
    if (parsed_list) {
        free(parsed_list);
    }
    if (number_list) {
        free(number_list);
    }
    if (err_name) {
        free(err_name);
    }
    return rv;
}


/*
 *  count = parseWordList(input_string, output_array, output_array_len);
 *
 *    Parses 'input_string' into a array of comma-separated tokens
 *    whose length is at least 1 (double comma is treated as a single
 *    comma).  Puts pointers to each word into 'output_array'.  The
 *    caller must have allocated 'output_array' and the caller should
 *    specify the length of 'output_array' in the 'output_array_len'
 *    parameter.  The function returns the number of words, or -1 if
 *    too more than 'output_array_len' words were found.  This
 *    function will destructively modify input_string; the caller
 *    should pass in a mutable copy.  The values in output_array will
 *    point into input_string; this function allocates no memory.
 */
static int parseWordList(char *string, char **word_list, int max_word_list)
{
    int word_count = 0;
    char *cp;
    char *ep;

    cp = string;
    while (*cp) {
        ep = strchr(cp, ',');
        if (ep == cp) {
            /* double comma, ignore */
            ++cp;
            continue;
        }
        if (ep != NULL) {
            *ep = '\0';
            ++ep;
        } else {
            /* force ep onto final '\0' */
            ep = cp + strlen(cp);
        }

        if (word_count >= max_word_list) {
            return -1;
        }

        word_list[word_count++] = cp;
        cp = ep;
    }

    return word_count;
}


/*
 *  is_ok = checkTypesForClass();
 *
 *    Will parse the user's "legacy" --class and --type input, or use
 *    the defaults if no class and/or type was given.  Return 0 on
 *    success, 1 on failure.
 *
 *    Will also verify that each listed type belongs to a class, and
 *    each class has at least one type.
 */
static int checkTypesForClass(void)
{
    char *type_str = NULL;
    char *type_names[NUM_FLOW_TYPES];
    int type_name_count;
    int num_classes;
    int class_list[FGLOB_NUM_CLASSES];
    int types_for_class_count[FGLOB_NUM_CLASSES];
    int i, j, k;
    int found_type;
    classInfoStruct *c_info;
    int rv = 0; /* return value */

    if (checkClass(&num_classes, class_list)) {
        return 1;
    }

    /* if user didn't give --type, use default types for each class */
    if (fList->fg_option[FGLOB_OPT_TYPE] == NULL) {
        for (i = 0; i < num_classes; ++i) {
            c_info = &(classInfo[class_list[i]]);
            for (j = 0; j < c_info->numDefaults; ++j) {
                fList->fg_classtype_list[fList->fg_classtype_count++] =
                    c_info->defaultList[j];
            }
        }
        return 0;
    }

    /* if user said --type=all, use all types for each class */
    if (strcmp(fList->fg_option[FGLOB_OPT_TYPE], "all") == 0) {
        for (i = 0; i < num_classes; ++i) {
            c_info = &(classInfo[class_list[i]]);
            for (j = 0; j < c_info->numTypes; ++j) {
                fList->fg_classtype_list[fList->fg_classtype_count++] =
                    c_info->typeList[j];
            }
        }
        return 0;
    }

    /* We will count the number of types we get for each class, so we
     * can complain if the user failed to specify types for a
     * class. */
    memset(types_for_class_count, 0, sizeof(types_for_class_count));

    /*
     * Match up the various types with the various classes.  Make a
     * modifiable copy of the types to parse into words.
     *
     * This probably needs to be thought through a bit better to
     * figure out what happens when types are given that don't
     * correspond to one class but not another.  warnings?
     */
    type_str = strdup(fList->fg_option[FGLOB_OPT_TYPE]);
    if (!type_str) {
        skAppPrintErr("Out of memory.  strdup() failed at %s:%d",
                      __FILE__, __LINE__);
        rv = 1;
        goto END;
    }

    type_name_count = parseWordList(type_str, type_names, NUM_FLOW_TYPES);

    if (type_name_count < 0) {
        skAppPrintErr("Too many types specified.  Only %d allowed",
                      NUM_FLOW_TYPES);
        rv = 1;
        goto END;
    }
    if (type_name_count == 0) {
        skAppPrintErr("No types were specified.");
        rv = 1;
        goto END;
    }

    for (k = 0; k < type_name_count; ++k) {
        /* look for type among all the classes.  FIXME: probably should do
         * something here to check for duplicates */
        found_type = 0;
        for (i = 0; i < num_classes; ++i) {
            c_info = &(classInfo[class_list[i]]);

            for (j = 0; j < c_info->numTypes; ++j) {
                if (0 == strcmp(type_names[k],
                                outFInfo[c_info->typeList[j]].dirPrefix))
                {
                    fList->fg_classtype_list[fList->fg_classtype_count++] =
                        c_info->typeList[j];
                    ++types_for_class_count[i];
                    ++found_type;
                }
            }
        }

        if (0 == found_type) {
            skAppPrintErr(("type '%s' not valid in specified class%s.\n"
                           "\tUse the --help option to see valid types "
                           "for each class"),
                          type_names[k], ((num_classes > 1) ? "es" : ""));
            rv = 1;
            goto END;
        }
    }

    /* For any classes specified for which we didn't find a type,
     * print and error and return. */
    for (i = 0; i < num_classes; ++i) {
        if (types_for_class_count[i] == 0) {
            skAppPrintErr("no corresponding types given for class '%s'",
                          classInfo[class_list[i]].className);
            rv = 1;
            goto END;
        }
    }

  END:
    if (type_str) {
        free(type_str);
    }
    return rv;
}


/*
 *  ok = checkClass(int *num_classes, int class_list[FGLOB_NUM_CLASSES]);
 *
 *    Check that the list of classes given by the user is valid.
 *    Returns 0 if it is value or 1 otherwise.
 *
 *    If no classes were specified, use the default class(es).
 *
 *    The value pointed to by 'num_classes' is set to the number of
 *    classes; the class_list[] is set to the class ids.
 */
static int checkClass(int *num_classes, int class_list[FGLOB_NUM_CLASSES])
{
    char *class_str = NULL;
    char *class_names[NUM_CLASSES];
    int class_name_count;
    int i, k;
    int found_class;
    int rv = 0; /* return value */

    /* check parameters */
    assert(num_classes);
    assert(class_list);

    /* put in known state */
    *num_classes = 0;

    /* if no class was specified by user, use the default class */
    if (NULL == fList->fg_option[FGLOB_OPT_CLASS]) {
        *num_classes = 1;
        class_list[0] = DEFAULT_CLASS;
        return 0;
    }

#if  FGLOB_NUM_CLASSES > 1
    /* Enable all classes if user's input is "all" */
    if (0 == strcmp(fList->fg_option[FGLOB_OPT_CLASS], "all")) {
        for (i = 0; i < NUM_CLASSES; ++i) {
            class_list[i] = i;
        }
        *num_classes = NUM_CLASSES;
        return 0;
    }
#endif /* FGLOB_NUM_CLASSES > 1 */

    /* make a modifiable copy of the class to parse into words */
    class_str = strdup(fList->fg_option[FGLOB_OPT_CLASS]);
    if (!class_str) {
        skAppPrintErr("Out of memory; strdup() failed at %s:%d",
                      __FILE__, __LINE__);
        rv = 1;
        goto END;
    }

    class_name_count = parseWordList(class_str, class_names, NUM_CLASSES);

#if  FGLOB_NUM_CLASSES <= 1
    if ((class_name_count < 0) || (class_name_count > 1)) {
        skAppPrintErr("Too many classes specified.  Only 1 allowed");
        rv = 1;
        goto END;
    }
#else
    if (class_name_count < 0) {
        skAppPrintErr("Too many classes specified.  Only %d allowed",
                      NUM_CLASSES);
        rv = 1;
        goto END;
    }
#endif /* FGLOB_NUM_CLASSES <= 1 */

    if (class_name_count == 0) {
        skAppPrintErr("No classes were specified");
        rv = 1;
        goto END;
    }

    for (k = 0; k < class_name_count; ++k) {
        /* look for the class */
        found_class = 0;
        for (i = 0; i < NUM_CLASSES; i++) {
            if (0 == strcmp(class_names[k], classInfo[i].className)) {
                class_list[(*num_classes)++] = i;
                ++found_class;
                break;
            }
        }

        /* print warning when class was not found and return */
        if (0 == found_class) {
            skAppPrintErr(("invalid class '%s'\n"
                           "\tUse the --help option to see valid classes\n"),
                          class_names[k]);
            rv = 1;
            goto END;
        }
    }

  END:
    if (class_str) {
        free(class_str);
    }
    return rv;
}


#if SK_FGLOB_CONFIG_FILE
/*
 *  is_ok = parseConfigurationFile();
 *
 *    Find, open, and parse the SiLK configuration file.  Return 0 if
 *    successful, or non-zero on these errors: file cannot be located,
 *    file cannot be opened, cannot create parser, file contains syntax
 *    errors, file does not contain a default domain.
 *
 *    Uses our standard findFile() routine to look for the configution
 *    file is SILK_PATH or in "../share" relative to the binary.
 */
static int parseConfigurationFile(void)
{
    char           *config_file = NULL;
    FILE           *config_fp = NULL;
    int             config_is_pipe;
    cfg_config_t   *config = NULL;
    int             rv = 0;

    if (fList->fg_config_file) {
        return 0;
    }

    /* find and open the configuration file */
    config_file = findFile(CFG_CONFIGURATION_FILE);
    if (!config_file) {
        /* library dumped error msg */
        rv = 1;
        goto END;
    }
    if (openFile(config_file, 0/*read*/, &config_fp, &config_is_pipe)) {
        skAppPrintErr("Cannot open configuration file '%s'",
                      config_file);
        rv = 1;
        goto END;
    }

    /* create parser and parse the configuration file */
    if (CFG_OK != cfg_config_create(&config)) {
        skAppPrintErr("Cannot create configuration parser");
        rv = 1;
        goto END;
    }
    if (CFG_OK != cfg_config_read( config, config_fp, skAppPrintErr)) {
        skAppPrintErr("Cannot parse configuration file '%s'",
                      config_file);
        rv = 1;
        goto END;
    }

    fList->fg_config_file = config;

  END:
    if (config_fp) {
        if (config_is_pipe) {
            pclose(config_fp);
        } else {
            fclose(config_fp);
        }
    }
    if (config_file) {
        free(config_file);
    }

    return rv;
}


/*
 *  is_ok = checkDomainDirectionAction();
 *
 *    Parse the values given by the user in the --domain, --direction,
 *    and --action switches.
 *
 *    For --domain: Make certain user's selection exists.  If not
 *    specified, use the default.
 *
 *    For --direction and --action: Convert the comma-separated list
 *    of terms into an array of terms.  If values are not given, use
 *    the defaults given in the configuration file.
 *
 *    Make certain data exists for the direction:action pairs that
 *    were given.
 *
 *    Return 0 on success, or non-zero on memory-allocation errors, if
 *    too many directions or actions are specified, or if the user
 *    does not give a value and no defaults were provided by the
 *    config file.
 */
static int checkDomainDirectionAction(void)
{
    cfg_map_t          *domain_list;
    cfg_map_t          *props;
    const cfg_array_t  *paths;
    char                flowpath[CFG_FLOWPATH_MAX_STRLEN];
    int                 i, j;
    int                 s;

    /* Get list of domains and properties */
    props = cfg_get_config_properties(fList->fg_config_file);
    domain_list = cfg_get_config_domains(fList->fg_config_file);

    /* Make certain domains exist */
    if ((domain_list == NULL) || (cfg_get_map_length(domain_list) < 1)) {
        skAppPrintErr("No domains found in configuration file");
        return 1;
    }

    /* For each of domain, direction, action, set the field on fList
     * to the user's value if provided.  If the user didn't provide
     * one, use the default value from the config file.  If one still
     * doesn't exist, whine and return an error. */

    /* *** DOMAIN *** */

    /* Check that we have a domain name */
    if (NULL != fList->fg_option[FGLOB_OPT_DOMAIN]) {
        /* value from user */
        fList->fg_domain = fList->fg_option[FGLOB_OPT_DOMAIN];
    } else if (props) {
        fList->fg_domain = (char*)cfg_map_assoc(props, CFG_DEFAULT_DOMAIN);
    }

    if (NULL == fList->fg_domain) {
        skAppPrintErr("No domain given and no default in config file.");
        return 1;
    }

    /* Check that we have data for that domain */
    if (NULL == cfg_map_assoc(domain_list, fList->fg_domain)) {
        skAppPrintErr("No data exists for the domain '%s'.",
                      fList->fg_domain);
        return 1;
    }

    /* *** DIRECTION *** */

    /* Check that we have a direction, and make a copy of it. */
    if (NULL != fList->fg_option[FGLOB_OPT_DIRECTION]) {
        /* value from user */
        fList->fg_direction = strdup(fList->fg_option[FGLOB_OPT_DIRECTION]);
    } else if (props) {
        char *temp = (char*)cfg_map_assoc(props, CFG_DEFAULT_DIRECTION);
        if (NULL != temp) {
            fList->fg_direction = strdup(temp);
        }
    }

    if (fList->fg_direction == NULL) {
        skAppPrintErr("No direction given and no default in config file.");
        return 1;
    }

    /* parse it into words */
    fList->fg_direction_count = parseWordList(fList->fg_direction,
                                              fList->fg_direction_list,
                                              FGLOB_NUM_DIRECTIONS);
    if (fList->fg_direction_count < 0) {
        skAppPrintErr("Too many directions specified.  Only %d allowed",
                      FGLOB_NUM_DIRECTIONS);
        return 1;
    }
    if (fList->fg_direction_count == 0) {
        skAppPrintErr("No directions were specified.");
        return 1;
    }

    /* *** ACTION *** */

    /* Check that we have a action, and make a copy of it. */
    if (NULL != fList->fg_option[FGLOB_OPT_ACTION]) {
        /* value from user */
        fList->fg_action = strdup(fList->fg_option[FGLOB_OPT_ACTION]);
    } else if (props) {
        char *temp = (char*)cfg_map_assoc(props, CFG_DEFAULT_ACTION);
        if (NULL != temp) {
            fList->fg_action = strdup(temp);
        }
    }

    if (fList->fg_action == NULL) {
        skAppPrintErr("No action given and no default in config file.");
        return 1;
    }

    /* parse it into words */
    fList->fg_action_count = parseWordList(fList->fg_action,
                                           fList->fg_action_list,
                                           FGLOB_NUM_ACTIONS);
    if (fList->fg_action_count < 0) {
        skAppPrintErr("Too many actions specified.  Only %d allowed",
                      FGLOB_NUM_ACTIONS);
        return 1;
    }
    if (fList->fg_action_count == 0) {
        skAppPrintErr("No actions were specified.");
        return 1;
    }

    /* *** CHECK *** */

    /* Make certain there is an entry for every direction:action pair
     * in the configuration file for this domain; warn if not.  This
     * involves a simple lookup, and does not require us to free()
     * anything when we are done. */
    for (i = 0; i < fList->fg_direction_count; ++i) {
        for (j = 0; j < fList->fg_action_count; ++j) {
            /* Create a flowpath and check its length */
            s = snprintf(flowpath, sizeof(flowpath), "%s:%s",
                         fList->fg_direction_list[i],
                         fList->fg_action_list[j]);
            if (s < 0 || (size_t)s >= sizeof(flowpath)) {
                skAppPrintErr(("direction/action combination %s:%s too long;\n"
                               "\t length of %d is longer than max (%lu)"),
                              fList->fg_direction_list[i],
                              fList->fg_action_list[j],
                              s, (unsigned long)sizeof(flowpath));
                return 1;
            }
            /* look it up */
            paths = cfg_config_lookup(fList->fg_config_file,
                                      fList->fg_domain, flowpath);
            if (paths == NULL) {
                skAppPrintErr("WARNING: invalid direction:action pair '%s'",
                              flowpath);
            }
        }
    }

    return 0;
}
#endif /* SK_FGLOB_CONFIG_FILE */


/*
 *  is_used = fglobValid();
 *
 *    Return 1 if fglob is to be used; 0 otherwise.
 */
int fglobValid(void)
{
    if ((fList != NULL) && (fList->fg_user_option_count > 0)) {
        return 1;
    }
    return 0;
}


/*
 *  option_count = fglobSetFilters(sensor_bitmap, file_info_bitmap);
 *
 *    This function is used when filtering a previous data pull.  It
 *    allows the --class, --type, and --sensor switches to work over
 *    this data.
 *
 *    Fill in the 'sensor_bitmap' with the values the user provided to
 *    fglob's --sensor switch, and indexed into the sensorInfo[] array
 *    from silk_site_*.h.
 *
 *    Fill in the 'file_info_bitmap' with the values the user provided
 *    to fglob's --class and --type switches and indexed into the
 *    outFInfo[] array from silk_site_*.h.
 */
int fglobSetFilters(
    uint32_t    sensor_bitmap[],
    uint32_t    finfo_bitmap[])
{
    int i;
    int rv = 0; /* no filters, no errors */

    /* Do we have sensor info? */
    if (fList->fg_option[FGLOB_OPT_SENSORS]) {
        if (parseSensorCmdLine(fList->fg_option[FGLOB_OPT_SENSORS],
                               sensor_bitmap))
        {
            return -1;
        }

        rv += 1;
    }


    if (fList->fg_option[FGLOB_OPT_CLASS] || fList->fg_option[FGLOB_OPT_TYPE])
    {
        /* User gave us at least class/type information */
        if (checkTypesForClass()) {
            /* error */
            return -2;
        }

        for (i = 0; i < fList->fg_classtype_count; ++i) {
            BITMAP_SETBIT(finfo_bitmap, fList->fg_classtype_list[i]);
        }

        /* if sensor information was given, compare it against the
         * class/type list that was given, but we treat any problems
         * as warnings only. */
        if (fList->fg_option[FGLOB_OPT_SENSORS]) {
            (void)checkSensors();
        }

        rv += 2;
    }

    /* Adjust the value of the 'fg_user_option_count' member */
    if (fList->fg_option[FGLOB_OPT_CLASS]) {
        --fList->fg_user_option_count;
    }
    if (fList->fg_option[FGLOB_OPT_TYPE]) {
        --fList->fg_user_option_count;
    }
    if (fList->fg_option[FGLOB_OPT_SENSORS]) {
        --fList->fg_user_option_count;
    }

    return rv;
}


#ifdef  TEST_FGLOB

int main(int argc, char **argv)
{
    char *fname;
    int numFiles;
    char arg[128];

    skAppRegister(argv[0]);
#if     0
    if (argc < 2) {
        fglobUsage();
        exit(1);
    }
#endif
    if (optionsSetup(fglobUsage)) {
        skAppPrintErr("unable to setup options module");
        exit(1);/* never returns */
    }

    if ( fglobSetup()){
        skAppPrintErr("fglob: error");
        exit(1);
    }
    numFiles = 0;
#if     1
    /* for testing command line options parsing */
    if (optionsParse(argc, argv) < 0) {
        skAppPrintErr("optionsParse error");
        fglobUsage();
        exit(1);
    }
#else
    /* for testing application passed options parsing */
    fprintf(stdout, "StartDate? ");fgets(arg, sizeof(arg), stdin);
    if (strlen(arg) > 1) {
        arg[strlen(arg) - 1] = '\0';
        fglobSetStartDate(arg);
    }
    fprintf(stdout, "EndDate? ");fgets(arg, sizeof(arg), stdin);
    if (strlen(arg) > 1) {
        arg[strlen(arg) - 1] = '\0';
        fglobSetEndDate(arg);
    }
    fprintf(stdout, "Class? ");fgets(arg, sizeof(arg), stdin);
    if (strlen(arg) > 1) {
        arg[strlen(arg) - 1] = '\0';
        fglobSetClass(arg);
    }
    fprintf(stdout, "Type? ");fgets(arg, sizeof(arg), stdin);
    if (strlen(arg) > 1) {
        arg[strlen(arg) - 1] = '\0';
        fglobSetType(arg);
    }
    fprintf(stdout, "SensorList? ");fgets(arg, sizeof(arg), stdin);
    if (strlen(arg) > 1) {
        arg[strlen(arg) - 1] = '\0';
        fglobSetSensors(arg);
    }
#endif

    while( ( fname = fglobNext()) != (char *)NULL) {
        fprintf(stdout, "%s\n", fname);
        numFiles++;
    }
    fprintf(stdout, "globbed %d files\n", numFiles);
    fglobTeardown();
    exit(0);
}
#endif
