/*
** Copyright (C) 2003-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwfglob
**
**  A utility to simply print to stdout the list of files that fglob
**  would normally return.  If CHECK_BLOCK_COUNT is defined, this
**  application code will stat() each file and print the string "ON
**  TAPE" if its block size is 0.
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwfilter/rwfglobapp.c,v 1.2 2005/06/10 15:18:48 thomasm Exp $");

#include "rwfilter.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* What to print when file has 0 blocksize */
#define ZERO_BLOCKSIZE_MSG "  \t*** ON_TAPE ***"

/* where to write --help output */
#define USAGE_FH stdout

/* where to write the output */
#define OUTPUT_FH stdout

/* LOCAL VARIABLES */

/* when this variable is nonzero, stat() to get the file's block size
 * and print ZERO_BLOCKSIZE_MSG if size==0 */
const int check_block_count = 1;


/*
 * appUsage:
 * 	print short usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 * 	print complete usage information to USAGE_FH and exit with
 * 	code.  passed to optionsSetup() to print usage when --help
 * 	option given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
    fprintf(USAGE_FH, "Usage: %s [options]\n", skAppName());
    fglobUsage(USAGE_FH);
    exit(EXIT_FAILURE);
}



/*
 *  printFileNames()
 *
 *    Output function: Simply prints the name of each globbed file.
 */
static void printFileNames(void)
{
    char *fname;
    int numFiles = 0;

    while( (fname = fglobNext()) != (char *)NULL ) {
        fprintf(OUTPUT_FH, "%s\n", fname);
        ++numFiles;
    }

    fprintf(OUTPUT_FH, "globbed %d files\n", numFiles);
}


/*
 *  printFileNamesCheckBlocks()
 *
 *    Output function: Prints the name of each globbed file.  In
 *    addition the function determines the file's blocksize.  If the
 *    blocksize is zero, it prints the ZERO_BLOCKSIZE_MSG after the
 *    file's name.
 */
static void printFileNamesCheckBlocks(void)
{
    struct stat statbuf;
    char *fname;
    int numFiles = 0;
    int numOnTape = 0;

    while( (fname = fglobNext()) != (char *)NULL ) {
        if (-1 == stat(fname, &statbuf)) {
            /* should never happen; fglob wouldn't have returned it */
            fprintf(stderr, "Cannot stat '%s'\n", fname);
            exit(1);
        }
        if (0 == statbuf.st_blocks) {
            fprintf(OUTPUT_FH, "%s%s\n", ZERO_BLOCKSIZE_MSG, fname);
            ++numOnTape;
        } else {
            fprintf(OUTPUT_FH, "%s\n", fname);
        }
        ++numFiles;
    }

    fprintf(OUTPUT_FH, "globbed %d files; %d on tape\n", numFiles, numOnTape);
}


int main(int argc, char **argv)
{
    int first_unhandled_option;

    skAppRegister(argv[0]);

    /* Set up the options module */
    if (optionsSetup(&appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);
    }

    /* Set up the fglob module */
    if (fglobSetup()){
        skAppPrintErr("unable to setup fglob module");
        exit(EXIT_FAILURE);
    }

    /* Parse the options */
    first_unhandled_option = optionsParse(argc, argv);
    assert(first_unhandled_option <= argc);
    if (first_unhandled_option < 0) {
        /* error in user's input.  option handler printed error */
        appUsage();
    }

    /* glob the files */
    if (check_block_count) {
        printFileNamesCheckBlocks();
    } else {
        printFileNames();
    }

    /* cleanup */
    fglobTeardown();
    optionsTeardown();

    return 0;
}
