/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwfilter.c
**
**  6/3/2001
**
**  Suresh L. Konda
**
**  Allows for selective extraction of records and fields from a rw
**  packed file.  This version, unlike rwcut, creates a binary file with
**  the filtered records.  A new file type is used.  The header does not
**  contain valid recCount and rejectCount values.  The other fields are
**  taken from the original input file.
**
**  A second header is also created which records the filter rules used
**  for each pass.  Thus this is a variable length header
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwfilter/rwfilter.c,v 1.37 2005/11/17 22:51:24 thomasm Exp $");

#include "rwfilter.h"


/* TYPEDEFS AND DEFINES */

/*
 *  INCR_REC_COUNT(count, rec)
 *
 *    Increment the values in the rec_count_t 'count' by the values in
 *    the rwRec 'rec'.
 *
 */
#define INCR_REC_COUNT(count, rec) \
    { (count).flows++; (count).pkts+=(rec).pkts; (count).bytes+=(rec).bytes; }


/* IMPORTED VARIABLES */

extern struct profile prof1;


/* IMPORTED FUNCTIONS */


/* EXPORTED VARIABLES */

struct profile prof2;

/* struct to hold all the io streams */
filter_io_t filter_io;

/* index into argv of first option that does not start with a '--'.
 * This assumes getopt rearranges the options, which gnu getopt will
 * do. */
int first_non_switch;

/* maximum number of records the user wants to see.  zero for ALL. */
uint64_t max_records[DESTINATION_TYPES];

/* counts needed for --print-stat option */
filterStatisticsV2_t stats;

/* number of checks to preform */
int checker_count = 0;

/* function pointers to handle checking and or processing */
int (*checker[MAX_CHECKERS])(rwRecPtr);

/* user shared object plug-in */
dynlibInfoStructPtr dlISP;

/* number of shared object plug-ins loaded */
int filterDynlibCount;

/* pointer to each plug-in */
dynlibInfoStructPtr filterDynlibList[FILTER_MAX_DYNLIBS];

/* names of plug-ins to attempt to load at startup */
const char *filterDynlibNames[] = {
    "libccfilter_new.so",
    "libipfilter.so",
    NULL /* sentinel */
};


/* LOCAL FUNCTIONS */

static int  filterFile(const char *inFName);
static int  readFileNamesFromFile(FILE *handle);


/* LOCAL VARIABLES */

/* cache of argc and argv used for dumping headers */
static int pargc;
static char **pargv;


/*
 *  int = readFileNamesFromFile(file);
 *
 *    Read the input from 'file' one line at a time.  Treat the line
 *    as the name of a file to be opened and filtered.  Return 0 on
 *    success and non-zero on error.
 *
 *    Spaces on the input lines are significant: leading, trailing,
 *    and embedded spaces are treated as part of the file name.
 *
 *    Lines that are too long are ignored, and a warning is printed to
 *    the stderr.
 */
static int readFileNamesFromFile(FILE *handle)
{
    char line[PATH_MAX+1];
    char *eol;
    unsigned long line_count = 0;
    int rv = 0; /* return value */

    /* read from handle until EOF */
    while (!feof(handle)) {
        if (!fgets(line, sizeof(line), handle)) {
            continue;
        }
        ++line_count;
        eol = strchr(line, '\n');
        if (eol == line) {
            /* empty line; ignore */
            continue;
        }

        if (NULL != eol) {
            /* expected behavior: read an entire line */
            *eol = '\0';
        } else if (feof(handle)) {
            /* okay: last line did not end in newline */
        } else {
            /* bad: line was longer than sizeof(line).  read until newline
             * or eof, then throw away the line */
            skAppPrintErr("xargs stream input line %lu too long. ignored",
                          line_count);
            while (fgets(line, sizeof(line), handle) && !strchr(line,'\n'))
                ; /* empty */
            continue;
        }
        if (0 != (rv = filterFile(line))) {
            goto END;
        }
    }

  END:
    return rv;
}


/*
 *  num_output_streams = closeOutputDests(dest_type, idx);
 *
 *    When idx is non-negative, close the destination stream at
 *    position 'idx' for the specified destination type 'dest_type'
 *    (e.g., DEST_ALL).
 *
 *    When idx is negative, close all the streams for the specified
 *    dest_type.
 *
 *    Adjust the count of output streams for 'dest_type'
 *    appropriately, and return the number of output streams across
 *    all destination types.
 */
static int closeOutputDests(
    const int   dest_type,
    int         idx)
{
    int i;
    int rv;
    int count;
    int num_outs;

    if (idx < 0) {
        /* close ALL streams of 'dest_type' */
        for (i = 0; i < filter_io.dest_count[dest_type]; ++i) {
            rv = rwCloseFile(filter_io.dest_stream[dest_type][i].ios);
            if (rv) {
                rwioPrintLastErr(filter_io.dest_stream[dest_type][i].ios,
                                 rv, &skAppPrintErr);
            }
        }

        filter_io.dest_count[dest_type] = 0;
    } else {
        /* close stream 'idx' of 'dest_type' */
        rv = rwCloseFile(filter_io.dest_stream[dest_type][idx].ios);
        if (rv) {
            rwioPrintLastErr(filter_io.dest_stream[dest_type][idx].ios,
                             rv, &skAppPrintErr);
        }

        /* decrement dest_type's stream count and move the last stream to
         * stream idx's position---unless 'idx' is the last stream. */
        --filter_io.dest_count[dest_type];
        count = filter_io.dest_count[dest_type];
        if (idx != count) {
            filter_io.dest_stream[dest_type][idx].ios = 
                filter_io.dest_stream[dest_type][count].ios;
            filter_io.dest_stream[dest_type][idx].fname = 
                filter_io.dest_stream[dest_type][count].fname;
        }
    }

    /* compute and return the number of output streams that we now
     * have. */
    num_outs = 0;
    for (i = 0; i < DESTINATION_TYPES; ++i) {
        num_outs += filter_io.dest_count[i];
    }

    return num_outs;
}


/*
 *  filterFile:
 *	This is the actual filtering of a file.  Also handles
 *	dumping out the header if required.
 */
static int filterFile(const char *fpath)
{
    static int first_file_type = -1;
    static int reading_records = 1;
    static rwRec rwrec;
    int i, j;
    int fail;
    int rv = LIBRW_OK;

    ++stats.fileCount;

    /* nothing to do in dry-run mode but print the file name*/
    if (filter_io.dryrun_fp) {
        fprintf(filter_io.dryrun_fp, "%s\n", fpath);
        return 0;
    }

    /* print filenames if requested */
    if (filter_io.filenames_fp) {
        fprintf(filter_io.filenames_fp, "%s\n", fpath);
    }

    if (!reading_records) {
        return 0;
    }

    /* open the input file */
    filter_io.input.ios = rwOpenFile(fpath, NULL);
    if (NULL == filter_io.input.ios) {
        /* some error. the library would have dumped a msg */
        return 1;
    }

    /* check that the files are all of the same type and version  */
    if (first_file_type < 0) {

        /* Processing first input file. Record the file type */
        first_file_type = rwGetFileType(filter_io.input.ios);

        /*
         * Add command line history to each binary output file and
         * print the header: If we are processing the result of
         * previous rwfilter command, add the history from that file
         * to each output.
         *
         * In addition, regardless of input, add this invocation's
         * command line to each output file's history.
         */
        for (i = 0; i < DESTINATION_TYPES; ++i) {
            for (j = 0; j < filter_io.dest_count[i]; ++j) {
                rwIOStruct *rwIOS = filter_io.dest_stream[i][j].ios;

                if (FT_RWFILTER == first_file_type) {
                    rv =rwioAppendToHistoryFromFile(rwIOS,filter_io.input.ios);
                }
                if (rv == LIBRW_OK) {
                    rv = rwioAppendToHistoryArgv(rwIOS, pargc, pargv);
                }
                if (rv == LIBRW_OK) {
                    rv = rwioWriteHeader(rwIOS);
                }

                if (rv) {
                    rwioPrintLastErr(rwIOS, rv, &skAppPrintErr);
                    rv = LIBRW_OK;
                }
            }
        }

    } else {
        /* if the file type is FT_RWFILTER, only 1 file is allowed */
        if (first_file_type == FT_RWFILTER) {
            skAppPrintErr("Only 1 input file of type filter allowed. Abort.");
            return 1;
        }

    }

    /* read and process each record */
    while (reading_records && rwRead(filter_io.input.ios, &rwrec)) {

        /* increment number of read records */
        INCR_REC_COUNT(stats.rCount, rwrec);

        /* the all-dest */
        if (filter_io.dest_count[DEST_ALL]) {
            i = 0;
            while (i < filter_io.dest_count[DEST_ALL]) {
                rv = rwWrite(filter_io.dest_stream[DEST_ALL][i].ios, &rwrec);
                if ( !LIBRW_ERROR_IS_FATAL(rv)) {
                    ++i;
                } else {
                    /* close the one stream and see if we need to
                     * continue reading.  DO NOT increment i here,
                     * since the stream at position i will be
                     * different after calling closeOutputDests. */
                    reading_records = closeOutputDests(DEST_ALL, i);
                }
            }

            /* close all streams for this destination type if we are
             * at user's requested max.  If max_records[] is 0, this
             * will never be true, and all records will be
             * processed. */
            if (stats.rCount.flows == max_records[DEST_ALL]) {
                reading_records = closeOutputDests(DEST_ALL, -1);
            }
        }
            

        /* run all checker()'s until end or one fails */
        for (i=0, fail=0; i < checker_count && !fail; ++i) {
            fail = (*(checker[i]))(&rwrec);
        }

        if (!fail) {

            /* increment number of record that pass */
            INCR_REC_COUNT(stats.pCount, rwrec);

            /* the pass-dest */
            if (filter_io.dest_stream[DEST_PASS]) {
                i = 0;
                while (i < filter_io.dest_count[DEST_PASS]) {
                    rv = rwWrite(filter_io.dest_stream[DEST_PASS][i].ios,
                                 &rwrec);
                    if ( !LIBRW_ERROR_IS_FATAL(rv)) {
                        ++i;
                    } else {
                        /* close the one stream and see if we need to
                         * continue reading.  DO NOT increment i here,
                         * since the stream at position i will be
                         * different after calling closeOutputDests. */
                        reading_records = closeOutputDests(DEST_PASS, i);
                    }
                }

                /* close all streams for this destination type if we are
                 * at user's requested max.  If max_records[] is 0, this
                 * will never be true, and all records will be
                 * processed. */
                if (stats.pCount.flows == max_records[DEST_PASS]) {
                    reading_records = closeOutputDests(DEST_PASS, -1);
                }
            }
        } else {

            /* the fail-dest */
            if (filter_io.dest_stream[DEST_FAIL]) {
                i = 0;
                while (i < filter_io.dest_count[DEST_FAIL]) {
                    rv = rwWrite(filter_io.dest_stream[DEST_FAIL][i].ios,
                                 &rwrec);
                    if ( !LIBRW_ERROR_IS_FATAL(rv)) {
                        ++i;
                    } else {
                        /* close the one stream and see if we need to
                         * continue reading.  DO NOT increment i here,
                         * since the stream at position i will be
                         * different after calling closeOutputDests. */
                        reading_records = closeOutputDests(DEST_FAIL, i);
                    }
                }

                /* close all streams for this destination type if we are
                 * at user's requested max.  If max_records[] is 0, this
                 * will never be true, and all records will be
                 * processed. */
                if ((stats.rCount.flows - stats.pCount.flows)
                    == max_records[DEST_FAIL])
                {
                    reading_records = closeOutputDests(DEST_FAIL, -1);
                }
            }
        }

    } /* while (reading_records && rwRead()) */

    /* close input */
    rwCloseFile(filter_io.input.ios);
    filter_io.input.ios = NULL;

    return 0;
}


int main(int argc, char ** argv)
{
    int rv = 0;
    int i, j;

    /*
      init_profile(&prof1, "OpenFile");
      init_profile(&prof2, "main");
      start_profile(&prof2);
    */

    appSetup(argc, argv);

    pargc = argc;
    pargv = argv;

    /* Get the files.  Only one of these should be active */
    if (fglobValid()) {
        const char * curFName;

        while ((rv == 0) && (curFName = fglobNext())) {
            rv = filterFile(curFName);
        }
    } else if (filter_io.input.fname) {
        rv = filterFile(filter_io.input.fname);
    } else if (filter_io.xargs.fp) {
        rv = readFileNamesFromFile(filter_io.xargs.fp);
    } else {
        for (i = first_non_switch; rv == 0 && i < argc; ++i) {
            rv = filterFile(argv[i]);
        }
    }

    /* We don't write the header to the destination file(s) until
     * we've read the first record.  However, if no files were read,
     * the destination files are empty, so dump the header to them
     * now. */
    if (stats.fileCount == 0) {
        for (i = 0; i < DESTINATION_TYPES; ++i) {
            for (j = 0; j < filter_io.dest_count[i]; ++j) {
                rwIOStruct *rwIOS = filter_io.dest_stream[i][j].ios;
                int iorv;

                if (rwIOS) {
                    if ((iorv = rwioAppendToHistoryArgv(rwIOS, pargc, pargv))
                        || (iorv = rwioWriteHeader(rwIOS)))
                    {
                        rwioPrintLastErr(rwIOS, iorv, &skAppPrintErr);
                    }
                }
            }
        }
    }

    if (filter_io.stats.fp) {
        filterStatisticsDump(filter_io.stats.fp, &stats);
    }

    appTeardown();

    /*
      stop_profile(&prof2);
      print_profile(stderr, &prof1);
      print_profile(stderr, &prof2);
    */

    return rv;
}
