/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
 *  filteropts.c
 *
 *  Suresh Konda
 *
 *  parsing and setting up filter options
 *
 *  10/21/01
 *
 */

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwfilter/rwfiltercheck.c,v 1.7 2005/11/16 14:57:30 thomasm Exp $");

#include "rwfilter.h"
#include "silk_site.h"


/* TYPEDEFS AND DEFINES */

/* number of ports and protocols */
#define MAX_PORTS      65536
#define MAX_PROTOCOLS    256

/* number of sensors and flow-types; these must agree with rwpack.h */
#define MAX_SENSORS    65536
#define MAX_FLOW_TYPES   256

/* number of TCP flag rules we support */
#define MAX_TCPFLAG_RULES  8

/* number of filter rules.  Approx equal to number of options */
#define NUM_RULE_CLASSES 64

/* multipler to convert a floating point number to a unit64_t */
#define PRECISION      10000

/* index into the IP arrays */
enum ip_array_index {
    SRC=0, DST, ANY,
    IP_INDEX_COUNT /* must be last */
};

/* holds a set of IP addresses ("10.11.12-14.x") for filtering */
typedef struct {
    uint32_t octets[4][256/32];
} ip_addr_range_t;

/* holds a range */
typedef struct {
    uint64_t max;
    uint64_t min;
} value_range_t;

/*
 *  CHECK_RANGE(v, t)
 *
 *    Returns 1 if value 'v' is withing the range 't'; 0 otherwise.
 */
#define CHECK_RANGE(v, t)   ((((v) >= (t).min) && ((v) <= (t).max)) ? 1 : 0)

/* holds TCP flags high/mask; e.g., S/SA */
typedef struct {
    uint8_t  high;
    uint8_t  mask;
} high_mask_t;

/* Set the 'high' and 'mask' members in 'high_mask' from values in 'arg' */
#define PARSE_TCP_HIGH_MASK(high_mask, arg) \
    skStringParseTCPFlagsHighMask(&(high_mask).high, &(high_mask).mask, (arg))

/* Test to see if TCP flags in 'var' meet the requirements of 'high_mask' */
#define CHECK_TCP_HIGH_MASK(var, high_mask) \
    TCP_FLAG_TEST_HIGH_MASK((var), (high_mask).high, (high_mask).mask)


/* The filters */
typedef struct {
    /* times */
    value_range_t sTime, eTime, activeTime, elapsed;

    /* flow volume */
    value_range_t bytes, pkts, bytesPerPacket;

    /*
     * IP values
     */
    ip_addr_range_t IP[IP_INDEX_COUNT];
    int negateIP[IP_INDEX_COUNT];

    /* Source and Dest ports */
    BITMAP_DECLARE(sPort, MAX_PORTS);
    BITMAP_DECLARE(dPort, MAX_PORTS);
    BITMAP_DECLARE(any_port, MAX_PORTS);

    /* IP Protocol */
    BITMAP_DECLARE(proto, MAX_PROTOCOLS);

    /* sensors and class/type */
    BITMAP_DECLARE(sID, MAX_SENSORS);
    BITMAP_DECLARE(flow_type, MAX_FLOW_TYPES);

    /* SNMP interfaces */
    BITMAP_DECLARE(input,  SK_SNMP_INDEX_LIMIT);
    BITMAP_DECLARE(output, SK_SNMP_INDEX_LIMIT);

    /* Next hop IP */
    ip_addr_range_t nhIP;

    /* tcp flags (old style tcp flags) */
    uint8_t flags;

    /* flags_all, flags_init, flags_session */
    high_mask_t flags_all[MAX_TCPFLAG_RULES];
    high_mask_t flags_init[MAX_TCPFLAG_RULES];
    high_mask_t flags_session[MAX_TCPFLAG_RULES];
    uint8_t count_flags_all;
    uint8_t count_flags_init;
    uint8_t count_flags_session;

    /* TCP state/attributes */
    uint8_t attributes;

    /* application */
    BITMAP_DECLARE(application, MAX_PORTS);

    /*
     *  entry in ruleSet[i] ==  ruleNumber if desired.  maxRuleseUsd is the
     *  number of rules actually used in this run
     */
    uint8_t ruleSet[NUM_RULE_CLASSES];
    uint8_t maxRulesUsed;

} filter_rules_t;



/* LOCAL FUNCTIONS */

static int parseTimeIntoValueRange(
    value_range_t  *p_vr,
    const char     *s_time);
static uint32_t parseIntegerRange(
    value_range_t  *tgtRange,
    const char     *rangeString);
static uint32_t parseDecimalRange(
    value_range_t  *tgtRange,
    const char     *rangeString);
static int parseFlags(
    filter_rules_t *rules,
    int             opt_index,
    const char     *opt_arg);
static int setFilterRuleBinaryFlag(
    filter_rules_t *rules,
    int             opt_index,
    const char     *opt_arg);
static int parseAttributes(
    filter_rules_t *rules,
    int             opt_index,
    const char     *opt_arg);
static int checkAddress(uint32_t tgtAddress, const ip_addr_range_t *maskRange);


/* LOCAL VARIABLES */

static filter_rules_t fltr_rules;
static filter_rules_t * const fltrRulesPtr = &fltr_rules;


/* OPTION SETUP */

/* struct implementer uses to define options */
typedef struct {
    struct option   option;
    const char     *help;
} filter_switch_t;

enum {
    OPT_STIME = 0, OPT_ETIME, OPT_ACTIVE_TIME, OPT_DURATION,
    OPT_SPORT, OPT_DPORT, OPT_APORT, OPT_PROTOCOL,
    OPT_BYTES, OPT_PACKETS, OPT_BYTES_PER_PACKET,
    OPT_SADDRESS, OPT_DADDRESS, OPT_ANY_ADDRESS,
    OPT_NOT_SADDRESS, OPT_NOT_DADDRESS, OPT_NOT_ANY_ADDRESS,
    OPT_NEXT_HOP_ID, OPT_INPUT_INDEX, OPT_OUTPUT_INDEX, OPT_TCP_FLAGS,
    OPT_FIN_FLAG, OPT_SYN_FLAG, OPT_RST_FLAG, OPT_PSH_FLAG,
    OPT_ACK_FLAG, OPT_URG_FLAG, OPT_ECE_FLAG, OPT_CWR_FLAG,
    OPT_FLAGS_ALL, OPT_FLAGS_INITIAL, OPT_FLAGS_SESSION, OPT_ATTRIBUTES,
    OPT_APPLICATION,
    OPT_SENSORS, OPT_FLOW_TYPE,
    _OPT_FINAL_OPTION_ /* must be last */
};

/* options as used by getopt_long() */
static struct option *filterOptions = NULL;

/* options and their help strings */
static filter_switch_t filterSwitch[] = {
    {{"stime",             REQUIRED_ARG, 0, OPT_STIME},
     ("start time is within this time window:\n"
      "\t YYYY/MM/DD[:HH[:MM[:SS]]][-YYYY/MM/DD[:HH[:MM[:SS]]]]\n"
      "\t If no window closing time, use start time. Window closing\n"
      "\t time is rounded to final second of specified precision.")},
    {{"etime",             REQUIRED_ARG, 0, OPT_ETIME},
     ("ending time is within this time window.")},
    {{"active-time",       REQUIRED_ARG, 0, OPT_ACTIVE_TIME},
     ("flow was active at any time during this time window.")},
    {{"duration",          REQUIRED_ARG, 0, OPT_DURATION},
     ("duration in seconds falls within integer range N-M.")},

    {{"sport",             REQUIRED_ARG, 0, OPT_SPORT},
     ("source port is contained in this list.\n"
      "\t A comma separated list of numbers and ranges between 0 and 65535.")},
    {{"dport",             REQUIRED_ARG, 0, OPT_DPORT},
     ("destination port is contained in this list.\n"
      "\t A comma separated list of numbers and ranges between 0 and 65535.")},
    {{"aport",             REQUIRED_ARG, 0, OPT_APORT},
     ("source or destination port is contained in this list.\n"
      "\t A comma separated list of numbers and ranges between 0 and 65535.")},
    {{"protocol",          REQUIRED_ARG, 0, OPT_PROTOCOL},
     ("protocol is contained in this list.\n"
      "\t A comma separated list of numbers and ranges between 0 and 256.")},

    {{"bytes",             REQUIRED_ARG, 0, OPT_BYTES},
     ("byte count is within the integer range N-M.")},
    {{"packets",           REQUIRED_ARG, 0, OPT_PACKETS},
     ("packet count is within the integer range N-M.")},
    {{"bytes-per-packet",  REQUIRED_ARG, 0, OPT_BYTES_PER_PACKET},
     ("byte-per-packet count is within decimal range X-Y.")},

    {{"saddress",          REQUIRED_ARG, 0, OPT_SADDRESS},
     ("source address matches this wildcard IP.\n"
      "\t Wildcard IP is a dotted quad where each octet is a number,\n"
      "\t a comma-separted list of numbers and/or ranges, or 'x' for 0-255")},
    {{"daddress",          REQUIRED_ARG, 0, OPT_DADDRESS},
     ("destination address matches this wildcard IP.")},
    {{"any-address",       REQUIRED_ARG, 0, OPT_ANY_ADDRESS},
     ("source or destination address matches this wildcard IP.")},
    {{"not-saddress",      REQUIRED_ARG, 0, OPT_NOT_SADDRESS},
     ("source address does not match this wildcard IP.")},
    {{"not-daddress",      REQUIRED_ARG, 0, OPT_NOT_DADDRESS},
     ("destination address does not match this wildcard IP.")},
    {{"not-any-address",   REQUIRED_ARG, 0, OPT_NOT_ANY_ADDRESS},
     ("neither source nor destination address matches this wildcard IP.")},
    {{"next-hop-id",       REQUIRED_ARG, 0, OPT_NEXT_HOP_ID},
     ("next hop address does not match this wildcard IP.")},

    {{"input-index",       REQUIRED_ARG, 0, OPT_INPUT_INDEX},
     ("SNMP input interface  is contained in this list.\n"
      "\t A comma separated list of numbers and ranges between 0 and 65535.")},
    {{"output-index",      REQUIRED_ARG, 0, OPT_OUTPUT_INDEX},
     ("SNMP output interface  is contained in this list.\n"
      "\t A comma separated list of numbers and ranges between 0 and 65535.")},

    {{"tcp-flags",         REQUIRED_ARG, 0, OPT_TCP_FLAGS},
     ("TCP flags are in the list in [CEUAPRSF] where\n"
      "\tC=CWR;E=ECE;U=URG;A=ACK;P=PSH;R=RST;S=SYN;F=FIN")},
    {{"fin-flag",          REQUIRED_ARG, 0, OPT_FIN_FLAG},
     ("FIN flag is present if arg is 1, absent if arg is 0")},
    {{"syn-flag",          REQUIRED_ARG, 0, OPT_SYN_FLAG},
     ("SYN flag is present if arg is 1, absent if arg is 0")},
    {{"rst-flag",          REQUIRED_ARG, 0, OPT_RST_FLAG},
     ("RST flag is present if arg is 1, absent if arg is 0")},
    {{"psh-flag",          REQUIRED_ARG, 0, OPT_PSH_FLAG},
     ("PSH flag is present if arg is 1, absent if arg is 0")},
    {{"ack-flag",          REQUIRED_ARG, 0, OPT_ACK_FLAG},
     ("ACK flag is present if arg is 1, absent if arg is 0")},
    {{"urg-flag",          REQUIRED_ARG, 0, OPT_URG_FLAG},
     ("URG flag is present if arg is 1, absent if arg is 0")},
    {{"ece-flag",          REQUIRED_ARG, 0, OPT_ECE_FLAG},
     ("ECE flag is present if arg is 1, absent if arg is 0")},
    {{"cwr-flag",          REQUIRED_ARG, 0, OPT_CWR_FLAG},
     ("CWR flag is present if arg is 1, absent if arg is 0")},
    {{"flags-all",         REQUIRED_ARG, 0, OPT_FLAGS_ALL},
     ("flags on all packets: <high-flags>/<care-flags>")},


    {{0, 0, 0, 0}, /* sentinel */
     NULL}
};


/* Print usage to specified file handle */
void filterUsage(FILE *fh)
{
    int i;

    fprintf(fh, ("\nPARTITIONING SWITCHES determine whether to pass"
                 " or fail a flow-record.\n\t The flow will fail unless"
                 " each of the following is true:\n\n"));
    for (i = 0; filterSwitch[i].option.name; i++) {
        fprintf(fh, "--%s %s. %s\n", filterSwitch[i].option.name,
                SK_OPTION_HAS_ARG(filterSwitch[i].option),
                filterSwitch[i].help);
    }
}


/*
 *    The options handler for the switches that this file registers.
 *
 *    Parses the user's values to the switches and fills in the global
 *    'fltrRulesPtr' variable.  Returns 0 on success or 1 on failure.
 */
static int filterOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    static int option_seen[NUM_RULE_CLASSES];
    int rule_key = opt_index;
    int found_dup = 0;
    int rv = 1;
    int i;

    /* Make certain this rule is not a repeat */
    switch (opt_index) {
      case OPT_SADDRESS:
      case OPT_NOT_SADDRESS:
        if (option_seen[OPT_NOT_SADDRESS]) {
            skAppPrintErr("A --%s filter has already been set",
                          filterOptions[OPT_NOT_SADDRESS].name);
            return 1;
        }
        if (option_seen[OPT_SADDRESS]) {
            skAppPrintErr("A --%s filter has already been set",
                          filterOptions[OPT_SADDRESS].name);
            return 1;
        }
        option_seen[opt_index] = 1;
        rule_key = OPT_SADDRESS;
        break;

      case OPT_DADDRESS:
      case OPT_NOT_DADDRESS:
        if (option_seen[OPT_NOT_DADDRESS]) {
            skAppPrintErr("A --%s filter has already been set",
                          filterOptions[OPT_NOT_DADDRESS].name);
            return 1;
        }
        if (option_seen[OPT_DADDRESS]) {
            skAppPrintErr("A --%s filter has already been set",
                          filterOptions[OPT_DADDRESS].name);
            return 1;
        }
        option_seen[opt_index] = 1;
        rule_key = OPT_DADDRESS;
        break;

      case OPT_ANY_ADDRESS:
      case OPT_NOT_ANY_ADDRESS:
        if (option_seen[OPT_NOT_ANY_ADDRESS]) {
            skAppPrintErr("A --%s filter has already been set",
                          filterOptions[OPT_NOT_ANY_ADDRESS].name);
            return 1;
        }
        if (option_seen[OPT_ANY_ADDRESS]) {
            skAppPrintErr("A --%s filter has already been set",
                          filterOptions[OPT_ANY_ADDRESS].name);
            return 1;
        }
        option_seen[opt_index] = 1;
        rule_key = OPT_ANY_ADDRESS;
        break;

      case OPT_FIN_FLAG:
      case OPT_SYN_FLAG:
      case OPT_RST_FLAG:
      case OPT_PSH_FLAG:
      case OPT_ACK_FLAG:
      case OPT_URG_FLAG:
      case OPT_ECE_FLAG:
      case OPT_CWR_FLAG:
        if (option_seen[opt_index]) {
            skAppPrintErr("A --%s filter has already been set",
                          filterOptions[opt_index].name);
            return 1;
        }
        rule_key = OPT_FLAGS_ALL;
        break;

      case OPT_FLAGS_ALL:
      case OPT_FLAGS_INITIAL:
      case OPT_FLAGS_SESSION:
        /* these can be repeated */
        break;

      default:
        if (option_seen[opt_index]) {
            skAppPrintErr("A --%s filter has already been set",
                          filterOptions[opt_index].name);
            return 1;
        }
        break;
    }

    /* check that this is not a repeated rule */
    for (i = 0; i < fltrRulesPtr->maxRulesUsed; ++i) {
        if (rule_key == fltrRulesPtr->ruleSet[i]) {
            found_dup = 1;
            break;
        }
    }

    if (found_dup == 0) {
        /* First time for this rule */
        fltrRulesPtr->ruleSet[fltrRulesPtr->maxRulesUsed] = rule_key;
        fltrRulesPtr->maxRulesUsed++;
    }


    /* Parse the parameter to the rule */
    switch (opt_index) {

      case OPT_STIME:
        rv = parseTimeIntoValueRange(&fltrRulesPtr->sTime , opt_arg);
        break;

      case OPT_ETIME:
        rv = parseTimeIntoValueRange(&fltrRulesPtr->eTime , opt_arg);
        break;

      case OPT_ACTIVE_TIME:
        rv = parseTimeIntoValueRange(&fltrRulesPtr->activeTime, opt_arg);
        break;

      case OPT_DURATION:
        rv = parseIntegerRange(&fltrRulesPtr->elapsed, opt_arg);
        break;

      case OPT_SPORT:
        rv = skStringParseNumberListToBitmap(fltrRulesPtr->sPort, opt_arg,
                                             MAX_PORTS);
        break;

      case OPT_DPORT:
        rv = skStringParseNumberListToBitmap(fltrRulesPtr->dPort, opt_arg,
                                             MAX_PORTS);
        break;

      case OPT_APORT:
        rv = skStringParseNumberListToBitmap(fltrRulesPtr->any_port, opt_arg,
                                             MAX_PORTS);
        break;

      case OPT_PROTOCOL:
        rv = skStringParseNumberListToBitmap(fltrRulesPtr->proto, opt_arg,
                                             MAX_PROTOCOLS);
        break;

      case OPT_BYTES:
        rv = parseIntegerRange(&fltrRulesPtr->bytes, opt_arg);
        break;

      case OPT_PACKETS:
        rv = parseIntegerRange(&fltrRulesPtr->pkts, opt_arg);
        break;

      case OPT_BYTES_PER_PACKET:
        rv = parseDecimalRange(&fltrRulesPtr->bytesPerPacket, opt_arg);
        break;

      case OPT_NOT_SADDRESS:
        fltrRulesPtr->negateIP[SRC] = 1;
        /* FALLTHROUGH */
      case OPT_SADDRESS:
        rv = skIPStringToBitmap(fltrRulesPtr->IP[SRC].octets, opt_arg);
        break;

      case OPT_NOT_DADDRESS:
        fltrRulesPtr->negateIP[DST] = 1;
        /* FALLTHROUGH */
      case OPT_DADDRESS:
        rv = skIPStringToBitmap(fltrRulesPtr->IP[DST].octets, opt_arg);
        break;

      case OPT_NOT_ANY_ADDRESS:
        fltrRulesPtr->negateIP[ANY] = 1;
        /* FALLTHROUGH */
      case OPT_ANY_ADDRESS:
        rv = skIPStringToBitmap(fltrRulesPtr->IP[ANY].octets, opt_arg);
        break;

      case OPT_NEXT_HOP_ID:
        rv = skIPStringToBitmap(fltrRulesPtr->nhIP.octets, opt_arg);
        break;

      case OPT_INPUT_INDEX:
        rv = skStringParseNumberListToBitmap(fltrRulesPtr->input, opt_arg,
                                             SK_SNMP_INDEX_LIMIT);
        break;

      case OPT_OUTPUT_INDEX:
        rv = skStringParseNumberListToBitmap(fltrRulesPtr->output, opt_arg,
                                             SK_SNMP_INDEX_LIMIT);
        break;

      case OPT_TCP_FLAGS:
        rv = skStringParseTCPFlags(&fltrRulesPtr->flags, opt_arg);
        break;

      case OPT_FLAGS_ALL:
      case OPT_FLAGS_INITIAL:
      case OPT_FLAGS_SESSION:
        rv = parseFlags(fltrRulesPtr, opt_index, opt_arg);
        break;

        /*
         * This section is all flag wonking
         */
      case OPT_FIN_FLAG:
      case OPT_SYN_FLAG:
      case OPT_RST_FLAG:
      case OPT_PSH_FLAG:
      case OPT_ACK_FLAG:
      case OPT_URG_FLAG:
      case OPT_ECE_FLAG:
      case OPT_CWR_FLAG:
        rv = setFilterRuleBinaryFlag(fltrRulesPtr, opt_index, opt_arg);
        break;

      case OPT_ATTRIBUTES:
        rv = parseAttributes(fltrRulesPtr, opt_index, opt_arg);
        break;

      case OPT_APPLICATION:
        rv = skStringParseNumberListToBitmap(fltrRulesPtr->application,
                                             opt_arg, MAX_PORTS);
        break;

      default:
        skAppPrintErr("Option %d not handled in switch() at %s:%d",
                      opt_index, __FILE__, __LINE__);
        abort();

    } /* switch */


    if (rv != 0) {
        skAppPrintErr("Error parsing --%s option: '%s'",
                      filterOptions[opt_index].name, opt_arg);
        return 1;
    }
    return 0;                     /* OK */
}


/*
 *  pass = filterCheck(&rwrec)
 *
 *    Check the rw record 'rwrec' against all of the rules the user
 *    specified.  If the record fails any rule, the (rule_number+1) is
 *    returned.  If the record passes all of the rules, 0 is returned.
 */
int filterCheck(const rwRec *rwrec)
{
    unsigned int i;
    int j;
    int wanted;


/* If (test) is zero, the record fails the filter and a non-zero value
 * is returned.  Else we continue with additional checks. */
#define FILTER_CHECK(test) \
    if (test) {/*pass*/} else return (1 + fltrRulesPtr->ruleSet[j])


    for (j = 0; j < fltrRulesPtr->maxRulesUsed; j++) {
        switch (fltrRulesPtr->ruleSet[j]) {

          case OPT_STIME:
            FILTER_CHECK(CHECK_RANGE(rwrec->sTime, fltrRulesPtr->sTime));
            break;

          case OPT_ETIME:
            FILTER_CHECK(CHECK_RANGE(rwrec->sTime + rwrec->elapsed,
                                     fltrRulesPtr->eTime));
            break;

          case OPT_ACTIVE_TIME:
            /* to pass the record; check that flow's start time is
             * less than the max value of range and that flow's end
             * time is greater than the min value of the range. */
            FILTER_CHECK((rwrec->sTime <= fltrRulesPtr->activeTime.max)
                         && ((rwrec->sTime + rwrec->elapsed)
                             >= fltrRulesPtr->activeTime.min));
            break;

          case OPT_DURATION:
            FILTER_CHECK(CHECK_RANGE(rwrec->elapsed, fltrRulesPtr->elapsed));
            break;

          case OPT_SPORT:
            FILTER_CHECK(BITMAP_GETBIT(fltrRulesPtr->sPort, rwrec->sPort));
            break;

          case OPT_DPORT:
            FILTER_CHECK(BITMAP_GETBIT(fltrRulesPtr->dPort, rwrec->dPort));
            break;

          case OPT_APORT:
            FILTER_CHECK(BITMAP_GETBIT(fltrRulesPtr->any_port, rwrec->sPort) ||
                         BITMAP_GETBIT(fltrRulesPtr->any_port, rwrec->dPort));
            break;

          case OPT_PROTOCOL:
            FILTER_CHECK(BITMAP_GETBIT(fltrRulesPtr->proto, rwrec->proto));
            break;

          case OPT_BYTES:
            FILTER_CHECK(CHECK_RANGE(rwrec->bytes, fltrRulesPtr->bytes));
            break;

          case OPT_PACKETS:
            FILTER_CHECK(CHECK_RANGE(rwrec->pkts, fltrRulesPtr->pkts));
            break;

          case OPT_BYTES_PER_PACKET:
            FILTER_CHECK(CHECK_RANGE((((uint64_t)(rwrec->bytes)) * PRECISION
                                      / ((uint64_t)(rwrec->pkts))),
                                     fltrRulesPtr->bytesPerPacket));
            break;

          case OPT_NOT_SADDRESS:
          case OPT_SADDRESS:
            /* Check if record's sIP matches the bitmap.  The record
             * FAILS the filter when the result of the check MATCHES
             * the status of the negate flag.  E.g., the record
             * matches the address-bitmap (checkAddress()==1) and the
             * user entered --not-saddr (negateIP==1).
             *
             * In the test below, when the two values are equal, the
             * XOR(^) of them will be FALSE, and !FALSE is TRUE.
             */
            FILTER_CHECK(checkAddress(rwrec->sIP.ipnum, &fltrRulesPtr->IP[SRC])
                         ^ fltrRulesPtr->negateIP[SRC]);
            break;

          case OPT_NOT_DADDRESS:
          case OPT_DADDRESS:
            FILTER_CHECK(checkAddress(rwrec->dIP.ipnum, &fltrRulesPtr->IP[DST])
                         ^ fltrRulesPtr->negateIP[DST]);
            break;

          case OPT_NOT_ANY_ADDRESS:
          case OPT_ANY_ADDRESS:
            FILTER_CHECK((checkAddress(rwrec->sIP.ipnum,
                                       &(fltrRulesPtr->IP[ANY]))
                          | checkAddress(rwrec->dIP.ipnum,
                                         &(fltrRulesPtr->IP[ANY])))
                         ^ fltrRulesPtr->negateIP[ANY]);
            break;

          case OPT_NEXT_HOP_ID:
            FILTER_CHECK(checkAddress(rwrec->nhIP.ipnum,
                                      &(fltrRulesPtr->nhIP)));
            break;

          case OPT_INPUT_INDEX:
            FILTER_CHECK(BITMAP_GETBIT(fltrRulesPtr->input, rwrec->input));
            break;

          case OPT_OUTPUT_INDEX:
            FILTER_CHECK(BITMAP_GETBIT(fltrRulesPtr->output, rwrec->output));
            break;

            /*
             * TCP check.  Passes if there's an intersection between
             * the raised flags and the filter flags.
             */
          case OPT_TCP_FLAGS:
            FILTER_CHECK(fltrRulesPtr->flags & rwrec->flags);
            break;

          case OPT_FLAGS_ALL:
            wanted = 0;
            for (i = 0; i < fltrRulesPtr->count_flags_all; ++i) {
                if (CHECK_TCP_HIGH_MASK(rwrec->flags,
                                        fltrRulesPtr->flags_all[i]))
                {
                    wanted = 1;
                    break; /* wanted */
                }
            }
            FILTER_CHECK(wanted);
            break;

          case OPT_FLAGS_INITIAL:
            wanted = 0;
            for (i = 0; i < fltrRulesPtr->count_flags_init; ++i) {
                if (CHECK_TCP_HIGH_MASK(rwrec->init_flags,
                                        fltrRulesPtr->flags_init[i]))
                {
                    wanted = 1;
                    break; /* wanted */
                }
            }
            FILTER_CHECK(wanted);
            break;

          case OPT_FLAGS_SESSION:
            wanted = 0;
            for (i = 0; i < fltrRulesPtr->count_flags_session; ++i) {
                if (CHECK_TCP_HIGH_MASK(rwrec->rest_flags,
                                        fltrRulesPtr->flags_session[i]))
                {
                    wanted = 1;
                    break;
                }
            }
            FILTER_CHECK(wanted);
            break;

          case OPT_ATTRIBUTES:
            FILTER_CHECK(fltrRulesPtr->attributes & rwrec->tcp_state);
            break;

          case OPT_APPLICATION:
            FILTER_CHECK(BITMAP_GETBIT(fltrRulesPtr->application,
                                       rwrec->application));
            break;

          case OPT_SENSORS:
            FILTER_CHECK(BITMAP_GETBIT(fltrRulesPtr->sID, rwrec->sID));
            break;

          case OPT_FLOW_TYPE:
            FILTER_CHECK(BITMAP_GETBIT(fltrRulesPtr->flow_type,
                                       rwrec->flow_type));
            break;

          default:
            skAppPrintErr("Unhandled value '%d' in switch() at %s:%d",
                          fltrRulesPtr->ruleSet[j], __FILE__, __LINE__);
            return (1 + fltrRulesPtr->ruleSet[j]);
        } /* switch */
    } /* outer for */

    return 0;                     /* WANTED! */
}


/*
 * int filterSetup()
 *
 * SUMMARY:
 *
 * Called by the application to let the filter library setup for
 * options processing.
 *
 * RETURNS:
 *
 * on failure, returns 1
 * on success, returns 0
 *
 * SIDE EFFECTS:
 *
 * Initializes the filter library options with the options processing library.
 *
 * NOTES:
 *
 */
int filterSetup(void)
{
    int i;

    /* make certain we have enough space  */
    assert(_OPT_FINAL_OPTION_ < NUM_RULE_CLASSES);

    /* clear memory */
    memset(fltrRulesPtr, 0, sizeof(filter_rules_t));

    /* copy options from filterSwitch[] to filterOptions[] */
    if (filterOptions == NULL) {
        filterOptions = calloc(sizeof(filterSwitch)/sizeof(filter_switch_t),
                               sizeof(struct option));
        for (i = 0; filterSwitch[i].option.name != NULL; ++i) {
            memcpy(&(filterOptions[i]), &(filterSwitch[i].option),
                   sizeof(struct option));
        }
    }

    /* register options */
    if (optionsRegister(filterOptions, (optHandler)filterOptionsHandler, NULL))
    {
        return 1;
    }

    return 0;
}


/*
 * void filterTeardown()
 *
 */
void filterTeardown(void)
{
    return;
}


/* Return number of rules active. */
int  filterGetRuleCount(void)
{
    return fltrRulesPtr->maxRulesUsed;
}


/*
 *  status = filterGetFGlobFilters();
 *
 *    Create filter rules that correspond to the --sensor, --class,
 *    --type switches from the fglob() code by getting those values
 *    from fglob and creating filters for them.
 */
int filterGetFGlobFilters(void)
{
    int rv;

    /* Let fglob fill in the bitmaps with the sensors and class/types
     * to filter over. */
    rv = fglobSetFilters(fltrRulesPtr->sID,
                         fltrRulesPtr->flow_type);

    if (rv < 0) {
        /* error */
        return rv;
    }
    if (rv & 1) {
        /* need to filter over sensor */
        fltrRulesPtr->ruleSet[fltrRulesPtr->maxRulesUsed] = OPT_SENSORS;
        fltrRulesPtr->maxRulesUsed++;
    }
    if (rv & 2) {
        /* need to filter over class/type */
        fltrRulesPtr->ruleSet[fltrRulesPtr->maxRulesUsed] = OPT_FLOW_TYPE;
        fltrRulesPtr->maxRulesUsed++;
    }

    return 0;
}


/*
 * static int parseTimeIntoValueRange(value_range_t *p_vr, char *s_time)
 *
 * SUMMARY:
 *
 * Parses a time string using skStringParseDatetimeRange().  Stores
 * the result in the value_range_t pointed to by p_vr.
 *
 * PARAMETERS:
 *
 * value_range_t *p_vr = pointer to the time range storage variable
 * char *s_time = a string containing a time value to parse
 *
 * RETURNS:
 *
 * 0 on success
 * -1 on error, such as a badly formatted string, an end time occurring
 *   before a start time, etc.
 *
 * SIDE EFFECTS:
 *
 * on success, p_vr is set to [start,end], or if no end-time was parsed,
 * p_vr is set to [start,start]
 *
 * on failures, the contents of p_vr are not modified
 */
static int parseTimeIntoValueRange(
    value_range_t  *p_vr,
    const char     *s_time)
{
    uint32_t min = 0;
    uint32_t max = 0;
    uint32_t max_rounded = 0;
    int max_precision = 0;

    /* parse the time */
    if (skStringParseDatetimeRange(&min, &max, s_time, NULL, &max_precision)) {
        return -1;
    }

    p_vr->min = min;

    /* maximum */
    if ( max == 0xFFFFFFFF ) {
        /* there was no max date parsed */
        p_vr->max = p_vr->min;
    } else {
        /* if the max date precision is less than (courser than)
         * second resolution, "round" the date up */
        if ( max_precision < 6 ) {
            if ( skDatetimeCeiling( &max_rounded, max, max_precision ) != 0 ) {
                return -1;
            }
            p_vr->max = max_rounded;
        } else {
            p_vr->max = max;
        }
    }

    return 0;
}


/*
 * uint32_t parseIntegerRange(value_range_t *tgtRange, char *rangeString)
 *
 * SUMMARY:
 *
 * parses a range of values (i.e.: a-b)
 *
 * PARAMETERS:
 *
 * value_range_t *tgtRange: the resulting range.
 * char *rangeString: a string containing a representation of the range.
 *
 * RETURNS:
 *
 * 0 on success; 1 on failure.
 *
 * SIDE EFFECTS:
 *
 * value_range_t is altered.
 */
static uint32_t parseIntegerRange(
    value_range_t  *tgtRange,
    const char     *rangeString)
{
    int rv;

    /* parse first part of range; rv, if positive, will be location of
     * start of upper half of range. */
    rv = skStringParseUint64(&tgtRange->min, rangeString, 0, 0);
    if (rv < 0) {
        /* error */
        skAppPrintErr("error parsing lower half of range");
        return 1;
    }

    if (rv == 0) {
        /* missing upper half */
        skAppPrintErr("range missing upper limit");
        return 1;
    }

    /* check for the hyphen */
    if (rangeString[rv - 1] != '-') {
        /* stopped on non-hyphen */
        skAppPrintErr("error parsing lower half of range");
        return 1;
    }

    /* see if open-ended range: "3-" */
    if (rangeString[rv] == '\0') {
        tgtRange->max = UINT64_MAX;
        return 0;
    }

    /* parse upper part of range */
    rv = skStringParseUint64(&tgtRange->max, &rangeString[rv], 0, 0);
    if (rv != 0) {
        /* error */
        skAppPrintErr("error parsing upper half of range");
        return 1;
    }

    if (tgtRange->max < tgtRange->min) {
        skAppPrintErr("range maximum less than minimum");
        return 1;
    }

    return 0;
}


/*
 * uint32_t parseDecimalRange(value_range_t *tgtRange, char *rangeString)
 *
 * SUMMARY:
 *
 * parses a range of values (i.e.: a-b)
 *
 * PARAMETERS:
 *
 * value_range_t *tgtRange: the resulting range.
 * char *rangeString: a string containing a representation of the range.
 *
 * RETURNS:
 *
 * 0 on success; 1 on failure.
 *
 * SIDE EFFECTS:
 *
 * value_range_t is altered.
 *
 * NOTES:
 *
 * rename this to parsefixedpointrange at some point.
 */
static uint32_t parseDecimalRange(
    value_range_t  *tgtRange,
    const char     *rangeString)
{
    const char *sp = rangeString;
    char *ep;
    double val;
    int i;

    /* eat leading whitespace */
    while (*sp && isspace((int)*sp)) {
        ++sp;
    }
    if ('\0' == *sp) {
        skAppPrintErr("range is empty");
        return 1;
    }

    /* loop twice: once for START of range and once for END */
    for (i = 0; i < 2; ++i) {
        if (('\0' == *sp) || ('-' == *sp)) {
            /* empty string, whitespace only, or begins with hyphen */
            skAppPrintErr("error parsing %s half of range",
                          ((i == 0) ? "lower" : "upper"));
            return 1;
        }

        /* parse the string */
        errno = 0;
        val = strtod(sp, &ep);
        if ((sp == ep) || (val == HUGE_VAL && errno == ERANGE)) {
            /* parse error or overflow */
            skAppPrintErr(("error parsing %s half of range: "
                           "cannot parse number"),
                          ((i == 0) ? "lower" : "upper"));
            return 1;
        }
        sp = ep;

        if (val > UINT32_MAX) {
            /* too big */
            skAppPrintErr(("error parsing %s half of range: "
                           "value too large"),
                          ((i == 0) ? "lower" : "upper"));
            return 1;
        }

        val *= PRECISION;
        if (i == 0) {
            /* start of range */
            if ('\0' == *sp) {
                skAppPrintErr("range missing upper half");
                return 1;
            }
            if ('-' != *sp) {
                /* junk at end */
                skAppPrintErr("error parsing lower half of range: "
                              "unexpected char after value");
                return 1;
            }
            /* skip hyphen */
            ++sp;
            tgtRange->min = (uint64_t)val;
            /* check for an open-ended range */
            if ('\0' == *sp) {
                tgtRange->max = UINT64_MAX;
                break;
            }
        } else {
            /* end of range */
            /* eat trailing whitespace */
            while (*sp && isspace((int)*sp)) {
                ++sp;
            }
            if ('\0' != *sp) {
                /* junk at end */
                skAppPrintErr("error parsing upper half of range: "
                              "unexpected char after value");
                return 1;
            }
            tgtRange->max = (uint64_t)val;
        }
    }

    if (tgtRange->max < tgtRange->min) {
        skAppPrintErr("range maximum less than range minimum");
        return 1;
    }

    return 0;
}


/*
 *  status = parseFlags(rules, opt_index, opt_arg)
 *
 */
static int parseFlags(
    filter_rules_t *rules,
    int             opt_index,
    const char     *opt_arg)
{
    uint8_t *rule_count;
    high_mask_t *h_m;

    switch (opt_index) {
      case OPT_FLAGS_ALL:
        rule_count = &rules->count_flags_all;
        h_m = &(rules->flags_all[*rule_count]);
        break;
      case OPT_FLAGS_INITIAL:
        rule_count = &rules->count_flags_init;
        h_m = &(rules->flags_init[*rule_count]);
        break;
      case OPT_FLAGS_SESSION:
        rule_count = &rules->count_flags_session;
        h_m = &(rules->flags_session[*rule_count]);
        break;
      default:
        skAppPrintErr("Unhandled value '%d' in switch() at %s:%d",
                      opt_index, __FILE__, __LINE__);
        assert(0);
        exit(EXIT_FAILURE);
    }

    if (*rule_count >= MAX_TCPFLAG_RULES) {
        skAppPrintErr("May only specify %d %s rules",
                      MAX_TCPFLAG_RULES, filterOptions[opt_index].name);
        return 1;
    }

    if (PARSE_TCP_HIGH_MASK(*h_m, opt_arg)) {
        return 1;
    }

    ++*rule_count;
    return 0;
}


/*
 *  status = setFilterRuleBinaryFlag(rules, opt_index, opt_arg);
 *
 *    Set the 'flagMark' and 'flagCare' for the TCP flag 'flag'
 *    according to whether 'opt_arg' parses as a 0 or a 1.
 *    'opt_index' is used to report an error.
 */
static int setFilterRuleBinaryFlag(
    filter_rules_t *rules,
    int             opt_index,
    const char     *opt_arg)
{
    static uint8_t binary_flag_pos = MAX_TCPFLAG_RULES;
    uint8_t flag;
    unsigned long i;

    if (binary_flag_pos == MAX_TCPFLAG_RULES) {
        /* first time we've called this function */
        if (rules->count_flags_all >= MAX_TCPFLAG_RULES) {
            skAppPrintErr("May only specify %d tcp-flags rules",
                          MAX_TCPFLAG_RULES);
            return 1;
        }
        binary_flag_pos = rules->count_flags_all;
        ++rules->count_flags_all;
    }

    switch (opt_index) {
      case OPT_FIN_FLAG:  flag = FIN_FLAG;  break;
      case OPT_SYN_FLAG:  flag = SYN_FLAG;  break;
      case OPT_RST_FLAG:  flag = RST_FLAG;  break;
      case OPT_PSH_FLAG:  flag = PSH_FLAG;  break;
      case OPT_ACK_FLAG:  flag = ACK_FLAG;  break;
      case OPT_URG_FLAG:  flag = URG_FLAG;  break;
      case OPT_ECE_FLAG:  flag = ECE_FLAG;  break;
      case OPT_CWR_FLAG:  flag = CWR_FLAG;  break;
      default:
        skAppPrintErr("Unhandled value '%d' in switch() at %s:%d",
                      opt_index, __FILE__, __LINE__);
        assert(0);
        exit(EXIT_FAILURE);
    }

    i = strtoul(opt_arg, (char **)NULL, 10);
    switch (i) {
      case 1:
        TCP_FLAG_SET_FLAG(rules->flags_all[binary_flag_pos].high, flag);
        /* FALLTHROUGH */
      case 0:
        TCP_FLAG_SET_FLAG(rules->flags_all[binary_flag_pos].mask, flag);
        break;
      default:
        skAppPrintErr("Error parsing --%s option: '%s'",
                      filterOptions[opt_index].name, opt_arg);
        return 1;
    }

    return 0;
}


/*
 *  status = parseAttributes(rules, opt_index, opt_arg);
 *
 */
static int parseAttributes(
    filter_rules_t *rules,
    int             opt_index,
    const char     *opt_arg)
{
    const char *cp = opt_arg;

    while (*cp) {
        switch (*cp) {
          case 'F':
            rules->attributes |= SK_TCPSTATE_FIN_FOLLOWED_NOT_ACK;
            break;
          case 'T':
            rules->attributes |= SK_TCPSTATE_TIMEOUT_KILLED;
            break;
          case 'C':
            rules->attributes |= SK_TCPSTATE_TIMEOUT_STARTED;
            break;
          default:
            skAppPrintErr("Unrecognized character '%c' in %s switch",
                          *cp, filterOptions[opt_index].name);
            return 1;
        }
        ++cp;
    }

    return 0;
}


/*
 *  checkAddress(uint32_t tgtAddress, const ip_addr_range_t *maskRange)
 *
 *    checkAddress verifies whether a particular address falls within
 *    an address range.
 *
 *    Returns 1 if a successful match. O otherwise.
 */
static int checkAddress(uint32_t tgtAddress, const ip_addr_range_t *maskRange)
{
    ipUnion address;
    address.ipnum = tgtAddress;

    return (BITMAP_GETBIT(maskRange->octets[0], address.O.o1) &&
            BITMAP_GETBIT(maskRange->octets[1], address.O.o2) &&
            BITMAP_GETBIT(maskRange->octets[2], address.O.o3) &&
            BITMAP_GETBIT(maskRange->octets[3], address.O.o4) &&
            1);
}
