/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwfilterutils.c
**
**  utility routines for rwfilter.c
**
**  Suresh L. Konda
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwfilter/rwfilterutils.c,v 1.38 2005/11/17 22:51:24 thomasm Exp $");

#include "rwfilter.h"


/* TYPEDEFS AND DEFINES */



/* LOCAL FUNCTIONS */

static void appUsage(void);
static void appUsageLong(void);
static void appOptionsUsage(FILE*);
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static int  filterAddDynlib(const char *dlpath, int warnOnError);
static int  filterCheckInputs(int argc);
static int  filterCheckOutputs(void);
static int  filterSetCheckers(void);


/* LOCAL VARIABLES */

/* OPTION SETUP */

/* these values used to index into following arrays.  need to keep
 * everything in sync */
typedef enum {
    RWFT_OPT_DRY_RUN, RWFT_OPT_INPUT_PIPE, RWFT_OPT_XARGS,
    RWFT_OPT_MAX_PASS_RECORDS, RWFT_OPT_PRINT_FILE, RWFT_OPT_DYNLIB,
    RWFT_OPT_PASS_DEST, RWFT_OPT_FAIL_DEST, RWFT_OPT_ALL_DEST,
    RWFT_OPT_PRINT_STAT, RWFT_OPT_PRINT_VOLUME
} appOptionsEnum;

static struct option appOptions[] = {
    {"dry-run",                 NO_ARG,       0, RWFT_OPT_DRY_RUN},
    {"input-pipe",              REQUIRED_ARG, 0, RWFT_OPT_INPUT_PIPE},
    {"xargs",                   REQUIRED_ARG, 0, RWFT_OPT_XARGS},
    {"max-pass-records",        REQUIRED_ARG, 0, RWFT_OPT_MAX_PASS_RECORDS},
    {"print-filenames",         NO_ARG,       0, RWFT_OPT_PRINT_FILE},
    {OPT_DYNAMIC_LIBRARY,       REQUIRED_ARG, 0, RWFT_OPT_DYNLIB},

    {"pass-destination",        REQUIRED_ARG, 0, RWFT_OPT_PASS_DEST},
    {"fail-destination",        REQUIRED_ARG, 0, RWFT_OPT_FAIL_DEST},
    {"all-destination",         REQUIRED_ARG, 0, RWFT_OPT_ALL_DEST},
    {"print-statistics",        OPTIONAL_ARG, 0, RWFT_OPT_PRINT_STAT},
    {"print-volume-statistics", OPTIONAL_ARG, 0, RWFT_OPT_PRINT_VOLUME},
    {0,0,0,0}                   /* sentinel entry */
};

static const char *appHelp[] = {
    "parse command line switches but do not process records",
    ("read SiLK flow records from a pipe: 'stdin' or\n"
     "\tpath to named pipe. No default"),
    ("read list of input file names from a file or pipe\n"
     "\tpathname or 'stdin'. No default"),
    "print at most this many records; 0 for all.  Def 0",
    "print names of input files during processing. Def. NO",
    ("augment processing with the specified dynamic\n"
     "\tlibrary. No default\n"
     "At least one of the following output switches is required:"),
    ("destination for records which pass the filter(s):\n"
     "\tpathname or 'stdout'. If pathname, it must not exist. No default"),
    ("destination for records which fail the filter(s):\n"
     "\tpathname or 'stdout'. If pathname, it must not exist. No default"),
    ("destination for all records regardless of pass/fail:\n"
     "\tpathname or 'stdout'. If pathname, it must not exist. No default"),
    ("print a count of total flows read to named file.\n"
     "\tIf no pathname provided, use stderr. No default"),
    ("print count of flows/packets/bytes read\n"
     "\tto named file. If no pathname provided, use stderr. No default"),
    (char *)NULL
};



/*
 *  appUsage:
 *      print usage information to stderr and exit with code 1
 *  Arguments: None
 *  Returns: None
 *  Side Effects: exits application.
 */
static void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 *  appUsageLong:
 *      print usage information to USAGE_FH and exit with code 1.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 *  Arguments: None
 *  Returns: None
 *  Side Effects: exits application.
 */
static void appUsageLong(void)
{
    int j;

#define USAGE_MSG                                                       \
    ("Usage %s <app-opts> <partition-opts> "                            \
     "{<selection-opts> | <inputFiles>}\n"                              \
     "\tPartitions SiLK Flow records into one or more 'pass' and/or\n"  \
     "\t'fail' output streams.  The source of the SiLK records can\n"   \
     "\tbe stdin, a named pipe, files listed on the command line, or\n" \
     "\tfiles selected from the data-store via the selection switches.\n" \
     "\tThere is no default input or output; these must be specified.\n")

    fprintf(USAGE_FH, USAGE_MSG, skAppName());
    appOptionsUsage(USAGE_FH);
    fglobUsage(USAGE_FH);
    filterUsage(USAGE_FH);

    for (j = 0; j < filterDynlibCount; ++j) {
        dynlibOptionsUsage(filterDynlibList[j], USAGE_FH);
    }
    if (dlISP) {
        dynlibOptionsUsage(dlISP, USAGE_FH);
    }

    exit(EXIT_SUCCESS);
}


/*
 *  appOptionsUsage:
 *      print help for options for this app to the specified file handle.
 *   Arguments: None.
 *   Returns: None.
 *   Side Effects: None.
 *   NOTE:
 *       do NOT add non-option usage here (i.e., required and other
 *       optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\nAPPLICATION SWITCHES control input and output\n\n");
    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }
}


/*
 *  appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 *  Arguments:
 *      argc, argv
 *  Returns: None.
 *  Side Effects:
 *      exits the application with code 1 if anything does not work.
 */
void appSetup(int argc, char **argv)
{
    int j;
    int input_count;
    int output_count;

    /* initialize variables */
    skAppRegister(argv[0]);
    memset(&filter_io, 0, sizeof(filter_io_t));
    memset(max_records, 0, sizeof(max_records));

    /* Verify same number of entries in option help and option struct */
    assert((sizeof(appHelp)/sizeof(char*)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* load options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);/* never returns */
    }

    /* load fglob module */
    if (fglobSetup()) {
        skAppPrintErr("unable to setup fglob module");
        exit(EXIT_FAILURE);
    }

    /* load filter module */
    if (filterSetup()) {
        skAppPrintErr("unable to setup filter module");
        exit(EXIT_FAILURE);
    }

    /* register options and the handler */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);
    }

    /* initialize stats data structure */
    filterStatisticsInit(&stats);

    /* add "hard-coded" dynamaic libraries */
    for (j = 0; filterDynlibNames[j]; ++j) {
        filterAddDynlib(filterDynlibNames[j], 0);
    }

    /* dlISP is specified on the command line; if specified it does all
     * filtering--to be consistent with legacy dynlibs.  Since we treat
     * dlISP specially, do not store it in the filterDynlibList[].
     */
    dlISP = dynlibCreate(DYNLIB_EXCL_FILTER);

    /* parse options */
    first_non_switch = optionsParse(argc, argv);
    assert(first_non_switch <= argc);
    if (first_non_switch < 0) {
        appUsage();
    }

    /* check all the output streams */
    output_count = filterCheckOutputs();
    if (output_count < 0) {
        /* fatal error. msg already printed */
        exit(EXIT_FAILURE);
    } else if (output_count == 0) {
        skAppPrintErr("No output(s) specified. Abort");
        appUsage();
    }

    /* Check that there is one and only one source of input to process */
    input_count = filterCheckInputs(argc);
    if (input_count < 0) {
        /* fatal error. msg already printed */
        exit(EXIT_FAILURE);
    } else if (input_count > 1) {
        skAppPrintErr("Multiple input sources were specified\n"
                      "\tOnly one of input-pipe, xargs, fglob, or explicit\n"
                      "\tfile names are permitted.");
        appUsage();
    } else if (input_count == 0) {
        skAppPrintErr("No input was specified.\n"
                      "\t No fglob, input-pipe, nor xargs options"
                      " specified,\n\t and no files present on the"
                      " command line.");
        appUsage();
    }

    /* Check we have a filtering rule--either built in or from a
     * dynamic library--or the --all-dest option */
    checker_count = filterSetCheckers();
    if ((checker_count <= 0) && (filter_io.dest_count[DEST_ALL] == 0)) {
        skAppPrintErr("No filtering rules given");
        appUsage();
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    (void) umask((mode_t) 0002);
    return;/* OK */
}


/*
 *  appOptionsHandler:
 *      called for each option the app has registered.
 *  Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 *  Returns:
 *      0 if OK. 1 else.
 *  Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    switch ((appOptionsEnum)opt_index) {
      case RWFT_OPT_PASS_DEST:
        if (filter_io.dest_count[DEST_PASS] == MAX_DESTINATIONS) {
            skAppPrintErr("Only %d --%s options allowed",
                          MAX_DESTINATIONS, appOptions[opt_index].name);
            return 1;
        }
        filter_io.dest_stream[DEST_PASS][filter_io.dest_count[DEST_PASS]++]
            .fname = opt_arg;
        break;

      case RWFT_OPT_FAIL_DEST:
        if (filter_io.dest_count[DEST_FAIL] == MAX_DESTINATIONS) {
            skAppPrintErr("Only %d --%s options allowed",
                          MAX_DESTINATIONS, appOptions[opt_index].name);
            return 1;
        }
        filter_io.dest_stream[DEST_FAIL][filter_io.dest_count[DEST_FAIL]++]
            .fname = opt_arg;
        break;

      case RWFT_OPT_ALL_DEST:
        if (filter_io.dest_count[DEST_ALL] == MAX_DESTINATIONS) {
            skAppPrintErr("Only %d --%s options allowed",
                          MAX_DESTINATIONS, appOptions[opt_index].name);
            return 1;
        }
        filter_io.dest_stream[DEST_ALL][filter_io.dest_count[DEST_ALL]++].fname
            = opt_arg;
        break;

      case RWFT_OPT_INPUT_PIPE:
        /*
         *  input-pipe. Delay check for multiple inputs sources till all
         *  options have been parsed
         */
        if (filter_io.input.fname) {
            skAppPrintErr("Only one --%s option is allowed",
                          appOptions[opt_index].name);
            return 1;
        }
        filter_io.input.fname = opt_arg;
        break;

      case RWFT_OPT_DYNLIB:
        if(dynlibLoad(dlISP, opt_arg)) {
            skAppPrintErr("fatal error loading library '%s':\n  %s",
                          opt_arg, dynlibLastError(dlISP));
            return 1;
        }
        break;

      case RWFT_OPT_DRY_RUN:
        filter_io.dryrun_fp = DRY_RUN_FH;
        break;

      case RWFT_OPT_PRINT_FILE:
        filter_io.filenames_fp = PRINT_FILENAMES_FH;
        break;

      case RWFT_OPT_PRINT_VOLUME:
        stats.volumeStats = 1;
        /* FALLTHROUGH */
      case RWFT_OPT_PRINT_STAT:
        if (filter_io.stats.fname) {
            skAppPrintErr("may only specified one of --%s or --%s",
                          appOptions[RWFT_OPT_PRINT_STAT].name,
                          appOptions[RWFT_OPT_PRINT_VOLUME].name);
            return 1;
        } else if (!opt_arg) {
            filter_io.stats.fname = "stderr";
        } else {
            filter_io.stats.fname = opt_arg;
        }
        break;

      case RWFT_OPT_MAX_PASS_RECORDS:
        if (skStringParseUint64(&(max_records[DEST_PASS]), opt_arg, 0, 0)) {
            skAppPrintErr("Cannot parse %s value '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

      case RWFT_OPT_XARGS:
        /* input-pipe for file names */
        if (filter_io.xargs.fname) {
            skAppPrintErr("Only one --%s option is allowed",
                          appOptions[opt_index].name);
            return 1;
        }
        filter_io.xargs.fname = opt_arg;
        break;

    } /* switch */

    return 0;                     /* OK */
}


/*
 *  appTeardown:
 *      Teardown all used modules and all application stuff.
 *  Arguments: None
 *  Returns: None
 *  Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Static variable teardownFlag is set.
 *  NOTE: This must be idempotent using static teardownFlag.
 */
void appTeardown(void)
{
    static uint8_t teardownFlag = 0;
    int i, j;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    dynlibTeardown(dlISP);

    for (j = 0; j < filterDynlibCount; ++j) {
        dynlibTeardown(filterDynlibList[j]);
        filterDynlibList[j] = NULL;
    }
    filterDynlibCount = 0;

    filterTeardown();

    fglobTeardown();

    optionsTeardown();

    /* close all the output files (pass, fail, all-dest) */
    for (i = 0; i < DESTINATION_TYPES; ++i) {
        for (j = 0; j < filter_io.dest_count[i]; ++j) {
            if (filter_io.dest_stream[i][j].ios) {
                rwCloseFile(filter_io.dest_stream[i][j].ios);
                filter_io.dest_stream[i][j].ios = NULL;
            }
            filter_io.dest_stream[i][j].fname = NULL;
        }
    }

    /* close stats */
    if (filter_io.stats.fp && filter_io.stats.closefn) {
        if (-1 == (*filter_io.stats.closefn)(filter_io.stats.fp)) {
                skAppPrintErr("Error closing --print-stat stream '%s': %s",
                              filter_io.stats.fname, strerror(errno));
        }
    }
    filter_io.stats.fname = NULL;

    /* close input */
    if (filter_io.input.ios) {
        rwCloseFile(filter_io.input.ios);
        filter_io.input.ios = NULL;
    }
    filter_io.input.fname = NULL;

    /* close xargs */
    if (filter_io.xargs.fp && filter_io.xargs.closefn) {
        (*filter_io.xargs.closefn)(filter_io.xargs.fp);
    }
    filter_io.xargs.fname = NULL;

    return;
}


/*
 * filterAddDynlib(dlpath, warnOnError);
 *
 *   Load the dynamic library at 'dlpath', store its information in
 *   the global 'filterDynlibList', and increment the global
 *   'filterDynlibCount'.  Return 0 on success.  Return 1 if the
 *   library cannot be loaded or if too many dynamic libraries have
 *   been loaded.
 */
static int filterAddDynlib(const char *dlpath, int warnOnError)
{
    dynlibInfoStructPtr dl;

    if (filterDynlibCount == FILTER_MAX_DYNLIBS) {
        skAppPrintErr("Too many dynlibs.  Only %u allowed",
                      FILTER_MAX_DYNLIBS);
        return 1;
    }

    dl = dynlibCreate(DYNLIB_SHAR_FILTER);
    if (dynlibLoad(dl, dlpath)) {
        /* failed to find library, missing symbols, or setup failed */
        if (warnOnError) {
            skAppPrintErr("Error loading dynamic library '%s': %s",
                          dlpath, dynlibLastError(dl));
        }
        dynlibTeardown(dl);
        return 1;
    }

    filterDynlibList[filterDynlibCount] = dl;
    ++filterDynlibCount;

    return 0;
}


/*
 *  stream_count = filterCheckInputs(argc)
 *
 *    Do basic checks for the input, and return the number of inputs
 *    selected from among: command-line files, fglob options, piped
 *    input, or xargs.  Returns -1 on error, such as trying to read
 *    binary records from a TTY.
 *
 *    If --xargs was given, opens that stream.
 */
static int filterCheckInputs(int argc)
{
    int count = 0;

    /* Check for file names on command line */
    if (first_non_switch < argc) {
        ++count;
    }

    /* the input-pipe must be 'stdin' or it must be an existing FIFO.
     * If 'stdin', stdin must not be a TTY, since we expect binary. */
    if (filter_io.input.fname) {
        ++count;
        if (strcmp(filter_io.input.fname, "stdin") == 0) {
            if (FILEIsATty(stdin)) {
                skAppPrintErr("stdin is connected to a terminal. Abort.");
                return -1;
            }
        } else if ( !fileExists(filter_io.input.fname)) {
            skAppPrintErr("input-pipe '%s' does not exist",
                          filter_io.input.fname);
            return -1;
        } else if ( !isFIFO(filter_io.input.fname)) {
            skAppPrintErr("input-pipe '%s' is not a pipe",
                          filter_io.input.fname);
            return -1;
        }
    }

    /* the xargs argument must be 'stdin' (terminal is ok) or it must
     * exist. Open it. */
    if (filter_io.xargs.fname) {
        ++count;
        if (strcmp(filter_io.xargs.fname, "stdin") == 0) {
            filter_io.xargs.fp = stdin;
        } else {
            int is_pipe;

            if (openFile(filter_io.xargs.fname, 0 /*read*/,
                         &filter_io.xargs.fp, &is_pipe))
            {
                skAppPrintErr("Cannot open --xargs stream '%s'",
                              filter_io.xargs.fname);
                return -1;
            }
            if (is_pipe) {
                filter_io.xargs.closefn = pclose;
            } else {
                filter_io.xargs.closefn = fclose;
            }
        }
    }

    if (count == 1) {
        /*
         * Let any traditional fglob options work as filters.
         */
        if (filterGetFGlobFilters()) {
            return -1;
        }
    }

    /* Check if fglob args were given */
    if (fglobValid()) {
        ++count;
    }

    return count;
}


/*
 *  stream_count = filterCheckOutputs()
 *
 *    Do basic checks for the output and open some outputs.  Make
 *    certain only one stream uses stdout.  Open the statistics and
 *    all-dest handles.
 *
 *    Returns -1 on error, or the number of output streams requested.
 *    Note that for this count --dry-run and a WILLPROCESS dynamic
 *    library (the global dlISP) are considered outputs.
 */
static int filterCheckOutputs(void)
{
    int count = 0;
    int stdout_used = 0;
    int i;
    int j;
    int rv;

    /* If the dynamic library is taking over all processing, warn if
       --pass or --fail was also given. */
    if (DYNLIB_WILLPROCESS == dynlibGetStatus(dlISP)) {
        ++count;
        if (filter_io.dest_count[DEST_PASS]) {
            skAppPrintErr("warning: dynamic library overriding "
                          "--pass-destinations");
            for (j = 0; j < filter_io.dest_count[DEST_PASS]; ++j) {
                filter_io.dest_stream[DEST_PASS][j].fname = NULL;
            }
        }
        if (filter_io.dest_count[DEST_FAIL]) {
            skAppPrintErr("warning: dynamic library overriding "
                          "--fail-destinations");
            for (j = 0; j < filter_io.dest_count[DEST_FAIL]; ++j) {
                filter_io.dest_stream[DEST_FAIL][j].fname = NULL;
            }
        }
        if (filter_io.stats.fp || filter_io.stats.fname) {
            skAppPrintErr("warning: dynamic library overriding "
                          "--print-statistics");
            filter_io.stats.fname = NULL;
            filter_io.stats.fp = NULL;
        }
    }

    if (filter_io.dryrun_fp) {
        ++count;
    }

    /* Basic checks: Only allow one output stream to use 'stdout'.
     * Make certain no one uses 'stderr'.  Make certain files either
     * do not exist or have no content if they exist.  For the binary
     * outputs, make certain the stream is not a terminal. */
    for (i = 0; i < DESTINATION_TYPES; ++i) {
        for (j = 0; j < filter_io.dest_count[i]; ++j) {
            if (0==strcmp("stderr", filter_io.dest_stream[i][j].fname)){
                skAppPrintErr("stderr is not a valid output device. Abort.");
                return -1;
            } else if (0==strcmp("stdout", filter_io.dest_stream[i][j].fname)){
                if (FILEIsATty(stdout)) {
                    skAppPrintErr("stdout is connected to a terminal. Abort.");
                    return -1;
                }
                if (stdout_used) {
                    skAppPrintErr("stdout is already allocated. Abort.");
                    return -1;
                }
                stdout_used = 1;
            } else if (fileExists(filter_io.dest_stream[i][j].fname)
                       && !isFIFO(filter_io.dest_stream[i][j].fname)
                       && (fileSize(filter_io.dest_stream[i][j].fname) > 0))
            {
                /* file exists, is regular file, non-zero size: error */
                skAppPrintErr(("Output file '%s' exists.\n"
                               "If you really want to overwrite this file, "
                               "remove it manually and then re-run %s."),
                              filter_io.dest_stream[i][j].fname, skAppName());
                return -1;
            }
        }
    }

    /* Check the STATISTICS stream */
    if (filter_io.stats.fname) {
        int fileid;
        if (strcmp(filter_io.stats.fname, "stdout") == 0) {
            if (stdout_used) {
                skAppPrintErr("stdout is already allocated. Abort.");
                return -1;
            }
            filter_io.stats.fp = stdout;
        } else if (strcmp(filter_io.stats.fname, "stderr") == 0) {
            filter_io.stats.fp = stderr;
        } else if (sscanf(filter_io.stats.fname, "[%d]", &fileid)) {
            /*
             * Check for entering a file id directly surrounded by square
             * brackets This is used in cluster environment to write back to
             * an open file on the master
             */
            /*fprintf(stderr,"opening printsats stream as fildes %d\n",fileid);*/
            filter_io.stats.fp = fdopen(fileid, "w");
            if ( filter_io.stats.fp == (FILE *)NULL) {
                skAppPrintErr("Can't open fildes %d"
                              " as stream for printstats", fileid);
                return -1;
            }
        } else if (!filter_io.dryrun_fp) {
            /* try open the file, unless this is a "dry run" */
            filter_io.stats.fp = fopen(filter_io.stats.fname, "w");
            if ( filter_io.stats.fp == (FILE *)NULL) {
                skAppPrintErr("Can't open file '%s'"
                              " for printstats output", filter_io.stats.fname);
                return -1;
            }
            filter_io.stats.closefn = fclose;
        }
    }
    if (filter_io.stats.fp) {
        ++count;
    }

    /* Open all the record output streams */
    for (i = 0; i < DESTINATION_TYPES; ++i) {
        for (j = 0; j < filter_io.dest_count[i]; ++j) {
            rwIOStruct *rwIOS;
            if ((rv = rwioCreate(&rwIOS, filter_io.dest_stream[i][j].fname,
                                 SK_RWIO_WRITE))
                || (rv = rwioSetFileType(rwIOS, FT_RWFILTER))
                || (rv = rwioOpen(rwIOS)))
            {
                rwioPrintLastErr(rwIOS, rv, &skAppPrintErr);
                rwioDestroy(&rwIOS);
                /* FIXME: Delete previously opened streams */
                return -1;
            }
            filter_io.dest_stream[i][j].ios = rwIOS;
            ++count;
        }
    }

    return count;
}


/*
 *  count = filterSetCheckers();
 *
 *    Set the array of function pointers to the pass/fail checking
 *    routines, and return the number of pointers that were set.  A
 *    return code of 0 means no filtering rules were specified.
 */
static int filterSetCheckers(void)
{
    int j;
    int count = 0;

    if (dynlibCheckLoaded(dlISP)) {
        /* dlISP is always exclusive for legacy reasons */
        checker[count] = dynlibGetRWProcessor(dlISP);
        ++count;
    } else {
        if (filterGetRuleCount() > 0) {
            checker[count] = (int (*)(rwRecPtr))filterCheck;
            ++count;
        }

        for (j = 0; j < filterDynlibCount; ++j) {
            if (dynlibCheckActive(filterDynlibList[j])) {
                checker[count] = dynlibGetRWProcessor(filterDynlibList[j]);
                ++count;
            }
        }
    }

    return count;
}
