/*
** Copyright (C) 2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
** rwgroup
**
** rwgroup is a 'fuzzy uniq' utility that can be used to group together
** records into 'groups' of fields; to work properly, it
** requires a sorted set of fields, because it can only do record-record
** comparison.
**
** command arguments:
**
** --id-fields: id fields are fields which have to be identical to
**   define a group.
** --delta-field: delta field is the field which is allowed to have a
**   delta value
** --delta-value: This is the maximum value by which two consecutive records
**   can differ before they're considered parts of different groups.  It
**   Is an integer, and represents an inclusive value, so if delta is 1 and
**   the field is stime, then records which are a second apart will be
**   considered part of the same group.
** --threshold: minimum number of records to output
** --summarize: Instead of printing out a record with a grouped id, this will
**              print out a summarized record with stime=min time,
**              etime=max time, bytes and packets summarized.
**
** The way that this application works is as follows: a 'group' is
** a set of records where the id-fields are identical, and the
** delta-field changes within delta.  Each group is identified by a
** unique group id stored in next hop ip; group id's start at 0
** and continue on from there.
**
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwgroup/rwgroup.c,v 1.6 2005/10/26 14:23:17 thomasm Exp $");

#include "utils.h"
#include "rwpack.h"


/* TYPEDEFS AND DEFINES */

/* File handle for --help output */
#define USAGE_FH stdout

#define MAX_FIELDS 8
#define MAX_THRESH 65535

typedef enum _fieldsEnum {
  RWSEQ_FID_SIP = 1, RWSEQ_FID_DIP, RWSEQ_FID_SPORT, RWSEQ_FID_DPORT,
  RWSEQ_FID_PROTO, RWSEQ_FID_PKTS, RWSEQ_FID_BYTES, RWSEQ_FID_FLAGS,
  RWSEQ_FID_STIME, RWSEQ_FID_DUR, RWSEQ_FID_ETIME, RWSEQ_FID_SENSOR
} fieldsEnum;



/* LOCAL FUNCTIONS */

static void appExit(int status);
static void appUsage(void);                     /* never returns */
static void appUsageLong(void);                 /* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);    /* never returns */
static void appOptionsUsage(FILE *fh);          /* application options usage*/
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static int cmp_field(
    int             field_id,
    const rwRec    *current_rec,
    const rwRec    *ref_rec,
    uint32_t        delta);
static int process_rec(rwRec *input_rec);


/* LOCAL VARIABLES */

static uint8_t mask_field_id[MAX_FIELDS];
static uint8_t field_count;
static uint8_t delta_field_id;
static uint16_t threshold;
static uint32_t delta, group_id, rec_id;
static rwRec *rec_thresh_buffer;
static rwIOStruct *out_rwios, *in_rwios;
static uint8_t objective_flag = 0;
static uint8_t summarize_flag = 0;
static const char *in_fname = NULL;
static const char *out_fname = NULL;


/* OPTIONS SETUP */

typedef enum _appOptionsEnum {
    RWSEQ_OPT_ID_FIELD, RWSEQ_OPT_DELTA_FIELD, RWSEQ_OPT_DELTA,
    RWSEQ_OPT_THRESH, RWSEQ_OPT_OBJECTIVE,
    RWSEQ_OPT_INPIPE, RWSEQ_OPT_OUTPIPE, RWSEQ_OPT_SUMMARIZE
} appOptionsEnum;

static struct option appOptions[] = {
    {"id-field",        REQUIRED_ARG, 0, RWSEQ_OPT_ID_FIELD},
    {"delta-field",     REQUIRED_ARG, 0, RWSEQ_OPT_DELTA_FIELD},
    {"delta-value",     REQUIRED_ARG, 0, RWSEQ_OPT_DELTA},
    {"rec-threshold",   REQUIRED_ARG, 0, RWSEQ_OPT_THRESH},
    {"objective",       NO_ARG,       0, RWSEQ_OPT_OBJECTIVE},
    {"summarize",       NO_ARG,       0, RWSEQ_OPT_SUMMARIZE},
    {0,0,0,0}           /* sentinel entry */
};

static char *appHelp[] = {
    "Fields to maintain identical; can be used multiple times",
    "Field which can differ within the group, only one allowed",
    "Maximum difference between consecutive values before they're different",
    "Minimum number of records before outputting",
    ("Count values objectively; all delta-field values\n"
     "  must be within delta of the first field"),
    "Output a summary rather than a sequence of records",
    (char *)NULL
};



/*
 *  appExit(status);
 *
 *    Exit the appliction with status.
 *
 */
static void appExit(int status)
{
    appTeardown();
    exit(status);
}


/*
 * appUsage:
 *      print short usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 *      print complete usage information to stderr and exit with code.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                             \
    ("%s: --id-field=<id field> [app opts] [files]\n"                         \
     "\tGroups flows together by specified id and delta fields; marks the\n"  \
     "\tgroup ID in next hop IP; requires pre-sorting.\n")

    fprintf(USAGE_FH, USAGE_MSG, skAppName());
    appOptionsUsage(USAGE_FH);
    exit(EXIT_SUCCESS);
}


/*
 * appOptionsUsage:
 *      print options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int nextOptIndex;

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *))
           == (sizeof(appOptions)/sizeof(struct option)));

    /* Initialize globals */
    skAppRegister(argv[0]);
    group_id = rec_id = 0;

    /* initialize options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);                     /* never returns */
    }

    /* register the options and handler */
    if (optionsRegister(appOptions, (optHandler) appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);                     /* never returns */
    }

    /* parse options */
    nextOptIndex = optionsParse(argc, argv);
    if (nextOptIndex < 0) {
        appUsage();             /* never returns */
    }

    /* check for extra options */
    if (nextOptIndex > argc) {
        skAppPrintErr("unexpected arguments; input must come from stdin.");
        appUsage();             /* never returns */
    }


    if ((field_count == 0) && (delta_field_id == 0)) {
        skAppPrintErr("No Fields Specified");
        appUsage();             /* never returns */
    };

    if ((delta_field_id > 0) && (delta == 0)) {
        skAppPrintErr("No delta specified for delta field.");
        appUsage();             /* never returns */
    };

    /*
     * Opens output IO Stream; if that fails, it terminates
     */
    if (out_fname == NULL) {
        out_fname = "stdout";
    }
    if (rwioCreate(&out_rwios, out_fname, SK_RWIO_WRITE)
        || rwioOpen(out_rwios))
    {
        skAppPrintErr("Couldn't open %s for writing.  Exiting", out_fname);
        appExit(EXIT_FAILURE);
    }

    /*
     * Deprecated for now, but functional.  IO input and output defaults
     * and can only be stdin/stdout.
     */
    if (in_fname == NULL) {
        in_fname = "stdin";
    }
    if (rwioCreate(&in_rwios, in_fname, SK_RWIO_READ)
        || rwioOpen(in_rwios))
    {
        skAppPrintErr("Couldn't open %s for reading.  Exiting.", in_fname);
        appExit(EXIT_FAILURE);
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;                     /* OK */
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    optionsTeardown();

    return;
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    uint8_t i,j;
    uint8_t *parsed_list;
    uint8_t count;
    uint32_t opt_value;

    switch ((appOptionsEnum)opt_index) {

      case RWSEQ_OPT_ID_FIELD:
        parsed_list = skParseNumberList(opt_arg, 1, 12, &count);
        if (NULL == parsed_list) {
            /*
             * Error: failed to successfully parse
             */
            return 1;
        }; /* Number list was successfully parsed */
        if ((0 == count) || (count > MAX_FIELDS)) {
            free(parsed_list);
            return 1;
        }; /* count failed on the number of fields */
        field_count = count;
        for ( i = j = 0 ; i < count ; i++) {
            if (parsed_list[i] > 12) {
                free(parsed_list);
                return 1;
            } else {
                mask_field_id[j] = parsed_list[i];
                j = j + 1;
            };
        };
        free(parsed_list);
        break;

      case RWSEQ_OPT_DELTA_FIELD:
        if (skStringParseUint32(&opt_value, optarg, 0, 255) < 0) {
            return 1;
        } else {
            delta_field_id = (uint8_t) (opt_value & 0x000000FF);
        };
        break;

      case RWSEQ_OPT_SUMMARIZE:
        summarize_flag = 1;
        break;

      case RWSEQ_OPT_DELTA:
        if (skStringParseUint32(&opt_value, optarg, 0, 255) < 0) {
            return 1;
        } else {
            delta = opt_value;
        };
        break;

      case RWSEQ_OPT_THRESH:
        if (skStringParseUint32(&opt_value, optarg, 0, MAX_THRESH) < 0) {
            return 1;
        } else {
            threshold = opt_value;
        };
        break;

      case RWSEQ_OPT_OBJECTIVE:
        objective_flag = 1;
        return 0;
        break;

      case RWSEQ_OPT_INPIPE:
        in_fname = opt_arg;
        break;

      case RWSEQ_OPT_OUTPIPE:
        out_fname = opt_arg;
        break;
    }

    return 0; /* OK */
}


/*
 * cmp_field
 *
 * There's probably somebody gasping in surprise that I didn't write this
 * as a giant macro, but I'm chanting "power law distribution" "power law
 * distribution" to myself here.
 * cmp_field takes an id, a value and a delta and returns 0 if the value is
 * equal to the field +/- delta, it returns 0 if the values are equal (within
 * delta), 1 otherwise.
 */
static int cmp_field(
    int             field_id,
    const rwRec    *current_rec,
    const rwRec    *ref_rec,
    uint32_t        delta)
{
    uint32_t result, ref_value, field_value;
    field_value = 0 ;
    ref_value = 0;
    switch (field_id) {
      case RWSEQ_FID_SIP:
        field_value = current_rec->sIP.ipnum;
        ref_value = ref_rec->sIP.ipnum;
        break;
      case RWSEQ_FID_DIP:
        field_value = current_rec->dIP.ipnum;
        ref_value = ref_rec->dIP.ipnum;
        break;
      case RWSEQ_FID_SPORT:
        field_value = current_rec->sPort;
        ref_value = ref_rec->sPort;
        break;
      case RWSEQ_FID_DPORT:
        field_value = current_rec->dPort;
        ref_value = ref_rec->dPort;
        break;
      case RWSEQ_FID_PROTO:
        field_value = current_rec->proto;
        ref_value = ref_rec->proto;
        break;
      case RWSEQ_FID_PKTS:
        field_value = current_rec->pkts;
        ref_value = ref_rec->pkts;
        break;
      case RWSEQ_FID_BYTES:
        field_value = current_rec->bytes;
        ref_value = ref_rec->bytes;
        break;
      case RWSEQ_FID_FLAGS:
        field_value = current_rec->flags;
        ref_value = ref_rec->flags;
        break;
      case RWSEQ_FID_STIME:
        field_value = current_rec->sTime;
        ref_value = ref_rec->sTime;
        break;
      case RWSEQ_FID_DUR:
        field_value = current_rec->elapsed;
        ref_value = ref_rec->elapsed;
        break;
      case RWSEQ_FID_ETIME:
        field_value = current_rec->sTime + current_rec->elapsed;
        ref_value = ref_rec->sTime + current_rec->elapsed;
        break;
      case RWSEQ_FID_SENSOR:
        field_value = current_rec->sID;
        ref_value = ref_rec->sID;
        break;
    };
    result = field_value > ref_value ?
        field_value - ref_value : ref_value - field_value;
    if (result <= delta) {
        return 0;
    } else {
        return 1;
    };
}


/*
 * int process_recs
 * Called on all records except the first; it compares the
 * new records against the group and generates a group
 * number as necessary.
 *
 */
static int process_rec(rwRec *current_rec)
{
    uint8_t i,increase;
    static rwRec last_rec;
    static uint32_t current_group_count;
    uint32_t old_etime, new_etime;
    increase = 0;
    if (rec_id > 0) {
        for (i = 0 ; i < field_count ; i++) {
            if (cmp_field(mask_field_id[i], current_rec, &last_rec, 0)) {
                increase = 1;
                break;
            };
        };

        if (!increase && delta > 0) {
            if (cmp_field(delta_field_id, current_rec, &last_rec, delta)) {
                increase = 1;
            };
        };
    } else {
        /*
         * Escape clause to manage the first record case, csc is set
         * to 0
         */
      if(summarize_flag) {
        memcpy(&last_rec, current_rec, sizeof(rwRec));
      };
      last_rec.bytes = 0;
      last_rec.pkts = 0;
      last_rec.flags = 0;
      current_group_count = 0;
    };
    group_id += increase;

    /*
     * Reset the current group count if we've moved on to the
     * next value
     */

    /*
     * Backup - store in last rec; if the threshold is being
     * used, then copy to the threshold buffer.
     */
    if(summarize_flag) {
      /*
       * 2005/09/28: New functionality added by MC.  This extends
       * functionality by adding the ability to just output one record
       * if the requirements are met.
       */
      current_rec->nhIP.ipnum = group_id;
      if (increase) {
        last_rec.nhIP.ipnum = group_id - 1;
        if(current_group_count > threshold) {
          rwWrite(out_rwios, &last_rec);
        };
        memcpy(&last_rec, current_rec, sizeof(rwRec));
      } else {
        last_rec.pkts += current_rec->pkts;
        last_rec.bytes += current_rec->bytes;
        last_rec.flags |= current_rec->flags;
        if (last_rec.sTime >= current_rec->sTime) {
          last_rec.sTime = current_rec->sTime;
        };
        old_etime = last_rec.sTime + last_rec.elapsed;
        new_etime = current_rec->sTime + current_rec->elapsed;
        if (last_rec.sTime > current_rec->sTime) {
          last_rec.sTime = current_rec->sTime;
        };
        if (old_etime < new_etime) {
          last_rec.elapsed = new_etime - last_rec.sTime;
        };
      };
    } else {
      if ((increase) ||
          ((!increase) && (!objective_flag))) {
        memcpy(&last_rec, current_rec, sizeof(rwRec));
      };
      /*
       * if the group count exceeds the threshold, print the current record
       */
      current_rec->nhIP.ipnum = group_id;
      if (current_group_count > threshold) {
        rwWrite(out_rwios, current_rec);
      } else {
        /*
         * If it's less, we just copy the record to the threshold buffer
         */
        if (current_group_count < threshold) {
          memcpy(rec_thresh_buffer + (current_group_count - 1),
                 current_rec, sizeof(rwRec));
        } else {
          for (i = 0 ; i < threshold - 1; i++) {
            rwWrite(out_rwios, rec_thresh_buffer + i);
          };
          rwWrite(out_rwios, current_rec);
        };
      };
    };
    current_group_count = increase ? 1 : current_group_count + 1;
    rec_id++;
    return 0;
}

int main(int argc, char **argv)
{
    rwRec input_rec;
    threshold = 0;

    appSetup(argc, argv); /* never returns on error */

    rec_thresh_buffer = (rwRec *) calloc(threshold + 1, sizeof(rwRec));

    while (rwRead(in_rwios, &input_rec)) {
        process_rec(&input_rec);
    };
    rwCloseFile(in_rwios);
    rwCloseFile(out_rwios);

    /* done */
    appExit(0);

    return 0;
}
