/*
** Copyright (C) 2003-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwfileinfo
**
**  Prints information about an rw-file:
**    file type
**    file version
**    byte order
**    compression level
**    header size
**    record size
**    record count
**    file size
**    command line args used to generate file
**
*/


#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwinfo/rwfileinfo.c,v 1.18 2005/10/26 14:23:17 thomasm Exp $");

#include "utils.h"
#include "rwpack.h"

/* LOCAL DEFINES AND TYPEDEFS */

/* where to write --help output */
#define USAGE_FH stdout

/* This is used to store information about a files */
struct rwinfoFileInfo {
    int compression;          /* our compression level (not gzip) */
    int version;              /* file version */
    int type;                 /* file type, (RWSPLIT, etc) as integer */
    off_t fileSize;           /* (uncompressed) file size */
    size_t recordLength;      /* size of one record */
    size_t headerLength;      /* size required for all headers */
    off_t countRecords;       /* count of records */
    const char *byteOrder;    /* byte order */
    const char *path;         /* filesystem path to file */
    const char *typeString;   /* file type as string "RWSPLIT" */
    const char *commandLines; /* command lines */
};

/* A list of the properties that can be printed in the order that they
 * will be printed. */
enum rwinfoPropertyEnum {
    RWINFO_TYPE,
    RWINFO_VERSION,
    RWINFO_BYTE_ORDER,
    RWINFO_COMPRESSION,
    RWINFO_HEADER_LENGTH,
    RWINFO_RECORD_LENGTH,
    RWINFO_COUNT_RECORDS,
    RWINFO_FILE_SIZE,
    RWINFO_COMMAND_LINES,
    /* Last item is used to get a count of the above; it must be last */
    RWINFO_PROPERTY_COUNT
};

/* Whether each property will be printed, and the label to use */
struct rwinfoProperties {
    int userFields;                            /* use gave fields? */
    int8_t willPrint[RWINFO_PROPERTY_COUNT];   /* nonzero if printed */
    const char *label[RWINFO_PROPERTY_COUNT];  /* property's label */
};


/* LOCAL FUNCTIONS */

static void appUsage(void);
static void appUsageLong(void);
static void appTeardown(void);
static void appSetup(int argc, char **argv);
static void appOptionsUsage(FILE *fh);
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static int  getFileInfo(struct rwinfoFileInfo *info);
static void printFileInfo(struct rwinfoFileInfo *info);

/* LOCAL VARIABLES */

static struct rwinfoProperties rwinfoProp = {
    0,
    {0, 0, 0, 0, 0, 0, 0, 0, 0},
    {
        "type",
        "version",
        "byte-order",
        "compression",
        "header-length",
        "record-length",
        "count-records",
        "file-size",
        "command-lines"
    }
};

static int print_summary = 0;

/* OPTIONS SETUP */

/* Create constants for the option processor */
enum rwinfoOptionIds {
    RWINFO_OPT_FIELDS,
    RWINFO_OPT_SUMMARY,
    /* Last item is used to get a count of the above; it must be last */
    RWINFO_OPTION_COUNT
};

static struct {
    /* a count of the options; from the enum above */
    int           count;
    /* index into argv of first non-option argument */
    int           nextIndex;
    /* the list of options; add one for sentinel */
    struct option options[1+RWINFO_OPTION_COUNT];
    /* their help strings; add one for sentinel */
    const char    *help[1+RWINFO_OPTION_COUNT];
} rwinfoOptions = {
    RWINFO_OPTION_COUNT,
    1,
    {
        {"fields",  REQUIRED_ARG, 0, RWINFO_OPT_FIELDS},
        {"summary", NO_ARG,       0, RWINFO_OPT_SUMMARY},
        {0,0,0,0} /* sentinal entry */
    },
    {
        /* add help strings here for the applications options */
        "list of fields to print.  Default ALL",
        "print a summary of total files, file sizes, and records",
        (char *)NULL /* sentinal entry */
    }
};


/*
 * appUsage:
 * 	print short usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 * 	print complete usage information to USAGE_FH and exit.
 * 	passed to optionsSetup() to print usage when --help option
 * 	given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
    int i;

#define USAGE_MSG                                                   \
    ("%s [field-values] <file1> [<file2> .. <fileN>]\n"             \
     "\tPrints info (number records, etc) about an rw-file\n")

    fprintf(USAGE_FH, USAGE_MSG, skAppName());

    appOptionsUsage(USAGE_FH);

    fprintf(USAGE_FH, "Field values:\n");
    for (i = 0; i < RWINFO_PROPERTY_COUNT; ++i) {
        fprintf(USAGE_FH, "  %3d %s\n", 1+i, rwinfoProp.label[i]);
    }

    exit(EXIT_SUCCESS);
}


/*
 * appTeardown:
 *	teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 * 	All modules are torn down. Then application teardown takes place.
 * 	Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    optionsTeardown();

    return;
}


/*
 * appSetup
 *	do all the setup for this application include setting up required
 *	modules etc.
 * Arguments:
 *	argc, argv
 * Returns: None.
 * Side Effects:
 *	exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int i;

    skAppRegister(argv[0]);

    /* Setup the options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);
    }

    /* Register the options */
    if (optionsRegister(rwinfoOptions.options, (optHandler)appOptionsHandler,
                        (clientData) &rwinfoProp))
    {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);
    }

    /* parse options */
    if ( (rwinfoOptions.nextIndex = optionsParse(argc, argv)) < 0) {
        appUsage();  /* never returns */
    }

    /* need at least one file */
    if (rwinfoOptions.nextIndex >= argc) {
        skAppPrintErr("expecting file name");
        appUsage();  /* never returns */
    }

    /* if user selected no fields to print, print all */
    if (!rwinfoProp.userFields) {
        for (i = 0; i < RWINFO_PROPERTY_COUNT; ++i) {
            rwinfoProp.willPrint[i] = -1;
        }
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;  /* OK */
}


/*
 * appOptionsUsage:
 * 	print options for this app to the named file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *	do NOT add non-option usage here (i.e., required and other
 *	optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\napp options\n");
    for (i = 0; i < RWINFO_OPTION_COUNT; i++ ) {
        fprintf(fh, "--%s %s. %s\n", rwinfoOptions.options[i].name,
                (rwinfoOptions.options[i].has_arg
                 ? (rwinfoOptions.options[i].has_arg == 1
                    ? "Required Arg"
                    : "Optional Arg")
                 : "No Arg"), rwinfoOptions.help[i]);
    }

    return;
}


/*
 * appOptionsHandler:
 * 	called for each option the app has registered.
 * Arguments:
 *	clientData cData: ignored here.
 *	int index: index into appOptions of the specific option
 *	char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *	0 if OK. 1 else.
 * Side Effects:
 *	Relevant options are set.
 */
static int appOptionsHandler(clientData cData, int opt_index, char *opt_arg)
{
    uint32_t *numberList;
    uint32_t numberCount;
    uint32_t i;

    switch(opt_index) {
      case RWINFO_OPT_FIELDS:
        if (skStringParseNumberList(&numberList, &numberCount, opt_arg,
                                    1, RWINFO_PROPERTY_COUNT, 0))
        {
            return 1;
        }
        ((struct rwinfoProperties*)cData)->userFields = 1;
        for (i = 0; i < numberCount; ++i) {
            ((struct rwinfoProperties*)cData)->willPrint[numberList[i]-1] = 1;
        }
        free(numberList);
        break;

      case RWINFO_OPT_SUMMARY:
        print_summary = 1;
        break;

      default:
        skAppPrintErr("Option '%d' not handled in switch() at %s:%d",
                      opt_index, __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }

    return 0;
}


/*
 * void getHeaderLength(rwIOStruct *rwIOS, struct rwinfoFileInfo *info);
 *
 * Determine the header length of the file and store this information
 * in the rwinfoFileInfo pointer.  If the file's has command line
 * information, obtain that as well.
 */
static void getHeaderLength(
    rwIOStruct             *rwIOS,
    struct rwinfoFileInfo  *info)
{
    div_t ratio;
    char *commandBuffer;
    char *bp;
    char *cp;
    uint32_t i;
    uint16_t bytecount;
    filterHeaderV1 *fhp;
  

    if ( 0 == rwIsCompressed(rwIOS) ) {
        /* file is not compressed; rwOpenFile() leaves us at end of header */
        info->headerLength = ftell(rwGetFILE(rwIOS));
        if (info->headerLength == (size_t)-1) {
            perror("tell failed");
        }
    } else {
        /* file is compressed.  get header length from rwIOS */
        if (FT_RWFILTER == rwGetFileType(rwIOS)) {
            info->headerLength =
                ((filterHeaderV1*)(rwIOS->hdr))->totFilterLen;
        } else {
            info->headerLength = rwIOS->hdrLen;
        }

        if ((2 == info->version) && (info->recordLength > 0)) {
            /* for version 2, the headerLength is always a multiple of the
             * recordLength */
            ratio = div(info->headerLength, info->recordLength);
            if (0 > ratio.rem) {
                /* headerLength is not exact multiple of recordLength */
                info->headerLength = info->recordLength * (1 + ratio.quot);
            }
        }
    }

    if (FT_RWFILTER == rwGetFileType(rwIOS)) {
        /* we know the length of commands will not be larger than the
         * entire header, so malloc that much space */
        commandBuffer = calloc(info->headerLength, 1);
        bp = commandBuffer;
        /* local copy of the header */
        fhp = ((filterHeaderV1*)(rwIOS->hdr));
        /* loop over each command */
        for (i = 0; i < fhp->filterCount; ++i) {
            /* get length and point cp at the first character in command */
            bytecount = fhp->fiArray[i].byteCount;
            cp = fhp->fiArray[i].info;
            while(bytecount--) {
                /* each command "phrase" is separated by a \0; change
                 * it to a space */
                if (*cp) {
                    *bp = *cp;
                } else {
                    *bp = ' ';
                }
                ++cp;
                ++bp;
            }
            /* finish this command and move forward to start the next
             * command */
            *bp = '\0';
            ++bp;
        }
        info->commandLines = commandBuffer;
    }
}


/*
 * void getNumberRecs(rwIOStruct *rwIOS, struct rwinfoFileInfo *info);
 *
 * Given rwIOS to the opened file, read the file to determine the
 * file's size and the number of records in the file.  Add this
 * information to the rwinfoFileInfo pointer.
 */
static void getNumberRecs(
    rwIOStruct             *rwIOS,
    struct rwinfoFileInfo  *info)
{
    off_t rc = 0;	/* record count */
    off_t fSize;

    if (rwIsCompressed(rwIOS)) {
        while (0 == rwSkip(rwIOS, 1)) {
            ++rc;
        }
        info->fileSize = info->headerLength + (info->recordLength * rc);
    } else {
        info->fileSize = fileSize(rwGetFileName(rwIOS));
        fSize = info->fileSize - (long)info->headerLength;
        if (info->recordLength > 0) {
            if ((fSize % info->recordLength) != 0) {
                fprintf(stderr,
                        ("Incomplete file."
                         " Size=%lld. HL=%lu. RL=%lu. %lld excess bytes\n"),
                        (long long)(info->headerLength+fSize),
                        (unsigned long)info->headerLength,
                        (unsigned long)info->recordLength,
                        (long long)(fSize%(info->recordLength)));
            }
            rc = fSize / info->recordLength;
        }
    }
    info->countRecords = rc;

    return;
}


/*
 *  getFileInfo(struct rwinfoFileInfo info);
 *
 *  Print the file information for the file at path.  Return 0 on
 *  success, or 1 on failure.
 */
static int getFileInfo(struct rwinfoFileInfo *info)
{
    rwIOStruct *rwIOS = NULL;
    int rv = LIBRW_OK;

    rv = rwioCreate(&rwIOS, info->path, SK_RWIO_READ);
    if (rv != LIBRW_OK) {
        rwioPrintLastErr(rwIOS, rv, &skAppPrintErr);
        goto END;
    }

    rv = rwioOpen(rwIOS);
    if ((rv != LIBRW_OK) && (rv != LIBRW_ERR_UNSUPPORT_TYPE)) {
        rwioPrintLastErr(rwIOS, rv, &skAppPrintErr);
        goto END;
    }

    /* Get the info in the generic header. */
    info->type = rwGetFileType(rwIOS);
    info->version = rwGetFileVersion(rwIOS);
    info->compression = ((genericHeader*)rwIOS->hdr)->cLevel;
    if (rwGetIsBigEndian(rwIOS)) {
        info->byteOrder = "BigEndian";
    } else {
        info->byteOrder = "littleEndian";
    }
    info->typeString = rwGetFileOutputFormatName(info->type);


    if (rv == LIBRW_ERR_UNSUPPORT_TYPE) {
        /* The type is not supported by librw, so we can't get any
         * information outside of that in the generic header. */
        rv = LIBRW_OK;
        goto END;
    }

    info->recordLength = rwIOS->recLen;
    getHeaderLength(rwIOS, info);
    getNumberRecs(rwIOS, info);

  END:
    if (rwIOS) {
        rwioDestroy(&rwIOS);
    }
    return (!!rv);
}


/*
 * printFileInfo(struct rwinfoFileInfo *info)
 *
 * Print information (as requested by the user in the rwinfoProp
 * structure) in 'info' to stdout.
 */
static void printFileInfo(struct rwinfoFileInfo *info)
{
    const char *indentString = "  ";
    const int labelWidth = -16;
    const char *comm;
    uint32_t count;

    printf("%s:\n", info->path);
    if (rwinfoProp.willPrint[RWINFO_TYPE]) {
        printf("%s%*s%s\n", indentString, labelWidth,
               rwinfoProp.label[RWINFO_TYPE], info->typeString);
    }
    if (rwinfoProp.willPrint[RWINFO_VERSION]) {
        printf("%s%*s%d\n", indentString, labelWidth,
               rwinfoProp.label[RWINFO_VERSION], info->version);
    }
    if (rwinfoProp.willPrint[RWINFO_BYTE_ORDER]) {
        printf("%s%*s%s\n", indentString, labelWidth,
               rwinfoProp.label[RWINFO_BYTE_ORDER], info->byteOrder);
    }
    if (rwinfoProp.willPrint[RWINFO_COMPRESSION]) {
        printf("%s%*s%d\n", indentString, labelWidth,
               rwinfoProp.label[RWINFO_COMPRESSION], info->compression);
    }
    if (rwinfoProp.willPrint[RWINFO_HEADER_LENGTH]) {
        printf("%s%*s%lu\n", indentString, labelWidth,
               rwinfoProp.label[RWINFO_HEADER_LENGTH],
               (unsigned long)info->headerLength);
    }
    if (rwinfoProp.willPrint[RWINFO_RECORD_LENGTH]) {
        printf("%s%*s%lu\n", indentString, labelWidth,
               rwinfoProp.label[RWINFO_RECORD_LENGTH],
               (unsigned long)info->recordLength);
    }
    if (rwinfoProp.willPrint[RWINFO_COUNT_RECORDS]) {
        printf("%s%*s%lld\n", indentString, labelWidth,
               rwinfoProp.label[RWINFO_COUNT_RECORDS],
               (long long)info->countRecords);
    }
    if (rwinfoProp.willPrint[RWINFO_FILE_SIZE]) {
        printf("%s%*s%lld\n", indentString, labelWidth,
               rwinfoProp.label[RWINFO_FILE_SIZE], (long long)info->fileSize);
    }
    if (rwinfoProp.willPrint[RWINFO_COMMAND_LINES] && info->commandLines) {
        printf("%s%*s\n", indentString, labelWidth,
               rwinfoProp.label[RWINFO_COMMAND_LINES]);
        comm = info->commandLines;
        count = 0;
        while (*comm) {
            ++count;
            printf("%*d%s%s\n", (-1*labelWidth), count, indentString, comm);
            comm += 1+strlen(comm);
        }
    }
}


/*
 *  For each file, get the file's info then print it
 */
int main(int argc, char **argv)
{
    int i;
    struct rwinfoFileInfo info;
    int64_t total_files = 0;
    int64_t total_bytes = 0;
    int64_t total_recs = 0;

    appSetup(argc, argv);			/* never returns on error */

    for (i = rwinfoOptions.nextIndex; i < argc; ++i) {
        memset(&info, 0, sizeof(struct rwinfoFileInfo));
        info.path = argv[i];
        if (0 == getFileInfo(&info)) {
            printFileInfo(&info);
            ++total_files;
            total_bytes += info.fileSize;
            total_recs += info.countRecords;
        } else {
            skAppPrintErr("perhaps '%s' is not a SiLK file?",
                          argv[i]);
        }
        if (info.commandLines) {
            free((char*)info.commandLines);
        }
    }

    if (print_summary) {
        printf("**SUMMARY**:\n  %-16s%lld\n",
               "number-files", (long long)total_files);
        if (rwinfoProp.willPrint[RWINFO_COUNT_RECORDS]) {
            printf("  %-16s%lld\n",
                   "total-records",
                   (long long)total_recs);
        }
        if (rwinfoProp.willPrint[RWINFO_FILE_SIZE]) {
            printf("  %-16s%lld\n", "all-file-sizes",
                   (long long)total_bytes);
        }
    }
    /* done */
    appTeardown();

    return 0;
}
