/*
** Copyright (C) 2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
** rwmatch
**
** rwmatch.c is a utility which takes two files (now called in and
** out) and produces a new file where the contents are linked together
** in the form (saddress, values) on one side and (daddress, values)
** on the other side.  apart from the saddress/daddress relationship,
** any other mating rules are specified by the command line.
**
** This is necessary because there are several operations where mating
** is not direct reversal; the particular case which is driving this
** applicatin is traceroute where the only rational is close
** relationship in time between s and d address, and one of them is a
** UDP signal with an ICMP response
**
** command arguments:
**
** --relate:    takes fields as pairs for relationship.  The first
**              value is the field in file 1, the second value is the
**              field in file ** 2.
**
**              for example, the default could be --relate=1,2
**
**              multiple --relate predicates are accepted.
**              Traditional tcp/ip matching would be:
**              --relate=1,2 --relate=2,1 --relate=3,4 --relate=4,3
**
** --delta:     the number of seconds difference we'll allow between
**              two records before considering them similar.  Defaults
**              to 2.
**
** --unmated:   output path for unmatched records; defaults to
**              /dev/null.  A record must have a response before it is
**              ejected.  So we're always sending out at least two
**              records.
**
** Apart from that; rwmate will annotate the nhip field with an index
** indicating the 'mating flow' that a record belongs to.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwmatch/rwmatch.c,v 1.3 2005/10/26 14:23:17 thomasm Exp $");

#include "utils.h"
#include "rwpack.h"


/* TYPEDEFS AND DEFINES */

/* Where to send --help output */
#define USAGE_FH stdout

#define MAX_FIELDS 8

#define RWMATE_REL_MAX 8
#define RWMATE_DELTA_MAX 3600
#define RWMATE_INCOMING 0
#define RWMATE_OUTGOING 1

typedef enum _fieldsEnum {
    RWMATE_FID_SIP = 1, RWMATE_FID_DIP, RWMATE_FID_SPORT, RWMATE_FID_DPORT,
    RWMATE_FID_PROTO, RWMATE_FID_PKTS, RWMATE_FID_BYTES, RWMATE_FID_FLAGS,
    RWMATE_FID_STIME, RWMATE_FID_ETIME
} fields_enum;

typedef struct _appState {
    char *in_fname;
    char *out_fname;
    char *mtd_fname;
    rwIOStruct *in_ptr, *out_ptr, *mtd_ptr;
} rwm_app_state_t;


/* LOCAL FUNCTIONS */

static void appExit(int status);
static void appUsage(void);                     /* never returns */
static void appUsageLong(void);                 /* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);    /* never returns */
static void appOptionsUsage(FILE *fh);          /* application options usage */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static uint32_t rwm_get_field(fields_enum field_id, rwRec *current_rec);
static int rwm_compare_records(rwRec *incoming_rec, rwRec *outgoing_rec);


/* LOCAL VARIABLES */

static uint8_t rwm_total_pairs;
static uint8_t rwm_mate_pairs[RWMATE_REL_MAX][2];
static uint8_t verbose_flag = 0;
static uint32_t rwm_delta;
static rwm_app_state_t app_state;


/* OPTIONS SETUP */

typedef enum _appOptionsEnum {
    RWMATE_OPT_RELATE, RWMATE_OPT_DELTA, RWMATE_OPT_UNMATED,
    RWMATE_OPT_VERBOSE
} appOptionsEnum;

static struct option appOptions[] = {
    {"relate",          REQUIRED_ARG, 0, RWMATE_OPT_RELATE},
    {"time-delta",      REQUIRED_ARG, 0, RWMATE_OPT_DELTA},
    {0,0,0,0}           /* sentinel entry */
};

static char *appHelp[] = {
    ("Pair of fields to relate when mating data,\n"
     "  given in the form of <query field id>,<response field id>"),
    "Maximum amount of time difference between two records for relation",
    (char *)NULL
};


/*
 *  appExit(status);
 *
 *    Exit the appliction with status.
 *
 */
static void appExit(int status)
{
    appTeardown();
    exit(status);
}


/*
 * appUsage:
 *      print short usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 *      print complete usage information to USAGE_FH and exit.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                               \
    ("%s: [--relate=RELATE-FIELD] [--time-delta=DELTA]"         \
     " query_file response_file output_path\n"                  \
     "\tGroups records as queries and responses\n")

    fprintf(USAGE_FH, USAGE_MSG, skAppName());
    appOptionsUsage(USAGE_FH);
    exit(EXIT_SUCCESS);
}


/*
 * appOptionsUsage:
 *      print options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int nextOptIndex;

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *))
           == (sizeof(appOptions)/sizeof(struct option)));

    /* initialize globals */
    skAppRegister(argv[0]);
    memset(&app_state, '\0', sizeof(rwm_app_state_t));

    /* initialize options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);                        /* never returns */
    }

    /* register the options and handler */
    if (optionsRegister(appOptions, (optHandler) appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);                        /* never returns */
    }

    /* parse options */
    nextOptIndex = optionsParse(argc, argv);
    if (nextOptIndex < 0) {
        appUsage();               /* never returns */
    }

    /* check for extra options */
    if (nextOptIndex > argc) {
        skAppPrintErr("unexpected arguments; input must come from stdin.");
        appUsage(); /* never returns */
    }

    /* check for too few options */
    if ((argc - nextOptIndex) < 3) {
        skAppPrintErr("Insufficient files specified");
        appUsage(); /* never returns */
    };

    /* get and open the query file (first input) */
    app_state.in_fname = argv[nextOptIndex];
    if ( !fileExists(app_state.in_fname)) {
        skAppPrintErr(("no query flow file named %s, "
                       "check the file and try again"),
                      app_state.in_fname);
        exit(EXIT_FAILURE);
    }

    app_state.in_ptr = rwOpenFile(app_state.in_fname, NULL);
    if (NULL == app_state.in_ptr) {
        skAppPrintErr("Trouble opening the query file.  Exiting");
        exit(EXIT_FAILURE);
    }

    /* get and open the response file (second input) */
    app_state.out_fname = argv[nextOptIndex + 1];
    if ( !fileExists(app_state.out_fname)) {
        skAppPrintErr(("no response flow file named %s, "
                       "check the file and try again"),
                      app_state.out_fname);
        exit(EXIT_FAILURE);
    }

    app_state.out_ptr = rwOpenFile(app_state.out_fname, NULL);
    if (NULL == app_state.out_ptr) {
        skAppPrintErr("Trouble opening the response file.  Exiting");
        exit(EXIT_FAILURE);
    }

    /* get and open the result file (output) */
    app_state.mtd_fname = argv[nextOptIndex + 2];
    if (fileExists(app_state.mtd_fname)) {
        skAppPrintErr("Output file %s already exists; delete it",
                      app_state.mtd_fname);
        exit(EXIT_FAILURE);
    }
            
    if (rwioCreate(&(app_state.mtd_ptr), app_state.mtd_fname, SK_RWIO_WRITE)
        || rwioOpen(app_state.mtd_ptr))
    {
        skAppPrintErr("Unable to open output file '%s'",
                      app_state.mtd_fname);
        exit(EXIT_FAILURE);
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;                     /* OK */
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    optionsTeardown();

    return;
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    uint32_t opt_value;
    char *index_pos;

    switch (opt_index) {

      case RWMATE_OPT_RELATE:
        if (rwm_total_pairs < RWMATE_REL_MAX) {
            index_pos = index(opt_arg, ',');
            if (index_pos == NULL) {
                skAppPrintErr("Error: Unparseable related string %s", opt_arg);
                return 1;
            } else {
                /*
                 * Store the first value before the comma as the initial field,
                 * the second value as the final field
                 */
                if (0 == (opt_value = strtoul(opt_arg, (char **) NULL, 10))) {
                    skAppPrintErr("Error parsing relation: %s", opt_arg);
                    return 1;
                };
                rwm_mate_pairs[rwm_total_pairs][0] = opt_value;
                opt_value = strtoul(index_pos + 1, (char **) NULL, 10);
                rwm_mate_pairs[rwm_total_pairs][1] = opt_value;
                rwm_total_pairs++;
            };
        } else {
            skAppPrintErr(("Error: Number of relations specified"
                           " exceeds maximum (%d).  Terminating"),
                          RWMATE_REL_MAX);
            return 1;
        };
        break;

      case RWMATE_OPT_VERBOSE:
        verbose_flag = 1;
        break;

      case RWMATE_OPT_DELTA:
        if (0 == (opt_value = strtoul(opt_arg, (char **) NULL, 10))) {
            skAppPrintErr("Error parsing delta value: %s", opt_arg);
            return 1;
        } else {
            rwm_delta = opt_value;
        };
        break;

      case RWMATE_OPT_UNMATED:
        skAppPrintErr("Unmated path not yet supported");
        break;

      default:
        skAppPrintErr("Unhandled option %d in options handler at %s:%d",
                      opt_index, __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }

    return 0; /* OK */
}

/*
 * uint32_t rwm_get_field(fields_enum current_field, rwRec *current_rec)
 *
 * Given a record and a field ID returns the value of the field for
 * that record.
 */
static uint32_t rwm_get_field(fields_enum field_id, rwRec *current_rec)
{
    uint32_t result = 0;

    switch (field_id) {
      case RWMATE_FID_SIP:
        result = current_rec->sIP.ipnum;
        break;
      case RWMATE_FID_DIP:
        result = current_rec->dIP.ipnum;
        break;
      case RWMATE_FID_SPORT:
        result = current_rec->sPort;
        break;
      case RWMATE_FID_DPORT:
        result = current_rec->dPort;
        break;
      case RWMATE_FID_PROTO:
        result = current_rec->proto;
        break;
      case RWMATE_FID_PKTS:
        result = current_rec->pkts;
        break;
      case RWMATE_FID_BYTES:
        result = current_rec->bytes;
        break;
      case RWMATE_FID_FLAGS:
        result = current_rec->flags;
        break;
      case RWMATE_FID_STIME:
        result = current_rec->sTime;
        break;
      case RWMATE_FID_ETIME:
        result = current_rec->sTime + current_rec->elapsed;
        break;
    };
    return result;
}

/*
 * int rwm_compare_records(rwRec *propose, rwRec *response)
 *
 * this uses the same hash information from rwm_hash to compare an
 * incoming and outgoing record in order to determine whether or not.
 * Note the global variable dependence.
 *
 * Returns 1 if the incoming > outing, -1 if outgoing > incoming, 0 if
 * equal
 */
static int rwm_compare_records(rwRec *incoming_rec, rwRec *outgoing_rec)
{
    uint8_t i;
    uint32_t ival, oval;

    for (i = 0 ; i < rwm_total_pairs; i++) {
        ival = rwm_get_field(rwm_mate_pairs[i][RWMATE_INCOMING], incoming_rec);
        oval = rwm_get_field(rwm_mate_pairs[i][RWMATE_OUTGOING], outgoing_rec);
        if (ival > oval) {
            return 1;
        } else if (ival < oval) {
            return -1;
        };
    };

    /*
     * We successfully matched all match fields that we care about.  Now we
     * compare timing.
     * The response should be within <delta> seconds of the end of the
     * message.
     */
    if (rwm_get_field(RWMATE_FID_STIME,incoming_rec) >
        rwm_get_field(RWMATE_FID_STIME, outgoing_rec))
    {
        return 1;
    };

    /*
     * We know outgoing.etime > outging.stime, and we know by now that
     * incoming.stime >= outgoing.stime; so we just have to compare etime.
     */
    if ((rwm_get_field(RWMATE_FID_ETIME,incoming_rec) + rwm_delta) <
        rwm_get_field(RWMATE_FID_ETIME, outgoing_rec))
    {
        return -1;
    };

    /*
     * we've passed all tests, so the two must be related.
     */
    return 0;
}

int main(int argc, char **argv)
{
    uint16_t threshold;
    uint32_t i;
    uint8_t printed_query;
    rwRec query_rec, response_rec;

    appSetup(argc, argv); /* never returns on error */

    /* initialize variables */
    threshold = 0;
    rwm_total_pairs = 0;
    i = 0;
    verbose_flag = 1;
    printed_query = 0;

    /*
     * The revised version of this application requires
     * sorted data that matches the mating.  So this means
     * that we're going to have records sorted slightly
     * differently (if we match 2/1, we sort --field=2,9 and
     * --field=1,9) on the two applications.
     *
     * The loop always begins with us
     * with a reference input record and an output record.
     * as long as the output record < the input record, we continue
     * to loop.
     */
    rwRead(app_state.out_ptr, &response_rec);
    while (rwRead(app_state.in_ptr, &query_rec)) {
        while (rwm_compare_records(&query_rec, &response_rec) < 0) {
            if (rwRead(app_state.in_ptr, &query_rec)) {
                break;
            };
        };
        printed_query = 0;
        while (rwm_compare_records(&query_rec, &response_rec) == 0) {
            if (!printed_query) {
                i++;
                query_rec.nhIP.ipnum = i;
                rwWrite(app_state.mtd_ptr, &query_rec);
                printed_query = 1;
            };
            response_rec.nhIP.ipnum = 0xFF000000 | i;
            rwWrite(app_state.mtd_ptr, &response_rec);
            rwRead(app_state.out_ptr, &response_rec);
        };
        while (rwm_compare_records(&query_rec, &response_rec) > 0) {
            rwRead(app_state.out_ptr, &response_rec);
        };
    };

    rwCloseFile(app_state.mtd_ptr);
    /* done */
    appExit(0);

    return 0;
}
