/*
** Copyright (C) 2003-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  A "plug-in" (currently compile-time only) for rwflowpack to
**  process NetFlow v5 PDUs from the network or from a file.
**
*/


#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwpack/pdureader.c,v 1.17 2005/10/04 17:41:08 thomasm Exp $");

#include "utils.h"
#include "rwpack.h"
#include "log.h"
#include "pdusource.h"
#include "rwflowpack.h"
#include "probeconf.h"
#include "rwpack-new.h"


/* MACROS and DATA TYPES */

/* Number of PDUs, used to create buffer for reading */
#define PDU_BUFFER_SIZE 60000

/* We need big socket buffers.  Let's try 8MB */
#define BIG_SOCKET_BUF  (1024 * 1024 * 8)


/* PRIVATE FUNCTIONS */

static int   pduReaderSetup(
    flow_reader_t      *flow_reader,
    uint8_t            *termination_flag,
    fp_daemon_mode_t   *is_daemon);
static int  pduReaderStartNetworkListener(void);
static int  pduReaderStartFileReader(void);
static int  pduReaderStop(void);
static void pduReaderCleanup(void);
static int  pduReaderOptionsValidate(probe_def_t *probe);
static void pduReaderOptionsUsage(FILE *fh);
static int  pduReaderOptionsHandler(clientData, int, char *);
static fp_get_record_result_t pduReaderGetRecordFile(rwRec *);
static fp_get_record_result_t pduReaderGetRecordNetwork(rwRec *);


/* PRIVATE VARIABLES */

/* Keep processing as long as this variable is non-zero */
static uint8_t *reading;

/* Struct to hold arguments to the options */
static struct {
    const char *netflow_port;
    const char *file;
} option_args = {
    NULL, NULL
};


/* The following variables are global for each PDU-reader; if we move
 * to having a single rwflowpack accept PDUs from multiple Netflow
 * sources, these will need to become thread-level globals. */

/* The source of PDUs. */
static pduSource_t pdu_src;

/* The probe the pdu-reader reads from. */
static const probe_def_t *t_probe = NULL;


/* OPTION SETUP */

/* Identifiers for each option */
typedef enum pduReaderOptEnum {
    FP_NETFLOW_PORT, FP_FILE
} pduReaderOptEnum_t;

/* names of options */
static struct option pduReaderOptions[] = {
    {"netflow-port", REQUIRED_ARG, 0, FP_NETFLOW_PORT},
    {"netflow-file", REQUIRED_ARG, 0, FP_FILE},
    {0,0,0,0}  /* sentinel entry */
};

/* option help strings */
static char *pduReaderHelp[] = {
    "port on which to accept netflow packets",
    "file from which to read netflow records",
    (char *)NULL  /* sentinel */
};



/*
 *  pduReaderCleanup();
 *
 *    Destroy the PDU reader.
 *
 *    Invoked by flow_reader->teardown_fn();
 */
static void pduReaderCleanup(void)
{
    if (pdu_src) {
        pduSourceDestroy(pdu_src);
        pdu_src = NULL;
    }
}


/*
 *  pduReaderStartNetworkListener()
 *
 *    Create a flowsource object that will read PDU records from a
 *    Berkeley socket.
 *
 *    One of two possible functions invoked by flow_reader->start_fn();
 */
static int pduReaderStartNetworkListener(void)
{
    in_addr_t bind_addr;
    uint16_t  bind_port;
    in_addr_t accept_from;

    /* if a pdu_src already exists, just return. */
    if (pdu_src != NULL) {
        return 0;
    }

    if (0 != skProbeGetListenAsHost(&bind_addr, &bind_port, t_probe)) {
        skAppPrintErr("Probe not configured for listening to network");
        logMsg("Probe not configured for listening to network");
        return -1;
    }

    (void)skProbeGetAcceptFromHost(&accept_from, t_probe);

    pdu_src = pduSourceCreate(bind_port, accept_from, bind_addr,
                              PDU_BUFFER_SIZE, BIG_SOCKET_BUF, &logMsg);
    if (pdu_src == NULL) {
        skAppPrintErr("Could not create PDU listener on %s:%d",
                      num2dot(bind_addr), bind_port);
        logMsg("Could not create PDU listener on %s:%d",
               num2dot(bind_addr), bind_port);
        return -1;
    }

    return 0;
}


/*
 *  pduReaderStartFileReader()
 *
 *    Create a flowsource object that will read PDU records from a
 *    file.
 *
 *    One of two possible functions invoked by flow_reader->start_fn();
 */
static int pduReaderStartFileReader(void)
{
    const char *filename;

    /* if a pdu_src already exists, just return. */
    if (pdu_src != NULL) {
        return 0;
    }

    /* Use the option_args.file argument over the FileSource in the
    probe.  This is to support rwdaemon.  It would be nice to do this
    in a less hackish way.  In the future, FIXME. */
    filename = option_args.file ? option_args.file :
               skProbeGetFileSource(t_probe);
    if (filename == NULL) {
        skAppPrintErr("Probe not configured for reading from file");
        logMsg("Probe not configured for reading from file");
        return -1;
    }

    pdu_src = pduFileSourceCreate(filename, &logMsg);
    if (pdu_src == NULL) {
        skAppPrintErr("Could not create PDU source from file '%s'",
                      filename);
        logMsg("Could not create PDU source from file '%s'",
               filename);
        return -1;
    }

    return 0;
}


/*
 *  pduReaderStop();
 *
 *    Stop reading records.
 *
 *    Invoked by flow_reader->stop_fn();
 */
static int pduReaderStop(void)
{
    pduReaderCleanup();
    return 0;
}


/*
 *  pduReaderGetRecordNetwork(&rwrec);
 *
 *    Fill 'rwrec' with an rw generic record from the underlying
 *    flowsource object connected to a socket.
 *
 *    One of two possible functions invoked by flow_reader->get_record_fn();
 */
static fp_get_record_result_t pduReaderGetRecordNetwork(rwRec *rwrec)
{
    int rv;

    do {
        /* get a record */
        rv = pduSourceGetGeneric(pdu_src, rwrec, SILK_ENDIAN_NATIVE);
        if (rv == -1) {
            return FP_GET_ERROR;
        }

        /* set it's sensor and flow_type */
        rv = skProbeDetermineFlowtype(t_probe, rwrec);
        assert(rv >= -1);
        assert(rv < NUM_FLOW_TYPES);
        if (rv == -1) {
            logMsg("Cannot determine flow_type probe %s, input %d, output %d",
                   skProbeGetName(t_probe), rwrec->input, rwrec->output);
        }
    } while (rv == -1);

    return FP_BREAK_POINT;
}


/*
 *  pduReaderGetRecordFile(&rwrec);
 *
 *    Fill 'rwrec' with an rw generic record from the underlying
 *    flowsource object connected to a file.
 *
 *    One of two possible functions invoked by flow_reader->get_record_fn();
 */
static fp_get_record_result_t pduReaderGetRecordFile(rwRec *rwrec)
{
    int rv;

    do {
        /* get a record */
        rv = pduSourceGetGeneric(pdu_src, rwrec, SILK_ENDIAN_NATIVE);
        if (rv == -1) {
            return FP_END_STREAM;
        }

        /* set it's sensor and flow_type */
        rv = skProbeDetermineFlowtype(t_probe, rwrec);
        assert(rv >= -1);
        assert(rv < NUM_FLOW_TYPES);
        if (rv == -1) {
            logMsg("Cannot determine flow_type probe %s, input %d, output %d",
                   skProbeGetName(t_probe), rwrec->input, rwrec->output);
        }
    } while (rv == -1);

    return FP_RECORD;
}


/*
 * pduReaderInitialize:
 *
 *      Set initial state of this flow reader.  Register its options.
 *
 *    Invoked by flow_reader->initialize_fn();
 */
int pduReaderInitialize(flow_reader_t *flow_reader)
{
    /* Set my name */
    flow_reader->reader_name = "PDU Reader";

    /* Initialize variables */
    reading = NULL;
    pdu_src = NULL;

    /* Set function pointers */
    flow_reader->teardown_fn         = &pduReaderCleanup;
    flow_reader->options_usage_fn    = &pduReaderOptionsUsage;
    flow_reader->options_validate_fn = &pduReaderOptionsValidate;
    flow_reader->setup_fn            = &pduReaderSetup;
    flow_reader->stop_fn             = &pduReaderStop;
    flow_reader->filename_fn         = &skPackBuildStorageFileName;
    /* flow_reader->start_fn is set below in setup. */
    /* flow_reader->get_record_fn is set below in setup. */

    /* check that we have the same number of options entries and help*/
    assert((sizeof(pduReaderHelp)/sizeof(char *)) == 
           (sizeof(pduReaderOptions)/sizeof(struct option)));

    /* register the options handler */
    if (optionsRegister(pduReaderOptions,(optHandler)pduReaderOptionsHandler,
                        (clientData)flow_reader))
    {
        skAppPrintErr("unable to register flow reader options");
        return 1;                   /* error */
    }

    return 0;
}


/*
 *  pduReaderSetup:
 *
 *    Validate the flow reader's options and do anything required
 *    prior to processing records.
 *
 *    Invoked by flow_reader->setup_fn();
 */
static int pduReaderSetup(
    flow_reader_t      *flow_reader,
    uint8_t            *termination_flag,
    fp_daemon_mode_t   *is_daemon)
{
    t_probe = flow_reader->probe_from_opts;

    /* The probe-verification should have made certain that one and
     * only one of the following collectors is active. */

    if (NULL != skProbeGetFileSource(t_probe) ||
        NULL != option_args.file)         /* Moved to front,
                                             option_args.file
                                             supercedes. FIXME. */
    {
        /* file based */
        flow_reader->start_fn = &pduReaderStartFileReader;
        flow_reader->get_record_fn = &pduReaderGetRecordFile;
        *is_daemon = FP_DAEMON_OFF;
    } else if (0 == skProbeGetListenAsHost(NULL, NULL, t_probe)) {
        /* network based */
        flow_reader->start_fn = &pduReaderStartNetworkListener;
        flow_reader->get_record_fn = &pduReaderGetRecordNetwork;
        *is_daemon = FP_DAEMON_ON;
    } else if (NULL != skProbeGetListenOnUnixDomainSocket(t_probe)) {
        /* unix domain socket not supported */
        skAppPrintErr("Cannot listen on unix domain socket for PDUs");
        return -1;
    } else {
        skAppPrintErr("Probe has unknown collector");
        assert(0);
        return -1;
    }

    reading = termination_flag;

    return 0;
}


/*
 *  status = pduReaderOptionsValidate(processor_properties);
 *
 *    Verify the options to the flowcap processor.
 *
 *    Invoked by flow_reader->options_validate_fn();
 */
static int pduReaderOptionsValidate(probe_def_t *p)
{
    uint32_t port;

    if (skPackGetRootDir() == NULL) {
        skAppPrintErr("--root-directory switch is required");
        return -1;
    }

    if (NULL != option_args.netflow_port) {
        if (NULL == p) {
            skAppPrintErr("Cannot give %s switch with sensor config file",
                          pduReaderOptions[FP_NETFLOW_PORT].name);
            return 1;
        }
        if (NULL != option_args.file) {
            skAppPrintErr("Cannot give both %s and %s switches",
                          pduReaderOptions[FP_NETFLOW_PORT].name,
                          pduReaderOptions[FP_FILE].name);
            return 1;
        }
        if (skStringParseUint32(&port, option_args.netflow_port, 0, 0xFFFF)) {
            skAppPrintErr("Invalid %s value '%s'",
                          pduReaderOptions[FP_NETFLOW_PORT].name,
                          option_args.netflow_port);
            return 1;
        }
        if (0 != skProbeSetListenOnPort(p, port)) {
            skAppPrintErr("Cannot set probe listen-on-port");
            return 1;
        }
    } else if (NULL != option_args.file) {
#if 0
        if (NULL == p) {
            skAppPrintErr("Cannot give %s switch with sensor config file",
                          pduReaderOptions[FP_FILE].name);
            return 1;
        }
        if (0 != skProbeSetFileSource(p, option_args.file)) {
            skAppPrintErr("Cannot set probe file source");
            return 1;
        }
#else
        /* This branch was put in to allow option_args.file to
        supercede the probe information.   Hackish.  FIXME. */
        if (NULL != p && 0 != skProbeSetFileSource(p, option_args.file)) {
            skAppPrintErr("Cannot set probe file source");
            return 1;
        }
#endif
    } else if (NULL != p) {
        skAppPrintErr("Must have sensor config file or %s or %s switch",
                      pduReaderOptions[FP_NETFLOW_PORT].name,
                      pduReaderOptions[FP_FILE].name);
        return 1;
    }

    return 0;
}


/*
 * pduReaderOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int optindex: index into pduReaderOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int pduReaderOptionsHandler(
    clientData  cData,
    int         opt_index,
    char       *opt_arg)
{
    flow_reader_t *flow_reader = (flow_reader_t*)cData;

    /*
     * Set the first_option_name field on clientData if we haven't set
     * it yet.  When multiple potential readers are available,
     * rwflowpack will examine this variable to make certain only one
     * reader was selected.
     */
    if (NULL == flow_reader->first_option_name) {
        flow_reader->first_option_name = pduReaderOptions[opt_index].name;
    }

    switch (opt_index) {
      case FP_NETFLOW_PORT:
        option_args.netflow_port = opt_arg;
        break;

      case FP_FILE:
        if (opt_arg[0] == '\0') {
            skAppPrintErr("Empty filename supplied");
            return 1;
        }
        option_args.file = opt_arg;
        break;

      default:
        skAppPrintErr("unhandled option %d in switch at %s:%d\n",
                      opt_index, __FILE__, __LINE__);
        assert(0);
        exit(EXIT_FAILURE);
    }

    return 0;                     /* OK */
}


/*
 * pduReaderOptionsUsage:
 *      print options for this pduReader to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 *
 *    Invoked by flow_reader->options_usage_fn();
 */
static void pduReaderOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\nPDU Reader-specific options\n");
    for (i = 0; pduReaderOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. %s\n", pduReaderOptions[i].name,
                (pduReaderOptions[i].has_arg
                 ? (pduReaderOptions[i].has_arg == 1
                    ? "Required Arg"
                    : "Optional Arg")
                 : "No Arg"), pduReaderHelp[i]);
    }

    return;
}
