/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwpack/probeconf.c,v 1.2 2005/05/20 17:54:56 thomasm Exp $");

#include "utils.h"
#include "skvector.h"
#include "probeconf.h"
#include "silk_site.h"
#include "probeconfscan.h"
#include "log.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* Maximum valid value for a port 2^16 - 1 */
#define PORT_VALID_MAX 0xFFFF

/* Set ports to this invalid value initially */
#define PORT_NOT_SET   0xFFFFFFFF


/* Web classification utilities; 'p' is a port in native byte order;
 * 'r' is an RWGENERIC record in native byte order. */
#define IS_WEB_PORT(p)  ((p) == 80 || (p) == 443 || (p) == 8080)
#define IS_WEB(r) \
    ((r)->proto == 6 && (IS_WEB_PORT((r)->sPort) || IS_WEB_PORT((r)->dPort)))


/* a map between probe types and printable names */
static struct {
    const char     *name;
    probe_enum_t    value;
} probe_type_name_map[PROBE_TYPE_COUNT] = {
    {"netflow",  PROBE_ENUM_NETFLOW}
};


/* The probe structure. */
struct _probe_def {
    /* A bitmap where an ON bit is an SNMP interface on the probe
     * that should be considered to have incoming traffic. */
    BITMAP_DECLARE(     incoming_if_map, SK_SNMP_INDEX_LIMIT);

    /* A bitmap where an ON bit is an SNMP interface on the probe
     * that should be considered to have outgoing traffic. */
    BITMAP_DECLARE(     outgoing_if_map, SK_SNMP_INDEX_LIMIT);

    /* Number of bits that are set in each of the above */
    uint32_t            incoming_if_count;
    uint32_t            outgoing_if_count;

    /* The host:port combination on which this probe should listen for
     * data, as an IP address in network-byte order and a
     * port-number. */
    in_addr_t           listen_as_addr;
    uint32_t            listen_on_port;

    /* The host that this probe should accept connections from, as a
     * C-string and an IP in network-byte order */
    char               *accept_from_name;
    in_addr_t           accept_from_addr;

    /* The unix domain socket on which this probe should listen for
     * data, as a UNIX pathname */
    char               *unix_domain_path;

    /* A file name from which to read flow data */
    char               *file_source;

    /* the name of the probe */
    char                probe_name[SENSOR_MAX_SENSOR_NAME_LEN+1];

    /* A list (and a count of the elements in the list) that contains
     * the IP addresses of the ISP's this probe talks to. */
    uint32_t           *isp_ip_list;
    int32_t             isp_ip_count;

    /* The SNMP interface on the probe to which non-routed traffic is
     * sent.  This is 0 on Cisco routers. */
    uint16_t            null_interface;

    /* The sensor ID.  Either from the sensorInfo[] structure in
     * silk_site_*.h, or a value we have created. */
    sensorID_t          sensor_id;

    /* Probe priority */
    probe_priority_t    priority;

    /* Probe type */
    probe_enum_t        probe_type;

};


/* LOCAL FUNCTIONS */

static void probeConfProbeDestroyAll(void);
static void probeConfSensorDestroyAll(void);
static const sensorinfo_t *probeConfSensorGetInfo(
    sensorinfo_t   *out_s_info,
    sensorID_t      sensor_id);
static int skProbeVerifyNetflow(probe_def_t *probe);
static int skProbeVerify(probe_def_t *probe);
static int skProbeVerifyClass(probe_def_t *probe);


/* LOCAL VARIABLES */

/* The probes that have been created and verified */
static sk_vector_t *probe_vec = NULL;

/* This list of sensors added in the config file. */
static sk_vector_t *sensor_vec = NULL;

/* The ID of the first sensor that gets added in the config file.
 * This is equal to the global 'numSensors' at the time the
 * application is invoked. */
static sensorID_t sensor_first_id = 0;


/*
 *  *****  Flow Type  **************************************************
 *
 *
 *  The probe is used to determine the flow-type---index into the
 *  outFInfo[] array in silk_site_*.h---of a flow record (rwRec) read
 *  from that probe.
 *
 *  Pull in the C file that defines that function and its supporting
 *  functions.
 */



#  define skProbeVerifyClassGeneric skProbeVerifyClass
#  define skProbeDetermineFlowtypeGeneric skProbeDetermineFlowtype
#  include "probeconf-generic.c"



/*
 *  *****  Probe configuration  **************************************
 */


/* setup the probes */
int probeConfSetup(void)
{
    if (NULL == probe_vec) {
        probe_vec = skVectorNew(sizeof(probe_def_t*));
        if (NULL == probe_vec) {
            return -1;
        }
    }

    return 0;
}


/* destroy everything */
void probeConfTeardown(void)
{
    /* clean up the parser */
    probeConfParseTeardown();

    /* destroy all probes */
    probeConfProbeDestroyAll();

    /* destroy all sensors */
    probeConfSensorDestroyAll();
}


/* return a count of verified probes */
size_t probeConfProbeCount(void)
{
    assert(probe_vec);
    return skVectorGetCount(probe_vec);
}


/* Verify that the probe is valid, and if it is, add it to the config
 * file list. */
int probeConfProbeVerify(probe_def_t *probe)
{
    assert(probe);
    assert(probe_vec);

    if (0 != skProbeVerify(probe)) {
        return -1;
    }
    return skVectorAppendValue(probe_vec, &probe);
}


/* get a probe by position */
const probe_def_t *probeConfProbeGetByPosition(size_t probe_idx)
{
    probe_def_t *p;

    assert(probe_vec);

    if (0 != skVectorGetValue(&p, probe_vec, probe_idx)) {
        return NULL;
    }

    return p;
}


/* return number of probes for the given sensor id.  fill if out_vec
 * if provided. */
int probeConfGetProbesForSensor(sk_vector_t *out_vec, sensorID_t sid)
{
    const probe_def_t *probe;
    int match_sid = 0; /* return value */
    size_t count;
    size_t i;

    assert(probe_vec);

    /* check input */
    if (sid == SENSOR_INVALID_SID) {
        return -1;
    }
    if (out_vec != NULL) {
        if (skVectorGetElementSize(out_vec) != sizeof(probe_def_t*)) {
            return -1;
        }
    }

    count = skVectorGetCount(probe_vec);
    for (i = 0; i < count; ++i) {
        skVectorGetValue(&probe, probe_vec, i);
        if (probe->sensor_id == sid) {
            ++match_sid;
            if (out_vec != NULL) {
                if (0 != skVectorAppendValue(out_vec, &probe)) {
                    return -1;
                }
            }
        }
    }

    return match_sid;
}


/* return a probe having the given sensor-name and probe-name. */
const probe_def_t *probeConfGetProbeByName(
    const char *sensor_name,
    const char *probe_name)
{
    const probe_def_t *probe;
    sensorID_t sid;
    size_t count;
    size_t i;

    assert(probe_vec);

    /* check input */
    if (sensor_name == NULL || probe_name == NULL) {
        return NULL;
    }

    /* get the sensor id for named sensor */
    sid = probeConfSensorGetID(sensor_name);
    if (sid == SENSOR_INVALID_SID) {
        return NULL;
    }

    /* loop over all probes until we find one with given name */
    count = skVectorGetCount(probe_vec);
    for (i = 0; i < count; ++i) {
        skVectorGetValue(&probe, probe_vec, i);
        if (probe->sensor_id == sid) {
            if (0 == strcmp(probe_name, probe->probe_name)) {
                return probe;
            }
        }
    }

    return NULL;
}


/*
 *  probeConfProbeDestroyAll();
 *
 *    Free all memory used to store the local probe list.
 */
static void probeConfProbeDestroyAll(void)
{
    size_t i, count;
    probe_def_t *probe;

    if (NULL == probe_vec) {
        return;
    }

    /* Free all the probes */
    count = skVectorGetCount(probe_vec);
    for (i = 0; i < count; ++i) {
        skVectorGetValue(&probe, probe_vec, i);
        skProbeDestroy(&probe);
    }

    /* destroy the vector itself */
    skVectorDestroy(probe_vec);
    probe_vec = NULL;
}


/*
 *  *****  Sensors  ****************************************************
 *
 *
 *  It's handy to be able to pack things to new sensors or to test
 *  packing without having to modify the sensorInfo[] array in
 *  silk_site_*.h and recompile the entire SiLK suite.  Currently
 *  (March-2005) the sensor information is only stored in the file
 *  name; however, internally rwflowpack needs to be able to refer to
 *  the sensor by an ID number.  To handle this, when rwflowpack is
 *  given a sensor name not in the sensorInfo[] array, it creates a
 *  temporary ID for the sensor; the skTempSensor*() functions below
 *  are an interface to these temporary sensors.
 *
 *  Threading Issues: The sensor list should only be modified during
 *  start-up (options and config-file parsing) and shut-down.  They
 *  should not change during processing.
 */



/*
 *  probeConfSensorDestroyAll();
 *
 *    Free all memory used to store the local sensor list.
 */
static void probeConfSensorDestroyAll(void)
{
    sensorinfo_t s_info;
    size_t i, count;

    if (NULL == sensor_vec) {
        return;
    }

    /* Free all the sensorNames */
    count = skVectorGetCount(sensor_vec);
    for (i = 0; i < count; ++i) {
        skVectorGetValue(&s_info, sensor_vec, i);
        free(s_info.sensorName);
    }

    /* destroy the vector itself */
    skVectorDestroy(sensor_vec);
    sensor_vec = NULL;
}


/*
 *  s_info = probeConfSensorGetInfo(&sensor_info, sensor_id);
 *
 *    Return the pointer to the 'sensorinfo_t' that contains
 *    information about the sensor numbered 'sensor_id'.
 *
 *    If the sensor is defined in the silk_site_*.h file,
 *    'sensor_info' REMAINS UNCHANGED and the return value is the ONLY
 *    reference the caller gets to the sensorinfo_t.  If the sensor
 *    was created with probeConfSensorCreate(), the 'sensor_info' will
 *    be filled in, and the return value will be the same as first
 *    parameter the caller passed to this function.
 */
static const sensorinfo_t *probeConfSensorGetInfo(
    sensorinfo_t   *out_s_info,
    sensorID_t      sensor_id)
{
    if (NULL == sensor_vec) {
        /* no custom sensor were created */
        if (sensor_id < numSensors) {
            /* sensor is defined in sensorInfo[] in silk_site_*.h */
            return &(sensorInfo[sensor_id]);
        }
    } else if (sensor_id < sensor_first_id) {
        /* sensor is defined in sensorInfo[] in silk_site_*.h */
        return &(sensorInfo[sensor_id]);
    } else {
        sensor_id -= sensor_first_id;
        if (0 == skVectorGetValue(out_s_info, sensor_vec, sensor_id)) {
            return out_s_info;
        }
    }

    return NULL;
}


/*
 *  sid = probeConfSensorCreate(name, class_id_list, class_id_count);
 *
 *    Create 'name' as a new sensor name, give it a new sensor id, and
 *    return the ID.  'class_id_list' is an array class IDs to which
 *    this sensor belongs; 'class_id_count', is the number of elements
 *    in that array.
 */
sensorID_t probeConfSensorCreate(
    const char         *name,
    const uint32_t     *class_id_list,
    uint32_t            class_id_count)
{
    sensorID_t sensor_id = SENSOR_INVALID_SID;
    sensorinfo_t s_info;
    uint32_t i;

    /* check input */
    if (name == NULL || class_id_count == 0 || class_id_list == NULL) {
        skAppPrintErr("Sensors require a name and a class list.");
        goto END;
    }
    if (class_id_count > NUM_CLASSES) {
        skAppPrintErr("Sensors only support %d classes", NUM_CLASSES);
        goto END;
    }
    for (i = 0; i < class_id_count; ++i) {
        if (class_id_list[i] >= NUM_CLASSES) {
            skAppPrintErr("Class number %d is not valid", class_id_list[i]);
            goto END;
        }
    }
    if (sensorNameIsLegal(name)) {
        goto END;
    }

    /* verify doesn't exist */
    if (SENSOR_INVALID_SID != probeConfSensorGetID(name)) {
        skAppPrintErr("Sensor '%s' already exists", name);
        goto END;
    }

    /* Do we need to initialize? */
    if (sensor_vec == NULL) {
        /* Create a vector of sensorinfo_t's */
        sensor_vec = skVectorNew(sizeof(sensorinfo_t));
        if (sensor_vec == NULL) {
            return SENSOR_INVALID_SID;
        }

        /* Our first sensor ID is the current number of sensors */
        sensor_first_id = numSensors;
    }

    /* fill in the sensorinfo_t structure */
    s_info.sensorName = strdup(name);
    if (NULL == s_info.sensorName) {
        skAppPrintErr("Cannot malloc at %s:%d", __FILE__, __LINE__);
        goto END;
    }

    s_info.numClasses = class_id_count;
    for (i = 0; i < class_id_count; ++i) {
        s_info.classNumbers[i] = class_id_list[i];
    }

    if (0 != skVectorAppendValue(sensor_vec, &s_info)) {
        goto END;
    }

    /* we were successful, set the return value */
    sensor_id = sensor_first_id - 1 + skVectorGetCount(sensor_vec);

  END:
    return sensor_id;
}


/* return count of all sensors */
size_t probeConfSensorGetCount(void)
{
    if (NULL == sensor_vec) {
        return numSensors;
    }
    return ((size_t)(sensor_first_id + skVectorGetCount(sensor_vec)));
}


/* return a sensor name given its id */
const char *probeConfSensorGetName(sensorID_t sensor_id)
{
    sensorinfo_t sensor_info;
    const sensorinfo_t *s_info;

    s_info = probeConfSensorGetInfo(&sensor_info, sensor_id);
    if (s_info != NULL) {
        return s_info->sensorName;
    }

    return NULL;
}


/* return a sensor id given its name */
sensorID_t probeConfSensorGetID(const char *name)
{
    sensorinfo_t s_info;
    sensorID_t sid;
    size_t i, count;

    sid = sensorNameToId(name);
    if (sid != SENSOR_INVALID_SID) {
        return sid;
    }

    if (sensor_vec != NULL) {
        count = skVectorGetCount(sensor_vec);
        for (i = 0; i < count; ++i) {
            skVectorGetValue(&s_info, sensor_vec, i);
            if (0 == strcmp(s_info.sensorName, name)) {
                return (i + sensor_first_id);
            }
        }
    }

    return SENSOR_INVALID_SID;
}


/* return a vector containing the sensor names contained in the string */
int probeConfSensorStringParseList(
    sk_vector_t    *out_vec,
    const char     *sensor_str)
{
    char sensor_name[SENSOR_MAX_SENSOR_NAME_LEN+1];
    int matches = 0; /* return value */
    const char *cp;
    const char *ep;
    size_t len;
    sensorID_t sid;

    /* check input */
    if (sensor_str == NULL) {
        return -1;
    }
    if (out_vec != NULL) {
        if (skVectorGetElementSize(out_vec) != sizeof(sensorID_t)) {
            return -1;
        }
    }

    /* treat sensor_str as comma delimted list of sensor names */
    cp = sensor_str;
    while (*cp) {
        /* find next comma-delimited token; set 'len' to its length;
         * move 'ep' to the start of the next token. */
        ep = strchr(cp, ',');
        if (ep == cp) {
            /* double comma */
            ++cp;
            continue;
        }
        if (ep) {
            len = (size_t)(ep - cp);
            ++ep;
        } else {
            len = strlen(cp);
            ep = cp+len;
        }

        if (len > SENSOR_MAX_SENSOR_NAME_LEN) {
            /* bad sensor name */
            return -1;
        }

        /* copy the sensor name */
        strncpy(sensor_name, cp, len);
        sensor_name[len] = '\0';

        /* get its id */
        sid = probeConfSensorGetID(sensor_name);
        if (sid == SENSOR_INVALID_SID) {
            /* bad sensor name */
            return -1;
        }

        /* add to vector */
        ++matches;
        if (out_vec != NULL) {
            if (0 != skVectorAppendValue(out_vec, &sid)) {
                return 0;
            }
        }

        cp = ep;
    }

    return matches;
}



/*
 *  *****  Probes  *****************************************************
 */

/* Create a probe */
int skProbeCreate(probe_def_t **probe)
{
    assert(probe);

    (*probe) = calloc(1, sizeof(struct _probe_def));
    if (NULL == (*probe)) {
        return -1;
    }

    (*probe)->accept_from_addr = INADDR_ANY;
    (*probe)->listen_as_addr = INADDR_ANY;
    (*probe)->listen_on_port = PORT_NOT_SET;
    (*probe)->sensor_id = SENSOR_INVALID_SID;
    (*probe)->probe_type = PROBE_ENUM_NETFLOW;
    (*probe)->priority = PROBE_MIN_PRIORITY;

    return 0;
}


/* Destroy a probe and free all memory associated with it */
void skProbeDestroy(probe_def_t **probe)
{
    assert(*probe);
    if ( !(*probe)) {
        return;
    }

    if ((*probe)->isp_ip_count) {
        free((*probe)->isp_ip_list);
        (*probe)->isp_ip_list = NULL;
        (*probe)->isp_ip_count = 0;
    }

    free(*probe);
    *probe = NULL;
}


/* Set and get the sensor that this probe belongs to. */
int skProbeSetSensor(probe_def_t *probe, sensorID_t sensor_id)
{
    assert(probe);
    if (sensor_id >= probeConfSensorGetCount()) {
        return -1;
    }

    probe->sensor_id = sensor_id;
    return 0;
}

sensorID_t skProbeGetSensorID(const probe_def_t *probe)
{
    assert(probe);
    return probe->sensor_id;
}

const char *skProbeGetSensorName(const probe_def_t *probe)
{
    assert(probe);
    if (probe->sensor_id == SENSOR_INVALID_SID) {
        return NULL;
    }
    return probeConfSensorGetName(probe->sensor_id);
}


/* Set and get the name of probe */
int skProbeSetName(probe_def_t *probe, const char *name)
{
    int rv;

    assert(probe);
    rv = sensorNameIsLegal(name);
    if (rv != 0) {
        return rv;
    }

    strcpy(probe->probe_name, name);
    return 0;
}

const char *skProbeGetName(const probe_def_t *probe)
{
    assert(probe);
    return probe->probe_name;
}


/* set and get probe priority */
int skProbeSetPriority(
    probe_def_t        *probe,
    probe_priority_t    priority)
{
    assert(probe);
    if (priority < PROBE_MIN_PRIORITY || priority > PROBE_MAX_PRIORITY) {
        return -1;
    }
    probe->priority = priority;
    return 0;
}

probe_priority_t skProbeGetPriority(const probe_def_t *probe)
{
    assert(probe);
    return probe->priority;
}


/* set and get the probe type */
int skProbeSetType(
    probe_def_t        *probe,
    probe_enum_t        probe_type)
{
    assert(probe);
    if (probe_type == PROBE_ENUM_INVALID || probe_type > PROBE_TYPE_COUNT) {
        return -1;
    }
    probe->probe_type = probe_type;
    return 0;
}

probe_enum_t skProbeGetType(const probe_def_t   *probe)
{
    assert(probe);
    return probe->probe_type;
}


/* Set and get the null-interface */
int skProbeSetNullInterface(
    probe_def_t        *probe,
    uint32_t            null_interface)
{
    assert(probe);
    if (null_interface >= SK_SNMP_INDEX_LIMIT) {
        return -1;
    }
    probe->null_interface = (uint16_t)null_interface;
    return 0;
}

uint16_t skProbeGetNullInterface(const probe_def_t   *probe)
{
    assert(probe);
    return probe->null_interface;
}


/* Set and get the list of IPs of the ISP */
int skProbeSetIspIps(
    probe_def_t        *probe,
    const sk_vector_t  *isp_ip_vec)
{
    size_t count;

    assert(probe);

    if (isp_ip_vec == NULL) {
        return -1;
    }

    count = skVectorGetCount(isp_ip_vec);
    if (count == 0) {
        return -1;
    }

    if (probe->isp_ip_count) {
        free(probe->isp_ip_list);
    }

    probe->isp_ip_list = malloc(count * skVectorGetElementSize(isp_ip_vec));
    if (!probe->isp_ip_list) {
        return -1;
    }
    skVectorToArray(probe->isp_ip_list, isp_ip_vec);
    probe->isp_ip_count = count;

    return 0;
}

uint32_t skProbeGetIspIps(
    const uint32_t    **out_ip_list,
    const probe_def_t  *probe)
{
    assert(probe);

    if (probe->isp_ip_count > 0) {
        if (out_ip_list != NULL) {
            *out_ip_list = probe->isp_ip_list;
        }
    }
    return probe->isp_ip_count;
}


/* Set and get the list of incoming SNMP interfaces */
int skProbeSetIncomingInterfaces(
    probe_def_t        *probe,
    const sk_vector_t  *if_vec)
{
    size_t count, i;
    uint32_t val;

    assert(probe);
    if (if_vec == NULL) {
        return -1;
    }

    count = skVectorGetCount(if_vec);
    if ((count == 0) || (skVectorGetElementSize(if_vec) != sizeof(uint32_t))) {
        return -1;
    }

    BITMAP_INIT(probe->incoming_if_map);

    for (i = 0; i < count; ++i) {
        skVectorGetValue(&val, if_vec, i);
        BITMAP_SETBIT(probe->incoming_if_map, val);
    }
    probe->incoming_if_count = count;

    return 0;
}

uint32_t skProbeGetIncomingInterfaces(
    const uint32_t    **out_if_map,
    const probe_def_t  *probe)
{
    assert(probe);

    if (probe->incoming_if_count > 0) {
        if (out_if_map != NULL) {
            *out_if_map = probe->incoming_if_map;
        }
    }
    return probe->incoming_if_count;
}


/* Set and get the list of outgoing SNMP interfaces */
int skProbeSetOutgoingInterfaces(
    probe_def_t        *probe,
    const sk_vector_t  *if_vec)
{
    size_t count, i;
    uint32_t val;

    assert(probe);
    if (if_vec == NULL) {
        return -1;
    }

    count = skVectorGetCount(if_vec);
    if ((count == 0) || (skVectorGetElementSize(if_vec) != sizeof(uint32_t))) {
        return -1;
    }

    BITMAP_INIT(probe->outgoing_if_map);

    for (i = 0; i < count; ++i) {
        skVectorGetValue(&val, if_vec, i);
        BITMAP_SETBIT(probe->outgoing_if_map, val);
    }
    probe->outgoing_if_count = count;

    return 0;
}

uint32_t skProbeGetOutgoingInterfaces(
    const uint32_t    **out_if_map,
    const probe_def_t  *probe)
{
    assert(probe);
    
    if (probe->outgoing_if_count > 0) {
        if (out_if_map != NULL) {
            *out_if_map = probe->outgoing_if_map;
        }
    }
    return probe->outgoing_if_count;
}


/* set and get host:port to listen on. */
int skProbeSetListenAsAddr(
    probe_def_t    *probe,
    in_addr_t       addr)
{
    assert(probe);

    if (addr == 0) {
        return -1;
    }

    probe->listen_as_addr = addr;
    return 0;
}


int skProbeSetListenOnPort(
    probe_def_t    *probe,
    uint32_t        port)
{
    assert(probe);
    if (port > PORT_VALID_MAX) {
        return -1;
    }

    probe->listen_on_port = port;
    return 0;
}


int skProbeGetListenAsHost(
    in_addr_t          *out_addr,
    uint16_t           *out_port,
    const probe_def_t  *probe)
{
    assert(probe);
    if (probe->listen_on_port == PORT_NOT_SET) {
        return -1;
    }

    if (out_addr) {
        *out_addr = probe->listen_as_addr;
    }
    if (out_port) {
        *out_port = (uint16_t)probe->listen_on_port;
    }
    return 0;
}


/* set and get the unix domain socket on which to listen. */
int skProbeSetListenOnUnixDomainSocket(
    probe_def_t        *probe,
    const char *        u_socket)
{
    assert(probe);
    if (u_socket == NULL || u_socket[0] == '\0') {
        return -1;
    }
    probe->unix_domain_path = strdup(u_socket);
    return 0;
}

const char *skProbeGetListenOnUnixDomainSocket(const probe_def_t *probe)
{
    assert(probe);
    return probe->unix_domain_path;
}


/* set and get the file name to read data from */
int skProbeSetFileSource(
    probe_def_t        *probe,
    const char         *pathname)
{
    assert(probe);
    if (pathname == NULL) {
        return -1;
    }

    probe->file_source = strdup(pathname);
    if (!probe->file_source) {
        return -1;
    }

    return 0;
}

const char *skProbeGetFileSource(const probe_def_t *probe)
{
    assert (probe);
    return probe->file_source;
}


/* set and get host to accept connections from */
int skProbeSetAcceptFromHost(
    probe_def_t        *probe,
    const char         *host_name,
    in_addr_t           host_addr)
{
    assert(probe);
    if (host_addr == 0) {
        return -1;
    }

    if (host_name) {
        probe->accept_from_name = strdup(host_name);
    }
    probe->accept_from_addr = host_addr;
    return 0;
}

const char *skProbeGetAcceptFromHost(
    in_addr_t          *out_addr,
    const probe_def_t  *probe)
{
    assert(probe);
    if (out_addr) {
        *out_addr = probe->accept_from_addr;
    }
    return probe->accept_from_name;
}


/*
 *  is_valid = skProbeVerifyNetflow(p);
 *
 *    Verify that probe has everything required to collect netflow
 *    data.
 */
static int skProbeVerifyNetflow(probe_def_t *probe)
{
    /* Netflow does not support unix sockets */
    if (probe->unix_domain_path != NULL) {
        skAppPrintErr("Cannot listen on a UNIX domain socket for PDUs");
        return -1;
    }

    return 0;
}


/*
 *  is_valid = skProbeVerify(p);
 *
 *    Verify that 'p' is a valid probe.
 */
static int skProbeVerify(probe_def_t *probe)
{
    size_t count, i;

    assert(probe);
    assert(probe_vec);

    /* check name */
    if ('\0' == probe->probe_name[0]) {
        skAppPrintErr("probe has no name");
        return -1;
    }

    /* check sensor id */
    if (probe->sensor_id == SENSOR_INVALID_SID) {
        skAppPrintErr("probe's sensor_id is not set");
        return -1;
    }

    /* check for duplicate probe names for this sensor */
    count = skVectorGetCount(probe_vec);
    for (i = 0; i < count; ++i) {
        const probe_def_t *valid_probe;
        skVectorGetValue(&valid_probe, probe_vec, i);
        if ((probe->sensor_id == valid_probe->sensor_id)
            && (0 == strcmp(probe->probe_name, valid_probe->probe_name)))
        {
            skAppPrintErr("there is already a probe for sensor %s named %s",
                          probeConfSensorGetName(probe->sensor_id),
                          probe->probe_name);
            return -1;
        }
    }

    /* check that one and only one of port, unix socket, and
     * file-source were given */
    count = 0;
    if (probe->listen_on_port != PORT_NOT_SET) {
        ++count;
    }
    if (probe->unix_domain_path != NULL) {
        ++count;
    }
    if (probe->file_source != NULL) {
        ++count;
    }

    if (count != 1) {
        if (count == 0) {
            skAppPrintErr("probe needs a collection source. Must give one\n"
                          "\t of IP-port, unix-domain socket, or file name");
        } else {
            skAppPrintErr("multiple collection sources given.  Give only one\n"
                          "\t of IP-port, unix-domain socket, or file name");
        }
        return -1;
    }

    /* return the result of verifying the probe by its class */
    return skProbeVerifyClass(probe);
}


/*
 *  *****  Probes Types  ***********************************************
 */

/* return a number give a probe type name */
probe_enum_t skProbetypeNameToEnum(const char *name)
{
    int i;

    if (NULL != name) {
        for (i = 0; i < PROBE_TYPE_COUNT; ++i) {
            if (0 == strcmp(name, probe_type_name_map[i].name)) {
                return probe_type_name_map[i].value;
            }
        }
    }

    return PROBE_ENUM_INVALID;
}


/* return a number give a probe type number */
const char *skProbetypeEnumtoName(probe_enum_t type)
{
    int i;

    for (i = 0; i < PROBE_TYPE_COUNT; ++i) {
        if (type == probe_type_name_map[i].value) {
            return probe_type_name_map[i].name;
        }
    }

    return NULL;
}
