%{
/*
** Copyright (C) 2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  Parser for probe configuration file
**
*/


#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwpack/probeconfparse.y,v 1.5.6.1 2005/11/29 20:08:37 thomasm Exp $");

#define DECLARE_CLASSINFO_VARIABLES
#include "probeconf.h"
#include "probeconfscan.h"
#include "silk_site.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* Set DEBUG to 1 to enable debugging printf messasges, 0 otherwise.
 * Generally best to leave this commented out so gcc -DDEBUG=1 will
 * work */
/* #define DEBUG 0 */

/* Verify DEBUG is set */
#ifndef DEBUG
#  define DEBUG 0
#endif

/* For printing messages when DEBUG is non-zero.  Use as:
 *    DEBUG_PRINTF(("x is %d\n", x));
 * Note ((double parens)) */
#if DEBUG
#  define DEBUG_PRINTF(x) printf x
#else
#  define DEBUG_PRINTF(x)
#endif


/* error message strings */
#define PARSE_MSG_ERROR "error while parsing file %s at line %d:\n"
#define PARSE_MSG_WARN  "warning while parsing file %s at line %d:\n"

/* magic value used to denote that a uint16_t---which are stored in
 * uint32_t's in the parser---has not yet been given a value. */
#define UINT16_NO_VALUE  0x10000  /* 0xFFFF + 1 */

/* to avoid repeatedly creating and destroying number lists, we keep a
 * pool of free number_lists.  this is the size of the pool. */
#define NUMBER_LIST_POOL_CAPACITY  16


/* parsing a probe fills in this struct; then we use it to create an
 * skProbe. */
typedef struct {
    char           *probe_name;
    number_list_t  *input_index;
    number_list_t  *output_index;
    number_list_t  *isp_ip;
    number_list_t  *class;
    uint32_t        null_interface;
    uint32_t        priority;
    name_ip_pair_t  accept_from_host;
    in_addr_t       listen_as_host;
    uint32_t        listen_on_port;
    char           *listen_on_usocket;
    char           *read_from_file;
    int             error_count;
    sensorID_t      sid;
    probe_enum_t    probe_type;
} probe_attr_t;


/* LOCAL FUNCTIONS */

/* wrappers around an sk_vector_t of integers */
static number_list_t *number_list_new(void);
static void number_list_append_value(number_list_t *nl, uint32_t value);

/* create a sensor */
static void sensor_new_sensor(char *s, number_list_t *nl);

/* functions to start and complete a probe_attr_t */
static void probe_start(void);
static void probe_end(void);

/* functions to set attributes of a probe_att_t */
static void sensor_probe(char *s);
static void probe_name(char *s);
static void probe_priority(uint32_t n);
static void probe_probe_type(probe_enum_t n);
static void probe_null_interface(uint32_t n);
static void probe_isp_ip(number_list_t *nl);
static void probe_input_index(number_list_t *nl);
static void probe_output_index(number_list_t *nl);
static void probe_listen_as_host(uint32_t a);
static void probe_listen_on_port(uint32_t n);
static void probe_listen_on_usocket(const char *s);
static void probe_read_from_file(const char *s);
static void probe_accept_from_host(name_ip_pair_t *a);

/* functions to convert string input to another form */
static uint32_t parse_int_u16(const char *s);
static uint32_t parse_ip_addr(const char *s);
static name_ip_pair_t *parse_resolvable_host(const char *s);
static uint32_t parse_class(const char *s);
static probe_enum_t parse_probe_type(const char *s);
static char *parse_sensor_name(const char *s);


/* LOCAL VARIABLES */

/* current file name */
static const char *probescan_filename;

/* the size of the free number_list pool and the number of entries in
 * the pool. */
static number_list_t *number_list_pool[NUMBER_LIST_POOL_CAPACITY];
static int number_list_pool_count = 0;

/* The parser works on a single global set of probe attributes */
static probe_attr_t g_static_probe;
static probe_attr_t *p = &g_static_probe;

/* Global error count for return status of probeConfParse */
static int g_probe_err;


%}
%union {
    char           *string;
    uint32_t        u32;
    number_list_t  *u32_list;
    name_ip_pair_t *addr;
    probe_enum_t    p_type;
}

%token SENSOR_T
%token CLASS_T

%token SENSOR_PROBE_T
%token PROBE_NAME_T
%token PRIORITY_T
%token PROBE_TYPE_T
%token NULL_INTERFACE_T
%token ISP_IP_T
%token INPUT_INDEX_T
%token OUTPUT_INDEX_T
%token LISTEN_AS_HOST_T
%token LISTEN_ON_PORT_T
%token LISTEN_ON_USOCKET_T
%token READ_FROM_FILE_T
%token ACCEPT_FROM_HOST_T
%token COMMA
%token <string> ID

%type <u32>      int_u16
%type <u32_list> int_u16_list
%type <addr>     resolvable_host
%type <u32>      dotted_quad
%type <u32_list> dotted_quad_list
%type <u32>      class
%type <u32_list> class_list
%type <p_type>   probe_type
%type <string>   sensor_name

%%

    /*
     * ******************  GRAMMAR RULES  ***********************************
     */

input:                                    input_block
                                        | input input_block
                                        | error
{
    skAppPrintErr((PARSE_MSG_ERROR
                   "\tMisplaced or unrecognized keyword"),
                  probescan_filename, probescan_line);
};


input_block:                              sensor_defn
                                        | probe_defn
;

    /*
     *  A sensor
     */

sensor_defn:  SENSOR_T sensor_name CLASS_T class_list
{
    probe_end();
    sensor_new_sensor($2, $4);
};


    /*
     * A probe
     */

probe_defn:                               sensor_probe_clause probe_clauses
;

probe_clauses:                           /* empty */
			                | probe_clauses probe_clause
;

probe_clause:	  		          probe_name_clause
					| priority_clause
					| probe_type_clause
					| null_interface_clause
					| isp_ip_clause
					| input_index_clause
					| output_index_clause
					| listen_host_clause
					| listen_port_clause
		                        | listen_usocket_clause
		                        | read_file_clause
					| accept_host_clause
		                        | error
{
    skAppPrintErr((PARSE_MSG_ERROR
                   "\tBad keyword or missing value"),
                  probescan_filename, probescan_line);
};



sensor_probe_clause:                      SENSOR_PROBE_T sensor_name
{
    probe_end();
    probe_start();
    sensor_probe($2);
};

probe_name_clause:                        PROBE_NAME_T sensor_name
{
    probe_name($2);
};

priority_clause:		          PRIORITY_T int_u16
{
    probe_priority($2);
};

probe_type_clause:		          PROBE_TYPE_T probe_type
{
    probe_probe_type($2);
};

null_interface_clause:		          NULL_INTERFACE_T int_u16
{
    probe_null_interface($2);
};

isp_ip_clause:                            ISP_IP_T dotted_quad_list
{
    probe_isp_ip($2);
};

input_index_clause:	                  INPUT_INDEX_T int_u16_list
{
    probe_input_index($2);
};

output_index_clause:	                  OUTPUT_INDEX_T int_u16_list
{
    probe_output_index($2);
};

listen_host_clause:	                  LISTEN_AS_HOST_T dotted_quad
{
    probe_listen_as_host($2);
};

listen_port_clause:	                  LISTEN_ON_PORT_T int_u16
{
    probe_listen_on_port($2);
};

listen_usocket_clause:                    LISTEN_ON_USOCKET_T ID
{
    probe_listen_on_usocket($2);
};

read_file_clause:                         READ_FROM_FILE_T ID
{
    probe_read_from_file($2);
};

accept_host_clause:	                  ACCEPT_FROM_HOST_T resolvable_host
{
    probe_accept_from_host($2);
};


    /*
     *  Parse a list of 16-bit integers.  Returns as number_list_t*
     */

int_u16_list:			  int_u16
{
    number_list_t *nl = number_list_new();
    number_list_append_value(nl, $1);
    $$ = nl;
}
				| int_u16_list int_u16
{
    number_list_append_value($1, $2);
    $$ = $1;
}
				| int_u16_list COMMA int_u16
{
    number_list_append_value($1, $3);
    $$ = $1;
};

int_u16:                          ID
{
    $$ = parse_int_u16($1);
};



    /*
     * Convert a list of class names to integers.  Return as a
     * number_list_t*
     */

class_list:                  class
{
    number_list_t *nl = number_list_new();
    number_list_append_value(nl, $1);
    $$ = nl;
}
                                | class_list class
{
    number_list_append_value($1, $2);
    $$ = $1;
}
                                | class_list COMMA class
{
    number_list_append_value($1, $3);
    $$ = $1;
};

class:                       ID
{
    $$ = parse_class($1);
};


    /*
     *  Parse a list of dotted-quad IP addresses.  Addresses do not
     *  have to resolve.  Return as a number_list_t*
     */

dotted_quad_list:                 dotted_quad
{
    number_list_t *nl = number_list_new();
    number_list_append_value(nl, $1);
    $$ = nl;
}
				| dotted_quad_list dotted_quad
{
    number_list_append_value($1, $2);
    $$ = $1;
};
				| dotted_quad_list COMMA dotted_quad
{
    number_list_append_value($1, $3);
    $$ = $1;
};


dotted_quad:			  ID
{
    $$ = parse_ip_addr($1);
};


    /*
     *  Parse a single resolvable hostname.
     */

resolvable_host:		  ID
{
    $$ = parse_resolvable_host($1);
};


    /*
     *  Parse probe type
     */

probe_type:			ID
{
    $$ = parse_probe_type($1);
};


    /*
     *  Sensor and/or probe name; just strdup() it for now.
     */

sensor_name:                       ID
{
    $$ = parse_sensor_name($$);
};


    /*
     *  Directory name
     */


%%

/*
 * *******************   SUPPORTING CODE  ******************************
 */


/*
 *  *****  Number Lists  ************************************************
 *
 *
 * number_list, a sk_vector_t of uint32_t's
 */


/* get a new number_list from the free-pool; if pool is empty, create
 * a new one */
static number_list_t *number_list_new(void)
{
    number_list_t *nl;

    /* if there are any in the pool, return one.  Otherwise create one. */
    if (number_list_pool_count) {
        --number_list_pool_count;
        nl = number_list_pool[number_list_pool_count];
        skVectorClear(nl);
    } else {
        nl = skVectorNew(sizeof(uint32_t));
    }

    return nl;
}


/* add the number_list to the free-pool.  If the pool is full, just
 * destroy the number_list. */
static void number_list_destroy(number_list_t *nl)
{
    /* If the pool is full, destroy the one we were handed; otherwise
     * add it to the pool. */
    if (number_list_pool_count == NUMBER_LIST_POOL_CAPACITY) {
        skVectorDestroy(nl);
    } else {
        number_list_pool[number_list_pool_count] = nl;
        ++number_list_pool_count;
    }
}


/* remove all number_list's from the pool and destroy them */
static void number_list_pool_empty(void)
{
    int i;

    for (i = 0; i < number_list_pool_count; ++i) {
        skVectorDestroy(number_list_pool[i]);
    }
    number_list_pool_count = 0;
}


/* append the value to the number_list */
static void number_list_append_value(number_list_t *nl, uint32_t value)
{
    skVectorAppendValue(nl, &value);
}


/* get the number of entries in the number_list */
static size_t number_list_entries(number_list_t *nl)
{
    return skVectorGetCount(nl);
}


/* get the value at the specified position */
static uint32_t number_list_get_value(number_list_t *nl, size_t position)
{
    uint32_t val;
    skVectorGetValue(&val, nl, position);
    return val;
}


/*
 *  *****  Sensors  ****************************************************
 */

/*
 *  sensor_new_sensor(name, class_id_list);
 *
 *    Create sensor 'name' that belongs to classes 'class_id_list'.
 */
static void sensor_new_sensor(char *s, number_list_t *nl)
{
    size_t i;
    uint32_t class_id_list[MAX_CLASSES_PER_SENSOR];
    size_t count;
    uint32_t val;

    /* make certain name is legal */
    if (0 != sensorNameIsLegal(s)) {
        /* message already printed */
        goto END;
    }

    /* check that we don't have too many or too few arguments */
    count = number_list_entries(nl);
    if (count == 0) {
        skAppPrintErr((PARSE_MSG_ERROR
                       "\tNo classes given for test sensor '%s'"),
                      probescan_filename, probescan_line, s);
        goto END;
    }
    if (count > MAX_CLASSES_PER_SENSOR) {
        skAppPrintErr((PARSE_MSG_ERROR
                       "\tA sensor can only belong to %d classes"),
                      probescan_filename, probescan_line,
                      MAX_CLASSES_PER_SENSOR);
        goto END;
    }

    /* check that all the values are valid */
    for (i = 0; i < count; ++i) {
        val = number_list_get_value(nl, i);
        if (val == UINT16_NO_VALUE) {
            /* error msg printed by parser */
            goto END;
        }
        class_id_list[i] = val;
    }

    probeConfSensorCreate(s, class_id_list, count);

  END:
    if (s) {
        free(s);
        s = NULL;
    }
    if (nl) {
        number_list_destroy(nl);
    }
}


/*
 *  *****  Probes  *****************************************************
 */


/* print the probe attributes -- for debugging */
static void probe_print(void)
{
#if DEBUG
    size_t input_count;
    size_t output_count;
    size_t isp_count;
    size_t i;

    input_count = ((p->input_index != NULL)
                   ? number_list_entries(p->input_index)
                   : 0);
    output_count = ((p->output_index != NULL)
                    ? number_list_entries(p->output_index)
                    : 0);
    isp_count = ((p->isp_ip != NULL)
                 ? number_list_entries(p->isp_ip)
                 : 0);

    DEBUG_PRINTF((("Got a probe (%d errors)\n"
                   "\t probe_name = '%s'\n"
                   "\t sensor_for_probe = '%u'\n"
                   "\t priority = '%u'\n"
                   "\t probe_type = '%u'\n"
                   "\t null_interface = '%u'\n"),
                  p->error_count,
                  p->probe_name,
                  p->sid,
                  p->priority,
                  p->probe_type,
                  p->null_interface));

    DEBUG_PRINTF(("\t input_index_count = '%u'", (unsigned int)input_count));
    if (input_count > 0) {
        for (i = 0; i < input_count; ++i) {
            DEBUG_PRINTF(("%s%u",
                          ((0 == i) ? " [" : ","),
                          number_list_get_value(p->input_index, i)));
        }
        DEBUG_PRINTF(("]"));
    }
    DEBUG_PRINTF(("\n"));

    DEBUG_PRINTF(("\t output_index_count = '%u'", (unsigned int)output_count));
    if (output_count > 0) {
        for (i = 0; i < output_count; ++i) {
            DEBUG_PRINTF(("%s%u",
                          ((0 == i) ? " [" : ","),
                          number_list_get_value(p->output_index, i)));
        }
        DEBUG_PRINTF(("]"));
    }
    DEBUG_PRINTF(("\n"));

    DEBUG_PRINTF(("\t isp_count = '%u'", (unsigned int)isp_count));
    if (isp_count > 0) {
        for (i = 0; i < isp_count; ++i) {
            uint32_t val = number_list_get_value(p->isp_ip, i);
            DEBUG_PRINTF(("%s%u(%s)",
                          ((0 == i) ? " [" : ","), val, num2dot(val)));
        }
        DEBUG_PRINTF(("]"));
    }
    DEBUG_PRINTF(("\n"));

    DEBUG_PRINTF((("\t listen_as_host = '%u' '%s'\n"
                   "\t listen_on_port = '%u'\n"
                   "\t listen_on_usocket = '%s'\n"
                   "\t read_from_file = '%s'\n"),
                  p->listen_as_host, num2dot(p->listen_as_host),
                  p->listen_on_port,
                  (p->listen_on_usocket ? p->listen_on_usocket : "NULL"),
                  (p->read_from_file ? p->read_from_file : "NULL")));

    DEBUG_PRINTF((("\t accept_from_host = '%s' '%u' '%s'\n"),
                  (p->accept_from_host.name
                   ? p->accept_from_host.name
                   : "NULL"),
                  p->accept_from_host.ip,
                  num2dot(p->accept_from_host.ip)));
#endif /* DEBUG */
}


/* Take all the data on the global probe_att_t and use it to create a
 * new skProbe. */
static void probe_attr_to_sk_probe(void)
{
    probe_def_t *probe = NULL;
    int rv = 0;

    rv = skProbeCreate(&probe);
    if (rv) { goto END; }

    rv = skProbeSetSensor(probe, p->sid);
    if (rv) { goto END; }

    rv = skProbeSetName(probe, p->probe_name);
    if (rv) { goto END; }

    if (p->priority != UINT16_NO_VALUE) {
        rv = skProbeSetPriority(probe, p->priority);
        if (rv) { goto END; }
    }
    if (p->probe_type != PROBE_ENUM_INVALID) {
        rv = skProbeSetType(probe, p->probe_type);
        if (rv) { goto END; }
    }
    if (p->null_interface != UINT16_NO_VALUE) {
        rv = skProbeSetNullInterface(probe, p->null_interface);
        if (rv) { goto END; }
    }
    if (p->isp_ip) {
        rv = skProbeSetIspIps(probe, p->isp_ip);
        if (rv) { goto END; }
    }
    if (p->input_index) {
        rv = skProbeSetIncomingInterfaces(probe, p->input_index);
        if (rv) { goto END; }
    }
    if (p->output_index) {
        rv = skProbeSetOutgoingInterfaces(probe, p->output_index);
        if (rv) { goto END; }
    }
    if (p->listen_as_host) {
        rv = skProbeSetListenAsAddr(probe, p->listen_as_host);
        if (rv) { goto END; }
    }
    if (p->listen_on_port != UINT16_NO_VALUE) {
        rv = skProbeSetListenOnPort(probe, p->listen_on_port);
        if (rv) { goto END; }
    }
    if (p->listen_on_usocket) {
        rv = skProbeSetListenOnUnixDomainSocket(probe,
                                                p->listen_on_usocket);
        if (rv) { goto END; }
    }
    if (p->read_from_file) {
        rv = skProbeSetFileSource(probe, p->read_from_file);
        if (rv) { goto END; }
    }
    if (p->accept_from_host.name) {
        rv = skProbeSetAcceptFromHost(probe, p->accept_from_host.name,
                                      p->accept_from_host.ip);
        if (rv) { goto END; }
    }

    rv = probeConfProbeVerify(probe);

  END:
    if ((rv != 0) && (NULL != probe)) {
        ++g_probe_err;
        skProbeDestroy(&probe);
    }
}


/* This should be called just before starting a new probe and when we
 * finish reading a file (to complete the last probe).  It will try to
 * convert the data from config file---which is stored in the global
 * probe_att_t---to an skProbe, and then it will free() all memory the
 * parser has allocated for this probe_attr_t */
static void probe_end(void)
{
    if (p->probe_name != NULL) {
        probe_print();

        if (p->error_count == 0) {
            probe_attr_to_sk_probe();
        }
    }

    /* Free everything we allocated */
    if (p->probe_name) {
        free(p->probe_name);
    }
    if (p->isp_ip) {
        number_list_destroy(p->isp_ip);
    }
    if (p->input_index) {
        number_list_destroy(p->input_index);
    }
    if (p->listen_on_usocket) {
        free(p->listen_on_usocket);
    }
    if (p->accept_from_host.name) {
        free(p->accept_from_host.name);
    }

    memset(p, 0, sizeof(probe_attr_t));
}


/* Begin a new probe by setting the memory of the global probe_attr_t
 * into its initial state. */
static void probe_start(void)
{
    memset(p, 0, sizeof(probe_attr_t));
    p->null_interface  = UINT16_NO_VALUE;
    p->probe_type      = PROBE_ENUM_INVALID;
    p->priority        = UINT16_NO_VALUE;
    p->listen_on_port  = UINT16_NO_VALUE;
}


/*
 *  sensor_probe(sensor_name);
 *
 *    Start a new probe that belongs to the sensor 'sensor_name'.  Set
 *    the probe's name to 'sensor_name' as well.
 */
static void sensor_probe(char *s)
{
    /* see if this is a defined sensor */
    p->sid = probeConfSensorGetID(s);
    if (p->sid == SENSOR_INVALID_SID) {
        skAppPrintErr((PARSE_MSG_ERROR
                       "\tName '%s' is not the name of a sensor"),
                      probescan_filename, probescan_line, s);
        ++p->error_count;
    }

    /* Set the probe's name to 's' */
    p->probe_name = s;
}


/*
 *  probe_name(probe_name);
 *
 *    Set the name of the global probe to 'probe_name'.
 */
static void probe_name(char *s)
{
    /* name is legal? */
    if (0 != sensorNameIsLegal(s)) {
        /* message already printed */
        ++p->error_count;
    }
    if (p->probe_name) {
        free(p->probe_name);
    }

    /* store info */
    p->probe_name = s;
}


/*
 *  probe_priority(n);
 *
 *    Set the priority of the global probe to n.
 */
static void probe_priority(uint32_t n)
{
    /* see if parser got valid value(s) */
    if (n == UINT16_NO_VALUE) {
        /* error msg printed by parser */
        ++p->error_count;
        return;
    }

    /* check for value out of range */
    if (n > PROBE_MAX_PRIORITY) {
        skAppPrintErr((PARSE_MSG_WARN
                       "\tPriority %u too large; use 0 <= n <= %d"),
                      probescan_filename, probescan_line,
                      n, PROBE_MAX_PRIORITY);
        ++p->error_count;
        return;
    }

    /* warn on overwrite */
    if (p->priority != UINT16_NO_VALUE) {
        skAppPrintErr((PARSE_MSG_WARN
                       "\tOverwriting previous %s value for probe %s"),
                      probescan_filename, probescan_line,
                      probescan_clause, p->probe_name);
    }

    /* store info */
    p->priority = n;
}


/*
 *  probe_probe_type(t);
 *
 *    Set the probe-type of the global probe to 't'.
 */
static void probe_probe_type(probe_enum_t t)
{
    /* see if parser got valid value(s) */
    if (t == PROBE_ENUM_INVALID) {
        /* error msg printed by parser */
        ++p->error_count;
        return;
    }

    /* warn on overwrite */
    if (p->probe_type != PROBE_ENUM_INVALID) {
        skAppPrintErr((PARSE_MSG_WARN
                       "\tOverwriting previous %s value for probe %s"),
                      probescan_filename, probescan_line,
                      probescan_clause, p->probe_name);
    }

    /* store info */
    p->probe_type = t;
}


/*
 *  probe_null_interface(n);
 *
 *    Set the null-interface of the global probe to n.
 */
static void probe_null_interface(uint32_t n)
{
    /* see if parser got valid value(s) */
    if (n >= UINT16_NO_VALUE) {
        /* error msg printed by parser */
        ++p->error_count;
        return;
    }

    /* warn on overwrite */
    if (p->null_interface != UINT16_NO_VALUE) {
        skAppPrintErr((PARSE_MSG_WARN
                       "\tOverwriting previous %s value (%u) for probe %s"),
                      probescan_filename, probescan_line,
                      probescan_clause, p->null_interface, p->probe_name);
    }

    /* store info */
    p->null_interface = n;
}


/*
 *  probe_isp_ip(list);
 *
 *    Set the isp-ip's on the global probe to nl.
 */
static void probe_isp_ip(number_list_t *nl)
{
    size_t i;
    size_t count = number_list_entries(nl);
    uint32_t val;

    /* see if parser got valid value(s) */
    for (i = 0; i < count; ++i) {
        val = number_list_get_value(nl, i);
        if (val == 0) {
            ++p->error_count;
            return;
        }
    }

    /* warn on overwrite */
    if (p->isp_ip != NULL) {
        skAppPrintErr((PARSE_MSG_WARN
                       "\tOverwriting previous %s value for probe %s"),
                      probescan_filename, probescan_line,
                      probescan_clause, p->probe_name);
        number_list_destroy(p->isp_ip);
    }

    /* store info */
    p->isp_ip = nl;
}


/*
 *  probe_input_index(list);
 *
 *    Set the input SNMP index list on probe to list.
 */
static void probe_input_index(number_list_t *nl)
{
    size_t i;
    size_t count = number_list_entries(nl);
    uint32_t val;

    /* see if parser got valid value(s) */
    for (i = 0; i < count; ++i) {
        val = number_list_get_value(nl, i);
        if (val == UINT16_NO_VALUE) {
            ++p->error_count;
            return;
        }
    }

    /* warn on overwrite */
    if (p->input_index != NULL) {
        skAppPrintErr((PARSE_MSG_WARN
                       "\tOverwriting previous %s value for probe %s"),
                      probescan_filename, probescan_line,
                      probescan_clause, p->probe_name);
        number_list_destroy(p->input_index);
    }

    /* store info */
    p->input_index = nl;
}


/*
 *  probe_output_index(list);
 *
 *    Set the output SNMP index list on probe to list.
 */
static void probe_output_index(number_list_t *nl)
{
    size_t i;
    size_t count = number_list_entries(nl);
    uint32_t val;

    /* see if parser got valid value(s) */
    for (i = 0; i < count; ++i) {
        val = number_list_get_value(nl, i);
        if (val == UINT16_NO_VALUE) {
            ++p->error_count;
            return;
        }
    }

    /* warn on overwrite */
    if (p->output_index != NULL) {
        skAppPrintErr((PARSE_MSG_WARN
                       "\tOverwriting previous %s value for probe %s"),
                      probescan_filename, probescan_line,
                      probescan_clause, p->probe_name);
        number_list_destroy(p->output_index);
    }

    /* store info */
    p->output_index = nl;
}


/*
 *  probe_listen_as_host(n);
 *
 *    Set the global probe to list as IP 'n'.
 */
static void probe_listen_as_host(uint32_t n)
{
    /* see if parser got valid value(s) */
    if (n == 0) {
        /* error msg printed by parser */
        ++p->error_count;
        return;
    }

    /* warn on overwrite */
    if (p->listen_as_host != 0) {
        skAppPrintErr((PARSE_MSG_WARN
                       "\tOverwriting previous %s value for probe %s"),
                      probescan_filename, probescan_line,
                      probescan_clause, p->probe_name);
    }

    /* store info */
    p->listen_as_host = n;
}


/*
 *  probe_listen_on_port(n);
 *
 *    Set the global probe to listen for flows on port 'n'.
 */
static void probe_listen_on_port(uint32_t n)
{
    /* see if parser got valid value(s) */
    if (n == UINT16_NO_VALUE) {
        /* error msg printed by parser */
        ++p->error_count;
        return;
    }

    /* warn on overwrite */
    if (p->listen_on_port != UINT16_NO_VALUE) {
        skAppPrintErr((PARSE_MSG_WARN
                       "\tOverwriting previous %s value for probe %s"),
                      probescan_filename, probescan_line,
                      probescan_clause, p->probe_name);
    }

    /* store info */
    p->listen_on_port = n;
}


/*
 *  probe_listen_on_usocket(s);
 *
 *    Set the global probe to listen for flows on the unix domain socket 's'.
 */
static void probe_listen_on_usocket(const char *s)
{
    /* warn on overwrite */
    if (p->listen_on_usocket != NULL) {
        skAppPrintErr((PARSE_MSG_WARN
                       "\tOverwriting previous %s value for probe %s"),
                      probescan_filename, probescan_line,
                      probescan_clause, p->probe_name);
    }

    /* store info */
    p->listen_on_usocket = strdup(s);
}


/*
 *  probe_read_from_file(s);
 *
 *    Set the global probe to read flows from the file 's'.
 */
static void probe_read_from_file(const char *s)
{
    /* warn on overwrite */
    if (p->read_from_file != NULL) {
        skAppPrintErr((PARSE_MSG_WARN
                       "\tOverwriting previous %s value for probe %s"),
                      probescan_filename, probescan_line,
                      probescan_clause, p->probe_name);
    }

    /* store info */
    p->read_from_file = strdup(s);
}


/*
 *  probe_accept_from_host(a);
 *
 *    Set the globe probe to accept data from the host a.
 */
static void probe_accept_from_host(name_ip_pair_t *a)
{
    /* see if parser got valid value(s) */
    if (a->ip == 0) {
        /* error msg printed by parser */
        ++p->error_count;
        if (a->name) {
            free(a->name);
        }
    } else {
        /* warn on overwrite */
        if (p->accept_from_host.name != NULL) {
            skAppPrintErr((PARSE_MSG_WARN
                           "\tOverwriting previous %s value for probe %s"),
                          probescan_filename, probescan_line,
                          probescan_clause, p->probe_name);
            free(p->accept_from_host.name);
        }

        /* store info */
        p->accept_from_host.ip   = a->ip;
        p->accept_from_host.name = a->name;
    }

    a->name = NULL;
}


/*
 *  val = parse_int_u16(s);
 *
 *    Parse 's' as a integer from 0 to 0xFFFF inclusive.
 */
static uint32_t parse_int_u16(const char *s)
{
    uint32_t num;
    int rv;

    rv = skStringParseUint32(&num, s, 0, 0xFFFF);
    if (rv == 0) {
        /* everything is good. */
    } else if (rv == -12) {
        skAppPrintErr((PARSE_MSG_ERROR
                       "\tValue (%s) for %s is too large; use 0 <= n < 65536"),
                      probescan_filename, probescan_line,
                      s, probescan_clause);
        num = UINT16_NO_VALUE;
    } else {
        skAppPrintErr((PARSE_MSG_ERROR
                       "\tUnable to parse %s value '%s'"),
                      probescan_filename, probescan_line,
                      probescan_clause, s);
        num = UINT16_NO_VALUE;
    }

    return num;
}


/*
 *  class_id = parse_class(s);
 *
 *    Parse 's' as the name of a class.  Return the class ID.
 */
static uint32_t parse_class(const char *s)
{
    uint32_t i;
    uint32_t class_id = UINT16_NO_VALUE;

    for (i = 0; i < NUM_CLASSES; ++i) {
        if (0 == strcmp(s, classInfo[i].className)) {
            class_id = i;
            break;
        }
    }

    if (class_id == UINT16_NO_VALUE) {
        skAppPrintErr((PARSE_MSG_ERROR
                       "\tDo not recognize class name '%s'"),
                      probescan_filename, probescan_line,
                      s);
    }

    return class_id;
}


/*
 *  ip = parse_ip_addr(s);
 *
 *    Parse 'ip' as an IP address.  Return the IP.
 */
static uint32_t parse_ip_addr(const char *s)
{
    uint32_t num;

    num = dot2num(s);
    if (num == 0xFFFFFFFF) {
        /* may be an error, or maybe we got "255.255.255.255" */
        if ( 0 != strcmp(s, "255.255.255.255")) {
            skAppPrintErr((PARSE_MSG_ERROR
                           "\tCannot parse IP addresses '%s'"),
                          probescan_filename, probescan_line,
                          s);
            num = 0;
        }
    }

    return num;
}


/*
 *  name_ip = parse_resolvable_host(name);
 *
 *    Parse 'name' as a resolvable host, i.e., pass 'name' to
 *    gethostbyname().  Returns a pointer to a static struct that
 *    contains the name and the IP address.
 */
static name_ip_pair_t *parse_resolvable_host(const char *s)
{
    static name_ip_pair_t a = {(in_addr_t)0, NULL};
    struct hostent *he;
    in_addr_t addr;

    if (a.name) {
        free(a.name);
    }
    memset(&a, 0, sizeof(name_ip_pair_t));
    a.name = strdup(s);

    he = gethostbyname(s);
    if (he == NULL || he->h_length != sizeof(in_addr_t)) {
        skAppPrintErr((PARSE_MSG_ERROR
                       "\tUnable to resolve %s value '%s'"),
                      probescan_filename, probescan_line,
                      probescan_clause, s);
    } else {
        memcpy(&addr, he->h_addr, sizeof(in_addr_t));
        a.ip = ntohl(addr);
    }

    return &a;
}


/*
 *  type = parse_probe_type(s);
 *
 *    Parse 's' as the type of a probe.  Return the type.
 */
static probe_enum_t parse_probe_type(const char *s)
{
    probe_enum_t type;

    type = skProbetypeNameToEnum(s);
    if (type == PROBE_ENUM_INVALID) {
        skAppPrintErr((PARSE_MSG_ERROR
                       "\tDo not recognize probe type '%s'"),
                      probescan_filename, probescan_line, s);
    }

    return type;
}


/*
 *  sensor_name = parse_sensor_name(s);
 *
 *    Complain if 's' is an illegal sensor name.  Return a copy of s.
 */
static char *parse_sensor_name(const char *s)
{
    int rv;

    rv = sensorNameIsLegal(s);
    if (rv != 0) {
        if (rv > 0) {
            skAppPrintErr((PARSE_MSG_ERROR
                           "\tSensor name '%s'"
                           " contains the illegal character '%c'"),
                          probescan_filename, probescan_line, s, s[rv-1]);
        } else if (rv == -1) {
            skAppPrintErr((PARSE_MSG_ERROR
                           "\tSensor name '%s'"
                           " is too short."),
                          probescan_filename, probescan_line, s);
        } else if (rv == -2) {
            skAppPrintErr((PARSE_MSG_ERROR
                           "\tSensor name '%s' "
                           " is too long; only %d characters permitted."),
                          probescan_filename, probescan_line,
                          s, SENSOR_MAX_SENSOR_NAME_LEN);
        }
    }

    return strdup(s);
}


int yyerror(char UNUSED(*s))
{
    return 0;
}


int probeConfParse(const char *filename)
{
    if (filename == NULL) {
        return -1;
    }

    /* bzero() the global probe */
    memset(p, 0, sizeof(probe_attr_t));

    /* clear the error count */
    g_probe_err = 0;

    /* open the file */
    probescan_filename = filename;
    if (0 == strcmp(probescan_filename, "stdin")) {
        yyin = stdin;
    } else {
        yyin = fopen(probescan_filename, "r");
        if (!yyin) {
            skAppPrintErr("Cannot open file '%s'", probescan_filename);
            return -1;
        }
    }

    /* set line number */
    probescan_line = 1;

    /* parse */
    while (!feof(yyin)) {
        yyparse();
    }

    /* force output of last probe */
    probe_end();

    /* close */
    if (yyin != stdin) {
        fclose(yyin);
    }

    /* return non-zero if we had errors */
    if (g_probe_err > 0) {
        return -1;
    }

    return 0;
}


void probeConfParseTeardown(void)
{
    number_list_pool_empty();
}
