/*
** Copyright (C) 2003-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  Reads PDUs from a router and writes the flows into packed files.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwpack/rwflowpack.c,v 1.32.4.1 2005/11/29 20:08:37 thomasm Exp $");

#define DECLARE_CLASSINFO_VARIABLES 1
#include "silk_site.h"
#include "utils.h"
#include "log.h"
#include "rwpack.h"
#include "rwflowpack.h"
#include "probeconf.h"
#include "rwpack-new.h"
#include "stream-cache.h"
#include "sktimer.h"
#include "skvector.h"


/* MACROS AND DATA TYPES */

/* #undef DEBUGGING */
/* #define DEBUGGING 1 */

/* Where to write usage (--help) information */
#define USAGE_FH stdout

/* When (what hour) to rotate the logs.  We will disable rotation when
 * running as a daemon. */
#define LOG_ROTATION_HOUR 1

/* Force the packer to use version 2 data files */
#define RWFLOWPACK_DEFAULT_VERSION  2
#ifndef RWFLOWPACK_DEFAULT_VERSION
#  define RWFLOWPACK_DEFAULT_VERSION SK_FILE_VERSION_ANY
#endif


typedef void (*cleanupHandler)(void *);

/* signal list */
struct siglist {
    int signal;
    char *name;
};


/* EXPORTED VARIABLES */

pthread_mutex_t pqmutex = PTHREAD_MUTEX_INITIALIZER;

#define STREAM_CACHE_DEPTH 5
streamCache_t *g_streamCache;


/* PRIVATE FUNCTIONS */

static void appUsage(void);
static void appUsageLong(void);
static void appOptionsUsage(FILE*);
static void appTeardown(void);
static void appSetup(int argc, char **argv);
static int  appOptionsHandler(clientData, int, char *);
static int  byteOrderParse(const char *endian_string);
static int  validateOptions(int argc, char **argv);
static int  startProcessor(void);
static void stopProcessor(void);
static void closeFiles(void);
static int  processRecord(const rwRec *rwrec);
static int  determineDefaultProbe(void);
static int  setSensorNameDefaultProbe(probe_def_t *probe);
static sensorID_t createSensorFromOptions(void);
static int  skSiteParseClassList(uint32_t out_list[], const char *class_str);
static int  setInputInterfaces(probe_def_t *probe, const char *if_string);
static int  setNullInterface(probe_def_t *probe, const char *if_string);
static int  setIspIP(probe_def_t *probe, const char *ip_string);


/* PRIVATE VARIABLES */

/* The possible flow readers */
static flow_reader_t flow_readers[] = {
    /* PDU READER */
    {NULL, NULL, NULL, &pduReaderInitialize, 
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL}
};

/* Get a total count of the flow readers */
static int num_flow_readers = sizeof(flow_readers)/sizeof(flow_reader_t);

/* An index into flow_readers[] giving the one that is in use.  The
 * value given here is the default.  */
static int flow_reader_used = 0;

/* Convenience pointer equivalent to &(flow_readers[flow_reader_used]) */
static flow_reader_t *flow_reader = NULL;

/* Name of PID file */
static char pidfile[PATH_MAX];

/* struct to hold command line arguments required to create a
 * temporary sensor. */
static struct {
    const char *sensor_name;
    const char *class;
} sensor_attr = {
    NULL, NULL
};


/* struct to hold command line arguments required to create a probe */
static struct {
    const char *isp_ipnums;
    const char *null_interface;
    const char *in_index;
} probe_attr = {
    NULL, NULL, NULL
};

/* True as long as we are reading.  A handle to this variable is
 * passed to the flow readers, and they may set it to 0 to stop
 * the processing. */
static uint8_t reading = 0;

/* non-zero when rwflowpack is shutting down */
static uint8_t shuttingDown = 0;

/* control thread */
static pthread_t main_thread;

/* thread that reads flows */
static pthread_t process_thread;

/* Timer that flushes files every so often */
static skTimer_t timing_thread = NULL;

/* Byte order for newly packed files.  When appending to existing
 * files, use the files' existing byte order. */
static silk_endian_t byte_order = SILK_ENDIAN_NATIVE;

/* When non-zero, ignore the packing rules specified in the
 * silk_site_*.h headers and pack using the FT_RWROUTED format which
 * includes rounter SNMP interface numbers. */
static uint8_t pack_interfaces = 0;

/* When FP_DAEMON_OFF, do not fork -- useful for debugging */
static fp_daemon_mode_t daemon_mode = FP_DAEMON_ON;

/* Full path to the log file.  This will be based on log_directory,
 * unless the user explicitly sets it with the --log-pathname
 * option. */
static const char *log_pathname;

/* Directory in which to store PID file and log messages---unless
 * the --log-pathname option was given */
static const char *log_directory;

/* Second half of log-file name: based on sensor name */
static const char *logFileSuffix = NULL;

/* Options for byte-order switch */
static struct {
    const char     *name;
    silk_endian_t   value;
} byte_order_opts[] = {
    {"native", SILK_ENDIAN_NATIVE},
    {"little", SILK_ENDIAN_LITTLE},
    {"big",    SILK_ENDIAN_BIG},
    {NULL,     SILK_ENDIAN_ANY} /* sentinel */
};


/* Signals to catch and their names for signals--do we need a library
 * for this? */
static struct siglist signalList[] = {
    {SIGHUP, "SIGHUP"},
    {SIGINT, "SIGINT"},
    {SIGQUIT, "SIGQUIT"},
    {SIGUSR1, "SIGUSR1"},
    {SIGUSR2, "SIGUSR2"},
    {SIGTERM, "SIGTERM"},
#ifdef SIGPWR
    {SIGPWR, "SIGPWR"},
#endif

#ifndef DEBUGGING
    {SIGILL, "SIGILL"},
#ifdef SIGTRAP
    {SIGTRAP, "SIGTRAP"},
#endif
#if 0
    {SIGABRT, "SIGABRT"},
#endif
#ifdef SIGIOT
    {SIGIOT, "SIGIOT"},
#endif
#if 0
#ifdef SIGBUS
    {SIGBUS, "SIGBUS"},
#endif
#endif
    {SIGFPE, "SIGFPE"},
    {SIGSEGV, "SIGSEGV"},
#ifdef SIGURG
    {SIGURG, "SIGURG"},
#endif
#ifdef SIGXCPU
    {SIGXCPU, "SIGXCPU"},
#endif
#ifdef SIGXFSZ
    {SIGXFSZ, "SIGXFSZ"},
#endif
#ifdef SIGSYS
    {SIGSYS, "SIGSYS"},
#endif
#endif /* !DEBUGGING */

    {0, NULL}                   /* end flag */
};


/* OPTIONS SETUP */

typedef enum _appOptionsEnum {
    FP_LOG_DIRECTORY=0, FP_LOG_PATHNAME, FP_ROOT_DIR, FP_BYTE_ORDER,
    FP_PACK_INTERFACES, FP_NO_DAEMON,
    FP_SENSOR_CONFIG, FP_SENSOR_NAME,
#if (NUM_CLASSES > 1)
    FP_CLASS,
#endif
    FP_IN_INDEX, FP_NULL_INTERFACE,  FP_ISP_IPNUMS, FP_READER_FUNCTION   
} appOptionsEnum;

static struct option appOptions[] = {
    {"log-directory",        REQUIRED_ARG, 0, FP_LOG_DIRECTORY},
    {"log-pathname",         REQUIRED_ARG, 0, FP_LOG_PATHNAME},
    {"root-directory",       REQUIRED_ARG, 0, FP_ROOT_DIR},
    {"byte-order",           REQUIRED_ARG, 0, FP_BYTE_ORDER},
    {"pack-interfaces",      NO_ARG,       0, FP_PACK_INTERFACES},
    {"no-daemon",            NO_ARG,       0, FP_NO_DAEMON},
    {"sensor-configuration", REQUIRED_ARG, 0, FP_SENSOR_CONFIG},
    {"sensor-name",          REQUIRED_ARG, 0, FP_SENSOR_NAME},
#if (NUM_CLASSES > 1)
    {"class",                REQUIRED_ARG, 0, FP_CLASS},
#endif
    {"in-index",             REQUIRED_ARG, 0, FP_IN_INDEX},
    {"null-interface",       REQUIRED_ARG, 0, FP_NULL_INTERFACE},
    {"isp-ipnums",           REQUIRED_ARG, 0, FP_ISP_IPNUMS},
    {"reader-function",      REQUIRED_ARG, 0, FP_READER_FUNCTION},
    {0,0,0,0}  /* sentinal entry */
};

static const char *appHelp[] = {
    ("directory for process ID file unless --no-daemon given.\n"
     "    Directory for log file unless --log-pathname argument provided"),
    ("complete path to log file; overrides the path based\n"
     "    on the --log-directory argument and disables log rotation"),
    "root of directory in which to store packed files",
    ("byte order to use for newly packed files:\n" /*CONCAT*/
     "    Choices: 'native'(Default), 'little', or 'big'"),
    ("include interface indices in packed records\n" /*CONCAT*/
     "    (useful for debugging the router configuration)"),
    "do not fork off as a daemon (for debugging)",
    "read sensor configuration from named file.",
    "name of the sensor--used in generating file names",
#if (NUM_CLASSES > 1)
    "class to use when packing to unrecognized sensor name",
#endif
    ("comma separated list of SNMP input interface numbers;\n" /*CONCAT*/
     "    traffic matching these is considered \"incoming\""),
    ("SNMP interface router sends blocked traffic to;\n" /*CONCAT*/
     "    traffic matching this is \"null\" or \"forrouter\" (Default 0)"),
    ("comma separated list of IPs for NICs facing ISP;\n" /*CONCAT*/
     "    divides \"forrouter\" traffic from \"null\" traffic"),
    NULL, /* generate 'reader-function' help on the fly */
    (char *)NULL  /* sentinal entry */
};


/*
 * appUsage:
 *      print short usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 *      print complete usage information to USAGE_FH and exit with code.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
    int i;

    fprintf(USAGE_FH, "%s [options]\n", skAppName());

    for (i = 0; i < num_flow_readers; ++i) {
        if ( flow_readers[i].options_usage_fn ) {
            (*flow_readers[i].options_usage_fn)(USAGE_FH);
        }
    }
    appOptionsUsage(USAGE_FH);

    exit(EXIT_FAILURE);
}


/*
 * appOptionsUsage:
 *      print options for this app to the named file handle
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i, j;

    fprintf(fh, "\nCommon options:\n");
    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. ", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]));
        if ( appHelp[i]) {
            fprintf(fh, "%s", appHelp[i]);
        } else {
            switch (i) {
              case FP_READER_FUNCTION:
                fprintf(fh,
                        ("function to read flows. Default %d.\n"
                         "    Choices: "),
                        1+flow_reader_used);
                for (j = 0; j < num_flow_readers; ++j) {
                    fprintf(fh, "%d=%s", j+1,
                            (flow_readers[j].reader_name
                             ? flow_readers[j].reader_name
                             : "NO_NAME"));
                    if (j + 1 < num_flow_readers) {
                        fprintf(fh, ", ");
                    }
                }
                break;
              default:
                fprintf(fh, "MISSING HELP");
                break;
            }
        }
        fprintf(fh, "\n");
    }
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    logMsg("Shutting Down");
    shuttingDown = 1;
    stopProcessor();
    closeFiles();
    probeConfTeardown();
    logTeardown();
    optionsTeardown();

    if (pidfile[0]) {
        (void)unlink(pidfile);
    }

    return;
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int logrotate = LOG_ROTATION_HOUR;
    const char *logdir = NULL;
    char logfile[PATH_MAX];
    int len = -1;
    int i;

    /* Register program and initialize globals */
    skAppRegister(argv[0]);
    pidfile[0] = '\0';

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* verify that the default flow reader is good */
    assert(flow_reader_used >= 0 && flow_reader_used < num_flow_readers);

    /* setup options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* register the options handler */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler,
                        (clientData)NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* Initialize packing library */
    if (skPackInitialize()) {
        skAppPrintErr("unable to initialize packing library");
        exit(EXIT_FAILURE);
    }

    /* setup the probe configuration parser */
    if (probeConfSetup()) {
        skAppPrintErr("unable to setup probe config file parser");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* Allow each reader to do its initial setup and register its options */
    for (i = 0; i < num_flow_readers; ++i) {
        /* set the probe_from_opts pointer */
        //flow_readers[i].probe_from_opts = g_probe_from_opts;

        /* call it's initialize function */
        if (0 != ((*flow_readers[i].initialize_fn)( &flow_readers[i]))) {
            if (flow_readers[i].reader_name) {
                skAppPrintErr("unable to setup the %s flow reader",
                              flow_readers[i].reader_name);
            } else {
                skAppPrintErr("unable to setup the flow reader number %d", i);
            }
            exit(EXIT_FAILURE);
        }
    }

    /* parse and validate options */
    if ( validateOptions(argc, argv) ) {
        appUsage();  /* never returns */
    }

    /* set the mask so that the mode is 0664 */
    (void) umask((mode_t) 0022);

    /* diable log rotation when not a daemon */
    if (daemon_mode == FP_DAEMON_OFF) {
        logrotate = -1;
    }

    /* set variables for log directory and log file */
    if (log_pathname != NULL) {
        /* full path given by user; set logfile to full path, set
         * logdir to NULL, and disable log rotation. */
        len = snprintf(logfile, sizeof(logfile), "%s", log_pathname);
        logdir = NULL;
        logrotate = -1;
    } else {
        /* we use the log_directory we were given */
        logdir = log_directory;
        if (logFileSuffix != NULL) {
            len = snprintf(logfile, sizeof(logfile), "%s-%s",
                           skAppName(), logFileSuffix);
        } else {
            len = snprintf(logfile, sizeof(logfile), "%s", skAppName());
        }
    }
    if (len < 0 || len >= (int)sizeof(logfile)) {
        skAppPrintErr("log file name too long");
        exit(EXIT_FAILURE);
    }

    /* Open log file; no mail is available */
    if (logSetup(logfile, logdir, logrotate, 0))
    {
        skAppPrintErr("unable to initialize log file");
        exit(EXIT_FAILURE);
    }

    main_thread = pthread_self();

    return;                       /* OK */
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    uint32_t opt_val;

    switch (opt_index) {

      case FP_LOG_DIRECTORY:
        if (!dirExists(opt_arg)) {
            skAppPrintErr("the %s '%s' does not exist",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        log_directory = opt_arg;
        break;

      case FP_LOG_PATHNAME:
        log_pathname = opt_arg;
        break;

      case FP_ROOT_DIR:
        if (!dirExists(opt_arg)) {
            skAppPrintErr("the %s '%s' does not exist",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        skPackSetRootDir(opt_arg);
        break;

      case FP_BYTE_ORDER:
        return byteOrderParse(opt_arg);
        break;

      case FP_PACK_INTERFACES:
        pack_interfaces = 1;
        break;

      case FP_NO_DAEMON:
        daemon_mode = FP_DAEMON_OFF;
        break;

      case FP_SENSOR_CONFIG:
        if (probeConfParse(opt_arg)) {
            return 1;
        }
        break;

      case FP_SENSOR_NAME:
        logFileSuffix = opt_arg;
        sensor_attr.sensor_name = opt_arg;
        break;

#if (NUM_CLASSES > 1)
      case FP_CLASS:
        sensor_attr.class = opt_arg;
        break;
#endif

      case FP_IN_INDEX:
        probe_attr.in_index = opt_arg;
        break;

      case FP_NULL_INTERFACE:
        probe_attr.null_interface = opt_arg;
        break;

      case FP_ISP_IPNUMS:
        probe_attr.isp_ipnums = opt_arg;
        break;

      case FP_READER_FUNCTION:
        if (flow_reader) {
            skAppPrintErr("Option %s given multiple times",
                          appOptions[opt_index].name);
            return 1;
        }
        if (skStringParseUint32(&opt_val, opt_arg, 1, num_flow_readers)) {
            skAppPrintErr("Invalid %s value '%s'.  Choose 1 <= n <= %d",
                          appOptions[opt_index].name,
                          opt_arg, num_flow_readers);
            return 1;
        }
        flow_reader_used = opt_val - 1;
        flow_reader = &(flow_readers[flow_reader_used]);
        break;

      default:
        skAppPrintErr("unhandled option %d in options switch at %s:%d",
                      opt_index, __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }

    return 0;                     /* OK */
}


/*
 *  ok = byteOrderParse(argument)
 *
 *    parse the argument to the --byte-order switch
 */
static int byteOrderParse(const char *endian_string)
{
    static int option_seen = 0;
    int i;
    size_t len;

    /* only process option one time */
    if (option_seen != 0) {
        skAppPrintErr("Option %s given multiple times",
                      appOptions[FP_BYTE_ORDER].name);
        return 1;
    }
    option_seen = 1;

    len = strlen(endian_string);
    if (len == 0) {
        skAppPrintErr("Empty string given for %s option",
                      appOptions[FP_BYTE_ORDER].name);
        return 1;
    }

    /* initialize byte order */
    byte_order = SILK_ENDIAN_ANY;

    /* parse user's input */
    for (i = 0; byte_order_opts[i].name != NULL; ++i) {
        if ((len <= strlen(byte_order_opts[i].name))
            && (0 == strncmp(byte_order_opts[i].name, endian_string, len)))
        {
            if (byte_order != SILK_ENDIAN_ANY) {
                skAppPrintErr("Ambiguous %s value '%s'",
                              byte_order_opts[i].name, endian_string);
                return 1;
            }
            byte_order = byte_order_opts[i].value;
        }
    }

    if (byte_order == SILK_ENDIAN_ANY) {
        skAppPrintErr("Cannot parse %s value '%s'",
                      appOptions[FP_BYTE_ORDER].name, endian_string);
        return 1;
    }

    return 0;
}


/*
 *  ok = validateOptions(argc, argv);
 *
 *    Call the options parser.  If options parsing succeeds, validate
 *    that all the required arguments are present, and that the user
 *    didn't give inconsistent arguments.  Returns 0 on success, or -1
 *    otherwise.
 */
static int validateOptions(int argc, char **argv)
{
    /* Probe that we create from options given on command line. */
    probe_def_t *tmp_probe = NULL;
    int nextOptIndex;
    int i;
    fp_daemon_mode_t is_daemon;

    /* parse the options */
    if ( (nextOptIndex = optionsParse(argc, argv)) < 0) {
        /* options handler has printed error */
        return -1;
    }

    /* check for additional arguments */
    if (argc != nextOptIndex) {
        skAppPrintErr("too many or unrecognized argument specified '%s'",
                      argv[nextOptIndex]);
        return -1;
    }

    /* Make certain flow_reader is set */
    if (flow_reader == NULL) {
        assert(flow_reader_used >= 0 && flow_reader_used < num_flow_readers);
        flow_reader = &(flow_readers[flow_reader_used]);
    }

    /* Check that an inconsistent set of options was not provided */
    for (i = 0; i < num_flow_readers; ++i) {
        if (i != flow_reader_used) {
            /* we are not using this flow-reader, so make certain the
             * user did not give an option from this flow-reader */
            if (flow_readers[i].first_option_name) {
                skAppPrintErr("Option '%s' incompatible with %s",
                              flow_readers[i].first_option_name,
                              flow_readers[flow_reader_used].reader_name);
                return -1;
            }
        }
    }

    /* the sensor configuration is incompatible with probes from the
     * command line */
    if (probeConfProbeCount() > 0) {
        /* probes on command line */
        if (NULL != probe_attr.in_index) {
            skAppPrintErr("Cannot give both %s and %s switches",
                          appOptions[FP_SENSOR_CONFIG].name,
                          appOptions[FP_IN_INDEX].name);
            return -1;
        }
        if (NULL != probe_attr.null_interface) {
            skAppPrintErr("Cannot give both %s and %s switches",
                          appOptions[FP_SENSOR_CONFIG].name,
                          appOptions[FP_NULL_INTERFACE].name);
            return -1;
        }
        if (NULL != probe_attr.isp_ipnums) {
            skAppPrintErr("Cannot give both %s and %s switches",
                          appOptions[FP_SENSOR_CONFIG].name,
                          appOptions[FP_ISP_IPNUMS].name);
            return -1;
        }
#if (NUM_CLASSES > 1)
        if (NULL != sensor_attr.class) {
            skAppPrintErr("Cannot give both %s and %s switches",
                          appOptions[FP_SENSOR_CONFIG].name,
                          appOptions[FP_NULL_INTERFACE].name);
            return -1;
        }
#endif
    } else {
        /* no probes from the sensor configuration file */
        if (NULL == probe_attr.in_index) {
            skAppPrintErr("Must give either %s or %s switch",
                          appOptions[FP_SENSOR_CONFIG].name,
                          appOptions[FP_IN_INDEX].name);
            return -1;
        }
        if (NULL == sensor_attr.sensor_name) {
            skAppPrintErr("Must give either %s or %s switch",
                          appOptions[FP_SENSOR_CONFIG].name,
                          appOptions[FP_SENSOR_NAME].name);
            return -1;
        }

        /* Create a probe from the command line */
        if (0 != skProbeCreate(&tmp_probe)) {
            skAppPrintErr("unable to create probe");
            return -1;
        }
        skProbeSetType(tmp_probe, PROBE_ENUM_NETFLOW);
        skProbeSetName(tmp_probe, skProbetypeEnumtoName(PROBE_ENUM_NETFLOW));
        if (setInputInterfaces(tmp_probe, probe_attr.in_index)) {
            return -1;
        }
        if ((NULL != probe_attr.null_interface)
            && (0 != setNullInterface(tmp_probe, probe_attr.null_interface)))
        {
            return -1;
        }
        if ((NULL != probe_attr.isp_ipnums)
            && (0 != setIspIP(tmp_probe, probe_attr.isp_ipnums)))
        {
            return -1;
        }
    }

    /* Call the selected flow reader's options_validate_fn. */
    if ( (*flow_reader->options_validate_fn)(tmp_probe) ) {
        return -1;
    }

    if (probeConfProbeCount() > 0) {
        /* Handle the probe and sensor */
        if (0 != determineDefaultProbe()) {
            return -1;
        }
    } else {
        if (setSensorNameDefaultProbe(tmp_probe)) {
            skProbeDestroy(&tmp_probe);
            return -1;
        }
    }

    /* Call the setup function for the reader; use a separate variable
     * here to determine if the flow-reader wants us to run as a
     * daemon so that the flow-reader doesn't override the user's
     * --nodaeamon choice. */
    if ( (*flow_reader->setup_fn)(flow_reader, &reading, &is_daemon) ) {
        exit(EXIT_FAILURE);
    }

    if (is_daemon == FP_DAEMON_OFF) {
        daemon_mode = FP_DAEMON_OFF;
    }

    /* must have a place for log messages; must have a log-directory
     * for the PID file unless we are not in daemon mode. */
    if (log_directory == NULL) {
        if (log_pathname == NULL) {
            skAppPrintErr("--%s or --%s switch is required",
                          appOptions[FP_LOG_DIRECTORY].name,
                          appOptions[FP_LOG_PATHNAME].name);
            return -1;
        }
        if (daemon_mode == FP_DAEMON_ON) {
            skAppPrintErr("--%s switch is required for daemons",
                          appOptions[FP_LOG_DIRECTORY].name);
            return -1;
        }
    }

    return 0;
}


/*
** signalHandler:
**      Trap all signals and shutdown when told to.
*/
static void signalHandler(int sigNum)
{
    struct siglist *s;
    sigset_t mask_set;    /* used to set a signal masking set. */
    sigset_t old_set;     /* used to store the old mask set.   */

    /* mask any further signals while we're inside the handler. */
    sigfillset(&mask_set);
    sigprocmask(SIG_SETMASK, &mask_set, &old_set);

    /* This is the body of our signal handler */
    for (s = signalList; s->signal != sigNum; s++);

    if (sigNum == SIGUSR2) {
        /* Internal shutdown. */
        logMsg("Shutting down.");
    } else {
        logMsg("Shut down by %s signal", s->name);
        printf("Signalled to death\n");
    }
    appTeardown();
}


/*
** installSignalHandler:
**      Trap all signals we can here with our own handler.
**      Exception: SIGPIPE.  Set this to SIGIGN.
** Inputs: None
** Outputs: 0 if OK. 1 else.
*/
static int installSignalHandler(void)
{
    struct siglist *s;

    if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
        skAppPrintErr("Couldn't ignore SIGPIPE");
        return 1;
    }

    for (s = signalList; s->signal; s++) {
        if (signal(s->signal, signalHandler) == SIG_ERR) {
            skAppPrintErr("Couldn't handle %s", s->name);
            return 1;
        }
    }

    return 0;
}


static skTimerRepeat_t flushFiles(void *UNUSED(dummy))
{
    pthread_cleanup_push((cleanupHandler)pthread_mutex_unlock,
                         (void *)&pqmutex);
    pthread_mutex_lock(&pqmutex);

    logMsg("Five minute flush.");
    skStreamCacheFlush(g_streamCache);
    
    pthread_cleanup_pop(1);

    return SK_TIMER_REPEAT;
}


static void closeFiles(void)
{
    pthread_cleanup_push((cleanupHandler)pthread_mutex_unlock,
                         (void *)&pqmutex);
    pthread_mutex_lock(&pqmutex);

    logMsg("Closing all files.");

    /* Delete the cache, flushing, closing and freeing all the open
     * streams */
    skStreamCacheDelete(g_streamCache);
  
    pthread_cleanup_pop(1);
}


/*
 ********  PARSE PROBE INFO  *******
 *
 *
 */


/*
 *  ok = determineDefaultProbe();
 *
 *    Called when we were given a sensor configuration file, this
 *    function determines a 'default probe'.  If given the name of a
 *    sensor on the command line (and it must be only one), that is
 *    the name of the default probe.  Otherwise use the first probe in
 *    the config file as the default probe.  Set the 'probe_from_opts'
 *    field on the flow_reader to this default probe.
 */
static int determineDefaultProbe(void)
{
    sk_vector_t *sensor_vec = NULL;
    sk_vector_t *probe_vec = NULL;
    sensorID_t sensor_id = SENSOR_INVALID_SID;
    const char *sensor_name;
    const probe_def_t *default_probe;
    const probe_def_t *p;
    int count;
    int i;
    int rv = -1;

    /* FIXME.  Better cleanup */

    /* create vector to hold probe list */
    probe_vec = skVectorNew(sizeof(probe_def_t*));
    if (NULL == probe_vec) {
        skAppPrintErr("vector create failed");
        goto END;
    }

    if (NULL == sensor_attr.sensor_name) {
        /* No sensor on command line.  Get sensor of first probe */
        p = probeConfProbeGetByPosition(0);
        if (NULL == p) {
            skAppPrintErr("Cannot get first probe from config file");
            goto END;
        }

        sensor_id = skProbeGetSensorID(p);
        sensor_name = probeConfSensorGetName(sensor_id);
    } else {
        sensor_name = sensor_attr.sensor_name;

        /* create vector to hold list of sensors */
        sensor_vec = skVectorNew(sizeof(sensorID_t));
        if (NULL == sensor_vec) {
            skAppPrintErr("vector create failed");
            goto END;
        }

        /* get the list of sensor-ids; there had better be exactly one. */
        count = probeConfSensorStringParseList(sensor_vec, sensor_name);
        if (count < 1) {
            skAppPrintErr("Bad sensor name '%s'", sensor_name);
            goto END;
        }
        if (count > 1) {
            skAppPrintErr("Only one sensor can be provided on command line");
            goto END;
        }

        skVectorGetValue(&sensor_id, sensor_vec, 0);
    }

    /* get probes for this sensor */
    count = probeConfGetProbesForSensor(probe_vec, sensor_id);
    if (count < 1) {
        skAppPrintErr("No probes exist for sensor '%s'",
                      sensor_name);
        goto END;
    }

    default_probe = NULL;
    for (i = 0; i < count; ++i) {
        skVectorGetValue(&p, probe_vec, i);
        if (PROBE_ENUM_NETFLOW == skProbeGetType(p)) {
            if (default_probe) {
                skAppPrintErr("Support only one netflow probe per sensor");
                goto END;
            }
            default_probe = p;
        }
    }

    if (NULL == default_probe) {
        skAppPrintErr("No netflow probes exist for sensor '%s'",
                      sensor_name);
        goto END;
    }

    flow_reader->probe_from_opts = default_probe;
    rv = 0;

  END:
    if (probe_vec) {
        skVectorDestroy(probe_vec);
    }
    if (sensor_vec) {
        skVectorDestroy(sensor_vec);
    }

    return rv;
}


/*
 *  ok = setSensorNameDefaultProbe(probe);
 *
 *    Check that the 'sensor_name' given on the command line meets all
 *    requirements for a sensor name; if so, set the name of 'probe'
 *    to that name, then verify the probe.  Set the 'probe_from_opts'
 *    field on the flow_reader to probe.  Finally, return 0.
 *
 *    If anything fails, return -1.
 */
static int setSensorNameDefaultProbe(probe_def_t *probe)
{
    sensorID_t sensor_id;

    sensor_id = sensorNameToId(sensor_attr.sensor_name);
    if (sensor_id == SENSOR_INVALID_SID) {
        sensor_id = createSensorFromOptions();
        if (sensor_id == SENSOR_INVALID_SID) {
            return -1;
        }
    }

    if (0 != skProbeSetSensor(probe, sensor_id)) {
        skAppPrintErr("Cannot set sensor on probe");
        return -1;
    }

    /* make certain the probe got everything it needs */
    if (0 != probeConfProbeVerify(probe)) {
        return -1;
    }

    flow_reader->probe_from_opts = probeConfProbeGetByPosition(0);

    return 0;
}


/*
 *  sid = createSensorFromOptions();
 *
 *    Create a new sensor from the --sensor-name and --class options,
 *    and return its ID.  Return SENSOR_INVALID_SID if unable to
 *    create the sensor.
 */
static sensorID_t createSensorFromOptions(void)
{
    uint32_t class_id[NUM_CLASSES];
    int class_count;
    sensorID_t sensor_id;
    int rv;

    rv = sensorNameIsLegal(sensor_attr.sensor_name);
    if (rv != 0) {
        return SENSOR_INVALID_SID;
    }

    if (sensor_attr.class == NULL) {
#if (NUM_CLASSES == 1)
        class_id[0] = DEFAULT_CLASS;
        class_count = 1;
#else
        skAppPrintErr("Unknown sensors require class name list");
        return SENSOR_INVALID_SID;
#endif
    } else {
        class_count = skSiteParseClassList(class_id,
                                           sensor_attr.class);
        if (class_count < 1) {
            return SENSOR_INVALID_SID;
        }
        if (class_count > MAX_CLASSES_PER_SENSOR) {
            skAppPrintErr("Sensors only support %d classes per sensor",
                          MAX_CLASSES_PER_SENSOR);
            return SENSOR_INVALID_SID;
        }
    }

    /* add the name and id to our local cache of sensors */
    sensor_id = probeConfSensorCreate(sensor_attr.sensor_name,
                                      class_id, (uint32_t)class_count);
    if (SENSOR_INVALID_SID == sensor_id) {
        skAppPrintErr(("Sensor name '%s' is unrecognized and cannot"
                       " create a temporary sensor ID."),
                      sensor_attr.sensor_name);
        return SENSOR_INVALID_SID;
    }

    /* print warning but allow processing to continue */
    skAppPrintErr(("warning: Unrecognized sensor name '%s';\n"
                   "\t using temporary ID %u for this sensor"),
                  sensor_attr.sensor_name, sensor_id);

    return sensor_id;
}


/*
 *  num_classes = skSiteParseClassList(out_array, input_string);
 *
 *    Parse 'input_string' as a comma separated list of class names,
 *    and fill out_array[] with the IDs of the classes.  Return the
 *    number of classes found, or -1 on error.  The empty string will
 *    return a count of 0.  Duplicate class names are ignored.  The
 *    out_array[] should have at least NUM_CLASSES elements.
 */
static int skSiteParseClassList(
    uint32_t        out_class_list[],
    const char     *class_str)
{
    int class_seen[NUM_CLASSES];
    const char *cp;
    const char *ep;
    size_t len;
    uint32_t i;
    int found_class;
    int class_count = 0;

    memset(class_seen, 0, sizeof(class_seen));

    if (NULL == out_class_list || NULL == class_str) {
        return -1;
    }

    cp = class_str;
    while (*cp) {
        /* find next comma-delimited token; set len to its length */
        ep = strchr(cp, ',');
        if (ep == cp) {
            /* double comma */
            ++cp;
            continue;
        }
        if (ep) {
            len = (size_t)(ep - cp);
            ++ep;
        } else {
            len = strlen(cp);
            ep = cp+len;
        }

        /* look for the class as a name among all the classes.
         * found_class is 0 if we don't find it; 1 if we do */
        found_class = 0;
        for (i = 0; i < NUM_CLASSES; ++i) {
            if ((0 == strncmp(cp, classInfo[i].className, len)) &&
                ('\0' == classInfo[i].className[len]))
            {
                if (class_seen[i]) {
                    skAppPrintErr("Ignoring repeat entry for class '%s'",
                                  classInfo[i].className);
                } else {
                    class_seen[i] = 1+class_count;
                    out_class_list[class_count] = i;
                    ++class_count;
                }
                found_class = 1;
                break;
            }
        }

        if (found_class == 0) {
            char *err_name = malloc(len+1);
            if (err_name) {
                /* only show user the current name that is giving
                 * problems */
                strncpy(err_name, cp, len);
                err_name[len] = '\0';
                cp = err_name;
            }
            skAppPrintErr("Unknown class name '%s'", cp);
            if (err_name) {
                free(err_name);
            }
            class_count = -1;
            goto END;
        }

        cp = ep;
    } /* while (*cp) */

  END:
    return class_count;
}


/*
 *  ok = setInputInterfaces(probe, if_string);
 *
 *    Parse 'if_string' as a list of SNMP interfaces that face the
 *    ISP---the incoming interfaces, and set this list on 'probe'.
 */
static int setInputInterfaces(
    probe_def_t    *probe,
    const char     *if_string)
{
    uint32_t *if_list;
    uint32_t if_count;
    sk_vector_t *v;
    int rv = 1;  /* return value; assume failure */

    if (skStringParseNumberList(&if_list, &if_count, if_string, 0,
                                (SK_SNMP_INDEX_LIMIT-1), 0))
    {
        skAppPrintErr("Cannot parse in-index interface list '%s'", if_string);
        goto END;
    }

    v = skVectorNewFromArray(sizeof(if_list[0]), if_list, if_count);
    free(if_list);
    if (NULL == v) {
        skAppPrintErr("Allocation failure");
        goto END;
    }

    rv = skProbeSetIncomingInterfaces(probe, v);
    skVectorDestroy(v);

  END:
    return rv;
}


/*
 *  ok = setNullInterface(probe, if_string);
 *
 *    Parse 'if_string' as an integer representing the NULL interface
 *    and set this list on 'probe'.
 */
static int setNullInterface(
    probe_def_t    *probe,
    const char     *if_string)
{
    uint32_t null_if;

    if (skStringParseUint32(&null_if, if_string, 0, (SK_SNMP_INDEX_LIMIT-1))) {
        skAppPrintErr("Invalid null-interface value '%s'",
                      if_string);
        return 1;
    }

    return skProbeSetNullInterface(probe, null_if);
}


/*
 *  ok = setIspIP(probe, ip_string);
 *
 *    Parse 'ip_string' as a list of dotted IP addresses representing
 *    the IP addresses of the ISP, and set these the 'probe'.
 */
static int setIspIP(
    probe_def_t    *probe,
    const char     *ip_string)
{
    uint32_t isp_ip_list[MAX_IP_NUMS];
    int32_t isp_ip_count;
    sk_vector_t *v;
    int rv = 1; /* return value; assume failure */

    isp_ip_count = parseIPinOptions(ip_string, isp_ip_list);
    if (isp_ip_count == 0) {
        skAppPrintErr("Invalid isp-ipnums IP spec '%s'", ip_string);
        goto END;
    }

    v = skVectorNewFromArray(sizeof(isp_ip_list[0]),
                             isp_ip_list, isp_ip_count);
    if (NULL == v) {
        skAppPrintErr("Allocation failure");
        goto END;
    }

    rv = skProbeSetIspIps(probe, v);
    skVectorDestroy(v);

  END:
    return rv;
}


/*
 *  rwIOS = getStream(finfo_idx,pack_type,pack_file_vers,endian,timestamp,sid);
 *
 *    Returns the stream in which to write a flow record that has the
 *    specified ('finfo_idx', 'timespamp', and 'sid').  'finfo_idx' is
 *    used to determine where in the file system to write the record;
 *    it is an index into outFInfo[] (from silk_site_*.h).
 *    'pack_type' (e.g., FT_RWWWW) is used to determine how the record
 *    is stored--how to pack the record. 'timestamp' is the UTC hour
 *    of the record.  'packed_file_vers' is the version of the
 *    'pack_type' to use.  'endian' is the byte order to use for newly
 *    created files.  'sid' is the ID of the sensor where the flow was
 *    collected.  Creates a new file if required.  Returns NULL on
 *    failure.
 */
static rwIOStruct *getStream(
    int             finfo_idx,
    int             pack_type,
    fileVersion_t   packed_file_version,
    silk_endian_t   endian,
    uint32_t        timestamp,
    sensorID_t      sid)
{
    rwIOStruct *rwIOS;

    /* See if we have an already open stream in the cache */
    rwIOS = skStreamCacheLookup(g_streamCache, finfo_idx, timestamp, sid);

    /* If file is not in cache, bring it in. opens or creates file as
     * necessary */
    if (rwIOS == NULL) {
        char filename[PATH_MAX];

        /* Build the file name */
        (*flow_reader->filename_fn)(filename, sizeof(filename), timestamp,
                                    outFInfo[finfo_idx].pathPrefix,
                                    outFInfo[finfo_idx].filePrefix,
                                    probeConfSensorGetName(sid));

        /* Create a stream and associated it with this file (creating
         * if necessary) */
        rwIOS = skStreamCreate(filename, pack_type, packed_file_version,
                               endian, timestamp, sid);
        if (rwIOS) {
            /* Add the stream to the cache */
            skStreamCacheAdd(g_streamCache, finfo_idx, rwIOS);
        }
    }

    return rwIOS;
}


/*
 *  type = processRecord(&rwrec);
 *
 *    Packs the given rwrec into the correct file using the correct
 *    packing type.  Returns the type (index into outFInfo[]) on
 *    success, or -1 on failure.
 */
static int processRecord(const rwRec *rwrec)
{
    uint8_t type;
    uint32_t timestamp; 
    rwIOStruct *rwIOS = NULL;
    int pack_type;

    /* Only one thread accesses the file queue at a time */
    pthread_mutex_lock(&pqmutex);

    type = rwrec->flow_type;
    assert(type < NUM_FLOW_TYPES);

    /* Figure out preferred format to pack into */
    if (pack_interfaces) {
        pack_type = FT_RWROUTED;
    } else {
        pack_type = outFInfo[type].packType;
    }

    /* Determine the hour this data is associated with. This is a UTC
     * value expressed in seconds, rounded (down) to the hour. */
    timestamp = rwrec->sTime - rwrec->sTime % 3600;

    /* Get a stream to write to. */
    rwIOS = getStream(type, pack_type, RWFLOWPACK_DEFAULT_VERSION,
                      byte_order, timestamp, rwrec->sID);
    if ( !rwIOS) {
        logMsg("DEBUG: getStream failed for type %d pack_type %d",
               type, pack_type);
        goto END;
    }
  
    /* Write record */
    if (rwWrite(rwIOS, (rwRec*)rwrec)) {
        logMsg("DEBUG: rwWrite() failed for type %d pack_type %d",
               type, pack_type);
        type = -1;
        goto END;
    }

  END:
    pthread_mutex_unlock(&pqmutex);
    return type;
}


static void *processor(void * UNUSED(dummy))
{ 
    sigset_t sigs;
    rwRec rec;

    sigfillset(&sigs);
    pthread_sigmask(SIG_SETMASK, &sigs, NULL); /* Only handle SIGPIPE */

    while(1)
    {
        fp_get_record_result_t rv;

        rv = flow_reader->get_record_fn(&rec);
        if (!reading && 
            (rv == FP_BREAK_POINT || rv == FP_GET_ERROR)) {
            /* We can only exit if we are at a break point (or in an
               error condition) */
            break;
        }
        if (rv == FP_END_STREAM) {
            reading = 0;
            /* Tell the main thread to quit. */
            pthread_kill(*(pthread_t *)main_thread, SIGUSR2);
            break;
        }
        if (rv == FP_GET_ERROR) {
            /* Ignore records with errors */
            continue;
        }

        processRecord(&rec);
    }

    return NULL;
}


static int startProcessor(void)
{
    pthread_attr_t attr;

    /* Create a cache of streams (file handles) so we don't have to
     * incur the expense of reopening files */
    logMsg("Creating stream cache");
    g_streamCache = skStreamCacheCreate(numFileTypes, STREAM_CACHE_DEPTH);
    if (NULL == g_streamCache) {
        return 1;
    }

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    if (skTimerCreate(&timing_thread, 300, flushFiles, NULL) == -1) {
        return 1;
    }

    reading = 1;

    /* For the multi-probe case, we'd spawn threads for each probe */
    if (flow_reader->start_fn() != 0) {
        reading = 0;
        return 1;
    }
    if (pthread_create(&process_thread, NULL, processor, NULL)) {
        reading = 0;
        return 1;
    }

    return 0;
}


static void stopProcessor(void)
{
    if (reading) {
        logMsg("Stopping processor");
        reading = 0;

        /* Give the thread a chance to quit on its own */
        sleep(2);

        skTimerDestroy(timing_thread);
        flow_reader->stop_fn();
        logMsg("Waiting for record handler.");
        pthread_join(process_thread, NULL); /* join */

        if (flow_reader->teardown_fn) {
            (*flow_reader->teardown_fn)();
        }
    }
}


/* Make it a daemon */
static void daemonize(void)
{
    pid_t pid;
    char pids[13];
    int fd;
    int len;

    if ((pid = fork()) == -1) {
        logMsg("Couldn't fork for daemon.");
        exit(EXIT_FAILURE);
    } else if (pid != 0) {
        exit(0);
    }

    setsid();
    chdir("/");
    umask(0);

    pid = getpid();

    logMsg("new rwflowpack: PID=%d", (int)pid);

    len = snprintf(pids, sizeof(pids), "%d", (int)pid);
    if (len <= 0) {
        logMsg("Couldn't generate pid.");
        goto ERROR;
    }

    if (logFileSuffix) {
        snprintf(pidfile, sizeof(pidfile), "%s/%s-%s.pid",
                 log_directory, skAppName(), logFileSuffix);
    } else {
        snprintf(pidfile, sizeof(pidfile), "%s/%s.pid",
                 log_directory, skAppName());
    }

    if ((fd = open(pidfile, O_CREAT | O_WRONLY | O_TRUNC, 0644)) == -1) {
        logMsg("Couldn't write pid file '%s': %s",
               pidfile, strerror(errno));
        goto ERROR;
    }

    if (write(fd, pids, len) != len) {
        logMsg("Error writing pid file.");
        goto ERROR;
    }

    close(fd);

    /* success */
    return;

    /* only execute this on an error */
  ERROR:
    appTeardown();
    exit(EXIT_FAILURE);
}


int main(int argc, char **argv)
{
    appSetup(argc, argv);                 /* never returns on error */

    /* install the signal hanlers */
    if (installSignalHandler()) {
        logMsg("Failed to start signal handler.  Exiting.");
        exit(EXIT_FAILURE);
    }

#ifndef DEBUGGING
    /* Run as a daemon */
    if (daemon_mode == FP_DAEMON_ON) {
        daemonize();
    }
#endif

    /* register after call to daemonize() so we don't run
     * appTeardown() in the parent process and write a confusing
     * "shutting down" message to the log */
    if (atexit(appTeardown) < 0) {
        logMsg("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    /* Choose which processing thread to run based on user choice */
    if (startProcessor() != 0) {
        logMsg("unable to start flow processor");
        exit(EXIT_FAILURE);
    }

    /* We now run forever, excepting signals */
    while (!shuttingDown) {
        pause();
    }

    /* done */
    appTeardown();

    return 0;
}
