/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/
#ifndef _RWFLOWPACK_H
#define _RWFLOWPACK_H

#include "silk.h"

RCSIDENTVAR(rcsID_RWFLOWPACK_H, "$SiLK: /src/rw/rwpack/rwflowpack.h,v 1.6 2005/05/20 17:54:58 thomasm Exp $");

/*
**  rwflowpack.h
**
**  rwflowpack takes flow records and packs them into binary files.
**  The flow records come from "flow readers".  Eventually we'd like
**  the flow readers to be shared objects that are loaded at run-time,
**  but currently the readers are compiled in.  Most interaction
**  between the flow readers and the main body of rwflowpack occurs
**  through a structure defined in this header.
*/

#include "rwpack-new.h"
#include "probeconf.h"


/*
 *  These values indicate whether the flow reader operates as a
 *  daemon.  When acting as a daemon, rwflowpack will fork().
 */
typedef enum {
    FP_DAEMON_OFF, FP_DAEMON_ON
} fp_daemon_mode_t;


/* 
 * These values give information back to the caller of the
 * get_record_fn as to whether the function worked, and whether it is
 * safe to stop pulling data.
 */
typedef enum {
    FP_GET_ERROR = -1,          /* An error occured */
    FP_BREAK_POINT = 0,         /* You may stop before processing the record */
    FP_RECORD,                  /* You may not stop after processing record */
    FP_END_STREAM               /* You must stop.  There is no record. */
} fp_get_record_result_t;

/*
 *  This is the structure that the main rwflowpack uses to communicate
 *  with the flow readers.  rwflowpack will have an array of these,
 *  with only the initialize_fn filled in---when we move to using
 *  dlopen(), this may be handled by the shared object's _init()
 *  function.  This struct gets passed into the initialize_fn which
 *  will fill in other values.  Since the struct is self-referencing,
 *  we must pre-declare it.
 */
typedef struct _flow_reader flow_reader_t;
struct _flow_reader {
    /* Name of this reader, for printing in --help output */
    const char *reader_name;

    /* When a reader's option handler is called the first time, this
     * field should set to the name of the option being processed.
     * The option checking code in rwflowpack will compare this field
     * across all readers to determine if multiple readers have been
     * specified.  */
    const char *first_option_name;

    /* For ease of testing and backward compatibility, a probe can be
     * created from the options given on the command line.  This
     * command-line probe is a global probe created in the
     * rwflowpack.c file, what we have here is pointer to that global
     * probe. */
    const probe_def_t *probe_from_opts;

    /* Function for the flow reader to initialize itself: register
     * options, initialize variables, etc.  It should set the
     * 'reader_name' and other fields on the 'flow_reader'
     * structure. */
    int       (*initialize_fn)(flow_reader_t *flow_reader);

    /* Function for flow reader to tear itself down */
    void      (*teardown_fn)(void);

    /* Function to print the options usage specific to this reader.
     * Argument is the file handle for output. */
    void      (*options_usage_fn)(FILE*);

    /* Function to do any final option validation once all options
     * have been processed.  This function should return 0 if
     * flow-reader has valid values for all the options it requires,
     * or non-zero otherwise.  After this is called, the main program
     * may do additional validation and setup, and then the main
     * program will the 'setup_fn'.
     */
    int       (*options_validate_fn)(probe_def_t *probe);

    /* Function to run to complete reader setup after the user has
     * chosen this reader.  The function set any other state to
     * prepare for the main processing loop in rwflowpack to call this
     * reader's 'get_record_fn'.  'flow_reader' is this structure.
     * 'termination_flag' is a pointer to the variable that that turns
     * false when it is time to terminate the program.  This
     * 'setup_fn' should set the 'is_daemon' flag depending on whether
     * the main processing loop should daemon-ize itself.
     */
    int       (*setup_fn)(flow_reader_t        *flow_reader,
                          uint8_t              *termination_flag,
                          fp_daemon_mode_t     *is_daemon);
    
    /* Function to start any necessary background record collection.
     * Returns 0 on success, -1 on failure. */
    int       (*start_fn)(void);

    /* Function to stop the previous background record collection.
     * Must also be sure to unblock a currently blocking call to
     * get_record_fn, if blocked. */
    int       (*stop_fn)(void);

    /* Retrieves the next record.  Return value indicates what actions
     * are safe after retrieving the value.  (See the enum definition
     * for fp_get_record_result_t above) */
    fp_get_record_result_t (*get_record_fn)(rwRec *);

    /* Function to get path to file where flow record will be written.
     * Path is returned in 'fname'; 'maxlen' is the length of fname;
     * 'sTime' is the file start time in UNIX epoch seconds; 'path' is
     * the path prefix from outFInfo[]; 'file' is the file prefix from
     * outFInfo[]; 'sID' is the name of the sensor. */
    void      (*filename_fn)(char*fname, int maxlen, uint32_t sTime,
                             const char*path,const char*file,const char*sID);
};


/*
 * The initialization functions for each flow reader.  Probably should
 * be in a header file for each reader, but seems overkill to have a
 * header for a single function declaration.
 */
extern int  pduReaderInitialize(flow_reader_t *flow_reader);


#endif /* _RWFLOWPACK_H */
