/*
** Copyright (C) 2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
** Take a tcpdump capture file as input, and output the same file, but
** with timestamps skewed by 0-4 milliseconds.
*/

#include "silk.h"

RCSIDENT("");

#include <pcap.h>
#include "utils.h"
#include "rwppacketheaders.h"

#define USAGE_FH stdout


/* LOCAL FUNCTIONS */

static void appTeardown(
    void );
static void appSetup(
    int argc, 
    char **argv );

static void appUsage(
    void );
static void appUsageLong(
    void );
static void appOptionsUsage(
    FILE* );
static int  appOptionsHandler(
    clientData  cData, 
    int         opt_index, 
    char       *opt_arg );


/* LOCAL VARIABLES */

static pcap_t *g_pkt_input = NULL;
static pcap_t *g_output;
static pcap_dumper_t *g_output_dumper = NULL;


/* OPTION SETUP */

typedef enum {
    RWPMATCH_OPT_FLOWFILE
} appOptionsEnum;


static struct option appOptions[] = {
    { 0, 0, 0, 0 } /* sentinal entry */
};


static const char *appHelp[] = {
    (char *)NULL
};


/*
 *  appUsage:
 *      print usage information to stderr and exit with code 1
 *  Arguments: None
 *  Returns: None
 *  Side Effects: exits application.
 */
static void appUsage( 
    void )
{
    fprintf( USAGE_FH, "Use `%s --help' for usage\n", skAppName() );
    exit( EXIT_FAILURE );
}


/*
 *  appUsageLong:
 *      print usage information to USAGE_FH and exit with code 1.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 *  Arguments: None
 *  Returns: None
 *  Side Effects: exits application.
 */
static void appUsageLong( 
    void )
{
    fprintf( USAGE_FH, "Usage: %s [APPLICATION SWITCHES] [FILE]\n", 
             skAppName() );
    appOptionsUsage( USAGE_FH );
    exit( EXIT_FAILURE );
}


/*
 *  appOptionsUsage:
 *      print help for options for this app to the specified file handle.
 *   Arguments: None.
 *   Returns: None.
 *   Side Effects: None.
 *   NOTE:
 *       do NOT add non-option usage here (i.e., required and other
 *       optional args)
 */
static void appOptionsUsage(
    FILE *fh )
{
    int i;

    fprintf( fh, "\nAPPLICATION SWITCHES control input and output\n\n" );
    for ( i = 0; appOptions[i].name; i++ ) {
        fprintf( fh, "--%s %s. %s\n", appOptions[i].name,
                 SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i] );
    }
}


/*
 *  appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 *  Arguments:
 *      argc, argv
 *  Returns: None.
 *  Side Effects:
 *      exits the application with code 1 if anything does not work.
 */
static void appSetup( 
    int argc, 
    char **argv )
{
    char errbuf[PCAP_ERRBUF_SIZE];

    int first_file;

    skAppRegister( argv[0] );

    assert( ( sizeof( appHelp ) / sizeof( char* ) ) ==
            ( sizeof( appOptions ) / sizeof( struct option ) ) );

    if ( optionsSetup( appUsageLong ) ) {
        skAppPrintErr( "unable to setup options module" );
        exit( EXIT_FAILURE ); /* never returns */
    }

    /* set permissions for output files */
    (void) umask( (mode_t) 0002 );

    if ( optionsRegister( appOptions, (optHandler) appOptionsHandler, NULL ) ) {
        skAppPrintErr( "unable to register application options" );
        exit( EXIT_FAILURE );
    }

    /* parse options */
    if ( ( first_file = optionsParse( argc, argv ) ) < 0 ) {
        appUsage(); /* never returns */
    }

    /*
    ** Open input files
    */
    if ( first_file >= argc ) {
        skAppPrintErr( "no input packet file found" );
        exit( EXIT_FAILURE );
    }
    g_pkt_input = pcap_open_offline( argv[first_file], errbuf );
    if ( g_pkt_input == NULL ) {
        skAppPrintErr( "unable to open input file %s: %s", argv[first_file],
                       errbuf );
        exit( EXIT_FAILURE );
    }

    /*
    ** Open output file (tcpdump format packet file)
    */
    if ( FILEIsATty( stdout ) ) {
        skAppPrintErr( "stdout is connected to a terminal" );
        exit( EXIT_FAILURE );
    }

    g_output = pcap_open_dead( pcap_datalink( g_pkt_input ), 
                               pcap_snapshot( g_pkt_input ) );
    if ( g_output == NULL ) {
        skAppPrintErr( "error opening stdout: %s", errbuf );
        exit( EXIT_FAILURE );
    }

    g_output_dumper = pcap_dump_open( g_output, "-" );
    if ( g_output_dumper == NULL ) {
        skAppPrintErr( "error opening stdout: %s", pcap_geterr( g_output ) );
        exit( EXIT_FAILURE );
    }

    if ( atexit( appTeardown ) < 0 ) {
        skAppPrintErr( "unable to register appTeardown() with atexit()" );
        appTeardown();
        exit( EXIT_FAILURE );
    }

    return; /* OK */
}

    

/*
 *  appOptionsHandler:
 *      called for each option the app has registered.
 *  Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 *  Returns:
 *      0 if OK. 1 else.
 *  Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData UNUSED( cData ),
    int opt_index,
    char *opt_arg )
{
    switch ( opt_index ) {
      default:
        break;
    }
    return 0; /* OK */
}


/*
 *  appTeardown:
 *      Teardown all used modules and all application stuff.
 *  Arguments: None
 *  Returns: None
 *  Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Static variable teardownFlag is set.
 *  NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(
    void )
{
    static uint8_t teardownFlag = 0;

    if ( teardownFlag ) {
        return;
    }
    teardownFlag = 1;
    optionsTeardown();

    /*
    ** Close all files
    */
    pcap_close( g_pkt_input );
    pcap_dump_close( g_output_dumper );
    pcap_close( g_output );

    return;
}



int main( 
    int argc, 
    char **argv )
{ 
    int offset = 0;
    struct pcap_pkthdr next_pkt_hdr;
    const u_char *next_pkt_data = NULL;

    appSetup( argc, argv );


    do {
        
        next_pkt_data = pcap_next( g_pkt_input, &next_pkt_hdr );
        if ( next_pkt_data != NULL ) {
            next_pkt_hdr.ts.tv_usec += offset % 5000;
            pcap_dump( (u_char *) g_output_dumper, &next_pkt_hdr, 
                       next_pkt_data );
            offset += 1000;
        }
    } while ( next_pkt_data != NULL );

    return 0;
}
