/*
** Copyright (C) 2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwpmatch.c
**
**  Filter a tcpdump file by outputting only packets whose 5-tuple and
**  timestamp match corresponding flows in a rw-file.  Outputs the
**  filtered tcpdump file to stdout.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwpmatch/rwpmatch.c,v 1.10 2005/11/16 19:22:54 leechr Exp $");

#include <pcap.h>
#include "rwpack.h"
#include "utils.h"
#include "rwppacketheaders.h"

#define USAGE_FH stdout


/* LOCAL FUNCTIONS */

static void appTeardown(
    void );
static void appSetup(
    int argc, 
    char **argv );

static void appUsage(
    void );
static void appUsageLong(
    void );
static void appOptionsUsage(
    FILE* );
static int  appOptionsHandler(
    clientData  cData, 
    int         opt_index, 
    char       *opt_arg );

static int timecmp_pkt_flow( 
    struct timeval *pkt, 
    rwRec *flow );
static int tuples_match( 
    struct pcap_pkthdr *hdr,
    const u_char *pkt,
    rwRec *flow );


/* LOCAL VARIABLES */

static pcap_t *g_pkt_input = NULL;
static rwIOStructPtr g_flow_input = NULL;
static pcap_t *g_output;
static pcap_dumper_t *g_output_dumper = NULL;


/* OPTION SETUP */

typedef enum {
    RWPMATCH_OPT_FLOWFILE
} appOptionsEnum;


static struct option appOptions[] = {
    { "flow-file", REQUIRED_ARG, 0, RWPMATCH_OPT_FLOWFILE },
    { 0, 0, 0, 0 } /* sentinal entry */
};


static const char *appHelp[] = {
    "Flow file to use to filter the tcpdump data",
    (char *)NULL
};


/*
 *  appUsage:
 *      print usage information to stderr and exit with code 1
 *  Arguments: None
 *  Returns: None
 *  Side Effects: exits application.
 */
static void appUsage( 
    void )
{
    fprintf( USAGE_FH, "Use `%s --help' for usage\n", skAppName() );
    exit( EXIT_FAILURE );
}


/*
 *  appUsageLong:
 *      print usage information to USAGE_FH and exit with code 1.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 *  Arguments: None
 *  Returns: None
 *  Side Effects: exits application.
 */
static void appUsageLong( 
    void )
{
    fprintf( USAGE_FH, "Usage: %s [APPLICATION SWITCHES] [FILE]\n", 
             skAppName() );
    appOptionsUsage( USAGE_FH );
    exit( EXIT_FAILURE );
}


/*
 *  appOptionsUsage:
 *      print help for options for this app to the specified file handle.
 *   Arguments: None.
 *   Returns: None.
 *   Side Effects: None.
 *   NOTE:
 *       do NOT add non-option usage here (i.e., required and other
 *       optional args)
 */
static void appOptionsUsage(
    FILE *fh )
{
    int i;

    fprintf( fh, "\nAPPLICATION SWITCHES control input and output\n\n" );
    for ( i = 0; appOptions[i].name; i++ ) {
        fprintf( fh, "--%s %s. %s\n", appOptions[i].name,
                 SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i] );
    }
}


/*
 *  appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 *  Arguments:
 *      argc, argv
 *  Returns: None.
 *  Side Effects:
 *      exits the application with code 1 if anything does not work.
 */
static void appSetup( 
    int argc, 
    char **argv )
{
    char errbuf[PCAP_ERRBUF_SIZE];

    int first_file;

    skAppRegister( argv[0] );

    assert( ( sizeof( appHelp ) / sizeof( char* ) ) ==
            ( sizeof( appOptions ) / sizeof( struct option ) ) );

    if ( optionsSetup( appUsageLong ) ) {
        skAppPrintErr( "unable to setup options module" );
        exit( EXIT_FAILURE ); /* never returns */
    }

    /* set permissions for output files */
    (void) umask( (mode_t) 0002 );

    if ( optionsRegister( appOptions, (optHandler) appOptionsHandler, NULL ) ) {
        skAppPrintErr( "unable to register application options" );
        exit( EXIT_FAILURE );
    }

    /* parse options */
    if ( ( first_file = optionsParse( argc, argv ) ) < 0 ) {
        appUsage(); /* never returns */
    }

    if ( g_flow_input == NULL ) {
        skAppPrintErr( "--flow-file=FILE flag required" );
        exit( EXIT_FAILURE );
    }

    /*
    ** Open input files
    */
    if ( first_file >= argc ) {
        skAppPrintErr( "no input packet file found" );
        exit( EXIT_FAILURE );
    }
    g_pkt_input = pcap_open_offline( argv[first_file], errbuf );
    if ( g_pkt_input == NULL ) {
        skAppPrintErr( "unable to open input file %s: %s", argv[first_file],
                       errbuf );
        exit( EXIT_FAILURE );
    }

    /*
    ** Open output file (tcpdump format packet file)
    */
    if ( FILEIsATty( stdout ) ) {
        skAppPrintErr( "stdout is connected to a terminal" );
        exit( EXIT_FAILURE );
    }

    g_output = pcap_open_dead( pcap_datalink( g_pkt_input ), 
                               pcap_snapshot( g_pkt_input ) );
    if ( g_output == NULL ) {
        skAppPrintErr( "error opening stdout: %s", errbuf );
        exit( EXIT_FAILURE );
    }

    g_output_dumper = pcap_dump_open( g_output, "-" );
    if ( g_output_dumper == NULL ) {
        skAppPrintErr( "error opening stdout: %s", pcap_geterr( g_output ) );
        exit( EXIT_FAILURE );
    }

    if ( atexit( appTeardown ) < 0 ) {
        skAppPrintErr( "unable to register appTeardown() with atexit()" );
        appTeardown();
        exit( EXIT_FAILURE );
    }

    return; /* OK */
}

    

/*
 *  appOptionsHandler:
 *      called for each option the app has registered.
 *  Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 *  Returns:
 *      0 if OK. 1 else.
 *  Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData UNUSED( cData ),
    int opt_index,
    char *opt_arg )
{
    switch ( (appOptionsEnum)opt_index ) {
      case RWPMATCH_OPT_FLOWFILE:
        /* the input-pipe must be 'stdin' or it must be an existing
         * FIFO.  If 'stdin', stdin must not be a TTY, since we expect
         * binary. */
        if ( strcmp( opt_arg, "stdin" ) == 0 ) {
            if ( FILEIsATty( stdin ) ) {
                skAppPrintErr( "stdin is connected to a terminal." );
                return -1;
            }
        }

        g_flow_input = rwOpenFile( opt_arg, NULL );
        if ( g_flow_input == NULL ) {
            exit( EXIT_FAILURE );
        }
        break;
    }

    return 0; /* OK */
}


/*
 *  appTeardown:
 *      Teardown all used modules and all application stuff.
 *  Arguments: None
 *  Returns: None
 *  Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Static variable teardownFlag is set.
 *  NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(
    void )
{
    static uint8_t teardownFlag = 0;

    if ( teardownFlag ) {
        return;
    }
    teardownFlag = 1;
    optionsTeardown();

    /*
    ** Close all files
    */
    pcap_close( g_pkt_input );
    pcap_dump_close( g_output_dumper );
    pcap_close( g_output );
    rwCloseFile( g_flow_input );

    return;
}



int main( 
    int argc, 
    char **argv )
{ 
    /*
    ** Flag to tell whether or not we should continue processing
    ** files.  0 if processing should continue, 1 if packet flow
    ** reached EOF, 2 if flow file reached EOF.  -1 on an error 
    ** condition. XXX
    */
    int done = 0;

    /*
    ** Flag to determine if the latest packet data has been consumed,
    ** and a new packet must be read from file.
    */
    int load_next_pkt = 1;
    /*
    ** Latest packet read from file.
    */
    struct pcap_pkthdr next_pkt_hdr;
    const u_char *next_pkt_data = NULL;

    /*
    ** Flag to determine if the latest flow data has been consumed,
    ** and a new flow must be read from file.
    */
    int load_next_flow = 1;

    /*
    ** Latest flow read from file.
    */
    rwRec next_flow;

    /*
    ** -1 if pkt < flow, 0 if pkt == flow, 1 if pkt > flow
    */
    int ts_compare;

    appSetup( argc, argv );


    while ( !done ) {
        /*
        ** If the current packet data is stale, load the next packet
        */
        if ( load_next_pkt ) {
            load_next_pkt = 0;
            next_pkt_data = pcap_next( g_pkt_input, &next_pkt_hdr );
            if ( next_pkt_data == NULL ) {
                done = 1;
            }
        }

        /*
        ** If the current flow data is stale, load the next flow
        */
        if ( load_next_flow ) {
            load_next_flow = 0;
            if ( !rwRead( g_flow_input, &next_flow ) ) {
                done = 2;
                continue; /* drops out of while loop */
            }
        }

        /*
        ** At this point, we are guaranteed that both the packet and
        ** flow data is fresh, so compare the records.
        */

        /*
        ** If the current packet matches the flow characteristics
        ** (5-tuple and timestamp), output the packet.
        */
        ts_compare = timecmp_pkt_flow( &next_pkt_hdr.ts, &next_flow );
        if ( ts_compare == 0 ) {
            if ( tuples_match( &next_pkt_hdr, next_pkt_data, &next_flow ) ) {
                pcap_dump( (u_char *) g_output_dumper, &next_pkt_hdr, 
                           next_pkt_data );
                load_next_pkt = 1;
                load_next_flow = 1;
                continue;
            }
        }

        /*
        ** The packet and flow didn't match.
        */
        if ( ts_compare < 0 ) {
            /*
            ** If the packet data has an earlier timestamp (or equal,
            ** but the 5-tuple didn't match), then the current packet
            ** does not exist in the flow file, so it can be skipped.
            ** Keep the same flow record, and try the next packet.
            */
            load_next_pkt = 1;
        } else {
            /*
            ** If the packet data has a later timestamp, then the flow
            ** contains an entry which was not derived from this
            ** packet data.  Report an error, and exit.
            */
            skAppPrintErr( "found a flow which does not have "
                           "corresponding packet data.  Aborting." );
            exit( EXIT_FAILURE );
        }
    }

    return 0;
}


/*
** returns:
**   -1 if packet is earlier than flow
**   0 if packet and flow occurred at the same time
**   1 if packet is later than flow
*/
static int timecmp_pkt_flow( 
    struct timeval *pkt, 
    rwRec *flow )
{
    uint32_t pkt_sec = pkt->tv_sec;

    /* should never get negative entries in timeval, but better safe
       than sorry */
    assert( pkt->tv_sec >= 0 );
    assert( pkt->tv_usec >= 0 );

    /* compare seconds */
    if ( pkt_sec < flow->sTime ) {
        return -1;
    } else if( pkt_sec > flow->sTime ) {
        return 1;
    } else {
#if 0 /* when msec support is in, compile this section */
        uint32_t pkt_msec_unrounded = pkt->tv_usec / 1000;
        uint16_t pkt_msec = pkt_msec_unrounded; 

        /* compare msec */
        if ( pkt_msec < flow->sTime_msec ) {
            return -1;
        } else if( pkt_msec > flow->sTime_msec ) {
            return 1;
        } else {
            return 0;
        }
#else
        return 0;
#endif
    }
}


/*
** Determines if a packet was aggregated into a flow using rwptoflow.
*/
static int tuples_match(
    struct pcap_pkthdr *hdr,
    const uint8_t *pkt,
    rwRec *flow )
{
    eth_header_t *eh;
    ip_header_t *ih;

    /*
    ** Check timestamp
    */
    if ( timecmp_pkt_flow( &hdr->ts, flow ) != 0 ) {
        return 0;
    }
    
    /*
    ** compare ethernet header
    */         
    eh = (eth_header_t *) pkt;
    if ( ntohs( eh->ether_type ) != ETHERTYPE_IP ) {
        return 0;
    }

    /* FUTURE: we don't store the ethernet identification field in the
       flow record, so we can't compare it when going backwards.  we
       might want to consider storing it in the future. */

    /* retrieve the position of the ip header (add length of
       ethernet header) */
    ih = (ip_header_t *) ( pkt + sizeof ( eth_header_t ) ); 

    /* compare size, saddr, daddr, proto */
    if ( ( ih->saddr.ipnum != htonl( flow->sIP.ipnum ) ) ||
         ( ih->daddr.ipnum != htonl( flow->dIP.ipnum ) ) ||
         ( ih->proto != flow->proto ) ) 
    {
        return 0;
    }
        
    return 1;    
}
