/*
** Copyright (C) 2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwptoflow.c
**
** Generates a flow for every IP packet.  Since IP packets can arrive
** out of order, though, some fragments are collapsed into a single
** flow.  (In particular, all fragments before the "zero" fragment are
** lumped into the "zero" fragment's flow.  Later fragments are output
** as their own flows.  We do this so that we can add OSI layer 4
** information to the flows we generate, like source and destination
** ports.)
**
** Future development:
**
**   Add mark-as-deleted call to hashlib so that we can purge the hash
**   table occasionally.  Currently, the hash table will grow until
**   system memory is exhausted.
**
**   In the event that the zero fragment is too small to contain TCP
**   flags, attempt to get them from the next fragment.  This will
**   require more sophisticated fragment reassembly.
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwpmatch/rwptoflow.c,v 1.19 2005/11/18 22:02:30 leechr Exp $");

#include <pcap.h>
#include "hashlib.h"
#include "rwpack.h"
#include "utils.h"
#include "rwppacketheaders.h"
#include "rwptoflow.h"

#define USAGE_FH stdout

/* LOCAL FUNCTIONS */

static void appSetup(
    int argc,
    char **argv );
static void appTeardown(
    void );


static void appUsage(
    void );
static void appUsageLong(
    void );
static void appOptionsUsage(
    FILE* );
static int  appOptionsHandler(
    clientData  cData,
    int         opt_index,
    char       *opt_arg );


static int handle_frag( 
    HashTable *fragments,
    rwRec *flow,
    struct pcap_pkthdr *hdr,
    const uint8_t *data );


/* LOCAL VARIABLES */

static rwptoflow_netflow_t g_netflow;
static pcap_t *g_input;
static rwIOStruct *g_output = NULL;

/* OPTION SETUP */

typedef enum {
    RWPTOFLOW_OPT_SENSORID,
    RWPTOFLOW_OPT_INPUTINDEX,
    RWPTOFLOW_OPT_OUTPUTINDEX,
    RWPTOFLOW_OPT_NEXTHOPIP
} appOptionsEnum;


static struct option appOptions[] = {
    {"set-sensorid", REQUIRED_ARG, 0, RWPTOFLOW_OPT_SENSORID},
    {"set-inputindex", REQUIRED_ARG, 0, RWPTOFLOW_OPT_INPUTINDEX},
    {"set-outputindex", REQUIRED_ARG, 0, RWPTOFLOW_OPT_OUTPUTINDEX},
    {"set-nexthopip", REQUIRED_ARG, 0, RWPTOFLOW_OPT_NEXTHOPIP},
    {0,0,0,0} /* sentinal entry */
};


static const char *appHelp[] = {
    "set the sensor id for all flows. Default 0.",
    "set the input index for all flows. Default 0.",
    "set the output index for all flows. Default 0.",
    "set the next hop ip address for all flows.\n  Default 0.0.0.0.",
    (char *)NULL
};


/*
 *  appUsage:
 *      print usage information to stderr and exit with code 1
 *  Arguments: None
 *  Returns: None
 *  Side Effects: exits application.
 */
static void appUsage(
    void )
{
    fprintf( USAGE_FH, "Use `%s --help' for usage\n", skAppName() );
    exit( EXIT_FAILURE );
}


/*
 *  appUsageLong:
 *      print usage information to USAGE_FH and exit with code 1.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 *  Arguments: None
 *  Returns: None
 *  Side Effects: exits application.
 */
static void appUsageLong(
    void )
{
#define USAGE_MSG                                                        \
    ("Usage: %s [APPLICATION SWITCHES] FILE\n"                           \
     "\tRead a TCPDUMP file and generate a SiLK Flow record for every\n" \
     "\tpacket.  Write the SiLK Flows to stdout.\n")

    fprintf( USAGE_FH, USAGE_MSG, skAppName() );
    appOptionsUsage( USAGE_FH );
    exit( EXIT_FAILURE );
}


/*
 *  appOptionsUsage:
 *      print help for options for this app to the specified file handle.
 *   Arguments: None.
 *   Returns: None.
 *   Side Effects: None.
 *   NOTE:
 *       do NOT add non-option usage here (i.e., required and other
 *       optional args)
 */
static void appOptionsUsage(
    FILE *fh )
{
    int i;

    fprintf( fh, "\nAPPLICATION SWITCHES control input and output\n\n" );
    for ( i = 0; appOptions[i].name; i++ ) {
        fprintf( fh, "--%s %s. %s\n", appOptions[i].name,
                 SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i] );
    }
}


/*
 *  appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 *  Arguments:
 *      argc, argv
 *  Returns: None.
 *  Side Effects:
 *      exits the application with code 1 if anything does not work.
 */
static void appSetup(
    int argc,
    char **argv )
{
    int rv;
    char errbuf[PCAP_ERRBUF_SIZE];
    int first_file;

    skAppRegister( argv[0] );

    assert( ( sizeof( appHelp ) / sizeof( char* ) ) ==
            ( sizeof( appOptions ) / sizeof( struct option ) ) );

    if ( optionsSetup( appUsageLong ) ) {
        skAppPrintErr( "unable to setup options module" );
        exit( EXIT_FAILURE ); /* never returns */
    }

    memset( &g_netflow, 0, sizeof( rwptoflow_netflow_t ) );

    /* set permissions for output files */
    (void) umask( (mode_t) 0002 );

    if ( optionsRegister( appOptions, (optHandler) appOptionsHandler, NULL )) {
        skAppPrintErr( "unable to register application options" );
        exit( EXIT_FAILURE );
    }

    /* parse options */
    if ( ( first_file = optionsParse( argc, argv ) ) < 0 ) {
        appUsage(); /* never returns */
    }

    /*
    ** Open input files
    */
    if ( argc - first_file != 1 ) {
        skAppPrintErr( "must have one input file" );
        appUsage(); /* never returns */
    }
    
    g_input = pcap_open_offline( argv[first_file], errbuf );
    if ( g_input == NULL ) {
        skAppPrintErr( "error opening input %s: %s", argv[first_file],
                       errbuf );
        exit( EXIT_FAILURE );
    }

    /*
    ** Open output file (rw-file which contains single packet flows
    */
    if ( ( rv = rwioCreate( &g_output, "stdout", SK_RWIO_WRITE ) )
         || ( rv = rwioSetFileType( g_output, FT_RWFILTER ) )
         || ( rv = rwioOpen( g_output ) )
         || ( rv = rwioAppendToHistoryArgv( g_output, argc, argv ) )
         || ( rv = rwioWriteHeader( g_output ) ) )
    {
        rwioPrintLastErr( g_output, rv, &skAppPrintErr );
        rwioDestroy( &g_output );
        exit( EXIT_FAILURE );
    }

    if ( atexit( appTeardown ) < 0 ) {
        skAppPrintErr( "unable to register appTeardown() with atexit()" );
        appTeardown();
        exit( EXIT_FAILURE );
    }

    return; /* OK */
}



/*
 *  appOptionsHandler:
 *      called for each option the app has registered.
 *  Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 *  Returns:
 *      0 if OK. 1 else.
 *  Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData UNUSED( cData ),
    int opt_index,
    char *opt_arg )
{
    int rv;
    uint32_t temp;

    switch ( (appOptionsEnum)opt_index ) {
      case RWPTOFLOW_OPT_SENSORID:
        rv = skStringParseUint32( &temp, opt_arg, 0, 0 );
        if( rv < 0 || temp > UINT8_MAX ) {
            skAppPrintErr( "invalid index in --set-inputindex" );
            return 1;
        }
        g_netflow.sensorID = (uint8_t) temp;
        break;

      case RWPTOFLOW_OPT_INPUTINDEX:
        rv = skStringParseUint32( &temp, opt_arg, 0, 0 );
        if( rv < 0 || temp > UINT8_MAX ) {
            skAppPrintErr( "invalid index in --set-inputindex" );
            return 1;
        }
        g_netflow.inputIndex = (uint8_t) temp;
        break;

      case RWPTOFLOW_OPT_OUTPUTINDEX:
        rv = skStringParseUint32( &temp, opt_arg, 0, 0 );
        if( rv < 0 || temp > UINT8_MAX ) {
            skAppPrintErr( "invalid index in --set-outputindex" );
            return 1;
        }
        g_netflow.outputIndex = (uint8_t) temp;
        break;

      case RWPTOFLOW_OPT_NEXTHOPIP:
        g_netflow.nextHopIP.ipnum = dot2num( opt_arg );
        if( g_netflow.nextHopIP.ipnum == UINT32_MAX ) {
            skAppPrintErr( "invalid IP address in --set-nexthopip" );
            return 1;
        }
        break;
    }

    return 0; /* OK */
}


/*
 *  appTeardown:
 *      Teardown all used modules and all application stuff.
 *  Arguments: None
 *  Returns: None
 *  Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Static variable teardownFlag is set.
 *  NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(
    void )
{
    static uint8_t teardownFlag = 0;

    if ( teardownFlag ) {
        return;
    }
    teardownFlag = 1;
    optionsTeardown();

    /*
    ** Close all files
    */

    /* inputs */
    pcap_close( g_input );

    /* output */
    rwCloseFile( g_output );

    return;
}



int main(
    int argc,
    char **argv )
{
    HashTable *fragtable;
    const int fragtable_initial_size = 1000;
    uint8_t *fragtable_no_value;
    
    struct pcap_pkthdr hdr;
    const u_char *data;
    
#if 0
    /* disabled, pending addition of a hash table delete function */
    
    /* every N records read, purge deleted records and rehash.  the
       size of a key/value pair is currently 24 bytes, so use that as
       a rough guide to how often you want to purge old records. */
    const int rehash_max = 1000000;
    int est_cur_size = 0;
#endif

    appSetup( argc, argv );

    /*
    ** initialize fragment dictionary
    */
    /* Allocate memory for and initialize special "empty" value */
    fragtable_no_value = (uint8_t *) malloc( sizeof ( dupe_hash_value_t ) );
    if ( fragtable_no_value == NULL ) {
        skAppPrintErr( "memory error" );
        exit( EXIT_FAILURE );
    }
    memset( fragtable_no_value, 0xFF, sizeof( dupe_hash_value_t ) );
    
    /* create fragment dictionary */
    fragtable = hashlib_create_table( sizeof ( dupe_hash_key_t ),
                                      sizeof ( dupe_hash_value_t ),
                                      HTT_INPLACE, /* values, not pointers */
                                      fragtable_no_value,
                                      NULL, 0, /* No user data */
                                      fragtable_initial_size, 
                                      DEFAULT_LOAD_FACTOR );
    
    while( NULL != ( data = pcap_next( g_input, &hdr ) ) ) {
        rwRec flow;
        memset( &flow, 0, sizeof ( rwRec ) );
        if ( handle_frag( fragtable, &flow, &hdr, data ) ) {
            /* write record to rw file */
            rwWrite( g_output, &flow );
        }

#if 0
        /* disabled, pending addition of hashlib_mark_deleted() */
        
        /*
        ** Occasionally purge old records from the table
        */
        ++est_cur_size;
        if ( est_cur_size > rehash_max ) {
            struct timeval timeout_ts;
            
            HASH_ITER iter;
            dupe_hash_key_t *key;
            dupe_hash_value_t *val;
            
            /*
            ** assume we don't underflow on time here, since timestamps
            ** should be more than a half an hour past epoch.
            */
            timeout_ts.tv_sec = hdr.ts.tv_sec - 1800;
            timeout_ts.tv_usec = hdr.ts.tv_usec;
            
            iter = hashlib_create_iterator( fragtable );
            while ( hashlib_iterate( fragtable, &iter, (uint8_t **) &key, 
                                     (uint8_t **) &val )
                    != ERR_NOMOREENTRIES )
            {
                if ( timercmp( &val->ts, &timeout_ts, < ) ) {
                    if ( hashlib_mark_deleted( fragtable, (uint8_t *) key ) 
                         != 0 ) 
                    {
                        skAppPrintErr( "error deleting from fragment table" );
                        exit( EXIT_FAILURE );
                    }
                }
            }
            if ( hashlib_rehash( fragtable ) != OK ) {
                skAppPrintErr( "error rehashing fragment table" );
                exit( EXIT_FAILURE );
            }
            est_cur_size = hashlib_count_entries( fragtable );
        }
#endif

    }
    
    hashlib_free_table( fragtable );
    free( fragtable_no_value );
    
    return 0;
}


/*
** 'fragments' is a hash table containing cached OSI layer 4
** information, so that we can properly output that information in
** flows.  it also contains aggregated byte/packet counts for packets
** for which we have not yet located the first fragment.
**
** 'flow' is where data is stored for output.  
**
** 'hdr' and 'data' contain the packet as read by pcap
**
** returns 1 if 'flow' should be output.  returns 0 if 'flow' should
** not be output.
*/
static int handle_frag( 
    HashTable *fragments,
    rwRec *flow,
    struct pcap_pkthdr *hdr,
    const uint8_t *data )
{
    /* pointer to the ethernet header inside of 'data' */
    eth_header_t *eh;
    /* pointer to the IP header inside of 'data' */
    ip_header_t *ih;
    /* the actual length of the IP header */
    uint32_t ih_len;

    /* the key to the fragments hash table */
    dupe_hash_key_t curkey;
    /* the current context, as retrieved from the fragments hash
       table.  if there was no entry for the current key, then a new
       context is created. */
    dupe_hash_value_t *context;
    int hash_rv;

    uint16_t stored_sport = 0;
    uint16_t stored_dport = 0;
    uint8_t stored_flags = 0;
    uint16_t stored_bytes;
    uint16_t stored_pkts;
    int f_current_is_zero_packet;

    /* make sure the packet has (and we captured) at least an ethernet
       and minimum IP header.  that gives us enough data to read the
       actual IP header size. */
    if ( hdr->caplen < ( sizeof ( eth_header_t ) + 
                         sizeof ( ip_header_t ) ) ||
         hdr->len < ( sizeof ( eth_header_t ) + 
                      sizeof ( ip_header_t ) ) )
    {
        skAppPrintErr( "dropping truncated packet, length %d "
                       "(%d bytes captured)", 
                       hdr->len, hdr->caplen );
        return 0;
    }
    
    /* get the ethernet header */
    eh = (eth_header_t *) data;
    if ( ntohs( eh->ether_type ) != ETHERTYPE_IP ) {
        /* ignoring non IP packet */
        return 0;
    }
    
    /* retrieve the position of the ip header (add length of
       ethernet header) */
    ih = (ip_header_t *) ( data + sizeof ( eth_header_t ) ); 
    
    /* retrieve the full size of the IP header. */
    ih_len = ( ih->ver_ihl & 0xf ) * 4;
    
    /*
    ** If the current fragment is the zero packet, get all the OSI
    ** layer 4 information out of it.  (Since a fragment must contain
    ** 8 bytes of information, and sport, dport, type and code all
    ** occur in the first 8 bytes of the packet, we only need to look
    ** for the first fragment.)
    */
    if ( ntohs( ih->flags_fo & IPHEADER_FO_MASK ) ) {
        /* we are the first fragment, so set port information */
        uint8_t *payload = ( (uint8_t *) ih ) + ih_len;
        icmp_header_t *icmphdr;
        tcp_header_t *tcphdr;
        udp_header_t *udphdr;

        /* make sure the packet has (and we captured) at least the
           first 8 bytes of payload.  if we have that, then we can
           extract ports. */
        if ( hdr->caplen < ( sizeof ( eth_header_t ) + 
                             sizeof ( ip_header_t ) + 8 ) || 
             hdr->len < ( sizeof ( eth_header_t ) +
                          sizeof ( ip_header_t ) + 8 ) )
        { 
            skAppPrintErr( "dropping truncated packet, length %d "
                           "(%d bytes captured)", 
                           hdr->len, hdr->caplen );
            return 0; 
        }

        
        switch ( ih->proto ) {
          case 1: /* ICMP */
            icmphdr = (icmp_header_t *) payload;
            stored_sport = 0;
            stored_dport = ( ( icmphdr->type << 8 ) | icmphdr->code );
            stored_flags = 0;
            break;
            
          case 6: /* TCP */

            /* see if we have enough of the TCP header to get the TCP
               flags.  If so, store them.  If not, set them to zero,
               but print a warning message. 

               NOTE: we are only guaranteed 8 bytes in the
               zero-fragment.  So, a legitimately small packet could
               fail this test.  (It is not indicative of malformed
               traffic.)  If it turns out that field data encounters
               this case, more complete fragment reassembly must be
               done. 
            */
            tcphdr = (tcp_header_t *) payload;
            if ( hdr->caplen < ( sizeof ( eth_header_t ) + 
                                 sizeof ( ip_header_t ) + 14 ) || 
                 hdr->len < ( sizeof ( eth_header_t ) +
                              sizeof ( ip_header_t ) + 14 ) )
            { 
                stored_flags = 0;
                skAppPrintErr( "Could not retrieve TCP flags from small"
                               " fragment.  TCP flags in the resulting flow"
                               " are invalid." );

            } else {
                stored_flags = ntohs( tcphdr->o_r_e ) & 0xFF;
            }

            stored_sport = ntohs( tcphdr->sport );
            stored_dport = ntohs( tcphdr->dport );
            break;
            
          case 17: /* UDP */
            udphdr = (udp_header_t *) payload;
            stored_sport = ntohs( udphdr->sport );
            stored_dport = ntohs( udphdr->dport );
            stored_flags = 0;

          default:
            stored_sport = 0;
            stored_dport = 0;
            stored_flags = 0;
        }
        f_current_is_zero_packet = 1;
    } else {
        f_current_is_zero_packet = 0;
    }
    stored_bytes = hdr->len - sizeof ( eth_header_t );
    stored_pkts = 1;

    /*
    ** Build the hash table lookup key for the current packet
    */
    curkey.sip.ipnum = ntohl( ih->saddr.ipnum );
    curkey.dip.ipnum = ntohl( ih->daddr.ipnum );
    curkey.id = ntohs( ih->identification );
    curkey.proto = ih->proto;

    /*
    ** Check for existing fragment entry in hash table
    */
    hash_rv = hashlib_lookup( fragments, (uint8_t *) &curkey, 
                              (uint8_t **) &context );
    switch ( hash_rv ) {
      case ERR_NOTFOUND:
        /* allocate the new context */
        context = (dupe_hash_value_t *) malloc( sizeof ( dupe_hash_value_t ) );
        memset( context, 0, sizeof ( dupe_hash_value_t ) );
        if ( context == NULL ) {
            skAppPrintErr( "memory allocation error" );
            exit( EXIT_FAILURE );
        }

        /* 
        ** initialize the new context 
        */
        memcpy( &context->ts, &hdr->ts, sizeof ( struct timeval ) );
        if ( f_current_is_zero_packet ) {
            context->val.data.sport = stored_sport;
            context->val.data.dport = stored_dport;
            context->val.data.flags = stored_flags;
            context->f_zero_packet_seen = 1;
        } else {
            context->f_zero_packet_seen = 0;
        }

        /*
        ** add the new context to the fragments table
        */
        hash_rv = hashlib_insert( fragments, (uint8_t *) &curkey, 
                                  (uint8_t **) &context );
        if ( hash_rv < 0 ) {
            skAppPrintErr( "error %d inserting into hash table", hash_rv );
            exit( EXIT_FAILURE );
        }

        /* FALLTHROUGH!  now that we have a context, we want to
           pretend we looked it up in the fragment table. */


      case OK:
        /* we either looked up or generated a valid context, so now we
           determine if we should ouput, and if so, set the flow
           variables */

        if ( !context->f_zero_packet_seen && !f_current_is_zero_packet ) {
            /* we still have no OSI layer 4 information.  store the
               byte and packet counts and hope we get a zero packet
               later. */
            context->val.count.oct += stored_bytes;
            context->val.count.oct += stored_pkts;
            /* do not set up the output flow record, and tell the
               caller that it should not attempt to output the flow
               record it passed in. */
            return 0;
        }

        /* if the context packet was not the zero packet, then we want
           to get its counters and then turn it into the zero-packet
           entry. */
        if ( !context->f_zero_packet_seen ) {
            stored_bytes += context->val.count.oct;
            stored_pkts += context->val.count.pkt;
            context->val.data.sport = stored_sport;
            context->val.data.dport = stored_dport;
            context->val.data.flags = stored_flags;
        }


        /*
        ** set up the flow record 
        */
        flow->sTime = hdr->ts.tv_sec; 
        flow->sTime_msec = hdr->ts.tv_usec / 1000; 
        flow->elapsed = 0; 
        flow->elapsed_msec = 0; 
        
        flow->pkts = stored_pkts;
        flow->bytes = stored_bytes;
        flow->proto = ih->proto;
        
        flow->sIP.ipnum = ntohl( ih->saddr.ipnum );
        flow->dIP.ipnum = ntohl( ih->daddr.ipnum );
        flow->sPort = stored_sport;
        flow->dPort = stored_dport;
        flow->flags = stored_flags;

        /* set netflow constants */
        flow->sID = (sensorID_t) g_netflow.sensorID;
        flow->nhIP = g_netflow.nextHopIP;
        flow->input = g_netflow.inputIndex;
        flow->output = g_netflow.outputIndex;
        break;


      default:
        skAppPrintErr( "error %d during hash table lookup", hash_rv );
        exit( EXIT_FAILURE );
    }

    return 1;
}
