/*
** Copyright (C) 2003-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
 * rwrandomizeip
 *
 * Read any rw file (rwpacked file, rwfilter output, etc) and output a
 * file with the source IPs and destination IPs randomly put into
 * 10.0.0.0/8, 172.16.0.0/12, or 192.168.0.0/16 space.
 *
 * Though the IPs are gone, the port numbers remain.  These randomized
 * files could provide some information to a malicious party, e.g.,
 * letting them know that a particular service is in use.
 *
 * TODO:
 *
 * --It would be nice if the user could optionally provide the cidr
 * block into which source and/or destination IPs should be placed.
 *
 * --Randomize the ports.
 *
 */

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwrandomizeip/rwrandomizeip.c,v 1.17 2005/10/26 18:26:08 thomasm Exp $");

#include "utils.h"
#include "rwpack.h"


/* TYPEDEFS AND DEFINES */

/* File handle for --help output */
#define USAGE_FH stdout

#define OCTETS_PER_IP 4
#define VALUES_PER_OCTET 256


/* LOCAL FUNCTIONS */

static void appUsage(void);                     /* never returns */
static void appUsageLong(void);                 /* never returns */
static void appOptionsUsage(FILE *fh);          /* application options usage */
static void appTeardown(void);
static void appSetup(int argc, char **argv);    /* never returns on failure */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static int  randomizeFile(const char *in, const char *out);
static void randomizeIP(uint32_t *ip);
static void consistentlyRandomizeIP(uint32_t *ip);
static void createShuffleTable(void);
static int  loadShuffleFile(const char *shuffleFileName);
static int  saveShuffleFile(const char *shuffleFileName);


/* LOCAL VARIABLES */

/* input and output file names */
static const char *inFName;
static const char *outFName;

/* whether to map ip addresses to a consistent random address */
static uint8_t consistentFlag = 0;

/* when mapping consistently, the table to use for the mapping */
static uint8_t ipShuffleTable[OCTETS_PER_IP][VALUES_PER_OCTET];


/* OPTIONS */

enum {
    RWRAND_OPT_CONSISTENT, RWRAND_OPT_LOAD_TABLE, RWRAND_OPT_SAVE_TABLE
};

static struct option appOptions[] = {
    {"consistent", NO_ARG,       0, RWRAND_OPT_CONSISTENT},
    {"load-table", REQUIRED_ARG, 0, RWRAND_OPT_LOAD_TABLE},
    {"save-table", REQUIRED_ARG, 0, RWRAND_OPT_SAVE_TABLE},
    {0,0,0,0}  /* sentinel entry */
};

static char *appHelp[] = {
    "Consistently randomize IP addresses",
    "Load up a randomization table from a previous run",
    "Save this run's randomization table for future use",
    (char *)NULL
};


/*
 * appUsage:
 *      print message on getting usage information and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 *      print usage information to USAGE_FH and exit with code 1.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
    fprintf(USAGE_FH,
            ("%s <input-file> <output-file>\n"
             "\t Substitute a non-routable IP address for the source\n"
             "\t and destination IP addresses of <input-file> and\n"
             "\t write result to <output-file>.\n"
             "\t You may use \"stdin\" for <input-file> and \"stdout\"\n"
             "\t for <output-file>.  Gzipped files are o.k.\n"),
            skAppName());

    appOptionsUsage(USAGE_FH);
    exit(EXIT_SUCCESS);
}


/*
 * appOptionsUsage:
 *      print options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s: %s\n", appOptions[i].name, appHelp[i]);
    }

    return;
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    optionsTeardown();

    return;
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int nextOptIndex;

    skAppRegister(argv[0]);

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* initialize the options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* register the options and handler */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* parse options */
    if ((nextOptIndex = optionsParse(argc, argv)) < 0) {
        appUsage();         /* never returns */
    }

    /* Check that we have input */
    if (nextOptIndex >= argc) {
        skAppPrintErr("Expecting input file name");
        appUsage();         /* never returns */
    }
    inFName = argv[nextOptIndex];
    ++nextOptIndex;

    /* check that we have an output location */
    if (nextOptIndex >= argc) {
        skAppPrintErr("Expecting output file name");
        appUsage();         /* never returns */
    }
    outFName = argv[nextOptIndex];
    ++nextOptIndex;

    /* check for extra options */
    if (argc != nextOptIndex) {
        skAppPrintErr("Got extra options");
        appUsage();         /* never returns */
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);            /* never returns */
    }

    return;                       /* OK */
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char        *opt_arg)
{
    switch(opt_index) {
      case RWRAND_OPT_CONSISTENT:
        if(consistentFlag) {
            skAppPrintErr("Choose --load, --consistent, OR --save.");
            return 1;
        }
        consistentFlag = 1;
        createShuffleTable();
        break;

      case RWRAND_OPT_LOAD_TABLE:
        if(consistentFlag) {
            skAppPrintErr("Choose --load, --consistent, OR --save.");
            return 1;
        }
        if (opt_arg[0] == '\0') {
            skAppPrintErr("Missing file name to load from.");
            return 1;
        }
        consistentFlag = 1;
        loadShuffleFile(opt_arg);
        break;

      case RWRAND_OPT_SAVE_TABLE:
        if(consistentFlag) {
            skAppPrintErr("Choose --load, --consistent, OR --save.");
            return 1;
        }
        if (opt_arg[0] == '\0') {
            skAppPrintErr("Missing file name to save to.");
            return 1;
        }
        consistentFlag = 1;
        createShuffleTable();
        saveShuffleFile(opt_arg);
        break;

      default:
        skAppPrintErr("Unhandled option %d in options handler at %s:%d",
                      opt_index, __FILE__, __LINE__);
        abort();
    };

    return 0;                     /* OK */
}


/*
 * void createShuffleTable(void)
 *
 * createShuffleTable generates a set of consistent mappings between IP
 * addresses.  Each table maps a corresponding octet value.  To
 * actually generate the maps we start with a 0-255 table and then
 * swap each value once.
 */
static void createShuffleTable(void)
{
    int i,j, swapIndex, stash;

    /*
     * First loop is pure initialization.  Set everything so
     * ipShuffleTable[i][j] = j.
     */
    for(i = 0; i < OCTETS_PER_IP; i++) {
        for(j = 0; j < VALUES_PER_OCTET; j++) {
            ipShuffleTable[i][j] = j;
        };
    };

    for(i = 0; i < OCTETS_PER_IP; i++) {
        for(j = 0; j < VALUES_PER_OCTET; j++) {
            swapIndex = ((double)VALUES_PER_OCTET * random() / (RAND_MAX+1.0));
            stash = ipShuffleTable[i][swapIndex];
            ipShuffleTable[i][swapIndex] = ipShuffleTable[i][j];
            ipShuffleTable[i][j] = stash;
        };
    };
}



/*
 *  void consistentlyRandomizeIP(uint32_t *ip)
 *
 *  Writes a new ip address to the location specified by 'ip' using
 *  the consistent mapping built using createShuffleTable().
 */
static void consistentlyRandomizeIP(uint32_t *ip)
{
    int tgtByte, i, mask;

    for (i = 0 ; i < OCTETS_PER_IP ; i++) {
        /*
         * isolate the targeted byte.
         */
        mask = (0xFF << (i * 8));
        tgtByte = (*ip & mask) >> (i * 8);
        tgtByte = ipShuffleTable[i][tgtByte];
        *ip = (*ip & ~mask) | (tgtByte << (i * 8));
    }
}



/*
 * int saveShuffleFile(char *shuffleFileName)
 *
 * Writes the ambient shuffle file data (the contents of the
 *
 * NOTES:
 * This routine is supposed to be called immediately after shuffling and
 * before any filtering has been done.
 */
static int saveShuffleFile(const char *shuffleFileName)
{
    /*
     * we don't have anything particularly special here, so we can
     * just use the generic header.
     */
    genericHeader gHdr;
    int i;
    FILE *shuffleFile;

    /*
     * First step - check to see if the file exists and if it does
     * complain and terminate.
     */
    if(fileExists(shuffleFileName)) {
        skAppPrintErr("Shuffle file %s already exists, delete it to continue",
                      shuffleFileName);
        exit(EXIT_FAILURE);
    }

    /* open file */
    if((shuffleFile = fopen(shuffleFileName, "wb")) == NULL) {
        skAppPrintErr("Error opening file %s: %s.  Terminating.",
                      shuffleFileName, strerror(errno));
        exit(EXIT_FAILURE);
    };

    /* write header */
    PREPHEADER(&gHdr);
    gHdr.type = FT_SHUFFLE;
    gHdr.version = 0;

    if (!fwrite((void *) &gHdr, sizeof(genericHeader), 1, shuffleFile)) {
        skAppPrintErr("Error writing to file %s: %s.  Terminating",
                      shuffleFileName, strerror(errno));
        exit(EXIT_FAILURE);
    }

    /*
     * We're always writing in the form 0,1,2,3
     *
     */
    for(i = 0 ; i < OCTETS_PER_IP; i++) {
        if(!fwrite((void *) ipShuffleTable[i], 1,
                   sizeof(uint8_t) * VALUES_PER_OCTET,
                   shuffleFile)) {
            skAppPrintErr("Error writing to file %s: %s.  Terminating",
                          shuffleFileName, strerror(errno));
            exit(EXIT_FAILURE);
        };
    };

    if(fclose(shuffleFile)) {
        skAppPrintErr("Error closing file %s: %s.  Terminating",
                      shuffleFileName, strerror(errno));
        exit(EXIT_FAILURE);
    };

    return 0; /* in defiance of all expectations, everything works */
}



/*
 *  int loadShuffleFile(char *shuffleFileName)
 *
 *    Loads a shuffle file off of disk by reading the octet streams,
 *    given the endian issue, this function will swap direction of the
 *    file was read in the opposite format.
 */
static int loadShuffleFile(const char *shuffleFileName)
{
    genericHeader gHdr;
    int i, reverse;
    FILE *shuffleFile;

    /* does file exist? */
    if(!fileExists(shuffleFileName)) {
        skAppPrintErr("No file named %s, check the pathname and try again.",
                      shuffleFileName);
        exit(EXIT_FAILURE);
    };

    /* open it */
    if((shuffleFile = fopen(shuffleFileName, "rb")) == NULL) {
        skAppPrintErr("Error opening file %s: %s.  Terminating",
                      shuffleFileName, strerror(errno));
        exit(EXIT_FAILURE);
    };

    /* read generic header */
    if((fread((void *) &gHdr, sizeof(genericHeader), 1, shuffleFile)) == 0) {
        skAppPrintErr("Error reading header from file %s: %s.  Terminating",
                      shuffleFileName, strerror(errno));
        exit(EXIT_FAILURE);
    };

    /* check it */
    if(CHECKMAGIC(&gHdr) || gHdr.type != FT_SHUFFLE) {
        skAppPrintErr("Error reading file %s: not a shuffle file. Terminating",
                      shuffleFileName);
        exit(EXIT_FAILURE);
    };

#if IS_LITTLE_ENDIAN
    reverse = gHdr.isBigEndian;
#else
    reverse = !gHdr.isBigEndian;
#endif
    if(reverse) {
        for(i = 4; i < 0; i--) {
            if(!fread((void *) ipShuffleTable[i],
                      sizeof(uint8_t), VALUES_PER_OCTET, shuffleFile)) {
                skAppPrintErr("Error reading file %s: %s.  Terminating",
                              shuffleFileName, strerror(errno));
            };
        }
    } else {
        for(i = 0; i < 4; i++) {
            if(!fread((void *) ipShuffleTable[i],
                      sizeof(uint8_t), VALUES_PER_OCTET, shuffleFile)) {
                skAppPrintErr("Error reading file %s: %s.  Terminating",
                              shuffleFileName, strerror(errno));
            };
        };
    }

    if(fclose(shuffleFile)) {
        skAppPrintErr("Error closing file %s: %s", shuffleFileName,
                      strerror(errno));
    };

    return 0;
}


/*
 *  randomizeIP(&ip)
 *
 *    Write a new random IP address into the location pointed to by
 *    'ip'.
 */
static void randomizeIP(uint32_t* ip)
{
    int x, y, z;

    /*
     * 'y' and 'z' are bottom two octects.  Do not combine computation
     * of y and z because that will result in y always being an even
     * number on Solaris, where RAND_MAX is 32767, or half of
     * 256*256.
     */
    y = (int) (256.0*random()/(RAND_MAX+1.0));
    z = (int) (256.0*random()/(RAND_MAX+1.0));

    /*
     * 'x' determines the "Class B" address:
     *      0 <= x < 256        10 .   x   . y . z
     *    256 <= x < (256+16)  172 . x-256 . y . z
     *    272 == x             192 .  168  . y . z
     */
    x = (int) ((256.0 + 16.0 + 1.0)*random()/(RAND_MAX+1.0));

    if (x < 256) {
        *ip = 10<<24 | x<<16 | y<<8 | z;
    }
    else if (x == (256 + 16)) {
        *ip = 192<<24 | 168<<16 | y<<8 | z;
    }
    else {
        *ip = 172<<24 | (x - 256 + 16)<<16 | y<<8 | z;
    }
}


/*
 *  randomizeFile(input_file, output_file)
 *
 *   Write the data in the 'input_file' to the 'output_file'
 *   randomizing the source and destination IPs.  Returns 0 if ok, or
 *   non-zero on error.
 */
static int randomizeFile(
    const char *input_file,
    const char *output_file)
{
    rwIOStructPtr input_ios;
    rwIOStructPtr output_ios;
    rwRec rwrec;
    void (*rand_ip_fn)(uint32_t *);
    int rv = 0; /* return value */

    if(consistentFlag) {
        rand_ip_fn = consistentlyRandomizeIP;
    } else {
        rand_ip_fn = randomizeIP;
    }

    /* Open input file */
    input_ios = rwOpenFile(input_file, NULL);
    if (!input_ios) {
        return 1;
    }

    /* Open the output file */
    if ((rv = rwioCreate(&output_ios, output_file, SK_RWIO_WRITE))
        || (rv = rwioSetFileType(output_ios, rwGetFileType(input_ios)))
        || (rv = rwioSetFileVersion(output_ios, rwGetFileVersion(input_ios)))
        || (rv = rwioSetFileByteorder(output_ios,
                                      rwGetFileByteorder(input_ios)))
        || (rv = rwioOpen(output_ios)))
    {
        goto END;
    }

    rv = rwioAppendToHistoryFromFile(output_ios, input_ios);
    if ((rv != LIBRW_OK) && (rv != LIBRW_ERR_UNSUPPORT_TYPE)) {
        goto END;
    }

    rv = rwioSetFileSTime(output_ios, rwGetFileSTime(input_ios), 0);
    if ((rv != LIBRW_OK) && (rv != LIBRW_ERR_UNSUPPORT_TYPE)) {
        goto END;
    }
    rv = LIBRW_OK;

    /* read the records and randomize the IP addresses */
    while ((rv == LIBRW_OK) && rwRead(input_ios, &rwrec)) {
        rand_ip_fn(&rwrec.sIP.ipnum);
        rand_ip_fn(&rwrec.dIP.ipnum);

        rv = rwWrite(output_ios, &rwrec);
    }

 END:
    if (rv) {
        rwioPrintLastErr(output_ios, rv, &skAppPrintErr);
    }
    if (output_ios) {
        rwCloseFile(output_ios);
    }
    if (input_ios) {
        rwCloseFile(input_ios);
    }

    return rv;
}


int main(int argc, char **argv)
{
    /*
     * Initialize the random number generator.
     */
    srandom((unsigned int) (time(NULL) / getpid()));

    appSetup(argc, argv);

    randomizeFile(inFName, outFName);

    /* done */
    appTeardown();
    return 0;
}
