/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
** bagtree.c
**
** Implementation of the bagtree library according to bagtree.h
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwset/bagtree.c,v 1.15 2005/10/12 18:01:12 leechr Exp $");

#include "bagtree.h"
#include "utils.h"


/*
**
** Internal macros and functions
**
*/
#define INCREMENT_COUNTER( p_counter ) ( ( *(p_counter) ) += 1 )
#define SET_COUNTER( counter, value ) ( ( counter ) = ( value ) )
#define SKBAG_COUNTER_COPY( dst, src ) ( ( dst ) = ( src ) )

#define SKBAG_KEY_COPY( p_src, p_dst ) ( *( p_dst ) = *( p_src ) )

/* mwd's bit shifting/masking functions */
#define BITMASK(s) (((s) >= 32) ? (~(uint32_t)0) : \
                   (~((~(uint32_t)0) << (s))))
#define GET_BITS(x, o, s) (((x) >> (o)) & BITMASK((s)))
#define SET_BITS(x, v, o, s) ((x) = ((x) & (~(BITMASK((s)) << (o)))) | \
                             (((v) & BITMASK((s))) << (o)))


static skBag_counter_t * _skBag_allocToCounter(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key );

static skBag_counter_t * _skBag_allocToCounterRecursive(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key ,
    skBag_node_t *p_subtree,
    skBag_level_t level_cur );

static int _skBag_counterIsNull(
    skBag_counter_t *p_counter );

static skBag_counter_t * _skBag_getCounterPointer(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key );

static skBag_counter_t * _skBag_getCounterPointerRecursive(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key,
    skBag_node_t *p_subtree,
    skBag_level_t level_cur );

static void _skBag_recurNodeFree(
    skBag_header_t *p_hdr,
    skBag_node_t *p_subtree,
    skBag_level_t level_cur );

static skBag_counter_t * _skBag_traverseSubtree(
    skBag_iterator_t *p_iter,
    skBag_node_t *p_subtree,
    skBag_level_t level,
    int f_use_iterator_start );

static void _skBag_updateContentStats(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key,
    skBag_counter_t *p_counter );

static void _skBag_getKeyFromPath(
    skBag_iterator_t *p_iter,
    skBag_key_t *p_key );

static skBag_levelsize_t _skBag_getLevelOffset(
    skBag_header_t *p_hdr,
    skBag_level_t level );

static skBag_levelsize_t _skBag_getLevelSize(
    skBag_header_t *p_hdr,
    skBag_level_t level );

static void _skBag_setLevelSize(
    skBag_header_t *p_hdr,
    skBag_level_t level,
    skBag_levelsize_t value );




/*
**
** Exported Bagtree API functions
**
*/


/*
** skBag_alloc
**
** allocate memory for a new bagtree, and set "p_p_hdr" to point to
** it.  the new bagtree should be "levels" levels deep (counting leaf
** nodes), and each level should encode "rg_level_size[level]" bits of
** data.
**
** for example, to create a 9/9/9/5 IP space tree:
**  BagTreeHeader *p_hdr;
**  uint8_t level_sizes[4] = { 9, 9, 9, 5 };
**  skBag_alloc( &p_hdr, 4, level_sizes );
**
** Input:
**
**   p_p_hdr - a pointer to the skBag_header_t pointer the caller wants
**   to fill
**
**   levels - the total number of levels of the bagtree, including
**   root node and leaf level
**
**   rg_level_size - an array of length levels, where each entry is
**   the number of key bits encoded at that level
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_alloc(
    skBag_header_t **p_p_hdr,
    skBag_level_t levels,
    skBag_levelsize_t *rg_level_size )
{
    uint8_t key_size = 0;
    skBag_level_t level_cur = 0;
    skBag_header_t *p_result;
    skBag_level_t level_set_cur = 0;

    /* check inputs */
    if( levels < MIN_LEVELS || levels > MAX_LEVELS ) {
        /* bagtree must have at least one level */
        return SKBAG_ERR_INPUT;
    }

    if( rg_level_size == NULL ) {
        /* must have a valid array of level sizes */
        return SKBAG_ERR_INPUT;
    }

    /*
    ** WARNING - this code can not guarantee that rg_level_size is
    ** large enough to contain levels elements.  if it is too small,
    ** the code will fault here.  We could change the API to make it
    ** more safe, but it would also make it more annoying to use.
    */

    /* each level must contain at least 1 bit of key. */

    for( ; level_cur < levels; ++level_cur ) {
        if( rg_level_size[level_cur] < 1 ) {
            return SKBAG_ERR_INPUT;
        }
        key_size += rg_level_size[level_cur];
    }

    /* the tree can encode at most a 32b key */
    if( key_size < MIN_KEY_SIZE || key_size > MAX_KEY_SIZE ) {
        return SKBAG_ERR_INPUT;
    }

    /* allocate header */
    p_result = (skBag_header_t *) malloc( sizeof( skBag_header_t ) );
    if( p_result == NULL ) {
        return SKBAG_ERR_MEMORY;
    }

    /* allocate level ranges */
    p_result->rg_level_size = (skBag_levelsize_t *)
                              malloc( levels *
                                      sizeof( skBag_levelsize_t ) );
    if( p_result->rg_level_size == NULL ) {
        free( p_result );
        return SKBAG_ERR_MEMORY;
    }

    /* allocate stats block */
    p_result->p_stats = (skBag_stats_t *)
                        malloc( sizeof( skBag_stats_t ) );
    if( p_result->p_stats == NULL ) {
        free( p_result->rg_level_size );
        free( p_result );
        return SKBAG_ERR_MEMORY;
    }

    /* memory is all allocated okay, initialize values */
    p_result->root.p_child = NULL;
    p_result->levels = levels;

    /*
    ** go through the level size array again, this time setting the
    ** values of the new array.  (could move this into the loop above
    ** for performance, but it is conceptually easier down here, and
    ** the performance hit is negligible because the number of levels
    ** is small.
    */
    for( ; level_set_cur < levels; ++level_set_cur ) {
        _skBag_setLevelSize( p_result,
                             level_set_cur + 1,
                             rg_level_size[level_set_cur] );
    }

    /* just zero out all stats counters */
    memset( p_result->p_stats, '\0', sizeof( skBag_stats_t ) );
    p_result->p_stats->min_key = ~( (skBag_key_t) 0 );

    /* set pointer and return OK */
    *p_p_hdr = p_result;

    return SKBAG_OK;
}


/*
** skBag_free
**
**   free all memory associated with a bagtree
**
** Input:
**
**   p_hdr - a pointer to the skBag_header_t pointer the caller wants
**   to free
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_free(
    skBag_header_t *p_hdr )
{
    if( p_hdr == NULL ) {
        return SKBAG_ERR_INPUT;
    }

    /* free nodes in tree */
    if( p_hdr->root.p_child != NULL ) {
        _skBag_recurNodeFree( p_hdr, p_hdr->root.p_child,
                              (skBag_level_t) 1 );
    }

    /* free level definitions */
    if( p_hdr->rg_level_size != NULL ) {
        free( p_hdr->rg_level_size );
    }

    /* free stats block */
    if( p_hdr->p_stats != NULL ) {
        free( p_hdr->p_stats );
    }

    /* free header */
    free( p_hdr );

    return SKBAG_OK;
}


/*
** skBag_getCounter
**
**   lookup tree at p_hdr, get counter associated with key p_key, and
**   put its value in p_counter.
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_key - the key whose counter is to be retrieved
**
**   p_counter - a pointer to the counter the consumer wants to fill
**
** Output:
**
**   a skBag_err_r error code
**
**   changes counter pointed to by p_counter
*/
skBag_err_t skBag_getCounter(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key,
    skBag_counter_t *p_counter )
{
    skBag_counter_t *p_result;

    if( p_hdr == NULL ) {
        SKBAG_COUNTER_COPY( *p_counter, 0 );
        return SKBAG_ERR_INPUT;
    }

    if( p_hdr->root.p_child == NULL ) {
        SKBAG_COUNTER_COPY( *p_counter, 0 );
        return SKBAG_ERR_KEY_NOT_FOUND;
    }

    p_result = _skBag_getCounterPointer( p_hdr, p_key );
    if( p_result == NULL ) {
        SKBAG_COUNTER_COPY( *p_counter, 0 );
        return SKBAG_ERR_KEY_NOT_FOUND;
    }

    SKBAG_COUNTER_COPY( *p_counter, *p_result );
    return SKBAG_OK;
}


/*
** skBag_addToCounter
**
**   given a pointer to a key, increment the counter associated with
**   that key
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_key - a pointer to the key to whose counter you want to add
**
**   p_counter - a pointer to the counter whose values you are adding
**
** Output:
**
**   a skBag_err_t error code
**   
**   SKBAG_ERR_OP_BOUNDS if the addition would exceed the counter size
**   and cause it to loop.
*/
skBag_err_t skBag_addToCounter( skBag_header_t *p_hdr, 
                                skBag_key_t *p_key,
                                skBag_counter_t *p_counter )
{
    skBag_counter_t counter_temp;
    skBag_counter_t *p_counter_existing;

    if( p_hdr == NULL ) {
        return SKBAG_ERR_INPUT;
    }
    
    if( p_key == NULL ) {
        return SKBAG_ERR_INPUT;
    }
    
    if( p_counter == NULL ) {
        return SKBAG_ERR_INPUT;
    }
    
    /* try to get existing counter */
    p_counter_existing = _skBag_getCounterPointer( p_hdr, p_key );
    
    if( p_counter_existing == NULL ) {
        /* not found, allocate bagtree down to counter */
        p_counter_existing = _skBag_allocToCounter( p_hdr, p_key );
       
        if( p_counter_existing == NULL ) {
            return SKBAG_ERR_MEMORY;
        }
    }

    if( _skBag_counterIsNull( p_counter ) ) {
        /* update new counter statistics */
        p_hdr->p_stats->unique_keys += 1;
    }
   
    /* increment */
    /* note, this will not work when switching from uint counters
       to structs. */
    counter_temp = *p_counter_existing + *p_counter;
    if( counter_temp < *p_counter_existing ) {
        /*
        ** counter wrapped, return error
        */
        return SKBAG_ERR_OP_BOUNDS;
    }

    /* update statistics */
    if( _skBag_counterIsNull( p_counter_existing ) ) {
        p_hdr->p_stats->keys_inserted += 1;
    }

    *p_counter_existing = counter_temp;

    _skBag_updateContentStats( p_hdr, p_key, p_counter_existing );    

    return SKBAG_OK;
}


/*
** skBag_incrCounter
**
**   given a pointer to a key, increment the counter associated with
**   that key
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_key - a pointer to the key whose counter is to be retrieved
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_incrCounter(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key )
{
    skBag_counter_t *p_counter;

    if( p_hdr == NULL ) {
        return SKBAG_ERR_INPUT;
    }

    /* try to get existing counter */
    p_counter = _skBag_getCounterPointer( p_hdr, p_key );

    if( p_counter == NULL ) {
        /* not found, allocate bagtree down to counter */
        p_counter = _skBag_allocToCounter( p_hdr, p_key );

        if( p_counter == NULL ) {
            return SKBAG_ERR_MEMORY;
        }
    }

    if( _skBag_counterIsNull( p_counter ) ) {
        /* update new counter statistics */
        p_hdr->p_stats->unique_keys += 1;
    }

    /* increment */
    INCREMENT_COUNTER( p_counter );

    /* update statistics */
    p_hdr->p_stats->keys_inserted += 1;
    _skBag_updateContentStats( p_hdr, p_key, p_counter );
    return SKBAG_OK;
}


/*
** skBag_setCounter
**
**   given a pointer to a key, set the counter associated with that
**   key to a given value
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_key - a pointer to the key whose counter is to be retrieved
**
**   p_value - a pointer to the value to which to set the counter
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_setCounter(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key,
    skBag_counter_t *p_value )
{
    skBag_counter_t *p_counter;

    if( p_hdr == NULL ) {
        return SKBAG_ERR_INPUT;
    }

    /* search for counter to see if it exists */
    p_counter = _skBag_getCounterPointer( p_hdr, p_key );
    if( p_counter == NULL ) {
        /* not found, allocate bagtree down to counter */
        p_counter = _skBag_allocToCounter( p_hdr, p_key );

        if( p_counter == NULL ) {
            return SKBAG_ERR_MEMORY;
        }
    }

    if( _skBag_counterIsNull( p_counter ) ) {
        /* update new counter statistics */
        p_hdr->p_stats->unique_keys += 1;
    }

    /* set */
    SET_COUNTER( *p_counter, *p_value );

    /* update statistics */
    p_hdr->p_stats->keys_inserted += 1;

    _skBag_updateContentStats( p_hdr, p_key, p_counter );
    return SKBAG_OK;
}


/*
** skBag_subtractFromCounter
**
**   given a pointer to a key, increment the counter associated with
**   that key
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_key - a pointer to the key to from whose counter you want to subtract
**
**   p_counter - a pointer to the counter whose value you are subtracting
**
** Output:
**
**   a skBag_err_t error code
**   
**   SKBAG_ERR_OP_BOUNDS if the subtraction would cause the counter to
**   become negative
*/
skBag_err_t skBag_subtractFromCounter( skBag_header_t *p_hdr, 
                                       skBag_key_t *p_key,
                                       skBag_counter_t *p_counter )
{
    skBag_counter_t counter_temp;
    skBag_counter_t *p_counter_existing;

    if( p_hdr == NULL ) {
        return SKBAG_ERR_INPUT;
    }
    
    if( p_key == NULL ) {
        return SKBAG_ERR_INPUT;
    }
    
    if( p_counter == NULL ) {
        return SKBAG_ERR_INPUT;
    }
    
    /* try to get existing counter */
    p_counter_existing = _skBag_getCounterPointer( p_hdr, p_key );
    
    if( p_counter_existing == NULL ) {
        return SKBAG_ERR_OP_BOUNDS;
    }
   
    /* note, this will not work when switching from uint counters
       to structs. */
    counter_temp = *p_counter_existing - *p_counter;
    if( counter_temp > *p_counter_existing ) {
        /*
        ** counter wrapped, return error
        */
        return SKBAG_ERR_OP_BOUNDS;
    }

    /*
    ** update statistics (bounds may be incorrect after this
    ** operation.  a full iteration over the tree would be necessary
    ** to recompute bounds.)
    */
    *p_counter_existing = counter_temp;

    if( _skBag_counterIsNull( p_counter_existing ) ) {
        p_hdr->p_stats->keys_inserted -= 1;
        p_hdr->p_stats->unique_keys -= 1;
    }

    return SKBAG_OK;
}





/*
**
** Exported Bagtree Iterator API functions
**
*/


/*
** skBag_allocIterator
**
**   create a new bagtree iterator, and initialize it so the first
**   call to skBag_getNext will return the counter associated with the
**   first key
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_iter - a pointer to the iterator pointer the consumer wants to
**   fill
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_allocIterator(
    skBag_header_t *p_hdr,
    skBag_iterator_t **p_iter )
{
    /* check inputs */
    if( p_hdr == NULL ) {
        return SKBAG_ERR_INPUT;
    }

    /* allocate iterator */
    *p_iter = (skBag_iterator_t *) malloc( sizeof( skBag_iterator_t ) );
    if( *p_iter == NULL ) {
        return SKBAG_ERR_MEMORY;
    }

    /* allocate level ranges */
    (*p_iter)->rg_offset_path = (skBag_blocksize_t *)
                                malloc( p_hdr->levels *
                                        sizeof( skBag_blocksize_t ) );
    if( (*p_iter)->rg_offset_path == NULL ) {
        free( *p_iter );
        return SKBAG_ERR_MEMORY;
    }

    (*p_iter)->p_hdr = p_hdr;
    skBag_resetIterator( *p_iter );
    return SKBAG_OK;
}


/*
** skBag_freeIterator
**
**   deallocate all memory associated with a bagtree iterator
**
** Input:
**
**   p_iter - a pointer to the iterator to free
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_freeIterator(
    skBag_iterator_t *p_iter )
{
    if( p_iter == NULL ) {
        return SKBAG_ERR_INPUT;
    }

    if( p_iter->rg_offset_path != NULL ) {
        free( p_iter->rg_offset_path );
    }
    free( p_iter );
    return SKBAG_OK;
}


/*
** skBag_resetIterator
**
**   reset the iterator so the first call to skBag_getNext will return
**   the counter associated with the first key
**
** Input:
**
**   p_iter - a pointer to the iterator to reset
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_resetIterator(
    skBag_iterator_t *p_iter )
{
    skBag_level_t i = (skBag_level_t) 0;
    if( p_iter == NULL ) {
        return SKBAG_ERR_INPUT;
    }
    /* don't want to skip the first counter */
    p_iter->f_skip_counter = 0;

    for( ; i < p_iter->p_hdr->levels; ++i ) {
        p_iter->rg_offset_path[i] = (skBag_blocksize_t) 0;
    }
    return SKBAG_OK;
}


/*
** skBag_getNext
**
**   get the next key/counter pair associated with a given iterator
**
** Input:
**
**   p_iter - a pointer to the iterator
**
**   p_key - a pointer to the key to fill with the next key
**
**   p_counter - a pointer to the counter to fill with the next
**   counter
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_getNext( skBag_iterator_t *p_iter,
                           skBag_key_t *p_key,
                           skBag_counter_t *p_counter )
{
    skBag_counter_t *p_result;

    /* check input */
    if( p_iter == NULL ) {
        return SKBAG_ERR_INPUT;
    }
    assert( p_iter->p_hdr != NULL );
    assert( p_iter->rg_offset_path != NULL );

    p_result = _skBag_traverseSubtree( p_iter, &p_iter->p_hdr->root,
                                       (skBag_level_t) 1, 1 );

    if( p_result == NULL ) {
        /* there was no next key. */
        return SKBAG_ERR_KEY_NOT_FOUND;
    }

    /* counter found */
    SKBAG_COUNTER_COPY( *p_counter, *p_result );
    /* generate key value from path */
    _skBag_getKeyFromPath( p_iter, p_key );

    return SKBAG_OK;
}





/*
**
** Exported Bagtree I/O API functions
**
*/


/*
** skBag_printTreeStats
**
**   print out metadata on how the bagtree is performing
**
** Input:
**
**   p_hdr - a pointer to the bagtree
**
**   fp_out - an open output stream
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_printTreeStats(
    skBag_header_t *p_hdr,
    FILE *fp_out )
{
    uint32_t total_counters;

    if( p_hdr == NULL ) {
        return SKBAG_ERR_INPUT;
    }

    if( fp_out == NULL ) {
        return SKBAG_ERR_INPUT;
    }

    fprintf( fp_out, "%18s:  %u (%llu bytes)\n",
             "nodes allocated",
             p_hdr->p_stats->nodes,
             (unsigned long long) p_hdr->p_stats->nodes_size );

    fprintf( fp_out, "%18s:  %u (%llu bytes)\n",
             "leaves allocated",
             p_hdr->p_stats->leaves,
             (unsigned long long) p_hdr->p_stats->leaves_size );

    fprintf( fp_out, "%18s:  %llu (%llu unique)\n",
             "keys inserted",
             p_hdr->p_stats->keys_inserted,
             (unsigned long long) p_hdr->p_stats->unique_keys );

    total_counters = (uint32_t) p_hdr->p_stats->leaves *
                     (uint32_t) 1 << _skBag_getLevelSize( p_hdr, 
                                                          p_hdr->levels );

    fprintf( fp_out, "%18s:  %.02f%%\n", 
             "counter density",
             (float) 100 * (
                 (float) p_hdr->p_stats->unique_keys / 
                 (float) total_counters ) );

    return SKBAG_OK;
}


/*
** skBag_writeBinary
**
**   serialize bag to output stream
**
** Input:
**
**   p_hdr - a pointer to the bagtree
**
**   fp_out - an open output stream
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_writeBinary(
    skBag_header_t *p_hdr,
    FILE *fp_out )
{
    skBag_iterator_t *p_iter;
    skBag_err_t err, err2;
    skBag_key_t key;
    skBag_counter_t counter;
    genericHeader file_header;

    if( p_hdr == NULL ) {
        return SKBAG_ERR_INPUT;
    }

    if( fp_out == NULL ) {
        return SKBAG_ERR_INPUT;
    }

    err = skBag_allocIterator( p_hdr, &p_iter );
    if( err != SKBAG_OK ) {
        return err;
    }

    /*
    ** output header
    */
    PREPHEADER( &file_header );
    file_header.type = FT_MACROBAGTREE;
    file_header.version = 2;
    fwrite( (void *) &file_header, sizeof( genericHeader ), 1, fp_out );

    /*
    ** output key/counter pairs
    */
    while ( (err = skBag_getNext( p_iter, &key, &counter )) == SKBAG_OK ) {
        fwrite( (void *) &key, sizeof( skBag_key_t ), 1, fp_out );
        fwrite( (void *) &counter, sizeof( skBag_counter_t ), 1, fp_out );
    }
    if( err == SKBAG_ERR_KEY_NOT_FOUND ) {
        /* expected behavior.  Reset 'err' */
        err = SKBAG_OK;
    } else {
        skAppPrintErr( "skBag_writeBinary:  error looping" );
    }

    err2 = skBag_freeIterator( p_iter );
    if( err2 != SKBAG_OK ) {
        skAppPrintErr( "skBag_writeBinary:  error %u freeing iterator",
                       err2 );
        /* Return err2 unless err already set */
        if (err == SKBAG_OK) {
            err = err2;
        }
    }

    return err;
}


/*
** skBag_readBinary
**
**   read serialized bag from input stream
**
** Input:
**
**   p_p_hdr - a pointer to the bagtree
**
**   fp_in - an open input stream
**
** Output:
**
**   a skBag_err_t error code
*/
skBag_err_t skBag_readBinary(
    skBag_header_t **p_p_hdr,
    FILE *fp_in )
{
    int f_swap_byte_order;
    skBag_level_t levels = (skBag_level_t) 4;
    skBag_levelsize_t rg_level_size[4] = { 9, 9, 9, 5 };
    skBag_err_t err;
    skBag_key_t key;
    uint32_t counter_32;
    skBag_counter_t counter;
    size_t bytes_read;
    genericHeader file_header;

    if( p_p_hdr == NULL ) {
        return SKBAG_ERR_INPUT;
    }

    if( fp_in == NULL ) {
        return SKBAG_ERR_INPUT;
    }

    /*
    ** read header
    */
    if( !fread( (void *) &file_header, sizeof( genericHeader ),
                1, fp_in ) )
    {
        skAppPrintErr( "skBag_readBinary:  cannot read header" );
        return SKBAG_ERR_INPUT;
    }

    if( CHECKMAGIC( &file_header ) ) {
        skAppPrintErr( "skBag_readBinary:  invalid header" );
        return SKBAG_ERR_INPUT;
    }

    if( file_header.type != FT_MACROBAGTREE ) {
        skAppPrintErr( "skBag_readBinary:  not a bag file" );
        return SKBAG_ERR_INPUT;
    }
    if (file_header.version > 2) {
        skAppPrintErr("skBag_readBinary: Bag file versions %d not supported",
                      file_header.version);
        return SKBAG_ERR_INPUT;
    }

    /*
    ** determine if endian swap is necessary
    */
#if IS_LITTLE_ENDIAN
    f_swap_byte_order = file_header.isBigEndian;
#else
    f_swap_byte_order = !file_header.isBigEndian;
#endif

    /*
    ** create bag
    */
    err = skBag_alloc( p_p_hdr, levels, rg_level_size );
    if( err ) {
        return err;
    }

    /*
    ** read key/counter pairs
    */
    while( 1 ) {
        bytes_read = fread( &key, sizeof( skBag_key_t ), 1, fp_in );
        if( bytes_read == 0 ) {
            break;
        }

        if( f_swap_byte_order ) {
            key = BSWAP32( key );
        }

        if( file_header.version == 1 ) {
            bytes_read = fread( &counter_32, sizeof( uint32_t ), 1, fp_in );
            if( f_swap_byte_order ) {
                counter = (skBag_counter_t) BSWAP32( counter_32 );
            } else {
                counter = (skBag_counter_t) counter_32;
            }
        } else {
            bytes_read = fread( &counter, sizeof( skBag_counter_t ), 1, 
                                fp_in );
            if( f_swap_byte_order ) {
                counter = BSWAP64( counter );
            }
        }

        if( bytes_read == 0 ) {
            break;
        }

        err = skBag_setCounter( *p_p_hdr, &key, &counter );
        if( err != SKBAG_OK ) {
            skAppPrintErr( "skBag_readBinary:  error %u setting counter",
                           err );
            return err;
        }
    }

    return SKBAG_OK;
}





/*
**
** Internal functions
**
*/


/*
** _skBag_allocToCounter
**
**   similar to getCounterPointer, but allocates nodes along the way
**   as necessary
**
** Input:
**
**   p_hdr - a pointer to the bagtree
**
**   p_key - a pointer to the key whose counter is to be incremented
**
** Output:
**
**   a pointer to the counter OR
**
**   NULL if allocation fails
*/
static skBag_counter_t * _skBag_allocToCounter(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key )
{
    assert( p_hdr != NULL );
    assert( p_key != NULL );
    return _skBag_allocToCounterRecursive( p_hdr, p_key, &p_hdr->root,
                                           (skBag_level_t) 1 );
}


/*
** _skBag_allocToCounterRecursive
**
**   recursive function to drill through the tree to a particular
**   counter, allocating nodes if they don't exist
**
** Input:
**
**   p_hdr - a pointer to the bagtree
**
**   p_key - a pointer to the key whose counter is to be incremented
**
**   p_p_subtree - a pointer to the place which will be the new root
**   node of the subtree being allocated
**
**   level_cur - the current level of the tree (should be between 1
**   and p_hdr->levels - 1)
**
** Output:
**
**   a pointer to the counter OR
**
**   NULL if allocation fails
*/
static skBag_counter_t * _skBag_allocToCounterRecursive(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key ,
    skBag_node_t *p_subtree,
    skBag_level_t level_cur )
{
    skBag_levelsize_t level_cur_offset;
    skBag_levelsize_t level_cur_size;
    uint32_t key_bits;

    assert( p_hdr != NULL );
    assert( p_key != NULL );
    assert( p_subtree != NULL );
    assert( level_cur >= 1 && level_cur <= p_hdr->levels );

    /*
    ** p_subtree does not point to the first node in the block.  it
    ** actually points to the parent of the first node in the block
    ** (so we can update the parent as we allocate new blocks
    */

    level_cur_offset = _skBag_getLevelOffset( p_hdr, level_cur );
    level_cur_size = _skBag_getLevelSize( p_hdr, level_cur );
    key_bits = GET_BITS( *p_key, level_cur_offset, level_cur_size );

    if( level_cur < p_hdr->levels ) {
        /* if we have are not the leaf level, allocate node */
        if( p_subtree->p_child == NULL ) {
            /* no child node exists, allocating block */
            uint32_t child_block_size =
                (uint32_t) sizeof( skBag_node_t ) *
                ( (uint32_t) 1 << _skBag_getLevelSize( p_hdr, level_cur ) );

            p_subtree->p_child = (skBag_node_t *) malloc( child_block_size );
            if( p_subtree->p_child == NULL ) {
                return NULL;
            }
            memset( p_subtree->p_child, '\0', child_block_size );
            /* update statistics */
            p_hdr->p_stats->nodes += 1;
            p_hdr->p_stats->nodes_size += child_block_size;
        }

        /* recurse one level deeper */
        return _skBag_allocToCounterRecursive(
            p_hdr, p_key, &p_subtree->p_child[key_bits], level_cur + 1 );

    } else {
        /*
        ** if we are currently on the last node level, our child
        ** should be a leaf
        */
        if( p_subtree->p_leaf == NULL ) {
            uint32_t leaf_block_size =
                (uint32_t) sizeof( skBag_counter_t ) *
                ( (uint32_t) 1 << _skBag_getLevelSize( p_hdr, level_cur ) );

            /* actually not a skBag_node_t we are allocating */
            p_subtree->p_leaf = (skBag_counter_t *) malloc( leaf_block_size );
            if( p_subtree->p_leaf == NULL ) {
                return NULL;
            }
            memset( p_subtree->p_leaf, 0, leaf_block_size );
            /* update statistics */
            p_hdr->p_stats->leaves += 1;
            p_hdr->p_stats->leaves_size += leaf_block_size;
        }

        return &p_subtree->p_leaf[key_bits];
    }
}


/*
** _skBag_counterIsNull
**
**   check a counter to see if it either doesn't exist, or exists but
**   has all counts of zero.  (could be a simple check now, but in the
**   future, may be more complex (if multiple counters exist).
**
** Input:
**
**   p_counter - a pointer to the counter to check
**
** Output:
**
**   1 if the counter does not exist, or if all counts are zero
**
**   0 if the counter exists and is positive
*/
static int _skBag_counterIsNull(
    skBag_counter_t *p_counter )
{
    assert( p_counter != NULL );
    return ( ( *p_counter == (skBag_counter_t) 0 ) ? 1 : 0 );
}


/*
** _skBag_getCounterPointer
**
**   given a pointer to a key, get a pointer to a counter
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_key - a pointer to the key whose counter is to be retrieved
**
** Output:
**
**   a pointer to the counter found or
**
**   NULL if no counter is found
*/
static skBag_counter_t * _skBag_getCounterPointer(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key )
{
    assert( p_hdr != NULL );
    assert( p_key != NULL );
    return _skBag_getCounterPointerRecursive(
        p_hdr, p_key, &p_hdr->root, (skBag_level_t) 1 );
}


/*
** _skBag_getCounterPointerRecursive
**
**   recursive function to drill through the tree.  access through the
**   _skBag_getCounterPointer wrapper.
**
** Input:
**
**   p_hdr - a pointer to the bagtree from which to get the counter
**
**   p_key - a pointer to the key whose counter is to be retrieved
**
**   p_subtree - a pointer to the current root of the subtree being
**   searched
**
**   level_cur - the current level of the tree (should be between 1
**   and p_hdr->levels - 1)
**
** Output:
**
**   a pointer to the counter found or
**
**   NULL if no counter is found
*/
static skBag_counter_t * _skBag_getCounterPointerRecursive(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key,
    skBag_node_t *p_subtree,
    skBag_level_t level_cur )
{
    skBag_levelsize_t level_cur_offset;
    skBag_levelsize_t level_cur_size;
    uint32_t key_bits;

    assert( p_hdr != NULL );
    assert( p_key != NULL );
    assert( p_subtree != NULL );
    assert( level_cur >= 1 && level_cur <= p_hdr->levels );

    /*
    ** p_subtree does not point to the first node in the block.  it
    ** actually points to the parent of the first node in the block
    */

    /* if there is no subtree, then the counter wasn't found */
    if( p_subtree->p_child == NULL ) {
        return NULL;
    }

    /* calculate the offset into this level */
    level_cur_offset = _skBag_getLevelOffset( p_hdr, level_cur );
    level_cur_size = _skBag_getLevelSize( p_hdr, level_cur );
    key_bits = GET_BITS( *p_key, level_cur_offset, level_cur_size );

    if( level_cur < p_hdr->levels ) {
        /* we are on a level of nodes, recurse down to the next level */
        return _skBag_getCounterPointerRecursive(
            p_hdr, p_key, &p_subtree->p_child[key_bits], level_cur + 1 );
    } else {
        /* in a leaf of counters */
        return &p_subtree->p_leaf[key_bits];
    }
}


/*
** _skBag_recurNodeFree
**
**   travel recursively through tree nodes, freeing all blocks
**
** Input:
**
**   p_hdr - a pointer to the skBag_header_t pointer the caller wants
**   to free
**
**   p_subtree - a pointer to the current root of the subtree being
**   freed
**
**   level_cur - the current level of the tree (should be between 1
**   and p_hdr->levels - 1)
**
** Output:
**
**   none
*/
static void _skBag_recurNodeFree(
    skBag_header_t *p_hdr,
    skBag_node_t *p_subtree,
    skBag_level_t level_cur )
{
    skBag_blocksize_t level_size;

    assert( p_hdr != NULL );
    assert( p_subtree != NULL );
    assert( level_cur >= 1 && level_cur <= p_hdr->levels );

    /* get the number of entries on this level */
    level_size = (skBag_blocksize_t) 1 <<
                 _skBag_getLevelSize( p_hdr, level_cur );

    /*
    ** for each pointer on this level which is not null, recurse this
    ** function using that pointer as the new tree root
    */

    if( level_cur < p_hdr->levels - 1 ) {
        /* we are in a node pointing to more nodes */
        uint32_t i_offset;
        for( i_offset = 0; i_offset < level_size; ++i_offset ) {
            if( p_subtree[i_offset].p_child != NULL ) {
                _skBag_recurNodeFree( p_hdr, p_subtree[i_offset].p_child,
                                      level_cur + 1 );
            }
        }
    } else {
        uint32_t i_offset;
        /* we are in a node pointing to leaves */
        for( i_offset = 0; i_offset < level_size; ++i_offset ) {
            if( p_subtree[i_offset].p_leaf != NULL ) {
                free( p_subtree[i_offset].p_leaf );
            }
        }
    }

    free( p_subtree );
}


/*
** _skBag_traverseSubtree
**
**   walk across the subtree and find the next non-zero counter.  if
**   the iterator has f_skip_cqqounter set to true, skip the first
**   counter (whether empty or not.  if skip_counter is false, then
**   start at the first counter.
**
** Input:
**
**   p_iter - a pointer to the iterator
**
**   p_subtree - a pointer to the current root of the subtree being
**   searched
**
**   skBag_level_t level_cur,
**
**   level_cur - the current level we are searching
**
**   f_use_iterator_start - boolean value.  if true, start offset from
**   the apropriate entry in rg_offset_path.  also, when making
**   recursive calls, only set to true once.  that way, the leftmost
**   path of the tree uses offsets, but once beyond that, search whole
**   blocks.
**
** Output:
**
**   a pointer to the next counter found OR
**
**   NULL if no counter was found
**
** also, sets p_iter->rg_offset_path to point to the counter found,
** and f_skip_counter if a counter is found (so that it doesn't
** re-find that counter on the next call.
*/
static skBag_counter_t * _skBag_traverseSubtree(
    skBag_iterator_t *p_iter,
    skBag_node_t *p_subtree,
    skBag_level_t level,
    int f_use_iterator_start )
{
    skBag_blocksize_t block_cur;
    skBag_blocksize_t block_size;
    skBag_counter_t *p_counter = NULL;

    /* check input */
    assert( p_iter != NULL );
    assert( p_iter->p_hdr != NULL );
    assert( p_iter->rg_offset_path != NULL );
    assert( p_subtree != NULL );
    assert( level >= 1 && level <= p_iter->p_hdr->levels );

    if( p_subtree->p_child == NULL ) {
        return NULL;
    }

    if( f_use_iterator_start ) {
        block_cur = p_iter->rg_offset_path[level - 1];
    } else {
        block_cur = (skBag_blocksize_t) 0;
    }

    block_size = (skBag_blocksize_t) 1 <<
                 _skBag_getLevelSize( p_iter->p_hdr, level );

    /*
    ** if we are at (or past!) the last node in the leaf, the counter
    ** will not be found in this subtree
    */
    if( block_cur >= block_size ) {
        return NULL;
    }

    if( level < p_iter->p_hdr->levels ) {
        /*
        ** if we are not at the bottom level of nodes, for each
        ** non-null child pointer, recurse one level deeper.
        */
        for( ; block_cur < block_size; ++block_cur ) {
            if( p_subtree->p_child[block_cur].p_child != NULL ) {
                /*
                ** if the counter is NULL, there is no subtree from
                ** this offset.
                */

                /*
                ** rg_offset_path[level] actually refers to the NEXT
                ** level, since the array starts at index 0 for level
                ** 1.
                */
                p_counter = _skBag_traverseSubtree(
                    p_iter, &p_subtree->p_child[block_cur],
                    level + 1, f_use_iterator_start );
                f_use_iterator_start = 0;
            }

            /*
            ** at this point, p_counter points to a found counter, or
            ** still is NULL.  if it points to a valid counter, set
            ** the iterator path for the current level, and return to
            ** the next shallow level.  (again, remember that the
            ** current level is index level - 1 in the offset path
            ** array.)
            */
            if( p_counter != NULL ) {
                p_iter->rg_offset_path[level - 1] = block_cur;
                return p_counter;
            }
        }
    } else {
        /*
        ** at the leaf level, so check for non-zero counters
        */

        /*
        ** if we follow the path straight down, we will re-find the
        ** same counter we found on the last iteration.  to avoid
        ** that, we skip the first counter on that straight path.
        */
        if( p_iter->f_skip_counter == 1 ) {
            ++block_cur;
            p_iter->f_skip_counter = 0;
        }

        for( ; block_cur < block_size; ++block_cur ) {
            if( ! _skBag_counterIsNull( &p_subtree->p_leaf[block_cur] ) ) {
                /*
                ** we found a counter.
                */
                p_iter->rg_offset_path[level - 1] = block_cur;
                if( level == p_iter->p_hdr->levels ) {
                    p_iter->f_skip_counter = 1;
                }
                return &p_subtree->p_leaf[block_cur];
            }
        }
    }

    /*
    ** no counter found.  if we are at the most shallow level, reset
    ** the iterator (so subsequent calls will re-start at the
    ** beginning of the tree.)
    */
    if( level == 1 ) {
        skBag_resetIterator( p_iter );
    }
    return NULL;
}


/*
** _skBag_updateContentStats
**
**   update the max_counter, min_key, max_key statistics after setting
**   a value.  (note that statistics bounds only tighten; there is no
**   way to loosen the value if a counter is set outside the bounds.
**   since statistics are generated every time the bagtree is created
**   in memory, serializing and deserializing will correct the
**   bounds.)
**
** Input:
**
**   p_hdr - a pointer to the bagtree
**
**   p_key - a pointer to the key just updated
**
**   p_counter a pointer to the counter just updated
**
** Output:
**   none
*/
static void _skBag_updateContentStats(
    skBag_header_t *p_hdr,
    skBag_key_t *p_key,
    skBag_counter_t *p_counter )
{
    assert( p_hdr != NULL );
    assert( p_hdr->p_stats != NULL );

    if( *p_counter > p_hdr->p_stats->max_counter ) {
        p_hdr->p_stats->max_counter = *p_counter;
    }

    if( *p_key > p_hdr->p_stats->max_key ) {
        p_hdr->p_stats->max_key = *p_key;
    }

    if( *p_key < p_hdr->p_stats->min_key ) {
        p_hdr->p_stats->min_key = *p_key;
    }
}





/*
**
** Macros / helper functions
**
*/


static void _skBag_getKeyFromPath(
    skBag_iterator_t *p_iter,
    skBag_key_t *p_key )
{
    skBag_level_t i;
    skBag_levelsize_t offset;

    assert( p_iter != NULL );
    assert( p_iter->p_hdr != NULL );
    assert( p_iter->rg_offset_path != NULL );
    assert( p_key != NULL );

    i = p_iter->p_hdr->levels;
    offset = 0;
    *p_key = 0;
    while( i > 0 ) {
        *p_key += ( 1 << offset ) * p_iter->rg_offset_path[i - 1];
        offset += _skBag_getLevelSize( p_iter->p_hdr, i );
        --i;
    }
}


/*
** calculate the offset this level encodes.  (the sum of the offsets
** that every level below this one encodes.)
*/
static skBag_levelsize_t _skBag_getLevelOffset(
    skBag_header_t *p_hdr,
    skBag_level_t level )
{
    skBag_levelsize_t offset = 0;

    assert( p_hdr != NULL );
    assert( level >= 1 && level <= p_hdr->levels );

    ++level; /* start at next level */
    while( level <= p_hdr->levels ) {
        offset += _skBag_getLevelSize( p_hdr, level );
        ++level;
    }
    return offset;
}


static skBag_levelsize_t _skBag_getLevelSize(
    skBag_header_t *p_hdr,
    skBag_level_t level )
{
    assert( p_hdr != NULL );
    assert( level >= 1 && level <= p_hdr->levels );

    if( level >= 1 && level <= p_hdr->levels ) {
        return p_hdr->rg_level_size[level - 1];
    } else {
        return (skBag_levelsize_t) 0;
    }
}


static void _skBag_setLevelSize(
    skBag_header_t *p_hdr,
    skBag_level_t level,
    skBag_levelsize_t value )
{
    assert( p_hdr != NULL );
    assert( level >= 1 && level <= p_hdr->levels );

    if( level >= 1 && level <= p_hdr->levels ) {
        p_hdr->rg_level_size[level - 1] = value;
    }
}
