/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
 * buildset
 *
 * Build an IP Set from textual data.
 */

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwset/buildset.c,v 1.19 2005/10/26 18:32:27 thomasm Exp $");

#include "iptree.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* where to write output from --help */
#define USAGE_FH stdout

/* return 1 if position 'i' in bitmap 'f' is set. */
#define CHECK_MARK(f,i) ( ((f)[(i)/32] & (1 << ((i) % 32))) ? 1 : 0 )


/* LOCAL FUNCTIONS */

static void appUsage(void);                     /* never returns */
static void appUsageLong(void);                 /* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);    /* never returns */
static void appOptionsUsage(FILE*);             /* application options usage */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);


/* LOCAL VARIABLES */

/* the IPset the application creates. */
static skIPTree_t *ipset = NULL;

/* name of input file */
static const char *input_fname = NULL;

/* name of output file */
const char *output_fname = NULL;


/* OPTIONS SETUP */

typedef enum {
    BUILD_HELP /* remove this when real options added */
} appOptionsEnum;

static struct option appOptions[] = {
    {"HELP", NO_ARG, 0, 0}, /* remove this when real options added */
    {0,0,0,0}                     /* sentinal entry */
};

static const char *appHelp[] = {
    "HELP", /* remove this when real options added */
    (char *)NULL
};


/*
 * appUsage:
 *      print usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 *      print usage information to USAGE_FH and exit with code 1.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                             \
    ("%s <input-file> <output-file>\n"                                        \
     "\tReads IP addresses in dotted-quad or CIDR notation from input-file\n" \
     "\tand writes a binary IPset file to output-file.  Use \"stdin\" as\n"   \
     "\tthe input-file to read the IPs from the standard input, and use\n"    \
     "\t\"stdout\" as the output-file to write the IPset to the standard\n"   \
     "\toutput when the standard output is not a terminal.\n")

    fprintf(USAGE_FH, USAGE_MSG, skAppName());
    appOptionsUsage(USAGE_FH);
    exit(EXIT_SUCCESS);
}


/*
 * appOptionsUsage:
 *      print options for this app to specified file handle
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 */
static void appOptionsUsage(FILE UNUSED(*fh))
{
/*     int i; */

/*     fprintf(fh, "\noptions:\n"); */
/*     for (i = 0; appOptions[i].name; i++ ) { */
/*         fprintf(fh, "--%s %s. %s\n", appOptions[i].name, */
/*                 SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]); */
/*     } */

    return;
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    if (ipset) {
        skIPTreeDelete(&ipset);
    }

    optionsTeardown();

    return;
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int next_option;

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* Register application and initialize globals */
    skAppRegister(argv[0]);

    /* initialize options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* register our options and our options handler */
    if (optionsRegister(appOptions, (optHandler) appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* parse options */
    next_option = optionsParse(argc, argv);
    if (next_option < 0) {
        appUsage(); /* never returns */
    }

    /* check correct number of args */
    if ((argc - next_option) != 2) {
        if ((argc - next_option) == 0) {
            skAppPrintErr("missing input and output file names");
        } else if ((argc - next_option) == 1) {
            skAppPrintErr("missing output file name.");
        } else {
            skAppPrintErr("too many arguments on command line");
        }
        appUsage(); /* never returns */
    }

    /* set input and output names */
    input_fname = argv[next_option++];
    output_fname = argv[next_option++];

    /* if output is to stdout; make certain it is not a tty; if stdout
     * is to a file, make certain file doesn't exist. */
    if (0 == strcmp("stdout", output_fname)) {
        if (FILEIsATty(stdout)) {
            skAppPrintErr("stdout is a terminal");
            exit(EXIT_FAILURE);
        }
    } else {
        if (fileExists(output_fname)) {
            /* file exists */
            skAppPrintErr(("Output IPset file '%s' already exists.\n"
                           "\t Delete or move it and try again."),
                          output_fname);
            exit(EXIT_FAILURE);
        }
    }

    /* create tree */
    if (skIPTreeCreate(&ipset)) {
        skAppPrintErr("Error allocating memory for IPset.");
        exit(EXIT_FAILURE);
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int  appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char        UNUSED(*opt_arg))
{

    switch ((appOptionsEnum)opt_index) {
      case BUILD_HELP: /* remove this when real options added */
        appUsageLong();
    }

    return 1;
}


/*
 * uint32_t parseIP(const char *ipString)
 *
 * parseIP uses the ip address specified in ipString to determine what
 * spots to mark in a macrotree.
 *
 * 'ipString' is a string representation of the IP address in CIDR
 * format or in ,- list format.
 *
 * Return code - 0 on success, nonzero on failure.
 *
 * Failures occur because of:
 * a) faulty parsing
 * b) too many IP addresses added.
 *
 * SIDE EFFECTS: on success, the global skIPTree_t 'ipset' will get an
 * address set added to it.
 */
static uint32_t parseIP(const char *ipString)
{
    uint32_t ipLatch[4];
    uint32_t value[4];
    uint32_t octet_bitmap[4][8];
    int rv;

    /* Fill in octet_bitmap */
    rv = skIPStringToBitmap(octet_bitmap, ipString);
    if (rv != 0) {
        /* not an IP address */
        if (rv < 0) {
            /* error */
            return 1;
        } else {
            /* whitespace only */
            return 0;
        }
    }

    /*
     *  Now we actually mark the ip addresses in question.
     *
     *  For each octet i, loop the ipLatch[i] variable through its
     *  bitmap---stored in octet_bitmap[i][].  If the bit is set, set
     *  value[i] to that octet's absolute value--i.e, shifted the
     *  correct number of bits--plus the value[i-1] from the previous
     *  octet.
     */
    for(ipLatch[0] = 0; ipLatch[0] < 256 ; ipLatch[0]++) {
        if( !CHECK_MARK(octet_bitmap[0], ipLatch[0])) {
            continue;
        }
        value[0] = ipLatch[0] << 24;

        for(ipLatch[1] = 0 ; ipLatch[1] < 256; ipLatch[1]++) {
            if( !CHECK_MARK(octet_bitmap[1], ipLatch[1])) {
                continue;
            }
            value[1] = (ipLatch[1] << 16) + value[0];

            for(ipLatch[2] = 0; ipLatch[2] < 256; ipLatch[2]++) {
                if( !CHECK_MARK(octet_bitmap[2], ipLatch[2])) {
                    continue;
                }
                value[2] = (ipLatch[2] << 8) + value[1];

                for(ipLatch[3] = 0; ipLatch[3] < 256; ipLatch[3]++) {
                    if( !CHECK_MARK(octet_bitmap[3], ipLatch[3])) {
                        continue;
                    }
                    value[3] = ipLatch[3] + value[2];

                    skIPTreeAddAddress(value[3], ipset);
                }
            }
        }
    }

    return 0;
}


/*
 *  buildIPTreeFromFile(input_file);
 *
 *    Read IP addresses from the file named by 'input_file' and use
 *    the to build the global IP Tree ipset.  Return 0 on success or
 *    -1 on failure.
 */
static int buildIPTreeFromFile(const char *input_file)
{
    FILE *asciiFile;
    int currentLine;
    char dataBuffer[255];
    char *eol;

    currentLine = 0;

    /* open input */
    if(!strcmp(input_file, "stdin")) {
        asciiFile = stdin;
    } else {
        if ((asciiFile = fopen(input_file, "r")) == NULL) {
            skAppPrintErr("Error opening file %s, exiting", input_file);
            return -1;
        }
    }

    /* read until end of file */
    while (!feof(asciiFile)) {
        memset(dataBuffer, '\0', sizeof(dataBuffer));
        if (fgets(dataBuffer, sizeof(dataBuffer), asciiFile) == NULL) {
            continue;
        }
        currentLine++;

        /* find end of line */
        eol = strchr(dataBuffer, '\n');
        if (dataBuffer == eol) {
            /* empty line; ignore */
            continue;
        }

        if (eol != NULL) {
            /* expected behavior: read an entire line */
            *eol = '\0';
        } else if (feof(asciiFile)) {
            /* okay: last line did not end in newline */
        } else {
            /* bad: line was longer than sizeof(dataBuffer).  read
             * until newline or eof, then throw away the line */
            skAppPrintErr("input line %d too long. ignored",
                          currentLine);
            while (fgets(dataBuffer, sizeof(dataBuffer), asciiFile)
                   && !strchr(dataBuffer, '\n'))
                ; /* empty */
            continue;
        }

        /* Terminate line at first comment char */
        if (NULL != (eol = strchr(dataBuffer, '#'))) {
            *eol = '\0';
        }

        /* parse the line */
        if(parseIP(dataBuffer)) {
            skAppPrintErr("Error in line %d: %s\n", currentLine, dataBuffer);
            return -1;
        };
    }

    /* close input */
    if (asciiFile != stdin) {
        fclose(asciiFile);
    }

    return 0;
}


int main(int argc, char **argv)
{
    appSetup(argc, argv);

    /* read input file */
    if (buildIPTreeFromFile(input_fname)) {
        return -1;
    }

    if (0 == strcmp("stdout", output_fname)) {
        /* write output to stdout */
        if (skIPTreeWrite(stdout, ipset)) {
            return -1;
        }
    } else {
        /* write output to file */
        if (skIPTreeSave(output_fname, ipset)) {
            return -1;
        }
    }

    /* done */
    return 0;
}
