/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
** ipfilter.c
**
**  This is a dynamic library which allows us to incorporate IP-sets
**  into rwfilter.
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwset/ipfilter.c,v 1.16 2005/10/26 18:26:08 thomasm Exp $");

#include "iptree.h"
#include "utils.h"
#include "dynlib.h"
#include "rwpack.h"


/* MACROS AND TYPEDEFS */

/* A count of the number of sets */
#define COUNT_SET_TYPES  4

/* Symbol names for the sets */
typedef enum {
    SET_SIP,
    SET_DIP,
    SET_ANY,
    SET_NHIP
} set_types_t;


/* EXPORTED FUNCTIONS */

/* called by dynlib */
int dynlib_api_version(void);
int setup(dynlibInfoStructPtr dlISP, dynlibSymbolId appType);
void teardown(dynlibSymbolId appType);
void optionsUsage(dynlibSymbolId appType, FILE *fh);
int filter(rwRecPtr rwrec);


/* LOCAL FUNCTIONS */

static int libOptionsSetup(dynlibInfoStructPtr dlISP, dynlibSymbolId aType);
static int optionsHandler(clientData cData, int opt_index, char *opt_arg);


/* LOCAL VARIABLES */

/* Plugin's name */
static const char *pluginName = "ipfilter";

/* The sets of IPs to accept or reject. */
static skIPTree_t *ipset[COUNT_SET_TYPES];

/* Whether the set is an accept (=1) or reject (=0) set. */
static int is_accept[COUNT_SET_TYPES];


/* OPTIONS SETUP */

/*
 * the options and help strings.  these vary depending on the type of
 * app that is using the shared library, and they get set to one of
 * the arrays below.
 */
static struct option *libOptions = NULL;
static char **optionsHelp = NULL;


/*
 * Options that get added to rwfilter.
 */
static struct option filterOptions[1+ 2*COUNT_SET_TYPES] = {
    {"sipset",      REQUIRED_ARG, 0, SET_SIP},
    {"dipset",      REQUIRED_ARG, 0, SET_DIP},
    {"anyset",      REQUIRED_ARG, 0, SET_ANY},
    {"nhipset",     REQUIRED_ARG, 0, SET_NHIP},
    {"not-sipset",  REQUIRED_ARG, 0, COUNT_SET_TYPES + SET_SIP},
    {"not-dipset",  REQUIRED_ARG, 0, COUNT_SET_TYPES + SET_DIP},
    {"not-anyset",  REQUIRED_ARG, 0, COUNT_SET_TYPES + SET_ANY},
    {"not-nhipset", REQUIRED_ARG, 0, COUNT_SET_TYPES + SET_NHIP},
    {0, 0, 0, 0}    /* sentinel */
};

static char *filterOptionsHelp[1+ 2*COUNT_SET_TYPES] = {
    ("Wanted Source IP Address Set\n"
     "\t pass the record if its source IP is in this set"),
    ("Wanted Destination IP Address Set\n"
     "\t pass the record if its destination IP is in this set"),
    ("Wanted Any IP Address Set\n"
     "\t pass the record if its source or destination IP is in this set"),
    ("Wanted Next Hop IP Address Set\n"
     "\t pass the record if its next hop IP is in this set"),
    ("Unwanted Source IP Address Set\n"
     "\t pass the record if its source IP is NOT in this set"),
    ("Unwanted Destination IP Address Set\n"
     "\t pass the record if its destination IP is NOT in this set"),
    ("Unwanted Any IP Address Set\n"
     "\t pass the record if neither its source nor dest IP is in this set"),
    ("Unwanted Next Hop IP Address Set\n"
     "\t pass the record if its next hop IP is NOT in this set"),
    (char*)NULL
};


/*
 *  int dynlib_api_version(void);
 *
 *    Return the dynlib API version that this plugin was compiled
 *    against.
 */
int dynlib_api_version(void)
{
    return DYNLIB_API_VERSION;
}


/*
 *  setup
 *      Called by dynlib interface code to setup this plugin.  This
 *      routine should set up options handling, if required.
 *  Arguments:
 *      --pointer to the dynamic library interface structure
 *      --Application type that is using this plugin.
 *  Returns:
 *      DYNLIB_FAILED - if processing fails
 *      DYNLIB_WONTPROCESS - if application should do normal output
 *      DYNLIB_WILLPROCESS - if this plugin takes over output
 *  Side Effects:
 */
int setup(dynlibInfoStructPtr dlISP, dynlibSymbolId appType)
{
    /* Make the plugin use the application's context, so we can add
     * options to main app. */
    skAppContextSet(dynlibGetAppContext(dlISP));

    /* Initialize variables */
    memset(ipset, 0, sizeof(ipset));
    memset(is_accept, 0, sizeof(is_accept));

    /* add the options */
    if (libOptionsSetup(dlISP, appType)) {
        return DYNLIB_FAILED;
    }

    return DYNLIB_WONTPROCESS;
}


/*
 *  teardown
 *      Called by dynlib interface code to tear down this plugin.
 *  Arguments:
 *      --Application type that is using this plugin
 *  Returns:
 *      None.
 *  Side Effects:
 *      None.
 */
void teardown(dynlibSymbolId UNUSED(appType))
{
    static int teardownFlag = 0;
    int i;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    /* destroy the sets */
    for (i = 0; i < COUNT_SET_TYPES; ++i) {
        if (ipset[i]) {
            skIPTreeDelete(&ipset[i]);
            ipset[i] = NULL;
        }
    }
}


/*
 *  optionsUsage
 *      Called by dynlib interface code to allow this plugin to print
 *      the options it accepts.
 *  Arguments:
 *      --Application type that is using this plugin
 *  Returns:
 *      None.
 *  Side Effects:
 *      None.
 */
void optionsUsage(dynlibSymbolId UNUSED(appType), FILE *fh)
{
    int i;

    if (libOptions == NULL) {
        return;
    }

    for (i = 0; libOptions[i].name != NULL; ++i) {
        fprintf(fh, "--%s %s. %s\n", libOptions[i].name,
                SK_OPTION_HAS_ARG(libOptions[i]), optionsHelp[i]);
    }
}


/*
 *  libOptionsSetup
 *      Sets up the options handler for this plugin
 *  Arguments:
 *      --pointer to the dynamic library interface structure
 *      --Application type that is using this plugin.
 *  Returns:
 *      0 if o.k.; nonzero otherwise.
 *  Side Effects:
 *      Options are registered.
 */
static int libOptionsSetup(dynlibInfoStructPtr dlISP, dynlibSymbolId appType)
{
    if (NULL != libOptions) {
        /* we've been here before */
        return 0;
    }

    /* check that we have the same number of options entries and help*/
    assert((sizeof(filterOptions)/sizeof(struct option))
           == (sizeof(filterOptionsHelp)/sizeof(char*)));

    /* this plug-in only supports rwfilter */
    switch (appType) {
      case DYNLIB_SHAR_FILTER:
        libOptions = filterOptions;
        optionsHelp = filterOptionsHelp;
        break;

      default:
        skAppPrintErr("Cannot use %s plug-in with %s application",
                      pluginName, skAppName());
        return DYNLIB_FAILED;
    }

    /* register the options */
    if (optionsRegister(libOptions, &optionsHandler, (clientData)dlISP)) {
        skAppPrintErr("%s: unable to register options",
                      pluginName);
        return 1;
    }

    return 0;
}


/*
 *  optionsHandler
 *      Called by options parser to handle one user option
 *  Arguments:
 *      clientData cData: the dynamic lib interface struct pointer
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int optionsHandler(clientData cData, int opt_index, char *opt_arg)
{
    static const char *set_name[COUNT_SET_TYPES] = {
        "Source", "Destination", "Any", "Next Hop"
    };
    skIPTree_t **target_set = NULL;

    /* Make sure option in range; else a fatal programmer error */
    if (opt_index < 0 || opt_index >= 2*COUNT_SET_TYPES) {
        skAppPrintErr("Option %d not handled at %s:%d",
                      opt_index, __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }

    /* Verify that we have a file name */
    if (opt_arg == NULL || opt_arg[0] == '\0') {
        skAppPrintErr("Expected name of %s IP-set file",
                      libOptions[opt_index].name);
        return 1;
    }

    /* Which set to read? */
    target_set = &ipset[opt_index % COUNT_SET_TYPES];

    /* For each, only one of accept or reject can be given */
    if (*target_set) {
        skAppPrintErr("%s IP-set already initialized",
                      set_name[opt_index % COUNT_SET_TYPES]);
        return 1;
    }

    /* Is this an accept set? */
    if (opt_index < COUNT_SET_TYPES) {
        is_accept[opt_index] = 1;
    }

    /* Read the set */
    if (skIPTreeLoad(opt_arg, target_set)) {
        skAppPrintErr("Error reading %s IP-set from file '%s'",
                      libOptions[opt_index].name, opt_arg);
        return 1;
    }

    /* Tell the dynlib glue-code that we are active. */
    dynlibMakeActive((dynlibInfoStructPtr)cData);

    return 0;
}


/*
 *  filter
 *      Called by dynlib interface code to filter records
 *  Arguments:
 *      a RW record
 *  Returns:
 *      0 to accept the record, 1 to reject the record
 *  Side Effects:
 *      None
 * Note:
 * A bug I should have caught earlier on; this function originally
 * only did the first applicable check for an IP, so if
 * you specified multiple trees it would only use the first one.
 *
 */
int filter(rwRecPtr rwrec)
{
    if (ipset[SET_SIP]) {
        /* reject the record if its IP is not in the set and the set
         * is an accept set, or if the record's IP is in the set and
         * the set is a reject (!accept) set. */
        if (skIPTreeCheckAddress(rwrec->sIP.ipnum, ipset[SET_SIP])
            ^ is_accept[SET_SIP])
        {
            return 1;
        }
    }

    if (ipset[SET_DIP]) {
        if (skIPTreeCheckAddress(rwrec->dIP.ipnum, ipset[SET_DIP])
            ^ is_accept[SET_DIP])
        {
            return 1;
        }
    }

    if (ipset[SET_ANY]) {
        /* reject the record if neither it's sIP nor dIP is in the set
         * and the set is an accept set, or if the record's sIP or dIP
         * is in the set and the set is a reject (!accept) set. */
        if (is_accept[SET_ANY]
            ^ (skIPTreeCheckAddress(rwrec->sIP.ipnum, ipset[SET_ANY])
               || skIPTreeCheckAddress(rwrec->dIP.ipnum, ipset[SET_ANY])))
        {
            return 1;
        }
    }

    if (ipset[SET_NHIP]) {
        if (skIPTreeCheckAddress(rwrec->nhIP.ipnum, ipset[SET_NHIP])
            ^ is_accept[SET_NHIP])
        {
            return 1;
        }
    }

    return 0;
}
