/*
** Copyright (C) 2003-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  Quick application to print the IPs in a set file
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwset/readset.c,v 1.31 2005/10/31 21:33:53 leechr Exp $");

#include "utils.h"
#include "iptree.h"
#include "print_network_structure.h"

/* LOCAL DEFINES AND TYPEDEFS */

/* Where to print usage (--help) */
#define USAGE_FH stdout


/* LOCAL FUNCTIONS */

/* main routine to process a file */
static void readsetPrintIndent(
    FILE   *fp,
    int     indent);
static void readsetPrintIP(
    FILE           *outstream,
    skBag_key_t     key);
static void readsetPrintKey(
    FILE           *outstream,
    skBag_key_t     key,
    print_state_t  *state);
static void readsetPrintNetwork(FILE *outstream, const skIPTree_t *tree);
static int readsetProcessFile(FILE *outstream, const char *fileName);

static void appUsage(void);                     /* never returns */
static void appUsageLong(void);                 /* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);    /* never returns */
static void appOptionsUsage(FILE *fh);          /* application options usage */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);


/* LOCAL VARIABLES */

/* where to send output */
static FILE *stream_out;

/* name of program to run to page output */
static char *pager = NULL;

/* whether we are using the pager */
static int using_pager = 0;

/* index of first option that is not handled by the options handler.
 * If this is equal to argc, then input is from stdin. */
static int nextOptIndex = 0;

static char output_delimiter = '|';

/* option flags */
static struct {
    /* how to print IPs from iptree.h: enum skIPTreeFormats; def. dot-quad */
    int ip_format;
    /* whether user explicitly gave the print-ips option */
    int print_ips;
    /* whether to count IPs: default no */
    int count_ips;
    /* whether to print statistics; default no */
    int statistics;
    /* output network breakdown of set contents */
    int f_print_network;
    /* print total breakdown */
    int f_print_T;
    /* print class A breakdown */
    int f_print_A;
    /* print class B breakdown */
    int f_print_B;
    /* print class C breakdown */
    int f_print_C;
    /* print /27 breakdown */
    int f_print_X;
    /* print individual hosts */
    int f_print_H;
    /* print summary */
    int f_print_S;
} optFlags = {
    SKIP_IPF_DOT, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};


/* OPTIONS SETUP */

typedef enum _appOptionsEnum {
    READSET_COUNT,
    READSET_PRINT,
    READSET_INTEGER_IPS,
    READSET_ZERO_PAD_IPS,
    READSET_CIDR_BLOCKS,
    READSET_NETWORK_STRUCTURE,
    READSET_STATISTICS,
    READSET_PAGER
} appOptionsEnum;

static struct option appOptions[] = {
    {"count-ips",         NO_ARG,       0, READSET_COUNT},
    {"print-ips",         NO_ARG,       0, READSET_PRINT},
    {"integer-ips",       NO_ARG,       0, READSET_INTEGER_IPS},
    {"zero-pad-ips",      NO_ARG,       0, READSET_ZERO_PAD_IPS},
    {"cidr-blocks",       NO_ARG,       0, READSET_CIDR_BLOCKS},
    {"network-structure", OPTIONAL_ARG, 0, READSET_NETWORK_STRUCTURE},
    {"print-statistics",  NO_ARG,       0, READSET_STATISTICS},
    {"pager",             REQUIRED_ARG, 0, READSET_PAGER},
    {0,0,0,0} /* sentinel entry */
};

static char *appHelp[] = {
    /* add help strings here for the applications options */
    "Print the number of IPs; disables default printing of IPs",
    "Also print IPs when count or statistics switch is given",
    "Print IPs as integers. Def. No",
    "Print IPs as dotted quad with 3 digits per octet. Def. No",
    "Print IPs in CIDR block notation. Def. No",
    ("Print the network structure of the set. Def. TS\n"
     "\tChoose from TABCXHS for counts by Total address space, /8, /16, /24,\n"
     "\t/27, and/or /32; with S provides roll-up summaries."),
    "Print set statistics (min-/max-ip, etc). Def. No",
    "Program to invoke to page output. Def. $SILK_PAGER or $PAGER",
    (char *)NULL /* sentinel entry */
};


/*
 *  appUsage();
 *
 *    Print a message to STDERR tell the user to run --help for usage,
 *    and then exit the application.
 */
void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 *  appUsageLong();
 *
 *    This function is called by the options processing code to
 *    generate usage output when the "--help" switch is given; the
 *    function should be passed in the optionsSetup() call.
 *
 *    The function prints a command summary and complete switch
 *    information to the USAGE_FH, and then exits the application.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                             \
    ("%s [options] [IPset-file]\n"                                            \
     "\tBy default, prints the IPs in the specified IPset-file.  Use\n"       \
     "\tswitches to control format of the outout and to optionally or\n"      \
     "\tadditionally print the number of IPs in the file, the network\n"      \
     "\tstructure, or other statistics.  If no IPset-file is given on\n"      \
     "\tthe command line, the IPset will be read from the standard input.\n")

    fprintf(USAGE_FH, USAGE_MSG, skAppName());
    appOptionsUsage(USAGE_FH);

    exit(1);
}


/*
 *  appOptionsUsage(fh);
 *
 *    Loops through all of the options to print their names and their
 *    help strings.  Output is written to the file handle 'fh'.
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\noptions:\n");
    for (i = 0; appOptions[i].name; i++) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }

    return;
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    if (using_pager) {
        skClosePager(stream_out, pager);
    }
    optionsTeardown();

    return;
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* Register application and initialize globals */
    skAppRegister(argv[0]);
    stream_out = stdout;

    /* Setup the options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);
    }

    /* Register our options and options handler */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);
    }

    /* parse options */
    nextOptIndex = optionsParse(argc, argv);
    assert(nextOptIndex <= argc);
    if (nextOptIndex < 0) {
        appUsage();             /* never returns */
    }

    /* either need name of set file(s) after options or a set file on stdin */
    if ((nextOptIndex == argc) && (FILEIsATty(stdin))) {
        skAppPrintErr("No files on the command line and"
                      " stdin is connected to a terminal");
        appUsage();
    }

    /* If no output was specified, set network printing to print all
     * hosts */
    if (!optFlags.statistics && !optFlags.count_ips
        && !optFlags.f_print_network && !optFlags.print_ips)
    {
        optFlags.f_print_network = 1;
        optFlags.f_print_H = 1;
    }

    /* If printing of IPs was requested but network-printing and CIDR
     * output were not, set network printing to output all hosts. */
    if (optFlags.print_ips && !optFlags.f_print_network
        && !(optFlags.ip_format & SKIP_IPF_CIDR))
    {
        optFlags.f_print_network = 1;
        optFlags.f_print_H = 1;
    }

    /* Open the pager when printing the IPs or network structure */
    if (optFlags.print_ips || optFlags.f_print_network) {
        using_pager = skOpenPagerWhenStdoutTty(&stream_out, &pager);
        if (using_pager < 0) {
            exit(EXIT_FAILURE);
        }
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;                     /* OK */
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    unsigned int i;

    switch ((appOptionsEnum)opt_index) {
      case READSET_STATISTICS:
        optFlags.statistics = 1;
        break;

      case READSET_COUNT:
        optFlags.count_ips = 1;
        break;

      case READSET_INTEGER_IPS:
        optFlags.print_ips = 1;
        if ((optFlags.ip_format & ~SKIP_IPF_CIDR) != SKIP_IPF_DOT) {
            skAppPrintErr("Warning: overriding previous output format;\n"
                          "\t will print IPs as integers.");
        }
        optFlags.ip_format = (SKIP_IPF_DEC
                              | (optFlags.ip_format & SKIP_IPF_CIDR));
        break;

      case READSET_ZERO_PAD_IPS:
        optFlags.print_ips = 1;
        if ((optFlags.ip_format & ~SKIP_IPF_CIDR) != SKIP_IPF_DOT) {
            skAppPrintErr("Warning: overriding previous output format;\n"
                          "\t will print IPs as three-digit dotted decimal.");
        }
        optFlags.ip_format = (SKIP_IPF_ZERO
                              | (optFlags.ip_format & SKIP_IPF_CIDR));
        break;

      case READSET_CIDR_BLOCKS:
        optFlags.print_ips = 1;
        optFlags.ip_format |= SKIP_IPF_CIDR;
        break;

      case READSET_NETWORK_STRUCTURE:
        /* Validate that this is a legal string */
        if (opt_arg != NULL) {
            for (i = 0; i < strlen(opt_arg); i++) {
                if (strchr("TABCXHS", opt_arg[i]) == NULL) {
                    skAppPrintErr("Illegal network structure output option.\n"
                                  "\t Should be some combination of TABCXHS.");
                    return 1;
                }
            }
            optFlags.f_print_T = (strchr(opt_arg, 'T') != NULL);
            optFlags.f_print_A = (strchr(opt_arg, 'A') != NULL);
            optFlags.f_print_B = (strchr(opt_arg, 'B') != NULL);
            optFlags.f_print_C = (strchr(opt_arg, 'C') != NULL);
            optFlags.f_print_X = (strchr(opt_arg, 'X') != NULL);
            optFlags.f_print_H = (strchr(opt_arg, 'H') != NULL);
            optFlags.f_print_S = (strchr(opt_arg, 'S') != NULL);
        } else {
            /*
             * default if --network-structure is used, but not
             * qualified with an option string.
             */
            optFlags.f_print_T = 1;
            optFlags.f_print_S = 1;
        }
        optFlags.f_print_network = 1;
        break;

      case READSET_PRINT:
        optFlags.print_ips = 1;
        break;

      case READSET_PAGER:
        pager = opt_arg;
        break;
    }

    return 0;                   /* OK */
}


/*
 * print an indentation to a given output stream
 */
#define INDENT_STRING "  "
static void readsetPrintIndent(
    FILE   *fp,
    int     indent)
{
    int i;
    for (i = 0; i < indent; i++) {
        fprintf(fp, INDENT_STRING);
    }
}


/*
 * print a single IP address according to output options
 */
static void readsetPrintIP(
    FILE           *outstream,
    skBag_key_t     key)
{
    switch (optFlags.ip_format) {
      case SKIP_IPF_DEC:
        fprintf(outstream, "%u", (uint32_t)key);
        break;
      case SKIP_IPF_HEX:
        fprintf(outstream, "0x%08x", (uint32_t)key);
        break;
      case SKIP_IPF_ZERO:
        fprintf(outstream, "%s", num2dot0((uint32_t)key));
        break;
      case SKIP_IPF_DOT:
      default:
        fprintf(outstream, "%s", num2dot((uint32_t)key));
        break;
    }
}


/*
 * readsetPrintKey
 *
 * print an individual address out of the set, and maintain state of
 * what addresses have been seen
 */
static void readsetPrintKey(
    FILE           *outstream,
    skBag_key_t     key,
    print_state_t  *state)
{
    int change_A, change_B, change_C, change_X;

    /*
     * calculate whether or not a given octet has changed.  since
     * the calculation is done to determine whether or not we need
     * to print new headers/footers, processing the first entry
     * and reaching the end of the data set counts as a change.
     */
    if (state->f_last_entry) {
        change_A = change_B = change_C = change_X = 1;
    } else if (state->f_first_entry) {
        state->f_first_entry = 0;
        change_A = change_B = change_C = change_X = 1;
    } else {
        change_A = (OCTET_A(key) != OCTET_A(state->prev_key));
        change_B = (OCTET_B(key) != OCTET_B(state->prev_key));
        change_C = (OCTET_C(key) != OCTET_C(state->prev_key));
        change_X = (OCTET_X(key) != OCTET_X(state->prev_key));
    }

    /*
     * if counting and summarizing, calculate changes in blocks
     * and close out blocks.
     */

    /* Close out old X block (/27)  */
    if (optFlags.f_print_X && ((state->class_X_seen_H > 0) &&
                               (change_X || change_C ||
                                change_B || change_A)))
    {
        readsetPrintIndent(outstream,
                           optFlags.f_print_T +
                           optFlags.f_print_A +
                           optFlags.f_print_B +
                           optFlags.f_print_C);
        readsetPrintIP(outstream, state->prev_key & MASK_ABCX);
        fprintf(outstream, "/27");
        fprintf(outstream, "%c", output_delimiter);
        if ( optFlags.f_print_S ) {
            fprintf( outstream, PRINT_X_SUMMARY, state->class_X_seen_H );
        } else {
            fprintf( outstream, PRINT_HOSTCOUNT, state->class_X_seen_H );
        }
        fprintf(outstream, "\n");

        /* reset sub-block counters */
        state->class_X_seen_H = 0;
    }

    /* Close out old C block */
    if (optFlags.f_print_C && ((state->class_C_seen_H > 0) &&
                               (change_C || change_B ||
                                change_A)))
    {
        readsetPrintIndent(outstream,
                           optFlags.f_print_T +
                           optFlags.f_print_A +
                           optFlags.f_print_B);
        readsetPrintIP(outstream, state->prev_key & MASK_ABC);
        fprintf(outstream, "/24");
        fprintf(outstream, "%c", output_delimiter);
        if ( optFlags.f_print_S ) {
            fprintf( outstream, "%c", output_delimiter );
            fprintf(outstream, PRINT_C_SUMMARY, state->class_C_seen_H,
                    state->class_C_seen_X);
        } else {
            fprintf( outstream, PRINT_HOSTCOUNT, state->class_C_seen_H );
        }
        fprintf(outstream, "\n");

        /* reset sub-block counters */
        state->class_C_seen_X = 0;
        state->class_C_seen_H = 0;
    }

    /* Close out old B block */
    if (optFlags.f_print_B && ((state->class_B_seen_H > 0) &&
                               (change_B || change_A)))
    {
        readsetPrintIndent(outstream,
                           optFlags.f_print_T +
                           optFlags.f_print_A);
        readsetPrintIP(outstream, state->prev_key & MASK_AB);
        fprintf(outstream, "/16");
        fprintf(outstream, "%c", output_delimiter);
        if ( optFlags.f_print_S ) {
            fprintf(outstream, PRINT_B_SUMMARY, state->class_B_seen_H,
                    state->class_B_seen_C, state->class_B_seen_X);
        } else {
            fprintf( outstream, PRINT_HOSTCOUNT, state->class_B_seen_H );
        }
        fprintf(outstream, "\n");

        /* reset sub-block counters */
        state->class_B_seen_C = 0;
        state->class_B_seen_X = 0;
        state->class_B_seen_H = 0;
    }


    /* Close out old A block */
    if (optFlags.f_print_A && ((state->class_A_seen_H > 0) && change_A))
    {
        readsetPrintIndent(outstream,
                           optFlags.f_print_T);
        readsetPrintIP(outstream, state->prev_key & MASK_A);
        fprintf(outstream, "/8");
        fprintf(outstream, "%c", output_delimiter);
        if ( optFlags.f_print_S ) {
            fprintf(outstream, PRINT_A_SUMMARY, state->class_A_seen_H,
                    state->class_A_seen_B, state->class_A_seen_C, 
                    state->class_A_seen_X);
        } else {
            fprintf( outstream, PRINT_HOSTCOUNT, state->class_A_seen_H );
        }
        fprintf(outstream, "\n");

        /* reset sub-block counters */
        state->class_A_seen_B = 0;
        state->class_A_seen_C = 0;
        state->class_A_seen_X = 0;
        state->class_A_seen_H = 0;
    }


    /*
     * Now that we've closed the footers, if we are at the end of
     * the data set, we can quit.
     */
    if (state->f_last_entry) {
        /* Close out old total block */
        if ( optFlags.f_print_T ) {
            fprintf( outstream, "TOTAL" );
            fprintf( outstream, "%c", output_delimiter );
            if ( optFlags.f_print_S ) {
                fprintf( outstream, PRINT_T_SUMMARY, 
                         state->seen_H, 
                         state->seen_A, state->seen_B, state->seen_C, 
                         state->seen_X );
            } else {
                /* don't use PRINT_HOSTCOUNT because seen_H is 64b */
                fprintf( outstream, ("%" PRIu64), state->seen_H );
            }
            fprintf( outstream, "\n" );
        }
        return;
    }

    /* Open new A block */
    if (change_A) {
        /* increment counters of class As */
        ++state->seen_A;

        /* wrt this class A, no class Bs, class Cs, /27s or hosts yet */
        state->class_A_seen_B = 0;
        state->class_A_seen_C = 0;
        state->class_A_seen_X = 0;
        state->class_A_seen_H = 0;
    }

    /* Open new B block */
    if (change_B || change_A) {
        /* increment counters of class Bs */
        ++state->seen_B;
        ++state->class_A_seen_B;

        /* wrt this class B, no class Cs, /27s, or hosts yet */
        state->class_B_seen_C = 0;
        state->class_B_seen_X = 0;
        state->class_B_seen_H = 0;
    }

    /* Open new C block */
    if (change_C || change_B || change_A) {
        /* increment counters of class Cs */
        ++state->seen_C;
        ++state->class_A_seen_C;
        ++state->class_B_seen_C;

        /* wrt this class C, no /27s or hosts yet */
        state->class_C_seen_X = 0;
        state->class_C_seen_H = 0;
    }

    /* Open new "X" block */
    if (change_X || change_C || change_B || change_A)
    {
        /* increment counters of /27s */
        ++state->seen_X;
        ++state->class_A_seen_X;
        ++state->class_B_seen_X;
        ++state->class_C_seen_X;

        /* wrt this /27, no hosts yet */
        state->class_X_seen_H = 0;
    }

    /* increment individual hosts */
    ++state->seen_H;
    ++state->class_A_seen_H;
    ++state->class_B_seen_H;
    ++state->class_C_seen_H;
    ++state->class_X_seen_H;

    /* print hosts if necessary. */
    if (optFlags.f_print_H) {
        readsetPrintIndent(outstream,
                           optFlags.f_print_T +
                           optFlags.f_print_A +
                           optFlags.f_print_B +
                           optFlags.f_print_C +
                           optFlags.f_print_X);
        readsetPrintIP(outstream, (uint32_t) key);
        fprintf(outstream, "\n");
    }

    state->prev_key = key;
}


/*
 * readsetPrintNetwork
 *
 */
static void readsetPrintNetwork(
    FILE               *outstream,
    const skIPTree_t   *tree)
{
    print_state_t state;
    skIPTreeIterator_t *iter;
    uint32_t ipaddr;

    memset(&state, 0, sizeof(state));
    skIPTreeIteratorCreate(&iter, tree);
    state.f_first_entry = 1;

    while (skIPTreeIteratorNext(&ipaddr, iter) == SKIP_ITERATOR_OK) {
        readsetPrintKey(outstream, ipaddr, &state);
    }

    /*
     * set the last key flag and call it once more, for good measure.
     * (that way, it closes out blocks after the last key.)
     */
    state.f_last_entry = 1;
    readsetPrintKey(outstream, ipaddr, &state);
}


/*
 *  readsetPrintStatistics
 *
 *    Prints, to outF, statistics of the IPTree tree.  Statistics
 *    printed are the minimum IP, the maximum IP, a count of the class
 *    A blocks, class B blocks (number of nodes), class C blocks, and
 *    a count of the addressBlocks (/27's) used.  If integerIP is 0,
 *    the min and max IPs are printed in dotted-quad form; otherwise
 *    they are printed as integers.
 */
static void readsetPrintStatistics(
    FILE               *outF,
    const skIPTree_t   *tree,
    int                 ip_format)
{
    skIPTreeIterator_t *iter;
    uint32_t ipaddr;
    uint32_t old_addr;
    uint32_t num_slash_8;
    uint32_t num_slash_16;
    uint32_t num_slash_24;
    uint32_t num_slash_27;
    uint32_t min_ip;
    uint32_t max_ip;
    uint64_t num_hosts;

    if (skIPTreeIteratorCreate(&iter, tree)) {
        return;
    }

    /* Get first IP */
    if (skIPTreeIteratorNext(&ipaddr, iter) != SKIP_ITERATOR_OK) {
        /* empty tree */
        skIPTreeIteratorDestroy(&iter);
        return;
    }

    /* first IP */
    min_ip = ipaddr;
    old_addr = ipaddr;
    num_hosts    = 1;
    num_slash_8  = 1;
    num_slash_16 = 1;
    num_slash_24 = 1;
    num_slash_27 = 1;

    /* Handle remaining IPs */
    while (skIPTreeIteratorNext(&ipaddr, iter) == SKIP_ITERATOR_OK) {
        if (OCTET_A(old_addr) != OCTET_A(ipaddr)) {
            ++num_slash_8;
            ++num_slash_16;
            ++num_slash_24;
            ++num_slash_27;
        } else if (OCTET_B(old_addr) != OCTET_B(ipaddr)) {
            ++num_slash_16;
            ++num_slash_24;
            ++num_slash_27;
        } else if (OCTET_C(old_addr) != OCTET_C(ipaddr)) {
            ++num_slash_24;
            ++num_slash_27;
        } else if (OCTET_X(old_addr) != OCTET_X(ipaddr)) {
            ++num_slash_27;
        }            
        ++num_hosts;
        old_addr = ipaddr;
    }

    /* Get the max */
    max_ip = ipaddr;

    fprintf(outF, "Network Summary\n");
    fprintf(outF, "\tminimumIP = ");
    skIPTreePrintAddr(outF, min_ip, ip_format);
    fprintf(outF, "\tmaximumIP = ");
    skIPTreePrintAddr(outF, max_ip, ip_format);

    fprintf(outF, ("\t%" PRIu64 " hosts, %f%% of 2^32\n"),
            num_hosts, 100 * ((double) num_hosts) /
            ((double) ((uint64_t) 1 << 32)));
    fprintf(outF, "\t%d occupied /8s, %f%% of 2^8\n",
            num_slash_8, 100 * ((double) num_slash_8) /
            ((double) (1 << 8)));
    fprintf(outF, "\t%d occupied /16s, %f%% of 2^16\n",
            num_slash_16, 100 * ((double) num_slash_16) /
            ((double) (1 << 16)));
    fprintf(outF, "\t%d occupied /24s, %f%% of 2^24\n",
            num_slash_24, 100 * ((double) num_slash_24) /
            ((double) (1 << 24)));
    fprintf(outF, "\t%d occupied /27s, %f%% of 2^27\n",
            num_slash_27, 100 * ((double) num_slash_27) /
            ((double) (1 << 27)));

    skIPTreeIteratorDestroy(&iter);
    return;
}


/*
 *  readsetProcessFile(fh, fileName);
 *
 *    Open the binary set-file given in 'fileName', read the ipTree
 *    from it; then print the output requested by the user to the
 *    stream 'fh'.  Return 0 on success, or 1 if the ipTree cannot be
 *    read from 'fileName'.
 */
static int readsetProcessFile(FILE *outstream, const char *fileName)
{
    uint64_t count;
    skIPTree_t *dataTree = NULL;

    /* Read ipTree from file */
    if (0 == strcmp("stdin", fileName)) {
        if (0 != skIPTreeRead(stdin, &dataTree)) {
            skAppPrintErr("error reading binary IP set from stdin");
            return 1;
        }
    } else if (0 != skIPTreeLoad(fileName, &dataTree)) {
        skAppPrintErr("error reading binary IP set from '%s'", fileName);
        return 1;
    }

    if (optFlags.count_ips) {
        count = skIPTreeCountIPs(dataTree);
        fprintf(outstream, ("%s: %" PRIu64 "\n"), fileName, count);
    }

    if (optFlags.ip_format & SKIP_IPF_CIDR) {
        skIPTreePrint(outstream, dataTree, optFlags.ip_format);
    }

    if (optFlags.f_print_network) {
        readsetPrintNetwork(outstream, dataTree);
    }

    if (optFlags.statistics) {
        readsetPrintStatistics(outstream, dataTree, optFlags.ip_format);
    }

    return 0;
}


int main(int argc, char **argv)
{
    int i;

    appSetup(argc, argv);                       /* never returns on error */

    if (nextOptIndex == argc) {
        (void)readsetProcessFile(stream_out, "stdin");
    } else {
        for (i = nextOptIndex; i < argc; ++i) {
            (void)readsetProcessFile(stream_out, argv[i]);
        }
    }

    /* done */
    appTeardown();

    return 0;
}
