/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwbagcat
**
**  Build binary file from flow records
**
*/


#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwset/rwbag.c,v 1.13 2005/05/20 17:55:05 thomasm Exp $");

#include "utils.h"
#include "rwpack.h"
#include "bagtree.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* where to write --help output */
#define USAGE_FH stdout

/* number of potential outputs: {src,dest} * {flows,pkts,bytes} */
#define NUM_OUTPUTS 6
#define NUM_ARRAY_OUTPUTS 9

typedef struct {
    skBag_header_t *bag;
    const char     *pathname;
    FILE           *filep;
    int             is_pipe;
} bag_io_t;

typedef struct {
    skBag_counter_t *array;
    skBag_key_t    size;
    const char     *pathname;
    FILE           *filep;
    int             is_pipe;
} array_io_t;

/* EXPORTED FUNCTIONS */


/* LOCAL FUNCTIONS */

static void appUsage(void);		     /* never returns */
static void appUsageLong(void);		     /* never returns */
static void appOptionsUsage(FILE*);	     /* application options usage */
static void appTeardown(void);
static void appSetup(int argc, char **argv); /* never returns when error */
static int  appOptionsHandler(clientData, int, const char *);
static int  prepareBagFile(bag_io_t *);
static int  prepareArrayBagFile(array_io_t *);


/* LOCAL VARIABLES */

/* The potential bag files to create */
static bag_io_t bag_io[NUM_OUTPUTS];
/* The potential portbag files to create */
static array_io_t array_io[NUM_ARRAY_OUTPUTS];

/* Magic constants to build the bags */
static const skBag_level_t levels = (skBag_level_t) 4;
static skBag_levelsize_t rg_level_size[4] = { 9, 9, 9, 5 };

/* First file index in argv[] after handling all options */
static int  first_file;

/* True if an overflow warning has been printed for given stream */
static int overflow_warning_printed[NUM_OUTPUTS];
static int array_overflow_warning_printed[NUM_ARRAY_OUTPUTS];


/* OPTIONS SETUP */

typedef enum _appOptionsEnum {
    /* These MUST be kept in order with the options */
    SRC_FLOWS=0,  SRC_PKTS,  SRC_BYTES,
    DEST_FLOWS, DEST_PKTS, DEST_BYTES,
    SRC_PORT_FLOWS, SRC_PORT_PKTS, SRC_PORT_BYTES,
    DEST_PORT_FLOWS, DEST_PORT_PKTS, DEST_PORT_BYTES,
    PROTO_FLOWS, PROTO_PKTS, PROTO_BYTES
} appOptionsEnum;

/* This #define is used to adjust the indices for the array-based
 * bags.  It must be the first enum item that represents an array bag
 * instead of a tree bag. */
#define AB_OFFSET SRC_PORT_FLOWS

static struct option appOptions[] = {
    {"sf-file", REQUIRED_ARG, 0, SRC_FLOWS},
    {"sp-file", REQUIRED_ARG, 0, SRC_PKTS},
    {"sb-file", REQUIRED_ARG, 0, SRC_BYTES},
    {"df-file", REQUIRED_ARG, 0, DEST_FLOWS},
    {"dp-file", REQUIRED_ARG, 0, DEST_PKTS},
    {"db-file", REQUIRED_ARG, 0, DEST_BYTES},
    {"port-sf-file", REQUIRED_ARG, 0, SRC_PORT_FLOWS},
    {"port-sp-file", REQUIRED_ARG, 0, SRC_PORT_PKTS},
    {"port-sb-file", REQUIRED_ARG, 0, SRC_PORT_BYTES},
    {"port-df-file", REQUIRED_ARG, 0, DEST_PORT_FLOWS},
    {"port-dp-file", REQUIRED_ARG, 0, DEST_PORT_PKTS},
    {"port-db-file", REQUIRED_ARG, 0, DEST_PORT_BYTES},
    {"proto-f-file", REQUIRED_ARG, 0, PROTO_FLOWS},
    {"proto-p-file", REQUIRED_ARG, 0, PROTO_PKTS},
    {"proto-b-file", REQUIRED_ARG, 0, PROTO_BYTES},
    {0,0,0,0}  /* sentinel entry */
};

static const char *appHelp[] = {
    "Write flow count by unique source IP into named file",
    "Write packet count by unique source IP into named file",
    "Write byte count by unique source IP into named file",
    "Write flow count by unique destination IP into named file",
    "Write packet count by unique destination IP into named file",
    "Write byte count by unique destination IP into named file",
    "Write flow count by unique source port into named file",
    "Write packet count by unique source port into named file",
    "Write byte count by unique source port into named file",
    "Write flow count by unique destination port into named file",
    "Write packet count by unique destination port into named file",
    "Write byte count by unique destination port into named file",
    "Write flow count by unique protocol into named file",
    "Write packet count by unique protocol into named file",
    "Write byte count by unique protocol into named file",
    (char *)NULL
};

/*
 * stdout_used is set to 1 by prepareBagFile and prepareArrayBagFile when
 * a file is to be written to stdout. 
 */
static int stdout_used = 0;


/*
 * appUsage:
 * 	print short usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 * 	print complete usage information to USAGE_FH and exit with
 * 	code.  passed to optionsSetup() to print usage when --help
 * 	option given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
    fprintf(USAGE_FH,
            ("Usage: %s [app-options] [files]\n"
             "\t Reads flow records and builds bag(s).  It supports using\n"
             "\t source or destination IP address or port as the key, and\n"
             "\t flow, packets, or bytes as the counter.  Reads from stdin\n"
             "\t or named files.\n"), skAppName());

    appOptionsUsage(USAGE_FH);
    exit(EXIT_FAILURE);
}


/*
 * appOptionsUsage:
 * 	print options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *	do NOT add non-option usage here (i.e., required and other
 *	optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\napp options\n");
    for (i = 0; appOptions[i].name; i++) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                (appOptions[i].has_arg
                 ? ((appOptions[i].has_arg == 1)
                    ? "Required Arg"
                    : "Optional Arg")
                 : "No Arg"), appHelp[i]);
    }

    return;
}


/*
 * appTeardown:
 *	teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 * 	All modules are torn down. Then application teardown takes place.
 * 	Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;
    int i;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    /* close all bag files */
    for (i = 0; i < NUM_OUTPUTS; ++i) {
        if (bag_io[i].bag) {
            skBag_free(bag_io[i].bag);
        }
        if (bag_io[i].filep) {
            if (bag_io[i].is_pipe == 1) {
                if (-1 == pclose(bag_io[i].filep)) {
                    skAppPrintErr("Error closing %s: %s",
                                  bag_io[i].pathname, strerror(errno));
                }
            } else if (bag_io[i].is_pipe == 0) {
                if (EOF == fclose(bag_io[i].filep)) {
                    skAppPrintErr("Error closing %s: %s",
                                  bag_io[i].pathname, strerror(errno));
                }
            }
            
        }
        if (bag_io[i].pathname) {
            free((char*)bag_io[i].pathname);
        }
        memset(&bag_io[i], 0, sizeof(bag_io_t));
    }

    /* close all array bag files */
    for (i = 0; i < NUM_ARRAY_OUTPUTS; ++i) {
	if (array_io[i].array) {
	    free(array_io[i].array);
	    array_io[i].array = NULL;
	}
	if (array_io[i].filep) {
	    if (array_io[i].is_pipe == 1) {
		if (-1 == pclose(array_io[i].filep)) {
		    skAppPrintErr("Error closing %s: %s",
				  array_io[i].pathname, strerror(errno));
		}
	    } else if (array_io[i].is_pipe == 0) {
		if (EOF == fclose(array_io[i].filep)) {
		    skAppPrintErr("Error closing %s: %s",
				  array_io[i].pathname, strerror(errno));
		}
	    }
	}
	if (array_io[i].pathname) {
	    free((char*)array_io[i].pathname);
	}
	memset(&array_io[i], 0, sizeof(bag_io_t));
    }

    optionsTeardown();
    return;
}


/*
 * appSetup
 *	do all the setup for this application include setting up required
 *	modules etc.
 * Arguments:
 *	argc, argv
 * Returns: None.
 * Side Effects:
 *	exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int output_count;
    int i;

    skAppRegister(argv[0]);

    /* initialize variables */
    memset(bag_io, 0, sizeof(bag_io));
    memset(array_io, 0, sizeof(array_io));
    memset(overflow_warning_printed, 0, sizeof(overflow_warning_printed));
    memset(array_overflow_warning_printed, 0,
	   sizeof(array_overflow_warning_printed));

    /* first do some sanity checks */

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* initialize options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);			/* never returns */
    }

    /* register the options handler */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL))
    {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);			/* never returns */
    }

    /* parse options */
    if ( (first_file = optionsParse(argc, argv)) < 0) {
        /* options parsing should print error */
        appUsage();		/* never returns */
    }

    /* first_file is looking at first file name to process */
    assert(first_file <= argc);
    if (first_file == argc) {
        if (FILEIsATty(stdin)) {
            skAppPrintErr("No input files on command line and"
                          " stdin is connected to a terminal");
            appUsage();
        }
    }

    /* Open the output files and allocate the bags */
    output_count = 0;
    for (i = 0; i < NUM_OUTPUTS; ++i) {
        if (bag_io[i].pathname) {
            if (0 != prepareBagFile(&bag_io[i])) {
                appUsage();
            }
            output_count++;
        }
    }

    for (i = 0; i < NUM_ARRAY_OUTPUTS; ++i) {
	if (array_io[i].pathname) {
	    if (0 != prepareArrayBagFile(&array_io[i])) {
		appUsage();
	    }
	    output_count++;
	}
    }

    if (output_count == 0) {
        skAppPrintErr("Must specify type of output(s) to generate.");
        appUsage();
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;			/* OK */
}


/*
 * appOptionsHandler:
 * 	called for each option the app has registered.
 * Arguments:
 *	clientData cData: ignored here.
 *	int index: index into appOptions of the specific option
 *	char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *	0 if OK. 1 else.
 * Side Effects:
 *	Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    const char *opt_arg)
{
    switch (opt_index) {
      case SRC_FLOWS:
      case SRC_PKTS:
      case SRC_BYTES:
      case DEST_FLOWS:
      case DEST_PKTS:
      case DEST_BYTES:
        if (!opt_arg || 0 == strlen(opt_arg)) {
            skAppPrintErr("missing file name for --%s option",
                          appOptions[opt_index].name);
            return 1;
        }
        if (bag_io[opt_index].pathname) {
            skAppPrintErr("Option %s given multiple times",
                          appOptions[opt_index].name);
            return 1;
        }
        bag_io[opt_index].pathname = strdup(opt_arg);
        if (!bag_io[opt_index].pathname) {
            skAppPrintErr("strdup failed");
            return 1;
        }
        break;

      case SRC_PORT_FLOWS:
      case SRC_PORT_PKTS:
      case SRC_PORT_BYTES:
      case DEST_PORT_FLOWS:
      case DEST_PORT_PKTS:
      case DEST_PORT_BYTES:
      case PROTO_FLOWS:
      case PROTO_PKTS:
      case PROTO_BYTES:
	  if (!opt_arg || 0 == strlen(opt_arg)) {
	      skAppPrintErr("missing file name for --%s option",
			    appOptions[opt_index].name);
	      return 1;
	  }
	  if (array_io[opt_index - AB_OFFSET].pathname) {
	      skAppPrintErr("Option %s given multiple times",
			    appOptions[opt_index].name);
	      return 1;
	  }
	  array_io[opt_index - AB_OFFSET].pathname =
	      strdup(opt_arg);
	  if (!array_io[opt_index - AB_OFFSET].pathname) {
	      skAppPrintErr("strdup failed");
	      return 1;
	  }
	  array_io[opt_index - AB_OFFSET].size = 0x10000;
	  if ( (opt_index == PROTO_FLOWS) || (opt_index == PROTO_PKTS) ||
	       (opt_index == PROTO_BYTES) ) {
	      array_io[opt_index - AB_OFFSET].size = 0x100;
	  }
	  break;

      default:
        skAppPrintErr("Option %d not handled in switch() at %s:%d",
                      opt_index, __FILE__, __LINE__);
        abort();
    }

    return 0;			/* OK */
}


/*
 *  ok = prepareBagFile(bag_io);
 *
 *    Prepare the output bag file 'bag_io', which is a bag_io_t with
 *    the pathname filled in.
 *
 *    This function opens the output file and allocates a bag that
 *    will be writing to that file.  The function makes sure a file
 *    with that name does not currently exist.  If pathname is
 *    "stdout" and no other bag files are writing to stdout, then
 *    stdout will be used.
 *
 *    Returns 0 on success.  Returns non-zero if allocation files, if
 *    we attempt to open an existing file, if more than one bag use
 *    stdout, or if the fopen() call fails.
 */
static int prepareBagFile(bag_io_t *bag)
{
    if (0 == strcmp("stdout", bag->pathname)) {
        if (stdout_used) {
            skAppPrintErr("Only one output may use stdout");
            return 1;
        }
        stdout_used = 1;
        bag->filep = stdout;
        bag->is_pipe = -1;
    } else if (fileExists(bag->pathname)) {
        skAppPrintErr("Will not overwrite existing file %s", bag->pathname);
        return 1;
    } else {
        if (openFile(bag->pathname, 1 /* write */,
                     &bag->filep, &bag->is_pipe))
        {
            return 1;
        }
    }

    if (skBag_alloc(&bag->bag, levels, rg_level_size) != SKBAG_OK) {
        skAppPrintErr("Error allocating bag tree");
        return 1;
    }

    return 0; /* OK */
}


/*
 *  ok = prepareArrayBagFile(array_io);
 *
 *    Prepare the output array bag file 'array_io', which is an
 *    array_io_t with the pathname and array size filled in.
 *
 *    This function opens the output file and allocates an array that
 *    will be written to that file.  The function makes sure a file
 *    with that name does not currently exist.  if pathname is
 *    "stdout" and no other files are writing to stdout, then stdout
 *    will be used.
 *
 *    Returns 0 on success.  Returns non-zero if allocation fails, if
 *    we attempt to open an existing file, if more than one bag uses
 *    stdout, or if the fopen() call fails.
 */
static int prepareArrayBagFile(array_io_t *arraybag)
{
    if (0 == strcmp("stdout", arraybag->pathname)) {
	if (stdout_used) {
	    skAppPrintErr("Only one output may use stdout");
	    return 1;
	}
	stdout_used = 1;
	arraybag->filep = stdout;
	arraybag->is_pipe = -1;
    } else if (fileExists(arraybag->pathname)) {
	skAppPrintErr("Will not overwrite existing file %s",
		      arraybag->pathname);
	return 1;
    } else {
	if (openFile(arraybag->pathname, 1 /* write */,
		     &arraybag->filep, &arraybag->is_pipe)) {
	    return 1;
	}
    }
    arraybag->array = (skBag_counter_t *) calloc(arraybag->size,
                                                 sizeof(skBag_counter_t));
    if (NULL == arraybag->array) {
	skAppPrintErr("Error allocating bag array");
	return 1;
    }
    return 0; /* OK */
}

/*
 *  ok = processFile(path);
 *
 *    Read the RW Records from the file at 'path'---path can be
 *    "stdin"---and potentially create bag files for {sIP,dIP} x
 *    {flows,pkts,bytes}.  Return 0 if successful; non-zero otherwise.
 */
static int processFile(const char *path)
{
    const skBag_counter_t max_val = (skBag_counter_t)(~0);
    rwIOStruct *rwIOS;
    skBag_err_t err;
    rwRec rwrec;
    int rv = -1; /* return value: assume error */
    skBag_counter_t count; /* used for overflow testing for array bags */
    skBag_counter_t old_count;

    skBag_counter_t temp;

    skBag_counter_t foo;

    if (! (rwIOS = rwOpenFile(path, NULL))) {
        /* rwOpenFile should print error */
        return rv;
    }

    while (rwRead(rwIOS, &rwrec)) {
        /* For Src and Dest flows, just increment the counter for each
           matching record */
        if (bag_io[SRC_FLOWS].bag) {
            if (SKBAG_OK != skBag_incrCounter(bag_io[SRC_FLOWS].bag,
                                              (skBag_key_t*)&rwrec.sIP.ipnum))
            {
                skAppPrintErr("Unexpected error getting value from bag");
                goto END;
            }
        }
        if (bag_io[DEST_FLOWS].bag) {
            if (SKBAG_OK != skBag_incrCounter(bag_io[DEST_FLOWS].bag,
                                              (skBag_key_t*)&rwrec.dIP.ipnum))
            {
                skAppPrintErr("Unexpected error getting value from bag");
                exit(EXIT_FAILURE);
            }
        }

        /* For Packets and Bytes on Src and Dest, caller helper
           function to add the value */
        if (bag_io[SRC_PKTS].bag) {
            temp = rwrec.pkts;
            err = skBag_addToCounter(bag_io[SRC_PKTS].bag, 
                                     (skBag_key_t*)&rwrec.sIP.ipnum,
                                     &temp);
            skBag_getCounter( bag_io[SRC_PKTS].bag, 
                              (skBag_key_t*)&rwrec.sIP.ipnum,
                              (skBag_counter_t*)&foo);

            if (err != SKBAG_OK) {
                if (err == SKBAG_ERR_OP_BOUNDS) {
                    /* Set key to its maximum value and print a
                     * warning if this is the first overflow for this
                     * output file. */
                    skBag_setCounter(bag_io[SRC_PKTS].bag, 
                                     (skBag_key_t*)&rwrec.sIP.ipnum,
                                     (skBag_counter_t*)&max_val);
                    if ( !overflow_warning_printed[SRC_PKTS] ) {
                        overflow_warning_printed[SRC_PKTS] = 1;
                        skAppPrintErr("**WARNING** Source packets overflow");
                    }
                } else {
                    skAppPrintErr("Error %d adding value to source packets",
                                  err);
                    goto END;
                }
            }
        }
        if (bag_io[SRC_BYTES].bag) {
            temp = rwrec.bytes;
            err = skBag_addToCounter(bag_io[SRC_BYTES].bag, 
                                     (skBag_key_t*)&rwrec.sIP.ipnum,
                                     &temp);
            if (err != SKBAG_OK) {
                if (err == SKBAG_ERR_OP_BOUNDS) {
                    /* Set key to its maximum value and print a
                     * warning if this is the first overflow for this
                     * output file. */
                    skBag_setCounter(bag_io[SRC_BYTES].bag, 
                                     (skBag_key_t*)&rwrec.sIP.ipnum,
                                     (skBag_counter_t*)&max_val);
                    if ( !overflow_warning_printed[SRC_BYTES] ) {
                        overflow_warning_printed[SRC_BYTES] = 1;
                        skAppPrintErr("**WARNING** Source bytes overflow");
                    }
                } else {
                    skAppPrintErr("Error %d adding value to source bytes",
                                  err);
                    goto END;
                }
            }
        }

        if (bag_io[DEST_PKTS].bag) {
            temp = rwrec.pkts;
            err = skBag_addToCounter(bag_io[DEST_PKTS].bag, 
                                     (skBag_key_t*)&rwrec.dIP.ipnum,
                                     &temp);
            if (err != SKBAG_OK) {
                if (err == SKBAG_ERR_OP_BOUNDS) {
                    /* Set key to its maximum value and print a
                     * warning if this is the first overflow for this
                     * output file. */
                    skBag_setCounter(bag_io[DEST_PKTS].bag, 
                                     (skBag_key_t*)&rwrec.dIP.ipnum,
                                     (skBag_counter_t*)&max_val);
                    if ( !overflow_warning_printed[DEST_PKTS] ) {
                        overflow_warning_printed[DEST_PKTS] = 1;
                        skAppPrintErr("**WARNING** Dest packets overflow");
                    }
                } else {
                    skAppPrintErr("Error %d adding value to dest packets",
                                  err);
                    goto END;
                }
            }
        }
        if (bag_io[DEST_BYTES].bag) {
            temp = rwrec.bytes;
            err = skBag_addToCounter(bag_io[DEST_BYTES].bag, 
                                     (skBag_key_t*)&rwrec.dIP.ipnum,
                                     &temp);
            if (err != SKBAG_OK) {
                if (err == SKBAG_ERR_OP_BOUNDS) {
                    /* Set key to its maximum value and print a
                     * warning if this is the first overflow for this
                     * output file. */
                    skBag_setCounter(bag_io[DEST_BYTES].bag, 
                                     (skBag_key_t*)&rwrec.dIP.ipnum,
                                     (skBag_counter_t*)&max_val);
                    if ( !overflow_warning_printed[DEST_BYTES] ) {
                        overflow_warning_printed[DEST_BYTES] = 1;
                        skAppPrintErr("**WARNING** Dest bytes overflow");
                    }
                } else {
                    skAppPrintErr("Error %d adding value to dest bytes",
                                  err);
                    goto END;
                }
            }
        }
	if (array_io[SRC_PORT_FLOWS-AB_OFFSET].array) {
	    old_count = array_io[SRC_PORT_FLOWS-AB_OFFSET].array[rwrec.sPort];
	    count = old_count + 1;
	    if ( count < old_count ) {
		if (!array_overflow_warning_printed
		        [SRC_PORT_FLOWS-AB_OFFSET]) {
		    array_overflow_warning_printed
			[SRC_PORT_FLOWS-AB_OFFSET] = 1;
		    skAppPrintErr("**WARNING** Source port flows overflow");
		}
	    } else {
		array_io[SRC_PORT_FLOWS-AB_OFFSET].array[rwrec.sPort] = count;
	    }
	}
	if (array_io[SRC_PORT_PKTS-AB_OFFSET].array) {
	    old_count = array_io[SRC_PORT_PKTS-AB_OFFSET].array[rwrec.sPort];
	    count = old_count + rwrec.pkts;
	    if ( count < old_count ) {
		if (!array_overflow_warning_printed
		        [SRC_PORT_PKTS-AB_OFFSET]) {
		    array_overflow_warning_printed
			[SRC_PORT_PKTS-AB_OFFSET] = 1;
		    skAppPrintErr("**WARNING** Source port packets overflow");
		}
	    } else {
		array_io[SRC_PORT_PKTS-AB_OFFSET].array[rwrec.sPort] = count;
	    }
	}
	if (array_io[SRC_PORT_BYTES-AB_OFFSET].array) {
	    old_count = array_io[SRC_PORT_BYTES-AB_OFFSET].array[rwrec.sPort];
	    count = old_count + rwrec.bytes;
	    if ( count < old_count ) {
		if (!array_overflow_warning_printed
		        [SRC_PORT_BYTES-AB_OFFSET]) {
		    array_overflow_warning_printed
			[SRC_PORT_BYTES-AB_OFFSET] = 1;
		    skAppPrintErr("**WARNING** Source port bytes overflow");
		}
	    } else {
		array_io[SRC_PORT_BYTES-AB_OFFSET].array[rwrec.sPort] = count;
	    }
	}

	if (array_io[DEST_PORT_FLOWS-AB_OFFSET].array) {
	    old_count = array_io[DEST_PORT_FLOWS-AB_OFFSET].array[rwrec.dPort];
	    count = old_count + 1;
	    if ( count < old_count ) {
		if (!array_overflow_warning_printed
		        [DEST_PORT_FLOWS-AB_OFFSET]) {
		    array_overflow_warning_printed
			[DEST_PORT_FLOWS-AB_OFFSET] = 1;
		    skAppPrintErr("**WARNING** Dest port flows overflow");
		}
	    } else {
		array_io[DEST_PORT_FLOWS-AB_OFFSET].array[rwrec.dPort] = count;
	    }
	}
	if (array_io[DEST_PORT_PKTS-AB_OFFSET].array) {
	    old_count = array_io[DEST_PORT_PKTS-AB_OFFSET].array[rwrec.dPort];
	    count = old_count + rwrec.pkts;
	    if ( count < old_count ) {
		if (!array_overflow_warning_printed
		        [DEST_PORT_PKTS-AB_OFFSET]) {
		    array_overflow_warning_printed
			[DEST_PORT_PKTS-AB_OFFSET] = 1;
		    skAppPrintErr("**WARNING** Dest port packets overflow");
		}
	    } else {
		array_io[DEST_PORT_PKTS-AB_OFFSET].array[rwrec.dPort] = count;
	    }
	}
	if (array_io[DEST_PORT_BYTES-AB_OFFSET].array) {
	    old_count = array_io[DEST_PORT_BYTES-AB_OFFSET].array[rwrec.dPort];
	    count = old_count + rwrec.bytes;
	    if ( count < old_count ) {
		if (!array_overflow_warning_printed
		        [DEST_PORT_BYTES-AB_OFFSET]) {
		    array_overflow_warning_printed
			[DEST_PORT_BYTES-AB_OFFSET] = 1;
		    skAppPrintErr("**WARNING** Dest port bytes overflow");
		}
	    } else {
		array_io[DEST_PORT_BYTES-AB_OFFSET].array[rwrec.dPort] = count;
	    }
	}
	
	if (array_io[PROTO_FLOWS-AB_OFFSET].array) {
	    old_count = array_io[PROTO_FLOWS-AB_OFFSET].array[rwrec.proto];
	    count = old_count + 1;
	    if ( count < old_count ) {
		if (!array_overflow_warning_printed
		        [PROTO_FLOWS-AB_OFFSET]) {
		    array_overflow_warning_printed
			[PROTO_FLOWS-AB_OFFSET] = 1;
		    skAppPrintErr("**WARNING** Protocol flows overflow");
		}
	    } else {
		array_io[PROTO_FLOWS-AB_OFFSET].array[rwrec.proto] = count;
	    }
	}
	if (array_io[PROTO_PKTS-AB_OFFSET].array) {
	    old_count = array_io[PROTO_PKTS-AB_OFFSET].array[rwrec.proto];
	    count = old_count + rwrec.pkts;
	    if ( count < old_count ) {
		if (!array_overflow_warning_printed
		        [PROTO_PKTS-AB_OFFSET]) {
		    array_overflow_warning_printed
			[PROTO_PKTS-AB_OFFSET] = 1;
		    skAppPrintErr("**WARNING** Proto packets overflow");
		}
	    } else {
		array_io[PROTO_PKTS-AB_OFFSET].array[rwrec.proto] = count;
	    }
	}
	if (array_io[PROTO_BYTES-AB_OFFSET].array) {
	    old_count = array_io[PROTO_BYTES-AB_OFFSET].array[rwrec.proto];
	    count = old_count + rwrec.bytes;
	    if ( count < old_count ) {
		if (!array_overflow_warning_printed
		        [PROTO_BYTES-AB_OFFSET]) {
		    array_overflow_warning_printed
			[PROTO_BYTES-AB_OFFSET] = 1;
		    skAppPrintErr("**WARNING** Proto bytes overflow");
		}
	    } else {
		array_io[PROTO_BYTES-AB_OFFSET].array[rwrec.proto] = count;
	    }
	}
	
    }

    /* Successful if we make it here */
    rv = 0;

  END:
    rwCloseFile(rwIOS);

    return rv;
}

static int writeArrayBagFile(array_io_t *arraybag)
{
    skBag_key_t i;
    genericHeader file_header;

    PREPHEADER(&file_header);
    file_header.type = FT_MACROBAGTREE;
    file_header.version = 2;
    fwrite((void *) &file_header, sizeof(genericHeader), 1, arraybag->filep);
    for ( i = 0; i < arraybag->size; ++i ) {
        if (arraybag->array[i] != (skBag_counter_t) 0) {
            if (!fwrite((void*) &i, sizeof(skBag_key_t), 1, arraybag->filep)) {
                skAppPrintErr("Error writing %s: %s",
                              arraybag->pathname, strerror(errno));
                return -1;
            }
            if (!fwrite((void*) &arraybag->array[i], sizeof(skBag_counter_t),
                        1, arraybag->filep)) {
                skAppPrintErr("Error writing %s: %s",
                              arraybag->pathname, strerror(errno));
                return -1;
            }
	}
    }
    return 0;
}

int main(int argc, char **argv)
{
    int i;

    appSetup(argc, argv);			/* never returns on error */

    if (first_file == argc) {
        if (0 != processFile("stdin")) {
            return 1;
        }
    } else {
        for (i = first_file; i < argc; ++i) {
            if (0 != processFile(argv[i])) {
                skAppPrintErr("Error processing input file %s",
                              argv[i]);
                return 1;
            }
        }
    }

    /* write out the bags */
    for (i = 0; i < NUM_OUTPUTS; ++i) {
        if (bag_io[i].bag) {
            if (SKBAG_OK != skBag_writeBinary(bag_io[i].bag, bag_io[i].filep)){
                skAppPrintErr("Error writing to %s",
                              bag_io[i].pathname);
            }
        }
    }

    /* write out the portbags */
    for (i = 0; i < NUM_ARRAY_OUTPUTS; ++i) {
	if (array_io[i].array) {
	    if (0 != writeArrayBagFile(&array_io[i])) {
		skAppPrintErr("Error writing to %s", array_io[i].pathname);
	    }
	}
    }

    /* done */
    appTeardown();

    return 0;
}
