/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
** rwbagbuild can read an IP set and generate a bag with a default
** count for each IP address, or it can read a pipe-separated text
** file representing a bag.
*/


#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwset/rwbagbuild.c,v 1.11 2005/10/10 17:29:11 thomasm Exp $");

#include "bagtree.h"
#include "iptree.h"
#include "utils.h"


/* LOCAL DEFINES AND TYPEDEFS */
#define checkMark( f, i ) ( ( f[i/32] & ( 1 << ( i % 32 ) ) ) ? 1 : 0 )

/* where to write --help output */
#define USAGE_FH stdout

/* EXPORTED FUNCTIONS */


/* LOCAL FUNCTIONS */
static void appUsage( void );                /* never returns */
static void appUsageLong( void );                    /* never returns */
static void appOptionsUsage( FILE* );        /* application options usage */
static void appTeardown( void );
static void appSetup( int argc, char **argv ); /* never returns when error */
static int  appOptionsHandler( clientData, int, char * );

static int rwbb_parseBagLine( char *line, skBag_header_t *p_bag );

int createBagFromTextBag( skBag_header_t *p_bag, FILE *fp_in );
int createBagFromSet( skBag_header_t *p_bag, FILE *fp_in );

/* LOCAL VARIABLES */
static const skBag_level_t levels = (skBag_level_t) 4;
static skBag_levelsize_t rg_level_size[4] = { 9, 9, 9, 5 };

static FILE *fp_out = NULL;
static char *filename_out = NULL;
static int f_out_is_pipe;

static FILE *fp_in = NULL;
static char *filename_in = NULL;
static int f_in_is_pipe;
/*
** 0 if input is an IP set.  1 if input is a text bag.
*/

static int f_use_default_count = 0;
static skBag_counter_t default_count = (skBag_counter_t) 1;


/* OPTIONS SETUP */

typedef enum _appOptionsEnum {
    RWBB_NO_INPUT,
    RWBB_SET_INPUT,
    RWBB_BAG_INPUT,
    RWBB_DEFAULT_COUNT,
    RWBB_OUTPUT
} appOptionsEnum;

static appOptionsEnum input_type = RWBB_NO_INPUT;

static struct option appOptions[] = {
    { "set-input",     REQUIRED_ARG, 0, RWBB_SET_INPUT },
    { "bag-input",     REQUIRED_ARG, 0, RWBB_BAG_INPUT },
    { "default-count", REQUIRED_ARG, 0, RWBB_DEFAULT_COUNT },
    { "output",        REQUIRED_ARG, 0, RWBB_OUTPUT },
    {0,0,0,0} /* sentinel entry */
};


static const char *appHelp[] = {
    "Create a bag from an IP set.",
    "Create a bag from a pipe separated text bag file.",
    "Default counters for each key in the new bag.\n  Default is 1 for sets.",
    "Destination for the new bag (stdout|pipe|filename).\n  Default stdout.",
    (char *)NULL
};


/*
 * appUsage:
 *      print short usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
void appUsage( void )
{
    fprintf( stderr, "Use `%s --help' for usage\n", skAppName() );
    exit( EXIT_FAILURE );
}


/*
 * appUsageLong:
 *      print complete usage information to USAGE_FH and exit with
 *      code.  passed to optionsSetup() to print usage when --help
 *      option given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong( void )
{
    fprintf(USAGE_FH,
            ("Usage: %s {--set-input=FILE | --bag-input=FILE} [options]....\n"
             "\tCreate a binary Bag file.\n"),
            skAppName());
    appOptionsUsage( USAGE_FH );
    fprintf( USAGE_FH, "\nEither --set-input or --bag-input is required.\n" );
    exit( EXIT_SUCCESS );
}


/*
 * appOptionsUsage:
 *      print options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 */
static void appOptionsUsage( FILE *fh )
{
    int i;
    fprintf( fh, "\noptions:\n" );
    for ( i = 0; appOptions[i].name; i++ ) {
        fprintf( fh, "--%s %s. %s\n", appOptions[i].name,
                 SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i] );
    }
    return;
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown( void )
{
    static int teardownFlag = 0;
    if ( teardownFlag ) {
        return;
    }
    teardownFlag = 1;
    optionsTeardown();
    /* local teardown segment */

    /*
    ** close pipes/files
    */
    if ( fp_out != NULL ) {
        if ( f_out_is_pipe ) {
            pclose( fp_out );
        } else {
            if ( fclose( fp_out ) == EOF ) {
                skAppPrintErr( "Error closing output file '%s': %s",
                               filename_out, strerror( errno ) );
            }
        }
    }

    if ( fp_in != NULL ) {
        if ( f_in_is_pipe ) {
            pclose( fp_in );
        } else {
            if ( fclose( fp_in ) == EOF ) {
                skAppPrintErr( "Error closing input file '%s': %s",
                               filename_in, strerror( errno ) );
            }
        }
    }

    return;
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int nextOptIndex;
    skAppRegister( argv[0] );

    /* first do some sanity checks */

    /* check that we have the same number of options entries and help*/
    assert( ( sizeof( appHelp ) / sizeof( char * ) ) ==
            ( sizeof( appOptions ) / sizeof( struct option ) ) );

    if ( optionsSetup( appUsageLong ) ) {
        skAppPrintErr( "unable to setup options module" );
        exit( EXIT_FAILURE ); /* never returns */
    }

    /* register the options handler */
    if ( optionsRegister( appOptions, (optHandler) appOptionsHandler, NULL ) )
    {
        skAppPrintErr( "unable to register application options" );
        exit( EXIT_FAILURE ); /* never returns */
    }

    fp_out = stdout;

    /* parse options */
    if ( ( nextOptIndex = optionsParse( argc, argv ) ) < 0 ) {
        /* options parsing should print error */
        appUsage(); /* never returns */
    }

    /*
    ** check final setup state
    */
    if ( input_type == RWBB_NO_INPUT ) {
        skAppPrintErr( "requires either set-input or bag-input.\n" );
        appUsage(); /* never returns */
    }

    if ( atexit( appTeardown ) < 0 ) {
        skAppPrintErr( "unable to register appTeardown() with atexit()" );
        appTeardown();
        exit( EXIT_FAILURE );
    }

    if ( fp_out == stdout && FILEIsATty( stdout ) ) {
        skAppPrintErr( "binary output not allowed to TTY stdout" );
        appTeardown();
        exit( EXIT_FAILURE );
    }

    return; /* OK */
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    switch ( (appOptionsEnum)opt_index ) {
      case RWBB_SET_INPUT:
        if ( opt_arg == NULL ) {
            return 1;
        }
        if ( strcmp( "stdin", opt_arg ) == 0 ) {
            fp_in = stdin;
        } else {
            filename_in = strdup( opt_arg );
            if ( filename_in == NULL ) {
                skAppPrintErr( "strdup failed!" );
                return 1;
            }
            if ( openFile(filename_in, 0 /* read */, &fp_in, &f_in_is_pipe) ) {
                return 1;
            }
        }
        input_type = RWBB_SET_INPUT;
        break;

      case RWBB_BAG_INPUT:
        if ( opt_arg == NULL ) {
            return 1;
        }
        if ( strcmp( "stdin", opt_arg ) == 0 ) {
            fp_in = stdin;
        } else {
            filename_in = strdup( opt_arg );
            if ( filename_in == NULL ) {
                skAppPrintErr( "strdup failed!" );
                return 1;
            }
            if (openFile(filename_in, 0 /* read */, &fp_in, &f_in_is_pipe)) {
                return 1;
            }
        }
        input_type = RWBB_BAG_INPUT;
        break;

      case RWBB_DEFAULT_COUNT:
        if ( opt_arg == NULL ) {
            return 1;
        }
        if (skStringParseUint32((uint32_t*)&default_count, opt_arg, 0, 0)) {
            skAppPrintErr("default count '%s' is not parsable", opt_arg);
            return 1;
        }
        f_use_default_count = 1;
        break;

      case RWBB_OUTPUT:
        if ( opt_arg == NULL ) {
            return 1;
        }
        if ( strcmp( "stdout", opt_arg ) == 0 ) {
            fp_out = stdout;
        } else {
            filename_out = strdup( opt_arg );
            if ( filename_out == NULL ) {
                skAppPrintErr( "strdup failed!" );
                return 1;
            }
            if ( openFile( filename_out,
                          1 /* write */,
                          &fp_out,
                          &f_out_is_pipe ) ) {
                return 1;
            }
        }
        break;

      case RWBB_NO_INPUT:
        /* impossible! */
        abort();
    }

    return 0; /* OK */
}


static int rwbb_parseBagLine(
    char           *line,
    skBag_header_t *p_bag)
{
    skBag_key_t key;
    skBag_counter_t counter;
    int f_is_integer;
    char *sz_key;
    char *sz_counter;
    char *temp;
    uint32_t octets[4][8];
    int rv;

    /*
    ** search for pipe, and break line into key and counter
    */
    sz_key = line;
    sz_counter = strchr( line, '|' );
    if ( sz_counter == NULL ) {
        /* not a pipe delimited line, assume that count is 1 */
        counter = (skBag_counter_t) 1;
    } else {
        /* change pipe into NULL char, to terminate first string */
        *sz_counter = '\0';
        /* move counter to point to next char (start of counter) */
        sz_counter++;

        /* parse second part of line to get count (if necessary) */
        if ( f_use_default_count == 0 ) {
            errno = 0;
            counter = strtoul( sz_counter, &temp, 10 );
            if ( (temp == sz_counter) || (errno != 0)
                || (*temp != '\0' && !isspace((int)*temp)) )
            {
                return 1;
            }
        } else {
            counter = default_count;
        }
    }

    /*
    ** parse key section of bag line
    */
    f_is_integer = 1;
    temp = sz_key;
    while ( *temp != '\0' ) {
        if ( !isdigit( (int) *temp ) ) {
            /* string contains a non-digit character, send to secondary
               parsing routine. */
            f_is_integer = 0;
        }
        ++temp;
    }

    if ( f_is_integer ) {
        /* integer ip */
        errno = 0;
        key = strtoul( sz_key, &temp, 10 );
        if ( errno != 0 || temp == sz_key || temp != '\0' ) {
            return 1;
        }
        if ( skBag_addToCounter( p_bag, &key, &counter ) != SKBAG_OK ) {
            return 1;
        }
    } else {
        int bit_A, bit_B, bit_C, bit_D;

        /* parse first part of line to get octet bitmaps */
        rv = skIPStringToBitmap( octets, sz_key );
        if ( rv != 0 ) {
            /* not parsable */
            return 1;
        }

        /*
        ** If the line was parsed as a dotted quad, CIDR block, etc.,
        ** then unroll the bitmaps into individual IPs and add them.
        */
        for ( bit_A = 0; bit_A < 256 ; bit_A++ ) {
            if ( !checkMark( octets[0], bit_A ) ) {
                continue;
            }
            for ( bit_B = 0 ; bit_B < 256; bit_B++ ) {
                if ( !checkMark( octets[1], bit_B ) ) {
                    continue;
                }
                for ( bit_C = 0; bit_C < 256; bit_C++ ) {
                    if ( !checkMark( octets[2], bit_C ) ) {
                        continue;
                    }
                    for ( bit_D = 0; bit_D < 256; bit_D++ ) {
                        if ( !checkMark( octets[3], bit_D ) ) {
                            continue;
                        }

                        key = ( ( bit_A << 24 ) +
                                ( bit_B << 16 ) +
                                ( bit_C <<  8 ) +
                                ( bit_D ) );
                        if ( skBag_addToCounter( p_bag, &key, &counter )
                            != SKBAG_OK )
                        {
                            return 1;
                        }
                    }
                }
            }
        }
    }

    return 0;
}



int createBagFromTextBag(
    skBag_header_t *p_bag,
    FILE           *fp_in)
{
    char buffer[128];
    char *p_line_end;
    int line_cur;

    /* read until end of file */
    line_cur = 0;
    while ( !feof( fp_in ) ) {
        memset( buffer, '\0', sizeof( buffer ) );
        if ( fgets( buffer, sizeof( buffer ), fp_in ) == NULL ) {
            continue;
        }
        ++line_cur;

        /* find end of line */
        p_line_end = strchr( buffer, '\n' );
        if ( buffer == p_line_end ) {
            /* ignore blank line */
            continue;
        }

        if ( p_line_end != NULL ) {
            /* okay, read whole line */
            *p_line_end = '\0';
        } else if ( feof( fp_in ) ) {
            /* okay, last line did not end in newline */
        } else {
            /*
            ** error, line was longer than buffer.  read until newline
            ** or eof, then throw away the line
            */
            skAppPrintErr( "ignoring input line %d, too long.",
                           line_cur );
            while ( fgets( buffer, sizeof( buffer ), fp_in )
                   && !strchr( buffer, '\n' ) )
            {
                ; /* empty */
            }
            continue;
        }

        /* parse the line */
        if ( rwbb_parseBagLine( buffer, p_bag ) ) {
            skAppPrintErr( "error parsing line %d", line_cur );
            return 1;
        }
    }

    return 0;
}


int createBagFromSet(
    skBag_header_t *p_bag,
    FILE           *fp_in )
{
    skIPTree_t *set = NULL;
    skBag_key_t ip;
    skIPTreeIterator_t *iter;

    /* Read macroTree from file */
    if ( skIPTreeRead( fp_in, &set ) != 0 ) {
        return 1;
    }

    if (skIPTreeIteratorCreate(&iter, set) != 0) {
        return 1;
    }

    while (skIPTreeIteratorNext((uint32_t*)&ip, iter) == SKIP_ITERATOR_OK) {
        if ( f_use_default_count == 0 ) {
            if ( skBag_incrCounter( p_bag, &ip )
                 != SKBAG_OK )
            {
                return 1;
            }
        } else {
            if ( skBag_setCounter( p_bag, &ip, &default_count ) )
            {
                return 1;
            }
        }
    }

    skIPTreeDelete( &set );
    return 0;
}


int main( int argc, char **argv )
{
    skBag_header_t *p_bag;

    appSetup( argc, argv ); /* never returns on error */

    if ( skBag_alloc( &p_bag, levels, rg_level_size ) != SKBAG_OK ) {
        skAppPrintErr( "error allocating tree" );
        appTeardown();
        exit( EXIT_FAILURE );
    }

    /*
    ** Handle set-file input
    */
    if ( input_type == RWBB_SET_INPUT ) {
        if ( createBagFromSet( p_bag, fp_in ) ) {
            skAppPrintErr( "error creating bag from set" );
            skBag_free( p_bag );
            appTeardown();
            exit( EXIT_FAILURE );
        }
    }

    /*
    ** Handle bag-file input
    */
    if ( input_type == RWBB_BAG_INPUT ) {
        if ( createBagFromTextBag( p_bag, fp_in ) ) {
            skAppPrintErr( "error creating bag from text bag" );
            skBag_free( p_bag );
            appTeardown();
            exit( EXIT_FAILURE );
        }
    }

    /*
    ** write output
    */
    if ( skBag_writeBinary( p_bag, fp_out ) != SKBAG_OK ) {
        skAppPrintErr( "error writing bag" );
        skBag_free( p_bag );
        appTeardown();
        exit( EXIT_FAILURE );
    }

    /* done */
    appTeardown();
    return EXIT_SUCCESS;
}
