/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**    rwbagcat reads a binary bag, converts it to text, and outputs it
**    to stdout.  It can also print various statistics and summary
**    information about the bag.  It attempts to read the bag(s) from
**    stdin or from any arguments.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwset/rwbagcat.c,v 1.32 2005/10/19 19:33:33 leechr Exp $");

#include "bagtree.h"
#include "iptree.h"
#include "utils.h"
#include "print_network_structure.h"

/* LOCAL DEFINES AND TYPEDEFS */

/* where to write --help output */
#define USAGE_FH stdout

/* width of count fields in columnar output */
#define COUNT_WIDTH 15

typedef struct _io_t {
    char *filename;
    FILE *stream;
    int (*closefn)(FILE *);
    struct _io_t *next;
} io_t;

typedef enum _binScheme {
    BINSCHEME_NONE=0,
    BINSCHEME_LINEAR=1,
    BINSCHEME_BINARY=2,
    BINSCHEME_DECIMAL=3
} binScheme_en;


/* EXPORTED FUNCTIONS */


/* LOCAL FUNCTIONS */

static void addInput( char* filename );
static void removeInputs( void );
static int  setOutput( const char* filename, io_t *p_out );

static int printStats(
    skBag_header_t *p_bag, 
    FILE           *fp_out );

static int getIPArgument( 
    const char *arg, 
    uint32_t   *ip );

static void appUsage(void);                  /* never returns */
static void appUsageLong(void);              /* never returns */
static void appOptionsUsage(FILE*);          /* application options usage */
static void appTeardown(void);
static void appSetup(int argc, char **argv); /* never returns when error */
static int  appOptionsHandler(clientData, int, char *);

static skBag_key_t calculateBinIndex( 
    skBag_counter_t    *p_counter, 
    binScheme_en        scheme );

static int calculateBinLabel(
    skBag_key_t     key,
    binScheme_en    scheme,
    char           *label,
    int             label_size );


/* LOCAL VARIABLES */

/* global I/O state */
static int file_index = 0; /* Index of current file argument in argv */
static io_t *inputs; /* linked list of all input files/streams/pipes */
static int f_stdin_active;
static io_t stats;
static io_t treestats;
static io_t output;

static struct {
    int f_print_stats;
    int f_print_treestats;
    int f_print_network;
    int f_print_T; /* print total */
    int f_print_A; /* print class A breakdown and subtotal */ 
    int f_print_B; /* print class B breakdown and subtotal */ 
    int f_print_C; /* print class C breakdown and subtotal */ 
    int f_print_X; /* print /27 breakdown and subtotal */ 
    int f_print_H; /* print individual hosts and counts */
    int f_print_S; /* print summary information on each breakdown */
} print_options = { 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 };

static binScheme_en bin_scheme = BINSCHEME_NONE;

/* delimiter between output columns for hosts/counts */
static char output_delimiter = '|';

/* whether host/count output is in columns or scrunched together */
static int output_is_columnar = 1;

/* whether to print keys as integers or IP addresses */
static int f_print_integer_keys = 0;

/* zero pad ips */
static int f_zero_pad_ips = 0;

/* print out keys whose count is zero (provided that minkey and maxkey are
   both set */
static int f_print_zero_counts = 0;

static skBag_counter_t mincount = 0;
static skBag_counter_t maxcount = ~( (skBag_counter_t) 0 );
static skBag_key_t minkey = 0;
static skBag_key_t maxkey = ~( (skBag_key_t) 0 );

/* name of program to run to page output */
static char *pager = NULL;


/* OPTIONS SETUP */

typedef enum _appOptionsEnum {
    RWBAGCAT_NETWORK_STRUCTURE,
    RWBAGCAT_BINIPS,
    RWBAGCAT_STATS,
    RWBAGCAT_TREESTATS,
    RWBAGCAT_MINCOUNT,
    RWBAGCAT_MAXCOUNT,
    RWBAGCAT_MINKEY,
    RWBAGCAT_MAXKEY,
    RWBAGCAT_OUTPUT,
    RWBAGCAT_DELIMITED,
    RWBAGCAT_INTEGER_KEYS,
    RWBAGCAT_ZERO_PAD_IPS,
    RWBAGCAT_PRINT_ZERO_COUNTS,
    RWBAGCAT_PAGER
} appOptionsEnum;


static struct option appOptions[] = {
    { "network-structure", OPTIONAL_ARG, 0, RWBAGCAT_NETWORK_STRUCTURE },
    { "bin-ips",           OPTIONAL_ARG, 0, RWBAGCAT_BINIPS },
    { "stats",             OPTIONAL_ARG, 0, RWBAGCAT_STATS },
    { "tree-stats",        OPTIONAL_ARG, 0, RWBAGCAT_TREESTATS },
    { "mincount",          REQUIRED_ARG, 0, RWBAGCAT_MINCOUNT },
    { "maxcount",          REQUIRED_ARG, 0, RWBAGCAT_MAXCOUNT },
    { "minkey",            REQUIRED_ARG, 0, RWBAGCAT_MINKEY },
    { "maxkey",            REQUIRED_ARG, 0, RWBAGCAT_MAXKEY },
    { "output",            REQUIRED_ARG, 0, RWBAGCAT_OUTPUT },
    { "delimited",         OPTIONAL_ARG, 0, RWBAGCAT_DELIMITED },
    { "integer-keys",      NO_ARG,       0, RWBAGCAT_INTEGER_KEYS },
    { "zero-pad-ips",      NO_ARG,       0, RWBAGCAT_ZERO_PAD_IPS },
    { "zero-counts",       NO_ARG,       0, RWBAGCAT_PRINT_ZERO_COUNTS },
    { "pager",             REQUIRED_ARG, 0, RWBAGCAT_PAGER },
    { 0, 0, 0, 0 } /* sentinel entry */
};


static const char *appHelp[] = {
    ("Print the network structure around the keys.\n"
     "\t Choose from TABCXHS for grand total, /8, /16, /24, /27, \n"
     "\t /32 headers and summary information"),
    ("Invert the bag and count by distinct volume values.\n"
     "\t linear:    volume => count(IP)\n"
     "\t binary:    log2(volume) => count(IP)\n"
     "\t decimal:   variation on log10(volume) => count(IP)"),
    ("Print general statistics about the bag to named stream.\n"
     "\t Send output to optional stream, or to stderr if none given"),
    ("Print bag performance data to named stream.\n"
     "\t Send output to optional stream, or to stdout if none given"),
    "Output records with count greater than this value.",
    "Output records with count less than this value.",
    "Output records with a key value greater than this value.",
    "Output records with a key value less than this value.",
    "Write output to named stream",
    "Do not print in columns, use delimiter between values",
    "Keys shouuld be printed as integers",
    "print IPs as dotted quad with 3 digits per octet",
    ("Print keys with a count of zero\n"
     "\t (requires --minkey and --maxkey)"),
    "Program to invoke to page output",
    (char *) NULL
};


/*
 * appUsage:
 *      print short usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
void appUsage( void )
{
    fprintf( stderr, "Use `%s --help' for usage\n", skAppName() );
    exit( EXIT_FAILURE );
}


/*
 * appUsageLong:
 *      print complete usage information to USAGE_FH and exit with
 *      code.  passed to optionsSetup() to print usage when --help
 *      option given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong( void )
{
    fprintf( USAGE_FH,
             ("Usage: %s [options] [files]\n"
              "\tPrint binary Bag files as text.\n"),
             skAppName() );
    appOptionsUsage( USAGE_FH );
    exit( EXIT_FAILURE );
}


/*
 * appOptionsUsage:
 *      print options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\napp options\n");
    for (i = 0; appOptions[i].name; i++) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }

    return;
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown( void )
{
    static int teardownFlag = 0;

    if ( teardownFlag ) {
        return;
    }
    teardownFlag = 1;
    optionsTeardown();

    /* local teardown segment */

    /* Close open files */
    removeInputs();
    
    if ( output.stream != NULL && output.closefn != NULL ) {
        if ( output.closefn( output.stream ) != 0 ) {
            skAppPrintErr( "Error closing output file '%s': %s",
                           output.filename, strerror( errno ) );
        }
    }

    if ( stats.stream != NULL && stats.closefn != NULL ) {
        if ( stats.closefn( stats.stream ) != 0 ) {
            skAppPrintErr( "Error closing output file '%s': %s",
                           stats.filename, strerror( errno ) );
        }
    }

    if ( treestats.stream != NULL && treestats.closefn != NULL ) {
        if ( treestats.closefn( treestats.stream ) != 0 ) {
            skAppPrintErr( "Error closing output file '%s': %s",
                           treestats.filename, strerror( errno ) );
        }
    }

    return;
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup( int argc, char **argv )
{
    int nextOptIndex;
    int using_pager = 0;

    skAppRegister( argv[0] );

    /* first do some sanity checks */

    /* check that we have the same number of options entries and help*/
    assert( ( sizeof( appHelp ) / sizeof( char * ) ) ==
            ( sizeof( appOptions ) / sizeof( struct option ) ) );

    if ( optionsSetup( appUsageLong ) ) {
        skAppPrintErr( "unable to setup options module" );
        exit( EXIT_FAILURE ); /* never returns */
    }

    /* register the options handler */
    if ( optionsRegister( appOptions, (optHandler) appOptionsHandler, NULL ) )
    {
        skAppPrintErr( "unable to register application options" );
        exit( EXIT_FAILURE ); /* never returns */
    }

    /*
     * set default output
     */
    output.stream = stdout;
    output.closefn = NULL;
    stats.stream = stderr;
    stats.closefn = NULL;
    treestats.stream = stdout;
    treestats.closefn = NULL;

    /* parse options */
    if ( ( nextOptIndex = optionsParse( argc, argv ) ) < 0 ) {
        /* options parsing should print error */
        appTeardown();
        appUsage(); /* never returns */
    }

    if ( print_options.f_print_network == 1 && bin_scheme != BINSCHEME_NONE ) {
        skAppPrintErr( "cannot have both --network-structure and --bin-ips" );
        appTeardown();
        appUsage(); /* never returns */
    }

    /* default to printing network hosts */
    if ( !print_options.f_print_stats && !print_options.f_print_treestats && 
         bin_scheme == BINSCHEME_NONE ) 
    {
        print_options.f_print_network = 1;
    }

    /* if minkey and maxkey aren't set, disable zero-count printing */
    if ( f_print_zero_counts && ( minkey == 0 || maxkey == UINT_MAX ) ) {
        skAppPrintErr( "to use --zero-counts, --minkey and --maxkey must "
                       "both be used" );
        f_print_zero_counts = 0;
    }

    /* don't try to print columnar output with --network-structure */
    if ( print_options.f_print_T || print_options.f_print_A || 
         print_options.f_print_B || print_options.f_print_C || 
         print_options.f_print_X ) 
    {
        output_is_columnar = 0;
    }

    file_index = nextOptIndex;

    inputs = NULL;
    f_stdin_active = 0;
    /*
     * add stdin as an input, if applicable
     */
    if ( !FILEIsATty( stdin ) ) {
        addInput( "stdin" );
    } else if (file_index == argc) {
        skAppPrintErr("no files given on command line and"
                      " stdin is connected to a terminal");
        appUsage();
    }

    /*
     * get filenames from command line, if present
     */
    while ( file_index < argc ) {
        addInput( argv[file_index] );
        ++file_index;
    }

    if ( inputs == NULL ) {
        skAppPrintErr( "no inputs to process" );
        appTeardown();
        exit( EXIT_FAILURE );
    }

    /* If the output.stream is to stdout and that is a terminal,
     * invoke the pager to read it.  If the pager is set, then have
     * the stats and treestats streams use it as well if they are set
     * to their default values, but do not set their closefn pointers
     * because we don't want to close the stream multiple times. */
    using_pager = skOpenPagerWhenStdoutTty( &output.stream, &pager );
    if ( using_pager < 0 ) {
        exit(EXIT_FAILURE);
    }
    if ( using_pager > 0 ) {
        output.filename = pager;
        output.closefn = pclose;
        if ( ( stats.stream == stderr ) && FILEIsATty( stderr ) ) {
            stats.stream = output.stream;
        }
        if ( treestats.stream == stdout ) {
            treestats.stream = output.stream;
        }
    }

    if ( atexit( appTeardown ) < 0 ) {
        skAppPrintErr( "unable to register appTeardown() with atexit()" );
        appTeardown();
        exit( EXIT_FAILURE );
    }

    return; /* OK */
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED( cData ),
    int         opt_index,
    char       *opt_arg )
{
    unsigned int i;
    uint32_t val;
    
    switch ( (appOptionsEnum)opt_index ) {
        
      case RWBAGCAT_NETWORK_STRUCTURE:
        /* Validate that this is a legal string */
        if ( opt_arg != NULL ) {
            for ( i = 0; i < strlen( opt_arg ); i++ ) {
                if ( strchr( "TABCXHS", opt_arg[i] ) == NULL ) {
                    skAppPrintErr( "Illegal network structure output option. "
                                   "Should be some combination of TABCXHS." );
                    return 1;
                }
            }
            print_options.f_print_T = ( strchr( opt_arg, 'T' ) != NULL );
            print_options.f_print_A = ( strchr( opt_arg, 'A' ) != NULL );
            print_options.f_print_B = ( strchr( opt_arg, 'B' ) != NULL );
            print_options.f_print_C = ( strchr( opt_arg, 'C' ) != NULL );
            print_options.f_print_X = ( strchr( opt_arg, 'X' ) != NULL );
            print_options.f_print_H = ( strchr( opt_arg, 'H' ) != NULL );
            print_options.f_print_S = ( strchr( opt_arg, 'S' ) != NULL );
        } else {
            /*
             * default if --network-structure is used, but not
             * qualified with an option string.
             */
            print_options.f_print_T = 1;
            print_options.f_print_S = 1;
        }
        print_options.f_print_network = 1;
        break;

      case RWBAGCAT_BINIPS:
        if ( opt_arg == NULL ) {
            bin_scheme = BINSCHEME_LINEAR;
        } else {
            size_t len = strlen(opt_arg);
            if ( len == 0 ) {
                skAppPrintErr( "--bin-ips requires an argument" );
                return 1;
            }
            if ( strncmp( opt_arg, "linear", len ) == 0 ) {
                bin_scheme = BINSCHEME_LINEAR;
            } else if ( strncmp( opt_arg, "binary", len ) == 0 ) {
                bin_scheme = BINSCHEME_BINARY;
            } else if ( strncmp( opt_arg, "decimal", len ) == 0 ) {
                bin_scheme = BINSCHEME_DECIMAL;
            } else {
                skAppPrintErr( "Illegal bin scheme. "
                               "Should be one of: linear, binary, decimal." );
                return 1;
            }
        }
        break;
        
      case RWBAGCAT_STATS:
        if ( opt_arg != NULL ) {
            if ( setOutput( opt_arg, &stats ) ) {
                skAppPrintErr("Error with %s value '%s'",
                              appOptions[opt_index].name, opt_arg);
                return 1;
            }
        }
        print_options.f_print_stats = 1;
        break;

      case RWBAGCAT_TREESTATS:
        if ( opt_arg != NULL ) {
            if ( setOutput( opt_arg, &treestats ) ) {
                skAppPrintErr("Error with %s value '%s'",
                              appOptions[opt_index].name, opt_arg);
                return 1;
            }
        }
        print_options.f_print_treestats = 1;
        break;          

      case RWBAGCAT_MINCOUNT:
        if ( skStringParseUint32( &val, opt_arg, 0, 0 ) ) {
            skAppPrintErr("Unable to parse %s value '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        mincount = (skBag_counter_t) val;
        break;

      case RWBAGCAT_MAXCOUNT:
        if ( skStringParseUint32( &val, opt_arg, 0, 0 ) ) {
            skAppPrintErr("Unable to parse %s value '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        maxcount = (skBag_counter_t) val;
        break;

      case RWBAGCAT_MINKEY:
        if ( getIPArgument( opt_arg, &val ) ) {
            skAppPrintErr("Unable to parse %s value '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        minkey = (skBag_key_t) val;
        break;

      case RWBAGCAT_MAXKEY:
        if ( getIPArgument( opt_arg, &val ) ) {
            skAppPrintErr("Unable to parse %s value '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        maxkey = (skBag_key_t) val;
        break;

      case RWBAGCAT_OUTPUT:
        if ( setOutput( opt_arg, &output ) ) {
            skAppPrintErr("Error with %s value '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

      case RWBAGCAT_DELIMITED:
        if (optarg) {
            output_delimiter = optarg[0];
        }
        output_is_columnar = 0;
        break;

      case RWBAGCAT_INTEGER_KEYS:
        f_print_integer_keys = 1;
        break;          

      case RWBAGCAT_ZERO_PAD_IPS:
        f_zero_pad_ips = 1;
        break;          

      case RWBAGCAT_PRINT_ZERO_COUNTS:
        f_print_zero_counts = 1;
        break;          

      case RWBAGCAT_PAGER:
        pager = opt_arg;
        break;
    }

    return 0; /* OK */
}




/*
 * parse a string argument, which may only be a string 32-bit unsigned
 * integer, or a string representation of an IP address (dotted quad
 * only, no CIDR, wildcards, ranges, etc. allowed.) into a unsigned
 * 32-bit integer
 *
 * returns 0 on a successful conversion, returns 1 otherwise.
 *
 * on a successful run, *ip is filled with the ip address parsed.  on
 * an unsuccessful run, *ip is unchanged.
 */
static int getIPArgument( 
    const char *arg, 
    uint32_t   *ip ) 
{
    char *temp;
    int f_is_integer; /* is argument an integer or not? */
    uint32_t parsed_ip;
    /*
     * if we find any non-digit characters, then the argument is not a
     * plain integer
     */
    f_is_integer = 1;
    temp = (char *) arg;            
    while ( *temp != '\0' ) {
        if ( !isdigit( (int) *temp ) ) {
            f_is_integer = 0;
        }
        ++temp;
    }

    if ( f_is_integer ) {
        /* integer ip */
        if ( skStringParseUint32( &parsed_ip, arg, 0, 0 ) ) {
            /*
             * integer was either too large or just an invalid value.
             * since we checked for non-digit characters, it was
             * probably just too large.
             */
            return 1;
        }
    } else {
        /* dotted ip */
        parsed_ip = dot2num( arg );
        if ( parsed_ip == ~( (uint32_t) 0 ) ) {
            /*
             * dot2num() couldn't parse the IP address.  since we
             * already checked for plain integers, we know the value
             * is unparsable.
             */
            return 1;
        }
    }
    *ip = parsed_ip;
    return 0;
}


/*
 * add entry to linked list of inputs
 */
static void addInput( char* filename ) 
{
    io_t **p_p_input_cur;
    int ispipe;

    p_p_input_cur = &inputs;

    /* go to last entry in linked list */
    while ( *p_p_input_cur != NULL ) {
        p_p_input_cur = &(*p_p_input_cur)->next;
    }

    *p_p_input_cur = (io_t *) malloc( sizeof( io_t ) );
    if ( *p_p_input_cur == NULL ) {
        skAppPrintErr( "Error allocating memory in addInput()" );
        appTeardown();
        exit( EXIT_FAILURE );
    }
    
    (*p_p_input_cur)->filename = strdup( filename );
    if ( (*p_p_input_cur)->filename == NULL ) {
        skAppPrintErr( "strdup failed!" );
        appTeardown();
        exit( EXIT_FAILURE );
    }

    if ( strcmp( filename, "stdin" ) == 0 ) {
        
        if ( f_stdin_active == 1 ) {
            assert( *p_p_input_cur != NULL );
            free( *p_p_input_cur );
            return;
        }

        f_stdin_active = 1;
        (*p_p_input_cur)->stream = stdin;
        (*p_p_input_cur)->closefn = NULL;
    } else {
        if ( openFile( filename, 
                       0 /* read */, 
                       &(*p_p_input_cur)->stream, 
                       &ispipe ) )
        {
            skAppPrintErr( "input file open failed" );
            appTeardown();
            exit( EXIT_FAILURE );
        }
        if ( ispipe ) {
            (*p_p_input_cur)->closefn = &pclose;
        } else {
            (*p_p_input_cur)->closefn = &fclose;
        }
    }

    (*p_p_input_cur)->next = NULL;
}


static void removeInputs(void) 
{
    io_t *p_input_cur;
    io_t *p_input_next;

    if ( inputs == NULL ) {
        return;
    }

    p_input_cur = inputs;
    while ( p_input_cur != NULL ) {
        p_input_next = p_input_cur->next;

        if ( p_input_cur->filename != NULL ) {
            if ( p_input_cur->closefn != NULL ) {
                p_input_cur->closefn( p_input_cur->stream );
                p_input_cur->closefn = NULL;
            }

            free( p_input_cur->filename );
            p_input_cur->filename = NULL;
        }

        free( p_input_cur );
        p_input_cur = p_input_next;
    }

    inputs = NULL;
}


/*
 *  status = setOutput(name, io_stream);
 *
 *    Set io_stream's output to 'name'.  Return 0 on success, 1
 *    otherwise.
 */
static int setOutput( const char* filename, io_t *p_out ) 
{
    int ispipe;

    if ( p_out == NULL ) {
        skAppPrintErr( "not a valid io_t structure" );
        return 1;
    }

    if ( filename == NULL ) {
        skAppPrintErr( "filename null" );
        return 1;
    }

    p_out->filename = (char*)filename;

    if ( strcmp( filename, "stdout" ) == 0 ) {
        p_out->stream = stdout;
        p_out->closefn = NULL;
    } else if ( strcmp( filename, "stderr" ) == 0 ) {
        p_out->stream = stderr;
        p_out->closefn = NULL;
    } else {
        if ( openFile( filename, 
                       0 /* read */, 
                       &p_out->stream, 
                       &ispipe ) )
        {
            skAppPrintErr( "output file open failed" );
            return 1;
        }
        if ( ispipe ) {
            p_out->closefn = &pclose;
        } else {
            p_out->closefn = &fclose;
        }
    }

    p_out->next = NULL;
    return 0;
}


/*
 * for a given bin scheme, calculate the bin index for a particular
 * counter value
 */
static skBag_key_t calculateBinIndex( 
    skBag_counter_t    *p_counter, 
    binScheme_en        scheme ) 
{
    switch ( scheme ) {
      case BINSCHEME_LINEAR:
        /* need to change if keys/counters are not equivalent */
        return (skBag_key_t) *p_counter;
        
      case BINSCHEME_BINARY:
        return (skBag_counter_t) floor( log10( (double) *p_counter ) / 
                                        log10( (double) 2 ) );
        
      case BINSCHEME_DECIMAL:
        if ( *p_counter < (skBag_counter_t) 100 ) {
            return (skBag_key_t) *p_counter;
        } else {
            return (skBag_counter_t)floor( ( log10( (double) *p_counter ) - 1 )
                                           * 100 );
        }

      case BINSCHEME_NONE:
        skAppPrintErr( "bad bin scheme" );
        assert(0);
        abort();
    }

    assert(0);
    abort();
    return 1;
}


/*
 * for a given bin index, write the label used to display that bin.
 * do not write more than label_size characters (including trailing
 * null)
 *
 * returns 1 if the string had to be truncated to fit in label_size.
 * returns 0 if the string fit.
 */
static int calculateBinLabel(
    skBag_key_t     key,
    binScheme_en    scheme,
    char           *label,
    int             label_size )
{
    double min;
    double max;
    skBag_key_t mid;

    switch ( scheme ) {
      case BINSCHEME_LINEAR:
        /* label is just bin number */
        if ( snprintf( label, (size_t) label_size, "%u", key ) > 
             label_size ) 
        {
            return 1;
        }
        return 0;

      case BINSCHEME_BINARY:
        /* label is range of values (2^03 to 2^04-1) */
        if ( snprintf( label, (size_t) label_size, "2^%u to 2^%u-1", 
                       key, key + 1 ) > label_size ) 
        {
            return 1;
        }
        return 0;

      case BINSCHEME_DECIMAL:
        /* label is the median value of possible keys in that bin */
        if ( key < (skBag_key_t) 100 ) {
            mid = key;
        } else {
            min = ceil( pow( 10, ( ( (double) key / 100.0 ) + 1.0 ) ) );
            max = floor( pow( 10, ( ( ( (double) key + 1.0 ) / 100.0 ) + 1 ) 
                              ) );
            mid = (skBag_key_t) floor( ( min + max ) / 2 );
        }

        if ( snprintf( label, (size_t) label_size, "%u", mid ) 
             > label_size ) 
        {
            return 1;
        }
        return 0;

      case BINSCHEME_NONE:
        skAppPrintErr( "bad bin scheme" );
        assert(0);
        abort();
    }

    assert(0);
    abort();
    return 0;
}


#define INDENT_STRING "  "
static void printIndent(
    FILE   *fp, 
    int     indent ) 
{
    int i;
    for ( i = 0; i < indent; i++ ) {
        fprintf( fp, INDENT_STRING );
    }
}


/*
 * print a single IP address according to output options
 */
static void rwbagcatPrintIP(
    FILE           *outstream, 
    skBag_key_t     key )
{
    if ( f_print_integer_keys ) {
        if ( output_is_columnar ) {
            fprintf( outstream, "%*u", COUNT_WIDTH,
                     (uint32_t) key );
        } else {
            fprintf( outstream, "%u", (uint32_t) key );
        }
    } else {
        if ( f_zero_pad_ips ) {
            fprintf( outstream, "%s", num2dot0( (uint32_t) key ) );
        } else {
            if ( output_is_columnar ) {
                fprintf( outstream, "%*s", COUNT_WIDTH,
                         num2dot( (uint32_t) key ) );
            } else {
                fprintf( outstream, "%s", num2dot( (uint32_t) key ) );
            }
        }
    }
}


static int printInvertedBag( skBag_header_t *p_bag )
{
    char s_label[32];
    skBag_iterator_t *p_iter;
    skBag_key_t key;
    skBag_counter_t count;

    if ( skBag_allocIterator( p_bag, &p_iter ) != SKBAG_OK ) {
        return 1;
    }
    
    while ( skBag_getNext( p_iter, &key, &count ) == SKBAG_OK ) {
        if ( calculateBinLabel( key, bin_scheme, s_label, 32 ) != 0 ) {
            skAppPrintErr( "bin label truncated" );
        }

        if (output_is_columnar) {
            fprintf( output.stream, "%*s%c%*llu%c\n",
                     COUNT_WIDTH, s_label, output_delimiter, 
                     COUNT_WIDTH, count, output_delimiter );
        } else {
            fprintf( output.stream, "%s%c%llu%c\n",
                     s_label, output_delimiter, count, output_delimiter );
        }
    }

    if ( skBag_freeIterator( p_iter ) != SKBAG_OK ) {
        return 1;
    }
    return 0;
}


/*
 * print an individual key
 */
static void printKey(
    skBag_key_t         key, 
    skBag_counter_t     count,
    print_state_t      *state ) 
{
    int change_A, change_B, change_C, change_X;

    /*
     * calculate whether or not a given octet has changed.  since
     * the calculation is done to determine whether or not we need
     * to print new headers/footers, processing the first entry
     * and reaching the end of the data set counts as a change.
     */
    if ( state->f_last_entry ) {
        change_A = change_B = change_C = change_X = 1;
    } else if ( state->f_first_entry ) {
        state->f_first_entry = 0;
        change_A = change_B = change_C = change_X = 1;
    } else {
        change_A = ( OCTET_A( key ) != OCTET_A( state->prev_key ) );
        change_B = ( OCTET_B( key ) != OCTET_B( state->prev_key ) );
        change_C = ( OCTET_C( key ) != OCTET_C( state->prev_key ) );
        change_X = ( OCTET_X( key ) != OCTET_X( state->prev_key ) );
    }
    
    /*
     * if counting and summarizing, calculate changes in blocks
     * and close out blocks.
     */

    /* Close out old X block (/27)  */
    if ( print_options.f_print_X && ( ( state->class_X_seen_H > 0 ) &&
                                      ( change_X || change_C || 
                                        change_B || change_A ) ) )
    {
        printIndent( output.stream, 
                     print_options.f_print_T + 
                     print_options.f_print_A + 
                     print_options.f_print_B + 
                     print_options.f_print_C );
        rwbagcatPrintIP( output.stream, state->prev_key & MASK_ABCX );
        fprintf( output.stream, "/27" );
        fprintf( output.stream, "%c", output_delimiter );
        fprintf( output.stream, PRINT_COUNT, state->class_X_count_H );
        if ( print_options.f_print_S ) {
            fprintf( output.stream, "%c", output_delimiter );
            fprintf( output.stream, PRINT_X_SUMMARY, state->class_X_seen_H );
        }
        fprintf( output.stream, "\n" );

        /* reset sub-block counters */
        state->class_X_seen_H = 0;
        state->class_X_count_H = 0;
    }
        
    /* Close out old C block */
    if ( print_options.f_print_C && ( ( state->class_C_seen_H > 0 ) && 
                                      ( change_C || change_B || 
                                        change_A ) ) )
    {
        printIndent( output.stream, 
                     print_options.f_print_T + 
                     print_options.f_print_A + 
                     print_options.f_print_B );
        rwbagcatPrintIP( output.stream, state->prev_key & MASK_ABC );
        fprintf( output.stream, "/24" );
        fprintf( output.stream, "%c", output_delimiter );
        fprintf( output.stream, PRINT_COUNT, state->class_C_count_H );
        if ( print_options.f_print_S ) {
            fprintf( output.stream, "%c", output_delimiter );
            fprintf( output.stream, PRINT_C_SUMMARY, state->class_C_seen_H,
                     state->class_C_seen_X );
        }
        fprintf( output.stream, "\n" );
            
        /* reset sub-block counters */
        state->class_C_seen_X = 0;
        state->class_C_count_X = 0;
        state->class_C_seen_H = 0;
        state->class_C_count_H = 0;
    }
    
    /* Close out old B block */
    if ( print_options.f_print_B && ( ( state->class_B_seen_H > 0 ) && 
                                      ( change_B || change_A ) ) )
    {
        printIndent( output.stream, 
                     print_options.f_print_T + 
                     print_options.f_print_A );
        rwbagcatPrintIP( output.stream, state->prev_key & MASK_AB );
        fprintf( output.stream, "/16" );
        fprintf( output.stream, "%c", output_delimiter );
        fprintf( output.stream, PRINT_COUNT, state->class_B_count_H );
        if ( print_options.f_print_S ) {
            fprintf( output.stream, "%c", output_delimiter );
            fprintf( output.stream, PRINT_B_SUMMARY, state->class_B_seen_H, 
                     state->class_B_seen_C, state->class_B_seen_X );
        }
        fprintf( output.stream, "\n" );
            
        /* reset sub-block counters */
        state->class_B_seen_C = 0;
        state->class_B_count_C = 0;
        state->class_B_seen_X = 0;
        state->class_B_count_X = 0;
        state->class_B_seen_H = 0;
        state->class_B_count_H = 0;
    }
        
    /* Close out old A block */
    if ( print_options.f_print_A && ( ( state->class_A_seen_H > 0 ) && 
                                      change_A ) )
    {
        printIndent( output.stream, 
                     print_options.f_print_T );
        rwbagcatPrintIP( output.stream, state->prev_key & MASK_A );
        fprintf( output.stream, "/8" );
        fprintf( output.stream, "%c", output_delimiter );
        fprintf( output.stream, PRINT_COUNT, state->class_A_count_H );
        if ( print_options.f_print_S ) {
            fprintf( output.stream, "%c", output_delimiter );
            fprintf( output.stream, PRINT_A_SUMMARY, state->class_A_seen_H, 
                     state->class_A_seen_B, state->class_A_seen_C,
                     state->class_A_seen_X );
        }
        fprintf( output.stream, "\n" );
            
        /* reset sub-block counters */
        state->class_A_seen_B = 0;
        state->class_A_count_B = 0;
        state->class_A_seen_C = 0;
        state->class_A_count_C = 0;
        state->class_A_seen_X = 0;
        state->class_A_count_X = 0;
        state->class_A_seen_H = 0;
        state->class_A_count_H = 0;
    }


    /*
     * Now that we've closed the footers, if we are at the end of
     * the data set, we can quit.
     */
    if ( state->f_last_entry ) {
        /* Close out old total block */
        if ( print_options.f_print_T ) {
            fprintf( output.stream, "TOTAL" );
            fprintf( output.stream, "%c", output_delimiter );
            fprintf( output.stream, PRINT_COUNT, state->count_H );
            if ( print_options.f_print_S ) {
                fprintf( output.stream, "%c", output_delimiter );
                fprintf( output.stream, PRINT_T_SUMMARY, 
                         state->seen_H, 
                         state->seen_A, state->seen_B, state->seen_C, 
                         state->seen_X );
            }
            fprintf( output.stream, "\n" );
        }
        return;
    }

    /* Open new A block */
    if ( change_A ) { 

        /* increment counters of class As */
        ++state->seen_A;
        state->count_A += count;

        /* wrt this class A, no class Bs, class Cs, /27s or hosts yet */
        state->class_A_seen_B = 0; 
        state->class_A_count_B = 0;
        state->class_A_seen_C = 0;
        state->class_A_count_C = 0;
        state->class_A_seen_X = 0;
        state->class_A_count_X = 0;
        state->class_A_seen_H = 0;
        state->class_A_count_H = 0;
    } 

    /* Open new B block */
    if ( change_B || change_A ) {
        /* increment counters of class Bs */
        ++state->seen_B;
        state->count_B += count;
        ++state->class_A_seen_B;
        state->class_A_count_B += count;
            
        /* wrt this class B, no class Cs, /27s, or hosts yet */
        state->class_B_seen_C = 0; 
        state->class_B_count_C = 0;
        state->class_B_seen_X = 0;
        state->class_B_count_X = 0;
        state->class_B_seen_H = 0;
        state->class_B_count_H = 0;
    }

    /* Open new C block */
    if ( change_C || change_B || change_A ) {
        /* increment counters of class Cs */
        ++state->seen_C;
        state->count_C += count;
        ++state->class_A_seen_C;
        state->class_A_count_C += count;
        ++state->class_B_seen_C;
        state->class_B_count_C += count;

        /* wrt this class C, no /27s or hosts yet */
        state->class_C_seen_X = 0; 
        state->class_C_count_X = 0;
        state->class_C_seen_H = 0;
        state->class_C_count_H = 0;
    }

    /* Open new "X" block */
    if ( change_X || change_C || change_B || change_A ) {
        /* increment counters of /27s */
        ++state->seen_X;
        state->count_X += count;
        ++state->class_A_seen_X;
        state->class_A_count_X += count;
        ++state->class_B_seen_X;
        state->class_B_count_X += count;
        ++state->class_C_seen_X;
        state->class_C_count_X += count;

        /* wrt this /27, no hosts yet */
        state->class_X_seen_H = 0; 
        state->class_X_count_H = 0;
    }


    /* print hosts if necessary. */
    if ( print_options.f_print_H ) {
        printIndent( output.stream, 
                     print_options.f_print_T + 
                     print_options.f_print_A + 
                     print_options.f_print_B + 
                     print_options.f_print_C +
                     print_options.f_print_X );
        rwbagcatPrintIP( output.stream, key );

        fprintf( output.stream, "%c%*llu%c\n",
                 output_delimiter,
                 (output_is_columnar ? COUNT_WIDTH : 0), count,
                 output_delimiter );
    }

    ++state->seen_H;
    state->count_H += count;
    ++state->class_A_seen_H;
    state->class_A_count_H += count;
    ++state->class_B_seen_H;
    state->class_B_count_H += count;
    ++state->class_C_seen_H;
    state->class_C_count_H += count;
    ++state->class_X_seen_H;
    state->class_X_count_H += count;

    state->prev_key = key;

}


static int printNetwork( skBag_header_t *p_bag ) 
{
    skBag_key_t key;
    skBag_counter_t count;

    skBag_iterator_t *p_iter;
    skBag_err_t rv;
    skBag_key_t cur_key;

    print_state_t state;
    memset( &state, 0, sizeof( print_state_t ) );
    state.f_first_entry = 1;

    if ( f_print_zero_counts ) {
        /*
         * if we are to print zero counts, then we just loop through
         * all values, printing them.
         */
        for ( cur_key = minkey; cur_key <= maxkey; ++cur_key ) {
            rv = skBag_getCounter( p_bag, &cur_key, &count );
            switch ( rv ) {
              case SKBAG_ERR_KEY_NOT_FOUND:
                count = 0;
                /* FALL THROUGH TO SKBAG_OK */
              case SKBAG_OK:
                printKey( cur_key, count, &state );
                break;

              default:
                skAppPrintErr( "Error %d reading from bag", rv );
            }
        }
    } else {
        /*
         * otherwise, only pull the values we will actually print
         */
        if ( skBag_allocIterator( p_bag, &p_iter ) != SKBAG_OK ) {
            return 1;
        }

        rv = SKBAG_OK;
        while ( rv == SKBAG_OK ) {
            /*
             * will run one more time than we have entries (in order to
             * close blocks).
             */
            rv = skBag_getNext( p_iter, &key, &count );

            /*
             * If the key is not in range, we want to skip the printing
             * loop and continue with the next key.  The only exception is
             * that on the last key, we always want to print the final
             * footers.
             */
            if ( ( key < minkey ) || ( key > maxkey ) || 
                 ( count < mincount ) || ( count > maxcount ) )
            {
                if ( rv == SKBAG_OK ) {
                    continue;
                }
            }

            if ( rv == SKBAG_ERR_KEY_NOT_FOUND ) {
                state.f_last_entry = 1;
            }

            printKey( key, count, &state );
        }

        skBag_freeIterator( p_iter );
    }

    /* print out final statistics */
    if ( print_options.f_print_stats ) {
        fprintf( stats.stream, "\n" );
        fprintf( stats.stream, "Network Summary\n" );
        fprintf( stats.stream, "\t%llu hosts, %f%% of 2^32\n",
                 state.seen_H, ( (double) state.seen_H ) / 
                 ( (double) ( (uint64_t) 1 << 32 ) ) );
        fprintf( stats.stream, "\t%d occupied /8s, %f%% of 2^8\n",
                 state.seen_A, ( (double) state.seen_A ) / 
                 ( (double) ( 1 << 8 ) ) );
        fprintf( stats.stream, "\t%d occupied /16s, %f%% of 2^16\n",
                 state.seen_B, ( (double) state.seen_B ) / 
                 ( (double) ( 1 << 16 ) ) );
        fprintf( stats.stream, "\t%d occupied /24s, %f%% of 2^24\n",
                 state.seen_C, ( (double) state.seen_C ) / 
                 ( (double) ( 1 << 24 ) ) );
        fprintf( stats.stream, "\t%d occupied /27s, %f%% of 2^32\n",
                 state.seen_X, ( (double) state.seen_X ) / 
                 ( (double) ( 1 << 27 ) ) );
    }

    return 0;
}


static int printStats(
    skBag_header_t *p_bag, 
    FILE           *fp_out ) 
{
    double count_temp =  0.0;
    double count_mult =  0.0;
    double sum = (uint64_t) 0.0; /* straight sum */
    double sum2 =  0.0; /* sum of squares */
    double sum3 =  0.0; /* sum of cubes */
    
    double key_count = 0.0;
    double mean =  0.0;
    double stddev =  0.0;
    double temp =  0.0;
    double variance =  0.0;
    double skew =  0.0;
    double kurtosis =  0.0;
    
    skBag_iterator_t *p_iter;
    skBag_key_t key;
    skBag_counter_t count;
    
    assert( p_bag != NULL );
    assert( fp_out != NULL );
    
    fprintf( fp_out, "\nStatistics\n" );
    
    if ( skBag_allocIterator( p_bag, &p_iter ) != SKBAG_OK ) {
        return 1;
    }

    
    while ( skBag_getNext( p_iter, &key, &count ) == SKBAG_OK ) {
        if ( ( key >= minkey ) && ( key <= maxkey ) &&
             ( count >= mincount ) && ( count <= maxcount ) )
        {
            /* straight sum */
            count_temp = (double) count;
            sum += count_temp;

            count_mult = count_temp;
            /* sum of squares */
            count_mult *= count_temp;
            sum2 += count_mult;

            /* sum of cubes */
            count_mult *= count_temp;
            sum3 += count_mult;

            ++key_count;
        }
    }

    if ( key_count == (double) 0.0 ) {
        fprintf( fp_out, "  No entries in bag.\n" );
        return 0;
    }
    
    if ( skBag_freeIterator( p_iter ) != SKBAG_OK ) {
        return 1;
    }

    /* formulae derived from HyperStat Online - David M. Lane */

    /* http://davidmlane.com/hyperstat/A15885.html (mean) */
    mean = sum / key_count;

    /* http://davidmlane.com/hyperstat/A16252.html (variance) */

    temp = sum2 
           - ( 2.0 * mean * sum )
           + ( key_count * mean * mean );

    variance = temp / ( key_count - 1.0 );
    
    /* http://davidmlane.com/hyperstat/A16252.html (standard deviation) */
    stddev = sqrt( variance );

    /* http://davidmlane.com/hyperstat/A11284.html (skew) */
    skew = ( ( sum3
               - ( 3.0 * mean * sum2 )
               + ( 3.0 * mean * mean * sum )
               - ( key_count * mean * mean * mean ) )
             / ( key_count * variance * stddev ) );



    /* http://davidmlane.com/hyperstat/A53638.html (kurtosis) */
    kurtosis = ( temp * temp ) / ( key_count * variance * variance );

    fprintf( fp_out, "%18s:  %u\n", "keys", (uint32_t) key_count );
    fprintf( fp_out, "%18s:  %u\n", "sum of counters", (uint32_t) sum );
    fprintf( fp_out, "%18s:  %.4g\n", "mean", mean );
    fprintf( fp_out, "%18s:  %.4g\n", "variance", variance );
    fprintf( fp_out, "%18s:  %.4g\n", "standard deviation", stddev );
    fprintf( fp_out, "%18s:  %.4g\n", "skew", skew );
    fprintf( fp_out, "%18s:  %.4g\n", "kurtosis", kurtosis );

    return 0;
}


/*
 * Output bag using current state of options
 */
static int processBag( skBag_header_t *p_bag ) 
{

    /* FIXME:  What's the right sort of bag to build for the inversion */
    skBag_level_t inverted_level_sizes[4] = { 9, 9, 9, 5 };
    
    if ( print_options.f_print_network != 0 ) {
        if ( printNetwork( p_bag ) != 0 ) {
            skAppPrintErr( "Cannot print network structure" );
            appTeardown();
            exit( EXIT_FAILURE );
        }
    }


    if ( bin_scheme != BINSCHEME_NONE ) {
        skBag_iterator_t *p_iter;
        skBag_key_t key;
        skBag_counter_t counter;

        /* Create an inverted bag */
        skBag_header_t *p_inverted_bag;
        if ( skBag_alloc( &p_inverted_bag, (skBag_level_t) 4, 
                          inverted_level_sizes ) != SKBAG_OK )
        {
            return 1;
        }

        if ( skBag_allocIterator( p_bag, &p_iter ) != SKBAG_OK ) {
            return 1;
        }
        
        /* add inverted entry to bag */
        while ( skBag_getNext( p_iter, &key, &counter ) == SKBAG_OK ) {

            if ( ( key < minkey ) || ( key > maxkey ) || 
                 ( counter < mincount ) || ( counter > maxcount ) )
            {
                continue;
            }


            key = calculateBinIndex( &counter, bin_scheme );
            if ( skBag_incrCounter( p_inverted_bag, &key ) != SKBAG_OK ) {
                return 1;
            }
        }

        skBag_freeIterator( p_iter );
        
        if ( printInvertedBag( p_inverted_bag ) != 0 ) {
            return 1;
        }

    }
    
    if ( print_options.f_print_stats ) {
        printStats( p_bag, stats.stream );
    }

    if ( print_options.f_print_treestats ) {
        skBag_printTreeStats( p_bag, treestats.stream );
    }
    
    return 0;
}


int main( int argc, char **argv ) 
{
    skBag_err_t err;
    skBag_header_t *p_bag;
    io_t *input_cur;

    appSetup( argc, argv ); /* never returns on error */

    input_cur = inputs;
    while ( input_cur != NULL ) {
        err = skBag_readBinary( &p_bag, input_cur->stream );
        if ( err != SKBAG_OK ) {
            skAppPrintErr("error reading bag from input stream" );
            exit( EXIT_FAILURE );
        }

        if ( processBag( p_bag ) ) {
            skAppPrintErr("error processing bag" );
            exit( EXIT_FAILURE );
        }

        err = skBag_free( p_bag );
        if ( err != SKBAG_OK ) {
            skAppPrintErr("error deallocating bag" );
            exit( EXIT_FAILURE );
        }

        input_cur = input_cur->next;
    }
    
    /* done */
    appTeardown();

    return 0;
}
