/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**    rwbagtool performs various operations on bags.  It can add bags
**    together, subtract a subset of data from a bag, perform key
**    intersection with an IP set, extract the key list of a bag as an
**    IP set, or filter bag records based on their counter value.
**
*/


#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwset/rwbagtool.c,v 1.11 2005/10/10 15:16:18 leechr Exp $");

#include "utils.h"
#include "bagtree.h"
#include "iptree.h"

/* LOCAL DEFINES AND TYPEDEFS */

/* where to write --help output */
#define USAGE_FH stdout

typedef enum _appOptionsEnum {
    RWBT_ADD,
    RWBT_INTERSECT,
    RWBT_COMPLEMENT,
    RWBT_DIVIDE,
    RWBT_SUBTRACT,
    RWBT_COVER_SET,
    RWBT_INVERT,
    RWBT_MIN_COUNT,
    RWBT_MAX_COUNT,
    RWBT_MIN_KEY,
    RWBT_MAX_KEY,
    RWBT_OUTPUT
} appOptionsEnum;


/* A struct to hold a bag and all information about its relation to
 * the file system. */
typedef struct {
    skBag_header_t *bag;
    const char     *pathname;
    FILE           *filep;
    int             is_pipe;
} bagfile_t;


/* EXPORTED FUNCTIONS */


/* LOCAL FUNCTIONS */

static void appUsage(void);                  /* never returns */
static void appUsageLong(void);              /* never returns */
static void appOptionsUsage(FILE*);          /* application options usage */
static void appTeardown(void);
static void appSetup(int argc, char **argv); /* never returns when error */
static int  appOptionsHandler(clientData, int, char *);

static int processBagDivide( 
    skBag_iterator_t *iter_ptr );

static int readBagFromFile( 
    skBag_header_t **bag, 
    const char *path );

static int file_index = 0; /* Index of current file argument in argv */

/* LOCAL VARIABLES */

/* What action the user wants to take (add, intersect, etc).  The
 * variable 'user_action' should be a value between RWBT_ADD and
 * RWBT_COVER_SET.  Set 'user_action_none' to a value outside that
 * range; this denotes that the user did not choose a value.
 */
static const appOptionsEnum user_action_none = RWBT_OUTPUT;
static appOptionsEnum user_action;

/* When outputting the new bag, only output entries whose keys and/or
 * whose values lie between these limits */
static skBag_counter_t min_count = 1;
static skBag_counter_t max_count = ~( (skBag_counter_t) 0 );
static skBag_key_t min_key = 0;
static skBag_key_t max_key = ~( (skBag_key_t) 0 );

/* When data appears on stdin, this var is 1; else it's 0 */
static int data_on_stdin = 0;

/* The bag we will output */
static bagfile_t output_bagfile;

/* Set to use for intersect and complement intersect */
static skIPTree_t *mask_set_ptr;

/* The file from which to read that set */
static char *op_filename = NULL;

/* OPTIONS SETUP */

static struct option appOptions[] = {
    {"add",                  NO_ARG,       0, RWBT_ADD},
    {"intersect",            REQUIRED_ARG, 0, RWBT_INTERSECT},
    {"complement-intersect", REQUIRED_ARG, 0, RWBT_COMPLEMENT},
    {"divide",               NO_ARG,       0, RWBT_DIVIDE},
    {"subtract",             NO_ARG,       0, RWBT_SUBTRACT},
    {"coverset",             NO_ARG,       0, RWBT_COVER_SET},
    {"invert",               NO_ARG,       0, RWBT_INVERT},
    {"mincount",             REQUIRED_ARG, 0, RWBT_MIN_COUNT},
    {"maxcount",             REQUIRED_ARG, 0, RWBT_MAX_COUNT},
    {"minkey",               REQUIRED_ARG, 0, RWBT_MIN_KEY},
    {"maxkey",               REQUIRED_ARG, 0, RWBT_MAX_KEY},
    {"output",               REQUIRED_ARG, 0, RWBT_OUTPUT},
    {0,0,0,0}  /* sentinel entry */
};

static const char *appHelp[] = {
    "Add all bag files read from stdin and the command line",
    "Masks keys in bag file using IPs in given SET file",
    "Masks keys in bag file using IPs NOT in given SET file",
    "Divide the first bag by all subsequent bags",
    "Substract from first bag file all subsequent bag files",
    "Extract the IPs from the bag file into a set file",
    "Count keys for each unique value",
    "Output records whose count is at least VALUE.  Default=1",
    "Output records whose count is less than VALUE.  Default=max",
    "Output records whose key is at least VALUE.  Default 0.0.0.0",
    "Output records whose key is less than VALUE.  Default 255.255.255.255",
    "Redirect output to specified file.",
    (char *)NULL
};



/*
 * appUsage:
 *      print short usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 *      print complete usage information to USAGE_FH and exit with
 *      code.  passed to optionsSetup() to print usage when --help
 *      option given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
    fprintf(USAGE_FH,
            ("Usage: %s [app_opts] bag-file [bag-files...] \n"
             "\t Perform operations on bag files.  Requires at least one bag\n"
             "\t file read from stdin or given on the command line.\n"),
            skAppName());

    appOptionsUsage(USAGE_FH);

    exit(EXIT_FAILURE);
}


/*
 * appOptionsUsage:
 *      print options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\napp options\n");
    for (i = 0; appOptions[i].name; i++) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                (appOptions[i].has_arg
                 ? ((appOptions[i].has_arg == 1)
                    ? "Required Arg"
                    : "Optional Arg")
                 : "No Arg"), appHelp[i]);
    }

    return;
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    optionsTeardown();

    if (output_bagfile.filep) {
        if (output_bagfile.is_pipe == 1) {
            if (-1 == pclose(output_bagfile.filep)) {
                skAppPrintErr("Error closing %s: %s",
                              output_bagfile.pathname, strerror(errno));
            }
        } else if (output_bagfile.is_pipe == 0) {
            if (EOF == fclose(output_bagfile.filep)) {
                skAppPrintErr("Error closing %s: %s",
                              output_bagfile.pathname, strerror(errno));
            }
        }
    }
    if (output_bagfile.pathname) {
        free((char*)output_bagfile.pathname);
    }

    if (op_filename) {
        free(op_filename);
    }

    return;
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    skAppRegister(argv[0]);

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* Initialize globals */
    user_action = user_action_none;
    memset(&output_bagfile, 0, sizeof(bagfile_t));

    /* Initialize the options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);                     /* never returns */
    }

    /* register the options handler */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL))
    {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);                     /* never returns */
    }

    /* parse the options; returns the index into argv[] of the first
     * non-option or < 0 on error  May re-arrange argv[]. */
    file_index = optionsParse(argc, argv);
    assert(file_index <= argc);
    if (file_index < 0) {
        /* options parsing should print error */
        appUsage();             /* never returns */
    }

    /* Make certain user selected some action to perform */
    if (user_action == user_action_none) {
        skAppPrintErr("One of --add, --subtract, --coverset, --intersect, "
                      "--divide, \n  --invert or --complement-intersect "
                      "must be specified.");
        appUsage();
    }

    /* file_index is looking at first file name to process.  We
     * require at least one bag file, either from stdin or from the
     * command line. */
    if ( !FILEIsATty(stdin) ) {
        data_on_stdin = 1;
    } else if (file_index == argc) {
        skAppPrintErr("No input files on command line and"
                      " stdin is connected to a terminal");
        appUsage();
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;                     /* OK */
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    char *ep;

    switch (opt_index) {
      case RWBT_ADD:
        if (user_action != user_action_none) {
            skAppPrintErr("switches %s and %s are incompatible",
                          appOptions[opt_index].name,
                          appOptions[user_action].name);
            return 1;
        }
        user_action = opt_index;
        break;

      case RWBT_INTERSECT:
        if (user_action != user_action_none) {
            skAppPrintErr("switches %s and %s are incompatible",
                          appOptions[opt_index].name,
                          appOptions[user_action].name);
            return 1;
        }
        user_action = opt_index;
        op_filename = strdup(opt_arg);
        break;

      case RWBT_COMPLEMENT:
        if (user_action != user_action_none) {
            skAppPrintErr("switches %s and %s are incompatible",
                          appOptions[opt_index].name,
                          appOptions[user_action].name);
            return 1;
        }
        user_action = opt_index;
        op_filename = strdup(opt_arg);
        break;

      case RWBT_DIVIDE:
        if (user_action != user_action_none) {
            skAppPrintErr("switches %s and %s are incompatible",
                          appOptions[opt_index].name,
                          appOptions[user_action].name);
            return 1;
        }
        user_action = opt_index;
        //op_filename = strdup(opt_arg);
        break;

      case RWBT_SUBTRACT:
        if (user_action != user_action_none) {
            skAppPrintErr("switches %s and %s are incompatible",
                          appOptions[opt_index].name,
                          appOptions[user_action].name);
            return 1;
        }
        user_action = opt_index;
        break;

      case RWBT_COVER_SET:
        if (user_action != user_action_none) {
            skAppPrintErr("switches %s and %s are incompatible",
                          appOptions[opt_index].name,
                          appOptions[user_action].name);
            return 1;
        }
        user_action = opt_index;
        break;

      case RWBT_INVERT:
        if (user_action != user_action_none) {
            skAppPrintErr("switches %s and %s are incompatible",
                          appOptions[opt_index].name,
                          appOptions[user_action].name);
            return 1;
        }
        user_action = opt_index;
        break;

      case RWBT_MIN_COUNT:
        min_count = strtoul(opt_arg, &ep, 10);
        if (ep == opt_arg || *ep != '\0') {
            skAppPrintErr("Could not parse '%s' as a number", opt_arg);
            return 1;
        }
        break;

      case RWBT_MAX_COUNT:
        max_count = strtoul(opt_arg, &ep, 10);
        if (ep == opt_arg || *ep != '\0') {
            skAppPrintErr("Could not parse '%s' as a number", opt_arg);
            return 1;
        }
        break;

      case RWBT_MIN_KEY:
        /* assume an integer; if character after first digit is '.',
         * then assume an IP */
        min_key = strtoul(opt_arg, &ep, 10);
        if (*ep == '.') {
            min_key = dot2num(opt_arg);
            if (min_key == 0xFFFFFFFF) {
                skAppPrintErr("Error parsing min key '%s' as IP", opt_arg);
            }
        } else if (ep == opt_arg || *ep != '\0') {
            skAppPrintErr("Could not parse min key '%s' as a number", opt_arg);
            return 1;
        }
        break;

      case RWBT_MAX_KEY:
        /* assume an integer; if character after first digit is '.',
         * then assume an IP */
        max_key = strtoul(opt_arg, &ep, 10);
        if (*ep == '.') {
            max_key = dot2num(opt_arg);
            if (max_key == 0xFFFFFFFF) {
                skAppPrintErr("Error parsing max key '%s' as IP", opt_arg);
            }
        } else if (ep == opt_arg || *ep != '\0') {
            skAppPrintErr("Could not parse max key '%s' as a number", opt_arg);
            return 1;
        }
        break;

      case RWBT_OUTPUT:
        if (opt_arg && (strcmp(opt_arg, "stdout")!=0)) {
            if (fileExists(opt_arg)) {
                skAppPrintErr("Output file '%s' exists. Remove the file first",
                              opt_arg);
                return 1;
            }
            output_bagfile.pathname = strdup(opt_arg);
        }
        break;

      default:
        skAppPrintErr("Option %d not handled in switch() at %s:%d",
                      opt_index, __FILE__, __LINE__);
        abort();
    }
    return 0;                   /* OK */
}


/*
 *  ok = processBag(bag);
 *
 *    Add or substract the contents of 'bag' to/from the global output
 *    bag 'output_bagfile'.  Whether to add or subtract is determined
 *    by the global 'user_action' variable.  Returns 0 on success or
 *    non-zero on error.
 */
static int processBag(skBag_header_t *bag_ptr)
{
    skBag_iterator_t *iter_ptr;
    skBag_key_t key;
    skBag_counter_t count;
    skBag_err_t rv_bag;

    /* Create an iterator to loop over the bag */
    rv_bag = skBag_allocIterator(bag_ptr, &iter_ptr);
    if (rv_bag != SKBAG_OK) {
        iter_ptr = NULL;
        goto END;
    }

    /* Iterate through the bag and process each record */
    switch (user_action) {
      case RWBT_COMPLEMENT:
      case RWBT_INTERSECT:
      case RWBT_ADD:
      case RWBT_COVER_SET:
        /* All of these require us to add all of the bags
           together; other option is to require some arguments to
           only take a single bag file */

        while (SKBAG_OK == (rv_bag = skBag_getNext(iter_ptr, &key, &count))) {
            rv_bag = skBag_addToCounter(output_bagfile.bag, &key, &count);
            if (rv_bag != SKBAG_OK) {
                if (rv_bag == SKBAG_ERR_OP_BOUNDS) {
                    skAppPrintErr("Overflow when adding bags");
                    goto END;
                }
                skAppPrintErr("Error %d when adding bags",
                              rv_bag);
                goto END;
            }
        }
        break;

      case RWBT_DIVIDE:
        if( processBagDivide( iter_ptr ) ) {
            /* an arbitrary errorm since processBagDivide() will print
               its own errors.  we just need to trigger a returned
               error code. */
            rv_bag = SKBAG_ERR_INPUT;
            goto END;
        }
        /* this is normal termination */
        rv_bag = SKBAG_ERR_KEY_NOT_FOUND;
        break;

      case RWBT_SUBTRACT:
        while (SKBAG_OK == (rv_bag = skBag_getNext(iter_ptr, &key, &count))) {
            rv_bag=skBag_subtractFromCounter(output_bagfile.bag, &key, &count);
            if (rv_bag != SKBAG_OK) {
                if (rv_bag == SKBAG_ERR_OP_BOUNDS) {
                    skAppPrintErr("Negative values in bag are not allowed");
                    goto END;
                }
                skAppPrintErr("Error %d when subtracting from bag",
                              rv_bag);
                goto END;
            }
        }
        break;

      default:
        skAppPrintErr("Option %d not handled in switch() at %s:%d",
                      user_action, __FILE__, __LINE__);
        abort();
    }

    /* We're either at the end of the bag or the getNext call failed */
    if (rv_bag == SKBAG_ERR_KEY_NOT_FOUND) {
        /* expected behavior: reached end of bag.  reset rv_bag */
        rv_bag = SKBAG_OK;
    } else {
        /* unexpected error: print message and set rv to that error */
        skAppPrintErr( "Error %d while iterating over bag", rv_bag );
    }

  END:
    if (iter_ptr) {
        skBag_freeIterator(iter_ptr);
    }

    return ((rv_bag == SKBAG_OK) ? 0 : -1);
}


/*
** processBagDivide
**
** Split off from processBag just for readability.  It divides the bag
** in outputbag by keys found in the iterator iter_ptr 
*/
static int processBagDivide( 
    skBag_iterator_t *divisor_iter_ptr ) 
{

    skBag_iterator_t *dividend_iter_ptr;
    skBag_err_t rv_dividend_bag;
    skBag_err_t rv_divisor_bag;

    skBag_key_t dividend_key;
    skBag_key_t divisor_key;

    skBag_counter_t dividend_count;
    skBag_counter_t divisor_count;
    skBag_counter_t result_count;

    skBag_err_t rv_bag;

    /* Create an iterator to loop over the dividend bag */
    rv_bag = skBag_allocIterator( output_bagfile.bag, &dividend_iter_ptr );
    if( rv_bag != SKBAG_OK ) {
        dividend_iter_ptr = NULL;
        goto END;
    }

    do {
        rv_dividend_bag = skBag_getNext( dividend_iter_ptr, &dividend_key, 
                                         &dividend_count );

        rv_divisor_bag = skBag_getNext( divisor_iter_ptr, &divisor_key, 
                                        &divisor_count );

        if( rv_dividend_bag != SKBAG_OK || rv_divisor_bag != SKBAG_OK ) {
            goto END;
        }

        if( dividend_key != divisor_key ) {
            skAppPrintErr( "Error dividing bags.  The keys in the two "
                           "bags do not match." );
            goto END;
        }

        if( dividend_count < divisor_count ) {
            skAppPrintErr( "Error dividing bags.  A divisor counter "
                           "is larger than a\n  dividend counter." );
            goto END;
        }
        
        result_count = dividend_count / divisor_count;
        
        rv_bag = skBag_setCounter( output_bagfile.bag, &dividend_key, 
                                   &result_count );
        if( rv_bag != SKBAG_OK ) {
            skAppPrintErr( "Unexpected error %d while setting counter after "
                           "division.", rv_bag );
            goto END;
        }
        
    } while( rv_dividend_bag == SKBAG_OK &&
             rv_divisor_bag == SKBAG_OK );

    /* We're either at the end of the bag or the getNext call failed */
    if( rv_dividend_bag == SKBAG_ERR_KEY_NOT_FOUND &&
        rv_divisor_bag == SKBAG_ERR_KEY_NOT_FOUND ) 
    {
        /* expected behavior:  reached end of both bags at same time.
           set rv_bag to SKBAG_OK */
        rv_bag = SKBAG_OK;
    } else if( ( rv_dividend_bag == SKBAG_ERR_KEY_NOT_FOUND &&
                 rv_divisor_bag == SKBAG_OK ) ||
               ( rv_dividend_bag == SKBAG_OK &&
                 rv_divisor_bag == SKBAG_ERR_KEY_NOT_FOUND ) )
    {
        /* one bag terminated before the other */
        skAppPrintErr( "Error dividing bags.  The keys in the two "
                       "bags do not match." );
        rv_bag = SKBAG_ERR_OP_BOUNDS;
    } else {
        if( rv_dividend_bag != SKBAG_OK ) {
            skAppPrintErr( "Error %d iterating over dividend bag", 
                           rv_dividend_bag );
            rv_bag = rv_dividend_bag;
        }

        if( rv_divisor_bag != SKBAG_OK ) {
            skAppPrintErr( "Error %d iterating over divisor bag", 
                           rv_divisor_bag );
            rv_bag = rv_divisor_bag;
        }
    }

  END:
    if( dividend_iter_ptr ) {
        skBag_freeIterator( dividend_iter_ptr );
    }
    
    return ( ( rv_bag == SKBAG_OK ) ? 0 : -1 );
}


/*
 *  ok = applyCutoffs(bag);
 *
 *    Run through the bag and zero out any entries not within range or
 *    which aren't in the masking set.  Return 0 on success, non-zero
 *    otherwise.
 *
 *    FIXME: We could do some of this during the insertion stage
 *    instead output to save ourselves some storage. This is not the
 *    most efficient implementation.
 */
static int applyCutoffs(skBag_header_t *bag_ptr)
{
    skBag_iterator_t *iter_ptr;
    skBag_key_t key;
    skBag_counter_t count;
    skBag_err_t rv_bag;

    /* Create an iterator to loop over the bag */
    rv_bag = skBag_allocIterator(bag_ptr, &iter_ptr);
    if (rv_bag != SKBAG_OK) {
        iter_ptr = NULL;
        goto END;
    }

    while ((rv_bag = skBag_getNext(iter_ptr, &key, &count)) == SKBAG_OK) {
        if ( (key < min_key)     || (key > max_key)     ||
             (count < min_count) || (count > max_count) ||
             ((user_action == RWBT_INTERSECT) &&
              !skIPTreeCheckAddress(key, mask_set_ptr)) ||
             ((user_action == RWBT_COMPLEMENT) &&
              skIPTreeCheckAddress(key, mask_set_ptr))  )
        {
            count = 0;
            rv_bag = skBag_setCounter(bag_ptr, &key, &count);
            if (rv_bag != SKBAG_OK) {
                skAppPrintErr("Error %d setting key %u to zero in bag",
                              rv_bag, key);
                goto END;
            }
        }
    }

    /* We're either at the end of the bag or the getNext call failed */
    if (rv_bag == SKBAG_ERR_KEY_NOT_FOUND) {
        /* expected behavior: reached end of bag.  reset rv_bag */
        rv_bag = SKBAG_OK;
    } else {
        /* unexpected error: print message and set rv to that error */
        skAppPrintErr("Error %d while interating over bag",
                      rv_bag);
    }

  END:
    if (rv_bag) {
        skBag_freeIterator(iter_ptr);
    }

    return ((rv_bag == SKBAG_OK) ? 0 : -1);
}


/*
 *  ok = writeOutput();
 *
 *    Generate the output as requested by the global 'user_action'
 *    variable.
 */
static int writeOutput(void)
{
    skBag_iterator_t *iter_ptr;
    skBag_key_t key;
    skBag_counter_t count;
    enum skIPTreeTypes rv_set;
    skBag_err_t rv_bag;
    skIPTree_t *cover_set_ptr;     /* Holds the current cover set */

    skBag_key_t temp_key;
    skBag_counter_t temp_count;

    /* Remove anything that's not in range or not in the intersecting
     * set (or complement) as appropriate */
    applyCutoffs(output_bagfile.bag);

    /* If we're not generating a cover set we're writing the output
     * bag, otherwise we generate and write out that set */
    if (user_action != RWBT_COVER_SET) {


        /* Inverting only takes one file, so we do that here */
        if (user_action == RWBT_INVERT) {
            /* Run through the bag */
            rv_bag = skBag_allocIterator(output_bagfile.bag, &iter_ptr);
            if (rv_bag != SKBAG_OK) {
                return -1;
            }

            temp_count = (skBag_counter_t) 1;
            while (SKBAG_OK == (rv_bag = skBag_getNext(iter_ptr, &key, 
                                                       &count))) 
            {
                temp_key = count;
                rv_bag = skBag_addToCounter(output_bagfile.bag, &temp_key, 
                                            &temp_count);
                if (rv_bag != SKBAG_OK) {
                    if (rv_bag == SKBAG_ERR_OP_BOUNDS) {
                        skAppPrintErr("Overflow when adding bags");
                        return -1;
                    }
                    skAppPrintErr("Error %d when adding bags",
                                  rv_bag);
                    return -1;
                }

                rv_bag = skBag_subtractFromCounter( output_bagfile.bag, &key, 
                                                    &count);
                if (rv_bag != SKBAG_OK) {
                    skAppPrintErr("Error %d when subtracting from bag",
                                  rv_bag);
                    return -1;
                }
            }
            skBag_freeIterator(iter_ptr);
        }

        if (SKBAG_OK != skBag_writeBinary(output_bagfile.bag,
                                          output_bagfile.filep))
        {
            skAppPrintErr("Error writing bag to output file '%s'",
                          output_bagfile.pathname);
            return -1;
        }
    } else {

        /* Create the cover set */
        rv_set = skIPTreeCreate(&cover_set_ptr);
        if (rv_set != SKIP_OK) {
            skAppPrintErr("Error creating cover set");
            return -1;
        }

        /* Run through the bag, adding items to the set */
        rv_bag = skBag_allocIterator(output_bagfile.bag, &iter_ptr);
        if (rv_bag != SKBAG_OK) {
            return -1;
        }

        while ((rv_bag = skBag_getNext(iter_ptr, &key, &count)) == SKBAG_OK) {
            skIPTreeAddAddress(key, cover_set_ptr);
        }
        if (rv_bag != SKBAG_ERR_KEY_NOT_FOUND) {
            /* unexpected error: print message and set rv to that error */
            skAppPrintErr("Error %d while interating over bag",
                          rv_bag);
        }

        skBag_freeIterator(iter_ptr);

        /* Write the set */
        if (skIPTreeWrite(output_bagfile.filep, cover_set_ptr)) {
            skAppPrintErr("Error writing set file.");
            return -1;
        }
    }

    return 0;
}


/*
 *  ok = readBagFromFile(&bag, path);
 *
 *    Fill in the bag object pointed to by 'bag' with the contents of
 *    'path'.  Return 0 on success, non-zero otherwise.
 */
static int readBagFromFile(skBag_header_t **bag, const char *path)
{
    FILE *fp;
    int is_pipe;

    if (openFile(path, 0/*read*/, &fp, &is_pipe)) {
        skAppPrintErr("Couldn't open file '%s' for reading",
                      path);
        return -1;
    }
    if (SKBAG_OK != skBag_readBinary(bag, fp)) {
        skAppPrintErr("Couldn't read bag from '%s'",
                      path);
        return -1;
    }
    if (is_pipe) {
        pclose(fp);
    } else {
        fclose(fp);
    }

    return 0;
}


int main(int argc, char **argv)
{
    skBag_header_t *input_bag;

    appSetup(argc, argv);                       /* never returns on error */

    /* If action is intersect or complement intersect read the set
     * file */
    if (user_action==RWBT_INTERSECT || user_action==RWBT_COMPLEMENT) {
        if ( skIPTreeLoad(op_filename, &mask_set_ptr) ) {
            skAppPrintErr("Couldn't open set file %s for reading",
                          op_filename);
            return 1;
        }
    }

    /* Open up main output file if necessary */
    if (output_bagfile.pathname == NULL) {
        if (FILEIsATty(stdout)) {
            skAppPrintErr("stdout is connected to a terminal");
            return 1;
        }
        output_bagfile.filep = stdout;
    } else {
        if (openFile(output_bagfile.pathname, 1/*write*/,
                     &output_bagfile.filep, &output_bagfile.is_pipe))
        {
            skAppPrintErr("Couldn't open file '%s' for writing",
                          output_bagfile.pathname);
            return 1;
        }
    }

    /* Read the first bag; this will be the basis of the output bag. */
    if ( data_on_stdin ) {
        /* use stdin as first bag */
        if (skBag_readBinary(&output_bagfile.bag, stdin) != SKBAG_OK) {
            skAppPrintErr("Couldn't read bag from stdin");
            return 1;
        }
    } else {
        if (readBagFromFile(&output_bagfile.bag, argv[file_index])) {
            return 1;
        }
        file_index++;
    }

    /* Open up each bag and process it appropriately */
    for ( ; file_index < argc; ++file_index) {
        if (readBagFromFile(&input_bag, argv[file_index])) {
            return 1;
        }

        /* Modify the output bag using the appropriate options */
        if (processBag(input_bag)) {
            return 1;
        }

        /* Blow the bag away. We're done with it */
        skBag_free(input_bag);
    }


    /* Write the output */
    writeOutput();

    /* free the output bag */
    skBag_free(output_bagfile.bag);

    /* done */
    appTeardown();

    return 0;
}
