/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
 * Rwset.c
 *
 * Michael Collins
 * May 6th
 *
 * rwset is an application which takes filter data and generates a
 * tree (rwset) of ip addresses which come out of a filter file.  This
 * tree can then be used to generate aggregate properties; rwsets will
 * be used for filtering large sets of ip addresses, aggregate properties
 * per element can be counted, &c.
 */

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwset/rwset.c,v 1.27 2005/10/26 18:33:42 thomasm Exp $");

#include "iptree.h"
#include "rwpack.h"
#include "utils.h"
#include "iochecks.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* where to write output from --help */
#define USAGE_FH stdout

/* number of set files that can be created (sip, dip, nhip) */
#define SET_FILE_TYPES 3


/* LOCAL FUNCTIONS */

static void appUsage(void);                     /* never returns */
static void appUsageLong(void);                 /* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);    /* never returns */
static void appOptionsUsage(FILE*);             /* application options usage */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);


/* LOCAL VARIABLES */

static int integerIPSFlag = 0;
static uint8_t print_ips = 0;
static uint8_t daddress_flag = 0;
static char *set_file_name = NULL;
static const char *file_name[SET_FILE_TYPES];
static skIPTree_t *iptree[SET_FILE_TYPES];
static iochecksInfoStructPtr ioISP;


/* OPTIONS SETUP */

typedef enum {
    RWSET_SIP_FILE,
    RWSET_DIP_FILE,
    RWSET_NHIP_FILE,
    RWSET_SADDRESS,
    RWSET_DADDRESS,
    RWSET_PRINT_IPS,
    RWSET_INTEGER_IPS,
    RWSET_SET_FILE
} appOptionsEnum;

static struct option appOptions[] = {
    {"sip-file",    REQUIRED_ARG, 0, RWSET_SIP_FILE},
    {"dip-file",    REQUIRED_ARG, 0, RWSET_DIP_FILE},
    {"nhip-file",   REQUIRED_ARG, 0, RWSET_NHIP_FILE},
    /* Deprecated switches follow: */
    {"saddress",    NO_ARG,       0, RWSET_SADDRESS},
    {"daddress",    NO_ARG,       0, RWSET_DADDRESS},
    {"print-ips",   NO_ARG,       0, RWSET_PRINT_IPS},
    {"integer-ips", NO_ARG,       0, RWSET_INTEGER_IPS},
    {"set-file",    REQUIRED_ARG, 0, RWSET_SET_FILE},
    {0,0,0,0}                     /* sentinal entry */
};

static const char *appHelp[] = {
    ("Create an IP set from source addresses\n"
     "\t and write it to the named file (file must not exist)"),
    ("Create an IP set from destination addresses\n"
     "\t and write it to the named file (file must not exist)"),
    ("Create an IP set from next-hop addresses\n"
     "\t and write it to the named file (file must not exist)\n"
     "\n"
     "The following options are deprecated and cannot be mixed with the above:"),
    ("Use source addresses as key (default behavior when no\n"
     "\t other options given.  Use --sip-set option instead."),
    ("Use destination addresses as key (use --dip-set instead)\n"
     "\n"
     "The following switches apply to --saddress and --daddress only:"),
    "Print IP addresses to stdout as dotted decimal. Def. No",
    "Print IP addresses to stdout as integers. Def. No",
    "Write binary IPset file to this location. Def. No",
    (char *)NULL
};


/*
 * appUsage:
 *      print usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 *      print usage information to USAGE_FH and exit with code 1.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                             \
    ("%s [options] [files]\n"                                                 \
     "\tRead SiLK Flow records and generate binary IPset file(s).  When no\n" \
     "\tfiles are given on command line, flows are read from STDIN.\n")

    fprintf(USAGE_FH, USAGE_MSG, skAppName());
    appOptionsUsage(USAGE_FH);
    exit(EXIT_SUCCESS);
}


/*
 * appOptionsUsage:
 *      print options for this app to specified file handle
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\noptions:\n");
    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }

    return;
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;
    int i;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    for (i = 0; i < SET_FILE_TYPES; ++i) {
        if (iptree[i]) {
            skIPTreeDelete(&iptree[i]);
            iptree[i] = NULL;
        }
    }

    iochecksTeardown(ioISP);
    optionsTeardown();

    return;
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int i;
    int stdout_used = 0;

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* Register application and initialize globals */
    skAppRegister(argv[0]);
    memset(file_name, 0, sizeof(file_name));
    memset(iptree, 0, sizeof(iptree));

    /* Initialize input checker */
    ioISP = iochecksSetup(0, 0, argc, argv);

    /* Initialize options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* Register options and the options handler */
    if (optionsRegister(appOptions, (optHandler) appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* Parse options */
    ioISP->firstFile = optionsParse(argc, argv);
    if (ioISP->firstFile < 0) {
        appUsage();/* never returns */
    }

    /* Make certain the user has requested some output.  Do not mix
     * old and new style arguments. */
    if (file_name[RWSET_SIP_FILE]
        || file_name[RWSET_DIP_FILE]
        || file_name[RWSET_NHIP_FILE])
    {
        if (set_file_name || print_ips) {
            skAppPrintErr("Cannot mix --set-file or --print-ips with the\n"
                          "  --sip, --dip, --nhip switches");
            appUsage();
        }
    } else if (!set_file_name && !print_ips) {
        skAppPrintErr("No output specified.  Use --sip-file, "
                      "--dip-file, and/or --nhip-file.");
        appUsage();
    }

    /* Convert legacy switches to new form */
    if (set_file_name) {
        /* convert set_file_name to sip or dip */
        if (daddress_flag) {
            file_name[RWSET_DIP_FILE] = set_file_name;
        } else {
            file_name[RWSET_SIP_FILE] = set_file_name;
        }
    } else if (print_ips) {
        /* convert print_ips to sip to dip file (as "/dev/null") */
        if (daddress_flag) {
            file_name[RWSET_DIP_FILE] = "/dev/null";
        } else {
            file_name[RWSET_SIP_FILE] = "/dev/null";
        }
    }

    /* Make certain we don't overwrite existing files, or that
     * multiple people have requested stdout, or that stdout is a
     * terminal. */
    for (i = 0; i < SET_FILE_TYPES; ++i) {
        if (NULL == file_name[i]) {
            /* output not requested */
        } else if (0 == strcmp("stdout", file_name[i])) {
            /* check stdout */
            if (stdout_used) {
                skAppPrintErr("Error: multiple sets going to stdout");
                exit(EXIT_FAILURE);
            } else if (FILEIsATty(stdout)) {
                skAppPrintErr("stdout is a terminal");
                exit(EXIT_FAILURE);
            }
            stdout_used = 1;
        } else if (fileExists(file_name[i])) {
            /* file exists */
            skAppPrintErr(("Output IPset file '%s' already exists.\n"
                           "\t Delete or move it and try again."),
                          file_name[i]);
            exit(EXIT_FAILURE);
        }
    }

    /* Create IP trees */
    for (i = 0; i < SET_FILE_TYPES; ++i) {
        if (file_name[i]) {
            if (skIPTreeCreate(&(iptree[i]))) {
                skAppPrintErr("Error allocating memory for ip tree '%s'",
                              file_name[i]);
                exit(EXIT_FAILURE);
            }
        }
    }

    /* Use STDIN as an input stream if it is not a TTY; make certain
     * we have some input and we are either reading from STDIN or
     * using files listed the command line, but not both. */
    if (iochecksAcceptFromStdin(ioISP) || iochecksInputs(ioISP, 0)) {
        appUsage();
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    (void) umask((mode_t) 0002);

    return;                       /* OK */
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int  appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    switch ((appOptionsEnum)opt_index) {
      case RWSET_SIP_FILE:
      case RWSET_DIP_FILE:
      case RWSET_NHIP_FILE:
        file_name[opt_index] = opt_arg;
        break;

      case RWSET_SADDRESS:
        daddress_flag = 0;
        break;

      case RWSET_DADDRESS:
        daddress_flag = 1;
        break;

      case RWSET_INTEGER_IPS:
        integerIPSFlag = 1;
        /* FALL THROUGH to set print flag as well */
      case RWSET_PRINT_IPS:
        print_ips = 1;
        break;

      case RWSET_SET_FILE:
        set_file_name = opt_arg;
        return 0;
    }

    return 0;                     /* OK */
}


/*
 *  status = rwsetProcessFile(filename);
 *
 *    Read the records from 'filename' and add them to the appropriate
 *    IPSet files.
 */
static int rwsetProcessFile(const char *filename)
{
    static rwRec rwrec;
    rwIOStruct *rwIOS;
    uint32_t *ip_p[SET_FILE_TYPES];
    int i;

    /* point ip_p[] at the address of the desired IP in the rwrec. */
    ip_p[RWSET_SIP_FILE]  = &rwrec.sIP.ipnum;
    ip_p[RWSET_DIP_FILE]  = &rwrec.dIP.ipnum;
    ip_p[RWSET_NHIP_FILE] = &rwrec.nhIP.ipnum;

    /* Open file */
    rwIOS = rwOpenFile(filename, NULL);
    if (NULL == rwIOS) {
        return 1;
    }

    /* Read in records */
    while (rwRead(rwIOS, &rwrec)) {
        for (i = 0; i < SET_FILE_TYPES; ++i) {
            if (iptree[i]) {
                skIPTreeAddAddress(*(ip_p[i]), iptree[i]);
            }
        }
    }

    rwCloseFile(rwIOS);
    return 0;
}


int main(int argc, char **argv)
{
    int rv = 0;
    int i;
    int counter;

    appSetup(argc, argv);                 /* never returns */

    /* Process files */
    for (counter = ioISP->firstFile; counter < ioISP->fileCount ; counter++) {
        if (rwsetProcessFile(ioISP->fnArray[counter])) {
            exit(EXIT_FAILURE);
        }
    }

    /* Generate output */
    for (i = 0; i < SET_FILE_TYPES; ++i) {
        if (iptree[i]) {
            if (print_ips) {
                skIPTreePrint(stdout, iptree[i], integerIPSFlag);
            }

            if (0 == strcmp("/dev/null", file_name[i])) {
                /* do nothing */
            } else if (0 == strcmp("stdout", file_name[i])) {
                if (skIPTreeWrite(stdout, iptree[i])) {
                    rv = 1;
                }
            } else if (skIPTreeSave(file_name[i], iptree[i])) {
                rv = 1;
            }

            skIPTreeDelete(&iptree[i]);
            iptree[i] = NULL;
        }
    }

    /* done */
    appTeardown();

    return rv;
}
