/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
 *
 * setintersect.c
 *
 * This is an application that takes multiple set files and intersects them,
 * generating a set of addresses that are in all sets.  Sets are
 * specified as 'removes' or adds, where a 'remove' file is one that
 * should have all of its addresses present removed, and an add file
 * will keep addresses that are in it.
 *
 */

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwset/setintersect.c,v 1.23 2005/10/26 14:23:18 thomasm Exp $");

#include "iptree.h"
#include "utils.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* where to write output from --help */
#define USAGE_FH stdout

/* maximum number of add and remove files */
#define ADD_FILES_LIMIT     4
#define REMOVE_FILES_LIMIT  4


/* LOCAL FUNCTIONS */

static void appUsage(void);                     /* never returns */
static void appUsageLong(void);                 /* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);    /* never returns */
static void appOptionsUsage(FILE *fh);          /* application options usage */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);


/* LOCAL VARIABLES */

static uint8_t printIPSFlag = 0;
static uint8_t integerIPSFlag = 0; /* whether to print ips as uint32_t */
static uint8_t addFiles, removeFiles;
static char *addFileNames[ADD_FILES_LIMIT];
static char *removeFileNames[REMOVE_FILES_LIMIT];
static char *targetFileName;


/* OPTIONS SETUP */

typedef enum {
    SIO_PRINT_IPS,
    SIO_INTEGER_IPS,
    SIO_ADD_FILE,
    SIO_REMOVE_FILE,
    SIO_TGTFILE
} appOptionsEnum;

static struct option appOptions[] = {
    {"print-ips",   NO_ARG,       0, SIO_PRINT_IPS},
    {"integer-ips", NO_ARG,       0, SIO_INTEGER_IPS},
    {"add-set",     REQUIRED_ARG, 0, SIO_ADD_FILE},
    {"remove-set",  REQUIRED_ARG, 0, SIO_REMOVE_FILE},
    {"set-file",    REQUIRED_ARG, 0, SIO_TGTFILE},
    {0,0,0,0}  /* sentinel entry */
};

static const char *appHelp[] = {
    "Print IP's as dotted quad to stdout",
    "Print IP's as 32 bit integers to stdout",
    "Specify a set of IPs to use",
    "Specify a set of IPs to remove",
    "File to write resulting set to",
    (char *)NULL
};



/*
 * appUsage:
 *      print usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
void appUsage(void) {
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 *      print usage information to USAGE_FH and exit with code 1.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                      \
    ("%s --add=<file> [options] {--print-ips | --set-file=<file>}\n"   \
     "\tGenerates a new IP-set by intersecting all --add-set binary\n" \
     "\tIP-set files, then removing all --remove-set binary IP-set\n"  \
     "\tfiles from the new IP-set.  Prints the resulting IP-set to\n"  \
     "\tstdout and/or writes it to the specified file.\n"              \
     "\tSupports 1-%d add sets and 0-%d remove sets.\n")

    fprintf(USAGE_FH, USAGE_MSG,
            skAppName(), ADD_FILES_LIMIT, REMOVE_FILES_LIMIT);
    appOptionsUsage(USAGE_FH);
    exit(EXIT_SUCCESS);
}


/*
 * appOptionsUsage:
 *      print options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\noptions:\n");
    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    optionsTeardown();

    return;
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int unhandled_option;

    /* initialize variables */
    skAppRegister(argv[0]);
    targetFileName = NULL;
    addFiles = 0;
    removeFiles = 0;

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* setup the options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* register our options and handler */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);                    /* never returns */
    }

    /* parse options */
    unhandled_option = optionsParse(argc, argv);
    assert(unhandled_option <= argc);
    if (unhandled_option < 0) {
        /* optionsParse() printed the error */
        appUsage();  /* never returns */
    }

    /* optionsParse() handles all arguments; if any are left over, it
     * is an error */
    if (unhandled_option != argc) {
        skAppPrintErr("Unrecognized argument '%s'", argv[unhandled_option]);
        appUsage();  /* never returns */
    }

    /* verify at least one add file */
    if (addFiles == 0) {
        skAppPrintErr("Please specify at least one add file");
        appUsage();  /* never returns */
    }

    /* verify output */
    if (!targetFileName && !printIPSFlag) {
        skAppPrintErr("Specify an output file name or use the ips flag");
        appUsage();
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    (void) umask((mode_t) 0002);

    return;                       /* OK */
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int  appOptionsHandler(
    clientData   UNUSED(cData),
    int                 opt_index,
    char               *opt_arg)
{
    switch ((appOptionsEnum)opt_index) {
      case SIO_INTEGER_IPS:
        integerIPSFlag = 1;
        /* FALLTHROUGH */
      case SIO_PRINT_IPS:
        printIPSFlag = 1;
        break;

      case SIO_TGTFILE:
        targetFileName = opt_arg;
        break;

      case SIO_ADD_FILE:
        if (addFiles == ADD_FILES_LIMIT) {
            skAppPrintErr("File limit exceeded. Only %d add files allowed",
                          ADD_FILES_LIMIT);
            return -1;
        }
        addFileNames[addFiles] = opt_arg;
        addFiles++;
        break;

      case SIO_REMOVE_FILE:
        if (removeFiles == REMOVE_FILES_LIMIT) {
            skAppPrintErr("File limit exceeded. Only %d remove files allowed",
                          REMOVE_FILES_LIMIT);
            return -1;
        }
        removeFileNames[removeFiles] = opt_arg;
        removeFiles++;
        break;
    }

    return 0;                     /* OK */
}


int main(int argc, char **argv)
{
    int i;
    skIPTree_t *newTree;
    skIPTree_t *tmpTree;
    newTree = tmpTree = NULL;

    appSetup(argc, argv);                 /* never returns on error */

    /* Read in initial --add file; appSetup should guarantee one add file */
    assert(addFiles > 0);
    if (skIPTreeLoad(addFileNames[0], &newTree)) {
        skAppPrintErr("Error reading tree from file '%s'", addFileNames[0]);
        return 1;
    }

    /* Process remaining --add files */
    for (i = 1; i < addFiles; i++) {
        skIPTreeDelete(&tmpTree);
        if (0 != skIPTreeLoad(addFileNames[i], &tmpTree)) {
            skAppPrintErr("Error reading tree from file '%s'",
                          addFileNames[i]);
            return 1;
        }
        skIPTreeIntersect(newTree, tmpTree);
    }

    /* Process --remove files */
    for (i = 0; i < removeFiles; i++) {
        skIPTreeDelete(&tmpTree);
        if (0 != skIPTreeLoad(removeFileNames[i], &tmpTree)) {
            skAppPrintErr("Error reading tree from file '%s'",
                          removeFileNames[i]);
            return 1;
        }
        skIPTreeSubstract(newTree, tmpTree);
    }

    /* Do output */
    if (printIPSFlag) {
        skIPTreePrint(stdout, newTree, integerIPSFlag);
    }

    if (targetFileName != NULL) {
        if (skIPTreeSave(targetFileName, newTree)) {
            appTeardown();
            return -1;
        };
    }

    /* done */
    appTeardown();

    return (0);
}
