/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwsort reads SiLK Flow Records from the standard input or from
**  named files and sorts them on one or more user-specified fields.
**
**  Up to MAX_IN_CORE_RECORDS will be stored in an in-memory
**  array---we will back off of the MAX_IN_CORE_RECORDS if we cannot
**  malloc enough RAM.  If less than that number of records are read,
**  the records are sorted and printed to standard out or to the named
**  output file.
**
**  When the 1+MAX_IN_CORE_RECORDS'th record is read, the array
**  containing the first MAX_IN_CORE_RECORDS is sorted and written to
**  a temporary file on disk; the array is cleared, the
**  1+MAX_IN_CORE_RECORDS'th record is added to array, then more
**  records are read up to the maximum, and another temp file is
**  written.  We continue until all records are read or until we write
**  MAX_TMP_FILES temporary files.  We then do a merge-sort on the
**  files, until we have two files left, and then those two files are
**  merged into the final output stream.
**
**  When the temporary files are written to the same volume (file
**  system) as the final output, the maximum disk usage will be
**  2-times the number of records read (times the size per record);
**  when different volumes are used, the disk space required for the
**  temporary files will be 1.5 times the number of records.
**
**  TO DO:
**
**    -- Should better handle the case where we reach the
**    MAX_TMP_FILES limit.
**
**    -- the temporary files should written and read using mmap()
**    instead of using librw.  One issue is that we need to be able to
**    handle very large (>2GB files); mmap will handle this, but we
**    will need to mmap() multiple times.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwsort/rwsort.c,v 1.22 2005/10/18 20:29:36 thomasm Exp $");

#include "rwpack.h"
#include "utils.h"
#include "iochecks.h"
#include "skstringmap.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* where to send --help output */
#define USAGE_FH stdout

/*
 *  For debugging.  Argument should be surrounded by double-parens:
 *      PRINTDEBUG((stderr, "My output\n"));
 */
/* #define PRINTDEBUG(args) fprintf args */
#ifndef PRINTDEBUG
#  define PRINTDEBUG(args)
#endif

/*
 *  Highest number of in-core records we support.  We will try to
 *  malloc a buffer to hold this records, but scale back if we fail.
 *
 *  Try to allocate almost 2GB of RAM.
 */
#define MAX_IN_CORE_RECORDS   (0x78000000 / sizeof(rwRec))

/*
 *  If we can't allocate space for at least this many records, give
 *  up.
 */
#define MIN_IN_CORE_RECORDS     1000

/*
 *  Maximum number of temporary files to create.  We need an array of
 *  this many char*'s to point to the file names.
 */
#define MAX_TMP_FILES  2048

/*
 *  Where to create temp files by default.  This can be overridden by
 *  the environment variable named in TMPDIR_ENVAR or by the
 *  --temp-dir switch.
 */
#define DEFAULT_TEMP_DIR "/tmp"

/*
 *  Name of envionment variable that holds name of temp directory.
 */
#define TMPDIR_ENVAR "TMPDIR"

/*
 *  Function type for signal handler
 */
typedef void (*sig_fn_t)(int);


/* LOCAL FUNCTIONS */

static void appUsage(void);                     /* never returns */
static void appUsageLong(void);                 /* never returns */
static void appTeardown(void);
static void appExit(int status);                /* never returns */
static void appSetup(int argc, char **argv);    /* never returns */
static void appOptionsUsage(FILE*);             /* application options usage */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static int  parseFields(const char *fields_string);
static void appHandleSignal(int signal_value);
static void appSetSignalHandler(sig_fn_t);

#define tempFileGetName(tmp_idx) \
    (tmp_names[(tmp_idx)] ? tmp_names[(tmp_idx)] : "NULL")
static int   tempFileMakeTemplate(char *template, size_t len);
static rwIOStruct *tempFileCreate(int tmp_idx);
static rwIOStruct *tempFileOpen(int tmp_idx);
static void tempFileRemove(int tmp_idx);
static int  dumpRecordsToTemp(int, const rwRec *, uint32_t);
static int  writeRecords(rwIOStruct *, const rwRec *, uint32_t);
static int  mergeFiles(rwIOStruct *, int, int);
static void rwqsort(char *a, size_t n, size_t es, int (*cmp)());


/* LOCAL VARIABLES */

/* available key fields; rwAsciiFieldMapAddDefaultFields() fills this */
static sk_stringmap_t *field_map = NULL;

/* number of fields to sort over; rwAsciiFieldMapParseFields() sets this */
static uint32_t num_fields = 0;

/* IDs of the fields to sort over; rwAsciiFieldMapParseFields() sets
 * it; values are from the rwrec_printable_fields_t enum. */
static uint32_t *sort_fields = NULL;

/* input checker */
static iochecksInfoStruct *ioISP;

/* name of output stream */
static char *output_path;

/* output stream */
static rwIOStruct *out_rwios;

/* temporary directory */
static char *g_user_temp_dir;

/* names of temporary files */
static char *tmp_names[MAX_TMP_FILES];


/* OPTIONS */

typedef enum {
    RWSO_OPT_FIELDS,
    RWSO_OPT_OUTPUT_PATH,
    RWSO_OPT_INPUT_PIPE,
    RWSO_OPT_TEMP_DIRECTORY
} appOptionsEnum;

static struct option appOptions[] = {
    {"fields",         REQUIRED_ARG, 0, RWSO_OPT_FIELDS},
    {"output-path",    REQUIRED_ARG, 0, RWSO_OPT_OUTPUT_PATH},
    {"input-pipe",     REQUIRED_ARG, 0, RWSO_OPT_INPUT_PIPE},
    {"temp-directory", REQUIRED_ARG, 0, RWSO_OPT_TEMP_DIRECTORY},
    {0,0,0,0}  /* sentinel entry */
};

static const char *appHelp[] = {
    NULL, /* generated dynamically */
    ("destination for sorted output (stdout|pipe).\n"
     "\tDefault is stdout if stdout is not a terminal"),
    ("get input byte stream from pipe (stdin|pipe).\n"
     "\tDefault is stdin if no filenames are given on the command line"),
    "temporary directory to use while sorting",
    (char *)NULL
};



/*
 * appUsage:
 *      print usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
void appUsage(void)
{
    fprintf(stderr, "Use `%s --help for usage\n", skAppName());
    appExit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 *      print usage information to USAGE_FH and exit with code 1.
 *      passed to optionsSetup() to print usage when --help option
 *      given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                             \
    ("%s --fields=<fields> [options] [files]\n"                               \
     "\tRead SiLK Flow records, sort them by the specified field(s), and\n"   \
     "\twrite the records to the named output path or to the standard\n"      \
     "\toutput.  When no files are given on command line, flows are read\n"   \
     "\tfrom the standard input\n.")

    fprintf(USAGE_FH, USAGE_MSG, skAppName());
    appOptionsUsage(USAGE_FH);
    fprintf(USAGE_FH, "NOTES:\n\t* The --fields switch is required.\n");
    appExit(EXIT_SUCCESS);
}


/*
 * appOptionsUsage:
 *      print options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *      do NOT add non-option usage here (i.e., required and other
 *      optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\noptions:\n");
    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. ", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]));
        switch (i) {
          case RWSO_OPT_FIELDS:
            /* Dynamically build the help */
            fprintf(fh, "Field(s) to use as key:\n");
            rwAsciiFieldMapPrintUsage(fh, field_map);
            break;
          default:
            /* Simple help text from the appHelp array */
            assert(appHelp[i]);
            fprintf(fh, "%s\n", appHelp[i]);
            break;
        }
    }

    return;
}


/*
 * appTeardown:
 *      teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 *      All modules are torn down. Then application teardown takes place.
 *      Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;
    int i;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    iochecksTeardown(ioISP);
    optionsTeardown();

    if (out_rwios) {
        if (rwCloseFile(out_rwios)) {
            skAppPrintErr("Exiting at %s:%d", __FILE__, __LINE__);
            appExit(EXIT_FAILURE);
        }
        out_rwios = NULL;
    }

    for (i = 0; i < MAX_TMP_FILES; ++i) {
        if (tmp_names[i]) {
            tempFileRemove(i);
        }
    }

    if (sort_fields != NULL) {
        free(sort_fields);
    }

    return;
}


/*
 *  appExit(status)
 *
 *  Exit the application with the given status.
 */
static void appExit(int status)
{
    appTeardown();
    exit(status);
}


/*
 *  appHandleSignal(signal_value)
 *
 *    Call appExit() to exit the program.  If signal_value is SIGPIPE,
 *    close cleanly; otherwise print a message that we've caught the
 *    signal and exit with EXIT_FAILURE.
 */
static void appHandleSignal(int sig)
{
    if (sig == SIGPIPE) {
        /* we get SIGPIPE if something downstream, like rwcut, exits
         * early, so don't bother to print a warning, and exit
         * successfully */
        appExit(EXIT_SUCCESS);
    } else {
        skAppPrintErr("caught signal..cleaning up and exiting");
        appExit(EXIT_FAILURE);
    }
}


/*
 *  appSetSignalHandler(sig_fn)
 *
 *    Call sig_fn to clean up any temporary files whenever we get a
 *    signal.
 */
static void appSetSignalHandler(sig_fn_t sig_fn)
{
    /* list of handled signals */
    int sigs[] = {SIGINT, SIGPIPE, SIGQUIT, SIGTERM};
    struct sigaction act;
    size_t i;

    act.sa_handler = sig_fn;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT;
#endif

    for (i = 0; i < (sizeof(sigs)/sizeof(int)); ++i) {
        if (sigaction(sigs[i], &act, NULL) < 0) {
            skAppPrintErr("Cannot register handler for signal #%d",
                          sigs[i]);
            appExit(EXIT_FAILURE);
        }
    }
}


/*
 * appSetup
 *      do all the setup for this application include setting up required
 *      modules etc.
 * Arguments:
 *      argc, argv
 * Returns: None.
 * Side Effects:
 *      exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int rv = LIBRW_OK;

    skAppRegister(argv[0]);

    /* Initialize variables */
    memset(tmp_names, 0, sizeof(tmp_names));

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* setup options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        appExit(EXIT_FAILURE);                    /* never returns */
    }

    /* register the options and the hander */
    if (optionsRegister(appOptions, (optHandler) appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        appExit(EXIT_FAILURE);                    /* never returns */
    }

    /* initialize string-map of field identifiers: add default fields,
     * then remove millisec fields, since the normal time sorting will
     * account for the millisecond values as well. */
    if (rwAsciiFieldMapAddDefaultFields(&field_map)) {
        skAppPrintErr("unable to setup fields stringmap");
        exit(EXIT_FAILURE);
    }
    (void)skStringMapRemoveByID(field_map, RWREC_FIELD_STIME_MSEC);
    (void)skStringMapRemoveByID(field_map, RWREC_FIELD_ETIME_MSEC);
    (void)skStringMapRemoveByID(field_map, RWREC_FIELD_ELAPSED_MSEC);

    /* allow for 0 input and 0 output pipes */
    ioISP = iochecksSetup(0, 0, argc, argv);

    /* parse options */
    if ( (ioISP->firstFile = optionsParse(argc, argv)) < 0) {
        appUsage();/* never returns */
    }

    /* Make sure the user specified at least one field */
    if (num_fields == 0) {
        skAppPrintErr("The sorting key (--%s switch) was not given",
                      appOptions[RWSO_OPT_FIELDS].name);
        appUsage();         /* never returns */
    }

    /* Use STDIN as an input stream if it is not a TTY; make certain
     * we have some input and we are either reading from STDIN or
     * using files listed the command line, but not both. */
    if (iochecksAcceptFromStdin(ioISP) || iochecksInputs(ioISP, 0)) {
        appUsage();
    }

    /* Check for an output stream; or default to stdout  */
    if (NULL == output_path) {
        output_path = "stdout";
    }

    /* open output */
    if ((rv = rwioCreate(&out_rwios, output_path, SK_RWIO_WRITE))
        || (rv = rwioOpen(out_rwios)))
    {
        rwioPrintLastErr(out_rwios, rv, NULL);
        skAppPrintErr("Couldn't open output file.  Exiting.");
        appExit(EXIT_FAILURE);
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        appExit(EXIT_FAILURE);
    }

    /* set signal handler to clean up temp files on SIGINT, SIGTERM, etc */
    appSetSignalHandler(&appHandleSignal);

    (void) umask((mode_t) 0002);
    return;                       /* OK */
}


/*
 * appOptionsHandler:
 *      called for each option the app has registered.
 * Arguments:
 *      clientData cData: ignored here.
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    switch ((appOptionsEnum)opt_index) {
      case RWSO_OPT_FIELDS:
        assert(opt_arg);
        if (parseFields(opt_arg)) {
            return 1;
        }
        break;

      case RWSO_OPT_OUTPUT_PATH:
        /* output file path: treat "stdout" as standard output---which
         * is the default; anything else as a file path */
        assert(opt_arg);
        if (0 != strcmp(opt_arg, "stdout")) {
            /* check for switch given multiple times */
            if (NULL != output_path) {
                skAppPrintErr("multiple output files: '%s' and '%s'",
                              output_path, opt_arg);
                return 1;
            }
            output_path = opt_arg;
        }
        break;

      case RWSO_OPT_INPUT_PIPE:
        assert(opt_arg);
        if (iochecksInputSource(ioISP, opt_arg)) {
            return 1;
        }
        break;

      case RWSO_OPT_TEMP_DIRECTORY:
        assert(opt_arg);
        g_user_temp_dir = opt_arg;
        break;
    }

    return 0;                     /* OK */
}


/*
 *  status = parseFields(fields_string);
 *
 *    Parse the user's option for the --fields switch and fill in the
 *    global g_fields[].  Return 0 on success; 1 on failure.
 */
static int parseFields(const char *field_string)
{
    /* have we been here before? */
    if (num_fields > 0) {
        skAppPrintErr("--%s switch provided multiple times",
                      appOptions[RWSO_OPT_FIELDS].name);
        return 1;
    }

    /* verify user gave us input */
    if (field_string == NULL || field_string[0] == '\0') {
        skAppPrintErr("missing a value for the --%s switch",
                      appOptions[RWSO_OPT_FIELDS].name);
        return 1;
    }

    /* parse the input */
    if (rwAsciiFieldMapParseFields(&sort_fields, &num_fields, field_string,
                                   field_map, RW_ASCII_DUPES_ERROR))
    {
        /* library printed the error */
        return 1;
    }
    assert(num_fields < RWREC_PRINTABLE_FIELD_COUNT);

    return 0;
}


/* Define our raw sorting functions */
#define RETURN_IF_SORTED(a, b)                  \
    if ((a) < (b)) { return -1; }               \
    else { if ((a) > (b)) { return 1; } }

/*
 *  recsort(a, b);
 *
 *     Returns an ordering on the recs pointed to `a' and `b' by
 *     comparing the fields listed in the sort_fields[] array.
 */
static int recsort(const rwRec *a, const rwRec *b)
{
    uint32_t i;

    for (i = 0; i < num_fields; i++) {
        switch ((rwrec_printable_fields_t)sort_fields[i]) {
          case RWREC_FIELD_SIP:
            RETURN_IF_SORTED(a->sIP.ipnum, b->sIP.ipnum);
            break;

          case RWREC_FIELD_DIP:
            RETURN_IF_SORTED(a->dIP.ipnum, b->dIP.ipnum);
            break;

          case RWREC_FIELD_SPORT:
            RETURN_IF_SORTED(a->sPort, b->sPort);
            break;

          case RWREC_FIELD_DPORT:
            RETURN_IF_SORTED(a->dPort, b->dPort);
            break;

          case RWREC_FIELD_PROTO:
            RETURN_IF_SORTED(a->proto, b->proto);
            break;

          case RWREC_FIELD_PKTS:
            RETURN_IF_SORTED(a->pkts, b->pkts);
            break;

          case RWREC_FIELD_BYTES:
            RETURN_IF_SORTED(a->bytes, b->bytes);
            break;

          case RWREC_FIELD_FLAGS:
            RETURN_IF_SORTED(a->flags, b->flags);
            break;

          case RWREC_FIELD_STIME:
          case RWREC_FIELD_STIME_MSEC:
            RETURN_IF_SORTED(a->sTime, b->sTime);
            RETURN_IF_SORTED(a->sTime_msec, b->sTime_msec);
            break;

          case RWREC_FIELD_ELAPSED:
          case RWREC_FIELD_ELAPSED_MSEC:
            RETURN_IF_SORTED(a->elapsed, b->elapsed);
            RETURN_IF_SORTED(a->elapsed_msec, b->elapsed_msec);
            break;

          case RWREC_FIELD_ETIME:
          case RWREC_FIELD_ETIME_MSEC:
            {
                uint32_t a_msec;
                uint32_t b_msec;
                uint32_t a_carry = 0;
                uint32_t b_carry = 0;

                a_msec = a->sTime_msec + a->elapsed_msec;
                if (a_msec > 1000) {
                    a_carry = 1;
                    a_msec -= 1000;
                }
                b_msec = b->sTime_msec + b->elapsed_msec;
                if (b_msec > 1000) {
                    b_carry = 1;
                    b_msec -= 1000;
                }

                RETURN_IF_SORTED((a->sTime + a->elapsed + a_carry),
                                 (b->sTime + b->elapsed + b_carry));
                RETURN_IF_SORTED(a_msec, b_msec);
            }
            break;

          case RWREC_FIELD_SID:
            RETURN_IF_SORTED(a->sID, b->sID);
            break;

          case RWREC_FIELD_INPUT:
            RETURN_IF_SORTED(a->input, b->input);
            break;

          case RWREC_FIELD_OUTPUT:
            RETURN_IF_SORTED(a->output, b->output);
            break;

          case RWREC_FIELD_NHIP:
            RETURN_IF_SORTED(a->nhIP.ipnum, b->nhIP.ipnum);
            break;

          case RWREC_FIELD_INIT_FLAGS:
            RETURN_IF_SORTED(a->init_flags, b->init_flags);
            break;

          case RWREC_FIELD_REST_FLAGS:
            RETURN_IF_SORTED(a->rest_flags, b->rest_flags);
            break;

          case RWREC_FIELD_TCP_STATE:
            RETURN_IF_SORTED(a->tcp_state, b->tcp_state);
            break;

          case RWREC_FIELD_APPLICATION:
            RETURN_IF_SORTED(a->application, b->application);
            break;

          case RWREC_FIELD_FTYPE_CLASS:
          case RWREC_FIELD_FTYPE_TYPE:
            RETURN_IF_SORTED(a->flow_type, b->flow_type);
            break;
        }
    }

    return 0;
}


/*
 *  ok = tempFileMakeTemplate(template, len);
 *
 *    Fill in the C string 'template' of size 'len' with the temp-file
 *    template.  Attempts to find a temporary directory, either from
 *    the user's input, from the TMPDIR environment variable, or from
 *    the DEFAULT_TEMP_DIR macro.  Also uses the application name.
 *    'template' is suitable for passing to mkstemp().  Returns 0 on
 *    success, or -1 if no directory could be found.
 */
static int tempFileMakeTemplate(char *template, size_t len)
{
    char *tmp_dir;
    int rv;

    /* First, figure out the directory to use.  Use the user's option
     * if given, or TMPDIR, or the default */
    if (g_user_temp_dir) {
        tmp_dir = g_user_temp_dir;
    } else if (NULL == (tmp_dir = getenv(TMPDIR_ENVAR))) {
        tmp_dir = DEFAULT_TEMP_DIR;
    }
    if ( !dirExists(tmp_dir)) {
        skAppPrintErr("Temporary directory %s does not exist", tmp_dir);
        return -1;
    }

    rv = snprintf(template, len, "%s/%s_tmp.XXXXXXXX", tmp_dir, skAppName());
    if ((size_t)rv >= len || rv < 0) {
        skAppPrintErr("snprintf error");
        return -1;
    }

    return 0;
}


/*
 *  rwIOS = tempFileCreate(tmp_idx);
 *
 *    Create and open a new temporary file indexed by 'tmp_idx.'  The
 *    name of the file can be found by tempFileGetName(), and the file
 *    can be re-opened with the tempFileOpen() function.  Returns a
 *    rwIOStruct* to the new file, or NULL on error: 'tmp_idx' out of
 *    range, mkstemp() fails.
 *
 */
static rwIOStruct *tempFileCreate(int tmp_idx)
{
    static char temp_file_template[PATH_MAX] = "";
    rwIOStruct *rwIOS;
    int rv = LIBRW_OK;

    /* check that tmp_idx is in range */
    if (tmp_idx < 0 || tmp_idx >= MAX_TMP_FILES) {
        skAppPrintErr("tempFileCreate index %d out of range 0 < n < %d",
                      tmp_idx, MAX_TMP_FILES);
        return NULL;
    }

    /* initialize the template from the tmpdir if we haven't */
    if (temp_file_template[0] == '\0') {
        if (tempFileMakeTemplate(temp_file_template, PATH_MAX)) {
            return NULL;
        }
    }

    if ((rv = rwioCreate(&rwIOS, temp_file_template, SK_RWIO_WRITE))
        || (rv = rwioMakeTemp(rwIOS)))
    {
        return NULL;
    }

    /* copy file name for later use */
    tmp_names[tmp_idx] = strdup(rwGetFileName(rwIOS));
    if (NULL == tmp_names[tmp_idx]) {
        skAppPrintErr("out of memory!");
        return NULL;
    }
    PRINTDEBUG((stderr, "new tmp %d => '%s'\n", tmp_idx, tmp_names[tmp_idx]));

    return rwIOS;
}


/*
 *  rwIOS = tempFileOpen(tmp_idx);
 *
 *    Returns a rwIOStruct* pointing to a unique temporary filename
 *    associated with 'tmp_idx' that has been previously created with
 *    tempFileCreate().  Returns NULL if file cannot be opened.
 */
static rwIOStruct *tempFileOpen(int tmp_idx)
{
    char *fn;

    assert (0 <= tmp_idx && tmp_idx < MAX_TMP_FILES);
    assert (tmp_names[tmp_idx]);

    fn = tempFileGetName(tmp_idx);
    PRINTDEBUG((stderr, "existing tmp %d => '%s'\n", tmp_idx, fn));

    return rwOpenFile(fn, NULL);
}


/*
 *  tempFileRemove(tmp_idx);
 *
 *    Remove the temporary file indexed by 'tmp_idx'.
 */
static void tempFileRemove(int tmp_idx)
{
    char *fn;

    assert (0 <= tmp_idx && tmp_idx < MAX_TMP_FILES);
    assert (tmp_names[tmp_idx]);

    fn = tempFileGetName(tmp_idx);

    PRINTDEBUG((stderr, "Removing file %s of size %lu\n",
                fn, (unsigned long)fileSize(fn)));
    if (-1 == remove(fn)) {
        skAppPrintErr("Cannot remove file '%s': %s",
                      fn, strerror(errno));
    }

    tmp_names[tmp_idx][0] = '\0';
    free(tmp_names[tmp_idx]);
    tmp_names[tmp_idx] = NULL;
}


/*
 *  ok = writeRecords(rwIOS, buffer, count);
 *
 *    Write 'count' records from 'buffer' to 'rwIOS'.  Return 0 if
 *    o.k.; non-zero otherwise.
 */
static int writeRecords(
    rwIOStruct     *rwIOS,
    const rwRec    *record_buffer,
    uint32_t        record_count)
{
    int rv = 0; /* return value */
    uint32_t c;
    const rwRec *rwrec;
    
    PRINTDEBUG((stderr, "Writing records to '%s'\n", rwIOS->fPath));
    for (c = 0, rwrec = record_buffer; c < record_count; c++, rwrec++) {
        if (0 != (rv = rwWrite(rwIOS, rwrec))) {
            break;
        }
    }

    return rv;
}


/*
 *  ok = dumpRecordsToTemp(tmp_idx, buffer, count);
 *
 *    Print 'count' records from 'buffer' into a newly created
 *    temporary file.  Index the temp file by 'tmp_idx'.  Returns 0 on
 *    success, or -1 if a temp file cannot be created, or the write or
 *    close calls fail.
 */
static int dumpRecordsToTemp(
    int             tmp_idx,
    const rwRec    *record_buffer,
    uint32_t        record_count)
{
    rwIOStruct *temp_file;
    int rv = 0; /* return value */

    temp_file = tempFileCreate(tmp_idx);
    if (!temp_file) {
        return -1;
    }

    rv = writeRecords(temp_file, record_buffer, record_count);

    if (rwCloseFile(temp_file)) {
        skAppPrintErr("Error closing temp file %s: %s",
                      tempFileGetName(tmp_idx), strerror(errno));
        if (rv == 0) {
            rv = -1;
        }
    }

    return rv;
}


/*
 * Merge two files into one, maintaining sorted order
 */
static int mergeFiles(
    rwIOStruct *rwIOS,
    int         tmp_idx_1,
    int         tmp_idx_2)
{
    int rv = 0; /* return value */
    rwIOStruct *rwios1 = NULL;
    rwIOStruct *rwios2 = NULL;
    rwRec r1;
    rwRec r2;
    int f1_has_data;
    int f2_has_data;

    PRINTDEBUG((stderr, "merging %d and %d to %s\n",
                tmp_idx_1, tmp_idx_2, rwIOS->fPath));

    if (NULL == (rwios1 = tempFileOpen(tmp_idx_1))) {
        rv = -1;
        goto END;
    }
    if (NULL == (rwios2 = tempFileOpen(tmp_idx_2))) {
        rv = -1;
        goto END;
    }

    f1_has_data = rwRead(rwios1, &r1);
    f2_has_data = rwRead(rwios2, &r2);
    while (f1_has_data && f2_has_data) {
        if (recsort( &r1, &r2) <= 0) {
            if (0 != rwWrite(rwIOS, &r1)) {
                rv = -1;
                goto END;
            }
            f1_has_data = rwRead(rwios1, &r1);
        } else {
            if (0 != rwWrite(rwIOS, &r2)) {
                rv = -1;
                goto END;
            }
            f2_has_data = rwRead(rwios2, &r2);
        }
    }
    if (f1_has_data) {
        do {
            if (0 != rwWrite(rwIOS, &r1)) {
                rv = -1;
                goto END;
            }
        } while (rwRead(rwios1, &r1));
    } else {
        assert(f2_has_data);
        do {
            if (0 != rwWrite(rwIOS, &r2)) {
                rv = -1;
                goto END;
            }
        } while (rwRead(rwios2, &r2));
    }

  END:
    tempFileRemove(tmp_idx_1);
    tempFileRemove(tmp_idx_2);
    if (rwios1) {
        rwCloseFile(rwios1);
    }
    if (rwios2) {
        rwCloseFile(rwios2);
    }

    return rv;
}


/*
 *  rwIOS = nextFile();
 *
 *    Return a IO Structure to the next input file to read.  Return
 *    NULL if no more input files.
 */
static rwIOStruct *nextFile(void)
{
    static int counter = -1;
    rwIOStruct *rwios = NULL;

    if (counter < 0) {
        counter = ioISP->firstFile;
    }

    /* Open next file; keep trying until good one found. */
    for (; (rwios == NULL && counter < ioISP->fileCount); counter++) {
        rwios = rwOpenFile(ioISP->fnArray[counter], ioISP->inputCopyFD);
        if (rwios == NULL) {
            skAppPrintErr("Ignoring input file '%s': Could not open",
                          ioISP->fnArray[counter]);
        }
    }

    return rwios;
}


int main(int argc, char **argv)
{
    int temp_file_no = 0;
    int merge_file_no = 0;
    rwIOStruct *input_rwios;    /* input stream */
    rwIOStruct *temp_file;
    void *record_buffer = NULL; /* Region of memory for records */
    rwRec *cur_rec = NULL;      /* Ptr into record_buffer */
    uint32_t buffer_max_recs;   /* buffer size (in number of records) */
    uint32_t record_count = 0;  /* Number of records read */

    appSetup(argc, argv);                 /* never returns on error */

    /* Allocate the buffer: We'll try to allocate a really BIG
     * chunk...if that fails, halve the chunk until we succeed. */
    buffer_max_recs = MAX_IN_CORE_RECORDS;
    while (1) {
        PRINTDEBUG((stderr, "buffer_max_recs = %d\n", buffer_max_recs));
        record_buffer = malloc(sizeof(rwRec) * buffer_max_recs);
        if (record_buffer) {
            /* malloc was successful */
            break;
        } else if (buffer_max_recs < MIN_IN_CORE_RECORDS) {
            /* give up at this point */
            skAppPrintErr("Could not malloc %lu bytes for record_buffer",
                          (sizeof(rwRec) * buffer_max_recs));
            appExit(EXIT_FAILURE);
        } else {
            /* halve it and try again */
            buffer_max_recs /= 2;
        }
    }

    cur_rec = (rwRec *)record_buffer;
    record_count = 0;

    /* open first file */
    input_rwios = nextFile();

    while (input_rwios != NULL) {
        /* read record */
        if (rwRead(input_rwios, cur_rec) == 0) {
            /* end of file: close current and open next */
            rwCloseFile(input_rwios);
            input_rwios = nextFile();
            continue;
        }

        cur_rec++;
        record_count++;

        if (record_count == buffer_max_recs) {
            /* Sort */
            rwqsort(record_buffer, record_count, sizeof(rwRec), recsort);

            /* Write to temp file */
            if(dumpRecordsToTemp( temp_file_no++,record_buffer,record_count)){
                appExit(EXIT_FAILURE);
            }

            /* Reset record buffer to 'empty' */
            cur_rec = (rwRec *)record_buffer;
            record_count = 0;
        }
    }

    /* if no records were read; exit the program */
    if (record_count == 0) {
        skAppPrintErr("No input records");
        free(record_buffer);
        return 0;
    }

    /* Sort last batch of records */
    rwqsort(record_buffer, record_count, sizeof(rwRec), recsort);

    if (temp_file_no == 0) {
        /* No temp files written, just output batch of records */
        if (writeRecords( out_rwios, record_buffer, record_count)) {
            appExit(EXIT_FAILURE);
        }
        free(record_buffer);
    } else {
        /* Write last batch to temp file */
        if (dumpRecordsToTemp( temp_file_no++, record_buffer, record_count)) {
            appExit(EXIT_FAILURE);
        }
        free(record_buffer);

        /* Merge all temp files, until two remain */
        while (merge_file_no + 2 < temp_file_no) {
            temp_file = tempFileCreate( temp_file_no++);
            if (!temp_file) {
                appExit(EXIT_FAILURE);
            }

            if (mergeFiles( temp_file, merge_file_no, merge_file_no+1)) {
                appExit(EXIT_FAILURE);
            }
            if (rwCloseFile(temp_file)) {
                skAppPrintErr("Exiting at %s:%d", __FILE__, __LINE__);
                appExit(EXIT_FAILURE);
            }

            merge_file_no += 2;
        }

        /* Merge last two temp files, pipe result to outF */
        if (mergeFiles( out_rwios, merge_file_no, merge_file_no+1)) {
            appExit(EXIT_FAILURE);
        }
    }

    if (rwCloseFile(out_rwios)) {
        skAppPrintErr("Exiting at %s:%d", __FILE__, __LINE__);
        appExit(EXIT_FAILURE);
    }
    out_rwios = NULL;

    appTeardown();
    return (0);
}


/*
  qsort()

  A la Bentley and McIlroy in Software - Practice and Experience,
  Vol. 23 (11) 1249-1265. Nov. 1993


*/

typedef long WORD;
#define W               sizeof(WORD)    /* must be a power of 2 */
#define SWAPINIT(a, es) swaptype = \
                        (( (a-(char *)0) | es) % W) ? 2 : (es > W) ? 1 : 0
#define exch(a, b, t)   (t = a, a = b, b = t)
#define swap(a, b)      swaptype != 0 ? swapfunc(a, b, es, swaptype) : \
                        (void )exch(*(WORD*)(a), *(WORD*)(b), t)
/* swap sequences of records */
#define vecswap(a, b, n)        if ( n > 0) swapfunc(a, b, n, swaptype)

#define PVINIT(pv, pm)  if (swaptype != 0) pv = a, swap(pv, pm); \
                        else pv = (char *)&v, v = *(WORD*)pm

#define min(a, b)       a <= b ? a : b

static char *med3(char *a, char *b, char *c, int (*cmp)())
{
    return(   cmp(a, b) < 0 ?
              (cmp(b, c) < 0 ? b : cmp(a, c) < 0 ? c : a)
              : (cmp(b, c) > 0 ? b : cmp(a, c) > 0 ? c : a)
              );
}


static void swapfunc(char *a, char *b, size_t n, int swaptype)
{
    if( swaptype <= 1)
        {
            WORD t;
            for( ; n > 0; a += W, b += W, n -= W)
                exch(*(WORD*)a, *(WORD*)b, t);
        }
    else
        {
            char t;
            for( ; n > 0; a += 1, b += 1, n -= 1)
                exch(*a, *b, t);
        }
}


static void rwqsort(char *a, size_t n, size_t es, int (*cmp)())
{
    char *pa, *pb, *pc, *pd, *pl, *pm, *pn, *pv;
    int r, swaptype;
    WORD t, v;
    size_t s;

    SWAPINIT(a, es);
    if(n < 7)
        {
            /* use insertion sort on smallest arrays */
            for (pm = a + es; pm < a + n*es; pm += es)
                for(pl = pm; pl > a && cmp(pl-es, pl) > 0; pl -= es)
                    swap(pl, pl-es);
            return;
        }

    pm = a + (n/2)*es;            /* small arrays middle element */
    if ( n > 7)                   /* SLK What about n == 7 ? */
        {
            pl = a;
            pn = a + (n - 1)*es;
            if( n > 40 )
                {
                    /* big arays.  Pseudomedian of 9 */
                    s = (n / 8) * es;
                    pl = med3(pl, pl + s, pl + 2 * s, cmp);
                    pm = med3(pm - s, pm, pm + s, cmp);
                    pn = med3(pn - 2 * s, pn - s, pn, cmp);
                }
            pm = med3(pl, pm, pn, cmp); /* mid-size, med of 3 */
        }
    PVINIT(pv, pm);               /* pv points to the partition value */
    pa = pb = a;
    pc = pd = a + (n - 1) * es;
    for(;;)
        {
            while(pb <= pc && (r = cmp(pb, pv)) <= 0)
                {
                    if (r == 0)
                        {
                            swap(pa, pb);
                            pa += es;
                        }
                    pb += es;
                }
            while (pc >= pb && (r = cmp(pc, pv)) >= 0)
                {
                    if( r == 0)
                        {
                            swap(pc, pd);
                            pd -= es;
                        }
                    pc -= es;
                }
            if( pb > pc)
                break;
            swap(pb, pc);
            pb += es;
            pc -= es;
        }
    pn = a + n * es;
    s = min(pa - a, pb - pa);
    vecswap(a, pb - s, s);
    s = min(pd - pc, pn -pd -(int32_t)es);
    vecswap(pb, pn - s, s);
    if (( s = pb - pa) > es) {
        rwqsort(a, s/es, es, cmp);
    }
    if (( s = pd - pc) > es) {
        rwqsort(pn - s, s/es, es, cmp);
    }

    return;
}


