/*
** Copyright (C) 2003-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
 * rwswapbytes
 *
 * Read any rw file (rwpacked file, rwfilter output, etc) and output a
 * file in the specified byte order.
 *
 */

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwswapbytes/rwswapbytes.c,v 1.11 2005/09/17 19:00:05 thomasm Exp $");

#include "rwpack.h"
#include "utils.h"
#include "rwswapbytes.h"


/* MACROS AND TYPEDEFS */

/* File handle for --help output */
#define USAGE_FH stdout

typedef enum _rwswapOptions {
    RWSW_BIG = 128, RWSW_LITTLE, RWSW_NATIVE, RWSW_SWAP,
    RWSW_COPY
} rwswapOptions_t;



/* LOCAL FUNCTIONS */

static void appUsage(void);			/* never returns */
static void appUsageLong(void);			/* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);	/* never returns on failure */
static void appOptionsUsage(FILE *); 		/* application options usage */
static int  appOptionsHandler(clientData, int, const char *);

/* LOCAL VARIABLES */

static rwswapOptions_t outEndian;
static const char *inFName;
static const char *outFName;


/* OPTIONS */

static struct option appOptions[] = {
    {"big-endian",       NO_ARG,       0, RWSW_BIG},
    {"little-endian",    NO_ARG,       0, RWSW_LITTLE},
    {"native-endian",    NO_ARG,       0, RWSW_NATIVE},
    {"swap-endian",      NO_ARG,       0, RWSW_SWAP},
    {0,0,0,0}                     /* sentinal entry */
};

static char *appHelp[] = {
    /* add help strings here for the applications options */
    "write output in big-endian format",
    "write output in little-endian format",
    "write output in this machine's native format",
    "unconditionally swap the byte-order of the input",
    (char *)NULL
};



/*
 * appUsage:
 * 	print message on getting usage information and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 * 	print usage information to stderr and exit with code 1.
 * 	passed to optionsSetup() to print usage when --help option
 * 	given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
    fprintf(USAGE_FH,
            ("%s <endian-switch> <input-file> <output-file>\n"
             "\tChange the endian-ness of <input-file> as specified\n"
             "\tby <endian-switch> and write result to <output-file>.\n"
             "\tYou may use \"stdin\" for <input-file> and \"stdout\"\n"
             "\tfor <output-file>.  Gzipped files are o.k.\n"),
            skAppName());
    appOptionsUsage(USAGE_FH);
    exit(EXIT_SUCCESS);
}


/*
 * appOptionsUsage:
 * 	print options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *	do NOT add non-option usage here (i.e., required and other
 *	optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\nendian-switch:\n");
    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s: %s\n", appOptions[i].name, appHelp[i]);
    }

    return;
}


/*
 * appTeardown:
 *	teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 * 	All modules are torn down. Then application teardown takes place.
 * 	Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    optionsTeardown();

    return;
}


/*
 * appSetup
 *	do all the setup for this application include setting up required
 *	modules etc.
 * Arguments:
 *	argc, argv
 * Returns: None.
 * Side Effects:
 *	exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int nextOptIndex;

    skAppRegister(argv[0]);

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* initialize the options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);			/* never returns */
    }

    /* register the options and handler */
    if (optionsRegister(appOptions, (optHandler) appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);			/* never returns */
    }

    /* parse options */
    if ((nextOptIndex = optionsParse(argc, argv)) < 0) {
        appUsage();		/* never returns */
    }

    /* Check that a swapping option was given */
    if (0 == outEndian) {
        skAppPrintErr("You must specify the output byte order.");
        appUsage();
    }

    /* Check that we have input */
    if (nextOptIndex >= argc) {
        skAppPrintErr("Expecting input file name");
        appUsage();		/* never returns */
    }
    inFName = argv[nextOptIndex];
    ++nextOptIndex;

    /* check that we have an output location */
    if (nextOptIndex >= argc) {
        skAppPrintErr("Expecting output file name");
        appUsage();		/* never returns */
    }
    outFName = argv[nextOptIndex];
    ++nextOptIndex;

    /* check for extra options */
    if (argc != nextOptIndex) {
        skAppPrintErr("Got extra options");
        appUsage();		/* never returns */
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);		/* never returns */
    }

    return;			/* OK */
}


/*
 * appOptionsHandler:
 * 	called for each option the app has registered.
 * Arguments:
 *	clientData cData: ignored here.
 *	int index: index into appOptions of the specific option
 *	char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *	0 if OK. 1 else.
 * Side Effects:
 *	Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    const char  UNUSED(*opt_arg))
{
    switch (opt_index) {
#if !IS_LITTLE_ENDIAN
      case RWSW_NATIVE:
#endif
      case RWSW_BIG:
        if ((0 != outEndian) && (RWSW_BIG != outEndian)) {
            skAppPrintErr("conflicting endian options given");
            return 1;
        }
        outEndian = RWSW_BIG;
        break;

#if IS_LITTLE_ENDIAN
      case RWSW_NATIVE:
#endif
      case RWSW_LITTLE:
        if ((0 != outEndian) && (RWSW_LITTLE != outEndian)) {
            skAppPrintErr("conflicting endian options given");
            return 1;
        }
        outEndian = RWSW_LITTLE;
        break;

      case RWSW_SWAP:
        if ((0 != outEndian) && (RWSW_SWAP != outEndian)) {
            skAppPrintErr("conflicting endian options given");
            return 1;
        }
        outEndian = RWSW_SWAP;
        break;

      default:
        skAppPrintErr("Unhandled option %d in options handler at %s:%d",
                      opt_index, __FILE__, __LINE__);
        abort();
    }

    return 0;			/* OK */
}


/*
 *  is_ok = rwswap_file(input_file, output_file);
 *
 *  Byte swap the file named 'input_file' and write it to
 *  'output_file' using the 'out_endian' to determine the byte-order
 *  of 'output_file'.
 */
static int rwswap_file(
    const char     *input_file,
    const char     *output_file,
    rwswapOptions_t out_endian)
{
    int32_t file_type;
    silk_endian_t byte_order;
    rwIOStructPtr input_ios;
    rwIOStructPtr output_ios;
    rwRec rwrec;
    int rv = 0;

    input_ios = rwOpenFile(input_file, NULL);
    if (!input_ios) {
        return 1;
    }
    file_type = rwGetFileType(input_ios);

    switch (out_endian) {
      case RWSW_BIG:
        byte_order = SILK_ENDIAN_BIG;
        break;
      case RWSW_LITTLE:
        byte_order = SILK_ENDIAN_LITTLE;
        break;
      case RWSW_NATIVE:
        byte_order = SILK_ENDIAN_NATIVE;
        break;
      case RWSW_SWAP:
        if (rwGetIsBigEndian(input_ios)) {
            byte_order = SILK_ENDIAN_LITTLE;
        } else {
            byte_order = SILK_ENDIAN_BIG;
        }
        break;
      default:
        skAppPrintErr("Case %d not handled in byte-order handler at %s:%d",
                      out_endian, __FILE__, __LINE__);
        abort();
    }

    if ((rv = rwioCreate(&output_ios, output_file, SK_RWIO_WRITE))
        || (rv = rwioSetFileType(output_ios, rwGetFileType(input_ios)))
        || (rv = rwioSetFileVersion(output_ios, rwGetFileVersion(input_ios)))
        || (rv = rwioSetFileByteorder(output_ios, byte_order))
        || (rv = rwioOpen(output_ios)))
    {
        goto END;
    }

    rv = rwioSetFileSTime(output_ios, rwGetFileSTime(input_ios), 0);
    if ((rv != LIBRW_OK) && (rv != LIBRW_ERR_UNSUPPORT_TYPE)) {
        goto END;
    }

    rv = rwioAppendToHistoryFromFile(output_ios, input_ios);
    if ((rv != LIBRW_OK) && (rv != LIBRW_ERR_UNSUPPORT_TYPE)) {
        goto END;
    }
    rv = LIBRW_OK;

    while ((rv == LIBRW_OK) && rwRead(input_ios, &rwrec)) {
        rv = rwWrite(output_ios, &rwrec);
    }

  END:
    if (rv) {
        rwioPrintLastErr(output_ios, rv, &skAppPrintErr);
    }
    if (output_ios) {
        rwCloseFile(output_ios);
    }
    if (input_ios) {
        rwCloseFile(input_ios);
    }

    return rv;
}


int main(int argc, char **argv)
{
    int rv;

    appSetup(argc, argv);

    rv = rwswap_file(inFName, outFName, outEndian);

    /* done */
    appTeardown();

    return rv;
}
