/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
** rwuniq.c
**
** Implementation of the rwuniq suite application. Takes a
** specification of the field(s) to count over on the command line,
** and writes an human-readable ascii representation (rwcut-style) to
** standard out.
**
** Will eventually support at least two binary formats (serialized
** hash data and a more compact record-oriented format with no "empty"
** entries).
*/


#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwuniq/rwuniq.c,v 1.32 2005/11/16 15:43:51 thomasm Exp $");

#include "utils.h"
#include "rwpack.h"
#include "iochecks.h"
#include "hashlib.h"


/* TYPEDEFS AND DEFINES */

/* #define GIVE_META_INFO */

/* For the compound key version of rwuniq, the maximum number of bytes
 * we allow in the key.  An arbitary number that should be big enough
 * for almost all cases. */
#define MAX_KEY_BYTE_WIDTH 20

/* Initializer for hash table creation */
#define HASH_INITIAL_SIZE 500000

/* Value to use as exit code when we run out of memory */
#define RWUNIQ_NO_MEMORY_EXIT_CODE 16

/* For ASCII output, maximum size of a single value column. Must be
 * big enough to hold printed uint64_t and timestamp(). */
#define VALUE_OUT_BUF_SIZE 64

/* default sTime bin size to use when --bin-time is requested */
#define DEFAULT_TIME_BIN  60
#define DEFAULT_TIME_BIN_STR "60"

/* file handle for --help usage message */
#define USAGE_FH stdout


/* LOCAL FUNCTIONS */

static void appUsage(void);
static void appUsageLong(void);                 /* never returns */
static void appOptionsUsage(FILE *);
static void appSetup(int argc, char **argv);    /* never returns */
static void appTeardown(void);
static int  appOptionsHandler(clientData, int, char *);
static void initializeKeyByteWidths(void);
static int  parseFields(const char *field_string);
static int  uniq_main(void);
#ifdef  ENABLE_BINARY
static void dumpBinary(void);
#endif
static void writeColTitles(FILE *fh);
static void dumpAscii(FILE *fh);
static void dumpRecordAscii(const rwRec *rwrec, const uint64_t val_ptr[]);

/* Functions that implement tuple counting */
static void createCompoundKey(const rwRec *rwrec, uint8_t *key);
static void decomposeCompoundKey(rwRec *rwrec, const uint8_t *key);
static int  countFileCompound(rwIOStruct*);
static int  countFileSingle(rwIOStruct*);

static int valueConfigInitialize(void);
static void valueConfigSetOffsets(void);
/* functions to manipulate the count values in the hash table */
static void packValue(
    uint8_t        *out_val_ptr,
    const rwRec    *rec_ptr);
static int unpackValues(
    uint64_t        out_value[],
    const uint8_t  *packed_val);


/* LOCAL VARIABLES */

/* hash table to count tuples */
static HashTable *g_hash_counter;

/* number of records read */
static uint64_t g_record_count = 0;

/* delimiter between output columns */
static char g_delimiter='|';

/* struct to hold the columns/fields that make up our key for this run */
static struct {
    /* the ids of the columns (from librw) that make up the key */
    uint8_t kf_fields[RWREC_PRINTABLE_FIELD_COUNT];
    /* the byte-offset for each requested field */
    size_t  kf_byte_offset[RWREC_PRINTABLE_FIELD_COUNT];
    /* total byte count in the key */
    uint8_t kf_byte_width;
    /* the number of actual columns/fields that make up the key */
    uint8_t kf_num_fields;
} key_fields;

/* byte width of each column/field. used to compute key_fields.kf_byte_width */
static size_t key_byte_width[RWREC_PRINTABLE_FIELD_COUNT];

/* available key fields */
static sk_stringmap_t *field_map = NULL;

/* name of program to run to page output */
static char *pager;

/* the floor of the sTime and/or eTime */
static uint32_t time_bin_size = 0;

/* when time-binning is active and all time fields---sTime, eTime,
 * elapased---are requested, we adjust the elapsed time so that it is
 * equal to eTime-sTime */
static int adjust_elapsed = 0;

/*
 *  Possible value fields to count w.r.t. the uniq key.  Keep this in
 *  sync with the value_col[] array.
 *
 *  In addition, the packValue() and unpackValue() functions must be
 *  modified, since they contains the mapping between the fields we
 *  are counting and their corresponding field in the rw record.
 *
 *  Since bytes is wider that the other fields (64b vs 32b), its
 *  offset must be 0 to avoid mis-alignment issues.
 */
typedef enum {
    COUNT_BYTES = 0,
    COUNT_PACKETS,
    COUNT_FLOWS,
    COUNT_STIME,
    COUNT_ETIME,
    TOTAL_COUNTS /* must be last item */
} value_types_t;

/* Information about each potential value field. */
static struct {
    /* field title for ASCII output */
    const char *vc_title;
    /* column width for ASCII output */
    int         vc_ascii_width;
    /* number of bytes required to hold this value */
    size_t      vc_byte_width;
} value_col[TOTAL_COUNTS] = {
    /* 0 */ {"Bytes",     20,  sizeof(uint64_t)},
    /* 1 */ {"Packets",   10,  sizeof(uint32_t)},
    /* 2 */ {"Records",   10,  sizeof(uint32_t)},
    /* 3 */ {"min_sTime", 19,  sizeof(uint32_t)},
    /* 4 */ {"max_eTime", 19,  sizeof(uint32_t)}
};

/* Array of structs to hold the information about the value_cols that
 * make up the value for this run */
static struct {
    /* only print sums if the minimum value is at least this value */
    uint64_t    vf_min;
    /* only print sums if the maximum value is no more than this value */
    uint64_t    vf_max;
    /* the byte-offset for each value field */
    size_t      vf_offset;
    /* whether the value is computed */
    int         vf_iscounted;
} value_fields[TOTAL_COUNTS];

/* total byte count for all value_cols in the value */
size_t value_byte_width = 0;


/* group input/output variables into a single struct */
static struct {
    FILE                 *outF;
    const char           *outFPath;
    int                   outIsPipe;
    iochecksInfoStruct   *ISP;
    rwAsciiStream_t      *rwas;
} uniq_io;

/* flags set by the user options */
static struct {
    int8_t binary_output;      /* Default to ascii */
    int8_t print_filenames;
    int8_t tabular_output;
    int8_t print_titles;
    int8_t epoch_time;
} uniq_flags = {0, 0, 1, 1, 0};


/* OPTIONS */

typedef enum {
    UNIQ_OPT_THRESHOLD,
    UNIQ_OPT_FIELDS,
    UNIQ_OPT_ALL_COUNTS,
    UNIQ_OPT_COUNT_FLOWS,
    UNIQ_OPT_COUNT_PACKETS,
    UNIQ_OPT_COUNT_BYTES,
    UNIQ_OPT_COUNT_STIME,
    UNIQ_OPT_COUNT_ETIME,
    UNIQ_OPT_BIN_TIME,
    UNIQ_OPT_NO_TITLES,
    UNIQ_OPT_NO_COLUMNS,
    UNIQ_OPT_COLUMN_SEPARATOR,
    UNIQ_OPT_DELIMITED,
    UNIQ_OPT_INTEGER_IPS,
    UNIQ_OPT_ZERO_PAD_IPS,
    UNIQ_OPT_INTEGER_SENSORS,
    UNIQ_OPT_EPOCH_TIME,
    UNIQ_OPT_OUTPUT_PATH,
    UNIQ_OPT_PRINT_FILENAMES,
    UNIQ_OPT_PAGER,
    UNIQ_OPT_COPY_INPUT
#ifdef ENABLE_BINARY
    , UNIQ_OPT_ENABLE_BINARY
#endif
} appOptionsEnum;

/* Skip deprecated switches in --help output */
static int app_option_first_index = UNIQ_OPT_FIELDS;

static struct option appOptions[] = {
    {"threshold",        REQUIRED_ARG, 0, UNIQ_OPT_THRESHOLD},
    {"fields",           REQUIRED_ARG, 0, UNIQ_OPT_FIELDS},
    {"all-counts",       NO_ARG,       0, UNIQ_OPT_ALL_COUNTS},
    {"flows",            OPTIONAL_ARG, 0, UNIQ_OPT_COUNT_FLOWS},
    {"packets",          OPTIONAL_ARG, 0, UNIQ_OPT_COUNT_PACKETS},
    {"bytes",            OPTIONAL_ARG, 0, UNIQ_OPT_COUNT_BYTES},
    {"stime",            NO_ARG,       0, UNIQ_OPT_COUNT_STIME},
    {"etime",            NO_ARG,       0, UNIQ_OPT_COUNT_ETIME},
    {"bin-time",         OPTIONAL_ARG, 0, UNIQ_OPT_BIN_TIME},
    {"no-titles",        NO_ARG,       0, UNIQ_OPT_NO_TITLES},
    {"no-columns",       NO_ARG,       0, UNIQ_OPT_NO_COLUMNS},
    {"column-separator", REQUIRED_ARG, 0, UNIQ_OPT_COLUMN_SEPARATOR},
    {"delimited",        OPTIONAL_ARG, 0, UNIQ_OPT_DELIMITED},
    {"integer-ips",      NO_ARG,       0, UNIQ_OPT_INTEGER_IPS},
    {"zero-pad-ips",     NO_ARG,       0, UNIQ_OPT_ZERO_PAD_IPS},
    {"integer-sensors",  NO_ARG,       0, UNIQ_OPT_INTEGER_SENSORS},
    {"epoch-time",       NO_ARG,       0, UNIQ_OPT_EPOCH_TIME},
    {"output-path",      REQUIRED_ARG, 0, UNIQ_OPT_OUTPUT_PATH},
    {"print-filenames",  NO_ARG,       0, UNIQ_OPT_PRINT_FILENAMES},
    {"pager",            REQUIRED_ARG, 0, UNIQ_OPT_PAGER},
    {"copy-input",       REQUIRED_ARG, 0, UNIQ_OPT_COPY_INPUT},
#ifdef ENABLE_BINARY
    {"binary",           NO_ARG,       0, UNIQ_OPT_ENABLE_BINARY},
#endif
    {0,0,0,0}            /* sentinal entry */
};

static char *appHelp[] = {
    "DEPRECATED.  Use --flows instead",
    NULL, /* built dynamically */
    ("Enable the next five switches--count everything.  If no\n"
     "\tcount is specified, flows are counted.  Def. No"),
    ("Count flow records; optionally set set minimum threshold\n"
     "\tfor output. Def. Yes unless other count switch given, 1"),
    ("Count packets; optionally set minimum packet threshold\n"
     "\tfor output. Def. No, 1"),
    ("Count bytes; optionally set minimum byte threshold\n"
     "\tfor output. Def. No, 1"),
    "Print earliest time for a flow with the given key. Def. No",
    "Print latest time for a flow with the given key. Def. No",
    ("When using 'sTime' or 'eTime' as a key, adjust time(s) to\n"
     "\tto appear in N-second bins (floor of time is used). Def. No, "
     DEFAULT_TIME_BIN_STR),
    "Do not print column titles. Def. Print titles",
    "Disable fixed-width columnar output. Def. Columnar",
    "Use specified character between columns. Def. '|'",
    "Shortcut for --no-columns --column-sep=CHAR",
    "Print IP numbers as integers. Def. Dotted decimal",
    "Print IP numbers as zero-padded dotted decimal. Def. No",
    "Print sensor as an integer. Def. Sensor name",
    "Print times in UNIX epoch seconds. Def. No",
    "Send output to given file path. Def. stdout",
    "Print names of input files as they are opened. Def. No",
    "Program to invoke to page output. Def. $SILK_PAGER or $PAGER",
    "Copy all input SiLK Flows to given pipe or file. Def. No",
#ifdef ENABLE_BINARY
    "Output in binary (hash serialization) format. Def. ASCII",
#endif
    (char *)NULL
};



/* FUNCTION DEFINITIONS */

/*
 *  appUsage();
 *
 *    Print short usage information---telling the user to use the help
 *    option---to stderr and exit the application with a FAILURE exit
 *    status.
 */
void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH and exit the
 *    application normally.  The function is passed to optionsSetup()
 *    to print the usage when the --help option is given.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                             \
    ("%s --fields=N [SWITCHES] [FILES]\n"                                     \
     "\tSummarize SiLK Flow records into user-defined keyed bins specified\n" \
     "\twith the --fields switch.  For each keyed bin, print byte, packet,\n" \
     "\tand/or flow counts and/or the time window when key was active.\n"     \
     "\tWhen no files are given on command line, flows are read from STDIN.\n")

    fprintf(USAGE_FH, USAGE_MSG, skAppName());
    appOptionsUsage(USAGE_FH);
    fprintf(USAGE_FH,
            "NOTES:\n"
            "\t* The --fields switch is required.\n"
            "\t* If --bytes or --packets is specified, rwuniq will no longer\n"
            "\t  count flows unless --flows is also specified.\n");
    exit(EXIT_SUCCESS);
}


/*
 *  appOptionsUsage(FILE *fh);
 *
 *    For each option defined in this application, print its name,
 *    whether it takes an option, and its help string.  Print to the
 *    specified file handle, 'fh'.
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\nSWITCHES:\n");
    for (i = app_option_first_index; appOptions[i].name; i++) {
        fprintf(fh, "--%s %s. ", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]));
        switch ((appOptionsEnum)i) {
          case UNIQ_OPT_FIELDS:
            /* Dynamically build the help */
            fprintf(fh, "Field(s) to use as key:\n");
            rwAsciiFieldMapPrintUsage(fh, field_map);
            break;
          default:
            /* Simple help text from the appHelp array */
            fprintf(fh, "%s\n", appHelp[i]);
            break;
        }
    }
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
static void appSetup(int argc, char **argv)
{
    int using_pager = 0;
    uint8_t i;

    /* register the application */
    skAppRegister(argv[0]);

    /* zero-out the key_fields */
    memset(&key_fields, 0, sizeof(key_fields));

    /* default output is stdout */
    uniq_io.outFPath = "stdout";
    uniq_io.outF = stdout;

    /* set key byte widths */
    initializeKeyByteWidths();

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* Set up and register the options */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);
    }

    /* Register our options handler */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);
    }

    /* Set up the ascii stream for printing results */
    if (rwAsciiStreamCreate(&uniq_io.rwas)) {
        skAppPrintErr("cannot create output stream");
        exit(EXIT_FAILURE);
    }
    rwAsciiSetDelimiter(uniq_io.rwas, g_delimiter);
    rwAsciiSetNoNewline(uniq_io.rwas);

    /* initialize string-map of field identifiers: add default fields,
     * then remove millisec fields, since unique-ing over them makes
     * little sense.
     *
     * Note that although we remove the MSEC fields from the available
     * fields here, the remainder of the code still supports MSEC
     * fields---which are mapped onto the non-MSEC versions of the
     * fields. */
    if (rwAsciiFieldMapAddDefaultFields(&field_map)) {
        skAppPrintErr("unable to setup fields stringmap");
        exit(EXIT_FAILURE);
    }
    (void)skStringMapRemoveByID(field_map, RWREC_FIELD_STIME_MSEC);
    (void)skStringMapRemoveByID(field_map, RWREC_FIELD_ETIME_MSEC);
    (void)skStringMapRemoveByID(field_map, RWREC_FIELD_ELAPSED_MSEC);

    /* allow for 0 input and 0 output pipes */
    uniq_io.ISP = iochecksSetup(0, 0, argc, argv);

    /* set up configuration structure for the current run */
    if (valueConfigInitialize()) {
        skAppPrintErr("unable to initialize value configuration");
        exit(EXIT_FAILURE);
    }

    /* parse options */
    uniq_io.ISP->firstFile = optionsParse(argc, argv);
    if (uniq_io.ISP->firstFile < 0) {
        appUsage();/* never returns */
    }

    /* Make sure the user specified at least one field */
    if (key_fields.kf_num_fields == 0) {
        skAppPrintErr("the --fields switch is required");
        appUsage();         /* never returns */
    }

    /* Warn if --bin-time requested but time not part of key.  Warn
     * that duration field will be modified when all time fields are
     * given. */
    if (time_bin_size != 0) {
        int saw_time = 0;
        for (i = 0; i < key_fields.kf_num_fields; ++i) {
            if ((key_fields.kf_fields[i] == RWREC_FIELD_ELAPSED)
                || (key_fields.kf_fields[i] == RWREC_FIELD_ELAPSED_MSEC))
            {
                saw_time |= 1;
            }
            if ((key_fields.kf_fields[i] == RWREC_FIELD_STIME)
                || (key_fields.kf_fields[i] == RWREC_FIELD_STIME_MSEC))
            {
                saw_time |= 2;
            }
            if ((key_fields.kf_fields[i] == RWREC_FIELD_ETIME)
                || (key_fields.kf_fields[i] == RWREC_FIELD_ETIME_MSEC))
            {
                saw_time |= 4;
            }
        }
        if (saw_time < 2) {
            if (FILEIsATty(stderr)) {
                skAppPrintErr("No time fields in key; %s switch ignored",
                              appOptions[UNIQ_OPT_BIN_TIME].name);
            }
        } else if (saw_time == 7) {
            /* must adjust elapsed to be eTime-sTime */
            adjust_elapsed = 1;
            if (FILEIsATty(stderr)) {
                skAppPrintErr("Warning: modifying duration field "
                              "to be eTime-sTime");
            }
        }
    }

    /* Use STDIN as an input stream if it is not a TTY; make certain
     * we have some input and we are either reading from STDIN or
     * using files listed the command line, but not both. */
    if (iochecksAcceptFromStdin(uniq_io.ISP)
        || iochecksInputs(uniq_io.ISP, 0))
    {
        appUsage();
    }

#ifdef ENABLE_BINARY
    /* if binary is selected, check that uniq_io.outF is not a tty */
    if (uniq_flags.binary_output) {
        if (FILEIsATty(uniq_io.outF)) {
            skAppPrintErr("Will not dump binary output to tty\n");
            exit(EXIT_FAILURE);
        }
    }
#endif

    /* Set up the value record: calculate the offsets based on the
     * command line switches seen; calculate total byte width of
     * value; set default value if no switches given */
    valueConfigSetOffsets();

    /* final check. See if stdout is being used for both --copy-input
     * and as the destination for the uniq data.  */
    if (uniq_io.ISP->stdoutUsed && uniq_io.outF == stdout) {
        skAppPrintErr("stdout used for both --copy-input and ascii output");
        exit(EXIT_FAILURE);
    }

    /* Invoke the pager */
    using_pager = skOpenPagerWhenStdoutTty(&(uniq_io.outF), &pager);
    if (using_pager < 0) {
        exit(EXIT_FAILURE);
    }
    if (using_pager) {
        uniq_io.outFPath = pager;
        uniq_io.outIsPipe = 1;
    }

    /* Finally, set the output handle */
    rwAsciiSetOutputHandle(uniq_io.rwas, uniq_io.outF);

    /* We don't need this any more.  Destroy it */
    if (field_map != NULL) {
        skStringMapDestroy(field_map);
        field_map = NULL;
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;                       /* OK */
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    iochecksTeardown(uniq_io.ISP);

    if (uniq_io.outF != stdout) {
        if (uniq_io.outIsPipe) {
            pclose(uniq_io.outF);
        } else {
            if (EOF == fclose(uniq_io.outF)) {
                skAppPrintErr("Error closing output file '%s': %s",
                              uniq_io.outFPath, strerror(errno));
            }
        }
    }

    rwAsciiStreamDestroy(&uniq_io.rwas);

    if (field_map != NULL) {
        skStringMapDestroy(field_map);
        field_map = NULL;
    }

    return;
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    Called by optionsParse(), this handles a user-specified switch
 *    that the application has registered, typically by setting global
 *    variables.  Returns 1 if the switch processing failed or 0 if it
 *    succeeded.  Returning a non-zero from from the handler causes
 *    optionsParse() to return a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    uint32_t temp;
    int i;

    switch ((appOptionsEnum)opt_index) {
      case UNIQ_OPT_FIELDS:
        if (parseFields(opt_arg)) {
            return 1;
        }
        break;

      case UNIQ_OPT_ALL_COUNTS:
        for (i = 0; i < TOTAL_COUNTS; ++i) {
            value_fields[i].vf_iscounted = 1;
        }
        break;

      case UNIQ_OPT_THRESHOLD:
#if 0
        /* print a warning if user is running interactively */
        if (FILEIsATty(stdout)) {
           skAppPrintErr("--threshold is deprecated.  use --flows instead");
        }
        /*@fallthrough@*/
#endif
      case UNIQ_OPT_COUNT_FLOWS:
        value_fields[COUNT_FLOWS].vf_iscounted = 1;
        if (opt_arg) {
            if (skStringParseUint32(&temp, opt_arg, 1, 0)) {
                skAppPrintErr("Error parsing %s threshold '%s'",
                              appOptions[opt_index].name, opt_arg);
                return 1;
            }
            value_fields[COUNT_FLOWS].vf_min = temp;
        }
        break;

      case UNIQ_OPT_COUNT_PACKETS:
        value_fields[COUNT_PACKETS].vf_iscounted = 1;
        if (opt_arg) {
            if (skStringParseUint32(&temp, opt_arg, 1,0)){
                skAppPrintErr("Error parsing %s threshold '%s'",
                              appOptions[opt_index].name, opt_arg);
                return 1;
            }
            value_fields[COUNT_PACKETS].vf_min = temp;
        }
        break;

      case UNIQ_OPT_COUNT_BYTES:
        value_fields[COUNT_BYTES].vf_iscounted = 1;
        if (opt_arg) {
            if (skStringParseUint32(&temp, opt_arg, 1, 0)) {
                skAppPrintErr("Error parsing %s threshold '%s'",
                              appOptions[opt_index].name, opt_arg);
                return 1;
            }
            value_fields[COUNT_BYTES].vf_min = temp;
        }
        break;

      case UNIQ_OPT_COUNT_STIME:
        /* earlieset stime */
        value_fields[COUNT_STIME].vf_iscounted = 1;
        break;

      case UNIQ_OPT_COUNT_ETIME:
        /* latest etime */
        value_fields[COUNT_ETIME].vf_iscounted = 1;
        break;

      case UNIQ_OPT_BIN_TIME:
        if (opt_arg == NULL || opt_arg[0] == '\0') {
            /* no time given; use default */
            time_bin_size = DEFAULT_TIME_BIN;
        } else {
            /* parse user's time */
            if (skStringParseUint32(&time_bin_size, opt_arg, 1, 0)) {
                skAppPrintErr("Cannot parse %s value '%s'",
                              appOptions[opt_index].name, opt_arg);
                return 1;
            }
        }
        break;

      case UNIQ_OPT_NO_TITLES:
        uniq_flags.print_titles = 0;
        rwAsciiSetNoTitles(uniq_io.rwas);
        break;

      case UNIQ_OPT_NO_COLUMNS:
        uniq_flags.tabular_output = 0;
        rwAsciiSetNoColumns(uniq_io.rwas);
        break;

      case UNIQ_OPT_COLUMN_SEPARATOR:
        g_delimiter = opt_arg[0];
        rwAsciiSetDelimiter(uniq_io.rwas, g_delimiter);
        break;

      case UNIQ_OPT_DELIMITED:
        if (opt_arg) {
            g_delimiter = opt_arg[0];
            rwAsciiSetDelimiter(uniq_io.rwas, g_delimiter);
        }
        uniq_flags.tabular_output = 0;
        rwAsciiSetNoColumns(uniq_io.rwas);
        break;

      case UNIQ_OPT_INTEGER_IPS:
        rwAsciiSetIntegerIps(uniq_io.rwas);
        break;

      case UNIQ_OPT_ZERO_PAD_IPS:
        rwAsciiSetZeroPadIps(uniq_io.rwas);
        break;

      case UNIQ_OPT_INTEGER_SENSORS:
        rwAsciiSetIntegerSensors(uniq_io.rwas);
        break;

      case UNIQ_OPT_EPOCH_TIME:
        uniq_flags.epoch_time = 1;
        rwAsciiSetEpochTime(uniq_io.rwas);
        break;

      case UNIQ_OPT_OUTPUT_PATH:
        if (!opt_arg || !strlen(opt_arg)) {
            skAppPrintErr("missing output path name");
            return 1;
        }
        uniq_io.outFPath = opt_arg;
        if (openFile(uniq_io.outFPath, 1 /* write */, &uniq_io.outF,
                      &uniq_io.outIsPipe))
        {
            return 1;
        }
        break;

      case UNIQ_OPT_PRINT_FILENAMES:
        uniq_flags.print_filenames = 1;
        break;

      case UNIQ_OPT_PAGER:
        pager = opt_arg;
        break;

      case UNIQ_OPT_COPY_INPUT:
        if (!opt_arg) {
            skAppPrintErr("Required arg to copy-input missing");
            return 1;
        }
        if (iochecksAllDestinations(uniq_io.ISP, opt_arg)) {
            return 1;
        }
        break;

#ifdef ENABLE_BINARY
      case UNIQ_OPT_ENABLE_BINARY:
        skAppPrintErr("Enabling binary output!");
        uniq_flags.binary_output = 1;
        break;
#endif
    }

    return 0;                     /* OK */
}


/*
 *  initializeKeyByteWidths();
 *
 *    Set the key widths in the key_byte_width[] global array
 */
static void initializeKeyByteWidths(void)
{
    rwRec rwrec; /* needed for field sizes */

    /* compute the length for each field */
    key_byte_width[RWREC_FIELD_SIP]          = sizeof(rwrec.sIP);
    key_byte_width[RWREC_FIELD_DIP]          = sizeof(rwrec.dIP);
    key_byte_width[RWREC_FIELD_SPORT]        = sizeof(rwrec.sPort);
    key_byte_width[RWREC_FIELD_DPORT]        = sizeof(rwrec.dPort);
    key_byte_width[RWREC_FIELD_PROTO]        = sizeof(rwrec.proto);
    key_byte_width[RWREC_FIELD_PKTS]         = sizeof(rwrec.pkts);
    key_byte_width[RWREC_FIELD_BYTES]        = sizeof(rwrec.bytes);
    key_byte_width[RWREC_FIELD_FLAGS]        = sizeof(rwrec.flags);
    key_byte_width[RWREC_FIELD_STIME]        = sizeof(rwrec.sTime);
    key_byte_width[RWREC_FIELD_ELAPSED]      = sizeof(rwrec.elapsed);
    /* eTime has same width as sTime */     
    key_byte_width[RWREC_FIELD_ETIME]        = sizeof(rwrec.sTime);
    key_byte_width[RWREC_FIELD_SID]          = sizeof(rwrec.sID);
    key_byte_width[RWREC_FIELD_INPUT]        = sizeof(rwrec.input);
    key_byte_width[RWREC_FIELD_OUTPUT]       = sizeof(rwrec.output);
    key_byte_width[RWREC_FIELD_NHIP]         = sizeof(rwrec.nhIP);
    key_byte_width[RWREC_FIELD_INIT_FLAGS]   = sizeof(rwrec.init_flags);
    key_byte_width[RWREC_FIELD_REST_FLAGS]   = sizeof(rwrec.rest_flags);
    key_byte_width[RWREC_FIELD_TCP_STATE]    = sizeof(rwrec.tcp_state);
    key_byte_width[RWREC_FIELD_APPLICATION]  = sizeof(rwrec.application);
    key_byte_width[RWREC_FIELD_FTYPE_CLASS]  = sizeof(rwrec.flow_type);
    key_byte_width[RWREC_FIELD_FTYPE_TYPE]   = sizeof(rwrec.flow_type);
    /* Ignore the milliseconds field */
    key_byte_width[RWREC_FIELD_STIME_MSEC]   = sizeof(rwrec.sTime);
    key_byte_width[RWREC_FIELD_ETIME_MSEC]   = sizeof(rwrec.sTime);
    key_byte_width[RWREC_FIELD_ELAPSED_MSEC] = sizeof(rwrec.elapsed);
}


/*
 *  status = parseFields(field_string);
 *
 *    Parse the string that represents the columns the user wishes to
 *    "uniq" over and fill in the key_fields structure.  Return 0 on
 *    success or non-zero on error.
 */
static int parseFields(const char *field_string)
{
    uint32_t field_count;
    uint32_t *parsed_list;
    uint32_t pad;
    uint32_t i;
    uint8_t j;
    int want_field;

    /* have we been here before? */
    if (key_fields.kf_num_fields > 0) {
        skAppPrintErr("--fields option given multiple times");
        return 1;
    }

    if (field_string == NULL || field_string[0] == '\0') {
        skAppPrintErr("missing --fields value");
        return 1;
    }

    if (rwAsciiFieldMapParseFields(&parsed_list, &field_count, field_string,
                                   field_map, RW_ASCII_DUPES_ERROR))
    {
        /* library printed the error */
        return 1;
    }
    assert(field_count < RWREC_PRINTABLE_FIELD_COUNT);

    /* set the fields on the AsciiStream */
    if (0 != rwAsciiSetFields(uniq_io.rwas, parsed_list, field_count)) {
        skAppPrintErr("Error setting fields");
        free(parsed_list);
        return 1;
    }

    /*
     * Copy values into the key_fields struct; compute the key width.
     */
    key_fields.kf_byte_width = 0;
    key_fields.kf_num_fields = 0;

    for (i = 0; i < field_count; i++) {
        assert(parsed_list[i] < RWREC_PRINTABLE_FIELD_COUNT);

        /* For values that key off of the same column, we only need
         * one of the values. */
        want_field = 1;
        switch (parsed_list[i]) {
          case RWREC_FIELD_FTYPE_CLASS:
          case RWREC_FIELD_FTYPE_TYPE:
            for (j = 0; j < key_fields.kf_num_fields; ++j) {
                if ((key_fields.kf_fields[j] == RWREC_FIELD_FTYPE_CLASS)
                    || (key_fields.kf_fields[j] == RWREC_FIELD_FTYPE_TYPE))
                {
                    want_field = 0;
                }
            }
            break;

          case RWREC_FIELD_STIME:
          case RWREC_FIELD_STIME_MSEC:
            for (j = 0; j < key_fields.kf_num_fields; ++j) {
                if ((key_fields.kf_fields[j] == RWREC_FIELD_STIME)
                    || (key_fields.kf_fields[j] == RWREC_FIELD_STIME_MSEC))
                {
                    want_field = 0;
                }
            }
            break;

          case RWREC_FIELD_ETIME:
          case RWREC_FIELD_ETIME_MSEC:
            for (j = 0; j < key_fields.kf_num_fields; ++j) {
                if ((key_fields.kf_fields[j] == RWREC_FIELD_ETIME)
                    || (key_fields.kf_fields[j] == RWREC_FIELD_ETIME_MSEC))
                {
                    want_field = 0;
                }
            }
            break;

          case RWREC_FIELD_ELAPSED:
          case RWREC_FIELD_ELAPSED_MSEC:
            for (j = 0; j < key_fields.kf_num_fields; ++j) {
                if ((key_fields.kf_fields[j] == RWREC_FIELD_ELAPSED)
                    || (key_fields.kf_fields[j] == RWREC_FIELD_ELAPSED_MSEC))
                {
                    want_field = 0;
                }
            }
            break;

          default:
            break;
        }

        if (want_field) {
            /* add this column to the key, and compute its byte-offset */
            key_fields.kf_fields[key_fields.kf_num_fields]
                = (uint8_t)parsed_list[i];
            key_fields.kf_byte_offset[key_fields.kf_num_fields]
                = key_fields.kf_byte_width;
            ++key_fields.kf_num_fields;

            /* compute new key width */
            key_fields.kf_byte_width += key_byte_width[parsed_list[i]];
        }
    }

    free(parsed_list);

    if (key_fields.kf_byte_width > MAX_KEY_BYTE_WIDTH) {
        skAppPrintErr("Fields make too large a key: %u bytes > %u max",
                      key_fields.kf_byte_width, MAX_KEY_BYTE_WIDTH);
        return 1;
    }

    /* pad key width to a 32-bit boundary to avoid bus errors on Sparc */
    pad = 4 - (key_fields.kf_byte_width % 4);
    if (pad != 4) {
        key_fields.kf_byte_width += pad;
    }

    return 0;
}


/*
 *  dumpRecordAscii(rwrec, values)
 *
 *    Print the rwrecord and its counts to output stream uniq_io.outF.
 *    values[] is an array of TOTAL_COUNTS values representing the
 *    various sums.
 */
static void dumpRecordAscii(
    const rwRec    *rwrec,
    const uint64_t  values[TOTAL_COUNTS])
{
    char buf[VALUE_OUT_BUF_SIZE];
    char *cp;
    int i;

    rwAsciiPrintRec(uniq_io.rwas, rwrec);

    for (i = 0; i < TOTAL_COUNTS; ++i) {
        if (value_fields[i].vf_iscounted) {
            cp = buf;
            switch (i) {
              case COUNT_BYTES:
              case COUNT_PACKETS:
              case COUNT_FLOWS:
                sprintf(cp, ("%" PRIu64), values[i]);
                break;

              case COUNT_STIME:
              case COUNT_ETIME:
                if (uniq_flags.epoch_time) {
                    sprintf(cp, ("%" PRIu64), values[i]);
                } else {
                    cp = timestamp((uint32_t)values[i]);
                }
                break;

              default:
                skAppPrintErr(("Programmer error: unhandled case '%d'"
                               " in switch() at %s:%d"),
                              i, __FILE__, __LINE__);
                assert(0);
                exit(EXIT_FAILURE);
            }

            if (uniq_flags.tabular_output) {
                fprintf(uniq_io.outF, "%*s%c", value_col[i].vc_ascii_width,
                        cp, g_delimiter);
            } else {
                fprintf(uniq_io.outF, "%s%c",
                        cp, g_delimiter);
            }
        }
    }

    fprintf(uniq_io.outF, "\n");
}


/*
 *  writeColTitles(fh);
 *
 *    Print the column titles to the output stream fh.
 */
static void writeColTitles(FILE *fh)
{
    int i;

    if (uniq_flags.print_titles == 0) {
        return;
    }

    rwAsciiPrintTitles(uniq_io.rwas);

    for (i = 0; i < TOTAL_COUNTS; ++i) {
        if (value_fields[i].vf_iscounted) {
            if (uniq_flags.tabular_output) {
                fprintf(fh, "%*.*s%c",
                        value_col[i].vc_ascii_width,
                        value_col[i].vc_ascii_width,
                        value_col[i].vc_title, g_delimiter);
            } else {
                fprintf(fh, "%s%c", value_col[i].vc_title, g_delimiter);
            }
        }
    }
    fprintf(fh, "\n");
}


/* Dump data in ascii form */
static void dumpAscii(FILE *fh)
{
    HASH_ITER it;
    uint8_t *key;
    uint8_t *val_ptr; /* the count */
    rwRec rwrec;
    uint64_t values[TOTAL_COUNTS];

#ifdef GIVE_META_INFO
    uint32_t tCount = 0;          /* table entries */
    uint32_t output_count = 0;    /* records printed */
#endif

#ifdef GIVE_META_INFO
    /* A little bit of header information */
    fprintf(fh, "Accumulated %u records.  ",
            hashlib_count_entries(g_hash_counter));
    fprintf(fh, "Thresholds:");
    for (i = 0; i < TOTAL_COUNTS; ++i) {
        if (value_fields[i].vf_iscounted) {
            fprintf(fh, (" %s >= %" PRIu64 ";"),
                    value_col[i].vc_title, value_fields[i].vf_min);
        }
    }
    fprintf(fh, "\n");
#endif

    /* Write out the headings */
    writeColTitles(fh);

    /* create the iterator */
    it = hashlib_create_iterator(g_hash_counter);

    /* Write out the records */
    while (hashlib_iterate(g_hash_counter, &it, &key, &val_ptr)
           != ERR_NOMOREENTRIES)
    {
#ifdef GIVE_META_INFO
        tCount++;
#endif
        if (unpackValues(values, val_ptr)) {
            decomposeCompoundKey(&rwrec, key);
            dumpRecordAscii(&rwrec, values);
#ifdef GIVE_META_INFO
            ++output_count;
#endif
        }
    }

#ifdef GIVE_META_INFO
    if (output_count == 0) {
        fprintf(fh, "No records with frequency > than cutoffs\n");
    } else {
        fprintf(fh, "%u of %u records met the thresholds\n",
                output_count, tCount);
    }
#endif

    return;
}


#ifdef  ENABLE_BINARY
/* Serialize the hash table data in binary form */
void dumpBinary(FILE *fd)
{
    genericHeader header;

    /* Build the header */
    PREPHEADER(&header);
    header.type = FT_RWUNIQ_DATA;
    header.version = RWUNIQ_FILE_VERSION;

    /* Write the data, prefixed with the header */
    hashlib_serialize_table(g_hash_counter, df,
                            (uint8_t*)&header, sizeof(header));
}
#endif


/*
 * createCompoundKey(rwrec, &key);
 *
 * Fill in the hash key 'key' using fields from 'rwrec'.  The fields
 * to use are given by the global key_fields.kf_fields[].
 */
static void createCompoundKey(
    const rwRec    *rwrec,
    uint8_t        *key)
{
    uint8_t i;

    for (i = 0; i < key_fields.kf_num_fields; ++i) {
        switch (key_fields.kf_fields[i]) {
          case RWREC_FIELD_SIP:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->sIP.ipnum,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_DIP:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->dIP.ipnum,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_SPORT:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->sPort,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_DPORT:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->dPort,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_PROTO:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->proto,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_PKTS:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->pkts,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_BYTES:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->bytes,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_FLAGS:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->flags,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_STIME:
          case RWREC_FIELD_STIME_MSEC:
            if (time_bin_size == 0) {
                /* no time bin */
                memcpy(key+key_fields.kf_byte_offset[i], &rwrec->sTime,
                       key_byte_width[key_fields.kf_fields[i]]);
            } else {
                /* find the floor of the time */
                uint32_t sTime;
                sTime = rwrec->sTime - (rwrec->sTime % time_bin_size);
                memcpy(key+key_fields.kf_byte_offset[i], &sTime,
                       key_byte_width[key_fields.kf_fields[i]]);
            }
            break;
          case RWREC_FIELD_ELAPSED:
          case RWREC_FIELD_ELAPSED_MSEC:
            if (adjust_elapsed == 0) {
                memcpy(key+key_fields.kf_byte_offset[i], &rwrec->elapsed,
                       key_byte_width[key_fields.kf_fields[i]]);
            } else {
                uint32_t elapsed;
                elapsed = (rwrec->elapsed
                           - ((rwrec->sTime + rwrec->elapsed) % time_bin_size)
                           + (rwrec->sTime % time_bin_size));
                memcpy(key+key_fields.kf_byte_offset[i], &elapsed,
                       key_byte_width[key_fields.kf_fields[i]]);
            }
            break;
          case RWREC_FIELD_ETIME:
          case RWREC_FIELD_ETIME_MSEC:
            {
                uint32_t eTime;
                /* eTime is computed */
                eTime = rwrec->sTime + rwrec->elapsed;
                if (time_bin_size != 0) {
                    /* compute floor */
                    eTime -= (eTime % time_bin_size);
                }
                memcpy(key+key_fields.kf_byte_offset[i], &eTime,
                       key_byte_width[key_fields.kf_fields[i]]);
            }
            break;
          case RWREC_FIELD_SID:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->sID,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_INPUT:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->input,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_OUTPUT:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->output,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_NHIP:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->nhIP.ipnum,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_INIT_FLAGS:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->init_flags,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_REST_FLAGS:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->rest_flags,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_TCP_STATE:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->tcp_state,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_APPLICATION:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->application,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_FTYPE_CLASS:
          case RWREC_FIELD_FTYPE_TYPE:
            memcpy(key+key_fields.kf_byte_offset[i], &rwrec->flow_type,
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          default:
            skAppPrintErr("Case %u not handled in switch() at %s:%d",
                          key_fields.kf_fields[i], __FILE__, __LINE__);
            abort();
        }
    }
}


/*
 *  decomposeCompoundKey(&rwrec, key);
 *
 *  Fill in 'rwrec' with the values in the hash key 'key'.  Uses the
 *  global key_fields.kf_fields[].  This is the opposite of
 *  createCompoundKey().
 */
static void decomposeCompoundKey(
    rwRec          *rwrec,
    const uint8_t  *key)
{
    uint32_t eTime = 0;
    uint8_t i;

    /* zero these fields for etime calc */
    rwrec->sTime = rwrec->elapsed = 0;

    for (i = 0; i < key_fields.kf_num_fields; ++i) {
        switch (key_fields.kf_fields[i]) {
          case RWREC_FIELD_SIP:
            memcpy(&rwrec->sIP.ipnum, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_DIP:
            memcpy(&rwrec->dIP.ipnum, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_SPORT:
            memcpy(&rwrec->sPort, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_DPORT:
            memcpy(&rwrec->dPort, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_PROTO:
            memcpy(&rwrec->proto, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_PKTS:
            memcpy(&rwrec->pkts, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_BYTES:
            memcpy(&rwrec->bytes, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_FLAGS:
            memcpy(&rwrec->flags, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_STIME_MSEC:
            rwrec->sTime_msec = 0;
            /* FALLTHROUGH */
          case RWREC_FIELD_STIME:
            memcpy(&rwrec->sTime, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_ELAPSED_MSEC:
            rwrec->elapsed_msec = 0;
            /* FALLTHROUGH */
          case RWREC_FIELD_ELAPSED:
            memcpy(&rwrec->elapsed, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_ETIME_MSEC:
            rwrec->sTime_msec = 0;
            rwrec->elapsed_msec = 0;
            /* FALLTHROUGH */
          case RWREC_FIELD_ETIME:
            /* break eTime into sTime and elapsed after processing all fields*/
            memcpy(&eTime, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_SID:
            memcpy(&rwrec->sID, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_INPUT:
            memcpy(&rwrec->input, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_OUTPUT:
            memcpy(&rwrec->output, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_NHIP:
            memcpy(&rwrec->nhIP.ipnum, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_INIT_FLAGS:
            memcpy(&rwrec->init_flags, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_REST_FLAGS:
            memcpy(&rwrec->rest_flags, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_TCP_STATE:
            memcpy(&rwrec->tcp_state, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_APPLICATION:
            memcpy(&rwrec->application, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          case RWREC_FIELD_FTYPE_CLASS:
          case RWREC_FIELD_FTYPE_TYPE:
            memcpy(&rwrec->flow_type, key+key_fields.kf_byte_offset[i],
                   key_byte_width[key_fields.kf_fields[i]]);
            break;
          default:
            skAppPrintErr("Case %u not handled in switch() at %s:%d",
                          key_fields.kf_fields[i], __FILE__, __LINE__);
            abort();
        }
    }

    /* handle eTime: records don't store eTime, so need to modify the
     * record's 'sTime' and 'elapsed' fields so that we get the
     * correct eTime out when we print. */
    if (eTime) {
        if (rwrec->sTime) {
            if (rwrec->elapsed == 0) {
                /* set elapsed based on start and end times */
                assert(rwrec->sTime <= eTime);
                rwrec->elapsed = eTime - rwrec->sTime;
            }
#ifndef NDEBUG
            else {
                /* nothing to do to the records; verify that the times
                 * all make sense. */
                assert(eTime == (rwrec->sTime + rwrec->elapsed));
            }
#endif
        } else if (rwrec->elapsed) {
            /* set start time based on end time and elapsed */
            assert(eTime > rwrec->elapsed);
            rwrec->sTime = eTime - rwrec->elapsed;
        } else {
            /* Just set sTime to eTime and elapsed is 0 */
            rwrec->sTime = eTime;
        }
    }
}


/*
 *  countFileCompound(rwIOS);
 *
 *  Count the number of uniq occurrences of a multiple fields in the
 *  records in the RW file pointed to by 'rwIOS'.  The fields are
 *  determined by the key_fields.kf_fields[] global.
 *
 *  Returns 0 on success.  If the hash table reports no memory,
 *  returns RWUNIQ_NO_MEMORY_EXIT_CODE.  If the hash table reports any
 *  other error, we exit.
 */
static int countFileCompound(rwIOStruct *rwIOS)
{
    static uint8_t *key = NULL;
    rwRec rwrec;
    uint8_t *val_ptr;
    int rv;

    /* create key */
    if (!key) {
        key = calloc(key_fields.kf_byte_width, sizeof(uint8_t));
        if (!key) {
            skAppPrintErr("Unable to allocate key");
            exit(EXIT_FAILURE);
        }
    }

    /* process file */
    while (rwRead(rwIOS, &rwrec)) {
        createCompoundKey(&rwrec, key);
        rv = hashlib_insert(g_hash_counter, key, &val_ptr);
        if (OK == rv) {
            /* new value */
            memset(val_ptr, 0, value_byte_width);
            packValue(val_ptr, &rwrec);
        } else if (OK_DUPLICATE == rv) {
            /* existing value */
            packValue(val_ptr, &rwrec);
        } else if ((ERR_OUTOFMEMORY == rv) || (ERR_NOMOREBLOCKS == rv)) {
            return RWUNIQ_NO_MEMORY_EXIT_CODE;
        } else {
            skAppPrintErr("Bad return code '%d' from hash table insert", rv);
            exit(EXIT_FAILURE);
        }

        ++g_record_count;
    }

    return 0;
}


/*
 *  countFileSingle(rwIOS);
 *
 *  Count the number of uniq occurrences of a single field in the
 *  records in the RW file pointed to by 'rwIOS'.  The field is
 *  determined by key_fields.kf_fields[0].
 *
 *  This is a copy-paste-edit version of the above; it's here to
 *  optimize for the common case where we are uniq-ing over a single
 *  key field.
 *
 *  Returns 0 on success.  If the hash table reports no memory,
 *  returns RWUNIQ_NO_MEMORY_EXIT_CODE.  If the hash table reports any
 *  other error, we exit.
 */
static int countFileSingle(rwIOStruct *rwIOS)
{
    rwRec rwrec;
    uint8_t *val_ptr;
    uint8_t key_val[sizeof(uint32_t)];
    uint8_t *key;
    int rv;

    if (time_bin_size != 0) {
        /* Must use the compound function when time-binning is
         * requested, since key is no longer a simple offset. */
        return countFileCompound(rwIOS);
    }

    /* make sure key_val is clean */
    memset(key_val, 0, sizeof(key_val));

    /* Figure out which field in the rwrec will be the key and store its
     * address in 'key' */
    switch (key_fields.kf_fields[0]) {
      case RWREC_FIELD_ETIME:
      case RWREC_FIELD_ETIME_MSEC:
        /* eTime must be computed, and we can't set key to the address of
         * anything, so use the compound version of count_file instead */
        return countFileCompound(rwIOS);
      case RWREC_FIELD_SIP:
        key = (uint8_t*)&rwrec.sIP.ipnum;
        break;
      case RWREC_FIELD_DIP:
        key = (uint8_t*)&rwrec.dIP.ipnum;
        break;
      case RWREC_FIELD_SPORT:
        key = (uint8_t*)&rwrec.sPort;
        break;
      case RWREC_FIELD_DPORT:
        key = (uint8_t*)&rwrec.dPort;
        break;
      case RWREC_FIELD_PROTO:
        key = (uint8_t*)&rwrec.proto;
        break;
      case RWREC_FIELD_PKTS:
        key = (uint8_t*)&rwrec.pkts;
        break;
      case RWREC_FIELD_BYTES:
        key = (uint8_t*)&rwrec.bytes;
        break;
      case RWREC_FIELD_FLAGS:
        key = (uint8_t*)&rwrec.flags;
        break;
      case RWREC_FIELD_STIME:
      case RWREC_FIELD_STIME_MSEC:
        key = (uint8_t*)&rwrec.sTime;
        break;
      case RWREC_FIELD_ELAPSED:
      case RWREC_FIELD_ELAPSED_MSEC:
        key = (uint8_t*)&rwrec.elapsed;
        break;
      case RWREC_FIELD_SID:
        key = (uint8_t*)&rwrec.sID;
        break;
      case RWREC_FIELD_INPUT:
        key = (uint8_t*)&rwrec.input;
        break;
      case RWREC_FIELD_OUTPUT:
        key = (uint8_t*)&rwrec.output;
        break;
      case RWREC_FIELD_NHIP:
        key = (uint8_t*)&rwrec.nhIP.ipnum;
        break;
      case RWREC_FIELD_INIT_FLAGS:
        key = (uint8_t*)&rwrec.init_flags;
        break;
      case RWREC_FIELD_REST_FLAGS:
        key = (uint8_t*)&rwrec.rest_flags;
        break;
      case RWREC_FIELD_TCP_STATE:
        key = (uint8_t*)&rwrec.tcp_state;
        break;
      case RWREC_FIELD_APPLICATION:
        key = (uint8_t*)&rwrec.application;
        break;
      case RWREC_FIELD_FTYPE_CLASS:
      case RWREC_FIELD_FTYPE_TYPE:
        key = (uint8_t*)&rwrec.flow_type;
        break;
      default:
        skAppPrintErr("Case %u not handled in switch() at %s:%d",
                      key_fields.kf_fields[0], __FILE__, __LINE__);
        assert(0);
        exit(EXIT_FAILURE);
    } /* switch */


    /* Process the file */
    while (rwRead(rwIOS, &rwrec)) {
        memcpy(key_val, key, key_byte_width[key_fields.kf_fields[0]]);
        rv = hashlib_insert(g_hash_counter, key_val, &val_ptr);
        if (OK == rv) {
            /* new value */
            memset(val_ptr, 0, value_byte_width);
            packValue(val_ptr, &rwrec);
        } else if (OK_DUPLICATE == rv) {
            /* existing value */
            packValue(val_ptr, &rwrec);
        } else if ((ERR_OUTOFMEMORY == rv) || (ERR_NOMOREBLOCKS == rv)) {
            return RWUNIQ_NO_MEMORY_EXIT_CODE;
        } else {
            skAppPrintErr("Bad return code '%d' from hash table insert", rv);
            exit(EXIT_FAILURE);
        }

        ++g_record_count;
    }

    return 0;
}


/*
 * initialize the value_config pointer so we can store option
 * state in it.
 *
 * modifies the global variable value_fields.
 */
static int valueConfigInitialize(void)
{
    int i;

    memset(value_fields, 0, sizeof(value_fields));

    /* thresholds default to 1 */
    for (i = 0; i < TOTAL_COUNTS ; ++i) {
        value_fields[i].vf_min = 1;
        value_fields[i].vf_max = 0;
    }
    return 0;
}



/*
 *  valueConfigSetOffsets();
 *
 *    Based on the command line flags seen for the value columns
 *    (value_fields[].vf_iscounted), set the byte-offset of each value
 *    field in the value-record.  Calculate it here so we don't have
 *    to recalculate it while we are reading records.
 */
static void valueConfigSetOffsets(void)
{
    int i;
    size_t current_offset = 0;

    while (current_offset == 0) {
        /* Compute the offset for each value field */
        for (i = 0; i < TOTAL_COUNTS; ++i) {
            if (value_fields[i].vf_iscounted) {
                value_fields[i].vf_offset = current_offset;
                current_offset += value_col[i].vc_byte_width;
            }
        }

        /* if no flags are specified, count by records (flows)---this
         * is to mimic legacy behavior that we may want to change in
         * the future.  Set the COUNT_FLOWS field to be counted, and
         * go around the while() loop again to set offset. */
        if (current_offset == 0) {
            value_fields[COUNT_FLOWS].vf_iscounted = 1;
        }
    }

    /* total bytes required to hold all value columns */
    value_byte_width = current_offset;
}


/*
 *  packValue(out_val_ptr, rec_ptr);
 *
 *    Depending on the value columns the user requested---i.e., the
 *    settings of value_field[].vf_iscounted---add the
 *    bytes/packets/flows/stime/etime in the rec_ptr to the values in
 *    the out_val_ptr.
 */
static void packValue(
    uint8_t        *out_val_ptr,
    const rwRec    *rec_ptr)
{
    /* bytes */
    if (value_fields[COUNT_BYTES].vf_iscounted) {
        uint64_t value;
        memcpy(&value, out_val_ptr + value_fields[COUNT_BYTES].vf_offset,
               value_col[COUNT_BYTES].vc_byte_width);
        value += rec_ptr->bytes;
        memcpy(out_val_ptr + value_fields[COUNT_BYTES].vf_offset, &value,
               value_col[COUNT_BYTES].vc_byte_width);
    }

    /* packets */
    if (value_fields[COUNT_PACKETS].vf_iscounted) {
        *((uint32_t *)(out_val_ptr + value_fields[COUNT_PACKETS].vf_offset))
            += rec_ptr->pkts;
    }

    /* flows */
    if (value_fields[COUNT_FLOWS].vf_iscounted) {
        ++*((uint32_t *)(out_val_ptr + value_fields[COUNT_FLOWS].vf_offset));
    }

    /* earliest sTime */
    if (value_fields[COUNT_STIME].vf_iscounted) {
        uint32_t *value =
            ((uint32_t*)(out_val_ptr + value_fields[COUNT_STIME].vf_offset));
        if (*value == 0) {
            *value = rec_ptr->sTime;
        } else if (rec_ptr->sTime < *value) {
            *value = rec_ptr->sTime;
        }
    }

    /* latest eTime */
    if (value_fields[COUNT_ETIME].vf_iscounted) {
        uint32_t *value =
            ((uint32_t*)(out_val_ptr + value_fields[COUNT_ETIME].vf_offset));
        if (*value < (rec_ptr->sTime + rec_ptr->elapsed)) {
            *value = (rec_ptr->sTime + rec_ptr->elapsed);
        }
    }
}


/*
 *  meets_thresholds = unpackValues(out_val, packed_val);
 *
 *    Does reverse of the packValues() subroutine: Unpacks the
 *    'packed_val' values into 'out_val', an array of TOTAL_COUNTS
 *    elements.  Values unpacked depend on the settings of the
 *    value_fields[].vf_iscounted flag.  Also checks the values
 *    against the min and max; returns 1 if the values are within the
 *    limits, and 0 otherwise.
 */
static int unpackValues(
    uint64_t        out_val[TOTAL_COUNTS],
    const uint8_t  *packed_val)
{
    int i;

    for (i = 0; i < TOTAL_COUNTS; ++i) {
        if (value_fields[i].vf_iscounted) {
            switch (i) {
              case COUNT_BYTES:
                memcpy(&out_val[i], (packed_val + value_fields[i].vf_offset),
                       value_col[i].vc_byte_width);
                if ((out_val[i] < value_fields[i].vf_min)
                    || ((value_fields[i].vf_max != 0)
                        && (out_val[i] > value_fields[i].vf_max)))
                {
                    return 0;
                }
                break;

              case COUNT_PACKETS:
              case COUNT_FLOWS:
                out_val[i] = 
                    *((uint32_t*)(packed_val + value_fields[i].vf_offset));
                if ((out_val[i] < value_fields[i].vf_min)
                    || ((value_fields[i].vf_max != 0)
                        && (out_val[i] > value_fields[i].vf_max)))
                {
                    return 0;
                }
                break;

              case COUNT_STIME:
              case COUNT_ETIME:
                out_val[i] =
                    *((uint32_t*)(packed_val + value_fields[i].vf_offset));
                break;

              default:
                skAppPrintErr(("Programmer error: unhandled case '%d'"
                               " in switch() at %s:%d"),
                              i, __FILE__, __LINE__);
                assert(0);
                exit(EXIT_FAILURE);
            }
        }
    }

    return 1;
}


/*
 *  uniq_main();
 *
 *    Main control function that creates a hash table, processes the
 *    input (files or stdin), and prints the results.
 */
static int uniq_main(void)
{
    rwIOStruct *rwIOS;
    const char *curFName;
    int ctr;
    uint8_t *no_val_ptr;
    int rv = 0;
    int (*count_fn)(rwIOStruct*);

    /* Create the value that the hash table uses to indicate a key
       without a value */
    no_val_ptr = (uint8_t *)malloc(value_byte_width);
    if (!no_val_ptr) {
        skAppPrintErr("Unable to allocate no-value default");
        exit(EXIT_FAILURE);
    }
    memset(no_val_ptr, 0xFF, value_byte_width);

    /* Create the hash table */
    g_hash_counter = hashlib_create_table(key_fields.kf_byte_width,
                                          value_byte_width,
                                          HTT_INPLACE,
                                          no_val_ptr,
                                          NULL,
                                          0,
                                          HASH_INITIAL_SIZE,
                                          DEFAULT_LOAD_FACTOR);
    if (NULL == g_hash_counter) {
        skAppPrintErr("Unable to create hash table");
        rv = 1;
        goto END;
    }

    /* Set a function pointer for the compound-key verses the
       single-key version of the counting code */
    if (key_fields.kf_num_fields > 1) {
        count_fn = &countFileCompound;
    } else {
        count_fn = &countFileSingle;
    }

    /* Iterate over the files to process */
    for (ctr = uniq_io.ISP->firstFile; ctr < uniq_io.ISP->fileCount ; ctr++) {

        /* open the file */
        curFName = uniq_io.ISP->fnArray[ctr];
        rwIOS = rwOpenFile(curFName, uniq_io.ISP->inputCopyFD);
        if (!rwIOS) {
            /* rwOpenFile dumps error */
            rv = 1;
            goto END;
        }

        /* print its name */
        if (uniq_flags.print_filenames) {
            fprintf(stderr, "%s\n", curFName);
        }

        /* process the file */
        rv = (*count_fn)(rwIOS);

        /* if we got an out-of-memory error, break out of the input loop */
        if (RWUNIQ_NO_MEMORY_EXIT_CODE == rv) {
            rwCloseFile(rwIOS);
            skAppPrintErr(("Out of memory after processing %" PRIu64
                           " records for\n"
                           "\t\t%u hash entries.  Printing partial results"),
                          g_record_count,
                          hashlib_count_entries(g_hash_counter));
            break;
        }

        rwCloseFile(rwIOS);
    }

    /* Generate output */
#ifdef  ENABLE_BINARY
    if (uniq_flags.binary_output) {
        dumpBinary(uniq_io.outF);
    } else {
        dumpAscii(uniq_io.outF);
    }
#else
    dumpAscii(uniq_io.outF);
#endif

  END:
    if (g_hash_counter) {
        hashlib_free_table(g_hash_counter);
    }
    free(no_val_ptr);

    return rv;
}


int main(int argc, char **argv)
{
    int rv;

    /* Global setup */
    appSetup(argc, argv);

    rv = uniq_main();

    /* Done, do cleanup */
    appTeardown();                /* AOK. Dump information */

    return rv;
}
