/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwnetmask
**
**  Read in rw-records and output rw-records, masking the Source IP
**  and Destination IP by the prefix-lengths given on the command
**  line.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/rw/rwutils/rwnetmask.c,v 1.10 2005/10/26 18:34:07 thomasm Exp $");

#include "utils.h"
#include "rwpack.h"


/* TYPEDEFS AND DEFINES */

/* File handle for --help output */
#define USAGE_FH stdout


/* LOCAL FUNCTIONS */

static void appUsage(void);			/* never returns */
static void appUsageLong(void);			/* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);	/* never returns */
static void appOptionsUsage(FILE *fh);		/* application options usage */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);


/* LOCAL VARIABLES */

/* The netmask of source/dest/next-hop IP.  For class A (/8), value
 * would be 0xFF000000. Value of 0 means to apply no mask--logically
 * the same as 0xFFFFFFFF. */
static struct {
    uint32_t  source;
    uint32_t  dest;
    uint32_t  next_hop;
} nmask_prefixes = {
    0, 0, 0
};


/* OPTIONS */

typedef enum {
    NMASK_SOURCE, NMASK_DEST, NMASK_NEXTHOP
}  appOptionsEnum;

static struct option appOptions[] = {
    {"source-prefix-length",      REQUIRED_ARG, 0, NMASK_SOURCE},
    {"destination-prefix-length", REQUIRED_ARG, 0, NMASK_DEST},
    {"next-hop-prefix-length",    REQUIRED_ARG, 0, NMASK_NEXTHOP},
    {0,0,0,0}			/* sentinal entry */
};

static char *appHelp[] = {
    /* add help strings here for the applications options */
    "High bits of the sIP to keep; Def 32",
    "High bits of the dIP to keep; Def 32",
    "High bits of the nhIP to keep; Def 32",
    (char *)NULL
};



/*
 * appUsage:
 * 	print short usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 * 	print complete usage information to stderr and exit with code.
 * 	passed to optionsSetup() to print usage when --help option
 * 	given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                             \
    ("%s [--source-prefix=N] [--destination-prefix=N] [--next-hop-prefix=N]\n"\
     "\tRead SiLK Flow records from STDIN, mask-off the lower (32-N) bits\n"  \
     "\tof the specified IP address(es), and write the resulting records to\n"\
     "\tstdout.  Does not process files.\n")

    fprintf(USAGE_FH, USAGE_MSG, skAppName());

    appOptionsUsage(USAGE_FH);
    exit(EXIT_SUCCESS);
}


/*
 * appOptionsUsage:
 * 	print options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *	do NOT add non-option usage here (i.e., required and other
 *	optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }
}


/*
 * appSetup
 *	do all the setup for this application include setting up required
 *	modules etc.
 * Arguments:
 *	argc, argv
 * Returns: None.
 * Side Effects:
 *	exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int nextOptIndex;

    skAppRegister(argv[0]);

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *))
           == (sizeof(appOptions)/sizeof(struct option)));

    /* initialize options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);			/* never returns */
    }

    /* register the options and handler */
    if (optionsRegister(appOptions, (optHandler) appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);			/* never returns */
    }

    /* parse options */
    nextOptIndex = optionsParse(argc, argv);
    if (nextOptIndex < 0) {
        appUsage();		/* never returns */
    }

    /* check for extra options */
    if (nextOptIndex > argc) {
        skAppPrintErr("unexpected arguments; input must come from stdin.");
        appUsage();
    }

    /* make certain at least one mask was specified */
    if ((0 == nmask_prefixes.source) && (0 == nmask_prefixes.dest)
        && (0 == nmask_prefixes.next_hop))
    {
        skAppPrintErr("must specify at least one prefix length option");
        appUsage();
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;			/* OK */
}


/*
 * appTeardown:
 *	teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 * 	All modules are torn down. Then application teardown takes place.
 * 	Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    optionsTeardown();

    return;
}


/*
 * appOptionsHandler:
 * 	called for each option the app has registered.
 * Arguments:
 *	clientData cData: ignored here.
 *	int index: index into appOptions of the specific option
 *	char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *	0 if OK. 1 else.
 * Side Effects:
 *	Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    uint32_t n;
    int rv;
    /* address of prefix to set */
    uint32_t *netmask_p = NULL;

    /* Store the address of the prefix to set in the switch(), then
     * set the referenced value below the switch() */
    switch ((appOptionsEnum)opt_index) {
      case NMASK_SOURCE:
        if (nmask_prefixes.source) {
            skAppPrintErr("warning: multiple source lengths given;"
                          " using final");
        }
        netmask_p = &nmask_prefixes.source;
        break;

      case NMASK_DEST:
        if (nmask_prefixes.dest) {
            skAppPrintErr("warning: multiple destination lengths given;"
                          " using final");
        }
        netmask_p = &nmask_prefixes.dest;
        break;

      case NMASK_NEXTHOP:
        if (nmask_prefixes.next_hop) {
            skAppPrintErr("warning: multiple next-hop lengths given;"
                          " using final");
        }
        netmask_p = &nmask_prefixes.next_hop;
        break;
    }

    /* we know the index is good, now set the value */
    assert(netmask_p);

    /* Parse it */
    rv = skStringParseUint32(&n, opt_arg, 1, 32);
    if (rv < -10) {
        skAppPrintErr("Prefix %lu out of range; use 1 <= n <= 32",
                      (unsigned long)n);
        return 1;
    }
    if (rv != 0) {
        skAppPrintErr("Cannot parse '%s' as a number", opt_arg);
        return 1;
    }

    *netmask_p = ~0 << (32-n);

    return 0; /* OK */
}


/*
 *  maskInput(input_file, output_file);
 *
 *    Read rw-records from the 'input_file', mask off the source,
 *    destination, and/or next-hop IP addresses, and print the records
 *    to the 'output_file'.
 */
static int maskInput(
    const char *input_file,
    const char *output_file)
{
    rwIOStructPtr input_ios;
    rwIOStructPtr output_ios;
    rwRec rwrec;
    int rv = LIBRW_OK;

    /* Open input */
    input_ios = rwOpenFile(input_file, NULL);
    if ( !input_ios) {
        return 1;
    }

    /* Open the output file */
    if ((rv = rwioCreate(&output_ios, output_file, SK_RWIO_WRITE))
        || (rv = rwioSetFileType(output_ios, rwGetFileType(input_ios)))
        || (rv = rwioSetFileVersion(output_ios, rwGetFileVersion(input_ios)))
        || (rv = rwioSetFileByteorder(output_ios,
                                      rwGetFileByteorder(input_ios)))
        || (rv = rwioOpen(output_ios)))
    {
        goto END;
    }

    rv = rwioAppendToHistoryFromFile(output_ios, input_ios);
    if ((rv != LIBRW_OK) && (rv != LIBRW_ERR_UNSUPPORT_TYPE)) {
        goto END;
    }

    rv = rwioSetFileSTime(output_ios, rwGetFileSTime(input_ios), 0);
    if ((rv != LIBRW_OK) && (rv != LIBRW_ERR_UNSUPPORT_TYPE)) {
        goto END;
    }
    rv = LIBRW_OK;

    /* read the records and mask the IP addresses */
    while ((rv == LIBRW_OK) && rwRead(input_ios, &rwrec)) {
        if (nmask_prefixes.source) {
            rwrec.sIP.ipnum &= nmask_prefixes.source;
        }
        if (nmask_prefixes.dest) {
            rwrec.dIP.ipnum &= nmask_prefixes.dest;
        }
        if (rwrec.nhIP.ipnum && nmask_prefixes.next_hop) {
            rwrec.nhIP.ipnum &= nmask_prefixes.next_hop;
        }

        rv = rwWrite(output_ios, &rwrec);
    }

 END:
    if (rv) {
        rwioPrintLastErr(output_ios, rv, &skAppPrintErr);
    }
    if (output_ios) {
        rwCloseFile(output_ios);
    }
    if (input_ios) {
        rwCloseFile(input_ios);
    }

    return rv;
}


int main(int argc, char **argv)
{
    int rv;

    appSetup(argc, argv); /* never returns on error */

    rv = maskInput("stdin", "stdout");

    /* done */
    appTeardown();

    return rv;
}
