#!/usr/bin/perl -w

# Copyright (C) 2005 by Carnegie Mellon University.
#
# @OPENSOURCE_HEADER_START@
# 
# Use of the SILK system and related source code is subject to the terms 
# of the following licenses:
# 
# GNU Public License (GPL) Rights pursuant to Version 2, June 1991
# Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
# 
# NO WARRANTY
# 
# ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
# PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
# PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
# "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
# KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
# LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
# MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
# OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
# SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
# TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
# WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
# LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
# CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
# CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
# DELIVERABLES UNDER THIS LICENSE.
# 
# Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
# Mellon University, its trustees, officers, employees, and agents from 
# all claims or demands made against them (and any related losses, 
# expenses, or attorney's fees) arising out of, or relating to Licensee's 
# and/or its sub licensees' negligent use or willful misuse of or 
# negligent conduct or willful misconduct regarding the Software, 
# facilities, or other rights or assistance granted by Carnegie Mellon 
# University under this License, including, but not limited to, any 
# claims of product liability, personal injury, death, damage to 
# property, or violation of any laws or regulations.
# 
# Carnegie Mellon University Software Engineering Institute authored 
# documents are sponsored by the U.S. Department of Defense under 
# Contract F19628-00-C-0003. Carnegie Mellon University retains 
# copyrights in all material produced under this contract. The U.S. 
# Government retains a non-exclusive, royalty-free license to publish or 
# reproduce these documents, or allow others to do so, for U.S. 
# Government purposes only pursuant to the copyright license under the 
# contract clause at 252.227.7013.
# 
# @OPENSOURCE_HEADER_END@

# geoip-csv-to-silk.pl
#
# John McClary Prevost
# June 1st, 2005
#
# RCSIDENT("$SiLK: /src/util/ccfilter_new/geoip-csv-to-silk.pl,v 1.2 2005/09/13 18:24:30 thomasm Exp $");
#
# Generate a silk-style prefixmap file from a CSV geoip data file.

use vars qw(@nodes);

@nodes = ();

$unknown_leaf = leaf('--');

$nodes[0] = [$unknown_leaf, $unknown_leaf];

while ( <> ) {
  chomp;
  s/"//g;                       # Remove '"' characters surrounding values
  my ($low_ip, $high_ip, $low_num, $high_num, $c_code, $c_name) = split /,/;
  my $low_bits = unpack('B*', (pack 'N', $low_num));
  my $high_bits = unpack('B*', (pack 'N', $high_num));
  add_to_tree($low_bits, $high_bits, lc $c_code, 0, 0, '');
}

output_tree();

sub output_tree {
  print(pack('C8', 0xDE, 0xAD, 0xBE, 0xEF, 0x00, 0x25, 0x01, 0x00));
  print(pack('V', $#nodes));
  for my $i ( 0 .. $#nodes ) {
    print(pack('VV', @{$nodes[$i]}));
  }
}

# Note that this can be done with linear allocation in an array:
# Allocated nodes are never destroyed, only modified and walked.
# It is possible for some nodes to be orphaned during the creation
# process, but only if the input data is poorly formed (overlapping
# regions.)

sub add_to_tree {
  # In order for us to be here:
  #   This node's range must overlap with the range of [$low_bits,$high_bits]
  #   This node's range must *not* be contained within [$low_bits,$high_bits]
  my ($low_bits, $high_bits, $c_code, $node, $bit, $bits) = @_;

  # Does the left subtree overlap [$low_bits,$high_bits]?
  if ( substr($low_bits, $bit, 1) eq '0' ) {

    # Is the left subtree completely contained in [$low_bits,$high_bits]?
    #   (Is the low end the low end of the left subtree?  And, is the
    #   high end either the high end of the left subtree or somewhere in
    #   the right subtree?)
    if ( (substr($low_bits, $bit+1, 31-$bit) eq ('0' x (31-$bit))) &&
         ((substr($high_bits, $bit, 1) eq '1') ||
          (substr($high_bits, $bit+1, 31-$bit) eq ('1' x (31-$bit)))) ) {
      # Left subtree is completely contained, set left subtree to country

      # Check first to warn of overlap:
      if ( $nodes[$node][0] ne $unknown_leaf ) {
        warn "Overlapping range: [$low_bits, $high_bits] x $bits";
      }

      # We're done--the left side is mapped entirely to this value
      $nodes[$node][0] = leaf($c_code);

    } else {
      # Left subtree overlaps but is not completely contained

      # Is the left subtree a leaf?
      if ( isleaf($nodes[$node][0]) ) {
        # It is: create a new node.  Initialize that node with both
        # branches pointing at the old leaf value.
        # Note: This should be $unknown_leaf with well-formed data.
        if ( $nodes[$node][0] ne $unknown_leaf ) {
          warn "Overlapping range: [$low_bits, $high_bits] x ${bits}0";
        }
        push @nodes, [$nodes[$node][0], $nodes[$node][0]];
        $nodes[$node][0] = $#nodes;
      }

      # Either way, there's a node there now, recurse down it.
      # If $high_bits is outside the left tree, chop it to be at the right edge
      if ( substr($high_bits, $bit, 1) eq '1' ) {
        $chopped = $low_bits;
        substr($chopped, $bit+1, 31-$bit, '1' x (31-$bit));
        add_to_tree($low_bits, $chopped, $c_code, $nodes[$node][0], $bit+1, $bits."0");
      } else {
        add_to_tree($low_bits, $high_bits, $c_code, $nodes[$node][0], $bit+1, $bits."0");
      }
    }
  }

  # Mirror image on right side, really

  # Does the right subtree overlap [$low_bits,$high_bits]
  if ( substr($high_bits, $bit, 1) eq '1' ) {

    # Is the right subtree completely contained in [$low_bits,$high_bits]?
    #   (Is the high end the high end of the right subtree?  And, is
    #   the low end either the low end of the right subtree or
    #   somewhere in the left subtree?)
    if ( (substr($high_bits, $bit+1, 31-$bit) eq ('1' x (31-$bit))) &&
         ((substr($low_bits, $bit, 1) eq '0') ||
          (substr($low_bits, $bit+1, 31-$bit) eq ('0' x (31-$bit)))) ) {
      # Right subtree is completely contained, set right subtree to country.

      # Check first to warn of overlap
      if ( $nodes[$node][1] ne $unknown_leaf ) {
        warn "Overlapping range: [$low_bits, $high_bits] x $bits";
      }

      # We're done--the right side is mapped entirely to this value
      $nodes[$node][1] = leaf($c_code);

    } else {
      # Right subtree overlaps but is not completely contained

      # Is the right subtree a leaf?
      if ( isleaf($nodes[$node][1]) ) {
        # It is: create a new node.  Initialize that node with both
        # branches pointing at the old leaf value.
        # Note: this should be $unknown_leaf with well-formed data.
        if ( $nodes[$node][1] ne $unknown_leaf ) {
          warn "Overlapping range: [$low_bits, $high_bits] x ${bits}1";
        }
        push @nodes, [$nodes[$node][1], $nodes[$node][1]];
        $nodes[$node][1] = $#nodes;
      }

      # Either way, there's a node there now, recurse down it.
      # If $low_bits is outside the right tree, chop it to be at the left edge
      if ( substr($low_bits, $bit, 1) eq '0' ) {
        my $chopped = $high_bits;
        substr($chopped, $bit+1, 31-$bit, '0' x (31-$bit));
        add_to_tree($chopped, $high_bits, $c_code, $nodes[$node][1], $bit+1, $bits."1");
      } else {
        add_to_tree($low_bits, $high_bits, $c_code, $nodes[$node][1], $bit+1, $bits."1");
      }
    }
  }
}

sub leaf {
  my $cc = shift;
  return 0x80000000 + ord(substr($cc,0,1)) * 256 + ord(substr($cc,1,1));
}

sub isleaf {
  my $val = shift;
  return $val & 0x80000000;
}
