/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
** prefixmap.c
**
** John McClary Prevost
** December 2nd, 2004
**
** Data structure for binding values to IP address prefixes (for
** example, mapping every value under 128.2/16 to a single value.)
** This implementation is intended to be used to provide the country
** code facility for rwfilter.
**
** The current implementation provides only lookups to the data
** structure, with the assumption that the structure has been built as
** a disk file by another tool.
*/

#include "silk.h"

RCSIDENT("$SiLK: /src/util/ccfilter_new/prefixmap.c,v 1.3 2005/02/23 21:18:13 thomasm Exp $");

#include "utils.h"
#include "prefixmap_priv.h"

/*
 * void skPrefixMapDelete(skPrefixMap_t *map);
 *
 * Frees the resources used by the prefixmap pointed to by map, then frees
 * map itself.  After calling, map is an invalid pointer.
 */

void skPrefixMapDelete(skPrefixMap_t *map)
{
    if ( map->buffer != NULL ) {
        free(map->buffer);
        map->buffer = NULL;
    }
    free(map);
}

/*
 * skPrefixMap_err_t skPrefixMapRead(FILE *in, skPrefixMap_t **map)
 *
 * Allocates a new prefixmap and assigns a pointer to it into the
 * provided space.  Then reads a prefixmap from the FILE* in.  Before
 * calling, map should not be NULL, *map may have any value, and in
 * should not be NULL.  If skPrefixMapRead returns successfully memory
 * has been allocated, and *map will contain a new valid skPrefixMap_t
 * pointer.  If skPrefixMapRead returns an error, *map will not
 * contain a valid pointer, and any allocated memory has been freed.
 */

skPrefixMap_err_t skPrefixMapRead(skPrefixMap_t **map, FILE *in)
{
    skPrefixMap_header_t header;
    int swapFlag;
    size_t i;

    /* Check arguments for sanity */
    if ( NULL == map ) {
        skAppPrintErr("No place was provided to store new prefixmap.");
        return SKPREFIXMAP_ERR_ARGS;
    }
    if ( NULL == in ) {
        skAppPrintErr("No input file provided from which to read prefixmap.");
        return SKPREFIXMAP_ERR_ARGS;
    }

    /* Read header and check that the input file is valid */
    if ( fread((void*)&header, sizeof(skPrefixMap_header_t), 1, in) != 1 ) {
        skAppPrintErr("Error reading header from input file.");
        return SKPREFIXMAP_ERR_IO;
    }
    if ( CHECKMAGIC(&(header.silkHeader)) ) {
        skAppPrintErr("Invalid header read from input file.");
        return SKPREFIXMAP_ERR_IO;
    }
    if ( FT_PREFIXMAP != header.silkHeader.type ) {
        skAppPrintErr("Input file is not a prefix map.");
        return SKPREFIXMAP_ERR_IO;
    }
    if ( 1 != header.silkHeader.version ) {
        skAppPrintErr("Unrecognized prefix map file version.");
        return SKPREFIXMAP_ERR_IO;
    }
    if ( 0 != header.silkHeader.cLevel ) {
        skAppPrintErr("Unrecognized prefix map compression level.");
        return SKPREFIXMAP_ERR_IO;
    }

#if IS_LITTLE_ENDIAN
    swapFlag = header.silkHeader.isBigEndian;
#else
    swapFlag = !header.silkHeader.isBigEndian;
#endif

    if ( swapFlag ) {
        header.recordCount = BSWAP32(header.recordCount);
    }

    if ( header.recordCount < 1 ) {
        skAppPrintErr("Input file contains invalid prefix map (no data).");
        return SKPREFIXMAP_ERR_IO;
    }

    /* Allocate a prefix map, and a storage buffer */
    *map = (skPrefixMap_t*) malloc(sizeof(skPrefixMap_t));
    if ( NULL == *map ) {
        skAppPrintErr("Failed to allocate memory for prefix map.");
        return SKPREFIXMAP_ERR_MEMORY;
    }

    (*map)->bufferSize = header.recordCount;
    (*map)->buffer = (skPrefixMap_record_t*)malloc(sizeof(skPrefixMap_record_t)
                                                   * header.recordCount);
    if ( NULL == (*map)->buffer ) {
        skAppPrintErr("Failed to allocate memory for prefix map data.");
        free(*map);
        *map = NULL;
        return SKPREFIXMAP_ERR_MEMORY;
    }

    /* Allocation completed successfully, read in the records. */
    if ( fread((void*)(*map)->buffer, sizeof(skPrefixMap_record_t),
               header.recordCount, in) != header.recordCount ) {
        skAppPrintErr("Failed to read all records from input file.");
        free((*map)->buffer);
        free(*map);
        *map = NULL;
        return SKPREFIXMAP_ERR_IO;
    }

    /* Swap the byte order of the data if needed. */
    if ( swapFlag ) {
        for ( i = 0; i < header.recordCount; i++ ) {
            (*map)->buffer[i].left = BSWAP32((*map)->buffer[i].left);
            (*map)->buffer[i].right = BSWAP32((*map)->buffer[i].right);
        }
    }

#if 0
    /* Check the invariants that ensure that the prefixmap is a total
     * mapping (every input maps to a single output.)
     *
     * The invariants are:
     * 1) No node may point to a non-existent child (>= header.recordCount)
     * 2) No node may point to a node earlier than itself (prevents cycles)
     */
    for ( i = 0; i < header.recordCount; i++ ) {
        uint32_t l = (*map)->buffer[i].left;
        uint32_t r = (*map)->buffer[i].right;
        if ( (SKPREFIXMAP_IS_NODE(l) && ((l >= header.recordCount) ||
                                         (l <= i))) ||
             (SKPREFIXMAP_IS_NODE(r) && ((r >= header.recordCount) ||
                                         (r <= i))) ) {
            skAppPrintErr("Prefix map is malformed (contains invalid child).");
            free((*map)->buffer);
            free(*map);
            *map = NULL;
            return SKPREFIXMAP_ERR_IO;
        }
    }

    /* Hrm.  We should also try to look for chains longer than 32
     * steps.  What's a good way to do that?
     */
#endif

    return SKPREFIXMAP_OK;
}

/* 
 * uint32_t skPrefixMapGet(skPrefixMap_t *map, uint32_t key);
 *
 * Returns the mapped value for the given 32 bit key.  A prefixmap will
 * give a result for any input value.  This function is guaranteed to
 * terminate.  The result is always less than 0x8000000.
 */
uint32_t skPrefixMapGet(skPrefixMap_t *map, uint32_t key)
{
    int depth = 31;             /* Start at the leftmost bit */
    uint32_t node = 0;          /* Start at the root node */
    while ( SKPREFIXMAP_IS_NODE(node) ) {
        if ( depth < 0 ) {
            /* This should be caught when the map is loaded. */
            skAppPrintErr("Corrupt prefix map.  No result found in 32 bits.");
            return 0xFFFFFFFF;
        }
        if ( key & (1 << depth) ) {
            /* Go right if the bit is 1. */
            node = map->buffer[node].right;
        } else {
            /* Go left if the bit is 0. */
            node = map->buffer[node].left;
        }
        --depth;
    }
    return SKPREFIXMAP_LEAF_VALUE(node);
}

/* Local Variables: */
/* c-basic-offset: 4 */
/* indent-tabs-mode: nil */
/* End: */
