/*
** Copyright (C) 2004-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
** prefixmap_test.c
**
** John McClary Prevost
** December 3rd, 2004
**
** Small application to test the prefixmap library by taking a
** prefixmap file and a dotted-quad-notation IP address and do a
** lookup on that file to print the result.
*/


#include "silk.h"

RCSIDENT("$SiLK: /src/util/ccfilter_new/prefixmap_test.c,v 1.2 2005/05/20 17:55:22 thomasm Exp $");

#include "utils.h"
#include "prefixmap.h"

/* LOCAL DEFINES AND TYPEDEFS */

/* where to write --help output */
#define USAGE_FH stdout


/* EXPORTED FUNCTIONS */


/* LOCAL FUNCTIONS */

static void appUsage(void);		     /* never returns */
static void appUsageLong(void);		     /* never returns */
static void appOptionsUsage(FILE*);	     /* application options usage */
static void appTeardown(void);
static void appSetup(int argc, char **argv); /* never returns when error */
static int  appOptionsHandler(clientData, int, const char *);


/* LOCAL VARIABLES */



/* OPTIONS SETUP */

typedef enum _appOptionsEnum {
    PREFIXMAP_TEST_MAP_FILE, PREFIXMAP_TEST_ADDRESS
} appOptionsEnum;

static struct option appOptions[] = {
    {"map-file", REQUIRED_ARG, 0, PREFIXMAP_TEST_MAP_FILE},
    {"address", REQUIRED_ARG, 0, PREFIXMAP_TEST_ADDRESS},
    {0,0,0,0}			/* sentinel entry */
};

static const char *appHelp[] = {
    /* add help strings here for the applications options */
    "path name of the map file.",
    "IP address to look up",
    (char *)NULL
};

static struct {
    const char *map_file;       /* filename of map file */
    uint32_t address;           /* IP address to look up */
} prefixmap_test_opt = {
    NULL, 0xFFFFFFFF
};

/*
 * appUsage:
 * 	print short usage information to stderr and exit with code 1
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
void appUsage(void)
{
    fprintf(stderr, "Use `%s --help' for usage\n", skAppName());
    exit(EXIT_FAILURE);
}


/*
 * appUsageLong:
 * 	print complete usage information to USAGE_FH and exit with
 * 	code.  passed to optionsSetup() to print usage when --help
 * 	option given.
 * Arguments: None
 * Returns: None
 * Side Effects: exits application.
 */
static void appUsageLong(void)
{
    fprintf(USAGE_FH, "Usage: %s [app_opts] ... \n", skAppName());
    appOptionsUsage(USAGE_FH);
    exit(EXIT_FAILURE);
}


/*
 * appOptionsUsage:
 * 	print options for this app to the specified file handle.
 * Arguments: None.
 * Returns: None.
 * Side Effects: None.
 * NOTE:
 *	do NOT add non-option usage here (i.e., required and other
 *	optional args)
 */
static void appOptionsUsage(FILE *fh)
{
    int i;

    fprintf(fh, "\napp options\n");
    for (i = 0; appOptions[i].name; i++) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                (appOptions[i].has_arg
                 ? ((appOptions[i].has_arg == 1)
                    ? "Required Arg"
                    : "Optional Arg")
                 : "No Arg"), appHelp[i]);
    }

    return;
}


/*
 * appTeardown:
 *	teardown all used modules and all application stuff.
 * Arguments: None
 * Returns: None
 * Side Effects:
 * 	All modules are torn down. Then application teardown takes place.
 * 	Global variable teardownFlag is set.
 * NOTE: This must be idempotent using static teardownFlag.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    /* local teardown segment */

    optionsTeardown();

    return;
}


/*
 * appSetup
 *	do all the setup for this application include setting up required
 *	modules etc.
 * Arguments:
 *	argc, argv
 * Returns: None.
 * Side Effects:
 *	exits with code 1 if anything does not work.
 */
static void appSetup(int argc, char **argv)
{
    int nextOptIndex;

    skAppRegister(argv[0]);

    /* initialize globals */

    /* check that we have the same number of options entries and help*/
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* Set up the options module */
    if (optionsSetup(appUsageLong)) {
        skAppPrintErr("unable to setup options module");
        exit(EXIT_FAILURE);			/* never returns */
    }

    /* register the options handler */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL))
    {
        skAppPrintErr("unable to register application options");
        exit(EXIT_FAILURE);			/* never returns */
    }

    /* parse the options */
    if ( (nextOptIndex = optionsParse(argc, argv)) < 0) {
        /* options parsing should print error */
        appUsage();		/* never returns */
    }

    if ( (NULL == prefixmap_test_opt.map_file) ||
         (0xFFFFFFFF == prefixmap_test_opt.address) ) {
        if ( NULL == prefixmap_test_opt.map_file ) {
            skAppPrintErr("required argument map-file not provided.");
        }
        if ( 0xFFFFFFFF == prefixmap_test_opt.address ) {
            skAppPrintErr("required argument address not provided.");
        }
        exit(EXIT_FAILURE);
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;  /* OK */
}


/*
 * appOptionsHandler:
 * 	called for each option the app has registered.
 * Arguments:
 *	clientData cData: ignored here.
 *	int index: index into appOptions of the specific option
 *	char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *	0 if OK. 1 else.
 * Side Effects:
 *	Relevant options are set.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    const char *opt_arg)
{

    switch (opt_index) {
    case PREFIXMAP_TEST_MAP_FILE:
        prefixmap_test_opt.map_file = opt_arg;
        break;

    case PREFIXMAP_TEST_ADDRESS:
        prefixmap_test_opt.address = dot2num(opt_arg);
        if ( 0xFFFFFFFF == prefixmap_test_opt.address ) {
            skAppPrintErr("Could not parse address.");
            exit(EXIT_FAILURE);
        }
        break;

      default:
        skAppPrintErr("Option %d not handled in switch() at %s:%d",
                      opt_index, __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }

    return 0;  /* OK */
}


int main(int argc, char **argv)
{
    FILE* inputFile;
    skPrefixMap_t* prefixMap;
    skPrefixMap_err_t map_error = SKPREFIXMAP_OK;

    appSetup(argc, argv);			/* never returns on error */

    /* Okay.  Now we should open the prefixmap file, read it in, and */
    /* then look up our address! */

    inputFile = fopen(prefixmap_test_opt.map_file, "r");
    if ( NULL == inputFile ) {
        skAppPrintErr("Failed to open file `%s': %s",
                      prefixmap_test_opt.map_file,
                      strerror(errno));
        exit(EXIT_FAILURE);
    }

    map_error = skPrefixMapRead(&prefixMap, inputFile);
    if ( SKPREFIXMAP_OK != map_error ) {
        if ( SKPREFIXMAP_ERR_ARGS == map_error ) {
            skAppPrintErr("Failed to read map file: Invalid arguments");
        } else if ( SKPREFIXMAP_ERR_MEMORY == map_error ) {
            skAppPrintErr("Failed to read map file: Out of memory");
        } else if ( SKPREFIXMAP_ERR_IO == map_error ) {
            skAppPrintErr("Failed to read map file: I/O error");
        }
        exit(EXIT_FAILURE);
    }

    printf("%d\n", skPrefixMapGet(prefixMap, prefixmap_test_opt.address));
    skPrefixMapDelete(prefixMap);

    /* done */
    appTeardown();

    return 0;
}

/* Local Variables: */
/* c-basic-offset: 4 */
/* indent-tabs-mode: nil */
/* End: */
