/*
** Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/


/*
**  log.c
**
**  Library to support file-based logging.  Cobbled together based on
**  the report.c code written by M. Duggan.
**
**  Suresh Konda
**  4/23/2003
*/


#include "silk.h"

RCSIDENT("$SiLK: /src/util/log/log.c,v 1.24 2005/09/26 15:25:06 dugganm Exp $");

#include "utils.h"
#include "log.h"
#include <sys/types.h>
#include <sys/wait.h>


#ifndef MAILX
#  ifdef SK_MAILER
#    define MAILX SK_MAILER
#  else
#    ifdef LINUX
#      define MAILX "mail"
#    else
#      define MAILX "mailx"
#    endif
#  endif
#endif


static int rolloverHour;
static char *logFilePrefix;
static char *logDir;
static FILE *logFS = NULL;
static char *curLogFilename;
static time_t rolloverTime;
static int mailAvailableFlag;
/* rotate mutex */
static pthread_mutex_t rotatemutex = PTHREAD_MUTEX_INITIALIZER;
/* logging mutex */
static pthread_mutex_t logmutex = PTHREAD_MUTEX_INITIALIZER;

static int openLogFS(void);
static void rotateLogs(void);
static void setRolloverTime(void);
static int compressLogs(void);

int logSetup(
    const char *in_logFilePrefix,
    const char *in_logDir,
    int         rHour,
    int         maFlag)
{
    /* Check input */
    if (in_logFilePrefix == NULL) {
        return 1;
    }
    if (rHour < 0 || in_logDir == NULL) {
        /* This means that logFilePrefix is actually a full path for a
           non-rotating log file. */
        if (in_logDir) {
            curLogFilename = (char *)malloc(2 + strlen(in_logFilePrefix)+
                                            strlen(in_logDir));
            strcpy(curLogFilename, in_logDir);
            strcat(curLogFilename, "/");
            strcat(curLogFilename, in_logFilePrefix);
        } else {
            curLogFilename = strdup(in_logFilePrefix);
        }
        logDir = NULL;
    } else {
        /* Normal rotating logfile. */
        if (!dirExists((char *)in_logDir)) {
            fprintf(stderr, "Dir %s does not exist\n", in_logDir);
            return 1;
        }
        if (rHour < 0 || rHour > 23) {
            return 1;
        }

        logDir = strdup(in_logDir);
        if (logDir == NULL) {
            return 1;
        }
        curLogFilename = (char *)malloc(32 + strlen(in_logFilePrefix)+
                                        strlen(in_logDir));
    }
    logFilePrefix = strdup(in_logFilePrefix);

    if (logFilePrefix == NULL || curLogFilename == NULL) {
        return 1;
    }

    logFS = (FILE *)NULL;

    if (openLogFS()) {
        free(logFilePrefix);
        if (logDir) {
            free(logDir);
        }
        free(curLogFilename);
        logFilePrefix = (char*)NULL;
        return 1;
    }
    if (logDir) {
        /* Rollover only used when using a rotating logFile. */
        rolloverHour = rHour;
        setRolloverTime();
    }
    mailAvailableFlag = maFlag;
    return 0;
}

void logTeardown(void)
{
    if(logFilePrefix) {
        free(logFilePrefix);
        if (logDir) {
            free(logDir);
        }
        free(curLogFilename);
        logFilePrefix = (char*)NULL;
    }

    if(logFS != stderr) {
        fclose(logFS);
    }
    logFS = (FILE *)NULL;
    return;
}


static int openLogFS(void)
{
    char date[9];
    struct tm *tmNow;
    time_t t;

    if (logDir) {
        /* Only calculate name if a ratating logFile. */
        t = time((time_t*)NULL);
        tmNow = localtime(&t);
        strftime(date, sizeof(date), "%Y%m%d", tmNow);
        sprintf(curLogFilename, "%s/%s-%s.log", logDir, logFilePrefix, date);
    } else if (logFS) {
        /* If non-rotating logfile, only open once. */
        return 0;
    }
    if ((logFS = fopen(curLogFilename, "a")) == (FILE *)NULL) {
        fprintf(stderr, "Cannot open log file %s\n", curLogFilename);
        logFS = stderr;
        return 1;
    }

    return 0;
}

static void rotateLogs(void)
{
    if (logFS == stderr) {
        return;
    }
    pthread_mutex_lock(&rotatemutex);
    if (logFS) {
        fclose(logFS);
    }
    if (compressLogs() < 0)
    {
        fprintf (stderr, "error forking compressLog\n");
    }
    logFS = (FILE*)NULL;
    if (openLogFS()) {
        pthread_mutex_unlock (&rotatemutex);
        return;
    }
    setRolloverTime();
    pthread_mutex_unlock (&rotatemutex);
    return;
}

static void setRolloverTime(void)
{
    time_t tNow;
    struct tm *tmNow, tmCopy;

    tNow = time((time_t*)NULL);
    tmNow = localtime(&tNow);
    memcpy(&tmCopy, tmNow, sizeof(tmCopy));
    tmCopy.tm_sec = 59;
    tmCopy.tm_min = 59;
    tmCopy.tm_hour = 23;
    rolloverTime = mktime(&tmCopy) + (rolloverHour * 60 * 60) + 1;
    return;
}

int vlogMsg(const char *fmt, va_list args)
{
    time_t ct;
    struct tm *t;
    char stime[10];
    int size = 0;

    pthread_cleanup_push((void (*)(void *))pthread_mutex_unlock,
                         (void *)&logmutex);
    pthread_mutex_lock(&logmutex);

    assert(logFS);

    ct = time(NULL);
    t = localtime(&ct);
    strftime(stime, sizeof(stime), "%H:%M:%S", t);
    size += fprintf(logFS, "%s: ", stime);
    size += vfprintf(logFS, fmt, args);
    fprintf(logFS, "\n");
    fflush(logFS);
    if (logDir && rolloverTime < ct) {
        /* If rotating logFile, log rollover, and then do so. */
        ct = time(NULL);
        t = localtime(&ct);
        strftime(stime, sizeof(stime), "%H:%M:%S", t);
        fprintf(logFS, "%s: Log rollover\n", stime);
        fflush(logFS);
        rotateLogs();
    }
    pthread_cleanup_pop(1);
    return size;
}

/* Subject may not include double quotes */
int vreport(
    const char *user,
    const char *subject,
    const char *message,
    va_list     args)
{
    char buf[4096];
    size_t rv;
    FILE *f;
    int ev;
    int nleft;
    char *pos;
    char command[4096];

    if (logFS) {
        vlogMsg(message, args);
    }

    if (mailAvailableFlag) {
        vsnprintf(buf, sizeof(buf), message, args);
        buf[sizeof(buf) - 1] = 0;
        pos = buf;
        nleft = strlen(buf);
        snprintf(command, sizeof(command), "%s -s \"%s\" %s", MAILX,
                 subject, user);
        if ((f = popen(command, "w")) == NULL) {
            return -1;
        }
        while (nleft && (rv = fwrite(buf, 1, strlen(buf), f))) {
            pos += rv;
            nleft -= rv;
        }
        if (nleft) {
            return -1;
        }
        fwrite("\n", 1, 1, f);
        fflush(f);
        ev = pclose(f);
        return (ev);
    } else {
        return 0;			/* OK */
    }
}


#ifndef logMsg
/*
 *  Sometimes it's useful to
 *
 *    #define skAppPrintErr printf
 *
 *  to see where we've messed up our formatting.  When doing that, we
 *  obviously do not want to compile this function.
 */
int logMsg(const char *fmt, ...)
{
    int retval;
    va_list args;

    va_start(args, fmt);
    retval = vlogMsg(fmt, args);
    va_end(args);
    return retval;
}
#endif /* !defined(logMsg) */


int report(
    const char *user,
    const char *subject,
    const char *message,
    ...)
{
    int retval;
    va_list args;
    va_start(args, message);
    retval = vreport(user, subject, message, args);
    va_end(args);
    return retval;
}


static int compressLogs()
{
    pid_t pid;
    char *file = strdup(curLogFilename);
    
    /* Fork */
    if ((pid = fork()) == -1) {
        fprintf(stderr, "Couldn't fork for compression.\n");
        free(file);
        return -1;
    } 
    /* Original process (parent) */
    if (pid != 0) {
        /* Wait for Child 1 to exit. */
        waitpid(pid, NULL, 0);
        free(file);
        return 0;
    } 
    /* Child 1 forks */
    if ((pid = fork()) == -1) {
        fprintf(stderr, "Couldn't fork for compression.\n");
        _exit(EXIT_FAILURE);
    }
    /* Child 1 */
    if (pid != 0) {
        /* Child 1 immediately exits, so parent can stop waiting. */
        _exit(EXIT_SUCCESS);
    }
    /* Child 2 executes the gzip */
    if (execlp("gzip", "gzip", "-f", file, NULL) <0 ) { 
        fprintf (stderr, "compressing error\n"); 
        _exit(EXIT_FAILURE);
    } 
    /* Should never get here. */
    assert(0);
    return -1;
}
