/*
**  Copyright (C) 2001-2005 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  a collection of utility routines.
**
**  Suresh L Konda
*/


#include "silk.h"

RCSIDENT("$SiLK: /src/util/utils/silkfilesys.c,v 1.20 2005/10/21 15:26:37 dugganm Exp $");

#include "utils.h"
#include <sys/mman.h>

#define NOZLIB



char *baseName_r(char *dest, const char *src, size_t dest_size)
{
    const char *startp;
    const char *endp;
    size_t src_len;

    /* check input: need space for {'.', '\0'} minimally */
    if (!dest || dest_size < 2) {
        return NULL;
    }

    /* degenerate cases */
    if (!src || (0 == (src_len = strlen(src)))) {
        return strcpy(dest, ".");
    }

    startp = strrchr(src, '/');
    if (!startp) {
        /* no slash; return what we were given */
        startp = src;
        endp = src + src_len;
    } else if ('\0' != *(startp+1)) {
        /* typical case: "/bin/cat" */
        ++startp;
        endp = src + src_len;
    } else {
        /* we could have "/", "///", "usr/", or "/usr/lib/" */
        while (startp > src && *startp == '/') {
            /* remove trailing '/' */
            --startp;
        }
        endp = startp + 1;
        /* go backward until find '/'; startp is char after the '/' */
        while (startp > src) {
            --startp;
            if (*startp == '/') {
                ++startp;
                break;
            }
        }
    }

    /* need to grab everything between startp and endp */
    src_len = endp - startp;
    if (src_len > dest_size-1) {
        return NULL;
    }
    strncpy(dest, startp, src_len);
    dest[src_len] = '\0';

    return dest;
}


char *dirName_r(char *dest, const char *src, size_t dest_size)
{
    const char *endp;
    size_t src_len;

    /* check input: need space for {'.', '\0'} minimally */
    if (!dest || dest_size < 2) {
        return NULL;
    }

    /* degenerate cases */
    if (!src || !(endp = strrchr(src, '/'))) {
        return strcpy(dest, ".");
    }

    if ('\0' == *(endp+1)) {
        /* we could have "/", "///", "usr/", or "/usr/lib/" */
        while (endp > src && *endp == '/') {
            /* remove trailing '/' */
            --endp;
        }
        while (endp > src && *endp != '/') {
            /* skip basename */
            --endp;
        }
        if (*endp != '/') {
            /* we're at start of string */
            return strcpy(dest, ".");
        }
    }

    /* handle duplicate '/' chars */
    while (endp > src && *endp == '/') {
        --endp;
    }

    src_len = endp - src + 1;
    if (src_len > dest_size-1) {
        return NULL;
    }

    strncpy(dest, src, src_len);
    dest[src_len] = '\0';

    return dest;
}


char *baseName(const char *src)
{
    static char dest[PATH_MAX]; /* return pointer */

    return baseName_r(dest, src, sizeof(dest));
}


char *dirName(const char *src)
{
    static char dest[PATH_MAX]; /* return pointer */

    return dirName_r(dest, src, sizeof(dest));
}


int isFIFO(const char *name)
{
    struct stat stBuf;
    if (stat(name, &stBuf) == -1) {
        return 0;
    }
    return (S_ISFIFO(stBuf.st_mode));
}


int dirExists(const char *dName)
{
    struct stat stBuf;
    if (stat(dName, &stBuf) == -1) {
        return 0;                   /* does not exist */
    }
    /* return a 1 only if this is a directory */
    return S_ISDIR(stBuf.st_mode);
}


int fileExists(const char *fName)
{
    struct stat stBuf;
    if (stat(fName, &stBuf) == -1) {
        return 0;                   /* does not exist */
    }
    /* return a 1 only if this is a regular file */
    return (S_ISREG(stBuf.st_mode) || S_ISFIFO(stBuf.st_mode));
}


off_t fileSize(const char *fName)
{
    struct stat stBuf;
    if (stat(fName, &stBuf) == -1) {
        return 0;                   /* does not exist */
    }
    /* it exists. return the size */
    return stBuf.st_size;
}


/*
 *  int ret = silkFileLocks(int fd, short type, int cmd);
 *
 *  Perform a locking operation on the opened file represented by the
 *  file descriptor fd.  type is the type of lock, it should be one of
 *  F_RDLCK for a read lock, F_WRLCK for a write lock, or F_UNLCK to
 *  unlock a previously locked file.  cmd should be one of F_SETLKW to
 *  wait indefinitely for a lock, or F_SETLK to return immediately.
 *  Return 0 if successful, 1 otherwise.
 *
 */
int silkFileLocks(int fd, short type, int cmd)
{
    struct flock lock;
    lock.l_type = type;
    lock.l_start = 0;             /* at SOF */
    lock.l_whence = SEEK_SET;     /* SOF */
    lock.l_len = 0;               /* EOF */

    if (fcntl(fd, cmd, &lock) != -1) {
        /* success */
        return 0;
    }

#if 0
    /* print warning */
    if (type == F_UNLCK) {
        /* cannot unlock */
        skAppPrintErr("Cannot release lock");
    } else if (cmd == F_SETLKW) {
        /* cannot get lock and willing to wait */
        skAppPrintErr("Cannot get lock");
    }
#endif

    return 1;
}


/*
 *  findFile(fName);
 *
 *  Find the given file in one of several places:
 *
 *  -- See if the environment variable named by the cpp macro
 *     ENV_SILK_PATH (normally SILK_PATH) is defined.  If so, look in
 *     $SILK_PATH/share/fName
 *  -- For historical reasons, look in $SILK_PATH/fName
 *  -- Take the full path to the application (/yadda/yadda/bin/app),
 *     lop off the app's parent directory (/yadda/yadda), and check
 *     the "/share" subdir (/yadda/yadda/share/fName
 *
 *  Inputs:
 *      filename to find
 *
 *  Outputs:
 *      Return char * if found NULL else.  When non-null is returned,
 *      the caller should free() the char* when finished.
 */
char *findFile(const char *fName)
{
    const char *app_name = skAppName();
    char buf[PATH_MAX];
    const size_t bufsize = sizeof(buf)-1;
    char *cp1;
    size_t len;

    /* Clear 'buf' */
    memset(buf, 0, sizeof(buf));

    /* check inputs */
    if (!fName) {
        return NULL;
    }

    /* Check in $SILK_PATH/share and $SILK_PATH */
    cp1 = getenv(ENV_SILK_PATH);
    if (cp1) {
        snprintf(buf, bufsize, "%s/share/%s", cp1, fName);
        if (fileExists(buf)) {
            return strdup(buf);
        }
        snprintf(buf, bufsize, "%s/%s", cp1, fName);
        if (fileExists(buf)) {
            return strdup(buf);
        }

        /* didn't find it.  clear 'buf' */
        memset(buf, 0, bufsize);
    }

    /* Look in binarypath/../share.  First, get the parent directory of
     * the executable. */
    if (app_name == (char *)NULL) {
        goto ERROR;
    }
    if (NULL == skAppDirParentDir(buf, bufsize)) {
        buf[0] = '\0';
        goto ERROR;
    }
    len = strlen(buf);

    /* Now append "/share/<file>" to it */
    snprintf((buf+len), (bufsize - len - 1), "/share/%s", fName);
    buf[bufsize] = '\0';
    if (fileExists(buf)) {
        return strdup(buf);
    }

  ERROR:
    skAppPrintErr(("Cannot find file '%s' in $" ENV_SILK_PATH "/share,"
                   " in $" ENV_SILK_PATH ","), fName);
    if (!app_name) {
        skAppPrintErr(" and application name not registered.");
    }
    else if ('\0' == buf[0]) {
        skAppPrintErr(" and cannot obtain full path to %s.", app_name);
    }
    else {
        skAppPrintErr(" nor in %s.", buf);
    }
    return ((char*)NULL);
}


char *skutilsFindPluginPath(const char *dlPath, char *path, size_t path_size)
{
#ifndef SILK_SUBDIR_PLUGINS
    return NULL;
#else
    const char *subdir[] = SILK_SUBDIR_PLUGINS;
    char *silkPath;
    size_t len;
    int i;
    int8_t checkSilkPath = 1;
    int8_t checkExec = 1;

    /* put path into known state */
    path[0] = '\0';

    if (strchr(dlPath, '/')) {
        return NULL;
    }

    /* if dlPath does not contain a slash, first look for the plugin in
     * the SILK_SUBDIR_PLUGINS subdirectory of the environment variable
     * named by ENV_SILK_PATH.  If the plugin does not exist there, pass
     * the dlPath as given to dlopen() which will use LD_LIBRARY_PATH or
     * equivalent.
     */
    while (checkSilkPath || checkExec) {
        if (checkSilkPath) {
            checkSilkPath = 0;
            if (NULL == (silkPath = getenv(ENV_SILK_PATH))) {
                continue;
            }
            strncpy(path, silkPath, path_size);
        } else if (checkExec) {
            checkExec = 0;
            if (NULL == skAppDirParentDir(path, path_size)) {
                /* cannot find executeable path */
                continue;
            }
        }
        path[path_size-1] = '\0';
        len = strlen(path);
        for (i = 0; subdir[i]; ++i) {
            snprintf(path + len, path_size - len - 1, "/%s/%s",
                     subdir[i], dlPath);
            path[path_size-1] = '\0';
            /* skAppPrintErr("DEBUG: %s", path); */
            if (fileExists(path)) {
                return path;
            }
        }
    }

    /* file does not exist.  Fall back to LD_LIBRARY_PATH */
    path[0] = '\0';
    return NULL;
#endif /* SILK_SUBDIR_PLUGINS */
}


/* open file 'FName' for read (mode==0) or write (mode==1).  See header. */
int openFile(const char *FName, int mode, FILE **fp, int *isPipe)
{
    char cmd[1024];
    const char *cp;

    /* after this while() loop, 'cp' will be NULL if 'FName' is NOT
     * compressed, or non-NULL if 'FName' is compressed. */
    cp = FName;
    while (NULL != (cp = strstr(cp, ".gz"))) {
        if (*(cp + 3) == '\0') {
            /* file ends with ".gz" */
            break;
        } else if (*(cp + 3) == '.') {
            /* Treat a file that contains ".gz." as a compressed file.
             * We do this to handle compressed files have had
             * mkstemp() extensions added to them.  This is hackish,
             * but it is simple and covers that common case with few
             * false positives.  Other solutions include looking for
             * the gzip magic number (31 139 (see RFC1952)) in the
             * first two bytes, but this only works when reading files
             * (not when writing), and it wouldn't work with named
             * pipes---no way to put back the two bytes for
             * non-gzipped files---although anyone using a named pipe
             * that contains ".gz" in its name and expects the reader
             * of the named pipe to handle compressed data is probably
             * asking for trouble.  Another solution would be to use
             * gzopen() and gzread()---which again won't work for
             * writing---and either accept their overhead when working
             * with uncompressed files, or have librw do special
             * things with compressed files.  There is really no good
             * solution here.  Hopefully soon we will be using LZO
             * compression in the body of the data files, but still
             * have an uncompresed SiLK header. */
            break;
        } else {
            cp += 3;
        }
    }

    if (NULL == cp) {
        /* Regular file or named pipe */
        *isPipe = 0;
        *fp = fopen(FName, mode ? "w" : "r");
    } else if (mode == 0 && fileExists(FName) == 0) {
        /* Attempting to read from non-existent gzip */
        *fp = NULL;
    } else {
        /* Either writing to gzip or reading from existing gzip */
        if ((sizeof(cmd) - 16u) < sizeof(FName)) {
            return 1;
        }
        *isPipe = 1;
        snprintf(cmd, sizeof(cmd), "gzip %s '%s'",
                 (mode ? ">" : "-d -c"), FName);
        *fp = popen(cmd, mode ? "w" : "r");
    }

    if (*fp == (FILE *)NULL) {
        if (mode == 0 && fileExists(FName) == 0) {
            skAppPrintErr("Cannot open non-existant file '%s'", FName);
        } else {
            skAppPrintErr("Unable to open file '%s' for %s",
                          FName, mode ? "writing" : "reading");
        }
        return 1;                   /* error */
    }

    return 0;
}


/*
 * mkDirPath:
 *      make full directory path including parent(s) if required
 * Input: char * path
 * Output: 0 on success, 1 on failure
 */
int mkDirPath(const char *directory)
{
    int rv = 0; /* return value */
    size_t dir_len;
    char *cp;
    char *dir_buf = NULL;
    char **slash_list = NULL;
    int slash_count = 0;
    const mode_t dirMode = /* 0775 */
        S_IRWXU | S_IRGRP | S_IWGRP |  S_IXGRP | S_IROTH | S_IXOTH;

    assert(directory);

    /* Common case is a one-off, so try that first.  AIX does not always
     * set errno to EEXIST for an existing directory, so call
     * dirExists() as a back-up test. */
    if (0==mkdir(directory, dirMode) || errno==EEXIST || dirExists(directory)){
        return rv;
    }

    dir_len = strlen(directory);
    if (0 == dir_len) {
        rv = 1;
        goto END;
    }

    /* make a copy of the directory name that we can modify, and malloc
     * an array of char*'s that will point to the slashes ('/') in that
     * dir_buf. */
    dir_buf = strdup(directory);
    slash_list = malloc(dir_len * sizeof(char*));
    if (!dir_buf || !slash_list) {
        rv = 1;
        goto END;
    }

    /* point cp at the end of the buffer, then work backward until we
     * find a slash.  Convert the slash to a '\0' and see if the parent
     * dir exists.  If it does not, keep shorting the directory
     * path--stashing the locations of the '/' in slash_list[]--until we
     * find an existing parent directory.  If the parent dir does exist,
     * change the '\0' back to a '/' and start making child
     * directories. */
    cp = &(dir_buf[dir_len]);

    while (1) {
        /* search for dir-sep */
        while (cp > dir_buf && *cp != '/') {
            --cp;
        }
        if (cp == dir_buf) {
            /* can't search past start of string */
            break;
        }
        /* see if parent dir exists... */
        *cp = '\0';
        if (dirExists(dir_buf)) {
            /* ...it does, we can start making child directories */
            *cp = '/';
            break;
        }
        /* ...else it does not. Store this location and continue up the path */
        slash_list[slash_count] = cp;
        ++slash_count;
    }

    /* dir_buf should contain a directory we can create */
    while (1) {
        /* make the child directory */
        if (0 != mkdir(dir_buf, dirMode)) {
            /* perhaps another thread or process created the directory? */
            int err = errno;
            if (err != EEXIST && !dirExists(dir_buf)) {
                skAppPrintErr("Cannot mkdir '%s': %s",
                              dir_buf, strerror(err));
                rv = 1;
                goto END;
            }
        }
        if (slash_count == 0) {
            /* we should have created the entire path */
            assert(0 == strcmp(dir_buf, directory));
            break;
        }
        /* convert this '\0' back to a '/' and make next child dir */
        --slash_count;
        *(slash_list[slash_count]) = '/';
    }

  END:
    /* cleanup buffers */
    if (dir_buf) {
        free(dir_buf);
    }
    if (slash_list) {
        free(slash_list);
    }
    return rv;
}

/*
 * copyFile:
 *      Copies the file "source" to "dest".  "Dest" may be a file or a
 *      directory.
 * Input: char * source
 *        char * dest
 * Output: 0 on success, errno on failure.
 */
int copyFile(const char *srcPath, const char *destPath)
{
    int fdin = -1, fdout = -1;
    void *src = NULL, *dst = NULL;
    const char *dest = NULL;
    char destBuf[PATH_MAX];
    char base[PATH_MAX];
    struct stat st;
    int saveerrno;
    int rv;
    ssize_t wrv;

    /* Open source */
    fdin = open(srcPath, O_RDONLY);
    if (fdin == -1) {
        goto copy_error;
    }

    /* Get source info */
    rv = fstat(fdin, &st);
    if (rv == -1) {
        goto copy_error;
    }

    /* Handle dest being a directory */
    if (dirExists(destPath)) {
        baseName_r(base, srcPath, sizeof(base));
        rv = snprintf(destBuf, sizeof(destBuf), "%s/%s", destPath, base);
        if (rv == -1) {
            close(fdin);
            return EIO;
        }
        if ((unsigned int)rv > (sizeof(destBuf) - 1)) {
            close(fdin);
            return ENAMETOOLONG;
        }
        dest = destBuf;
    } else {
        dest = destPath;
    }

    /* Open dest */
    fdout = open(dest, O_RDWR | O_CREAT | O_TRUNC, st.st_mode);
    if (fdout == -1) {
        goto copy_error;
    }

    /* Resize dest to source size (For mmap.  See APUE [Stevens].) */
    rv = lseek(fdout, st.st_size - 1, SEEK_SET); 
    if (rv == -1) {
        goto copy_error;
    }
    wrv = write(fdout, "", 1);
    if (wrv != 1) {
        goto copy_error;
    }

    /* Map source */
    src = mmap(0, st.st_size, PROT_READ, MAP_SHARED, fdin, 0);
    if (src == MAP_FAILED) {
        goto copy_error;
    }
    rv = close(fdin);
    if (rv == -1) {
        goto copy_error;
    }
    fdin = -1;

    /* Map dest */
    dst = mmap(0, st.st_size, PROT_READ | PROT_WRITE, MAP_SHARED, fdout, 0);
    if (dst == MAP_FAILED) {
        goto copy_error;
    }
    rv = close(fdout);
    if (rv == -1) {
        goto copy_error;
    }
    fdout = -1;

    /* Copy src to dest */
    memcpy(dst, src, st.st_size);

    /* Close src and dest */
    munmap(src, st.st_size);
    munmap(dst, st.st_size);

    return 0;

  copy_error:
    saveerrno = errno;
    
    if (fdin != -1) {
        close (fdin);
    }
    if (fdout != -1) {
        close (fdout);
        unlink(dest);
    }
    if (src) {
        munmap(src, st.st_size);
    }
    if (dst) {
        munmap(dst, st.st_size);
        unlink(dest);
    }

    return saveerrno;
}


/*
 * moveFile:
 *      Moves the file "source" to "dest".  "Dest" may be a file or a
 *      directory.
 * Input: char * source
 *        char * dest
 * Output: 0 on success, errno on failure.
 */
int moveFile(const char *srcPath, const char *destPath)
{
    const char *dest;
    char destBuf[PATH_MAX];
    char base[PATH_MAX];
    int rv;
    int saveerrno;

    /* Handle dest being a directory */
    if (dirExists(destPath)) {
        baseName_r(base, srcPath, sizeof(base));
        rv = snprintf(destBuf, sizeof(destBuf), "%s/%s", destPath, base);
        if (rv == -1) {
            return EIO;
        }
        if ((unsigned int)rv > (sizeof(destBuf) - 1)) {
            return ENAMETOOLONG;
        }
        dest = destBuf;
    } else {
        dest = destPath;
    }
   
    /* Attempt a simple move */
    rv = rename(srcPath, dest);
    if (rv == -1) {
        if (errno != EXDEV) {
            return errno;
        }
        
        /* Across filesystems, so copy and delete. */
        rv = copyFile(srcPath, dest);
        if (rv != 0) {
            return rv;
        }
        rv = unlink(srcPath);
        if (rv == -1) {
            saveerrno = errno;
            unlink(dest);
            return saveerrno;
        }
    }

    return 0;
}


/* paginate the output.  see utils.h */
int skOpenPagerWhenStdoutTty(
    FILE  **output_stream,
    char  **pager)
{
    FILE *out;
    char *pg;
    pid_t pid;
    int wait_status;

    /* verify input; deference the input variables */
    assert(output_stream);
    assert(pager);
    out = *output_stream;
    pg = *pager;

    /* don't page if output is not "stdout" */
    if (NULL == out) {
        out = stdout;
    } else if (out != stdout) {
        /* no change */
        return 0;
    }

    /* don't page a non-terminal */
    if ( !FILEIsATty(out)) {
        if (pg) {
            /* pager explicitly given but ignored */
            skAppPrintErr("pager switch ignored");
        }
        return 0;
    }

    /* get pager from environment if not passed in */
    if (NULL == pg) {
        pg = getenv("SILK_PAGER");
        if (NULL == pg) {
            pg = getenv("PAGER");
        }
    }

    /* a NULL or an empty string pager means do nothing */
    if ((NULL == pg) || ('\0' == pg[0])) {
        return 0;
    }

#if 1
    /* invoke the pager */
    out = popen(pg, "w");
    if (NULL == out) {
        skAppPrintErr("Unable to invoke pager '%s'", pg);
        return -1;
    }

    /* see if pager started.  There is a race condition here, and this
     * assumes we have only one child, which should be true. */
    pid = wait4(0, &wait_status, WNOHANG, NULL);
    if (pid) {
        skAppPrintErr("Unable to invoke pager '%s'", pg);
        return -1;
    }
#else
    {
    int pipe_des[2];

    /* create pipe and fork */
    if (pipe(pipe_des) == -1) {
        skAppPrintErr("Cannot create pipe: %s", strerror(errno));
        return -1;
    }
    pid = fork();
    if (pid < 0) {
        skAppPrintErr("Cannot fork: %s", strerror(errno));
        return -1;
    }

    if (pid == 0) {
        /* CHILD */
            
        /* close output side of pipe; set input to stdin */
        close(pipe_des[1]);
        if (pipe_des[0] != STDIN_FILENO) {
            dup2(pipe_des[0], STDIN_FILENO);
            close(pipe_des[0]);
        }

        /* invoke pager */
        execlp(pg, NULL);
        skAppPrintErr("Unable to invoke pager '%s': %s",
                      pg, strerror(errno));
        _exit(EXIT_FAILURE);
    }

    /* PARENT */

    /* close input side of pipe */
    close(pipe_des[0]);

    /* try to open the write side of the pipe */
    out = fdopen(pipe_des[1], "w");
    if (NULL == out) {
        skAppPrintErr("Cannot fdopen: %s", strerror(errno));
        return -1;
    }

    /* it'd be nice to have a slight pause here to give child time to
     * die if command cannot be exec'ed, but it's not worth the
     * trouble to use select(), and sleep(1) is too long. */

    /* see if child died unexpectedly */
    if (waitpid(pid, &wait_status, WNOHANG)) {
        skAppPrintErr("Unable to invoke pager '%s'", pg);
        return -1;
    }
    }
#endif /* 1: whether to use popen() */

    /* looks good. set the output variables and return */
    *pager = pg;
    *output_stream = out;
    return 1;
}


/* Close the pager */
void skClosePager(
    FILE       *pager_stream,
    const char *pager)
{
    if (pager_stream && (pager_stream != stdout)) {
        if (-1 == pclose(pager_stream)) {
            skAppPrintErr("Error closing pager '%s'", pager);
        }
    }
}
